//--- ECROven.c --- 12/16/05 ---(01/19/05)---
#define DeviceVers     2

#define DevName        "ECR Oven"
#define DeviceType     0x2d

#define ECROven

//-----------------------------------------------------------------------------
//
//  Copyright (c) 2005 National Superconducting Cyclotron Laboratory,
//            Michigan State University, and Mark Davis
//
//-----------------------------------------------------------------------------
//  Program to control serveral power supplies that drive an induction heating
//  oven for the Electron Cyclotron Resonance (ECR) facility.
//
//  For notes on program structure, see the file:
//    ..\ZWORLD\NSCL_ZWORLD_LIBS\BASE_DLP_DLM_C.LIB
//-----------------------------------------------------------------------------
//  !!!!!! DON'T FORGET TO CHANGE DeviceVers ABOVE !!!!!!
//-----------------------------------------------------------------------------
//
//  2005-12-16 - Mark Davis -
//               - increased "max" values for biasSetV and sptrSetV to -10000
//
//  2005-09-21 - Mark Davis -
//               - added offsets for each scaled value
//               - changed meaning of gain/offset values so all computations
//                 are as follows (for both reads and sets):
//                   Actual = (ADC|DAC) * gain + offset
//                   [so, for sets:  DAC = (Actual - offset) / gain ]
//
//  2005-09-09 - Mark Davis -
//               - changed to produce (nominally) pulses for the external
//                 watchdog timer every 100ms (was every 1000ms)
//
//  2005-01-19 - Mark Davis -
//               - first version of new ECR Oven and Bias controller code
//
//-----------------------------------------------------------------------------
//
//  Operational and Programming NOTES:
//
//  - The Host expects the most significant bits at lower addresses while the
//    ZWorld keeps the most significant bits at the higher address
//
//  - This code assumes that, when recieving values via ModBus TCP protocol,
//    any values that span multiple registers (such as floating point values)
//    will be recieved via a Write Multiple Registers command rather than
//    multiple Write Single Register commands (the latter can cause anomolous
//    behavior).
//
//-----------------------------------------------------------------------------

#class auto
#memmap xmem
#nodebug

#ifdef  COMPILE_PRIMARY_PROG
  #error "Cannot compile as primary program."
  #error "Remove COMPILE_PRIMARY_PROG macro in 'Options | Compiler | #defines' options"
  #error "Must Recompile BIOS also. Hit Ctrl-Y then recompile"
#endif

#ifdef  DevelopMode
  #ifdef  COMPILE_SECONDARY_PROG
  #error "DevelopMode set - must NOT compile as secondary program."
  #error "Remove COMPILE_SECONDAY_PROG macro in 'Options | Compiler | #defines' options"
  #error "Must Recompile BIOS also. Hit Ctrl-Y then recompile"
  #endif
#else
  #ifndef  COMPILE_SECONDARY_PROG
  #error "Must compile as secondary program or DevelopMode."
  #error "Put COMPILE_SECONDARY_PROG macro in 'Options Compiler | #defines' options"
  #error "Must Recompile BIOS also. Hit Ctrl-Y then recompile"
  #endif
#endif


#define  NumADC    6  // # of A/D inputs used
#define  NumDAC    4  // # of D/A outputs used

#define  NumChans  1  // # logical devices

//#define DEBUG  true


//#define  UseEncoder  // uses optical encoder knob?

//#define NSCLSerial     // support for NSCL RS-485 serial commands

#define SERIAL_A_AVAIL
#if (_BOARD_TYPE_ != RCM3200)
  #define SERIAL_B_AVAIL
#endif
#define SERIAL_C_AVAIL
#define SERIAL_D_AVAIL
//#define SERIAL_E_AVAIL
//#define SERIAL_F_AVAIL


#define  ChanMult     100  // multipler for chan number or'd to reg & coil #'s

#define  RELAY_PULSE_LEN  500  // ms to energize relays for momentary contacts


//=============================================================================
//                 Definitions the same for all controllers
//=============================================================================


//=============================================================================
//          Common stuff but with Controller-specific variations
//=============================================================================

//----- bit usage masks for digitalIn -----
#define  RF_PS_HotShot      0x020000 // 17 0 ==> HotShot (2kW) Power Supply
#define  RF_PS_NovaStar     0x040000 // 18 0 ==> NovaStar (3kW) Power Supply
#define  RF_PS_NOT_FAULT    0x080000 // 19 K7A - RF Power Supply Fault
#define  RF_PS_NOT_HEAT_ON  0x100000 // 20 K6A - RF Power Supply Heat On
#define  RF_PS_NOT_READY    0x200000 // 21 K5A - RF Power Supply Ready

//----- bit usage masks for digitalOut -----
// driven by K1, connected to +5 VDC
#define  ACTIVITY_LED    0x000001  // which bit is used for the activity LED
#define  EXT_485_RX      0x000040  // 06 enable Tx on extern 2/4-wire adapter
#define  EXT_485_TX      0x000080  // 07 enable Rx on extern 2/4-wire adapter
// driven by K2, connected to +24 VDC
#define  RF_PS_RESET     0x000200  // 09 Reset pulse for some RF PwrSupply's
#define  RF_PS_STOP      0x000400  // 10 Stop pulse for some RF PwrSupply's
#define  EXT_INTERLOCK   0x000800  // 11 Extern Interlock/Watchdog timer relay
#define  OVEN_PS_OFF     0x001000  // 12 K1 - Oven Power Supply On/Off
#define  BIAS_PS_OFF     0x002000  // 13 K2 - Bias Power Supply On/Off
#define  SPTR_PS_OFF     0x004000  // 14 K3 - Sputterring Disk Pwr Supply On/Off
#define  RF_PS_START     0x008000  // 15 K4 - RF Gen Power Supply Start pulse

#define  EXT_485_CTL     // explicity control Tx on ext 2/4 wire adapter
#define  EXT_485_RX_TX     0x0000C0  // both Tx and Rx bits
#define  NOT_EXT_485_RX_TX 0x00003F  // all BUT Tx and Rx bits (for 8-bit value)


//--- structure for Event messages ----
typedef struct  {
  int   msgID;        // which msg is it?
  unsigned int  seqNum;  // to identify last msg in seq of interdependent ones
  union {
    float dummy;     //
  } msgData;
}  EventMsg;

//--- values for EventMsg.msgID ---



#use "Base_DLP_H.lib"


//----- values we want to be PERSISTENT (i.e. stored in flash) -----
//--- (search for PERSISTENT for places code normally added to support them) ---
//-- !!!!! Adding to this struct is OK, but never touch existing items !!!! ---
typedef struct  {
  uint   vers;            // version of this structure
  uint   devType;         // device type this data came from

  float  ovenReadVGain;
  float  biasReadVGain;
  float  sptrReadVGain;

  float  sptrReadAGain;
  float  biasReadAGain;
  float  ovenReadAGain;

  float  ovenSetVGain;
  float  biasSetVGain;
  float  sptrSetVGain;
  float  rfGenSetVGain;

  float  ovenReadVOffset;
  float  biasReadVOffset;
  float  sptrReadVOffset;

  float  sptrReadAOffset;
  float  biasReadAOffset;
  float  ovenReadAOffset;

  float  ovenSetVOffset;
  float  biasSetVOffset;
  float  sptrSetVOffset;
  float  rfGenSetVOffset;
}  ApplData;

#define  ApplDataVers   1


//========================================
//===== Coil or ModBus bit registers =====
//========================================
//--- writing to a coil causes an action - values are not actually saved ---
//  See NewCoilValue() for details

//--- NOTE: 0-19 (00001-000020) are reserved for common values ---
#define  OvenOn      20  // [00021]
#define  OvenOff     21  // [00022]
#define  BiasOn      22  // [00023]
#define  BiasOff     23  // [00024]
#define  SptrOn      24  // [00025]
#define  SptrOff     25  // [00026]
#define  RFGenStart  26  // [00027]
#define  RFGenStop   27  // [00028]
#define  RFGenReset  28  // [00029]
//                          =======
      #define    NUMCOILS       29


//======================================
//===== Read-Only ModBus registers =====
//======================================
//--- 30001-30050 (0-49) are reserved for generic/common values ---
//--- 30051-3xxxx are for controller specific values ---

RO_REG_U32( devStatus,      30051 )  // Bit-mapped device status
RO_REG_U16( rfSupplyType,   30053 )  // type of RF Power Supply in use
//                          =====
     #define  NUM_READ_REGS    53

//--- bit usage for devStatus --- (30051#U32)
#define  RF_PS_FAULT       0x0001
#define  RF_PS_HEAT_ON     0x0002
#define  RF_PS_READY       0x0004
#define  OVEN_PS_ENABLED   0x0008
#define  BIAS_PS_ENABLED   0x0010
#define  SPTR_PS_ENABLED   0x0020
#define  RF_PS_NO_RESP     0x0040


//--- bit usage for rfSupplyType --- (30052#U16)
#define  RF_TYPE_UNKNOWN   0  //
#define  RF_TYPE_NovaStar  1  // Ameritherm NovaStar (3kW) power supply
#define  RF_TYPE_HotShot   2  // Ameritherm HotShot (2kW) power supply


//=======================================
//===== Read/Write ModBus registers =====
//=======================================
//--- Note that read/write registers can be defined as "Read Only" ---
//--- This allows the client to send write commands for a register to ---
//--- initiate some action without actually changing the value of the ---
//--- register.  Also, devices that have read-only values but multiple ---
//--- logical devices need to use the R/W registers, as the Read-Only ---
//--- set does not support multiple devices, and generally is limited ---
//--- to values that apply to all logical devices handled by one CPU ---

//--- Special handling for chgs to reg values is done by DevNewRegValue() ---

//--- The ChanMult value determines how many registers can be defined ---
//--- per logical device.  A value of 100 means 40001-40100 are for ---
//--- device 0, 40101-40200 are for device 1, etc.  Note that only ---
//--- one set of register definitions is required (i.e. define 40001 ---
//--- but not 40101, 40201, etc).

RW_REG_F32( sptrSetV,        40001, RW )  // Sputter Disk PwrSupply Voltage Set
RW_REG_F32( sptrReadV,       40003, RW )  // Sputter Disk PwrSupply Voltage Read
RW_REG_F32( sptrReadA,       40005, RW )  // Sputter Disk PwrSupply Current Read

RW_REG_F32( biasSetV,        40007, RW )  // Bias Disk PwrSupply Voltage Set
RW_REG_F32( biasReadV,       40009, RW )  // Bias Disk PwrSupply Voltage Read
RW_REG_F32( biasReadA,       40011, RW )  // Bias Disk PwrSupply Current Read

RW_REG_F32( ovenSetV,        40013, RW )  // Oven Disk PwrSupply Voltage Set
RW_REG_F32( ovenReadV,       40015, RW )  // Oven Disk PwrSupply Voltage Read
RW_REG_F32( ovenReadA,       40017, RW )  // Oven Disk PwrSupply Current Read

RW_REG_F32( rfGenSetV,       40019, RW )  // RF Gen Disk PwrSupply Voltage Set
RW_REG_F32( rfGenReadV,      40021, RO )  // RF Gen Disk PwrSupply Voltage Read
RW_REG_F32( rfGenReadA,      40023, RO )  // RF Gen Disk PwrSupply Current Read
RW_REG_F32( rfGenReadW,      40025, RO )  // RF Gen Disk PwrSupply Power Read

RW_REG_F32( ovenReadVGain,   40027, RW )  // multiplier for Oven Volt reading
RW_REG_F32( ovenReadAGain,   40029, RW )  // multiplier for Oven Curr reading

RW_REG_F32( biasReadVGain,   40031, RW )  // multiplier for Bias Volt reading
RW_REG_F32( biasReadAGain,   40033, RW )  // multiplier for Bias Curr reading

RW_REG_F32( sptrReadVGain,   40035, RW )  // multiplier for Sputter Volt reading
RW_REG_F32( sptrReadAGain,   40037, RW )  // multiplier for Sputter Curr reading

RW_REG_F32( ovenSetVGain,    40039, RW )  // multiplier for Oven Volt setting
RW_REG_F32( biasSetVGain,    40041, RW )  // multiplier for Bias Volt setting
RW_REG_F32( sptrSetVGain,    40043, RW )  // multiplier for Sputter Volt setting
RW_REG_F32( rfGenSetVGain,   40045, RW )  // multiplier for RF Gen Volt setting

RW_REG_F32( ovenReadVOffset, 40047, RW )  // multiplier for Oven Volt reading
RW_REG_F32( ovenReadAOffset, 40049, RW )  // multiplier for Oven Curr reading

RW_REG_F32( biasReadVOffset, 40051, RW )  // multiplier for Bias Volt reading
RW_REG_F32( biasReadAOffset, 40053, RW )  // multiplier for Bias Curr reading

RW_REG_F32( sptrReadVOffset, 40055, RW )  // multiplier for Sputter Volt reading
RW_REG_F32( sptrReadAOffset, 40057, RW )  // multiplier for Sputter Curr reading

RW_REG_F32( ovenSetVOffset,  40059, RW )  // multiplier for Oven Volt setting
RW_REG_F32( biasSetVOffset,  40061, RW )  // multiplier for Bias Volt setting
RW_REG_F32( sptrSetVOffset,  40063, RW )  // multiplier for Sputter Volt setting
RW_REG_F32( rfGenSetVOffset, 40065, RW )  // multiplier for RF Gen Volt setting
//                           =====
             #define  NUMREGS   66  // don't forget _x32 types use 2 regs!

// search for _PERSISTENT_ to see where code normally added for saved values

RegInfo  rwRegs[NUMREGS+1];
RegInfo  roRegs[NUM_READ_REGS+1];

//=============================================================================
//              Definitions generally unique to this controller
//=============================================================================


//===== function prototypes for controller-specific functions =====

void PSOnOff(int onCoil, int offCoil, int ctlBit);

void PulseBit(int coilNum, int digOutBit, ulong *startTime);

void Process_RF_PS_Resp(char *resp);


//=============================================================================
//  GLOBAL variables
//=============================================================================

ApplData  applData;  // for storing application specific config data


//==================== device specific stuff ====================

bool  devInitialized, vacuumOK, psNoResponse, restartPS_Service;
int   externWatchDog, psSrvIdx;

ulong  startHeatTime, stopHeatTime, resetHeatTime;

xstring DevStateStr {
"Serial RF PS: ###       RF Heat Reset:   ###     Oven Supply Volts:  -##.###V",
"NonSer RF PS: ###       RF Heat Stop:    ###     Bias Disk Volts:    -##.###V",
"RF PS Fault:  #####     Oven Power:      ###     Sputter Disk Volts: -##.###V",
"RF PS Heat:   ###       Bias Disk Power: ###     Sputter Disk Amps:  -##.###V",
"RF PS Ready:  ###       Sptr.Disk Power: ###     Bias Disk Amps:     -##.###V",
"                        RF Gen Start:    ###     Oven Supply Amps:   -##.###V",
" ",
"RF Read Volts: ###.#V                            Oven Supply Set:    -##.###V",
"RF Read Amps:  ###.#A                            Bias Disk Set:      -##.###V",
"RF Read Watts: ###.#W                            Sputter Disk Set:   -##.###V",
"                                                 RF Generator Set:   -##.###V",
"RF Supply type: [........]",
""
};

const DigFmtInfo digInFmt[] = {
  {  0, 0, ""       ,0,  ""      ,0 }, // 00
  {  0, 0, ""       ,0,  ""      ,0 }, // 01
  {  0, 0, ""       ,0,  ""      ,0 }, // 02
  {  0, 0, ""       ,0,  ""      ,0 }, // 03
  {  0, 0, ""       ,0,  ""      ,0 }, // 04
  {  0, 0, ""       ,0,  ""      ,0 }, // 05
  {  0, 0, ""       ,0,  ""      ,0 }, // 06
  {  0, 0, ""       ,0,  ""      ,0 }, // 07
  {  0, 0, ""       ,0,  ""      ,0 }, // 08
  {  0, 0, ""       ,0,  ""      ,0 }, // 09
  {  0, 0, ""       ,0,  ""      ,0 }, // 00
  {  0, 0, ""       ,0,  ""      ,0 }, // 11
  {  0, 0, ""       ,0,  ""      ,0 }, // 12
  {  0, 0, ""       ,0,  ""      ,0 }, // 13
  {  0, 0, ""       ,0,  ""      ,0 }, // 14
  {  0, 0, ""       ,0,  ""      ,0 }, // 15
  {  0, 0, ""       ,0,  ""      ,0 }, // 16
  { 14, 3, "YES"    ,0,  "NO "   ,0 }, // 17  Serial interface to RF Supply
  { 14, 4, "YES"    ,0,  "NO "   ,0 }, // 18  Non-serial RF Power Supply
  { 14, 5, "FAULT"  ,1,  "OK   " ,0 }, // 19  RF PS Fault
  { 14, 6, "ON   "  ,0,  "OFF  " ,1 }, // 20  RF PS Heat On
  { 14, 7, "YES  "  ,0,  "NO   " ,1 }, // 21  RF PS Ready
  { -1,-1, "",0, "",0 }
};

const DigFmtInfo digOutFmt[] = {
  {  0, 0, ""     ,0,  ""    ,0 },  // 00
  {  0, 0, ""     ,0,  ""    ,0 },  // 01
  {  0, 0, ""     ,0,  ""    ,0 },  // 02
  {  0, 0, ""     ,0,  ""    ,0 },  // 03
  {  0, 0, ""     ,0,  ""    ,0 },  // 04
  {  0, 0, ""     ,0,  ""    ,0 },  // 05
  {  0, 0, ""     ,0,  ""    ,0 },  // 06
  {  0, 0, ""     ,0,  ""    ,0 },  // 07
  {  0, 0, ""     ,0,  ""    ,0 },  // 08
  { 41, 3, "YES"  ,1,  "NO " ,0 },  // 09  // RF Heat Reset: 0 energizes relay
  { 41, 4, "YES"  ,1,  "NO " ,0 },  // 10  // RF Heat Stop: 0 energizes relay
  {  0, 0, ""     ,0,  ""    ,0 },  // 11
  { 41, 5, "ON "  ,0,  "OFF" ,1 },  // 12  // Oven PS: 0 energizes relay
  { 41, 6, "ON "  ,0,  "OFF" ,1 },  // 13  // Bias PS: 0 energizes relay
  { 41, 7, "ON "  ,0,  "OFF" ,1 },  // 14  // Sptr.PS: 0 energizes relay
  { 41, 8, "YES"  ,1,  "NO " ,0 },  // 15  // RF Heat Start: 0 energizes relay
  { -1,-1, "",0, "",0 }
};


const AnalogFmtInfo  adcFmt[] = {
  { 69, 3, 7,3 },
  { 69, 4, 7,3 },
  { 69, 5, 7,3 },
  { 69, 6, 7,3 },
  { 69, 7, 7,3 },
  { 69, 8, 7,3 },
  { -1,-1, 0,0 }
};

const AnalogFmtInfo  dacFmt[] = {
  { 69,10, 7,3 },
  { 69,11, 7,3 },
  { 69,12, 7,3 },
  { 69,13, 7,3 },
  { -1,-1, 0,0 }
};


xstring ApplStrs  {
#define   HelpStr                 ApplStr(0)
  "A=ADC fmt, H=Help, S=Sock/Ser, X=eXt cmd, Z=Zero errs",

#define   ExtHelpStr              ApplStr(1)
  "Bye, Calib, Help, ID, IP, Reboot, Reg, Serial, Track, <reg>",

#define   DevInitMsg              ApplStr(2)
  "Device Initialized...",

#define   ResetApplDataMsg        ApplStr(3)
  "Resetting Application data...",

#define   CmdParserErrMsg         ApplStr(4)
  "cmd parser error",

#define   UnrecogCmdMsg           ApplStr(5)
  "Unrecognized command: %s",

#define   IncompleteCmdMsg        ApplStr(6)
  "Incomplete command",

#define   SetCoilMsg              ApplStr(7)
  "set coil [%05d] = %d",

#define   CalibStateMachErrMsg    ApplStr(8)
  "Error in calibration state machine",

#define   PSNoRespMsg             ApplStr(9)
  "RF PS Not Responding on serial interface",

""
};   //NewStr


xstring ApplHelpStrs {
  "PS cmd                       - Send cmd to RF PwrSupply and display response",
  ""
};

#use "Base_DLP_C.lib"


//----------------------------------------------------------------
void RF_PS_Server(LinkProc *lp);
SerialLinkParams  RF_PS_Params; // {  // baud rate not fixed
//  3,               // Port D
//  9600 or 38400    // 9600 for NovaStar (3kW) 38400 for HotShot (2kW)
//  8,               // data bits
//  PARAM_NOPARITY   // even parity, 1 stop bit
//  SLF_HALFDUPLEX   // ctl of send/rcv mode required
//};
const ServiceDesc  RF_PS_Service = {
  #define  MaxRF_PS_Clients   1    // Max # clients
  "RF_PS_Ser",       // Name
  &RF_PS_Params,     // parameters needed to construct a ComLink object
  LT_ASYNC,          // type of ComLink object/connection
  LogYes,            // log std event messages
  NetClientNo,       // don't incl in count of appl-type network clients
  0,                 // no timeout for idle connections (secs)
  MaxRF_PS_Clients,  // max # of concurrent clients
  &RF_PS_Server      // process that manages this connection
};


#define  RCV_BUF_SIZE  128
#define  MaxCmdSize     64

typedef struct {
  bool   inUse;            // in-use flag
  ulong  cmdCount;         // # cmds rcvd on this connection
  ulong  startIdleTime;    // last time started waiting for idle period
  ulong  xmitStartTime;    // time last cmd was sent
  ulong  rcvdBeforeCmd;    // rcvd count just before last xmit
  bool   userCmd;          // last cmd sent was entered by the user
  int    rcvCount;         // # chars of respBuf[] used so far
  char   respBuf[RCV_BUF_SIZE];
} RF_PS_Client;

RF_PS_Client   RF_PS_Clients[MaxRF_PS_Clients];

char  psCmd[MaxCmdSize];



//=============================================================================
//          Standard functions with APPLICATION SPECIFIC variations
//=============================================================================


//-----------------------------------------------------------------------------
//  Initialization for processing of commands
//-----------------------------------------------------------------------------
int InitCmds(int chanNum)
{
  devInitialized = true;
  ShowMsg(DevInitMsg);

  return 0;
}

//-----------------------------------------------------------------------------
//  Put device in state that is safe for controller to reboot
//-----------------------------------------------------------------------------
void PrepForReboot()
{
  ulong  lastWriteTime;


//... make changes to digitalOut, voltsOut[], etc ...

//--- wait for latest changes to take effect ---
  lastWriteTime = writeTime;  while (writeTime == lastWriteTime);
}


//-----------------------------------------------------------------------------
//  check validity of application/configuration data.  If any of it is not
//  usuable, change it to a reasonable default value and save any changes
//  to flash memory
//-----------------------------------------------------------------------------
void ValidateApplData(bool readFirst)
{
  int  stat;
  bool  dataChgs;
  char  *msg;


  msg = getBuf(90);

  dataChgs = false;

  if (readFirst)
    if ((stat = Block_Read(BID_APPL_DATA, 0, &applData, sizeof(applData))) < 2)
      sprintf(msg, ApplDataReadErrMsg, stat);

  if (applData.devType != DeviceType)  {
    LogMsg(ResetApplDataMsg);
    memset(&applData, 0, sizeof(applData));
    applData.vers = ApplDataVers;  applData.devType = DeviceType;
    Block_SetSize(BID_APPL_DATA, sizeof(applData));  dataChgs = true;
  }

#define CheckRange(param, minVal, maxVal, defValue)  \
  { if ((applData.param < (minVal)) OR (applData.param > (maxVal)))  {  \
    applData.param = (defValue);  dataChgs = true;  } }

#define CheckZero(param, defValue)  \
  { if (applData.param == 0.0)  {  \
    applData.param = (defValue);  dataChgs = true;  } }


//  if (applData.param not valid)  {
//    applData.param = defaultValue;
//    dataChgs = true;
//  }

  CheckZero( ovenReadVGain, 6.0 );
  CheckZero( ovenReadAGain, 66.7 );
  CheckZero( ovenSetVGain, 0.0666667 );

  CheckZero( biasReadVGain, 100.0 );
  CheckZero( biasReadAGain, 0.001 );
  CheckZero( biasSetVGain,  -0.01 );

  CheckZero( sptrReadVGain, 100.0 );
  CheckZero( sptrReadAGain, 0.001 );
  CheckZero( sptrSetVGain,  -0.01 );

  CheckZero( rfGenSetVGain, -0.01 );

// search for _PERSISTENT_ to see where code normally added for saved values


  freeTo(msg);

  if (dataChgs)  SaveApplData(true);
}



//-----------------------------------------------------------------------------
//  process a multi-character command from the diagnostic terminal/console
//-----------------------------------------------------------------------------
void ProcessStrCmd(char *cmd, LinkProc *lp)
{
  char  *ptr;
  int  i, chan, state, digState;
  ulong  mask;
  float  volts;
  bool  validCmd;
  char  *msg, *temp, *cmdStr;


  msg = getBuf(90 + 90 + 90);  temp = msg + 90;  cmdStr = temp + 90;

  strcpy(cmdStr, cmd);  ptr = strtok(cmd, " ");
  state = 0;  validCmd = false;
  strcpy(msg, CmdParserErrMsg);

  while (ptr AND (state != 999))  {

    switch (state)  {

      case 0:  // which command did they type?
        if (strcmpi(ptr, "ps") == 0)  { state = 104;  break; }

        sprintf(msg, UnrecogCmdMsg, ptr);
        state = 999;  break;


      case 104:  // set next cmd to send to RF Power Supply
        strncpy(psCmd, ptr, sizeof(psCmd));
        psCmd[sizeof(psCmd)-2] = 0;  strcat(psCmd, "\r");
        *msg = 0;  validCmd = true;  state = 999;  break;
    }

    if (ptr)  ptr = strtok(NULL, " ");
  }
  strCmdState = false;



//--- what state were we in when we ran out of cmd string to parse? ---
  switch (state)  {

    case 999:  // msg supplied by cmd parser state machine
      break;

    default:
      strcpy(msg, IncompleteCmdMsg);
  }

  if (! validCmd)
    ProcessStrCmdCommon(cmdStr, msg, lp);  // not valid dev cmd - check generic ones
  else
    if (*msg)
      if (StreamDisp)
        ShowMsg(msg);
      else  {
        respShownTime = MS_TIMER - 5000;
        DispStr(0,23, msg, true);
        respShownTime = MS_TIMER;
      }

  freeTo(msg);
}


//-----------------------------------------------------------------------------
//  process single-char commands from the diagnostic console
//-----------------------------------------------------------------------------
void CharCmd(char cmd)
{
  CharCmdCommon(cmd);

  cmd |= 0x20;

//--- device-specfic stuff ---
//  if (cmd == '?') ...

}


//-----------------------------------------------------------------------------
//  Update data based on digital and analog input values and determine new
//  values for the digital and analog outputs
//
//  !!!!! WARNING !!!!!
//
//  This function is called by the timer-interrupt driven function that reads
//  and writes the Analog and Digital I/O values.  It should be limited to
//  processing current input values and updating output values that will become
//  the active values on the next interrupt (currently, this interuupt occurs
//  every 10ms, so this routine must take considerably less time than that to
//  do EVERYTHING it needs to).
//-----------------------------------------------------------------------------
nodebug void UpdateDeviceState()
{
  ulong  secTime, mask;
  int  i, rate, newType;
  static int  prevComCount;


//--- update register values ---
  ovenReadV = voltsIn[0] * ovenReadVGain + ovenReadVOffset;
  ovenReadA = voltsIn[5] * ovenReadAGain + ovenReadAOffset;

  biasReadV = voltsIn[1] * biasReadVGain + biasReadVOffset;
  biasReadA = voltsIn[4] * biasReadAGain + biasReadAOffset;

  sptrReadV = voltsIn[2] * sptrReadVGain + sptrReadVOffset;
  sptrReadA = voltsIn[3] * sptrReadAGain + sptrReadAOffset;

  vacuumOK = true;  //finish - should read from PLC device


  devStatus = 0;

  if (!(digitalIn & RF_PS_HotShot))   newType = RF_TYPE_HotShot;   else
  if (!(digitalIn & RF_PS_NovaStar))  newType = RF_TYPE_NovaStar;  else
                                      newType = RF_TYPE_UNKNOWN;
  if (newType != rfSupplyType)  {
    rfSupplyType = newType;
    if (newType == RF_TYPE_HotShot)
      RF_PS_Params.baud = 38400;
    else
      RF_PS_Params.baud = 9600;
    restartPS_Service = true;
  }


  if (! (digitalIn & RF_PS_NOT_FAULT))    devStatus |= RF_PS_FAULT;
  if (! (digitalIn & RF_PS_NOT_HEAT_ON))  devStatus |= RF_PS_HEAT_ON;
  if (! (digitalIn & RF_PS_NOT_READY))    devStatus |= RF_PS_READY;

  if (! (digitalOut & OVEN_PS_OFF)   )    devStatus |= OVEN_PS_ENABLED;
  if (! (digitalOut & BIAS_PS_OFF))       devStatus |= BIAS_PS_ENABLED;
  if (! (digitalOut & SPTR_PS_OFF))       devStatus |= SPTR_PS_ENABLED;

  if (psNoResponse)  devStatus |= RF_PS_NO_RESP;


//--- most the supplies have simple On/Off control ---
  PSOnOff(OvenOn, OvenOff, OVEN_PS_OFF);
  PSOnOff(BiasOn, BiasOff, BIAS_PS_OFF);
  PSOnOff(SptrOn, SptrOff, SPTR_PS_OFF);

//--- RF Gen supply uses Start/Stop/Reset pulses ---
  PulseBit(RFGenStart, RF_PS_START, &startHeatTime);
  PulseBit(RFGenStop,  RF_PS_STOP,  &stopHeatTime);
  PulseBit(RFGenReset, RF_PS_RESET, &resetHeatTime);


//--- An external watchdog timer/relay is used to trip a safety interlock ---
//--- if it does not receive a pulse every few seconds.  This pulse is ---
//--- generated here by setting one of the output lines low for one control ---
//--- cycle based on a counter.  Note that the counter is reset only if ---
//--- a value controlled by the background ComLoop task changes value. ---
//--- This insures that the pulses will stop if either the ComLoop or the ---
//--- ControlLoop tasks have major problems ---
  if (externWatchDog != 1)
    SetBits(digitalOut, EXT_INTERLOCK);
  else
    ClearBits(digitalOut, EXT_INTERLOCK);
  if (externWatchDog > 0)
    --externWatchDog;
  else
    if (comCount != prevComCount)  externWatchDog = 10;
  prevComCount = comCount;


#define DisableAll  (OVEN_PS_OFF | BIAS_PS_OFF | SPTR_PS_OFF )

  if (! vacuumOK)  {
    SetBits(digitalOut, DisableAll);  SetBit(coils[0], RFGenStop);  }

  voltsOut[0] = (ovenSetV - ovenSetVOffset) / ovenSetVGain;
  voltsOut[1] = (biasSetV - biasSetVOffset) / biasSetVGain;
  voltsOut[2] = (sptrSetV - sptrSetVOffset) / sptrSetVGain;
  voltsOut[3] = (rfGenSetV - rfGenSetVOffset) / rfGenSetVGain;

  for (i=0; i<NumDAC; ++i)  {
    if (voltsOut[i] < DACMin)  voltsOut[i] = DACMin;
    if (voltsOut[i] > DACMax)  voltsOut[i] = DACMax;
  }
}


//-----------------------------------------------------------------------------
//  Do anything specific to this controller that is not time-critical.
//  This is called once during each pass of the ComLoop() function
//-----------------------------------------------------------------------------
void BackGroundProcess()
{
  if (recursiveBackGround)  return;
  recursiveBackGround = true;

  StackPtr();


  recursiveBackGround = false;
}


//-----------------------------------------------------------------------------
//  Do device specific formatting of current state data that cannot be
//  handled by entries in generic table for I/O values
//
//  Returns:
//      true:  last portion of data has been displayed
//      false:  more data to display during next pass
//-----------------------------------------------------------------------------
bool ShowDeviceState()
{
  static devShowState;
  char  *msg, *str;


  msg = getBuf(90);

  switch (++devShowState)  {

    case 1:
      if (psNoResponse)  {
        if (blinkOff)  str = "     ";  else  str = "*N/A*";
        DispStr(15,10, str, false);
        DispStr(15,11, str, false);
        DispStr(15,12, str, false);
      }
      else  {
        DispStr(15,10, fmtFloat(rfGenReadV,5,1,msg), false);
        DispStr(15,11, fmtFloat(rfGenReadA,5,1,msg), false);
        DispStr(15,12, fmtFloat(rfGenReadW,5,1,msg), false);
      }

      if (rfSupplyType == RF_TYPE_HotShot)  str = "HotShot ";  else
      if (rfSupplyType == RF_TYPE_NovaStar) str = "NovaStar";  else
      if (blinkOff)  str = "        ";  else  str = "UNKNOWN ";
      DispStr(17,14, str, false);

      break;

  //--- to catch all non-active state values ---
    default:
      devShowState = 0;  break;
  }

  freeTo(msg);

  return (devShowState == 0);
}


//-----------------------------------------------------------------------------
//  Determine what, if any, action should be taken as a result of one of the
//  coil values being changed
//-----------------------------------------------------------------------------
void NewCoilValue(int chanNum, int coilNum, int coilVal, int cmdRow)
{
  char  *msg;
  RegInfo  *reg;


  StackPtr();

  msg = getBuf(80);

  sprintf(msg, SetCoilMsg, coilNum+1, coilVal);
//  if (GenericDisp)  DispStr(CmdCol,cmdRow, msg, true);
  LogMsg(msg);

  if ((coilNum < 0) || (coilNum >= NUMCOILS))  { freeTo(msg);  return; }

  if (coilVal)  switch (coilNum)  {

    case DoReboot:
      Reboot(10);  coilVal = 0;  break;

    case DoInit:
      InitCmds(0);  coilVal = 0;  break;

    case AbortCalib:
      if (calibrating)  cancelCalib = true;
      coilVal = 0;  break;

  }
  freeTo(msg);

  ++criticalSection;
  if (coilVal)
    SetBit(coils[chanNum], coilNum);
  else
    ClearBit(coils[chanNum], coilNum);
  --criticalSection;
}


//-----------------------------------------------------------------------------
//  Determine what, if any, action should be taken as a result of one of the
//  registers being written to.
//
//  Called by the generic function NewRegValue() in the base code which is
//  responsible for actually copying the <regName>_New value to <regName>
//
//  Returns true if the new value should be used, false if it should not
//
//  If any value used during execution of the periodic ControlLoop() function
//  (besides the new register value) is altered here, you MUST increase the
//  criticalSection counter 1st (it will be restored to its previous value
//  after the new register value is in effect).  This is to avoid having the
//  ControlLoop() processes act on unsynchronized values.
//
//    Example:  DevNewRegValue() alters some non-register value in response
//    to a new register value.  Since the new register value does not go
//    into effect until DevNewRegValue() returns, any such values used within
//    ControlLoop() would be out of sync with the current register value.
//
//  NOTE:  Assumes that, when a register that is part of a larger data type
//         (e.g. a 32-bit float or long integer) that the lower numbered
//         registers will be written first.
//
//  Any string in *msg will be logged after returning from this func.
//-----------------------------------------------------------------------------
bool  DevNewRegValue(int chanNum, int regNum, int regVal,
                     int clientType, LinkProc *lp, char *msg)
{
  int  cmdRow;


  cmdRow = CmdRow(LinkProc_Idx(lp));

//--- Check the New register value (or set of registers comprising a ---
//--- single value) and either reject it (in which case the new value ---
//--- will be ignored), or do whatever processing is necessary using the ---
//--- new value.  NOTE: The new value does not actually go in to effect ---
//--- until sometime after this function returns (one of the reasons that ---
//--- proper handling of the criticalSection counter is required). ---

  do  {

//  if (regNum == regName_Idx)  {
//    if (regName_New not usable)  {
//      sprintf(msg, "optional msg if default 'ignored' msg insufficient");
//      return false;
//    }
//    ... whatever else needs to be done with regName_New value ...
//    sprintf(msg, "...optional msg about action taken...");
//    (NOTE: regName = regName_New is handled by caller of this function)
//    break;
//  }

#define  CheckReg(regName, minVal, maxVal)  \
    if (regNum == regName ## _Idx)  {  \
      if (regName ## _New < (minVal))  regName ## _New = (minVal);  \
      if (regName ## _New > (maxVal))  regName ## _New = (maxVal);  \
      break;  \
    }

    CheckReg( ovenSetV, 0.0, 60.0);
    CheckReg( biasSetV, -10000.0, 0.0);
    CheckReg( sptrSetV, -10000.0, 0.0);
    CheckReg( rfGenSetV, 0, 10.0);


#define  SaveRegNZ(name)  \
   if (regNum == name ## _Idx)  {  \
     if (name ## _New == 0.0)  return false;  \
     ++criticalSection;  \
     applData.name = name = name ## _New;  ChgWait(1);  \
   }

   SaveRegNZ( ovenReadVGain );
   SaveRegNZ( ovenReadAGain );
   SaveRegNZ( biasReadVGain );
   SaveRegNZ( biasReadAGain );
   SaveRegNZ( sptrReadVGain );
   SaveRegNZ( sptrReadAGain );
   SaveRegNZ( ovenSetVGain );
   SaveRegNZ( biasSetVGain );
   SaveRegNZ( sptrSetVGain );
   SaveRegNZ( rfGenSetVGain );

#define  SaveReg(name)  \
   if (regNum == name ## _Idx)  {  \
     ++criticalSection;  \
     applData.name = name = name ## _New;  ChgWait(1);  \
   }

   SaveReg( ovenReadVOffset );
   SaveReg( ovenReadAOffset );
   SaveReg( biasReadVOffset );
   SaveReg( biasReadAOffset );
   SaveReg( sptrReadVOffset );
   SaveReg( sptrReadAOffset );
   SaveReg( ovenSetVOffset );
   SaveReg( biasSetVOffset );
   SaveReg( sptrSetVOffset );
   SaveReg( rfGenSetVOffset );


// search for _PERSISTENT_ to see where code normally added for saved values

  } while (false);

  return true;
}


//---------------------------------------------------------------------------
//  Multiple calibration processes can be supported simply by having a
//  different starting state for each one.
//---------------------------------------------------------------------------
void calib_tick()
{
  int  i;
  ulong  waited;
  static int  returnState, cancelState;
  static ulong  startTime, waitStart, minWaitMS, maxWaitMS, lastReadTime;
  static char  msg[90];


#define  StopBecause(str)  {  \
  strcpy(msg, (str));  calibState = cancelState;  break;  }


  if (recursiveCalib)  return;
  recursiveCalib = true;

//--- wait until we have new values to work with ---
  if (RdTime == lastReadTime)  { recursiveCalib = false;  return; }
  lastReadTime = rdTime;

//--- start a new calibration? ---
  if (startCalibState AND !calibrating)  {
    *msg = 0;  calibState = startCalibState;  cancelState = CALIB_IDLE; }
  startCalibState = CALIB_IDLE;

//--- external cancelation of current calibration? ---
  if (cancelCalib)  {
    if (calibrating)  { strcpy(msg, "Canceled");  calibState = cancelState; }
    cancelCalib = false;
  }


  switch (calibState)  {

    case CALIB_IDLE:
      break;


//  case CALIB_xxx_INIT:  // prep for specific calibration process
//    ... set state of machine as needed to begin calibration ...
//    cancelState = CALIB_xxx_ABORT;   calibState = CALIB_xxx_1;
//    ShowMsg("Starting ... calibration");   break;


//  case CALIB_xxx_1:  // check progress, "wait" if necessary ...
//    if (! (calibration-friendly-conditions))   StopBecause("xxx");
//    if (1st-state-reached)  {
//      ShowMsg("1st state achieved!!!");
//      calibState = CALIB_xxx_2;  break;
//    }
//    tweak-state-as-needed;
//    if (tweaked-too-far)  StopBecause("xxx");
//    calibState = CALIB_xxx_WAIT;  break;


//  case CALIB_xxx_2:  // check progress, "wait" if necessary ...
//    if (! (calibration-friendly-conditions))   StopBecause("xxx");
//    if (1st-state-reached)  {
//      ShowMsg("2nd state achieved!!!");
//      calibState = CALIB_xxx_RETURN;  break;
//    }
//    tweak-state-as-needed;
//    if (tweaked-too-far)  StopBecause("xxx");
//    calibState = CALIB_xxx_WAIT;  break;


//  case CALIB_xxx_WAIT:  // do asynchronous "wait" for an event
//    waited = elapsedMS(busyStart);
//    if (waited < minWaitMS)  break;
//    if ((event-not-yet-happened) AND (waited < maxWaitMS))  break;
//    calibState = returnState;  break;


//  case CALIB_xxx_ABORT:
//    if (*msg)  {  strcat(msg, ": Aborting Calibration...");  LogErr(msg); }
//    calibState = CALIB_xxx_RETURN;

//  case CALIB_xxx_RETURN:
//    put things back the way they were before calibration began
//    calibState = cancelState = CALIB_IDLE;  break;



    default:
      LogErr(CalibStateMachErrMsg);  calibState = CALIB_IDLE;
  }

  recursiveCalib = false;
}



//=============================================================================
//                   Functions unique to this application
//=============================================================================

//-----------------------------------------------------------------------------
//  Turn Power Supply control bit on or off to reset or energize a relay that
//  turns a Power Supply Off/On
//-----------------------------------------------------------------------------
void PSOnOff(int onCoil, int offCoil, int ctlBit)
{
//--- if ON coil is set, then energize the related relay and reset coil ---
  if (GetBit(coils[0], onCoil))  {
    if (vacuumOK)  ClearBits(digitalOut, ctlBit);
    ClearBit(coils[0], onCoil);
  }

//--- if OFF coil is set, then de-energize the related relay and reset coil ---
  if (GetBit(coils[0], offCoil))  {
    SetBits(digitalOut, ctlBit);
    ClearBit(coils[0], offCoil);
  }
}


//-----------------------------------------------------------------------------
//  Start or stop a 0-level pulse on one of the output bits
//-----------------------------------------------------------------------------
void PulseBit(int coilNum, int digOutBit, ulong *startTime)
{
//--- if specified coil set, then set output bit to gnd level & reset coil ---
  if (GetBit(coils[0], coilNum))  {
    ClearBits(digitalOut, digOutBit);
    ClearBit(coils[0], coilNum);
    *startTime = MS_TIMER;  return;
  }

//--- if pulse not currently active, just return ---
  if (! *startTime)  return;

//--- if pulse not yet finished, return ---
  if (elapsedMS(*startTime) < RELAY_PULSE_LEN)  return;

//--- end the pulse by setting output bit back to high-level ---
  SetBits(digitalOut, digOutBit);  *startTime  = 0;
}


//-----------------------------------------------------------------------------
//  Parse the response from the Ameritherm Power Supply
//-----------------------------------------------------------------------------
void Process_RF_PS_Resp(char *resp)
{
  char  *ptr;
  int  len;



  switch  (rfSupplyType)  {

    case RF_TYPE_HotShot:
      // scan bkwds from end for 1st space char
      for (ptr = resp+strlen(resp)-1; (ptr > resp) AND (*ptr > ' '); --ptr);

      if (strncmp(resp, "Power Out", 9) == 0)  rfGenReadW = atof(ptr);  else
      if (strncmp(resp, "Amps Out", 8) == 0)  rfGenReadA = atof(ptr);

      if (rfGenReadA == 0.0)
        rfGenReadV = 0.0;
      else
        rfGenReadV = rfGenReadW / rfGenReadA;

      break;


    case RF_TYPE_NovaStar:
      ptr = strtok(resp, " ");
      while (ptr)  {
        if ((len = strlen(ptr)) < 1)  break;
        while (*ptr AND (*ptr == '0'))  { ++ptr;  --len; } // skip any leading 0's

        switch (ptr[len-1])  {
          case 'V':  rfGenReadV = atof(ptr);  break;
          case 'A':  rfGenReadA = atof(ptr);  break;
          case 'W':  rfGenReadW = atof(ptr);  break;
        }

        if (ptr)  ptr = strtok(NULL, " ");
      }
      break;

  }

}

//-----------------------------------------------------------------------------
//  send commands to and interpret responses from the RF Power Supply
//-----------------------------------------------------------------------------
void RF_PS_Server(LinkProc *lp)
{
  short  i, y, numBytes, err, cmdLen;
  char  ch;
  bool  showDetails;
  char  *msg, *temp, *statCmd;
  ComLink  *cl;
  RF_PS_Client  *ps;
  SerialLink  *sl;


#define  RFS_SEND_CMD     1
#define  RFS_XMIT_ACTIVE  2
#define  RFS_WAIT_START   3
#define  RFS_WAIT_END     4
#define  RFS_PROCESS      5
#define  RFS_WAIT_IDLE    6

#define  MaxRespSize  ( sizeof(RF_PS_Clients[0].respBuf) )

  StackPtr();

  msg = getBuf(90 + 90);  temp = msg + 90;

  y = CmdRow(LinkProc_Idx(lp));

  cl = lp->comLink;  sl = (SerialLink *) cl;
  ps = (RF_PS_Client *) lp->srvData;

  showDetails = GenericDisp AND lp->verboseDisp;

  if (err = cl->getErr(cl))  {
    ShowMsg(cl->errDesc(cl, err, temp));  ++serialErrs;  ++ttlErrs;  }

  if (restartPS_Service)  {
    LinkProc_StartClose(lp);  restartPS_Service = false;  }


  switch (lp->srvState)  {

    case SRV_NEW:  // initialize data for new client
      if ((ps = New_(RF_PS_Client)) == NULL)  {
        LinkProc_StartClose(lp);  break; }
      lp->srvData = (void *) ps;  lp->tempUsed = 0;
      lp->srvState = RFS_SEND_CMD;


    case RFS_SEND_CMD:  // send a command
      if (cl->wrFree(cl) < MaxCmdSize)  break;
      if (! rfSupplyType)  { psNoResponse = true;  return; }
      if (ps->userCmd = *psCmd)
        statCmd = psCmd;
      else
        if (rfSupplyType == RF_TYPE_HotShot)
          statCmd = "1,data\r";  else  statCmd = "r001\r";
      cmdLen = strlen(statCmd);   cl->write(cl, statCmd, cmdLen);
      *psCmd = 0;  ps->xmitStartTime = MS_TIMER;
      ps->rcvdBeforeCmd = cl->readCount;  lp->srvState = RFS_XMIT_ACTIVE;
      break;


    case RFS_XMIT_ACTIVE:  // wait for transmission to finish
      if (cl->unsent(cl) <= 0)  lp->srvState = RFS_WAIT_START;
      break;


    case RFS_WAIT_START:  // wait for start of next line of resp from supply
      if (cl->unread(cl) <= 0)  {
        if (elapsedMS(cl->lastActive) < 500)  break;  // timeout for 1st char
        if (cl->readCount == ps->rcvdBeforeCmd)  {
          psNoResponse = true;
          if (GenericDisp)  DispStr(CmdCol,y, "No Response", true);
          if (!StreamDisp)  DispStr(0,21, blinkOff ? " " : PSNoRespMsg, true);
        }
        lp->srvState = RFS_SEND_CMD;  break;
      }
      if ((ch = cl->getc(cl)) > '~')  ch = '*';
      *ps->respBuf = ch;
      if (*ps->respBuf < ' ')  break;  // LF indicates start of msg
      ps->rcvCount = 1;  lp->srvState = RFS_WAIT_END;
      break;


    case RFS_WAIT_END:  // wait for the end of the current resp line
      if ((numBytes = cl->unread(cl)) <= 0)  // nothing new
        if (elapsedMS(cl->lastActive) > 9)  {  // max 10ms between chars
          ps->respBuf[ps->rcvCount] = 0;
          sprintf(msg, "Timeout: %s", ps->respBuf);
          ShowMsg(msg);  ++serialErrs;  ++ttlErrs;
          lp->srvState = RFS_WAIT_IDLE;  break;
        }
      for (i=0; i<numBytes; ++i)  {
        if (ps->rcvCount >= MaxRespSize)  {
          ShowMsg("Msg too long");  ++serialErrs;  ++ttlErrs;
          lp->srvState = RFS_WAIT_IDLE;  break;
        }
        if ((ch = cl->getc(cl)) > '~')  ch = '*';
        ps->respBuf[ps->rcvCount] = ch;
        if (ch >= ' ')  { ++ps->rcvCount;  continue; }  // CR marks EOL
        ps->respBuf[ps->rcvCount] = 0;  psNoResponse = false;
        lp->srvState = RFS_PROCESS;  break;
      }
      if (elapsedMS(ps->xmitStartTime) > 10000)  // in case of endless resp
        lp->srvState = RFS_SEND_CMD;
      break;


    case RFS_PROCESS:
      if (showDetails)  DispStr(CmdCol,y, ps->respBuf, true);
      if (ps->userCmd)
        ShowMsg(ps->respBuf);
      else
        Process_RF_PS_Resp(ps->respBuf);
      lp->srvState = RFS_WAIT_START;  break;


    case RFS_WAIT_IDLE:  // wait for idle/no data rcvd after an error
      if (!ps->startIdleTime)  ps->startIdleTime = MS_TIMER;
      if (cl->unread(cl) > 0)  {
        while (cl->read(cl, ps->respBuf, MaxRespSize) > 0);  //flush read buf
        break;
      }
      if (elapsedMS(cl->lastActive) > 100)  {
        ps->startIdleTime = 0;  lp->srvState = RFS_SEND_CMD;  }
      if (elapsedMS(ps->xmitStartTime) > 10000)  {
        ShowMsg("PS won't shut up!");  lp->srvState = SRV_CLEANUP; }
      break;


    case SRV_CLEANUP:
      if (ps)  { lp->srvData = NULL;  Free_(ps, RF_PS_Client);  }
      cl->startClose(cl);  lp->srvState = SRV_DONE;  break;
  }


  freeTo(msg);
}


//-----------------------------------------------------------------------------
//  Handle any device-specific device special file open requests
//-----------------------------------------------------------------------------
int DLP_DevFile_Open(LinkProc *lp, int argc, char *argv[])
{
  return -999;
}


//============================================================================
main()
{
  int  i;
  char  *msg;


  firstTime = true;

  while (true)  {
    CommonInit();

    msg = getBuf(128);   // do NOT free up - "backup" in case of failed getBuf()


    #if (_BOARD_TYPE_ != RCM3200)
      ioBoardType = SmartCat;
    #else
      ioBoardType = NSCL16;
    #endif

    digOutBootState =  OVEN_PS_OFF | BIAS_PS_OFF | SPTR_PS_OFF | EXT_485_RX;

    CommonHdwInit(digOutBootState);  DLPHdwInit();


  //--- make sure output voltages are set as needed at startup ---
    for (i=0; i<NumADC; ++i)  voltsIn[i] = 0.0;
    for (i=0; i<NumDAC; ++i)  voltsOut[i] = 0.0;

    DLPInit();

  //========== Remaining Device-specific Initialization ==========

  #define RestoreReg(regName)  regName = applData.regName

    RestoreReg( ovenReadVGain );
    RestoreReg( ovenReadAGain );

    RestoreReg( biasReadVGain );
    RestoreReg( biasReadAGain );

    RestoreReg( sptrReadVGain );
    RestoreReg( sptrReadAGain );

    RestoreReg( ovenSetVGain );
    RestoreReg( biasSetVGain );
    RestoreReg( sptrSetVGain );
    RestoreReg( rfGenSetVGain );


    RestoreReg( ovenReadVOffset );
    RestoreReg( ovenReadAOffset );

    RestoreReg( biasReadVOffset );
    RestoreReg( biasReadAOffset );

    RestoreReg( sptrReadVOffset );
    RestoreReg( sptrReadAOffset );

    RestoreReg( ovenSetVOffset );
    RestoreReg( biasSetVOffset );
    RestoreReg( sptrSetVOffset );
    RestoreReg( rfGenSetVOffset );

// search for _PERSISTENT_ to see where code must be added for saved values

    devInitialized = vacuumOK = psNoResponse = false;

    startHeatTime = stopHeatTime = resetHeatTime = MS_TIMER;

    calibState = CALIB_IDLE;

    RF_PS_Params.port = 3;
    RF_PS_Params.baud = 38400;  // different for different supplies
    RF_PS_Params.bits = 8;
    RF_PS_Params.parity = PARAM_NOPARITY;
    RF_PS_Params.flags = SLF_HALFDUPLEX;

    rfSupplyType = RF_TYPE_UNKNOWN;  *psCmd = 0;  restartPS_Service = false;
    memset(RF_PS_Clients, 0, sizeof(RF_PS_Clients));
    psSrvIdx = Service_Add( &RF_PS_Service );  *psCmd = 0;

    controlEnabled = true;  //initiate active control of the device

    while (ComLoop());

    freeTo(msg);  firstTime = false;
  }
}

