//----- TCPSAppl.c ----- 12/20/12 -----(02/17/09)----

// EE Bundle #096

// Sections that typically require application-specific code are marked with
// the comment:  //APPLCODE

//-----------------------------------------------------------------------------
//  Application specific code for the main application task that is run at
//  a fixed interval and application specific commands run on demand.
//
//APPLCODE
//  Application code to manage the mechanical or SCR controlled reversing
//  switches for Trim Coil Power Supplies.
//-----------------------------------------------------------------------------
//  Change log (most recent first):
//
//
//  2012-12-20 - Mark Davis
//               - Rather than simply not allowing a softReset when the variac
//                 is moving in the UP direction, the direction of movement
//                 is simply reversed to insure the bias voltage doesn't
//                 increase to a dangerous level during the restart.
//
//  2012-12-14 - Mark Davis
//               - Added periodic logging of the device state.
//
//  2011-12-20 - Mark Davis
//               - Fixed the switchState logic so it can't get stuck in the
//                 wait-for-zero state.  When the polarity of the setpoint is
//                 changed so that it is different than the current polarity
//                 of the switch, switchState is set to wait-for-zero.  Prior
//                 to this fix, if the polarity of the setpoint is changed
//                 again while switchState is still set to wait-for-zero, it
//                 could remain stuck in the wait-for-zero state (which would
//                 never be satisfied as long as the setpoint remained well
//                 above 0).
//
//  2011-03-24 - Mark Davis
//               - Enabled softReset mode for this application.  The softReset
//                 option is temporarily disabled at times when it may not be
//                 safe to reboot and leave the outputs in the state they are
//                 in (such as when the bias voltage is being adjusted).
//               - Disabled redundant log msg for new setpoint when it is
//                 received from an external client (a process that already
//                 logs the event).  This improves response time a bit when
//                 interactively changing the setpoint using a knob.
//
//  2011-02-08 - Mark Davis
//               - Removed the extra "Interlock cleared" message that was
//                 logged at the end of a reset pulse if there is no
//                 external interlock asserted.  This not only is redundant
//                 (all the digital inputs are already monitored and changes
//                 logged), but would show that an Interlock was cleared even
//                 if there wasn't one.
//               - The existing ON bit in the devStatus now indicates the
//                 state of the controller logic rather than the supply, and
//                 a new bit was added to provide the reported state of the
//                 supply.
//
//  2011-01-19 - Mark Davis
//               - Added minBiasChgRate and enhanced the bias voltage error
//                 tracking to allow bias voltage errors to exceed the
//                 specified interval as long as the error is shrinking by at
//                 least the specified min rate.  Prior to this change, the
//                 normal jump in the bias voltage caused by a rapid decrease
//                 in the current (followed by a much slower drop when the
//                 new current was close to 0) would require the biasErrMS
//                 to be set so high that real problems could be ignored
//                 long enough to endanger the transistors.
//
//  2011-01-18 - Mark Davis
//               - Fixed bug that stopped bias voltage error interlock events
//                 from being logged after the first occurance.
//               - Bias voltage is now adjust if it is too large even when the
//                 supply is not active.
//
//  2010-08-09 - Mark Davis
//               - The applied setpoint now follows the ramprate setting when
//                 the requested setpoint changes polarity, and the process of
//                 actually changing the polarity of the switch is now delayed
//                 until the readback drops below the safeSwAmps value.
//               - Added maxRTZms to support ramping to zero before actually
//                 turning the supply off in response to an Off request.  A
//                 value of zero means turn off with no delay.
//
//  2010-08-06 - Mark Davis
//               - Moved the call to logDigIOChgs() to AFTER the updated
//                 Next is copied to Cur (and Cur to Prev) so the messages
//                 make sense!
//               - Added the debouncing of the digital inputs that was in the
//                 older (pre-DEBROS) version.
//
//  2010-08-05 - Mark Davis
//               - The interlocks input (which indicates a Water Flow,
//                 Temperature, or OFF Button interlock) now asserts an
//                 interlocked state when it occurs.  This is necessary
//                 because these interlocks are latched and require a reset
//                 pulse from the controller to reset them.
//
//  2010-08-04 - Mark Davis
//               - Moved default read/set calib values from setDefaultConfig()
//                 to readDefaults() and setDefaults() (where they are supposed
//                 to be).
//               - Added polThreshVolts to avoid confusion and allowing testing
//                 with a short for a load.  When a short is used for a load,
//                 the load voltage is still close to zero even for significant
//                 current levels, so this guards against erroneous polarity
//                 errors that would otherwise occur when testing with a short.
//               - Basic testing/debugging completed on SCR vers of supply
//                 using a DC load (a short).
//
//  2010-08-03 - Mark Davis
//               - Added checks for whether the bias voltage is doing what it
//                 should:
//                   - A logged message for when it varies from the setpoint
//                     for too long a period
//                   - An interlock if its amplitude exceeds the setpoint by
//                     more than the deadband amount for too long a period
//
//  2010-07-28 - Mark Davis
//               - First tests on real hardware (with a short for a load)
//               - Added line that insures that Variac Down is asserted when
//                 the supply is not on (it was leaving it unchanged).
//               - Added defaults for the read/set gains.
//
//  2010-07-26 - Mark Davis
//               - The older-style devStatus value is now updated each pass (it
//                 was previously always 0).
//               - Updated/revised comments.
//
//  2009-12-14 - Mark Davis
//               - Changed order of param files to deal with problems caused
//                 by the order in which they were restored at startup.
//
//  2009-02-17 - Mark Davis -
//               - Started port of older DynamicC version of application to
//                 the newer Softools + DEBROS RTOS platform
//
//-----------------------------------------------------------------------------
//  To-Do:
//
//   - Rather than clearing the allowSoftReset when the variac is increasing
//     the transistor voltage, just insure the variac is not moving in the
//     positive direction during a restart.  This would be the same as allowing
//     a soft reset when the variac is NOT moving in the + direction, so it
//     posses no greater risk than it does now by allowing a soft reset while
//     the variac is not being moved.
//-----------------------------------------------------------------------------


//=============================================================================
//
//  The purpose of the TrimCoil Power Supply (TCPS) controller is to provide a
//  bi-polar output to a load using a uni-poloar Power Supply.  This is
//  achieved by using a pair of relays to swap the (+) and (-) connections
//  from the power supply to the load.
//
//
//  Current control
//
//    A variac and a transistor bank control the output of the power supply.
//    The input to the variac is an AC voltage.  The output of the variac is
//    rectified and provides the collector bias voltage to the transistor bank.
//    An analog PI circuit controls the transistor bank, regulating the current
//    output.  The emitter output from the transistor bank is connected to a
//    reversing switch, which allows control of the polarity of the output.
//
//    The ZWorld controls the position of the variac via digital outputs that
//    enable movement in one of two directions.  The rectified bias voltage is
//    monitored via an analog input and the position of the variac is altered as
//    needed to maintain the voltage within specified parameters.  The setting
//    for the analog PI circuit that regulates the current output is controlled
//    by one of the ZWorld's analog outputs.
//
//    The output of the transistor bank is connected to a shunt and the voltage
//    across the shunt monitored via an analog input (which also provides the
//    feedback for the analog PI circuit).  The resulting output voltage is
//    monitored via an analog input.
//
//
//  Changing Polarity
//
//    To avoid possible damage to the load or the power supply, a change in the
//    output that requires changing the polarity is done by 1st ramping the
//    output of the supply to zero, switching the polarity only when the actual
//    output is close to zero, and then ramping the output level to the desired
//    level using the new polarity.
//
//
//    There are 2 different styles of Trim Coil Power Supplies.  The older
//    types use a motor-driven switch to control the relays, while the newer
//    ones use SCRs.
//
//
//    For the older motorized ones, the desired position of the motorized switch
//    is set via a single digital output bit.  The motor moves from one position
//    to the other (Normal to Reverse or Reverse to Normal) based on the digital
//    output.  During the movement, the power supply is disabled for a short
//    time.  The change in the enabled/disabled state of the supply can be read
//    by the software but ONLY when it is asserting the ON command.  So if the
//    supply is not logically on during the change, there is no way to tell
//    if/when the polarity actually changed.
//
//    If such a case, about all the software can do is change the polarity
//    command bit, wait for a short amount of time, and assume that it worked.
//    The only backup is the polarity of the load voltage.  This voltage is
//    read on the load side so its polarity tells us the actual polarity of the
//    current.  But this can only be used when the voltage is large enough that
//    its polarity is stable, which is generally not the case for small
//    currents.
//
//
//    The newer supplies have SCRs connected to the Normal and Reverse polarity
//    relays, controlled by 2 digital outputs on the ZWorld.  For these
//    supplies, the ZWorld is responsible for turning the supply off, changing
//    the state of the switches, and then turning the supply back on.  This is
//    an improvement over the mechanical switch, but can still be a problem
//    since there is no way to verify that an SCR that was conducting actually
//    turned off when we told it to (which can happen if the current through
//    the SCR is large enough).
//
//
//  Interlocks
//
//    There are a number of relays external to the controller board that
//    monitor external conditions and provide interlocks for some of the
//    conditions.  The water flow (to cool the transistors), temperature,
//    and local off button pressed are latched conditions that have to be
//    reset by the controller once they have occured.
//
//
//  Bias Voltage:
//
//    The transistor collector bias voltage is a function of the load current
//    and the variac position.  The controller monitors the bias voltage and
//    adjusts the position of the variac when needed to keep it within
//    specified limits.  If the voltage is too large for too long, it will
//    burn out a transistor, which causes the current through the remaining
//    ones to increase, buring out another, then another, etc. until the
//    current becomes large enough that (sometimes) the fuse for a transistor
//    will burn out before the transistor does (but don't count on it).
//
//    Because the voltage is a function of not just the variac position but
//    the load voltage, rapid changes in the load current will cause the
//    bias voltage to change much faster than it can be adjusted by the
//    variac.  A rapid increase in the load current causes the bias voltage
//    to drop quickly (which is not dangerous to the transistors), while a
//    sudden drop in the current causes the bias voltage to increase.
//
//    The controller code has always monitored the bias voltage so that it
//    could adjust it by changing the position of the variac as needed.  But
//    the latest version now generates warning and error messages based on
//    range and time settings.  In addition, if the error exceeds the set
//    threshold for longer than a specified interval, and does not decrease
//    by at least a specified rate, than an interlock condition is
//    invoked.
//
//-----------------------------------------------------------------------------
//  Determining the state of the polarity switch:
//
//  For the SCR type supplies, we are in control of the "ON" command for both
//  of the SCRs (one for positive/normal polarity, one for negative/reverse).
//  The only real unknown is whether or not an SCR continues to conduct after
//  we remove the "ON" command for it (which it WILL do until the current
//  drops below a certain level).
//
//  In addition, we need to turn off the supply when changing which SCR is ON
//  (to make sure both ARE off before turning one back on).  So it doesn't
//  matter at all whether or not the supply is logically on or off before,
//  during or after the change:  We can do the switch at any time that the
//  current is low enough and be reasonably certain that it worked.
//
//
//  The version with the motor-driven polarity switch is another matter.  For
//  it, we have a single bit that tells the motor which polarity we want the
//  switch to be in.  When the state of that bit is changed, the motor begins
//  to move to the opposite state (a task that takes several seconds), during
//  which it disables and re-enables the supply.  There are no inputs that
//  indicate the state of the motor.  The state can only be determined
//  indirectly by monitoring the state of the supply and the polarity of
//  the load voltage.
//
//  Monitoring the enabled/disabled state of the supply requires that the
//  on command is asserted before and during the change.  And using the
//  polarity of the load voltage requires a significant current to be
//  flowing through the load (and that the calibrations are correct so the
//  readings are reliable).
//
//  The later method would require us to not only have the supply on, but
//  to run a significant current through the load just to determine the
//  state of the switch.  Therefore this method is not a practical means of
//  determining the actual polarity except when there is already a significant
//  level of current flowing through the load.
//
//  Monitoring the on/off state of the supply is is generally sufficient,
//  but only works while we are asserting the ON command to the supply.  If
//  supply is not already on, or we are commanded to turn it off during the
//  change, then we have no confirmation of when the switch completes.
//
//  So, the best we can do for the motor-driven version is this:
//
//    - If we are asserting the ON command all through the change period,
//      then monitor the actual on/off status of the supply, waiting for it
//      to turn off and then back on.
//
//    - If the assertion of the ON command is removed at any time during the
//      change, then wait until enough time has passed since we started the
//      change that we can be reasonably certain the change has finished.
//
//  In both cases, when the change state is over, we assume that the result
//  is that the polarity is what we told it to be.  The only exception is
//  when the load current is high enough to produce a load voltage with a
//  stable polarity, in which case we use that polarity to determine the
//  actual state of the switch.
//
//-----------------------------------------------------------------------------

#include <common.h>

#include <unistd.h>
#include <sys/types.h>


//APPLCODE - change to match the name of this file
#ifndef  ApplicSource
  #define  ApplicSource  "TCPSAppl.c"
#endif

#ifndef DidRegDefs
  #define  DidRegDefs
  #include <modbusRegs.c>
#endif
#ifndef RegDefs


//APPLCODE - change to the application name
const char far applName[] = { "TCPS" };
const char ApplAbbrev[] = { "TCPS" };


//APPLCODE - add ID to unistd.h and place it here
#define DeviceType  DT_TCPS_2009

//APPLCODE - set the # of input and outputs used
#define  NUM_APPL_AIN      4   // analog inputs
#define  NUM_APPL_AOUT     1   // analog outputs
#define  NUM_APPL_DIN     12   // digital inputs
#define  NUM_APPL_DOUT    11   // digital outputs

#define  ADC_MASK_APPL  0x0000000FL
#define  DAC_MASK_APPL  0x00000001L


#define SharedINC
#include <sharedAppl.c>



//=============================================================================
//  definitions related to use of the Digital and Analog I/O signals
//=============================================================================

//APPLCODE - define macros for each Digital bit and each Analog index

//=============== digitalIn bit usage ===============
#define  CTL_POWER_OFF   0x000001  // 00: Outside control power: 0 = ON
#define  WATER_FLOW_BAD  0x000002  // 01: Water Flow: 0 = Water Flow OK
#define  TEMP_BAD        0x000004  // 02: Temperature:  0 = Temperature OK
#define  TRANS_FUSES_BAD 0x000008  // 03: Transister Bank Fuses: 0 = Fuses OK
#define  INTERLOCKS_BAD  0x000010  // 04: Interlocks:  0 = OK
#define  OFF_ASSERTED    0x000020  // 05: On/Off fdback: 0 = ZWorld asserted ON
#define  OVERLOADS_BAD   0x000040  // 06: Primary Overloads:  0 = Overloads OK
#define  MANUAL_OFF      0x000080  // 07: OFF button: 0 = NOT pressed, 1 = pressed
#define  DI_UNUSED_08    0x000100  // 08: unused
#define  DI_UNUSED_09    0x000200  // 09: unused
#define  EXT_INTERLOCK   0x000400  // 10: External Interlock: 1 = Interlock
#define  PS_IS_OFF       0x000800  // 11: PS Status: 0 = PS ON, no interlocks

const uchar  LogicOrderDIn[NUM_APPL_DIN] =  {
   0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11 };


/*EXAMPLE:
#define  ClosedLoop(din)        ( (bool)((din) & CLOSED_LOOP) )
#define  FinalBtnPushed(din)    ( (bool)((din) & FINAL_BTN) )
#define  InitialBtnPushed(din)  ( (bool)((din) & INITIAL_BTN) )
#define  RFisOff(din)           ( (bool)((din) & RF_IS_OFF) )
#define  LocalMode(din)         ( (bool)((din) & LOC_REM_SW) )
*/

//--- msgs to log automatically when a digital input changes state ---
//--- use NULL for bits that we don't want to log changes for ---
const char far *digInDescr[NUM_APPL_DIN*2] = {
//--0 state descr--     --1 state descr--
  "Control Power On",   "Control Power Off",
  "Water Flow OK",      "Water Flow Bad",
  "Temp OK",            "Temp Bad",
  "Fuses OK",           "Fused Bad",
  "Interlocks cleared", "Interlocks asserted",
  "ON bit asserted",    "ON bit cleared",
  "No Overloads",       "Overload Bad",
  "OFF Btn released",   "OFF Btn pressed",
   NULL,                 NULL,
   NULL,                 NULL,
  "Ext Intlk Cleared",  "Ext Intlk Asserted",
  "PS is ON",           "PS is Off"
};


//=============== digitalOut bit usage ===============
#define  VARIAC_UP_OFF      0x000001  // 00: Variac Up: 0 = Move Variac Up
#define  VARIAC_DOWN_OFF    0x000002  // 01: Variac Down: 0 = Move Variac Down
#define  ALLOW_INTLKS       0x000004  // 02: Intlks: 0 pulse = Reset Interlocks
#define  REV_POLARITY       0x000008  // 03* Non-SCR Polarity: 1 = Reversed
#define  TURN_PS_OFF        0x000010  // 04: Power Supply On/Off: 1 = OFF
#define  DO_UNUSED_05       0x000020  // 05: unused
#define  DO_UNUSED_06       0x000040  // 06: unused
#define  DO_UNUSED_07       0x000080  // 07: unused
#define  ActivityLED        0x000100  // 08: activity LED
#define  NORM_POLARITY_OFF  0x000200  // 09* disable Normal Polarity output
#define  REV_POLARITY_OFF   0x000400  // 10* disable Reverse Polarity output
// * NOTE: The "bump" supplies on the K1200 use SCR's that are controlled
//         by the Polarity Relays rather than via a single digital output

const uchar  LogicOrderDOut[NUM_APPL_DOUT] =  {
   0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10 };

//APPLCODE - set to state we want outputs in during initialization
#define InitialDigOutState  (  VARIAC_UP_OFF      \
                             | VARIAC_DOWN_OFF    \
                             | TURN_PS_OFF        \
                             | NORM_POLARITY_OFF  \
                             | REV_POLARITY_OFF   \
                            )

#define  TurnOnActivityLED()    ClearBits(Next.digitalOut, ActivityLED)
#define  TurnOffActivityLED()   SetBits(Next.digitalOut, ActivityLED)

 //--- msgs to log automatically when a digital output changes state ---
//--- use NULL for bits that we don't want to log changes for ---
const char far *digOutDescr[NUM_APPL_DOUT*2] = {
//--0 state descr--       --1 state descr--
  "Start Variac UP",      "Stop Variac UP",        // 00
  "Start Variac Down",    "Stop Variac Down",      // 01
  "Start Reset pulse",    "End Reset pulse",       // 02
  "Move to [+] polarity", "Move to [-] polarity",  // 03
  "Assert PS ON",         "Assert PS OFF",         // 04
   NULL,                   NULL,                   // 05
   NULL,                   NULL,                   // 06
   NULL,                   NULL,                   // 07
   NULL,                   NULL,                   // 08
  "Enable [+} polarity",  "Disable [+] polarity",  // 09
  "Enable [-] polarity",  "Disable [-] polarity"   // 10
};


//=============== Analog inputs ===============
//--- indexes for anaIn[] and readVal[] ---
#define  SHUNT_VOLTS        0   // volt across shunt resistor in series w/load
#define  SUPPLY_VOLTAGE     1   // volt in to Reversing Switch
#define  BIAS_VOLTS         2   // volts in to transistor bank collectors
#define  LOAD_VOLTAGE       3   // volt to load (after relays, shunt resistor)

const uchar  LogicOrderAIn[NUM_APPL_AIN] = { 0, 1, 2, 3 };


//=============== Anlog outputs ===============
//--- indexes for anaOut[] and setVal[] ---
#define  OUTPUT_CTL         0   // setpoint - ctl volt to analog PI circuit

const uchar  LogicOrderAOut[NUM_APPL_AOUT] = { 0 };


//=============================================================================
//  Application-specific data structures and definitions
//=============================================================================

// Each value in the ApplState structure should be throught of as representing
// a "signal" in a clocked logic circuit, with the "applic" task representing
// the logic implemented by the circuit elements such as gates, latches, etc.

// The main "applic" task is the highest priority task and executes the code
// that implements the application logic on a real-time, periodic basis.

//                 ====== IMPORTANT NOTE on LOCKING ======
// Remember to use a semaphore or some other form of locking mechanism to
// protect all access to shared values by more than one task.  The only
// exception to this is when the applic task needs to access a value protected
// by other tasks using only a scheduler/kernel lock (the KLOCK and KUNLOCK
// macros).  Because the applic task is the only task that runs at the highest
// priority, no other task can preempt it, and if the scheduler is locked,
// the applic task cannot preempt another task.  But scheduler locks must be
// kept short to avoid creating large latency in the timely execution of other
// tasks (the applic task in particular).


// ***** WARNING ***** WARNING ***** WARNING ***** WARNING *****
//  Changing the applState structure means that a soft reset will fail!  A
//  change to the size will cause a soft restart to abort and do a CPU
//  restart instead.  And while a change that does NOT result in a new
//  size will not cause the soft restart to abort, it does mean that any
//  shift in the position of values can cause all kinds of mayhem.
// ***** WARNING ***** WARNING ***** WARNING ***** WARNING *****


typedef struct {
 //-----------------------------------------------------------------------
 // SOFTWARE SETTING INPUTS
 //
 //  These values represent Software Inputs to the application logic.
 //
 //  These values are set in the "Next" instance by tasks OTHER than the
 //  "applic" task to change some value used by the application logic.
 //  Remember that all tasks besides the main "applic" task must use a
 //  scheduler lock to protect reads and writes to these values.
 //
 //  Note that the application task may alter these values (e.g. to change
 //  an invalid value in to a valid one).
 //-----------------------------------------------------------------------
  bool   onRequest;
  bool   offRequest;
  bool   resetRequest;

  float  ampSet;          // most recent setpoint


 //-----------------------------------------------------------------------
 // READINGS
 //
 //  These values represent "passive" inputs to the application logic.
 //
 //  These values are set in the "Next" instance by the applic task just
 //  after reading the state of the hardware inputs.
 //
 //  These values generally come from one of two sources:
 //    - The just-read state of analog and/or digital hardware inputs
 //    - Values maintained by other tasks
 //
 //  While some of these values are simply copies of "external" values,
 //  including them in the ApplState structure allows us to see how
 //  they change between cycles.  Also, some of these values are the result
 //  of applying logic to "external" values.  Applying that logic once at
 //  at the beginning of the applic loop and saving the result eliminates
 //  the need to repeat the same logic in multiple blocks of code later on,
 //  avoiding unnecessary redundancy.
 //-----------------------------------------------------------------------
  ulong  processTime;     // time (in ms) when processing was started
  ulong  digitalIn;       // state of digital inputs

 //-----------------------------------------------------------------------
 // OUTPUTS from UN-CLOCKED (asynchronous, gate-level) LOGIC ELEMENTS
 //
 //  These values are outputs from one of the blocks of code in the applic
 //  task that represent a non-clocked logic element (such as a gate).
 //
 //  Each value here is set by one (and ONLY one) of the blocks of code in
 //  the applic task.  This group contains ONLY values that are set by a
 //  block of code that represents a NON-clocked logic element.
 //
 //  These values are set in the "Next" instance by the applic task just
 //  after reading the state of the hardware inputs.  They must NOT be
 //  altered by other tasks (although they can read them as long as they
 //  use appropriate locking to do so).
 //
 //  Because these values represent the outputs of un-clocked logic elements,
 //  and such elements often need to be cascaded (the outputs used as inputs
 //  to other non-clocked elements), the result is often dependent on the
 //  order of execution of the blocks of code.
 //-----------------------------------------------------------------------
  bool   ctlPowerOn;       // outside control power is ON
  bool   waterFlowOK;      // Water Flow is OK
  bool   tempOK;           // Temp is OK
  bool   fusesOK;          // Fuses are OK
  bool   interlocks;       // Interlock asserted
  bool   onIsAsserted;     // On/Off output bit reads ON
  bool   overloadsOK;      // No Overloads asserted
  bool   offPressed;       // Off button is being pressed
  bool   externIntlk;      // External Interlock is being asserted
  bool   biasWarn;         // Bias error long enough to warrant a warning
  bool   biasIntlk;        // Bias error long enough to warrant an interlock
  bool   psIsOn;           // PS is ON

  bool   polarityErr;      // is load polarity what it should be

  float  biasError;        // diff between setpoint and actual bias volts
  float  biasErrorABS;     // diff between setpoint and actual bias volts

  int    ampSetPolarity;   // polarity of current ampSet
  bool   needPolarityChg;  // need to chg the state of the polarity switch

  int    switchPolarity;   // which polarity mode is the switch in
  int    loadPolarity;     // polarity of current through the load

 //-----------------------------------------------------------------------
 // OUTPUTS from CLOCKED (latched inputs and outputs) LOGIC ELEMENTS
 //
 //  These values are outputs from one of the blocks of code in the applic
 //  task that represent a clocked logic element (such as latches).
 //
 //  Each value here is set by one (and ONLY one) of the blocks of code in
 //  the applic task.  This group contains only values that are set by a
 //  block of code that represents a CLOCKED logic element.
 //
 //  These values are set in the "Next" instance by the applic task AFTER
 //  all the other ApplState values are set.  Because they represent only
 //  the clocked logic elements, and each block of code reads only values
 //  from the Cur instance, the blocks can be executed in any order and the
 //  result will always be the same.
 //
 //  These values are used as inputs by other blocks and/or to set values
 //  used by other tasks (such as values read by the ModBus server task).
 //
 //  Again, these values must NOT be altered by other tasks (although they
 //  can read them as long as they use appropriate locking to do so).
 //-----------------------------------------------------------------------
  ulong  digitalOut;      // state for digital outputs

  bool   assertReset;     // assert reset cmd for error/"BAD" inputs
  ulong  resetStarted;    // time resetul pulse was initiated

  int    state;           // state of application logic
  ulong  rtzStarted;      // time (in ms) started ramping to 0 before going off

  int    switchState;     // state of polarity switch
  int    assertPolarity;  // -1, 0, or 1  (0 value applies only to SCR type)
  ulong  switchStarted;   // time (in ms) when a chg in polarity was started

  bool  assertOn;         // assert ON cmd for PS

  float  ampApply;        // value to apply to regulation circuit

  bool  adjustBias;       // move the variac to adjust the bias voltage
  bool  incrBiasVolts;    // increase the bias voltage
  bool  decrBiasVolts;    // decrease the bias voltage

} ApplState;


ApplState  prevState, curState, nextState;

#define  Prev  prevState
#define  Cur   curState
#define  Next  nextState


const uint ApplStateDataSize =  ( sizeof(ApplState)*3 + sizeof(setVal) );


//===== "cur" values that we don't need to compare to "prev" instance =====
//----- (so we don't need them to be in the ApplState structure) -----
float  lastRampRate, maxAmpChg;

ulong  polarityErrStartTime, biasAdjustTime, biasErrTime;
bool  loggedPolarityErr;

ulong  lastBiasVoltsChkMS;
float  avgdBiasVolts, lastAvgdBiasVolts, biasVoltsChg, prevBiasVoltsErr,
       biasVoltsChgRate;
bool   ignoreBiasErr;

uint  lastValLogTime, logInterval;





//----- state values -----
#define  STATE_OFF     0   // PS off, idle
#define  STATE_ON      1   // PS on or polarity switch in progress
#define  STATE_INTLK   2   // error occured that needs acknowledgement

const char *StateDesc[] = { "OFF    ", "ON     ", "INTERLK" };


//----- switchState values -----
#define  SW_UNKNOWN       0  // unknown state (requires waiting a fixed period)
#define  SW_OFF           1  // load disconnected (SCR type only)
#define  SW_ON            2  // on in either positive or negative state
#define  SW_WAIT_FOR_OFF  3  // ON-to-OFF phase of polarity change
#define  SW_WAIT_FOR_ON   4  // OFF-to-ON phase of polarity change
#define  SW_WAIT_FOR_ZERO 5  // wait for cur to drop to safe switching level

const char *SwitchStateDesc[] = { "UNKNOWN      ",
                                  "OFF          ",
                                  "ON           ",
                                  "Wait-for-OFF ",
                                  "Wait-for-ON  ",
                                  "Wait-for-Zero",
                                };



//=============================================================================
//  List of daemon processes.  The inetd process listens for connection
//  requests on each port in the list and creates a new instance of the
//  specified process to service each new connection.
//=============================================================================
#include <services.h>

//APPLCODE - add tasks that service network clients

//--- service funcs declared the same as standard command funcs:
//     e.g.:   void _task fastdatad(int argc, char **argv);

void _task tsh(int argc, char **argv);


ServiceInfo inetdConf[] = {
//--- add any application-specific inetd-type network services here ---
//--- (add port # names to services.h) ---
//  NAME          PORT           PROTOCOL  FUNC       PRI  -1  Config Params
//  ------------  -------------  --------  ---------- ---  --  ----------------

//--- standard ones defined in sharedAppl.c ---
//{ "console",    TTY0,          ASYNC,    tsh,        10, -1, &consCfg },
//{ "fastdatad",  FASTDATA_PORT, ProtoTCP, fastdatad,   5, -1, NULL },
//{ "modbusTCPd", MODBUS_PORT,   ProtoTCP, modbusTCPd,  6, -1, NULL },
//{ "tsh",        TELNET_PORT,   ProtoTCP, tsh,        10, -1, NULL }

  #define SharedServices
  #include <sharedAppl.c>
};

const int  NumServices = sizeof(inetdConf) / sizeof(ServiceInfo);





//=============================================================================
//  Defintions for variables that are accessable as ModBus registers
//=============================================================================
#else  // end of before-RegDefs portion

/*
  The purpose of this section is to define a set of C variables that are
  accessable as ModBus registers.  This approach avoids the need to use
  special functions or macros throughout the code when accessing register
  values, which keeps the application code fast and simple.

  The model used provides support for multiple logical devices.  The 2nd
  parameter in each register definition macro indicates if one instanace
  applies to all logical devices (specified by an "x" for "don't care"), or
  if each logical device requires its own instance (specified by an under-
  score "_" to signify "fill in the blank").  In the 1st case, the variable
  is defined as a scalar.  In the 2nd case, it is defined as an array.

  The definition macros support data types (floats and longs in particular)
  that require more than one register to store them.  To avoid the possibility
  of corrupted values, all reads of and writes to such values must be done
  using the multi-register commands and including all the registers that
  make up the value in the command.  Differences in register order between
  different client types is handled by a table that automatically swaps
  register values as needed based on the client type.

  The definition for each writable register also includes the function to
  call when the value is set via the ModBus interface.  Many registers can
  use the default setXXX() functions, but you can specify your own to provide
  any special logic needed (including ignoring the new setting if invalid).

  For registers, the format for the function definitions is:

    bool mySetFunc(SetRegFuncParams);

       where SetRegFuncParams is currently defined as:
              void *info, uchar chanNum, void *newVal, fd_t eventsfd

  For coils, the format for the function definitions is:

     bool (*SetCoilFunc)(SetCoilFuncParams);

         where SetCoilFuncParams is currently define as:
               void *info, uchar chanNum, uchar newVal, fd_t eventsfd

  The parameters to the call are as follows:
    *info:     A pointer to the Coil's/Register's CoilInfo/RegInfo structure
    chanNum:   The logical device # (0 is the 1st)
    *newVal:   A pointer to the new value requested in the Modbus packet
               (changed to native-byte order where needed)
    eventsfd:  The fd for the FIFO used to send msgs to the main application
               process

  The function returns true if the new value was accepted, false if not.

  The register numbers for multiple "channels" or logical devices are
  handled automatically based on the xxChanMult values.  These values specify
  the offset between a register value for one logical device and the "same"
  register for the Next.

  Example:

    Assume the following definitions:
      #define  NumLogicalDevices  2
      #define  R4ChanMult  100
      REG4( motorPos,       _, 40050, F32, RW, setMotorPos )

    Then we would get:
      global variable definition:  float motorPos[2];
      register pair:  40050,40051  refers to:  motorPos[0]
      register pair:  40150.40151  refers to:  motorPos[1]

    If instead we had specified:
      REG4( motorPos,       x, 40050, F32, RW, setMotorPos )

    Then we would get:
      global variable definition:  float motorPos;
      register pairs:  40050,40051 & 40150,40151  BOTH refer to:  motorPos


  To keep the definitions as simple as possible, and avoid requiring multiple
  macros for each register definition, each macro is used more than once,
  with a different definition (#define) each time (a practice generally
  referred to as X-Macros).  While trying to follow the order of inclusion
  of files (including a recursive reference back to this file) can be rather
  confusing, the X-Macro method keeps the source code simple to read and
  maintain by doing everything with a single macro for each register.

  ==========================================================================
                             !!! WARNING !!!
  ==========================================================================
    Each sequence of register definitions MUST be contiguous!  There must
    NOT be any holes (i.e. no unused register #'s between used ones) and no
    overlap!
-----------------------------------------------------------------------------*/


//APPLCODE - set # of devices for which we want duplicate register sets
#ifndef NumLogicalDevices
  #define  NumLogicalDevices  1
#endif

//--------------------------------------------------------------
//APPLCODE - set max # of registers per set of coils
#ifndef NumC0Chans
  #define NumC0Chans_  NumLogicalDevices
  #define C0ChanMult_  100
#endif

#ifdef COIL0

COIL0( DoReboot,       x, 1, setCoil )
//COIL0( Coil00002,      _, 2, setCoil )
//COIL0( Coil00003,      _, 3, setCoil )

#undef COIL0
#endif


//--------------------------------------------------------------
//APPLCODE - set max # of registers per set of read-only registers
#ifndef NumR3Chans
  #define NumR3Chans_  NumLogicalDevices
  #define R3ChanMult_  100
#endif

#ifdef REG3

REG3( dlmVersTime,    x, 30001, U32 )
REG3( rtosVersTime,   x, 30003, U32 )
REG3( applVersTime,   x, 30005, U32 )
REG3( devType_,       x, 30007, U16 )
REG3( deviceID_,      x, 30008, U32 )
REG3( upSecs,         x, 30010, U32 )
REG3( percentIdle_,   x, 30012, U16 )

REG3( numChans,       x, 30013, U16 )
REG3( coilChanMult,   x, 30014, U16 )
REG3( reg3ChanMult,   x, 30015, U16 )
REG3( reg4ChanMult,   x, 30016, U16 )

REG3( digInBits,      x, 30017, U32 )

REG3( statusBitsStdA, x, 30019, U16 )
REG3( statusBitsStdB, x, 30020, U16 )
REG3( statusBitsAppl, x, 30021, U32 )

REG3( latency,        x, 30023, U16 )  // actual vs scheduled appl start in ms
REG3( tcpTickTime,    x, 30024, U16 )  // # ms inside tcp_tick() call
REG3( applLoopTime,   x, 30025, U16 )  // # ms used by applic task loop
REG3( kernelLoopTime, x, 30026, U16 )  // # ms used by last kernel pass


//APPLCODE - add registers for externally available read-only values

#undef REG3
#endif


#ifdef RegMacros
//===== application-specific defintions related to register values =====

//  #define  READING(idx)  REG_ARRAY(float, rdg_1, idx)

#endif


//--------------------------------------------------------------
//APPLCODE - set max # of registers per set of read/write registers
#ifndef NumR4Chans
  #define NumR4Chans_  NumLogicalDevices
  #define R4ChanMult_  100
#endif

/*
  The "standard" functions for processing writes to Modbus register values:

    setS16, setU16, setS32, setU32, setF32

  These functions simply log a message for the change and update the
  value in a critical block (so the change cannot be interrupted).

  In many cases, just allowing any setting to be written to a register (or
  multiple registers comprising a larger value) is not acceptable.


  If any of the following apply, then a custom func must be used instead:

    - The order of changes is important.  For example, if you get ON,
      OFF, and RESET commands in rapid succession, you have to be sure
      they are processed in the order they were recieved.

    - Not all possible settings are acceptable.  The acceptable settings
      for a value are often a subset of the possible ones, and frequently
      are dependent upon the current state of the controller.


  IMPORTANT NOTES:

    The "applic" task periodically executes the code that makes up the
    bulk of the application logic.  Because it runs at the highest priority
    there is no chance that a lower priority task (like the daemon handling
    Modbus register writes) can change a value in the middle of this
    processing.

    However, the applic task CAN (and often does) preempt other processes,
    so any process that does read or make changes to the register values
    must lock the register values before doing so.  This is currently done
    by disabling the scheduler using the KLOCK() and KUNLOCK() macros.

    Use of such a lock insures that changes occur only between passes
    through the application logic, but there are still several issues
    that must be addressed when writing functions to handle writes to
    register values:

      - Priority inversion and latency:  Allowing a lower priority task
        to lock out higher ones (or, in this case, ALL tasks) allows
        the lower priority task to run when a higher-priority one should
        be running.  Care must be taken to keep such periods as short
        as possible to avoid disrupting the timing of the control logic.

      - Loss of encapsulation:  The last thing we want is to break the
        application logic in to small pieces spread through a bunch of
        functions executed by different tasks at different times.

        While the logic for processing writes to register values is often
        dependent upon the state of the control logic, functions outside
        the applic task must NOT directly alter the state of the application
        logic.  To achieve this, such functions must limit changes to
        values that act as input to the applic task.

        To this end, you need to treat the register values that act as
        inputs just like they were knobs, switches, analog inputs, or
        latches.  Like their hardware equivalent, the applic task logic
        reads their state at the beginning of each pass and uses this
        information when determining what the next state of the control
        logic should be.  And just like a hardware latch, values can be
        reset after reading them so the next write to the value can be
        treated as another instance of a request.

    The trickiest circumstances to handle are ones where the order of separate
    writes is important.  The most common example are commands for ON, OFF,
    and RESET.  If, for example, you had 3 separate flags, and simply set
    each once corresponding to which of the 3 commands you received, you
    end up with multiple flags set when the applic logic is processed, and
    no way to determine what the end result should be.  In these cases,
    the logic that controls the state of each flag should be placed in its
    own function and called during the processing of any event that could
    affect the flag's state.

    Ideally, the logic for setting each variable would be triggered each
    time an event occurs that could affect it:  A publish-subscribe type
    of mechanism.  One possibility to do this efficiently?  Use a bit-mask
    "array" to determine which functions to call for each register that is
    changed.  That way it doesn't take many bytes to determine which functions
    to call, and each function can update its corresponding register value
    based on its current state and the event that just occured.

    Or should I simply combine such values in to one (a struct, if necessary),
    and have a single function that is called for all the related values
    and updates the state of the single logical entity based on its
    current state and what event called it?  (But the # of combinations could
    get out of hand easily...)
*/

#ifdef REG4

REG4( timeZone_,        x, 40001, S16, RW, MB_timeZone )
REG4( curTime,          x, 40002, U32, RW, setU32 )

REG4( ipAddr,           x, 40004, U32, RW, setU32 )
REG4( netmask,          x, 40006, U32, RW, setU32 )
REG4( gateway,          x, 40008, U32, RW, setU32 )

REG4( setACoil,         _, 40010, U16, RW, setU16 )
REG4( resetACoil,       _, 40011, U16, RW, setU16 )

REG4( digOutBits,       x, 40012, U32, RW, setU32 )

REG4( cmdBitsStd,       x, 40014, U32, RW, MB_cmdBitsStd )
REG4( cmdBitsAppl,      x, 40016, U32, RW, MB_cmdBitsAppl )

//APPLCODE - add registers for externally available read/write values
REG4( ampSet,           x, 40018, F32, RW, MB_ampSet )  /* desired output */

REG4( switchType,       x, 40020, U16, RW, setU16 )  /* relay or SCR switch */
REG4( devStatus,        x, 40021, U16, RO, setU16 )  /* bit-mapped status */

REG4( loadAmps,         x, 40022, F32, RO, setF32 )  /* Amps thru shunt measured via ADC */
REG4( loadVolts,        x, 40024, F32, RO, setF32 )  /* volt to load measured via ADC */
REG4( psVolts,          x, 40026, F32, RO, setF32 )  /* PS output volt measured via ADC */

REG4( maxVolts,         x, 40028, F32, RW, setF32 )  /* max voltage allowed */
REG4( maxAmps,          x, 40030, F32, RW, setF32 )  /* max current allowed */

REG4( safeSwAmps,       x, 40032, F32, RW, setF32 )  /* max A allowed for polarity swap */
REG4( connectLoadDelay, x, 40034, U32, RW, setU32 )  /* PS On --> load connected delay */

REG4( biasVolts,        x, 40036, F32, RO, setF32 )  /* transistor bank collector bias */

REG4( biasVoltsSet,     x, 40038, F32, RW, setF32 )  /* desired transistor bank bias V */
REG4( biasErrorStart,   x, 40040, F32, RW, setF32 )  /* threshold to start correcting */
REG4( biasErrorStop,    x, 40042, F32, RW, setF32 )  /* threshold to stop correcting */

REG4( rampRate,         x, 40044, F32, RW, setF32 )  /* setpoint chg rate in amps/sec */

REG4( switchTime,       x, 40046, U32, RW, setU32 )  /* # ms req for polarity switch */

REG4( polThreshAmps,    x, 40048, F32, RW, setF32_ABS )  /* min load cur for reading polarity */
REG4( polThreshVolts,   x, 40050, F32, RW, setF32_ABS )  /* min load volts for reading polarity */

REG4( biasWarnMS,       x, 40052, U16, RW, setU16 )  /* max ms before warn msg */
REG4( biasErrMS,        x, 40053, U16, RW, setU16 )  /* max ms before interlock */

REG4( debounceMS,       x, 40054, U16, RW, setU16 )  /* ms period for digIn debouncing */

REG4( maxRTZms,         x, 40055, U32, RW, setU32 )  /* max time to Ramp-To-0 before off */

REG4( minBiasChgRate,   x, 40057, F32, RW, setF32_ABS)  /* min chg rate for bias v corrections */

#undef REG4
#endif


#ifdef RegMacros
//===== application-specific defintions related to register values =====

//  #define  ISET(idx)   REG_ARRAY(float, iSet_1, idx)

#endif

//--------------------------------------------------------------
#undef RegMacros

#endif
#ifndef RegDefs // start of after-RegDefs portion


//--- definitions for cmdBitsAppl ---
//#define  CBA_DoReboot   0x00000001  /* reboot ctlr */
//#define  CBA_DoInit     0x00000002  /* tell ctlr to (re)initialize */
#define  CBA_TurnOn     0x00000100  /* connect PS to load, turn it On */
#define  CBA_TurnOff    0x00000200  /* disconnect from load, turn it Off */
#define  CBA_Reset      0x00000400  /* clear Interlocks */


//--- bit usage for switchType ---
#define  MOTOR_SWITCH    1  // polarity controlled via relays
#define  SCR_SWITCH      2  // polarity controlled via SCRs

//finish
//--- bit usage for devStatus --- (some bits based on old TrimCoil PS device) ---
#define  DS_INTLK      0x0001   // Interlock condition active
#define  DS_ON         0x0002   // Controller is in ON state
#define  DS_NORM_POL   0x0004   // Normal (+) Polarity state in effect
#define  DS_RTZ        0x0008   // Returning-To-Zero before shutting off
#define  DS_PS_IS_OFF  0x0010   // PS is not ON (as of 2011-02-08)

#define  DS_SER_ERR    0x0100   // Serial PS not responding
#define  DS_CALIB_ERR  0x0200   // serial & analog readings vastly different



//=============================================================================
//  defintions for "parameterized" data values - lists of values that we want
//  to be "persistent" (i.e. saved to a file and restored on startup)
//=============================================================================
//APPLCODE - settings/values we want saved to/restored from files


//--- parameter ID's for data in Device Data/calibration file ---
//--------------------------------------------------------------------------
// WARNING:  ID's can be added, but existing ID's (and the format of the
//           data they represent) should NEVER change!  NOTE:  Existing ones
//           can be "eliminated" by changing their name and no longer
//           using them, but never REUSE an old one for something new!!!
//--------------------------------------------------------------------------
//  NOTE:  ID's >= 1000 are rsvd for values shared by most applications
//--------------------------------------------------------------------------

const ParamData far  devDataParams[] = {
//APPLDATA
//                ID   ---value---  --RESTORE_FUNC--
//PERSISTENT_VAR(  1,  myVarName,     NULL ),
//                ID   ---value---  granularity  --RESTORE_FUNC--
//PERSISTENT_TBL(  2   myArrayName, elemSize,     NULL ),

  PERSISTENT_VAR(  1,  switchType,       NULL ),
  PERSISTENT_VAR(  2,  maxVolts,         NULL ),
  PERSISTENT_VAR(  3,  maxAmps,          NULL ),
  PERSISTENT_VAR(  4,  safeSwAmps,       NULL ),
  PERSISTENT_VAR(  5,  connectLoadDelay, NULL ),
  PERSISTENT_VAR(  6,  biasVoltsSet,     NULL ),
  PERSISTENT_VAR(  7,  biasErrorStart,   NULL ),
  PERSISTENT_VAR(  8,  biasErrorStop,    NULL ),
  PERSISTENT_VAR(  9,  rampRate,         NULL ),
  PERSISTENT_VAR( 10,  switchTime,       NULL ),
  PERSISTENT_VAR( 11,  polThreshAmps,    NULL ),
  PERSISTENT_VAR( 12,  biasWarnMS,       NULL ),
  PERSISTENT_VAR( 13,  biasErrMS,        NULL ),
  PERSISTENT_VAR( 14,  polThreshVolts,   NULL ),
  PERSISTENT_VAR( 15,  debounceMS,       NULL ),
  PERSISTENT_VAR( 16,  maxRTZms,         NULL ),
  PERSISTENT_VAR( 17,  minBiasChgRate,   NULL ),

  #define SharedDevParams
  #include <sharedAppl.c>

  END_PARAM_DATA
};

//----------------------------------------------------
RESTORE_FUNC(ampSet);

const ParamData far  settingsParams[] = {
//APPLDATA
//                ID   ---value---  --RESTORE_FUNC--
//  PERSISTENT_VAR(  1,  rfOffSetpoint,   NULL ),
//  PERSISTENT_VAR(  2,  rfOnSetpoint,    NULL ),

  PERSISTENT_VAR(  1,  ampSet,        PVRF(ampSet) ),

  END_PARAM_DATA
};


//-----------------------------------------------------------------
//  Application-specific values needed to restore the state of the
//  controller following a "soft" reset (i.e. a restart without a
//  CPU reset)
//-----------------------------------------------------------------
const ParamData far  ctlrStateParams[] = {
// sample values from sharedAppl.c, common to all applications
//                   --ID--  ---value,granularity---  --RESTORE_FUNC--
//    PERSISTENT_VARx(  1000,  prevState,               NULL ),
//    PERSISTENT_VARx(  1001,  curState,                NULL ),
//    PERSISTENT_VARx(  1002,  nextState,               NULL ),
//    PERSISTENT_TBL (  1003,  setVal,4,                NULL ),
// // PERSISTENT_TBL (  1004,  shadowRegValues,4,       NULL ),

//APPLDATA
//    PERSISTENT_VAR(  1,  myApplStateVar,        NULL ),


  #define SharedCtlrStateParams
  #include <sharedAppl.c>

  END_PARAM_DATA
};


//-----------------------------------------------------------------------------
//  A list of parameterized "persistent" data.  Backup copies of such values
//  are stored in files and restored on startup.
//-----------------------------------------------------------------------------

//    Base File Name    Category  ParamData
//    ----------------  --------  ---------
//--- add any application-specific parameterized file defs here ---
//ParamFile  myAppCfgFile =
//  { "myFileName",     PDC_CTLR, myCfgParams,   0 };

ParamFile  settingsFile =
    { "settings",       PDC_SET,  settingsParams  };


#define SharedParams
#include <sharedAppl.c>


ParamFile *paramFiles[] = {
  #define SharedParamsList
  #include <sharedAppl.c>
//--- add any application-specific ParamData references here ---
//  &myAppCfgFile,
  &settingsFile
};

const int NumParamFiles = sizeof(paramFiles) / sizeof(paramFiles[0]);


//=============================================================================
//  Declarations for functions that handle setting "next" values that serve
//  as inputs to the application control logic (run by the "applic" task in
//  a periodically executed loop).
//
//  While care must be taken in these functions to use locks to avoid the
//  possibility of being preempted in the middle of an operation, such locks
//  must be kept to as short a time as possible to avoid excessively latency
//  in the main control loop.
//=============================================================================

void onRequest(void);
void offRequest(void);
void resetRequest(void);

void newSetpoint(float newAmps, bool logIt);



//=============================================================================
//  This is where misc global variables should be defined
//=============================================================================
#define SharedGlobals
#include <sharedAppl.c>

//APPLCODE
//--- add definitions for application-specific global variables here ---


//=============================================================================
//  Functions referenced in the Modbus register definitions to handle writes to
//  specific global values defined as ModBus registers
//=============================================================================

//APPLCODE - add functions to process writes to ModBus register values

//-----------------------------------------------------------------------------
//  External ModBus client sent a new value for "cmdBitsAppl"
//
//  Check the bits that where set and trigger the corresponding events as
//  appropriate (but do NOT save the new value - just use it to trigger
//  events)
//-----------------------------------------------------------------------------
bool _nearcall MB_cmdBitsAppl(SetRegFuncParams)
{
  ulong  cmdBits = MB_VAL(ulong);

  if (cmdBits & CBA_TurnOn)    onRequest();
  if (cmdBits & CBA_TurnOff)   offRequest();
  if (cmdBits & CBA_Reset)     resetRequest();

  cmdBitsAppl = 0;

  return true;
}

//-----------------------------------------------------------------------------
//  External ModBus client sent a new setpoint
//
//  Check the new setting and alter output values as appropriate
//-----------------------------------------------------------------------------
bool MB_ampSet(SetRegFuncParams)
{
  newSetpoint(MB_VAL(float), false);

  return true;
}


//=============================================================================
//  functions to handle restoration of saved "persistent" values
//=============================================================================

//-----------------------------------------------------------------------------
//  restore an ampSet value from the settings file
//-----------------------------------------------------------------------------
RESTORE_FUNC(ampSet)
{
  newSetpoint(NEW_VAL(float), true);
}


/*
//-----------------------------------------------------------------------------
//  one of the setpoint values needs restoring from the settings file
//-----------------------------------------------------------------------------
RESTORE_FUNC(myVal)
{
  uint  u;
  float  fval;
  char  temp[32];

  for (u=0; u<NumPS; ++u)  if (curVal == (void far *) &ISET(u))  {
    fval = *(float far *)newVal;
    printf("restoring iSet_%u to: %s\r\n", u, fmtFloat(fval, 8,2, temp));
    newSetpoint(fval, true);
  }

}

*/

//-----------------------------------------------------------------------------


//-----------------------------------------------------------------------------
#define SharedFuncs
#include <sharedAppl.c>


//=============================================================================
//  Misc Application-specific functions
//=============================================================================
//APPLCODE



//=============================================================================
//  Task functions to handle application-specific commands
//=============================================================================
//APPLCODE

//-----------------------------------------------------------------------------
//  example func to process a command entered via a user shell
//-----------------------------------------------------------------------------
/*
void _task miscTest(int argc, char **argv)
{
  int  idx;
  bool  validCmd;

//used by EVENT macro
  fd_t  eventsfd = -1;
  EventMsg  eventMsg;


  validCmd = false;

  for (;;)  {
    if (argc < 3)  break;
    if (! isdigit(argv[1][0]))  break;

    idx = atoi(argv[1]);
    if ((idx < 1) OR (idx > NumChan))  {
      printf("Invalid chan #\n");  Exit(-2); }

    --idx;  validCmd = true;

    if (StriEq(argv[2], "x"))     { EVENT(MyEventX, idx, NULL, 0);  break; }
    if (StriEq(argv[2], "y"))     { EVENT(MyEventY, idx, NULL, 0);  break; }

    validCmd = false;  break;
  }

  if (! validCmd)  { printf("Usage:  misc # x|y\r\n");  Exit(-1); }

  Exit(0);
}
*/

//=============================================================================
//  Generic structure to handle application-specific presentation of the
//  digital and analog input/output values.  Code to handle additional
//  formatting that cannot be handled in this generic manner should be
//  placed in the showState() task routine.
//=============================================================================
//APPLCODE - text and formatting info for application state summary

//-----------------------------------------------------------------------------
_nearcall char *fmtSwType(DispElem far *elem, char *buf)
{
  switch (switchType)  {
    case MOTOR_SWITCH:  return  BlueText"MOTOR";
    case SCR_SWITCH:    return  BlueText"SCR  ";
  }

  return RedText"?????";
}

//-----------------------------------------------------------------------------
_nearcall char *fmtSWState(DispElem far *elem, char *buf)
{
  return (char *) SwitchStateDesc[ELEM_VAL(U16)];
}

//-----------------------------------------------------------------------------
_nearcall char *fmtState(DispElem far *elem, char *buf)
{
  if (Cur.rtzStarted)  return   RedText"RTZ  ";

  switch (Cur.state)  {
    case STATE_OFF:    return   RedText"OFF  ";
    case STATE_ON:     return  BlueText"ON   ";
    case STATE_INTLK:  return   RedText"INTLK";
  }

  return RedText"?????";
}

//-----------------------------------------------------------------------------
_nearcall char *fmtPolState(DispElem far *elem, char *buf)
{
  bool  posPol, negPol;

  switch (switchType)  {
    case MOTOR_SWITCH:
      return (Cur.digitalOut & REV_POLARITY) ? BlueText"[-]" : BlueText"[+]";

    case SCR_SWITCH:
      posPol = !(Cur.digitalOut & NORM_POLARITY_OFF);
      negPol = !(Cur.digitalOut & REV_POLARITY_OFF);
      if (posPol AND !negPol)   return BlueText"[+]";
      if (!posPol AND negPol)   return BlueText"[-]";
      if (posPol AND negPol)    return RedText"err";
      return "n/a";
  }

  return RedText"???";
}


//-----------------------------------------------------------------------------
  const char far DevStateStr[] = {
"---- Digital In ----     -- Digital Out --             ----- ADC Volts ----\r\n"
"Extern Power: ###        Variac Up:   ###              Shunt Volts: -##.###\r\n"
"Water Flow:   ###        Variac Down: ###              Supply Volt: -##.###\r\n"
"Temperature:  ###        Reset Intlk: ###              Trans.BankV: -##.###\r\n"
"Trans. Fuses: ###        Polarity:    ###              Output Volt: -##.###\r\n"
"Interlocks:   ###        Assert ON:   ###              ----- DAC Volts ----\r\n"
"ON Asserted:  ###                                      Output Ctl:  -##.###\r\n"
"Overloads:    ###        ---- Transistor Bias ----\r\n"
"OFF Button:   #######    Reading:         -##.## V     ------- Scaled -------\r\n"
"                         Setpoint:        -##.## V     Setpoint:   -###.### A\r\n"
"ExtInterlock: ###        Error:           -##.## V     Shunt Curr: -###.### A\r\n"
"PS Status:    ###        ---- Error Threshold ----     PS Voltage: -###.### V\r\n"
"                         Start Correction:-##.## V     Output Volt:-###.### V\r\n"
"Sw.Type:   #####         Stop Correction: -##.## V\r\n"
"Sw.State:  #############\r\n"
"Ctl.State: #####\r\n"
" \r\n"
};

const DispElem  far dispElems[] = {
  { 38, 7, VFPV(digitalOut),        "", fmtPolState,      3,3 },

  { 42,11, VFPV(biasVolts),         DARK_BLUE, fmtF32,    6,2 },
  { 42,12, VFPV(biasVoltsSet),      DARK_BLUE, fmtF32,    6,2 },
  { 42,13, VFPV(Cur.biasError),     DARK_BLUE, fmtF32,    6,2 },

  { 42,15, VFPV(biasErrorStart),    DARK_BLUE, fmtF32,    6,2 },
  { 42,16, VFPV(biasErrorStop),     DARK_BLUE, fmtF32,    6,2 },

  { 68, 4, VFPV(anaIn[SHUNT_VOLTS]),    DARK_BLUE, fmtF32,    7,3 },
  { 68, 5, VFPV(anaIn[SUPPLY_VOLTAGE]), DARK_BLUE, fmtF32,    7,3 },
  { 68, 6, VFPV(anaIn[BIAS_VOLTS]),     DARK_BLUE, fmtF32,    7,3 },
  { 68, 7, VFPV(anaIn[LOAD_VOLTAGE]),   DARK_BLUE, fmtF32,    7,3 },

  { 68, 9, VFPV(anaOut[OUTPUT_CTL]),    DARK_BLUE, fmtF32,    7,3 },

  { 67,12, VFPV(ampSet),            DARK_BLUE, fmtF32,    8,3 },
  { 67,13, VFPV(loadAmps),          DARK_BLUE, fmtF32,    8,3 },
  { 67,14, VFPV(psVolts),           DARK_BLUE, fmtF32,    8,3 },
  { 67,15, VFPV(loadVolts),         DARK_BLUE, fmtF32,    8,3 },

  { 11,16, VFPV(switchType),         "",       fmtSwType,  5,5 },
  { 11,17, VFPV(Cur.switchState),   DARK_BLUE, fmtSWState, 12,12 },
  { 11,18, VFPV(Cur.state),         "",        fmtState,   5,5 },

  { 0,0, NULL, NULL, NULL, 0,0 }
};

//--- ATTR_NAME macros currently defined in fcntl.h ---
const DigDispElem far digInElems[] = {
//  xx,yy,  BIT_MASK,       "0STR"  ,ATTR_NAME,  "1STR"  ,ATTR_NAME

  { 14, 4,  CTL_POWER_OFF,   "ON ",DARK_BLUE,     "OFF",DARK_RED   },
  { 14, 5,  WATER_FLOW_BAD,  "OK ",DARK_BLUE,     "BAD",DARK_RED   },
  { 14, 6,  TEMP_BAD,        "OK ",DARK_BLUE,     "BAD",DARK_RED   },
  { 14, 7,  TRANS_FUSES_BAD, "OK ",DARK_BLUE,     "BAD",DARK_RED   },
  { 14, 8,  INTERLOCKS_BAD,  "OK ",DARK_BLUE,     "BAD",DARK_RED   },
  { 14, 9,  OFF_ASSERTED,    "YES",DARK_BLUE,     "NO ",DARK_RED   },
  { 14,10,  OVERLOADS_BAD,   "OK ",DARK_BLUE,     "BAD",DARK_RED   },
  { 14,11,  MANUAL_OFF,      "NORMAL ",DARK_BLUE, "PRESSED",DARK_RED   },

  { 14,13,  EXT_INTERLOCK,   "NO ",DARK_BLUE,     "YES",DARK_RED   },
  { 14,14,  PS_IS_OFF,       "ON ",DARK_BLUE,     "OFF",DARK_RED   },

  { 0,0, 0, NULL,NULL, NULL,NULL }
};

const DigDispElem far digOutElems[] = {
//  xx,yy,  BIT_MASK,        "0STR",ATTR_NAME, "1STR",ATTR_NAME

  { 38, 4,  VARIAC_UP_OFF,   "ON ",DARK_RED,   "OFF",DARK_BLUE },
  { 38, 5,  VARIAC_DOWN_OFF, "ON ",DARK_RED,   "OFF",DARK_BLUE },
  { 38, 6,  ALLOW_INTLKS,    "YES",DARK_RED,   "NO ",DARK_BLUE },

  { 38, 8,  TURN_PS_OFF,     "YES",DARK_BLUE,  "NO ",DARK_RED  },

  { 0,0,  0,  NULL,NULL,  NULL,NULL }
};

//---------------------------------------------------------------
//---- older style - do not use for newer applications -----
const DigFmtInfo digInFmt[] = {
//{ xx,yy, "0-str"  ,1,  "1-str" ,0 },  //  descr
  { -1,-1, "",0, "",0 }
};

const DigFmtInfo digOutFmt[] = {
//{ xx,yy, "0-str"  ,1,  "1-str" ,0 },  //  descr
  { -1,-1, "",0, "",0 }
};


const AnalogFmtInfo  adcFmt[] = {
  { -1,-1, 0,0 }
};

const AnalogFmtInfo  dacFmt[] = {
  { -1,-1, 0,0 }
};


//-----------------------------------------------------------------------------
//  "state":  Show continuously updated state of the system
//-----------------------------------------------------------------------------
void _task showState(int argc, char **argv)
{
  bool  blinkOff;
  char  temp[128];


  printf(clsCmd);  blinkOff = false;

  CursorTo(0, 3);  printf(DevStateStr);

  for (;;)  {
    CursorTo(0, 0);  blinkOff = !blinkOff;

    stdHeader();  displayElems((DispElem far *)dispElems, blinkOff, temp);

    displayDigElems(digitalIn, (DigDispElem far *)digInElems, blinkOff, temp);
    displayDigElems(digitalOut, (DigDispElem far *)digOutElems, blinkOff, temp);

/* //tdebug
    CursorTo(25,18);
    printf("assertP: %d  ampSetP: %d  switchP: %d  pErr: %d  ",
           Cur.assertPolarity, Cur.ampSetPolarity, Cur.switchPolarity,
           Cur.polarityErr);  */

    CursorTo(0,22);  fflush(stdout);  CheckETX();  sleepMS(500);
  }

}


//-----------------------------------------------------------------------------
//  Application-specific commands
//-----------------------------------------------------------------------------
#pragma warn -775

void _task cmd(int argc, char **argv)
{
  bool  validCmd;
  char  *cmdStr;

#define SetTrue(val)  { \
  _EnterCritical_();  val = true;  _ExitCritical_();  \
  validCmd = true;  break; \
}
#define SetVal(val)  { \
  KLOCK();  val = atof(argv[3]);  KUNLOCK();  \
  validCmd = true;  break; \
}


  validCmd = false;

  do  {
    if (argc < 2)  break;

    cmdStr = argv[1];

    if (StriEq(cmdStr, "on"))  { onRequest();  validCmd = true;  break; }

    if (StriEq(cmdStr, "off"))  { offRequest();  validCmd = true;  break; }

    if (StriEq(cmdStr, "reset"))  { resetRequest();  validCmd = true;  break; }

    if (StriEq(cmdStr, "set") AND (argc > 2))  {
      newSetpoint(atof(argv[2]), true);  validCmd = true;  break; }

  }  while(0);

  if (! validCmd)  {
    printf("\r\n"
           "Usage:  cmd on|off|reset|set #\r\n"
          );
    Exit(-1);
  }

  Exit(0);
}

#pragma warn 775


//=============================================================================
//  Event-Handler functions called by the appl task to process event messages
//=============================================================================
//APPLCODE

//-----------------------------------------------------------------------------
//  Event-Handler for a request to ...
//-----------------------------------------------------------------------------
/*
void EVENT_MyEvent(EventMsg *event)
{
  uint  idx;
  float  setpoint;
  char  temp[16];


  idx = event->chanNum;

  if (error)  {
    fprintf(DiagLog, "error msg, chan:\n", idx+1);
    return;
  }

  setpoint = *((float *)event->data);

  fprintf(DiagLog, "New SetPoint for PS%d: %s\n",
     idx+1, fmtFloat(setpoint, 8,3, temp) );

  *((float *)((&iSet_1_RegInfo + idx)->regVal)) = setpoint;

  anaOut[idx] = ISETtoVolts(idx, setpoint);
}
*/



//=============================================================================
//  Update the the state of the application logic in well-defined phases
//=============================================================================
// IMPORTANT NOTES and RULES:
//
//  - The code to change ANY value must be in ONE place.  NEVER spread the
//    logic for updating a single value around.  There should be ONE place
//    to look to see how a value gets set.
//
//  - "cur" values are basically "inputs" - literally (they are set based
//    on the just-read state of a digital or analog input) and/or figuratively
//    (they hold a value set by some external source, such as a cmd from a
//    remote client or another user-interface task).
//
//  - "next" values are basically "outputs" - literally (they determine the
//    next state of a digital or analog output) and/or figuratively (they
//    can affect the state of the logic for the next pass).
//
//  - The "cur" OR the "next" instance of a value may be changed - NEVER
//    both of them (with the exception of the first step below that copies
//    all of cur to prev and next to cur).
//
//  - A value MUST be in the ApplState struct if one of the following applies:
//     - The application logic is dependent on if or how much the value has
//       changed between passes.
//     - The value is used as both input (used in in a conditional and/or in
//       an expression that can affect the result of an assignment) and output
//       (the target of an assignment).
//
//  - Variables that are NOT in the ApplState struct but are USED (in
//    conditionals or other expressions) by the application logic must be
//    treated just like "cur" values.  Variables SET (the target in an
//    assignment statement) by the application logic must be treated just
//    like "next" values.
//
//  - When the logic requires comparing an "input" to a related "output", be
//    sure to use the "cur" value for the "input", and the "prev" value for
//    the "output".
//
//    This is because during the processing of the loop, the "cur" values
//    represent the input values just read (what they are "now") and the
//    "prev" values represent the values the outputs were set to at the
//    beginning of the prev pass (i.e. they are the values that "caused"
//    the inputs to have their "cur" values).
//
//    During a pass the output values in "cur" represent what the outputs
//    where set to AFTER the "cur" inputs were updated, so comparing them
//    makes no sense.  Doing so is the equivalent of comparing a past response
//    to a condition that didn't exist at the time.
//=============================================================================


//=============================================================================
//  Definitions for functions that handle setting "next" values that serve
//  as inputs to the application control logic (run by the "applic" task in
//  a periodically executed loop).
//
//  While care must be taken in these functions to use locks to avoid the
//  possibility of being preempted in the middle of an operation, such locks
//  must be kept to as short a time as possible to avoid excessively latency
//  in the main control loop.
//
//  For REALLY short locks, using a critical section is best, because this
//  has the least overhead (but it has to be REALLY short to avoid causing
//  problems for interrupt handlers).
//
//  For longer locks (ones that involve anything but the most trivial
//  conditional logic), use either KLOCK() and KUNLOCK() (which disable the
//  scheduler) or a more specific global semaphore/mutex lock.
//=============================================================================

//-----------------------------------------------------------------------------
void onRequest(void)
{
  _EnterCritical_();
  Next.onRequest = true;  Next.offRequest = false;
  _ExitCritical_();

  fprintf(DiagLog, "ON request\n");
}

//-----------------------------------------------------------------------------
void offRequest(void)
{
  _EnterCritical_();
  Next.offRequest = true;  Next.onRequest = false;
  _ExitCritical_();

  fprintf(DiagLog, "OFF request\n");
}

//-----------------------------------------------------------------------------
void resetRequest(void)
{
  _EnterCritical_();  Next.resetRequest = true;  _ExitCritical_();

  fprintf(DiagLog, "RESET request\n");
}

//-----------------------------------------------------------------------------
void newSetpoint(float newAmps, bool logIt)
{
  char  temp[16];


  KLOCK();

  if (FABS(newAmps) > maxAmps)  newAmps = (float)SIGN(newAmps) * maxAmps;

  Next.ampSet = newAmps;

  KUNLOCK();

  if (logIt)
    fprintf(DiagLog, "New Setpoint: %s\n", fmtFloat(newAmps, 10,3, temp) );
}

//-----------------------------------------------------------------------------



int  bitCounts[32];

//-----------------------------------------------------------------------------
//  Minimize the effect of noise on the digitial inputs in trade for some
//  added latency in response time to actual "legitimate" changes in state.
//
//  What this func does is add or subtract the periodic timer interval value
//  to/from a running total for each input bit, with a cap on the total being
//  the configured # of ms for debouncing.
//
//  The current state of each input is determined by whether the running total
//  for each input is positive or negative.  A positive value indicates a
//  predominately ON state, while a negative value indicates a predominiately
//  OFF state.
//
//  Inputs that are mostly stable (anomolies not very frequent) will quickly
//  reach the positive or negative limit and stay close to it until the "real"
//  state changes to the opposite state.  When that happens, the total will
//  shrink, cross 0, and grow toward the opposite limit.
//
//  NOTE the latency effect of this approach on the two extremes:
//    - No noise:  No errors, but Max latency (== the debounce period)
//    - Max noise:  Min latency, but constant errors.
//
//  NOTE that debounce period also effectively sets a limit on how short a
//  period that a change in state must last for it to be recognized:  Any
//  change that lasts for less than twice the debounceMS setting will
//  generally not be detected.
//
//  The setting for the debounce period effectively allows you to ajust the
//  inherent compromise between delays introduced for the correct readings vs
//  the probability of getting a false reading.  A setting of 0 effectively
//  disables the filter, while larger values reduce the effect of any noise.
//
//  This approach behaves aproximately like connecting a capacitor across a
//  noisy input.  As long as the noise level is low enough, the capacitor
//  will tend to charge up to the level of the ON voltage, or dischage to
//  0.  Noise will increase or decrease the voltage, but so long as it is
//  less frequent than the "real" value, the voltage across the capacitor
//  will tend to stay near the voltage level for the "real" state.
//-----------------------------------------------------------------------------
void debounceDigIn(void)
{
  int  i;
  int  *bitCount;
  ulong  mask, digIn;


  bitCount = bitCounts;  digIn = 0;  mask = 1;

  for (i=0; i<32;  ++i)  {
    if (digitalIn & mask)
      *bitCount += (int)SchedMS;
    else
      *bitCount -= (int)SchedMS;

    if (*bitCount >  (int)debounceMS)  *bitCount =  (int)debounceMS;
    if (*bitCount < -(int)debounceMS)  *bitCount = -(int)debounceMS;

    if (*bitCount > 0)  digIn |= mask;

    ++bitCount;  mask <<= 1;
  }

  digitalIn = digIn;
}


//-----------------------------------------------------------------------------
void showBiasVolts(void)
{
  char  temp[32];

  printf("  Bias Volts Chg rate: %.16s\r\n",
           Engineering(biasVoltsChgRate, 4, temp+16) );
}


//=============================================================================
//  Copy the "cur" state from last time to "prev", and update "cur"
//=============================================================================
//  Just before this function is called, readings were taken of the current
//  state of the hardware, and any changes to the output states specified by
//  the last pass were applied (just after taking the readings).
//
//  This func first copies "cur" to "prev" and then "next" to "cur" which does
//  2 things for us:
//
//    - We have a copy of the controller's logical state from the last pass
//      (so we can determine what has changed)
//
//    - We have a copy of values set by other tasks indicating events/requests
//      that happend/were made since the beginning of the last pass
//
//  We then reset the values (as needed) that are set by other tasks to
//  indicate events and requests, and update the current state based on the
//  readings just taken and the requests and events communicated by other
//  tasks.
//
//  All of this is in preperation for the updateNextState() function, which
//  implements the main "decision making" portion of the application logic,
//  determining what the next state should be based on the latest conditions
//  (as determined by the updateCurState() function).
//
//=============================================================================
void updateCurState(void)
{
  bool  loggedStateChg;
  float  curErr;
  ulong  interval;
  char  temp[64];


  if (current != applTask)  return;  // must only be called by "applic" task

//-----------------------------------------------------------------
// The following conditions exist at this point:
// - The inputs have just been read
// - The outputs were just set to values computed during the last pass
// - "cur" contains the state the inputs were in at the START of
//   of the previous pass, and the state the outputs were in DURING
//   the previous pass
// - "next" contains the values that the outputs were just set
//   to and any "input" values that were changed by other tasks
//------------------------------------------------------------------


//  Next.processTime = GetMS();  curTime = time(NULL);

//  msSincePrev = msSince(Cur.processTime);

  debounceDigIn();  Next.digitalIn = digitalIn;

//APPLCODE

  Next.ctlPowerOn   = !(digitalIn & CTL_POWER_OFF);
  Next.waterFlowOK  = !(digitalIn & WATER_FLOW_BAD);
  Next.tempOK       = !(digitalIn & TEMP_BAD);
  Next.fusesOK      = !(digitalIn & TRANS_FUSES_BAD);
  Next.interlocks   = BOOL(digitalIn & INTERLOCKS_BAD);
  Next.onIsAsserted = !(digitalIn & OFF_ASSERTED);
  Next.overloadsOK  = !(digitalIn & OVERLOADS_BAD);
  Next.offPressed   = BOOL(digitalIn & MANUAL_OFF);
  Next.externIntlk  = BOOL(digitalIn & EXT_INTERLOCK);
  Next.psIsOn       = !(digitalIn & PS_IS_OFF);

  loadAmps  = readVal[SHUNT_VOLTS];  loadAmps = FABS(loadAmps);
  psVolts   = readVal[SUPPLY_VOLTAGE];
  biasVolts = readVal[BIAS_VOLTS];
  loadVolts = readVal[LOAD_VOLTAGE];


  Next.biasError = biasVolts - biasVoltsSet;
  Next.biasErrorABS = FABS(Next.biasError);

  Next.ampSetPolarity = SIGN(Next.ampSet);

//--- Compute the rate of chg for the bias voltage aprox 10 times/sec. ---
//--- If the error is decreasing fast enough, then ignore the error ---
  if ((interval = msSince(lastBiasVoltsChkMS)) >= 100)  {
    lastBiasVoltsChkMS = GetMS();
    avgdBiasVolts = GetAvgdRead(BIAS_VOLTS);
    curErr = avgdBiasVolts - biasVoltsSet;  curErr = FABS(curErr);
    biasVoltsChgRate = (prevBiasVoltsErr - curErr) * (1000.0/(float)interval);
    ignoreBiasErr = (biasVoltsChgRate > minBiasChgRate);
    prevBiasVoltsErr = curErr;
  }



//--- Check to see if the bias voltage is out of spec for too long ---
  do  {
    if (!Next.adjustBias)  {
      biasAdjustTime = 0;  Next.biasWarn = false;  break; }
    if (!biasAdjustTime)  { biasAdjustTime = GetMS();  break; }
    if (msSince(biasAdjustTime) < biasWarnMS)  break;
    Next.biasWarn = true;
  } while(0);

//--- Check to see if the AMPLITUDE of the bias voltage is dangerously high ---
//--- (Dangerous, that is, to the operation of the transistors.  Too high ---
//--- a voltage, and one will burn out - which causes another to burn out, ---
//--- and then another, ... ---)
  do {
    if ( (Next.biasErrorABS < biasErrorStart) OR ignoreBiasErr OR
        ((biasVolts < biasVoltsSet) AND (!Cur.psIsOn OR !Cur.assertOn)) ) {
      biasErrTime = 0;  Next.biasIntlk = false;  break; }
    if (!biasErrTime)  { biasErrTime = GetMS();  break; }
    if (msSince(biasErrTime) < biasErrMS)  break;
    Next.biasIntlk = true;
  } while(0);



//--- Because the shunt resistor used to determine the output current ---
//--- is not on the load side of the polarity switch it always gives us ---
//--- a positive reading, so we need to set the polarity of the reading ---
//--- based on the state of the polarity switch. ---

//--- Unfortunately, the polarity switch has been known to "stick", so ---
//--- simply setting the polarity of loadAmps based on what we THINK the ---
//--- state of the switch is can result in an incorrect value. ---

//--- The load voltage, which is read from the load-side of the polarity ---
//--- switch should be able to tell us the right polarity, but for small ---
//--- currents the load voltage can be small enough that it bounces around ---
//--- zero, so its polarity is only stable/usable for larger values. ---

//--- So, for lower currents we use what we BELIEVE the state of the ---
//--- polarity switch is (so the polarity of loadAmps doesn't bounce ---
//--- around), and for higher currents we use the polarity of loadVolts ---
//--- (which should be stable for higher currents) ---

  if ((loadAmps < polThreshAmps) OR (FABS(loadVolts) < polThreshVolts))  {
    if (Prev.switchPolarity < 0)  loadAmps = -loadAmps;  }
  else  // use polarity of load voltage
    if (loadVolts < 0.0)  loadAmps = -loadAmps;

  Next.loadPolarity = SIGN(loadAmps);

  if ((Next.switchState == SW_ON) AND (Next.state == STATE_ON))
    Next.polarityErr = (Next.loadPolarity != Next.switchPolarity);
  else
    Next.polarityErr = false;

/*
  Only allow a softReset if we are in a "safe" state (i.e. leaving the
  outputs in their current state for about 30 seconds is not a risk to the
  equipment).  Keep in mind that what we are choosing here is what happens
  when the OS discovers a fatal problem that requires a restart.  There
  are only 2 options:

    - A hard/CPU reset:  This puts all the digital outputs back to some
      default state, and resets the analog outputs right after rebooting.

    - A soft reset:  This leaves all the outputs (digital and analog) in
      the state they are in during the restart (for aprox 30 seconds) and
      the control logic picks up right where it left off.  The result is
      basically the same as what would happen if the application task
      was suspended for 30 seconds.

  So, even if a soft reset might not be ideal, it is probably better than
  a hard reset, which insures everything turns off and stays that way until
  someone notices and manually turns things back on.

  For the TCPS, the only real concern is that the UP control bit for the
  variac is not being asserted, as that could cause the bias voltage to keep
  climbing, burning up transistors.  And simply disablng the command before
  a soft reset is not a sufficient safeguard because it might have been
  asserted because the current setpoint was causing a magnet to ramp up
  (and a soft reset would suspend the ramping, which generally causes the
  bias voltage to suddenly jump without any change to the variac position).
*/
  // now handled in applicShutdown()
//allowSoftReset = BOOL(digitalOut & VARIAC_UP_OFF);



//--- If the rampRate setting changed, update the max amount we should ---
//--- change the active setpoint each pass ---
  if (rampRate != lastRampRate)  {
    maxAmpChg = rampRate * (float)SchedMS / 1000.0;  lastRampRate = rampRate; }


// Note that a copy is done (rather than just switching pointer values) to
// insure all values are set in both the source and destination.  If we
// simply kept pointers and swapped the pointer values, then EVERY value
// would have to be updated independently by the code, which is a pain to
// program, adds overhead and code, and is a VERY easy way to introduce odd
// bugs.  Overall, copying everything here is actually faster overall,
// simpler, and much safer.

  Prev = Cur;  Cur = Next;

//--- At this point the Cur copy contains what the inputs were at the start ---
//--- of this pass, and the Prev copy contains the output values from the ---
//--- the previous pass (i.e. the output values that resulted in the input ---
//--- values in Cur) ---


//--- log msgs for digital I/O bit chgs (but only if not starving other ---
//--- processes for cycles) ---
  /* if (applLoopTime < 8) */  logDigIOChgs(false);


  loggedStateChg = false;

//--- log msgs for various changes ---
  if (Cur.switchState != Prev.switchState)  {
    printf("Switch state: %.32s\r\n", SwitchStateDesc[Cur.switchState]);
    loggedStateChg = true;
  }

  if (Cur.polarityErr != Prev.polarityErr)
    if (Cur.polarityErr)  {
      polarityErrStartTime = GetMS();  loggedPolarityErr = false; }
    else  if (loggedPolarityErr)  {
      polarityErrStartTime = 0;  printf("*** Load Polarity OK ***\r\n");
      loggedStateChg = true;
    }

  if (Cur.polarityErr)  if (msSince(polarityErrStartTime) > 5000)
    if (!loggedPolarityErr)  {
      printf("*** Load Polarity WRONG ***\r\n");
      loggedPolarityErr = loggedStateChg = true;
    }

  if (Cur.adjustBias != Prev.adjustBias)  {
    printf("Bias Adjust: %s", Cur.adjustBias ? "START" : "END");
    showBiasVolts();  loggedStateChg = true; }

  if (Cur.biasWarn != Prev.biasWarn)  {
    printf("***** Bias voltage Warning: %s *****\r\n",
           Cur.biasWarn ? "SET" : "CLEARED" );
    showBiasVolts();  loggedStateChg = true;
  }
  if (Cur.biasIntlk != Prev.biasIntlk)  {
    printf("***** Bias voltage INTERLOCK: %s *****\r\n",
           Cur.biasIntlk ? "SET" : "CLEARED" );
    showBiasVolts();  loggedStateChg = true;
  }

/*  //tdebug
    if ( (Cur.assertPolarity != Prev.assertPolarity)
      OR (Cur.ampSetPolarity != Prev.ampSetPolarity)
      OR (Cur.switchPolarity != Prev.switchPolarity)
      OR (Cur.polarityErr != Prev.polarityErr)  )
     printf("assertP:%d, ampSetP:%d, switchP:%d, pErr:%d\r\n",
            Cur.assertPolarity, Cur.ampSetPolarity, Cur.switchPolarity,
            Cur.polarityErr
           );  */

  logInterval = (Cur.state != STATE_ON) ? 3600 : 600;

  if ( loggedStateChg OR (SecSince_(lastValLogTime) > logInterval) )  {
    if (!loggedStateChg)  printf("Current state:\n");
    printf("  ampSet:%s   loadAmps:%s\n"
           "  biasVolts:%s   biasVoltsSet:%s\n"
           "  psVolts:%s   loadVolts:%s\n"
           "  Assert:%s  Actual:%s\n",
           fmtFloat(ampSet, 7,2, temp),
           fmtFloat(loadAmps, 7,2, temp+10),
           fmtFloat(biasVolts, 6,2, temp+20),
           fmtFloat(biasVoltsSet, 6,2, temp+30),
           fmtFloat(psVolts, 6,2, temp+40),
           fmtFloat(loadVolts, 6,2, temp+50),
           digitalOut & TURN_PS_OFF ? "OFF" : "ON ",
           digitalIn & PS_IS_OFF ? "OFF" : "ON "
          );
    lastValLogTime = GetSeconds_();
  }

}



//=============================================================================
//  Now we update the values in "next" based on the "cur" and "prev"
//  values.
//=============================================================================
//  All values changed by updateNextState() are "outputs" from this func.
//  All other values referenced in this function are "inputs" to this func.
//
//  All the values in "next" provide information about the next logical state
//  for the controller:  Values to be applied during the next pass of the
//  control logic.  All "next" values fall in to one of 2 categories:
//
//    - They are set/updated ONLY by this function (which is executed only
//      by the "applic" task)
//
//    - They are set ONLY by other tasks (within a lock), and possibly
//      reset/cleared by the "applic" task in updateCurState() (after copying
//      "next" to "cur")
//
//
//  All values that act as input to this func must fall in to one of these
//  categories:
//
//    - "prev" values
//
//    - "cur" values updated directly by statements in updateCurState()
//
//    - "cur" values updated indirectly by updateCurState() due to the copy
//      of "next" to "cur", which includes the values in "next" that are
//      set by other tasks
//
//    - a shared/global value that is the equivalent of a "cur" value but
//      for which we don't need to track changes
//
//
//  All "outputs" from this function must be altered ONLY within this function.
//
//  All "inputs" to this function must NEVER be altered within this function.
//
//  Each output value (or VERY small "set" of output values) must be updated
//  in a single, self-contained block of code.  NO output may be referenced
//  OR altered in ANOTHER block of code:  This is a declarative style of
//  programming and has multiple advantages:
//
//     - There are no order-of-execution dependencies across nodes/blocks
//       within a pass.  They can be executed in any order and the result
//       will be the same (get used to it - this method will be the norm as
//       computer languages and hardware add support for atomic transactions
//       and computers evolve in to neural-net-like processor arrays - FPGAs
//       are just the beginning of this trend).
//
//     - Encapsulating all logic for an output within a single node/block
//       leaves just one place to start when tracking a problem backward from
//       the result (far simpler than guessing at the possible causes and
//       trying to work forward, looking for paths to the observed result).
//
//  Each node/block-of-code should be prefaced with the following information:
//   - The names of the values it modifies
//   - A descr of the meaning of the values and the logic used to set them
//   - A list of the values that serve as "inputs" to the node
//
//=============================================================================
void updateNextState(void)
{
  float  chg;
  char  temp[32];


  if (current != applTask)  return;  // must only be called by "applic" task

//--------------------------------------------------------------------
// Next.assertReset, Next.resetStarted
// Next.resetRequest (cleared)
//    Description:
//       Should we be asserting a reset pulse for the error inputs?
//    Inputs:
//       Cur.resetRequest
//       Cur.resetStarted
//--------------------------------------------------------------------
  if (Cur.resetRequest)  {
    Next.resetRequest = false;
    Next.resetStarted = GetMS();  Next.assertReset = true;
  }
  if (Cur.resetStarted AND (msSince(Cur.resetStarted) >= 500))  {
    Next.resetStarted = 0;  Next.assertReset = false;  }


//--------------------------------------------------------------------
// Next.state
// Next.rtzStarted
// Next.onRequest, Next.offRequest (cleared)
//    Description:
//       Determines the next state of for the application logic
//       (STATE_ON, STATE_OFF, STATE_INTLK)
//    Inputs:
//       Cur.onRequest
//       Cur.offRequest
//       Prev.assertReset
//       Cur.assertReset
//       Cur.externIntlk
//       Cur.biasIntlk
//----------------------------------------------------------------------
//--- chg state from OFF to ON? ---
  if (Cur.onRequest)  do  {
    Next.onRequest = false;
    if (!switchType)  {
      printf("Must set switchType 1st!\n");  break;  }
    if ((Cur.state == STATE_INTLK) OR Cur.externIntlk)  {
      printf("Must clear Interlock first\n\n");  break; }
    if (Cur.state == STATE_ON)  { printf("Already ON\n\n");  break; }
    Next.state = STATE_ON;  printf("Turning ON\n\n");
  } while(0);

//--- chg state from ON to OFF (or ON while ramping to zero) ? ---
  if (Cur.offRequest)  do {
    Next.offRequest = false;
    if ((Cur.state == STATE_INTLK) OR Cur.externIntlk)  {
      printf("Must clear Interlock first\n\n");  break; }
    if (Cur.state == STATE_OFF)  { printf("Already OFF\n\n");  break; }
    if (maxRTZms)  {
      if (Cur.rtzStarted)  break;
      printf("Ramping to Zero\n\n");  Next.rtzStarted = GetMS();  break;
    }
    Next.state = STATE_OFF;  printf("Turning OFF\n\n");
  } while(0);

//--- chg state from ON (while ramping to zero) to OFF? ---
  if (Cur.rtzStarted)  do  {
    if (Cur.state != STATE_ON)  { Next.rtzStarted = 0;  break; }
    if ((msSince(Cur.rtzStarted) < maxRTZms)
      AND (FABS(loadAmps) > safeSwAmps))  break;
    Next.rtzStarted = 0;  Next.state = STATE_OFF;
    printf("Turning OFF with loadAmps at: %s\n\n", fmtFloat(loadAmps, 8,3, temp) );
  } while(0);

//--- chg state from INTLK to OFF? ---
  if (Prev.assertReset AND !Cur.assertReset)  do {  // end of reset pulse
    if (Cur.state != STATE_INTLK)  {
      printf("No Interlock to clear\n\n");  break;  }
    if (Cur.interlocks)  {
      printf("Interlocks still asserted\n\n");  break; }
    if (Cur.externIntlk)  {
      printf("External Interlock still asserted\n\n");  break; }
    if (Cur.biasIntlk)  {
      printf("Bias voltage interlock still exists\n\n");  break; }
    Next.state = STATE_OFF;
  } while(0);

//--- chg state to INTLK? ---
  if (Cur.externIntlk OR Cur.biasIntlk OR Cur.interlocks)
    Next.state = STATE_INTLK;


//--------------------------------------------------------------------
// Next.switchState
// Next.assertPolarity
// Next.switchStarted
// Next.switchPolarity
//    Description:
//       Determines the next state for the polarity switch
//    Inputs:
//       Cur.switchState
//       Cur.switchPolarity
//       Cur.switchStarted
//       Cur.state
//       Cur.psIsOn
//       Cur.assertOn
//----------------------------------------------------------------------
  do  {

  // The motor-driven polarity switch does not have a stable OFF
  // position.  It is either in the normal or reverse state, or
  // in transition from one to the other.  If we are asserting
  // the ON command for the supply during the transition, we can
  // monitor the progress because the motorized switch will
  // shut off the supply and then turn it back on.  Otherwise,
  // all we can do is wait for sufficient time to pass and assume
  // that a transition finished.

    if (switchType == MOTOR_SWITCH)  {

    //--- We can't track the actual state during the change unless we ---
    //--- are asserting the ON command the whole time ---
      if (!Cur.assertOn AND Cur.switchStarted)  {
        Next.switchState = SW_UNKNOWN;  break; }

      switch (Cur.switchState)  {
        case SW_UNKNOWN:
          Next.switchPolarity = 0;
          if (msSince(Cur.switchStarted) < switchTime)  break;
          Next.switchState = SW_ON;  Next.switchStarted = 0;
          break;

        case SW_OFF:  // an error!
          printf("ERROR: switchState == OFF for MOTOR type switch!!!\r\n");
          Next.switchState = SW_UNKNOWN;  Next.switchStarted = GetMS();
          break;

        case SW_ON:  // do we need to change the polarity?
          if ((Cur.state == STATE_ON)
              AND (Cur.switchPolarity != Cur.ampSetPolarity))
            Next.switchState = SW_WAIT_FOR_ZERO;
          break;

        case SW_WAIT_FOR_OFF:  // wait for switch to turn PS off
          if (!Cur.psIsOn)  Next.switchState = SW_WAIT_FOR_ON;
          break;

        case SW_WAIT_FOR_ON:  // wait for switch to turn PS back on
          if (!Cur.psIsOn)  break;
          Next.switchState = SW_ON;  Next.switchStarted = 0;
          break;

        case SW_WAIT_FOR_ZERO:  // wait until it is safe to start change
          if ((Cur.state == STATE_ON)  // abort if polarity chg no longer needed
              AND (Cur.switchPolarity == Cur.ampSetPolarity))  {
            Next.switchState = SW_ON;  break; }
          if (Cur.ampApply != 0.0)  break;
          if (FABS(loadAmps) > safeSwAmps)  break;
          Next.assertPolarity = Cur.ampSetPolarity;  // start moving the switch
          Next.switchState = SW_WAIT_FOR_OFF;  Next.switchStarted = GetMS();
          break;
      }

      if (!Cur.psIsOn)  Next.switchPolarity = 0;
      if ((Cur.switchState != SW_ON) AND (Next.switchState == SW_ON))
        Next.switchPolarity = Next.assertPolarity;
    }

  // For the SCR-controlled polarity switch version, we MUST shut
  // off the supply (so the SCR's will stop conducting), change
  // the polarity commands, then turn the supply back on.

    if (switchType == SCR_SWITCH)  {

      if (Cur.state != STATE_ON)  {
        Next.assertPolarity = 0;  Next.switchState = SW_OFF;  break; }

      switch (Cur.switchState)  {
        case SW_UNKNOWN:
          Next.switchPolarity = 0;
          if (msSince(Cur.switchStarted) < switchTime)  break;
          Next.switchState = SW_ON;  Next.switchStarted = 0;  break;
          break;

        case SW_OFF:  // do we need to turn on one of the SCRs?
          if (Cur.state != STATE_ON)  break;
          Next.assertPolarity = Cur.ampSetPolarity;
          Next.switchState = SW_WAIT_FOR_ON;  Next.switchStarted = GetMS();
          break;

        case SW_ON:  // do we need to change the polarity?
          if ((Cur.state == STATE_ON) AND Cur.assertOn
              AND (Cur.switchPolarity != Cur.ampSetPolarity))
            Next.switchState = SW_WAIT_FOR_ZERO;
          break;

        case SW_WAIT_FOR_OFF:  // wait sufficient time for SCR to turn off
          if (Cur.psIsOn)  { Next.switchStarted = GetMS();  break; }
          if (msSince(Cur.switchStarted) < switchTime/2)  break;
          Next.switchState = SW_WAIT_FOR_ON;
          Next.assertPolarity = Cur.ampSetPolarity;
          break;

        case SW_WAIT_FOR_ON:  // wait a bit before applying setpoint
          if (msSince(Cur.switchStarted) < switchTime)  break;
          Next.switchState = SW_ON;  Next.switchStarted = 0;
          break;

        case SW_WAIT_FOR_ZERO:  // wait until it is safe to start change
          if (Cur.ampApply != 0.0)  break;
          if (FABS(loadAmps) > safeSwAmps)  break;
          Next.assertPolarity = 0;  // turn off both SCRs
          Next.switchState = SW_WAIT_FOR_OFF;  Next.switchStarted = GetMS();
          break;
      }

      if (Next.switchState == SW_ON)
        Next.switchPolarity = Next.assertPolarity;
    }

  } while(0);


//-------------------------------------------------------------------
// Next.assertOn
//    Description:
//       Should the ON bit be asserted?
//    Inputs:
//       switchType - motor-driven or SCR-controlled?
//       Cur.state
//       Cur.switchState
//-------------------------------------------------------------------
  Next.assertOn = (Cur.state == STATE_ON);

  // If we are in the process of changing the polarity for the SCR version,
  // when we reach the point where the ON commands for both the SCR's have
  // been disabled, we turn off the supply to help inusure that both the
  // SCRs stop conducting.

  // Remember that, for the motor-driven version, we WANT the ON command
  // asserted whenever allowed because that is the only way we can tell what
  // state the switch is in (the motor-driven one turns the supply off and
  // back on during the change, but we can only read the actual on/off state
  // if we are asserting the on command while the motor changes it).

  if (switchType == SCR_SWITCH)
    if ((Cur.switchState != SW_ON) AND !Cur.assertPolarity)
       Next.assertOn = false;


//-------------------------------------------------------------------
// Next.ampApply
//    Description:
//       The setpoint that we want applied to the regulation circuit.
//       NOTE that this value must be 0 while a change in the polarity
//       is in progress.
//    Inputs:
//       Cur.ampApply - cur active setpoint for regulation circuit
//       Cur.psIsOn   - is the PS currently ON
//       Cur.assertOn - have we told the PS to turn on
//       Cur.switchState - logical state of the polarity switch
//       Cur.switchPolarity - polarity state of switch
//       Cur.ampSet   - desired setpoint when supply is ON
//
//-------------------------------------------------------------------
  {
    float  target;

//--- if the PS is (or should be) OFF, then apply a 0 setpoint ---
    if  (!Cur.psIsOn OR !Cur.assertOn)
      Next.ampApply = 0.0;
    else  {
    //--- adjust ampApply in direction of target setpoint ---
      target = (Cur.rtzStarted ? 0.0 : Cur.ampSet);
      chg = target - Cur.ampApply;
      if (FABS(chg) > maxAmpChg)  chg = (chg > 0.0) ? maxAmpChg : -maxAmpChg;
      Next.ampApply = Cur.ampApply + chg;

    //--- A change that results in a new setpoint whose polarity is NOT the ---
    //--- same as the current switch polarity is not allowed.  This forces ---
    //--- the applied setpoint to stop at 0 until the polarity of the ---
    //--- switch matches the direction we are trying to go. ---
      if (Cur.switchPolarity != SIGN(Next.ampApply))  Next.ampApply = 0.0;
    }
  }


//-------------------------------------------------------------------
// Next.adjustBias
// Next.incrBiasVolts
// Next.decrBiasVolts
//    Description:
//       An analog feedback circuit regulates the output current by
//       controlling the base voltage on a bank of transistors.  But
//       the collector-bias voltage for the transistor bank is set by
//       a software-controlled mechanically-driven variac.  Whenever
//       the supply is on, the software is responsible for keeping
//       the difference between the actual bias voltage and its
//       setpoint within a specified range.
//    Inputs:
//       Cur.psIsOn   - is the PS currently ON
//       Cur.assertOn - have we told the PS to turn on
//       Cur.biasErr  - how far is bias voltage from the setpoint
//       biasErrorStart - error level above which correction begins
//       biasErrorStop - error level below which correction stops
//-------------------------------------------------------------------
  if (Cur.biasErrorABS > biasErrorStart)  Next.adjustBias = true;
  if (Cur.biasErrorABS < biasErrorStop)   Next.adjustBias = false;

//--- If actual < setting and PS is not on, then don't adjust it ---
  if (!Cur.psIsOn OR !Cur.assertOn)  if (Cur.biasError < 0)
    Next.adjustBias = false;

  Next.incrBiasVolts = Next.decrBiasVolts = false;
  if (Next.adjustBias)
    if (Cur.biasError < 0.0)
      Next.incrBiasVolts = true;
    else
      Next.decrBiasVolts = true;


}  // end updateNextState()



//=============================================================================
//  Next we update output signal values to be written to the outputs at
//  the beginning of the next pass (just after the inputs are read).
//=============================================================================
void updateOutputState(void)
{
  if (current != applTask)  return;  // must only be called by "applic" task

  Next.digitalOut = 0xffffffffL;


//--- update output values that are based on current conditions (values ---
//--- in "cur" or any values used in a conditional in the applic logic) ---

/*EXAMPLE:
  // (insured earlier that only ONE of these can be true when setting them)
  if (initialBtnPressed)  setVal[FP_METER] = rfOffSetpoint;  else
  if (finalBtnPressed)    setVal[FP_METER] = rfOnSetpoint;  else
                          setVal[FP_METER] = cavityVolts;


  SetOffApplLEDs(Next.digitalOut);

  if (limited)  SetOnLimitLED(Next.digitalOut);
  if (regErr)   SetOnErrorLED(Next.digitalOut);
*/


//--- update output values based on values that were set by the application ---
//--- logic (e.g. values in "Next" or variables that are the target of an ---
//--- assignment statement in the application logic) ---

//APPLCODE

  if (Next.assertReset)   ClearBits(Next.digitalOut, ALLOW_INTLKS);

  if (Next.assertOn)  {
    ClearBits(Next.digitalOut, TURN_PS_OFF);

    if (Next.incrBiasVolts)  ClearBits(Next.digitalOut, VARIAC_UP_OFF);  else
    if (Next.decrBiasVolts)  ClearBits(Next.digitalOut, VARIAC_DOWN_OFF);

    if (switchType == SCR_SWITCH)  {  // separate bits for + and -
      if (Next.assertPolarity < 0)
        ClearBits(Next.digitalOut, REV_POLARITY_OFF);
      else  if (Next.assertPolarity > 0)
        ClearBits(Next.digitalOut, NORM_POLARITY_OFF);
    }
    else  // one bit sets either + or -
      if (Next.assertPolarity >= 0)  ClearBits(Next.digitalOut, REV_POLARITY);
  }
  else
    ClearBits(Next.digitalOut, VARIAC_DOWN_OFF);


  setVal[OUTPUT_CTL] = FABS(Next.ampApply);
}




//=============================================================================
// The last thing we do is update values read by other tasks
//=============================================================================
// IMPORTANT NOTES and WARNINGS:
//
//   To avoid concurrency issues, other tasks need to use locks when reading
//   values set here.  But be careful to keep such locks short to avoid
//   delaying the timing of the main application task.  If necessary, split
//   such code in to sections, getting and releasing the lock(s) in each
//   section to avoid increasing latency for the main application task.
//
//   Values in the "next" instance of the application state fall in to 2
//   categories:  Inputs and Outputs.
//
//   The "next" Outputs are those that may be changed in updateOutputState().
//   The reset are Inputs, and may be changed by other tasks within a lock.
//
//   Since each pass through the application logic starts by setting
//   cur = next, any "next" values set by other tasks become part of the
//   "cur" instance used by the rest of the applic logic.  Because the
//   "applic" task at a higher priority than any other real-time task, the
//   changes to "next" values by other tasks can ONLY happen between passes
//   through the applic logic loop.
//
//   NOTES:
//     - Some of these "next" input values values must be initialized in
//       updateCurState() in order to tell when a value is set by another
//       task.  For example, if we have a boolean flag that is set when we
//       get an ON command, then (after its "next" value is copied to "cur"),
//       we have to clear the flag in order to tell if another ON command is
//       received between passes.
//
//     - As with reads to these values, other tasks must use locks when
//       writing to the "next" input values to insure they are not preempted
//       by the "applic" task in the middle of a write.
//
//=============================================================================
void updateExternals(void)
{
  if (current != applTask)  return;  // must only be called by "applic" task

  digInBits = Cur.digitalIn;
  digOutBits = Cur.digitalOut;

//APPLCODE
  ampSet = Cur.ampSet;

  devStatus = 0;
  if (Cur.externIntlk)             devStatus |= DS_INTLK;
  if (Cur.state == STATE_INTLK)    devStatus |= DS_INTLK;
  if (Cur.state == STATE_ON)       devStatus |= DS_ON;
  if (Cur.switchPolarity >= 0)     devStatus |= DS_NORM_POL;
  if (Cur.rtzStarted)              devStatus |= DS_RTZ;
  if (!Cur.psIsOn)                 devStatus |= DS_PS_IS_OFF;

//  if (serReadErr)                  devStatus |= DS_SER_ERR;
//  if (calibErr)                    devStatus |= DS_CALIB_ERR;

}



//=============================================================================
_nearcall float readGetRaw(uint ui)  {
  return (ui < NumHdwADC ? anaIn[ui] : 0.0);  }

_nearcall float readMinRaw(uint ui)  {
  return (ui < NumHdwADC) ? (adcMinRaw(ui) - adcOffset[ui]) * adcGain[ui] : 0.0; }

_nearcall float readMaxRaw(uint ui)  {
  return (ui < NumHdwADC) ? (adcMaxRaw(ui) - adcOffset[ui]) * adcGain[ui] : 9.8; }

_nearcall void readDefaults(uint ui, float *gain, float *offset)
{
  switch (ui)  {
    case SHUNT_VOLTS:     *gain = 40.0;  break;
    case SUPPLY_VOLTAGE:  *gain = -1.0;  break;
    case BIAS_VOLTS:      *gain = -5.5;  break;
    case LOAD_VOLTAGE:    *gain =  5.5;  break;
    default:              *gain =  1.0;
  }
  *offset = 0.0;
}

_nearcall void readCalibPoints(uint ui, float *low, float *high)  {
  *low = -9.5;  *high = 9.5; }

_nearcall char *readUnits(uint ui, char *buf)  { return "V"; }


//=============================================================================
_nearcall float setGetRaw(uint ui)  {
  return (ui < NumHdwADC ? anaOut[ui] : 0.0);  }

_nearcall float setMinRaw(uint ui)  {
  return (ui < NumHdwDAC) ? (dacMinRaw(ui) - dacOffset[ui]) * dacGain[ui] : 0.0; }

_nearcall float setMaxRaw(uint ui)  {
  return (ui < NumHdwDAC) ? (dacMaxRaw(ui) - dacOffset[ui]) * dacGain[ui] : 9.8; }

_nearcall void setDefaults(uint ui, float *gain, float *offset)
{
  *gain = (ui == OUTPUT_CTL) ? -40.0 : 1.0;
  *offset = 0.0;
}

_nearcall void setCalibPoints(uint ui, float *low, float *high)  {
  *low = -9.5;  *high = 9.5;  }

_nearcall char *setUnits(uint ui, char *buf)  { return "V"; }


//=============================================================================
//  Set default labels for all the I/O signals
//=============================================================================
void setDefaultLabels(void)
{
  printf("Setting default labels for I/O signals\r\n");

//--- set application-appropriate default labels ---
//APPLCODE

//--- set default labels for the I/O signals ---
  fstrcpy(ainLabel[0], "Load Current");
  fstrcpy(ainLabel[1], "Supply Volts");
  fstrcpy(ainLabel[2], "Bias Volts");
  fstrcpy(ainLabel[3], "Load Volts");

  fstrcpy(aoutLabel[0], "Ctl Setpoint");

  fstrcpy(dinLabel[0], "Ctl Power Off");
  fstrcpy(dinLabel[1], "Water Flow Bad");
  fstrcpy(dinLabel[2], "Temp. Bad");
  fstrcpy(dinLabel[3], "Trans Fuses Bad");
  fstrcpy(dinLabel[4], "Interlocks");
  fstrcpy(dinLabel[5], "PS Off Asserted");
  fstrcpy(dinLabel[6], "Overload");
  fstrcpy(dinLabel[7], "OFF Pressed");
  fstrcpy(dinLabel[8], "DI-08");
  fstrcpy(dinLabel[9], "DI-09");
  fstrcpy(dinLabel[10], "Ext Interlock");
  fstrcpy(dinLabel[11], "PS is OFF");

  fstrcpy(doutLabel[0], "Variac Up OFF");
  fstrcpy(doutLabel[1], "Variac Down OFF");
  fstrcpy(doutLabel[2], "Allow Intlks");
  fstrcpy(doutLabel[3], "Mech Rev Pol");
  fstrcpy(doutLabel[4], "Turn PS Off");
  fstrcpy(doutLabel[5], "DO-05");
  fstrcpy(doutLabel[6], "DO-06");
  fstrcpy(doutLabel[7], "DO-07");
  fstrcpy(doutLabel[8], "Activity LED");
  fstrcpy(doutLabel[9], "SCR Plus Pol Off");
  fstrcpy(doutLabel[10], "SCR Neg Pol Off");
}


//=============================================================================
//  This function is called to force software configurable values to default
//  values.
//=============================================================================
void setDefaultConfig(void)
{
  printf("Setting in-RAM device config to default values\r\n");

//===== set default values for persistent data values =====
//APPLCODE
  switchType = MOTOR_SWITCH;

  maxVolts = 50.0;
  maxAmps = 100.0;

  safeSwAmps = 2.0;

  connectLoadDelay = 500;  // # of ms to wait after chging polarity

  biasVoltsSet = 10.0;
  biasErrorStart = 2.0;
  biasErrorStop = 0.50;

  biasWarnMS = biasErrMS = 2000;

  rampRate = 50.0;

  switchTime = 8000;  // based on possibility of being a motor-driven switch

  polThreshAmps = 2.0;  // min load cur to reliably use load volt polarity
  polThreshVolts = 0.2;  // min output voltage to reliably read polarity

  debounceMS = 250;  // interval (in ms) for debouncing digitalIn bits

  maxRTZms = 10000;  // max ms to wait for PS to reach zero before turning Off */

  minBiasChgRate = 0.5;  // min chg rate for bias voltage is 0.5V/sec */
}


//=============================================================================
//  This function is called by the application loop just before a shutdown.
//  The appliation must do whatever is necessary for a clean shutdown and
//  then exit
//=============================================================================
void applicShutdown(void)
{
  printf("Application prepping for shutdown\r\n");  fflush(stdout);

//APPLCODE

/*
  If the bias voltage was being increased, then we are not in a steady state
  and we need to start the variac moving in the opposite direction to insure
  that the bias voltage doe snot increase to a dangerous point during the
  soft reset.  Note that simply stopping the variac is not enough, as it might
  have been asserted because the current setpoint was causing a magnet to ramp
  up (and a soft reset would suspend the ramping, which generally causes the
  bias voltage to suddenly jump without any change to the variac position).

  So, while switching its direction MAY not be necessary, it is the only way
  to insure a SAFE restart short of shutting off the supply (which would cause
  the magnet to dump, which would be even more disruptive then the temporary
  change in output this approach results in).
*/
  if (!(digitalOut & VARIAC_UP_OFF))  {
    SetBits(Next.digitalOut, VARIAC_UP_OFF);
    ClearBits(Next.digitalOut, VARIAC_DOWN_OFF);
    digitalOut = Next.digitalOut;  WriteDigOutputs(digitalOut);
  }


  Exit(0);
}


//=============================================================================
//  This is the highest priority task, usually resumed at a regular interval
//  so that it can do what it needs and sleep again until the start of the
//  next interval.
//=============================================================================
void _task applic(int argc, char **argv)
{
  #define SharedApplLocals
  #include <sharedAppl.c>

//--- add local variable definitions here ---
//APPLCODE


  #define SharedApplLocals2
  #include <sharedAppl.c>



  #define SharedApplInit
  #include <sharedAppl.c>

  BZERO(bitCounts);  // counts for debouncing the digital inputs

//--- add application initialization code here ---
//APPLCODE
  allowSoftReset = true;


  #define SharedApplInit2
  #include <sharedAppl.c>

  lastRampRate = rampRate;
  maxAmpChg = rampRate * (float)SchedMS / 1000.0;
  polarityErrStartTime = biasAdjustTime = biasErrTime = 0;
  lastValLogTime = 0;


//--- Since we don't know what state the polarity switch is in when we ---
//--- first start, we err on the side of caution and behave as if we ---
//--- just started a change (a 50% probability since the polarity ---
//--- could have been the opposite of our default start-up setting) ---
  Next.switchStarted = GetMS();


//--- set averaging period for monitoring bias voltage for errors ---
  setAvgPeriodADC(BIAS_VOLTS, 500);  // 500 ms

  lastBiasVoltsChkMS = GetMS();
  avgdBiasVolts = lastAvgdBiasVolts = readVal[BIAS_VOLTS];
  biasVoltsChg = 0.0;  ignoreBiasErr = false;


//--- The following loop is run at a periodic interval (normally 100 times ---
//--- per second, or every 10ms).  The application process normally ---
//--- consists of reading the current state of the digital and analog ---
//--- inputs, changing the state of the analog and digital outputs to the ---
//--- values computed during the previous pass, then computing new values ---
//--- for the outputs and updating values that outside clients can view ---
//--- (currently accessable as a set of ModBus registers) ---

  for (;;)  {
    #define SharedApplLoopStart
    #include <sharedAppl.c>


  //--- read the current state of the analog and digital Input signals, and ---
  //--- write the values computed at the end of the last pass ---
    digitalOut = Next.digitalOut;  rwIO();

  //--- process next event message (if any) ---
    processMsg(eventsfd, (EventMsg *)buf, sizeof(buf));

  //--- update the current state and determine the next state ---
    updateCurState();  updateNextState();
    updateOutputState();  updateExternals();


    #define SharedApplLoopEnd
    #include <sharedAppl.c>
  }

}


//=============================================================================
//  list of application-specific commands that are normally entered via a
//  user shell
//=============================================================================
//void _task nsclSeriald(int argc, char **argv);

//--- add any application-specific commands here ---
//APPLCODE
const CmdInfo far applCmds[] = {
//  { "misc",        miscTest,       "..." },
  { "cmd",       cmd,         "Application-specific commands" },
  { "state",     showState,   "Show state of controller" }
};

int const NumApplCmds = ( sizeof(applCmds) / sizeof(CmdInfo) );

//-----------------------------------------------------------------------------

#endif
