//----- ps8rackAppl.c ----- 03/14/11 -----(03/05/06)----

// EE Documents bundle # 038

//-----------------------------------------------------------------------------
//  Application specific code for the main application task that is run at
//  a fixed interval and application specific commands run on demand.
//
//  Application code to monitor 8 analog power supplies
//-----------------------------------------------------------------------------
//
//  2011-03-14 - Mark Davis
//               - Suppressed logging of many of the new messages related to
//                 changes in the setpoint until there is an idle period.
//                 Prior to this, the tiem required to generate the messages
//                 was slowing response time too much to make interactive
//                 adjustment unusable (again).
//               - For bipolar type supplies, the logic that makes a change in
//                 the polarity of the setpoint a special case is now disabled.
//                 This was also causing a significant delay that interferred
//                 with interactive adjustment of the setpoint.
//
//  2011-03-07 - Mark Davis
//               - Moved the logging of state changes out of updateCurState()
//                 in to a separate function because updateCurState() had
//                 gotten larger than the 4K max.
//
//  2011-01-11 - Mark Davis
//               - A message is now logged when the active setpoint reaches
//                 a new setpoint goal and the readback gets close to the
//                 new goal.
//
//  2011-01-10 - Mark Davis
//               - Removed DARK_ prefix from text color attributes specified
//                 in showState() (because Mac's show them as black rather than
//                 the specified color).
//
//  2011-01-07 - Mark Davis
//               - Added msg regarding the need for a change in polarity,
//                 including repeating it if the condition persists.
//               - Log msgs for the rampTo0 state are now also repeated if
//                 the condition persists.
//
//  2011-01-06 - Mark Davis
//               - Added logStateSecs and supporting code to periodically log
//                 the state of all the power supplies.
//               - The ramping of the applied setpoint is now based on the
//                 ideal SchedMS value rather than the msSincePrev value.
//                 While the result may not be as close (on average) to a
//                 linear ramp rate, it avoids big (or small), uneven size
//                 jumps due to variations in the timing of the applic loop.
//
//  2011-01-04 - Mark Davis
//               - Added application-specific ApplStateDataSize which provides
//                 the number of bytes needed to save the state of the
//                 application logic (used by the soft reset process).
//
//  2010-12-21 - Mark Davis
//               - Existing read/set errors are now repeated in the log on
//                 a periodic basis and after a new setpoint or an on or off
//                 command is received.  This makes it much easier to look at
//                 the log and see what is causing various unexpected
//                 behaviors.  Part of this change is the addition of the
//                 repeatMsgsSecs value the determines how often persistent
//                 read/set errors are logged.
//               - Once set, the set/read error is not cleared until the
//                 error level drops below the configured threshold.  Before
//                 this it would clear if the readback started ramping in the
//                 direction of the active setpoint, which made for confusing
//                 messages about the read/set error being cleared even if
//                 there was still a huge descrepency bewteeen the two.
//
//  2010-12-20 - Mark Davis
//               - Readded the digital input and output xxxDescr[] strings so
//                 their initial state would be loged, but commented out the
//                 call in the main loop so they would not create redundant
//                 messages (the appl loop currently logs more meaningful msgs
//                 for changes to these values).
//               - A supply is turned off if an OFF command is received and
//                 the supply does not appear to be responding to changes in
//                 the active setpoint (in this case, the ramping to 0 that
//                 preceeds turning a supply off).
//
//  2010-12-14 - Mark Davis
//               - The "ON" string shown in the state screen to indicate the
//                 controller is asserting the ON command is now shown in red
//                 (rather than blue) if there is an pending OFF request.
//
//  2010-10-01 - Mark Davis
//               - Shifted the decimal point by 1 position for the current
//                 set and read values in showState().
//
//  2010-03-07 - Mark Davis
//               - Added "PS is now ON/OFF" msg to the messages logged for
//                 changes in the input states.  Previous to this you could
//                 only tell indirectly if the asserted ON command was working
//                 by the absence of an error msg saying the asserted and
//                 read states were not in agreement (and/or msgs indicating
//                 that a RESET request was ignored because the supply is ON).
//               - Also changed the order state-change msgs are logged to
//                 make more sense.
//
//  2010-02-08 - Mark Davis
//               - The "state" command now shows "???" for the Actual On status
//                 if the supply type is not yet known.
//
//  2009-12-16 - Mark Davis
//               - Simplifications to read/set monitoring (previously the
//                 ramp interlock logic).
//               - Added instantOff settings to allow selected supplies to
//                 ignore the readback and shut off immediately on request.
//               - A Reset request is now ignored ONLY if BOTH the asserted
//                 ON and the ON feedback from the supply are true.  It was
//                 based only on the asserted one, which meant a Reset request
//                 would be ignored for supplies in a confused state where a
//                 reset might have cleared the problem.
//               - An error message is logged if the asserted and feedback ON
//                 states do not agree for more than 1 second.
//
//  2009-12-14 - Mark Davis
//               - Changed order of param files to deal with problems caused
//                 by the order in which they were restored at startup.
//
//  2009-10-26 - Mark Davis
//               - Fixes for set/read error msg logic.
//
//  2009-10-16 - Mark Davis
//               - The newer ramp interlock logic is now disabled by default.
//                 It must be enabled for selected supplies by setting the
//                 ignoreRampIlk_# register to 0 for specific supplies.
//               - A couple of diagnostic/log msgs that were commented out have
//                 been restored
//
//  2009-09-29 - Mark Davis
//               - Refined rampInterlock handling to minimize likelihood of
//                 a false interlock, and also added the averaged reading and
//                 saveSwAmps value to the logged message when it is triggered.
//               - added iRead_#_avgd registers to make monitoring the avgd
//                 readbacks possible.
//               - added ability to disable the rampInterlock logic
//
//  2009-09-23 - Mark Davis
//               - Added log message warning when the current reading while
//                 off if > safeSwAmps_# at the time an on command is received.
//
//  2009-09-22 - Mark Davis
//               - Added per-supply setpoint ramprates
//               - Ramp-to-0 interlock now checks for a problem any time the
//                 readback should be ramping toward 0.  Previously this was
//                 done only for unipolar supplies with an external PLC
//                 managed polarity switch.
//               - Added code to monitor the set/read error and log a message
//                 if the error exceeds 5% of the possible range whenever the
//                 setpoint is stable.
//
//  2009-06-08 - Mark Davis
//               - Changed SchedMS to 20 to reduce overhead
//
//  2009-05-28 - Mark Davis
//               - Added latency as a Modbus register
//
//  2009-04-20 - Mark Davis
//               - removed redundant digital input and output xxxDescr[] msgs
//                 (the applic code handles these more thoroughly).
//
//  2008-03-25 - Mark Davis
//               - fixed problem with Ramp Interlock msg string (the format
//                 string expected values not provided in the printf call).
//               - reduced default setting for minRampRate to 0.250 (amps/sec)
//               - added log messages for start and end of ramp-to-0 state
//
//  2008-03-23 - Mark Davis
//               - finished changes for ramp-to-0 behavior (when an Off command
//                 is received, it now sets the supply output to 0 and waits
//                 for the load current to get near 0).  Also asserts a
//                 ramp interlock if load does not appear to be ramping
//                 toward zero when it should be.
//               - setpoints are now stored in the settings file and restored
//                 on startup
//
//  2008-02-29 - Mark Davis
//               - reset request now acted on again when an interlock exists
//
//  2008-02-25 - Mark Davis
//               - fixed statusBitsAppl so it properly sets the on status bit
//                 for unipolar supplies
//               - the supply type is no longer set unless a type test was
//                 actually run
//
//  2008-02-06 - Mark Davis
//               - updated to latest std structures
//               -
//
//  2008-01-11 - Mark Davis
//               - More corrections:  Modbus register values where not being
//                 updated properly.  Also added the latched interlock status
//                 for each supply to statusBitsAppl.
//
//  2007-12-07 - Mark Davis
//               - corrections to changes for supporting unipolar supply.
//                 Previous version did not assert a reset pulse on command.
//
//  2007-11-28 - Mark Davis
//               - finished changes for supporting both unipolar and bipolar
//                 supplies
//               - added ability to assign a name to each supply
//
//  2007-10-04 - Mark Davis
//               - added max setting limits for each of the 8 supplies
//
//  2007-08-08 - Mark Davis -
//               - changed load voltage default gain (apparently the gain on
//                 the board was changed from the prototype boards, so now
//                 the default gain is appropriate for the newer boards).
//
//  2007-08-07 - Mark Davis -
//               - better defaults for device-level calibration test points
//
//  2007-06-12 - Mark Davis -
//               - corrected code to use correct names for computing scaled
//                 values (and shifted ID's around in the devdata file so
//                 the values from a previous calibration would still be used
//                 for the correct values).
//               - added pscalib option to automatically calculate all the
//                 gains and offsets given the actual voltage and current
//                 for 2 set points.
//
//  2007-03-14 - Mark Davis -
//               - added call to rwIO() (replaces "invisible" code in the
//                 sharedAppl.c file that did this for all applications).
//
//  2007-03-04 - Mark Davis -
//               - reworked X-Macro usage to allow application-specific
//                 register definitions to reside in the same file as the
//                 rest of the application-specific code.
//
//  2006-10-26 - Mark Davis -
//               - started adding actual application-specific logic and
//                 fleshing out structure for all future controllers that
//                 will be based on the new DEBROS RTOS code.
//
//-----------------------------------------------------------------------------

/*
Tests to perform after changing code:

 - Ability to generate a reset pulse when off, with or without an interlock

 - Type test happens and produces correct results for an on, off, or
   (when off) a reset command

 - Active setpoint ramps to 0 for an off command, but supply stays on until
   one of the following happens:
     - The readback drops below maxReadErr setting
     - The readback error condition is asserted
     - An interlock occurs

 - Changes in the applied setpoint happen at the maxRampRate_# setting

 - A readSetErr occurs if both the following are true every cycle for more
   than maxErrTime milliseconds:
     - readSetDiff > maxReadErr
     - readChgRate < minRampRate OR readSetDiff got larger



_____________________________________________________________________

 Outputs:
   Polarity_Reset
   On/Off cmd
   current setpoint

 Inputs:
   Polarity_On/Off feedback
   PS Interlock
   Ext Interlock
   current readback
   voltage readback
_____________________________________________________________________

 EVENTS:

  - On command
  - Off command
  - reset command
  - clear interlock command
  - new setpoint
  - supply type changed

 Conditions/states/values that affect state of nodes:

  - supply type
  - polarity of setpoint != state of polarity switch
  - actual ON/OFF state of power supply
  - requested ON/OFF state of power supply
  - external interlock asserted
  - requested setpoint
  - active setpoint
  - is the readback behaving as expected (stable and close to the setpoint,
    or ramping toward the setpoint)
  -


 Behavioral descriptions:

   Safely turning off a supply:
     - When an OFF request is received, the setpoint is ramped to 0 and the
       readback should then also ramp to 0.  When the readback gets close
       to zero (defined by maxReadErr) if is safe to stop asserting the ON
       command.
     - If the active setpoint is 0, and the readback is not close to 0 and
       is not ramping toward 0 for some period of time, then assert an
       interlock (unless this feature is disabled)

   Determining supply type, reducing risk of operating in wrong mode:
    - Assuming a bipolar supply when a unipolar is connected is dangerous,
      so assume a unipolar supply until testing indicates this is not
      the case.
    - For unipolar supplies, the Polarity_On/Off input will follow the
      state of the Polarity_Reset output
    - For bipolar supplies, the Polarity_On/Off input should follow the
      state of the On/Off output
    - Continually check state of following:
       - Polarity_On/Off input
       - Polarity_Reset output
       - On/Off output
       - unipolarType and bipolarType flags
       If (Polarity_On/Off input != Polarity_Reset output) {
         bipolarType = true; unipolarType = false;  }
       If (unipolarType
           AND (lastTime(Polarity_On/Off input == On/Off output) > threshold))
         unipolarType = false;
    - A test cycle consists of pulsing the Polarity_Reset output while
      the supply is turned off, and observing the Polarity_On/Off input.
      If the input follows the output, then the wiring is for a Unipolar
      supply.
    - If neither supply type flag is true, then assert an interlock condition
    - run test cycle whenever the supply on/off state is changed

   Reset command:
    - Ignored whenever the supply output is not near-zero
    - implemented as a supply-type-test cycle, which pulses the reset output
      to test for unipolar type wiring.

   New setpoint:
    - if unipolarType AND polarity of requested setpoint != state of
      polarity switch:
        - set active setpoint = 0
        - wait for supply output to drop to near-zero
        - change polarity switch to agree with active setpoint
    - assert reset pulse for specified period when requested
_____________________________________________________________________

*/


#include <common.h>

#include <unistd.h>
#include <sys/types.h>



#ifndef  ApplicSource
  #define  ApplicSource  "ps8rackAppl.c"
#endif

#ifndef DidRegDefs
  #define  DidRegDefs
  #include <modbusRegs.c>
#endif
#ifndef RegDefs


//APPLCODE - change to the application name
const char far applName[] = { "PS8-Rack Controller" };
const char ApplAbbrev[] = { "PS8" };

//APPLCODE - add ID to unistd.h and place it here
#define DeviceType  DT_PS8Rack

//APPLCODE - set the # of input and outputs used
#define  NUM_APPL_AIN     16   // analog inputs
#define  NUM_APPL_AOUT     8   // analog outputs
#define  NUM_APPL_DIN     24   // digital inputs
#define  NUM_APPL_DOUT    16   // digital outputs

#define  ADC_MASK_APPL  0x0000FFFFL
#define  DAC_MASK_APPL  0x000000FFL

//APPLCODE - set to state we want outputs in during initialization
#define InitialDigOutState  ( 0 )


#define SharedINC
#include <sharedAppl.c>


#define  NumPS  8  // # of power supply "channels" supported

#define  ISetGainDefault   2.5
#define  IReadGainDefault  2.5
#define  VReadGainDefault  2.5

#define  V_RD_IDX(chan)  ((uint)(chan) << 1)
#define  I_RD_IDX(chan)  (V_RD_IDX(chan) + 1)



void logChgMsgs(ulong curTime);


//=============================================================================
//  definitions related to use of the Digital and Analog I/O signals
//=============================================================================

//=============== digitalIn bit usage ===============
//  ----- These must be shifted by 3 * chan idx -----
#define  EXT_INTLK_OK   0x00000001L  // 0=fault, 1=OK
#define  PS_INTLK_OK    0x00000002L  // 0=fault, 1=OK
#define  PS_IS_ON       0x00000004L  // 0=OFF, 1=ON

#define  INTLK_BITS     0x00000003L  // shift by 3 * chan idx
#define  ALL_INTLK_BITS  033333333L  // easier to read in octal

const uchar  LogicOrderDIn[NUM_APPL_DIN] = {
  0,1,2,  3,4,5,  6,7,8,  9,10,11,  12,13,14,  15,16,17,  18,19,20,  21,22,23 };

#define  PS_IN_BITS(din, chan)     ( ((din) >> ((uint)(chan) * 3)) & 7 )

#define  ExtInterlock(chan)   \
                   BOOL((~digitalIn >> (uint)(chan)*3) & EXT_INTLK_OK)
#define  PSInterlock(chan)    \
                   BOOL((~digitalIn >> (uint)(chan)*3) & PS_INTLK_OK)

#define  ActiveInterlock(chan)  \
                   BOOL((~digitalIn >> (uint)(chan)*3) & INTLK_BITS)


//--- msgs to log automatically when a digital input changes state ---
//--- use NULL for bits that we don't want to log changes for ---
const char far *digInDescr[NUM_APPL_DIN*2] = {
//--0 state descr--         --1 state descr--
  "PS1: Extern Interlock",  "PS1: External Interlock Cleared",
  "PS1: PS Interlock",      "PS1: PS Interlock Cleared",
  "PS1: OFF",               "PS1: ON",

  "PS2: Extern Interlock",  "PS2: External Interlock Cleared",
  "PS2: PS Interlock",      "PS2: PS Interlock Cleared",
  "PS2: OFF",               "PS2: ON",

  "PS3: Extern Interlock",  "PS3: External Interlock Cleared",
  "PS3: PS Interlock",      "PS3: PS Interlock Cleared",
  "PS3: OFF",               "PS3: ON",

  "PS4: Extern Interlock",  "PS4: External Interlock Cleared",
  "PS4: PS Interlock",      "PS4: PS Interlock Cleared",
  "PS4: OFF",               "PS4: ON",

  "PS5: Extern Interlock",  "PS5: External Interlock Cleared",
  "PS5: PS Interlock",      "PS5: PS Interlock Cleared",
  "PS5: OFF",               "PS5: ON",

  "PS6: Extern Interlock",  "PS6: External Interlock Cleared",
  "PS6: PS Interlock",      "PS6: PS Interlock Cleared",
  "PS6: OFF",               "PS6: ON",

  "PS7: Extern Interlock",  "PS7: External Interlock Cleared",
  "PS7: PS Interlock",      "PS7: PS Interlock Cleared",
  "PS7: OFF",               "PS7: ON",

  "PS8: Extern Interlock",  "PS8: External Interlock Cleared",
  "PS8: PS Interlock",      "PS8: PS Interlock Cleared",
  "PS8: OFF",               "PS8: ON"
};


//=============== digitalOut bit usage ===============
#define  ON_BIT         0x00000001L  // shift by 2 * chan idx
#define  RESET_BIT      0x00000002L  // shift by 2 * chan idx (bipolar)
#define  POL_SW_BIT     0x00000002L  // shift by 2 * chan idx (unipolar)

// NOTE: The ON_BIT digital output is is overridden by the EXT_INTLK_OK
//       digital input (this wiring is on the rear panel circuit board
//       that the rear connectors are mounted to).


#define  ActivityLED    0x80000000L

const uchar  LogicOrderDOut[NUM_APPL_DOUT] = {
  0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15 };


#define  TurnOnActivityLED()    ClearBits(Next.digitalOut, ActivityLED)
#define  TurnOffActivityLED()   SetBits(Next.digitalOut, ActivityLED)

#define  PS_OUT_BITS(dout, chan)   ( ((dout) >> ((uint)(chan) << 1)) & 3 )


#define  SetPosPolarity(chan)  SetBits(Next.digitalOut, POL_SW_BIT<<((chan)<<1))
#define  SetNegPolarity(chan)  ClearBits(Next.digitalOut, POL_SW_BIT<<((chan)<<1))

#define  ApplyReset(chan)     SetBits(Next.digitalOut, RESET_BIT<<((chan)<<1))
#define  ClearReset(chan)     ClearBits(Next.digitalOut, RESET_BIT<<((chan)<<1))

#define  TurnSupplyOn(chan)   SetBits(Next.digitalOut, ON_BIT<<((chan)<<1))
#define  TurnSupplyOff(chan)  ClearBits(Next.digitalOut, ON_BIT<<((chan)<<1))

#define  OnIsAsserted(chan)   (digitalOut & ON_BIT<<((chan)<<1))

  //--- msgs to log automatically when a digital output changes state ---
//--- use NULL for bits that we don't want to log changes for ---
const char far *digOutDescr[NUM_APPL_DOUT*2] = {
//--0 state descr--            --1 state descr--
  "PS1: not asserting ON",     "PS1: Assert ON",
  "PS1: not asserting RESET",  "PS1: Asserting RESET",

  "PS2: not asserting ON",     "PS2: Assert ON",
  "PS2: not asserting RESET",  "PS2: Asserting RESET",

  "PS3: not asserting ON",     "PS3: Assert ON",
  "PS3: not asserting RESET",  "PS3: Asserting RESET",

  "PS4: not asserting ON",     "PS4: Assert ON",
  "PS4: not asserting RESET",  "PS4: Asserting RESET",

  "PS5: not asserting ON",     "PS5: Assert ON",
  "PS5: not asserting RESET",  "PS5: Asserting RESET",

  "PS6: not asserting ON",     "PS6: Assert ON",
  "PS6: not asserting RESET",  "PS6: Asserting RESET",

  "PS7: not asserting ON",     "PS7: Assert ON",
  "PS7: not asserting RESET",  "PS7: Asserting RESET",

  "PS8: not asserting ON",     "PS8: Assert ON",
  "PS8: not asserting RESET",  "PS8: Asserting RESET"
};



//=============== Analog inputs ===============
//  inputs 0, 2, 4, 6, 8, 10, 12, 14  are Voltage readbacks
//  inputs 1, 3, 5, 7, 9, 11, 13, 15  are Current readbacks

const uchar  LogicOrderAIn[NUM_APPL_AIN] =
   { 0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15 };


//=============== Analog outputs ===============
//  each output controls the current output of a single supply

const uchar  LogicOrderAOut[NUM_APPL_AOUT] = { 0, 1, 2, 3, 4, 5, 6, 7 };





//=============================================================================
//  List of daemon processes.  The inetd process listens for connection
//  requests on each port in the list and creates a new instance of the
//  specified process to service each new connection.
//=============================================================================
#include <services.h>

//--- service funcs declared the same as standard command funcs:
//     e.g.:   void _task fastdatad(int argc, char **argv);

void _task tsh(int argc, char **argv);


ServiceInfo inetdConf[] = {
//--- add any application-specific inetd-type network services here ---
//--- (add port # names to services.h) ---
//  NAME          PORT           PROTOCOL  FUNC       PRI  -1  Config Params
//  ------------  -------------  --------  ---------- ---  --  ----------------

//--- standard ones defined in sharedAppl.c ---
//{ "console",    TTY0,          ASYNC,    tsh,        10, -1, &consCfg },
//{ "fastdatad",  FASTDATA_PORT, ProtoTCP, fastdatad,   5, -1, NULL },
//{ "modbusTCPd", MODBUS_PORT,   ProtoTCP, modbusTCPd,  6, -1, NULL },
//{ "tsh",        TELNET_PORT,   ProtoTCP, tsh,        10, -1, NULL }

  #define SharedServices
  #include <sharedAppl.c>
};

const int  NumServices = sizeof(inetdConf) / sizeof(ServiceInfo);



//=============================================================================
//  Application-specific data structures and definitions
//=============================================================================

// Each value in the ApplState structure should be throught of as representing
// a "signal" in a clocked logic circuit, with the "applic" task representing
// the logic implemented by the circuit elements such as gates, latches, etc.

// The main "applic" task is the highest priority task and executes the code
// that implements the application logic on a real-time, periodic basis.

//                 ====== IMPORTANT NOTE on LOCKING ======
// Remember to use a semaphore or some other form of locking mechanism to
// protect all access to shared values by more than one task.  The only
// exception to this is when the applic task needs to access a value protected
// by other tasks using only a scheduler/kernel lock (the KLOCK and KUNLOCK
// macros).  Because the applic task is the only task that runs at the highest
// priority, no other task can preempt it, and if the scheduler is locked,
// the applic task cannot preempt another task.  But scheduler locks must be
// kept short to avoid creating large latency in the timely execution of other
// tasks (the applic task in particular).


// ***** WARNING ***** WARNING ***** WARNING ***** WARNING *****
//  Changing the applState structure means that a soft reset will fail!  A
//  change to the size will cause a soft restart to abort and do a CPU
//  restart instead.  And while a change that does NOT result in a new
//  size will not cause the soft restart to abort, it does mean that any
//  shift in the position of values can cause all kinds of mayhem.
// ***** WARNING ***** WARNING ***** WARNING ***** WARNING *****


//===== Most the ApplState values are ones duplicated for each supply =====

typedef struct {
 //-----------------------------------------------------------------------
 // SOFTWARE SETTING INPUTS
 //
 //  These values represent Software Inputs to the application logic.
 //
 //  These values are set in the "Next" instance by tasks OTHER than the
 //  "applic" task to change some value used by the application logic.
 //  Remember that all tasks besides the main "applic" task must use a
 //  scheduler lock to protect reads and writes to these values.
 //
 //  Note that the application task may alter these values (e.g. to change
 //  an invalid value in to a valid one).
 //-----------------------------------------------------------------------
  bool   onRequest;          // unprocessed ON request rcvd
  bool   offRequest;         // unprocessed OFF request rcvd
  ulong  resetRequested;     // the time a reset was requested

  float  requestedAmps;      // last requested setpoint

  bool   trace;              // enables detailed logging of state chgs


 //-----------------------------------------------------------------------
 // READINGS
 //
 //  These values represent "passive" inputs to the application logic.
 //
 //  These values are set in the "Next" instance of ApplState by the applic
 //  task just after reading the state of the hardware inputs.
 //
 //  These values generally come from one of two sources:
 //    - The just-read state of analog and/or digital hardware inputs
 //    - Values maintained by other tasks
 //
 //  While some of these values are simply copies of "external" values,
 //  including them in the ApplState structure allows us to see how
 //  they change between cycles.  Also, some of these values are the result
 //  of applying logic to "external" values.  Applying that logic once at
 //  at the beginning of the applic loop and saving the result eliminates
 //  the need to repeat the same logic in multiple blocks of code later on,
 //  avoiding unnecessary redundancy.
 //-----------------------------------------------------------------------
  bool   psIsOn;             // ON status from PS asserted

  float  actualAmps;         // last-known amperage readback
  float  avgdAmps;           // avgd version of actualAmps
  float  actualVolts;        // last-known voltage readback

  float  readChgRate;        // extrapolated chg in readback/sec

  bool   psInterlock;        // active interlock from PS
  bool   externInterlock;    // active interlock from other source

  bool   needPolarityChg;    // asserted polarity != requested polarity

  bool   testIn;
  bool   testOut;


 //-----------------------------------------------------------------------
 // OUTPUTS from UN-CLOCKED (asynchronous, gate-level) LOGIC ELEMENTS
 //
 //  These values are outputs from one of the blocks of code in the applic
 //  task that represent a non-clocked logic element (such as a gate).
 //
 //  Each value here is set by one (and ONLY one) of the blocks of code in
 //  the applic task.  This group contains ONLY values that are set by a
 //  block of code that represents a NON-clocked logic element.
 //
 //  These values are set in the "Next" instance of ApplState by the applic
 //  task just after reading the state of the hardware inputs.  These values
 //  must  NOT be altered by other tasks (although they can read them as long
 //  as they use appropriate locking to do so).
 //
 //  Because these values represent the outputs of un-clocked logic elements,
 //  and such elements often need to be cascaded (the outputs used as inputs
 //  to other non-clocked elements), the result is often dependent on the
 //  order of execution of the blocks of code.
 //-----------------------------------------------------------------------
  ulong  readCheckTime;      // time (in ms) many of the following last updated

  float  readCheckAmps;      // amp reading at readCheckTime

  bool   setpointChgd;       // the active setpoint changed between passes
  bool   nearZero;           // reading below maxReadErr
  bool   newGoal;            // readback not yet close to latest setpoint goal

  ulong  nextRepeatMsgsSecs; // next time to re-log a msg for an existing state

  bool   setReadErr;         // readback not following setpoint
  float  smallestErr;        // smallest err been since last setpoint chg
  bool   acceptableErr;      // read/set error < max when last cheched

  bool   ramping;            // set/read error reduction rate >= minRampRate
  ulong  lastReadOKTime;     // last time err < max or chg rate >= min
  bool   readErrIntlk;       // Interlocked due to a setReadErr

  bool   onCtlErr;           // psIsOn != assertOn for a significant period
  ulong  lastOnOKTime;       // last time psIsOn == assertOn

  bool   activeIntlk;        // one of the interlock states is active



 //-----------------------------------------------------------------------
 // OUTPUTS from CLOCKED (latched inputs and outputs) LOGIC ELEMENTS
 //
 //  These values are outputs from one of the blocks of code in the applic
 //  task that represent a clocked logic element (i.e. it acts like a
 //  clocked latch but is in software and the clock is one pass through
 //  the main application loop).
 //
 //  Each value here is set by one (and ONLY one) of the blocks of code in
 //  the applic task.  This group contains only values that are set by a
 //  block of code that represents a CLOCKED logic element.
 //
 //  These values are set in the "Next" instance of ApplState by the applic
 //  task AFTER all the other ApplState values are set.  Because they represent
 //  only the clocked logic elements, and each block of code reads only values
 //  from the Cur instance, the blocks can be executed in any order and the
 //  result will always be the same.
 //
 //  These values are used as inputs by other blocks and/or to set values
 //  used by other tasks (such as values read by the ModBus server task).
 //
 //  Again, these values must NOT be altered by other tasks (although they
 //  can read them as long as they use appropriate locking to do so).
 //-----------------------------------------------------------------------
  ulong  testStarted;        // time the test/reset pulse FSA was activated
  bool   onPending;          // on request waiting for type test to finish

  bool   logicallyOn;        // actively controlling supply output
  bool   assertOn;           // ON cmd to PS is asserted

  ulong  rampTo0Started;     // time (in ms) started ramping to 0
  bool   logSetReached;      // log a msg when read close to cur stable goal

  bool   prepForOff;         // waiting for near-0 readback before turning off
  bool   offPending;         // reached end of ramp-to-0 cycle

  bool   assertPosPolarity;  // assert positive polarity for unipolar supply
  ulong  polChgStarted;      // externally ctld polarity switch told to chg

  float  setpointGoal;       // value we are ramping activeSetpoint toward
  float  activeSetpoint;     // value actually being applied to control the PS

  bool   interlocked;        // existing or uncleared interlock condition

  bool   cantBeUnipolar;     // wiring not for Unipolar supply
  uint   supplyType;         // Unipolar or Bipolar supply

} PSState;



typedef struct  {
  PSState  ps[NumPS];

 //-----------------------------------------------------------------------
 // SOFTWARE SETTING INPUTS
 //
 //  These values represent Software Inputs to the application logic.
 //
 //  These values are set in the "Next" instance by tasks OTHER than the
 //  "applic" task to change some value used by the application logic.
 //  Remember that all tasks besides the main "applic" task must use a
 //  scheduler lock to protect reads and writes to these values.
 //
 //  Note that the application task may alter these values (e.g. to change
 //  an invalid value in to a valid one).
 //-----------------------------------------------------------------------


 //-----------------------------------------------------------------------
 // READINGS
 //
 //  These values represent "passive" inputs to the application logic.
 //
 //  These values are set in the "Next" instance by the applic task just
 //  after reading the state of the hardware inputs.
 //
 //  These values generally come from one of two sources:
 //    - The just-read state of analog and/or digital hardware inputs
 //    - Values maintained by other tasks
 //
 //  While some of these values are simply copies of "external" values,
 //  including them in the ApplState structure allows us to see how
 //  they change between cycles.  Also, some of these values are the result
 //  of applying logic to "external" values.  Applying that logic once at
 //  at the beginning of the applic loop and saving the result eliminates
 //  the need to repeat the same logic in multiple blocks of code later on,
 //  avoiding unnecessary redundancy.
 //-----------------------------------------------------------------------
  ulong  digitalIn;

  ulong processTime;

 //-----------------------------------------------------------------------
 // OUTPUTS from UN-CLOCKED (asynchronous, gate-level) LOGIC ELEMENTS
 //
 //  These values are outputs from one of the blocks of code in the applic
 //  task that represent a non-clocked logic element (such as a gate).
 //
 //  Each value here is set by one (and ONLY one) of the blocks of code in
 //  the applic task.  This group contains ONLY values that are set by a
 //  block of code that represents a NON-clocked logic element.
 //
 //  These values are set in the "Next" instance by the applic task just
 //  after reading the state of the hardware inputs.  These values must
 //  NOT be altered by other tasks (although they can read them as long
 //  as they use appropriate locking to do so).
 //
 //  Because these values represent the outputs of un-clocked logic elements,
 //  and such elements often need to be cascaded (the outputs used as inputs
 //  to other non-clocked elements), the result is often dependent on the
 //  order of execution of the blocks of code.
 //-----------------------------------------------------------------------


 //-----------------------------------------------------------------------
 // OUTPUTS from CLOCKED (latched inputs and outputs) LOGIC ELEMENTS
 //
 //  These values are outputs from one of the blocks of code in the applic
 //  task that represent a clocked logic element (such as latches).
 //
 //  Each value here is set by one (and ONLY one) of the blocks of code in
 //  the applic task.  This group contains only values that are set by a
 //  block of code that represents a CLOCKED logic element.
 //
 //  These values are set in the "Next" instance by the applic task AFTER
 //  all the other ApplState values are set.  Because they represent only
 //  the clocked logic elements, and each block of code reads only values
 //  from the Cur instance, the blocks can be executed in any order and the
 //  result will always be the same.
 //
 //  These values are used as inputs by other blocks and/or to set values
 //  used by other tasks (such as values read by the ModBus server task).
 //
 //  Again, these values must NOT be altered by other tasks (although they
 //  can read them as long as they use appropriate locking to do so).
 //-----------------------------------------------------------------------

  ulong  digitalOut;  // state for digital outputs

} ApplState;


ApplState prevState, curState, nextState;

#define  Prev  prevState
#define  Cur   curState
#define  Next  nextState


const uint ApplStateDataSize =  ( sizeof(ApplState)*3 + sizeof(setVal) );


//--- values for ApplState.supplyType ---
#define  ST_Unknown    0
#define  ST_Bipolar    1
#define  ST_Unipolar   2


bool ignoreResetReqs[NumPS];

ulong  lastLogStateSecs;

const char far *TypeDesc[] = { "Unknown", "Bipolar", "Unipolar" };

const char TypeDescCh[3] = { '?', 'B', 'U' };



#define  MAX_PS_NAME_LEN  12

char far psNames[NumPS][MAX_PS_NAME_LEN+1];

ulong msSincePrev;


//=============================================================================
//  Defintions for variables that are accessable as ModBus registers
//=============================================================================
#else  // end of before-RegDefs portion

/*
  The purpose of this section is to define a set of C variables that are
  accessable as ModBus registers.  This approach avoids the need to use
  special functions or macros throughout the code when accessing register
  values, which keeps the application code fast and simple.

  The model used provides support for multiple logical devices.  The 2nd
  parameter in each register definition macro indicates if one instanace
  applies to all logical devices (specified by an "x" for "don't care"), or
  if each logical device requires its own instance (specified by an under-
  score "_" to signify "fill in the blank").  In the 1st case, the variable
  is defined as a scalar.  In the 2nd case, it is defined as an array.

  The definition macros support data types (floats and longs in particular)
  that require more than one register to store them.  To avoid the possibility
  of corrupted values, all reads of and writes to such values must be done
  using the multi-register commands and including all the registers that
  make up the value in the command.  Differences in register order between
  different client types is handled by a table that automatically swaps
  register values as needed based on the client type.

  The definition for each writable register also includes the function to
  call when the value is set via the ModBus interface.  Many registers can
  use the default setXXX() functions, but you can specify your own to provide
  any special logic needed (including ignoring the new setting if invalid).

  For registers, the format for the function definitions is:

    bool mySetFunc(SetRegFuncParams);

       where SetRegFuncParams is currently defined as:
              void *info, uchar chanNum, void *newVal, fd_t eventsfd

  For coils, the format for the function definitions is:

     bool (*SetCoilFunc)(SetCoilFuncParams);

         where SetCoilFuncParams is currently define as:
               void *info, uchar chanNum, uchar newVal, fd_t eventsfd

  The parameters to the call are as follows:
    *info:     A pointer to the Coil's/Register's CoilInfo/RegInfo structure
    chanNum:   The logical device # (0 is the 1st)
    *newVal:   A pointer to the new value requested in the Modbus packet
               (changed to native-byte order where needed)
    eventsfd:  The fd for the FIFO used to send msgs to the main application
               process

  The function returns true if the new value was accepted, false if not.

  The register numbers for multiple "channels" or logical devices are
  handled automatically based on the xxChanMult values.  These values specify
  the offset between a register value for one logical device and the "same"
  register for the Next.

  Example:

    Assume the following definitions:
      #define  NumLogicalDevices  2
      #define  R4ChanMult  100
      REG4( motorPos,       _, 40050, F32, RW, setMotorPos )

    Then we would get:
      global variable definition:  float motorPos[2];
      register pair:  40050,40051  refers to:  motorPos[0]
      register pair:  40150.40151  refers to:  motorPos[1]

    If instead we had specified:
      REG4( motorPos,       x, 40050, F32, RW, setMotorPos )

    Then we would get:
      global variable definition:  float motorPos;
      register pairs:  40050,40051 & 40150,40151  BOTH refer to:  motorPos


  To keep the definitions as simple as possible, and avoid requiring multiple
  macros for each register definition, each macro is used more than once,
  with a different definition (#define) each time (a practice generally
  referred to as X-Macros).  While trying to follow the order of inclusion
  of files (including a recursive reference back to this file) can be rather
  confusing, the X-Macro method keeps the source code simple to read and
  maintain by doing everything with a single macro for each register.

  ==========================================================================
                             !!! WARNING !!!
  ==========================================================================
    Each sequence of register definitions MUST be contiguous!  There must
    NOT be any holes (i.e. no unused register #'s between used ones) and no
    overlap!
-----------------------------------------------------------------------------*/

#ifndef NumLogicalDevices
  #define  NumLogicalDevices  1
#endif

//-----------------------------------------------------------------------------
#ifndef NumC0Chans
  #define NumC0Chans_  NumLogicalDevices
  #define C0ChanMult_  100
#endif

#ifdef COIL0

COIL0( DoReboot,       x, 1, setCoil )
//COIL0( Coil00002,      _, 2, setCoil )
//COIL0( Coil00003,      _, 3, setCoil )

#undef COIL0
#endif


//-----------------------------------------------------------------------------
#ifndef NumR3Chans
  #define NumR3Chans_  NumLogicalDevices
  #define R3ChanMult_  100
#endif

#ifdef REG3

REG3( dlmVersTime,    x, 30001, U32 )
REG3( rtosVersTime,   x, 30003, U32 )
REG3( applVersTime,   x, 30005, U32 )
REG3( devType_,       x, 30007, U16 )
REG3( deviceID_,      x, 30008, U32 )
REG3( upSecs,         x, 30010, U32 )
REG3( percentIdle_,   x, 30012, U16 )

REG3( numChans,       x, 30013, U16 )
REG3( coilChanMult,   x, 30014, U16 )
REG3( reg3ChanMult,   x, 30015, U16 )
REG3( reg4ChanMult,   x, 30016, U16 )

REG3( digInBits,      x, 30017, U32 )

REG3( statusBitsStdA, x, 30019, U16 )
REG3( statusBitsStdB, x, 30020, U16 )
REG3( statusBitsAppl, x, 30021, U32 )

REG3( iRead_1,        x, 30023, F32 )  /* current (amperage) readback */
REG3( vRead_1,        x, 30025, F32 )  /* voltage readback */

REG3( iRead_2,        x, 30027, F32 )  /* current (amperage) readback */
REG3( vRead_2,        x, 30029, F32 )  /* voltage readback */

REG3( iRead_3,        x, 30031, F32 )  /* current (amperage) readback */
REG3( vRead_3,        x, 30033, F32 )  /* voltage readback */

REG3( iRead_4,        x, 30035, F32 )  /* current (amperage) readback */
REG3( vRead_4,        x, 30037, F32 )  /* voltage readback */

REG3( iRead_5,        x, 30039, F32 )  /* current (amperage) readback */
REG3( vRead_5,        x, 30041, F32 )  /* voltage readback */

REG3( iRead_6,        x, 30043, F32 )  /* current (amperage) readback */
REG3( vRead_6,        x, 30045, F32 )  /* voltage readback */

REG3( iRead_7,        x, 30047, F32 )  /* current (amperage) readback */
REG3( vRead_7,        x, 30049, F32 )  /* voltage readback */

REG3( iRead_8,        x, 30051, F32 )  /* current (amperage) readback */
REG3( vRead_8,        x, 30053, F32 )  /* voltage readback */

REG3( statusBitsAppl2,x, 30055, U32 )  /* more per-chan status bits */

REG3( latency,        x, 30057, U16 )  // actual vs scheduled appl start in ms
REG3( tcpTickTime,    x, 30058, U16 )  // # ms inside tcp_tick() call
REG3( applLoopTime,   x, 30059, U16 )  // # ms used by applic task loop

REG3( iRead_1_avgd,   x, 30060, F32 )  /* avgd readback */
REG3( iRead_2_avgd,   x, 30062, F32 )  /* avgd readback */
REG3( iRead_3_avgd,   x, 30064, F32 )  /* avgd readback */
REG3( iRead_4_avgd,   x, 30066, F32 )  /* avgd readback */
REG3( iRead_5_avgd,   x, 30068, F32 )  /* avgd readback */
REG3( iRead_6_avgd,   x, 30070, F32 )  /* avgd readback */
REG3( iRead_7_avgd,   x, 30072, F32 )  /* avgd readback */
REG3( iRead_8_avgd,   x, 30074, F32 )  /* avgd readback */

REG3( kernelLoopTime, x, 30076, U16 )  // # ms used by last kernel pass

#undef REG3
#endif


#ifdef RegMacros
//===== application-specific defintions related to register values =====
  #define  IREADAVGD(idx)  REG_ARRAY(float, iRead_1_avgd, idx)

//WARNING: Cannot use REG_ARRAY() macro since they are in pairs!
  #define  IREAD(idx)  ( *((float *)((&iRead_1_RegInfo + (idx<<1))->regVal)) )
  #define  VREAD(idx)  ( *((float *)((&vRead_1_RegInfo + (idx<<1))->regVal)) )


//===== bit usage for statusBitsAppl =====
//---- the first 24 bits are just the digitalIn bits ---
#define  EXT_INTLK_OK_1   0x00000001L  // 0=fault, 1=OK
#define  PS_INTLK_OK_1    0x00000002L  // 0=fault, 1=OK
#define  PS_IS_ON_1       0x00000004L  // 0=OFF, 1=ON

#define  EXT_INTLK_OK_2   0x00000008L  // 0=fault, 1=OK
#define  PS_INTLK_OK_2    0x00000010L  // 0=fault, 1=OK
#define  PS_IS_ON_2       0x00000020L  // 0=OFF, 1=ON

#define  EXT_INTLK_OK_3   0x00000040L  // 0=fault, 1=OK
#define  PS_INTLK_OK_3    0x00000080L  // 0=fault, 1=OK
#define  PS_IS_ON_3       0x00000100L  // 0=OFF, 1=ON

#define  EXT_INTLK_OK_4   0x00000200L  // 0=fault, 1=OK
#define  PS_INTLK_OK_4    0x00000400L  // 0=fault, 1=OK
#define  PS_IS_ON_4       0x00000800L  // 0=OFF, 1=ON

#define  EXT_INTLK_OK_5   0x00001000L  // 0=fault, 1=OK
#define  PS_INTLK_OK_5    0x00002000L  // 0=fault, 1=OK
#define  PS_IS_ON_5       0x00004000L  // 0=OFF, 1=ON

#define  EXT_INTLK_OK_6   0x00008000L  // 0=fault, 1=OK
#define  PS_INTLK_OK_6    0x00010000L  // 0=fault, 1=OK
#define  PS_IS_ON_6       0x00020000L  // 0=OFF, 1=ON

#define  EXT_INTLK_OK_7   0x00040000L  // 0=fault, 1=OK
#define  PS_INTLK_OK_7    0x00080000L  // 0=fault, 1=OK
#define  PS_IS_ON_7       0x00100000L  // 0=OFF, 1=ON

#define  EXT_INTLK_OK_8   0x00200000L  // 0=fault, 1=OK
#define  PS_INTLK_OK_8    0x00400000L  // 0=fault, 1=OK
#define  PS_IS_ON_8       0x00800000L  // 0=OFF, 1=ON

//--- and the last 8 are set based on the latched interlock states ---
#define  PS_INTLKD_1      0x0100000L   // PS 1 is interlocked
#define  PS_INTLKD_2      0x0200000L   // PS 2 is interlocked
#define  PS_INTLKD_3      0x0400000L   // PS 3 is interlocked
#define  PS_INTLKD_4      0x0800000L   // PS 4 is interlocked
#define  PS_INTLKD_5      0x1000000L   // PS 5 is interlocked
#define  PS_INTLKD_6      0x2000000L   // PS 6 is interlocked
#define  PS_INTLKD_7      0x3000000L   // PS 7 is interlocked
#define  PS_INTLKD_8      0x4000000L   // PS 8 is interlocked

//===== bit usage for statusBitsAppl2 =====
#define  PS_RAMP0_1       0x00000001L  // 1=ramping to zero
#define  PS_RAMP0_2       0x00000002L  // 1=ramping to zero
#define  PS_RAMP0_3       0x00000004L  // 1=ramping to zero
#define  PS_RAMP0_4       0x00000008L  // 1=ramping to zero
#define  PS_RAMP0_5       0x00000010L  // 1=ramping to zero
#define  PS_RAMP0_6       0x00000020L  // 1=ramping to zero
#define  PS_RAMP0_7       0x00000040L  // 1=ramping to zero
#define  PS_RAMP0_8       0x00000080L  // 1=ramping to zero

#endif


//-----------------------------------------------------------------------------
#ifndef NumR4Chans
  #define NumR4Chans_  NumLogicalDevices
  #define R4ChanMult_  200
#endif

#ifdef REG4

REG4( timeZone_,      x, 40001, S16, RW, MB_timeZone )
REG4( curTime,        x, 40002, U32, RW, setU32 )

REG4( ipAddr,         x, 40004, U32, RW, setU32 )
REG4( netmask,        x, 40006, U32, RW, setU32 )
REG4( gateway,        x, 40008, U32, RW, setU32 )

REG4( setACoil,       _, 40010, U16, RW, setU16 )
REG4( resetACoil,     _, 40011, U16, RW, setU16 )

REG4( digOutBits,     x, 40012, U32, RW, setU32 )

REG4( cmdBitsStd,     x, 40014, U32, RW, MB_cmdBitsStd )
REG4( cmdBitsAppl,    x, 40016, U32, RW, MB_cmdBitsAppl )

REG4( iSet_1,         x, 40018, F32, RW, MB_iSet )
REG4( iSet_2,         x, 40020, F32, RW, MB_iSet )
REG4( iSet_3,         x, 40022, F32, RW, MB_iSet )
REG4( iSet_4,         x, 40024, F32, RW, MB_iSet )
REG4( iSet_5,         x, 40026, F32, RW, MB_iSet )
REG4( iSet_6,         x, 40028, F32, RW, MB_iSet )
REG4( iSet_7,         x, 40030, F32, RW, MB_iSet )
REG4( iSet_8,         x, 40032, F32, RW, MB_iSet )

REG4( iSet_min_1,     x, 40034, F32, RW, setF32 )
REG4( iSet_min_2,     x, 40036, F32, RW, setF32 )
REG4( iSet_min_3,     x, 40038, F32, RW, setF32 )
REG4( iSet_min_4,     x, 40040, F32, RW, setF32 )
REG4( iSet_min_5,     x, 40042, F32, RW, setF32 )
REG4( iSet_min_6,     x, 40044, F32, RW, setF32 )
REG4( iSet_min_7,     x, 40046, F32, RW, setF32 )
REG4( iSet_min_8,     x, 40048, F32, RW, setF32 )

REG4( iSet_max_1,     x, 40050, F32, RW, setF32 )
REG4( iSet_max_2,     x, 40052, F32, RW, setF32 )
REG4( iSet_max_3,     x, 40054, F32, RW, setF32 )
REG4( iSet_max_4,     x, 40056, F32, RW, setF32 )
REG4( iSet_max_5,     x, 40058, F32, RW, setF32 )
REG4( iSet_max_6,     x, 40060, F32, RW, setF32 )
REG4( iSet_max_7,     x, 40062, F32, RW, setF32 )
REG4( iSet_max_8,     x, 40064, F32, RW, setF32 )

REG4( maxReadErr_1,   x, 40066, F32, RW, MB_maxReadErr )
REG4( maxReadErr_2,   x, 40068, F32, RW, MB_maxReadErr )
REG4( maxReadErr_3,   x, 40070, F32, RW, MB_maxReadErr )
REG4( maxReadErr_4,   x, 40072, F32, RW, MB_maxReadErr )
REG4( maxReadErr_5,   x, 40074, F32, RW, MB_maxReadErr )
REG4( maxReadErr_6,   x, 40076, F32, RW, MB_maxReadErr )
REG4( maxReadErr_7,   x, 40078, F32, RW, MB_maxReadErr )
REG4( maxReadErr_8,   x, 40080, F32, RW, MB_maxReadErr )

REG4( polChgDelay,    x, 40082, U16, RW, setU16 )
REG4( resetGraceTime, x, 40083, U16, RW, setU16 )

REG4( unused_084,     x, 40084, F32, RW, setF32 )

REG4( maxRampRate_1,  x, 40086, F32, RW, setF32 )
REG4( maxRampRate_2,  x, 40088, F32, RW, setF32 )
REG4( maxRampRate_3,  x, 40090, F32, RW, setF32 )
REG4( maxRampRate_4,  x, 40092, F32, RW, setF32 )
REG4( maxRampRate_5,  x, 40094, F32, RW, setF32 )
REG4( maxRampRate_6,  x, 40096, F32, RW, setF32 )
REG4( maxRampRate_7,  x, 40098, F32, RW, setF32 )
REG4( maxRampRate_8,  x, 40100, F32, RW, setF32 )

REG4( ilkIfReadErr_1, x, 40102, U16, RW, setU16 )
REG4( ilkIfReadErr_2, x, 40103, U16, RW, setU16 )
REG4( ilkIfReadErr_3, x, 40104, U16, RW, setU16 )
REG4( ilkIfReadErr_4, x, 40105, U16, RW, setU16 )
REG4( ilkIfReadErr_5, x, 40106, U16, RW, setU16 )
REG4( ilkIfReadErr_6, x, 40107, U16, RW, setU16 )
REG4( ilkIfReadErr_7, x, 40108, U16, RW, setU16 )
REG4( ilkIfReadErr_8, x, 40109, U16, RW, setU16 )

REG4( minRampRate_1,  x, 40110, F32, RW, setF32 )
REG4( minRampRate_2,  x, 40112, F32, RW, setF32 )
REG4( minRampRate_3,  x, 40114, F32, RW, setF32 )
REG4( minRampRate_4,  x, 40116, F32, RW, setF32 )
REG4( minRampRate_5,  x, 40118, F32, RW, setF32 )
REG4( minRampRate_6,  x, 40120, F32, RW, setF32 )
REG4( minRampRate_7,  x, 40122, F32, RW, setF32 )
REG4( minRampRate_8,  x, 40124, F32, RW, setF32 )

REG4( maxErrTime_1,   x, 40126, U16, RW, setU16 )
REG4( maxErrTime_2,   x, 40127, U16, RW, setU16 )
REG4( maxErrTime_3,   x, 40128, U16, RW, setU16 )
REG4( maxErrTime_4,   x, 40129, U16, RW, setU16 )
REG4( maxErrTime_5,   x, 40130, U16, RW, setU16 )
REG4( maxErrTime_6,   x, 40131, U16, RW, setU16 )
REG4( maxErrTime_7,   x, 40132, U16, RW, setU16 )
REG4( maxErrTime_8,   x, 40133, U16, RW, setU16 )

REG4( repeatMsgsSecs, x, 40134, U32, RW, MB_repeatMsgsSecs)
REG4( logStateSecs,   x, 40136, U32, RW, setU32)

#undef REG4
#endif

#ifdef RegMacros
//===== application-specific defintions related to register values =====

  #define  ISET(idx)              REG_ARRAY(float, iSet_1, idx)
  #define  ISET_MIN(idx)          REG_ARRAY(float, iSet_min_1, idx)
  #define  ISET_MAX(idx)          REG_ARRAY(float, iSet_max_1, idx)
  #define  MaxReadErr(idx)        REG_ARRAY(float, maxReadErr_1, idx)
  #define  MaxRampRate(idx)       REG_ARRAY(float, maxRampRate_1, idx)
  #define  IlkIfReadErr(idx)      REG_ARRAY(uint, ilkIfReadErr_1, idx)
  #define  MinRampRate(idx)       REG_ARRAY(uint, minRampRate_1, idx)
  #define  MaxErrTime(idx)        REG_ARRAY(uint, maxErrTime_1, idx)

//--- bit usage for cmdBitsAppl[] (4-bits x 8 supplies = 32 bits) ---
  #define  RESET_PS      1  // asserting PS Reset cmd
  #define  TURN_ON_PS    2  // turn supply ON
  #define  TURN_OFF_PS   4  // turn supply OFF
  #define  UNUSED_CMD    8  // unused

#endif

//--------------------------------------------------------------
#undef RegMacros

#endif
#ifndef RegDefs // start of after-RegDefs portion


//=============================================================================
//  defintions for "parameterized" data values - lists of values that we want
//  to be "persistent" (i.e. saved to a file and restored on startup)
//=============================================================================


//--- parameter ID's for data in Device Data/calibration file ---
//--------------------------------------------------------------------------
// WARNING:  ID's can be added, but existing ID's (and the format of the
//           data the represent) should NEVER change!  NOTE:  Existing ones
//           can be "eliminated" by changing their name and no longer
//           using them, but never REUSE an old one for something new!!!
//--------------------------------------------------------------------------
//  NOTE:  ID's >= 1000 are rsvd for values shared by most applications
//--------------------------------------------------------------------------
const ParamData far  devDataParams[] = {
//                ID   ---value---  --RESTORE_FUNC--
  PERSISTENT_VAR(  7,  iSet_min_1,       NULL ),
  PERSISTENT_VAR(  8,  iSet_min_2,       NULL ),
  PERSISTENT_VAR(  9,  iSet_min_3,       NULL ),
  PERSISTENT_VAR( 10,  iSet_min_4,       NULL ),
  PERSISTENT_VAR( 11,  iSet_min_5,       NULL ),
  PERSISTENT_VAR( 12,  iSet_min_6,       NULL ),
  PERSISTENT_VAR( 13,  iSet_min_7,       NULL ),
  PERSISTENT_VAR( 14,  iSet_min_8,       NULL ),

  PERSISTENT_VAR( 15,  iSet_max_1,       NULL ),
  PERSISTENT_VAR( 16,  iSet_max_2,       NULL ),
  PERSISTENT_VAR( 17,  iSet_max_3,       NULL ),
  PERSISTENT_VAR( 18,  iSet_max_4,       NULL ),
  PERSISTENT_VAR( 19,  iSet_max_5,       NULL ),
  PERSISTENT_VAR( 20,  iSet_max_6,       NULL ),
  PERSISTENT_VAR( 21,  iSet_max_7,       NULL ),
  PERSISTENT_VAR( 22,  iSet_max_8,       NULL ),

  PERSISTENT_VAR( 23,  maxReadErr_1,     NULL ),
  PERSISTENT_VAR( 24,  maxReadErr_2,     NULL ),
  PERSISTENT_VAR( 25,  maxReadErr_3,     NULL ),
  PERSISTENT_VAR( 26,  maxReadErr_4,     NULL ),
  PERSISTENT_VAR( 27,  maxReadErr_5,     NULL ),
  PERSISTENT_VAR( 28,  maxReadErr_6,     NULL ),
  PERSISTENT_VAR( 29,  maxReadErr_7,     NULL ),
  PERSISTENT_VAR( 30,  maxReadErr_8,     NULL ),

  PERSISTENT_VAR( 31,  polChgDelay,      NULL ),

  PERSISTENT_TBL( 32,  psNames, MAX_PS_NAME_LEN+1, NULL ),

  PERSISTENT_VAR( 33,  resetGraceTime,   NULL ),

//  PERSISTENT_VAR( 34,  minRampRate,  NULL ),

  PERSISTENT_VAR( 35,  maxRampRate_1,    NULL ),
  PERSISTENT_VAR( 36,  maxRampRate_2,    NULL ),
  PERSISTENT_VAR( 37,  maxRampRate_3,    NULL ),
  PERSISTENT_VAR( 38,  maxRampRate_4,    NULL ),
  PERSISTENT_VAR( 39,  maxRampRate_5,    NULL ),
  PERSISTENT_VAR( 40,  maxRampRate_6,    NULL ),
  PERSISTENT_VAR( 41,  maxRampRate_7,    NULL ),
  PERSISTENT_VAR( 42,  maxRampRate_8,    NULL ),

  PERSISTENT_VAR( 43,  ilkIfReadErr_1,   NULL ),
  PERSISTENT_VAR( 44,  ilkIfReadErr_2,   NULL ),
  PERSISTENT_VAR( 45,  ilkIfReadErr_3,   NULL ),
  PERSISTENT_VAR( 46,  ilkIfReadErr_4,   NULL ),
  PERSISTENT_VAR( 47,  ilkIfReadErr_5,   NULL ),
  PERSISTENT_VAR( 48,  ilkIfReadErr_6,   NULL ),
  PERSISTENT_VAR( 49,  ilkIfReadErr_7,   NULL ),
  PERSISTENT_VAR( 50,  ilkIfReadErr_8,   NULL ),

  PERSISTENT_VAR( 51,  minRampRate_1,    NULL ),
  PERSISTENT_VAR( 52,  minRampRate_2,    NULL ),
  PERSISTENT_VAR( 53,  minRampRate_3,    NULL ),
  PERSISTENT_VAR( 54,  minRampRate_4,    NULL ),
  PERSISTENT_VAR( 55,  minRampRate_5,    NULL ),
  PERSISTENT_VAR( 56,  minRampRate_6,    NULL ),
  PERSISTENT_VAR( 57,  minRampRate_7,    NULL ),
  PERSISTENT_VAR( 58,  minRampRate_8,    NULL ),

  PERSISTENT_VAR( 59,  maxErrTime_1,     NULL ),
  PERSISTENT_VAR( 60,  maxErrTime_2,     NULL ),
  PERSISTENT_VAR( 61,  maxErrTime_3,     NULL ),
  PERSISTENT_VAR( 62,  maxErrTime_4,     NULL ),
  PERSISTENT_VAR( 63,  maxErrTime_5,     NULL ),
  PERSISTENT_VAR( 64,  maxErrTime_6,     NULL ),
  PERSISTENT_VAR( 65,  maxErrTime_7,     NULL ),
  PERSISTENT_VAR( 66,  maxErrTime_8,     NULL ),

  PERSISTENT_VAR( 67,  repeatMsgsSecs,   NULL),
  PERSISTENT_VAR( 68,  logStateSecs,     NULL),

  #define SharedDevParams
  #include <sharedAppl.c>

  END_PARAM_DATA
};

//----------------------------------------------------
RESTORE_FUNC(iSet_X);

const ParamData far  settingsParams[] = {
//APPLDATA
//                ID   ---value---  --RESTORE_FUNC--
  PERSISTENT_VAR(  1,  iSet_1,       PVRF(iSet_X) ),
  PERSISTENT_VAR(  2,  iSet_2,       PVRF(iSet_X) ),
  PERSISTENT_VAR(  3,  iSet_3,       PVRF(iSet_X) ),
  PERSISTENT_VAR(  4,  iSet_4,       PVRF(iSet_X) ),
  PERSISTENT_VAR(  5,  iSet_5,       PVRF(iSet_X) ),
  PERSISTENT_VAR(  6,  iSet_6,       PVRF(iSet_X) ),
  PERSISTENT_VAR(  7,  iSet_7,       PVRF(iSet_X) ),
  PERSISTENT_VAR(  8,  iSet_8,       PVRF(iSet_X) ),

  END_PARAM_DATA
};


//-----------------------------------------------------------------
//  Application-specific values needed to restore the state of the
//  controller following a "soft" reset (i.e. a restart without a
//  CPU reset)
//-----------------------------------------------------------------
const ParamData far  ctlrStateParams[] = {
// sample values from sharedAppl.c, common to all applications
//                   --ID--  ---value,granularity---  --RESTORE_FUNC--
//    PERSISTENT_VARx(  1000,  prevState,               NULL ),
//    PERSISTENT_VARx(  1001,  curState,                NULL ),
//    PERSISTENT_VARx(  1002,  nextState,               NULL ),
//    PERSISTENT_TBL (  1003,  setVal,4,                NULL ),
// // PERSISTENT_TBL (  1004,  shadowRegValues,4,       NULL ),

//APPLDATA
//    PERSISTENT_VAR(  1,  myApplStateVar,        NULL ),


  #define SharedCtlrStateParams
  #include <sharedAppl.c>

  END_PARAM_DATA
};


//-----------------------------------------------------------------------------
//  A list of parameterized "persistent" data.  Backup copies of such values
//  are stored in files and restored on startup.
//-----------------------------------------------------------------------------

//    Base File Name    Category  ParamData
//    ----------------  --------  ---------
//--- add any application-specific parameterized file defs here ---
//ParamFile  myAppCfgFile =
//  { "myFileName",     PDC_CTLR, myCfgParams,   0 };

ParamFile  settingsFile =
    { "settings",       PDC_SET,  settingsParams  };


#define SharedParams
#include <sharedAppl.c>


// NOTE: The ctlrStateFile is NOT included here as we don't EVER want
//       it to be restored manually.  It is restored ONLY once by the
//       application code on startup and ONLY following a "soft" reset
//       (i.e. one that avoided a CPU reset, which would have caused
//       all the digital outputs to change to their default states).

ParamFile *paramFiles[] = {
  #define SharedParamsList
  #include <sharedAppl.c>
//--- add any application-specific ParamData references here ---
//  &myAppCfgFile,
  &settingsFile
};

const int NumParamFiles = sizeof(paramFiles) / sizeof(paramFiles[0]);


//=============================================================================
//  Definitions for event msgs.  Event msgs are the means that other processes
//  use to inform the application process of an event that it needs to handle.
//
//  In general, allowing multiple processes to concurrently modify the state
//  of the controller is a sure recipie for chaos.  By restricting all such
//  changes to the context of a single process, and sending messages to that
//  process to request changes, we eliminate all such concurrency issues (not
//  the least of which is keeping track of the order of events).
//
//  The most common event is a request for change of state that corresponds to
//  a message received from an external client via a protocol like ModBus.
//=============================================================================

// EVENT_DEF( MyEvent )

//=============================================================================
//  Generic structure to handle application-specific presentation of the
//  digital and analog input/output values.  Code to handle additional
//  formatting that cannot be handled in this generic manner should be
//  placed in the showState() task routine.
//=============================================================================
  const char far DevStateStr[] = {
"\r\n"
  };


const DigFmtInfo digInFmt[] = {
  { -1,-1, "",0, "",0 }
};

const DigFmtInfo digOutFmt[] = {
  { -1,-1, "",0, "",0 }
};


const AnalogFmtInfo  adcFmt[] = {
  { -1,-1, 0,0 }
};

const AnalogFmtInfo  dacFmt[] = {
  { -1,-1, 0,0 }
};


//=============================================================================
//  Declarations for functions that handle setting "next" values that serve
//  as inputs to the application control logic (run by the "applic" task in
//  a periodically executed loop).
//
//  While care must be taken in these functions to use locks to avoid the
//  possibility of being preempted in the middle of an operation, such locks
//  must be kept to as short a time as possible to avoid excessively latency
//  in the main control loop.
//=============================================================================

void onRequest(uint psIdx);
void offRequest(uint psIdx);
void resetRequest(uint psIdx);

void newSetpoint(uint psIdx, float newAmps);




//=============================================================================
//  This is where misc global variables should be defined
//=============================================================================
#define SharedGlobals
#include <sharedAppl.c>

//--- add definitions for application-specific global variables here ---

//float far isetMin[NumPS], isetMax[NumPS];


//=============================================================================
//  Functions referenced in the Modbus register definitions to handle writes to
//  specific global values defined as ModBus registers
//=============================================================================

//-----------------------------------------------------------------------------
//  External ModBus client sent a new value for "cmdBitsAppl"
//
//  Check the bits that where set and trigger the corresponding events as
//  appropriate (but do NOT save the new value - just use it to trigger
//  events)
//-----------------------------------------------------------------------------
bool MB_cmdBitsAppl(SetRegFuncParams)
{
  uchar  i;
  ulong  cmdBits = MB_VAL(ulong);

  for (i=0; i<NumPS; ++i)  {
    if ((uint)cmdBits & TURN_ON_PS)   onRequest(i);
    if ((uint)cmdBits & TURN_OFF_PS)  offRequest(i);
    if ((uint)cmdBits & RESET_PS)     resetRequest(i);
    cmdBits >>= 4;
  }

  return true;
}


//-----------------------------------------------------------------------------
//  External ModBus client sent a new setpoint for one of the 8 power supplies.
//
//  Check the new setting and alter output values as appropriate
//-----------------------------------------------------------------------------
bool MB_iSet(SetRegFuncParams)
{
  uint  idx;

  idx = (REG->regNum - iSet_1_RegInfo.regNum) / 2;
  if (idx >= NumPS)  return false;

  newSetpoint(idx, MB_VAL(float));

  return true;
}


//-----------------------------------------------------------------------------
//  External ModBus client sent a new value for maxReadErr for a supply
//
//  Check the new setting and alter output values as appropriate
//-----------------------------------------------------------------------------
bool MB_maxReadErr(SetRegFuncParams)
{
  uint  idx;
  float  maxReadErr;


  idx = (REG->regNum - maxReadErr_1_RegInfo.regNum) / 2;
  if (idx >= NumPS)  return false;

  maxReadErr = MB_VAL(float);  maxReadErr = FABS(maxReadErr);

  _EnterCritical_();
  MaxReadErr(idx) = maxReadErr;
  _ExitCritical_();

  return true;
}


//-----------------------------------------------------------------------------
//  External ModBus client sent a new value for repeatMsgsSecs
//
//  Update the repeat time for each supply as necessary
//-----------------------------------------------------------------------------
bool MB_repeatMsgsSecs(SetRegFuncParams)
{
  uint  ui;
  ulong  newSecs;
  PSState *psNext;


  newSecs = MB_VAL(ulong);

  for (ui=0; ui<NumPS; ++ui)  {
    psNext  = Next.ps + ui;

    KLOCK();

    if (newSecs >= repeatMsgsSecs)
      psNext->nextRepeatMsgsSecs += newSecs - repeatMsgsSecs;
    else
      psNext->nextRepeatMsgsSecs -= repeatMsgsSecs - newSecs;

    repeatMsgsSecs = newSecs;

    KUNLOCK();
  }

  return true;
}



//=============================================================================
//  functions to handle restoration of saved "persistent" values
//=============================================================================

//-----------------------------------------------------------------------------
//  one of the setpoint values needs restoring from the settings file
//-----------------------------------------------------------------------------
RESTORE_FUNC(iSet_X)
{
  uint  u;

  for (u=0; u<NumPS; ++u)  if (curVal == (void far *) &ISET(u))  {
    newSetpoint(u, NEW_VAL(float));  break; }
}


//-----------------------------------------------------------------------------





//-----------------------------------------------------------------------------
#define SharedFuncs
#include <sharedAppl.c>



//=============================================================================
//  Task functions to handle application-specific commands
//=============================================================================


//-----------------------------------------------------------------------------
//  example func to process a command entered via a user shell
//-----------------------------------------------------------------------------
/*
void _task miscTest(int argc, char **argv)
{
  if (argc < 3)  { printf("Usage:  misc arg1 arg2\r\n");  Exit(-1); }

  printf("misc args: [%s] [%s]\r\n", argv[1], argv[2]);

  Exit(0);
}
*/

//-----------------------------------------------------------------------------
//  "state":  Show continuously updated state of the system
//-----------------------------------------------------------------------------
void _task showState(int argc, char **argv)
{
  uint  ui;
  int  y, leftFill;
  bool  blinkOff;
  char  interlockedFlag;
  PSState  *psCur;
  char  typeCh;
  char  *isetAttr,     *ireadAttr,
        *polResetAttr, *polResetStr,
        *onOffReqAttr, *onOffReqStr,
        *onOffActAttr, *onOffActStr,
        *psIntlkAttr,  *psIntlkStr,
        *extIntlkAttr, *extIntlkStr;
  char  temp[32];


  printf(clsCmd);  blinkOff = false;

  for (;;)  {
    CursorTo(0, 0);    stdHeader();

    blinkOff = !blinkOff;

    CursorTo(0, 3);  printf(
"                                                Pol/  On/Off  Interlock \r\n"
"                 ISET      IREAD     VREAD   T Reset  Req/Act  PS /EXT  \r\n"
"               --------  --------  --------  - -----  ------- --------- \r\n");/*
"xxxxxxxxxxS1: -#####.## -#####.##  -###.###  X  xxxx  xxx/xxx  xxx/xxx ",
"xxxxxxxxxxS2: -#####.## -#####.##  -###.###  X  xxxx  xxx/xxx  xxx/xxx ",
"xxxxxxxxxxS3: -#####.## -#####.##  -###.###  X  xxxx  xxx/xxx  xxx/xxx ",
"xxxxxxxxxxS4: -#####.## -#####.##  -###.###  X  xxxx  xxx/xxx  xxx/xxx ",
"xxxxxxxxxxS5: -#####.## -#####.##  -###.###  X  xxxx  xxx/xxx  xxx/xxx ",
"xxxxxxxxxxS6: -#####.## -#####.##  -###.###  X  xxxx  xxx/xxx  xxx/xxx ",
"xxxxxxxxxxS7: -#####.## -#####.##  -###.###  X  xxxx  xxx/xxx  xxx/xxx ",
"xxxxxxxxxxS8: -#####.## -#####.##  -###.###  X  xxxx  xxx/xxx  xxx/xxx ",
*/

    for (ui=0; ui<NumPS; ++ui)  {
      psCur = Cur.ps + ui;

      y = ui + 6;

      leftFill = MAX_PS_NAME_LEN - (uint) fstrlen(psNames[ui]);
      if (leftFill > 0)  printf("%*s", leftFill, "                    ");

      printf("%ls:", psNames[ui]);

      isetAttr = (psCur->assertOn
                    AND (psCur->requestedAmps != psCur->activeSetpoint))
                        ? RED : BLUE;
      DispStrAttr(14,y, isetAttr,
                   fmtFloat(psCur->requestedAmps, 9,2,temp), false);

      ireadAttr = (psCur->setReadErr OR psCur->readErrIntlk)
                        ? RED : BLUE;
      DispStrAttr(24,y, ireadAttr,
                  fmtFloat(psCur->actualAmps, 9,2,temp), false);

      DispStrAttr(35,y, BLUE,
                  fmtFloat(psCur->actualVolts, 8,3,temp), false);

      typeCh = TypeDescCh[psCur->supplyType];

    //--- Polarity (unipolar)/Reset (biploar) /Test (both) digital output ---
      if (psCur->testStarted)  {
        polResetAttr = RED;  polResetStr = "test"; }
      else  {
        polResetAttr = (psCur->assertOn AND psCur->needPolarityChg) ?
                          RED : BLUE;
        if (psCur->rampTo0Started)  {
          polResetStr = "RTZ ";
          if (psCur->setReadErr)  polResetAttr = RED;
        }
        else
          polResetStr = (psCur->assertPosPolarity ? "[+] " : "[-] ");
      }

    //--- On/Off digital output ---
      if (psCur->assertOn)  {
        onOffReqAttr = (psCur->prepForOff ? RED : BLUE);
        onOffReqStr = "ON ";
      }
      else  {
        onOffReqAttr = RED;  onOffReqStr = "OFF";  }

    //--- On/Off digital input ---
      if (psCur->supplyType == ST_Unknown)  {
        onOffActAttr = RED;  onOffActStr = "???";  }
      else
        if (psCur->psIsOn)  {
          onOffActAttr = BLUE;  onOffActStr = "ON ";  }
        else  {
          onOffActAttr = RED;  onOffActStr = "OFF";  }


    //--- PS Interlock digital input ---
      if (psCur->psInterlock)  {
        psIntlkAttr = RED;  psIntlkStr = "YES"; }
      else  {
        psIntlkAttr = BLUE;  psIntlkStr = "NO "; }

    //--- External Interlock digital input ---
      if (psCur->externInterlock)  {
        extIntlkAttr = RED;  extIntlkStr = "YES"; }
      else  {
        extIntlkAttr = BLUE;  extIntlkStr = "NO "; }

    //--- check for uncleared interlock ---
      interlockedFlag = psCur->interlocked ? '*' : ' ';


      sprintf(temp, "%c ", typeCh);
      DispStrAttr(45,y, BLUE, temp, false);

      DispStrAttr(48,y, polResetAttr, polResetStr, false);

      DispStrAttr(54,y, onOffReqAttr, onOffReqStr, false);
      printf("/");
      DispStrAttr(58,y, onOffActAttr, onOffActStr, false);

      DispStrAttr(63,y, psIntlkAttr, psIntlkStr, false);
      printf("/");
      DispStrAttr(67,y, extIntlkAttr, extIntlkStr, false );
      printf("%c\r\n", interlockedFlag);
    }

    CursorTo(0,15);  fflush(stdout);  CheckETX();  sleepMS(500);
  }

}



//-----------------------------------------------------------------------------
//  "tracesup":  Generate detailed state chg msgs for a supply
//-----------------------------------------------------------------------------
void _task traceSup(int argc, char **argv)
{
  uint  psNum;


  if (showHelp(1, argc, argv))  {
    printf("\nusage:  tracesup [ps#]  (no ps# means disable)\r\n");
    Exit(-1);
  }

  if (argc < 2)  {
    for (psNum=0; psNum<NumPS; ++psNum)  Next.ps[psNum].trace = false;
    Exit(0);
  }

  psNum = (uint) atoi(argv[1]);
  if ((psNum < 1) OR (psNum > NumPS))  Exit(-1);

  Next.ps[psNum-1].trace = true;  Exit(0);
}


//-----------------------------------------------------------------------------
//  Turn On/Off, or Reset a Power Supply
//-----------------------------------------------------------------------------
void _task cmd(int argc, char **argv)
{
  int  idx;
  bool  validCmd;


  validCmd = false;

  for (;;)  {
    if (argc < 3)  break;
    if (! isdigit(argv[1][0]))  break;

    idx = atoi(argv[1]);
    if ((idx < 1) OR (idx > NumPS))  { printf("Invalid supply #\n");  Exit(-2); }

    --idx;  validCmd = true;

    if (StriEq(argv[2], "on"))     { onRequest(idx);  break; }
    if (StriEq(argv[2], "off"))    { offRequest(idx);  break; }
    if (StriEq(argv[2], "reset"))  { resetRequest(idx);  break; }

    if (StriEq(argv[2], "set") AND (argc > 3)) {
      newSetpoint(idx, atof(argv[3]));  break; }

    validCmd = false;  break;
  }

  if (! validCmd)  {
    printf("Usage:  cmd # on|off|reset|set #.#\r\n");  Exit(-1); }

  Exit(0);
}


//-----------------------------------------------------------------------------
//  Show/Set the name for a power supply
//-----------------------------------------------------------------------------
void _task psname(int argc, char **argv)
{
  uint  ui;


  if (argc < 2)  { printf("usage:  psname # [newName]\r\n");  Exit(-1); }

  ui = (uint) atoi(argv[1]);
  if ((ui < 1) OR (ui > NumPS))  {
    printf("Invalid supply #: %s\r\n", argv[1]);  Exit(-1); }

  --ui;

  if (argc < 3)  { printf("%ls\r\n", psNames[ui]);  Exit(0); }

  if (strlen(argv[2]) > MAX_PS_NAME_LEN)  argv[2][MAX_PS_NAME_LEN] = 0;

  fstrcpy(psNames[ui], (uchar const far *)argv[2]);

  Exit(0);
}


//=============================================================================
//  Event-Handler functions called by the appl task to process event messages
//=============================================================================




//=============================================================================
//  Application-specific logic
//=============================================================================

//-----------------------------------------------------------------------------
//  Compare a boolean polarity switch setting (positivePolarity) and a floating
//  setpoint value (setpoint).
//  Returns true if the values are compatible, false if not
//-----------------------------------------------------------------------------
bool polarityDifference(bool positivePolarity, float setpoint)
{
  bool  diff;


  diff = false;

  if (positivePolarity)  {
    if (setpoint < 0.0)  diff = true;  }
  else
    if (setpoint > 0.0)  diff = true;

  return diff;
}


#pragma warn -775

//=============================================================================
//  Update the the state of the application logic in well-defined phases
//=============================================================================
// IMPORTANT NOTES and RULES:
//
//  - The code to change ANY value must be in ONE place.  NEVER spread the
//    logic for updating a single value around.  There should be ONE place
//    to look to see how a value gets set.
//
//  - "cur" values are basically "inputs" - literally (they are set based
//    on the just-read state of a digital or analog input) and/or figuratively
//    (they hold a value set by some external source, such as a cmd from a
//    remote client or another user-interface task).
//
//  - "next" values are basically "outputs" - literally (they determine the
//    next state of a digital or analog output) and/or figuratively (they
//    can affect the state of the logic for the next pass).
//
//  - The "cur" OR the "next" instance of a value may be changed - NEVER
//    both of them (with the exception of the first step below that copies
//    all of cur to prev and next to cur).
//
//  - A value MUST be in the ApplState struct if one of the following applies:
//     - The application logic is dependent on if or how much the value has
//       changed between passes.
//     - The value is used as both input (used in in a conditional and/or in
//       an expression that can affect the result of an assignment) and output
//       (the target of an assignment).
//
//  - Variables that are NOT in the ApplState struct but are USED (in
//    conditionals or other expressions) by the application logic must be
//    treated just like "cur" values.  Variables SET (the target in an
//    assignment statement) by the application logic must be treated just
//    like "next" values.
//
//  - When the logic requires comparing an "input" to a related "output", be
//    sure to use the "cur" value for the "input", and the "prev" value for
//    the "output".
//
//    This is because during the processing of the loop, the "cur" values
//    represent the input values just read (what they are "now") and the
//    "prev" values represent the values the outputs were set to at the
//    beginning of the prev pass (i.e. they are the values that "caused"
//    the inputs to have their "cur" values).
//
//    During a pass the output values in "cur" represent what the outputs
//    where set to AFTER the "cur" inputs were updated, so comparing them
//    makes no sense.  Doing so is the equivalent of comparing a past response
//    to a condition that didn't exist at the time.
//=============================================================================


//=============================================================================
//  Definitions for functions that handle setting "next" values that serve
//  as inputs to the application control logic (run by the "applic" task in
//  a periodically executed loop).
//
//  While care must be taken in these functions to use locks to avoid the
//  possibility of being preempted in the middle of an operation, such locks
//  must be kept to as short a time as possible to avoid excessively latency
//  in the main control loop.
//
//  For REALLY short locks, using a critical section is best, because this
//  has the least overhead (but it has to be REALLY short to avoid causing
//  problems for interrupt handlers).
//
//  For longer locks (ones that involve anything but the most trivial
//  conditional logic), use either KLOCK() and KUNLOCK() (which disable the
//  scheduler) or a more specific global semaphore/mutex lock.
//=============================================================================

//-----------------------------------------------------------------------------
void onRequest(uint idx)
{
  PSState *psNext = Next.ps + idx;

  KLOCK();

  psNext->onRequest = true;
  psNext->offRequest = false;

  KUNLOCK();

  fprintf(DiagLog, "PS%d (%.16ls): ON request\n", idx+1, psNames[idx]);
}

//-----------------------------------------------------------------------------
void offRequest(uint idx)
{
  PSState *psNext = Next.ps + idx;

  KLOCK();

  psNext->onRequest = false;
  psNext->offRequest = true;

  KUNLOCK();

  fprintf(DiagLog, "PS%d (%.16ls): OFF request\n", idx+1, psNames[idx]);
}

//-----------------------------------------------------------------------------
void resetRequest(uint idx)
{
  _EnterCritical_();  Next.ps[idx].resetRequested = GetMS();  _ExitCritical_();

  fprintf(DiagLog, "PS%d (%.16ls): RESET request\n", idx+1, psNames[idx]);
}

//-----------------------------------------------------------------------------
void newSetpoint(uint idx, float reqAmps)
{
  float  newAmps, minAmps, maxAmps, applMin, applMax;
//  char  temp[32];


  newAmps = reqAmps;

//--- determine min and max based on hdw and applic limits ---
  applMin = ISET_MIN(idx);  applMax = ISET_MAX(idx);
  minAmps = SetMin(idx);  if (minAmps < applMin)  minAmps = applMin;
  maxAmps = SetMax(idx);  if (maxAmps > applMax)  maxAmps = applMax;

  if (newAmps < minAmps)  newAmps = minAmps;
  if (newAmps > maxAmps)  newAmps = maxAmps;

//tooslow
//  fprintf(DiagLog, "PS%d (%.16ls): Setpoint: requested: %s  used: %s\n",
//          idx+1, psNames[idx],
//          Engineering(reqAmps, 6, temp),
//          Engineering(newAmps, 6, temp+16) );

  _EnterCritical_();
  nextState.ps[idx].requestedAmps = newAmps;
  _ExitCritical_();
}


//-----------------------------------------------------------------------------
//  log the current state of one of the power supplies
//-----------------------------------------------------------------------------
void logPSState(uint psNum)
{
  int  leftFill;
  PSState  *psCur;
  char  *polResetStr,  *onOffActStr;
  char  temp[64];


  if (!psNum)  {
    printf("\r\n");
    printf(
"                                                Pol/  On/Off  Interlock \r\n"
"                 ISET      IREAD     VREAD   T Reset  Req/Act  PS /EXT  \r\n"
"               --------  --------  --------  - -----  ------- --------- \r\n");/*
"xxxxxxxxxxS1: -#####.## -#####.##  -###.###  X  xxxx  xxx/xxx  xxx/xxx* ",
"xxxxxxxxxxS2: -#####.## -#####.##  -###.###  X  xxxx  xxx/xxx  xxx/xxx* ",
"xxxxxxxxxxS3: -#####.## -#####.##  -###.###  X  xxxx  xxx/xxx  xxx/xxx* ",
"xxxxxxxxxxS4: -#####.## -#####.##  -###.###  X  xxxx  xxx/xxx  xxx/xxx* ",
"xxxxxxxxxxS5: -#####.## -#####.##  -###.###  X  xxxx  xxx/xxx  xxx/xxx* ",
"xxxxxxxxxxS6: -#####.## -#####.##  -###.###  X  xxxx  xxx/xxx  xxx/xxx* ",
"xxxxxxxxxxS7: -#####.## -#####.##  -###.###  X  xxxx  xxx/xxx  xxx/xxx* ",
"xxxxxxxxxxS8: -#####.## -#####.##  -###.###  X  xxxx  xxx/xxx  xxx/xxx* ",
*/
  }

  psCur = Cur.ps + psNum;

//--- Polarity (unipolar)/Reset (biploar) /Test (both) digital output ---
  if (psCur->testStarted)
    polResetStr = "test";
  else
    if (psCur->rampTo0Started)
      polResetStr = "RTZ ";
    else
      polResetStr = psCur->assertPosPolarity ? "[+] " : "[-] ";

//--- On/Off digital input ---
  if (psCur->supplyType == ST_Unknown)
    onOffActStr = "???";
  else
    onOffActStr = (psCur->psIsOn ? "ON " : "OFF");


  leftFill = MAX_PS_NAME_LEN - (uint) fstrlen(psNames[psNum]);
  if (leftFill > 0)  printf("%*s", leftFill, "                    ");

  printf("%ls: %s %s  %s  %c  %s  %s/%s  %s/%s%c\r\n",
         psNames[psNum],
         fmtFloat(psCur->requestedAmps, 9,2, temp),
         fmtFloat(psCur->actualAmps,    9,2, temp+10),
         fmtFloat(psCur->actualVolts,   8,3, temp+20),
         TypeDescCh[psCur->supplyType],
         polResetStr,
         (psCur->assertOn ? "ON " : "OFF"), onOffActStr,
         (psCur->psInterlock ? "YES" : "NO "),
         (psCur->externInterlock ? "YES" : "NO "),
         (psCur->interlocked ? '*' : ' ')
        );
}


//-----------------------------------------------------------------------------
//  write log messages for changes in state for the specified supply that are
//  not covered by the normal logging
//-----------------------------------------------------------------------------
typedef struct  {
  uint             offset;    // # bytes from start of struct
  const char far  *name;      // name of value
  uint             type;      // type of value
}  ValueDescr;

#define  PS_STATE_VAL(name, type)  { offsetof(PSState, name), #name, type }

const ValueDescr  far psStateValues[] =  {
  PS_STATE_VAL( activeSetpoint,     DT_F32 ),
  PS_STATE_VAL( smallestErr,        DT_F32 ),
  PS_STATE_VAL( activeIntlk,        DT_U8  ),
  PS_STATE_VAL( needPolarityChg,    DT_U8  ),
  PS_STATE_VAL( nearZero,           DT_U8  ),
  PS_STATE_VAL( setReadErr,         DT_U8  ),
  PS_STATE_VAL( ramping,            DT_U8  ),
  PS_STATE_VAL( onPending,          DT_U8  ),
  PS_STATE_VAL( rampTo0Started,     DT_U32 ),
  PS_STATE_VAL( prepForOff,         DT_U8  ),
  PS_STATE_VAL( offPending,         DT_U8  ),
  PS_STATE_VAL( assertPosPolarity,  DT_U8  ),
  PS_STATE_VAL( polChgStarted,      DT_U32 ),
  PS_STATE_VAL( cantBeUnipolar,     DT_U8  ),
  PS_STATE_VAL( supplyType,         DT_U16 )
};

#define  NUM_PS_VALS  ( sizeof(psStateValues) / sizeof(ValueDescr) )

//--------------------------------------------------
void logPSChgs(uint psNum)
{
  bool  logAll;
  uint  ui;
  PSState *psPrev, *psCur;  //, *psNext;
  ValueDescr valDescr;
  void  *prevVal, *curVal;
  char  temp[64];


  psPrev = Prev.ps + psNum;
  psCur  = Cur.ps + psNum;
//  psNext  = Next.ps + psNum;


  logAll = !psPrev->trace;

  for (ui=0; ui<NUM_PS_VALS; ++ui)  {
    valDescr = psStateValues[ui];

    prevVal = (char *)psPrev + valDescr.offset;
    curVal = (char *)psCur + valDescr.offset;

    switch (valDescr.type)  {

      case DT_S16:
        if (logAll OR (*(S16 *)curVal != *(S16 *)prevVal))
          printf("PS%u: %ls: %d\r\n", psNum+1, valDescr.name, *(S16 *)curVal );
        break;

      case DT_U16:
        if (logAll OR (*(U16 *)curVal != *(U16 *)prevVal))
          printf("PS%u: %ls: %u\r\n", psNum+1, valDescr.name, *(U16 *)curVal );
        break;

      case DT_S32:
        if (logAll OR (*(S32 *)curVal != *(S32 *)prevVal))
          printf("PS%u: %ls: %ld\r\n", psNum+1, valDescr.name, *(S32 *)curVal );
        break;

      case DT_U32:
        if (logAll OR (*(U32 *)curVal != *(U32 *)prevVal))
          printf("PS%u: %ls: %lu\r\n", psNum+1, valDescr.name, *(U32 *)curVal );
        break;

      case DT_F32:
        if (logAll OR (*(F32 *)curVal != *(F32 *)prevVal))
          printf("PS%u: %ls: %.16s\r\n", psNum+1, valDescr.name,
                 Engineering(*(float *)curVal, 6, temp) );
        break;

      case DT_STR:
        if (logAll OR !StrEq((char *)curVal, (char *)prevVal))
          printf("PS%u: %ls: %s\r\n", psNum+1, valDescr.name, (char *)curVal );
        break;

      case DT_U8:
        if (logAll OR (*(U8 *)curVal != *(U8 *)prevVal))
          printf("PS%u: %ls: %u\r\n", psNum+1, valDescr.name, *(U8 *)curVal );
        break;

      case DT_S8:
        if (logAll OR (*(S8 *)curVal != *(S8 *)prevVal))
          printf("PS%u: %ls: %d\r\n", psNum+1, valDescr.name, *(S8 *)curVal );
        break;
    }

  }

}

/*
=========== NOTES on Locking ============

The applic task runs at the highest priority (even above the kernel), so it
cannot be preempted by any other task.  As long as this is true, the applic
task does not need to use locks when accessing global variables.

However, all other tasks MUST disable the scheduler when accessing values
used by the applic task to avoid being preempted while doing so.  Care must
be taken to keep these times short to avoid latency issues.

For REALLY short times, a critical section can be used as this is much more
efficient than disabling and reenabling the scheduler.  But critical sections
disable all interrupts, so it is not practical in most cases.

*/

//=============================================================================
//  Update the "Cur"rent state of the controller
//
//  Be sure to read the "NOTES on Locking" section above....
//=============================================================================
//
//  Just before this function is called, readings were taken of the current
//  state of the hardware, and any changes to the output states specified by
//  the last pass were applied (just after taking the readings).  These values
//  are copied to Next as needed and constitute the Hardware Inputs.
//
//  Other tasks can affect the state of the control logic by changing values
//  in Next as well.  These values constitute Software Inputs.
//
//  This function is responsible for the Asynchronous logic that is applied
//  each pass.  It updates values in Next (or other global variables) based
//  on the Software and Hardware Input values in Next.  This portion of the
//  logic is analogous to unclocked gates in an electronic circuit.
//
//  After the Asychronous logic is applied, we set Prev = Cur and Cur = Next.
//  The result is that Prev contains the controller state as it was during the
//  previous pass, and Cur indicates the current state of the controller.
//
//  Also, see the "NOTES on Locking" section above....
//=============================================================================
void updateCurState(void)
{
  uint  ui, readIdx;
  uchar  psInBits, psOutBits;
  float  ampErr, setpointChg;
  PSState  *psCur, *psNext;
  static uint  loopCount = 0;



  if (current != applTask)  return;  // must only be called by "applic" task

//-----------------------------------------------------------------
// The following conditions exist at this point:
// - The analog and digital hardware inputs have just been read
// - The outputs were just set to values computed during the last pass
// - "Cur" contains the state the inputs were in at the START of
//   of the previous pass, and the state the outputs were in DURING
//   the previous pass
// - "Next" contains the values that the outputs were just set
//   to and any "input" values that were changed by other tasks
//------------------------------------------------------------------

// Now we update the values in "Next" that are based solely on the
// state of the hardware inputs we just read and values maintained by
// by other tasks (basically, we deal with READINGS here: Values that
// we need but are not written to "Next" by another task)

  Next.processTime = GetMS();

  msSincePrev = msSince(Cur.processTime);

  Next.digitalIn = digitalIn;


//APPLCODE

//--- update the state of each of the independent power supplies ---
  for (ui=0; ui<NumPS; ++ui)  {
//    psPrev = Prev.ps + ui;
    psCur  = Cur.ps + ui;
    psNext  = Next.ps + ui;

    readIdx = I_RD_IDX(ui);

    psInBits = (uchar) PS_IN_BITS(Next.digitalIn, ui);
    psOutBits = (uchar) PS_OUT_BITS(Cur.digitalOut, ui);

    if (psNext->supplyType == ST_Bipolar)
      psNext->psIsOn = BOOL((int)psInBits & (int)PS_IS_ON);
    else
      psNext->psIsOn = psNext->assertOn;


  //--- update current and voltage readings ---
    psNext->actualAmps = readVal[readIdx];
    psNext->avgdAmps = GetAvgdRead(readIdx);

    if (psCur->supplyType == ST_Unipolar)
      if (!psCur->assertPosPolarity)  psNext->actualAmps = -psNext->actualAmps;

    psNext->actualVolts = readVal[V_RD_IDX(ui)];

    psNext->nearZero = ( FABS(psNext->avgdAmps) <= MaxReadErr(ui) );

    setpointChg = psNext->activeSetpoint - psCur->activeSetpoint;
    psNext->setpointChgd = (setpointChg != 0.0);


    if (!psNext->readCheckTime)  {
      psNext->readCheckTime = Next.processTime;
      psNext->readCheckAmps = psNext->avgdAmps;
      psNext->setReadErr = false;
    }

  //===== Check the readback for unexpected behavior aprox 5 times/sec. =====
  //===== Limited to one PS/pass to further reduce CPU usage =====

    if (((loopCount % NumPS) == ui)
        AND (msSince(psNext->readCheckTime) >= 200))  {

    //--- rate of chg in the avgd readback extrapolated to 1 second ---
      psNext->readChgRate = (psNext->avgdAmps - psNext->readCheckAmps) * 5.0;

      ampErr = psCur->activeSetpoint - psNext->avgdAmps;
      ampErr = FABS(ampErr);

    //--- The readback is ramping if it is changing by some minimum rate ---
    //--- and the change gets it closer to a steady active setpoint or is ---
    //--- in the same direction as the last change in the active setpoint ---
      psNext->ramping = (FABS(psNext->readChgRate) >= MinRampRate(ui));
      if (!psNext->setpointChgd)  {
        if (ampErr >= psCur->smallestErr)  psNext->ramping = false; }
      else
        if (SIGN(setpointChg) != SIGN(psNext->readChgRate))
          psNext->ramping = false;

      psNext->acceptableErr = (ampErr <= MaxReadErr(ui));

      psNext->readCheckAmps = psNext->avgdAmps;
      psNext->readCheckTime = Next.processTime;

    //--- Make sure read/set err is either shrinking or not too large ---
      if (psNext->setpointChgd OR !psNext->psIsOn)  {
        psNext->smallestErr = ampErr;
        psNext->lastReadOKTime = Next.processTime;
      }
      else  {
        if (ampErr < psNext->smallestErr)  psNext->smallestErr = ampErr;
        if (psNext->acceptableErr OR psNext->ramping)
          psNext->lastReadOKTime = Next.processTime;
      }

      if (psNext->setReadErr)  {
        if (psNext->acceptableErr)  psNext->setReadErr = false;  }
      else
        if (msSince(psNext->lastReadOKTime) > MaxErrTime(ui))
          psNext->setReadErr = true;
    }


  //--- If the requested ON/OFF state or the setpoint changed, wait 10 ---
  //--- seconds before repeating any msgs for existing conditions ---
    if (psNext->onRequest OR psNext->offRequest
        OR (psNext->setpointGoal != psCur->setpointGoal) )
      psNext->nextRepeatMsgsSecs = curTime + 10;


  //--- Assert an interlock for set/read errors? ---
    psNext->readErrIntlk =
      (psNext->setReadErr AND psNext->psIsOn AND IlkIfReadErr(ui));

    psNext->psInterlock =     !((int)psInBits & (int)PS_INTLK_OK);
    psNext->externInterlock = !((int)psInBits & (int)EXT_INTLK_OK);

    psNext->activeIntlk =
      psNext->psInterlock OR psNext->externInterlock OR psNext->readErrIntlk;

  //--- is the supply not turning on when it should? ---
    if (psNext->psIsOn == psNext->assertOn)
      psNext->lastOnOKTime = Next.processTime;
    psNext->onCtlErr = (msSince(psNext->lastOnOKTime) > MaxErrTime(ui));


    psNext->needPolarityChg =
      (psNext->supplyType != ST_Bipolar) AND
      polarityDifference(psNext->assertPosPolarity, psNext->requestedAmps);

    psNext->testIn = BOOL((int)psInBits & (int)PS_IS_ON);
    psNext->testOut = BOOL((int)psOutBits & (int)RESET_BIT);


    ignoreResetReqs[ui] = psNext->logicallyOn AND !psNext->onCtlErr;
  }


// Next we make a copy of the application state before we apply all the
// logic that updates the overall state.  This allows us to have a copy of
// all the relevant values that remain unchanged during the processing,
// avoiding all kinds of order-of-execution dependencies.

// We also make a copy of the state from last time before overwriting it
// so we can see if/how values have changed since the last pass

// Note that a copy is done (rather than just switching pointer values) to
// insure all values are set in both the source and destination.  If we
// simply kept pointers and swapped the pointer values, then EVERY value
// would have to be updated independently by the code, which is a pain to
// program, adds overhead and code, and is a VERY easy way to introduce odd
// bugs.  Overall, copying everything here is actually faster overall,
// simpler, and much safer (and is now required by the new softReset code).

  Prev = Cur;  Cur = Next;

//--- At this point the Cur copy contains what the inputs were at the start ---
//--- of this pass, and the Prev copy contains the output values from the ---
//--- the previous pass (i.e. the output values that resulted in the input ---
//--- values in Cur) ---



//===== log msgs about current state and/or changes in state =====

  logChgMsgs(curTime);

  ++loopCount;
}

//-----------------------------------------------------------------------------
//
//-----------------------------------------------------------------------------
void logChgMsgs(ulong curTime)
{
  bool  repeatMsgs;
  uint  ui;
  PSState  *psPrev, *psCur, *psNext;
  char  *stateStr;
  char  buf[64];

  static bool  logState = false;
  static uint  psCount = 0;


//--- log msgs for bit chgs only if not starving other processes for cycles ---
//  if (applLoopTime < 14)  logDigIOChgs(false);  // handled better by code

  if (curTime - lastLogStateSecs > logStateSecs)  {
    logState = true;  lastLogStateSecs = curTime; }



  for (ui=0; ui<NumPS; ++ui)  {
    psPrev = Prev.ps + ui;
    psCur  = Cur.ps + ui;
    psNext  = Next.ps + ui;

    if (logState AND (ui == psCount))  logPSState(ui);

    repeatMsgs = (curTime >= psCur->nextRepeatMsgsSecs);

    if (psCur->trace)  logPSChgs(ui);

  //--- ON request while PS is OFF but the reading is not close to 0 ---
    if (psCur->onRequest AND !psPrev->onRequest AND !psCur->psIsOn
        AND !psCur->nearZero)
      printf("PS%d (%.16ls): OFF reading > maxReadErr_%d\r\n"
             "  Current reading: %s  maxReadErr_%d: %s\r\n",
             ui+1, psNames[ui], ui+1,
             Engineering(psCur->avgdAmps, 4, buf),
             ui+1, Engineering(MaxReadErr(ui), 4, buf+16) );

  //--- set/read error state changed or its still active and its been ---
  //--- awhile since we logged a message about it ---
    stateStr = NULL;
    if (psCur->setReadErr != psPrev->setReadErr)
      stateStr = (psCur->setReadErr ? "SET" : "CLEARED");
    else
      if (psCur->setReadErr AND repeatMsgs)  stateStr = "PERSISTS";
    if (stateStr)  {
      printf("PS%d (%.16ls): Set/Read error %s\r\n"
             "  Set:  Requested: %s  Asserted: %s\r\n"
             "  Read: %s  Rate of chg: %s\r\n",
             ui+1, psNames[ui], stateStr,
             Engineering(psPrev->requestedAmps,  6, buf),
             Engineering(psPrev->activeSetpoint, 6, buf+16),
             Engineering(psCur->readCheckAmps,   6, buf+32),
             Engineering(psCur->readChgRate,     6, buf+48)
            );
      psNext->nextRepeatMsgsSecs = curTime + repeatMsgsSecs;
    }

  //--- change in polarity required ---
    stateStr = NULL;
    if (psCur->polChgStarted != psPrev->polChgStarted)
      stateStr = (psCur->polChgStarted ? "SET" : "CLEARED");
    else
      if (psCur->polChgStarted AND repeatMsgs)  stateStr = "PERSISTS";
    if (stateStr)  {
      printf("PS%d (%.16ls): Need polarity change: %s\r\n",
             ui+1, psNames[ui], stateStr );
      psNext->nextRepeatMsgsSecs = curTime + repeatMsgsSecs;
    }


  //--- PS is not responding to asserted ON/OFF control output ---
    stateStr = NULL;
    if (psCur->onCtlErr != psPrev->onCtlErr)
      stateStr = (psCur->onCtlErr ? "SET" : "CLEARED");
    else
      if (psCur->onCtlErr AND repeatMsgs)  stateStr = "PERSISTS";
    if (stateStr)  {
      printf("PS%d (%.16ls): ON/OFF control Error: %s\r\n"
             "   assertOn: %s  psIsOn: %s\r\n",
             ui+1, psNames[ui], stateStr,
             (psCur->assertOn ? "YES" : "NO"),
             (psCur->psIsOn ? "YES" : "NO")
            );
      psNext->nextRepeatMsgsSecs = curTime + repeatMsgsSecs;
    }


    if (psCur->readErrIntlk != psPrev->readErrIntlk)  {
      printf("PS%d (%.16ls): Read Error Interlock %s\r\n",
             ui+1, psNames[ui], psCur->readErrIntlk ? "SET" : "CLEARED");
      if (psCur->readErrIntlk)
        printf("  Current reading: %s  maxReadErr_%d: %s\r\n",
           Engineering(psCur->avgdAmps, 4, buf),
           ui+1, Engineering(MaxReadErr(ui), 4, buf+16) );
    }

    if (psCur->psInterlock != psPrev->psInterlock)
      printf("PS%d (%.16ls): Power Supply Interlock %s\r\n",
             ui+1, psNames[ui], psCur->psInterlock ? "SET" : "CLEARED");

    if (psCur->externInterlock != psPrev->externInterlock)
      printf("PS%d (%.16ls): External Interlock %s\r\n",
             ui+1, psNames[ui], psCur->externInterlock ? "SET" : "CLEARED");

    if (psCur->interlocked != psPrev->interlocked)
      printf("PS%d (%.16ls): Interlock %s\r\n",
             ui+1, psNames[ui], psCur->interlocked ? "SET" : "CLEARED");


    if (psCur->assertOn != psPrev->assertOn)
      printf("PS%d (%.16ls): Turning PS %s\r\n",
             ui+1, psNames[ui], psCur->assertOn ? "ON" : "OFF");

    if (psCur->logicallyOn != psPrev->logicallyOn)
      printf("PS%d (%.16ls): Changed to %s mode\r\n",
             ui+1, psNames[ui], psCur->logicallyOn ? "ON" : "OFF");

    if (psCur->psIsOn != psPrev->psIsOn)
      printf("PS%d (%.16ls): PS is now %s\r\n",
             ui+1, psNames[ui], psCur->psIsOn ? "ON" : "OFF");


    if (psCur->testStarted != psPrev->testStarted)
      printf("PS%d (%.16ls): %s type test\r\n",
             ui+1, psNames[ui], psCur->testStarted ? "Started" : "Finished" );

    if (psCur->supplyType != psPrev->supplyType)
      printf("PS%d (%.16ls): Supply type now: %ls\r\n",
             ui+1, psNames[ui], TypeDesc[psCur->supplyType]);


    if (psCur->resetRequested AND !psPrev->resetRequested
        AND ignoreResetReqs[ui])
      printf("PS%d (%.16ls): RESET ignored while supply is ON\r\n",
             ui+1, psNames[ui]);


    if (psCur->assertPosPolarity != psPrev->assertPosPolarity)
      printf("PS%d (%.16ls): Asserted Polarity now: [%c]\r\n",
             ui+1, psNames[ui], psCur->assertPosPolarity ? '+':'-');


    stateStr = NULL;
    if (psCur->rampTo0Started != psPrev->rampTo0Started)
      stateStr = (psCur->rampTo0Started ? "STARTED" : "ENDED");
    else
      if (psCur->rampTo0Started AND repeatMsgs)  stateStr = "PERSISTS";
    if (stateStr)  {
      printf("PS%d (%.16ls): Ramp-to-Zero: %s\r\n",
             ui+1, psNames[ui], stateStr);
      psNext->nextRepeatMsgsSecs = curTime + repeatMsgsSecs;
    }

//tooslow
//    if (psCur->setpointGoal != psPrev->setpointGoal)
//      printf("PS%d (%.16ls): Setpoint Goal now: %s\r\n",
//            ui+1, psNames[ui], Engineering(psCur->setpointGoal, 6, buf) );


    if (psCur->interlocked AND psCur->onRequest AND !psPrev->onRequest)
      printf("PS%d (%.16ls): Must clear interlocks first!\r\n",
             ui+1, psNames[ui]);
  }

//--- used to reduce lower priority per-PS processing to one PS per pass ---
  if (logState)  if (++psCount == NumPS)  { psCount = 0;  logState = false; }
}


//=============================================================================
//  Now we update the values in "next" based on the "cur" and "prev"
//  values.
//=============================================================================
//  All values changed by updateNextState() are "outputs" from this func.
//  All other values referenced in this function are "inputs" to this func.
//
//  All the values in "next" provide information about the next logical state
//  for the controller:  Values to be applied during the next pass of the
//  control logic.  All "next" values fall in to one of 2 categories:
//
//    - They are set/updated ONLY by this function (which is executed only
//      by the "applic" task)
//
//    - They are set ONLY by other tasks (within a lock), and possibly
//      reset/cleared by the "applic" task in updateCurState() (after copying
//      "next" to "cur")
//
//
//  All values that act as input to this func must fall in to one of these
//  categories:
//
//    - "prev" values
//
//    - "cur" values updated directly by statements in updateCurState()
//
//    - "cur" values updated indirectly by updateCurState() due to the copy
//      of "next" to "cur", which includes the values in "next" that are
//      set by other tasks
//
//    - a shared/global value that is the equivalent of a "cur" value but
//      for which we don't need to track changes
//
//
//  All "outputs" from this function must be altered ONLY within this function.
//
//  All "inputs" to this function must NEVER be altered within this function.
//
//  Each output value (or VERY small "set" of ouput values) must be updated
//  in a single, self-contained block of code.  NO output may be referenced
//  OR altered in ANOTHER block of code:  This is a declarative style of
//  programming and has multiple advantages:
//
//     - There are no order-of-execution dependencies across nodes/blocks
//       within a pass.  They can be executed in any order and the result
//       will be the same (get used to it - this method will be the norm as
//       computer languages and hardware add support for atomic transactions
//       and computers evolve in to neural-net-like processor arrays - FPGAs
//       are just the beginning of this trend).
//
//     - Encapsulating all logic for an output within a single node/block
//       leaves just one place to start when tracking a problem backward from
//       the result (far simpler than guessing at the possible causes and
//       trying to work forward, looking for paths to the observed result).
//
//     - Drastically reduces the complexity of the code by greatly reducing
//       the number of possible paths through the code.  This simplifies
//       testing and verification processes.
//
//  Each node/block-of-code should be prefaced with the following information:
//   - The names of the values it modifies
//   - A descr of the meaning of the values and the logic used to set them
//   - A list of the values that serve as "inputs" to the node
//
//=============================================================================
void updateNextState(void)
{
  uint  ui;
  PSState *psPrev, *psCur, *psNext;
//  char  buf[64];


  if (current != applTask)  return;  // must only be called by "applic" task

  for (ui=0; ui<NumPS; ++ui)  {
    psPrev = Prev.ps + ui;
    psCur  = Cur.ps + ui;
    psNext = Next.ps + ui;


  //---------------------------------------------------------------------
  // psNext->assertPosPolarity, psNext->polChgStarted
  //    Description:
  //       Determines the desired state for the external polarity switch that
  //       is controlled by a PLC (for unipolar supplies only).
  //       If the power supply is on or the load current is too high to safely
  //       change the polarity, then do not change the polarity.  Otherwise,
  //       set according to the polarity of the requested setpoint.
  //    Inputs:
  //      psCur->nearZero
  //      psCur->psIsOn
  //      psCur->assertOn
  //      psCur->activeSetpoint
  //      psCur->logicallyOn
  //      psCur->supplyType
  //      psCur->requestedAmps
  //---------------------------------------------------------------------
    if (!psCur->logicallyOn)
      psNext->assertPosPolarity = false;
    else
      if (psCur->nearZero AND
          ( (!psCur->psIsOn AND !psCur->assertOn)
             OR
          ((psCur->supplyType == ST_Bipolar)
            AND (psCur->activeSetpoint == 0.0))))
        psNext->assertPosPolarity = (psCur->requestedAmps >= 0.0);

    if ( (psNext->assertPosPolarity != psCur->assertPosPolarity)
      AND (psCur->supplyType == ST_Unipolar) )  psNext->polChgStarted = GetMS();

    if (psCur->polChgStarted)  if (msSince(psCur->polChgStarted) > polChgDelay)
      psNext->polChgStarted = 0;


  //--------------------------------------------------------------------
  // psNext->rampTo0Started
  //    Description:
  //       Time we started ramping toward 0
  //    Inputs:
  //       psCur->setpointGoal
  //       psCur->nearZero
  //       psCur->psIsOn
  //--------------------------------------------------------------------
    if ((psCur->setpointGoal != 0.0) OR psCur->nearZero OR !psCur->psIsOn)
      psNext->rampTo0Started = 0;
    else
      if (!psNext->rampTo0Started)  psNext->rampTo0Started = GetMS();


  //--------------------------------------------------------------------
  // psNext->newGoal
  //    Description:
  //       Flag to track when we have a new active setpoint goal.  This
  //       is used to log a message the first time the readback gets
  //       close enough to the new goal.
  //    Inputs:
  //       psCur->setpointChgd
  //       psCur->newGoal
  //       psCur->activeSetpoint
  //       psCur->setpointGoal
  //       psCur->acceptableErr
  //--------------------------------------------------------------------
    if (psCur->setpointChgd)  psNext->newGoal = true;

    if (psCur->newGoal)  if (psCur->activeSetpoint == psCur->setpointGoal)
      if (psCur->acceptableErr)  if (psCur->readCheckTime == Cur.processTime) {
//tooslow
//        printf("PS%d (%.16ls): Setpoint goal reached:\r\n"
//               "  Goal: %s   Read: %s\r\n", ui+1, psNames[ui],
//                Engineering(psCur->setpointGoal, 4, buf),
//                Engineering(psCur->actualAmps, 4, buf+16) );
        psNext->newGoal = false;
      }

  //--------------------------------------------------------------------
  // psNext->setpointGoal, psNext->activeSetpoint
  //    Description:
  //       Determine the value that we should be ramping the control
  //       voltage toward and the value to actually apply next
  //    Inputs:
  //       psCur->interlocked
  //       psCur->assertOn
  //       psCur->needPolarityChg
  //       psCur->prepForOff
  //       psCur->requestedAmps
  //       psCur->setpointGoal
  //       psCur->activeSetpoint
  //--------------------------------------------------------------------
    if (psCur->interlocked OR !psCur->assertOn
        OR psCur->needPolarityChg OR psCur->prepForOff)
      psNext->setpointGoal = 0.0;
    else
      psNext->setpointGoal = psCur->requestedAmps;

    if (psCur->interlocked OR !psCur->assertOn)
      psNext->activeSetpoint = 0.0;
    else  {
      float chg = psNext->setpointGoal - psCur->activeSetpoint;
      float maxChg = MaxRampRate(ui) * (float)SchedMS / 1000.0;
      if (FABS(chg) > maxChg)  chg = (chg < 0.0) ? -maxChg : maxChg;
      psNext->activeSetpoint += chg;
    }


  //--------------------------------------------------------------------
  // psNext->assertOn
  //    Description:
  //       Specifies whether or not the ON cmd to the supply should be
  //       actively asserted.  This is NOT the same as the supply being
  //       "logically" on, as the supply needs to be physically off
  //       during certain periods when we are logically in an ON state.
  //    Inputs:
  //       psCur->logicallyOn
  //       psCur->interlocked
  //       psCur->polChgStarted
  //       psCur->testStarted
  //       psCur->supplyType
  //       psCur->nearZero
  //       psCur->needPolarityChg
  //----------------------------------------------------------------------
    psNext->assertOn =
      ( psCur->logicallyOn AND !psCur->interlocked AND
        !psCur->polChgStarted AND !psCur->testStarted AND
        ((psCur->supplyType == ST_Bipolar)
                 OR !(psCur->nearZero AND psCur->needPolarityChg)) );


  //----------------------------------------------------------------------
  // psNext->testStarted
  //    Description:
  //       The time (in ms) that a type-test cycle was started.  A test
  //       for the supply type is started for the following events:
  //         - The supply is being turned off
  //         - The supply is off and is about to be turned on
  //         - A reset request was rcvd while the supply was off
  //    Inputs:
  //       psCur->onRequest
  //       psCur->resetRequested
  //       psCur->logicallyOn
  //       psCur->assertOn
  //       psCur->offPending
  //-----------------------------------------------------------------------
    if (psCur->testStarted)
      if (msSince(psCur->testStarted) >= 1000)  psNext->testStarted = 0;

    if (psCur->onRequest OR psCur->resetRequested)
      if (!ignoreResetReqs[ui])
        if (!psCur->testStarted)  psNext->testStarted = GetMS();

    if (psCur->offPending)
      if (!psCur->testStarted)  psNext->testStarted = GetMS();



  //----------------------------------------------------------------------
  // psNext->onPending, psNext->onRequest(reset)
  //    Description:
  //       onPending:  A change to ON mode is pending the end of a type-test
  //                   cycle
  //       onRequest:  reset once request is acknowledged
  //    Inputs:
  //       psCur->interlocked
  //       psCur->offPending
  //       psCur->logicallyOn
  //       psCur->onRequest
  //       psCur->testStarted
  //-----------------------------------------------------------------------
    if (psCur->interlocked OR psCur->offPending OR psCur->logicallyOn)
      psNext->onPending = psNext->onRequest = false;
    else
      if (psCur->onRequest AND psCur->testStarted)  {
        psNext->onPending = true;  psNext->onRequest = false;  }



  //--------------------------------------------------------------------
  // psNext->prepForOff, psNext->offRequest(reset)
  //    Description:
  //       Set when an off cmd received while logically on to indicate
  //       the supply output should be set to 0 and the supply left on
  //       until the load current drops to near zero.
  //    Inputs:
  //       psCur->offRequest
  //       psCur->onRequest
  //       psCur->logicallyOn
  //--------------------------------------------------------------------
    if (psCur->offRequest)  {
      psNext->offRequest = false;
      if (psCur->logicallyOn)  psNext->prepForOff = true;
    }
    if (psCur->onRequest OR !psCur->logicallyOn)  psNext->prepForOff = false;


  //--------------------------------------------------------------------
  // psNext->offPending
  //    Description:
  //       Set to cause the supply to be turned off.  This is happens
  //       when a prefForOff cycle has ended because the load current has
  //       dropped below the set threshold.
  //    Inputs:
  //       psCur->prepForOff
  //       psCur->nearZero
  //       psCur->logicallyOn
  //--------------------------------------------------------------------
    if (psCur->prepForOff AND (psCur->nearZero OR psCur->setReadErr))
      psNext->offPending = true;

    if (!psCur->logicallyOn)  psNext->offPending = false;


  //----------------------------------------------------------------------
  // psNext->logicallyOn
  //    Description:
  //       The supply should be on if not in the middle of changing the
  //       polarity
  //    Inputs:
  //       psCur->interlocked
  //       psCur->offPending
  //       psCur->onPending
  //       psCur->testStarted
  //-----------------------------------------------------------------------
    if (psCur->interlocked OR psCur->offPending)
      psNext->logicallyOn = false;
    else
      if (psCur->onPending AND !psCur->testStarted)  psNext->logicallyOn = true;


  //-----------------------------------------------------------------------
  // psNext->interlocked, psNext->resetRequested(reset)
  //    Description:
  //       Indicates a fault condition exists or did exist and has not
  //       been cleared.
  //    Inputs:
  //       psCur->activeIntlk
  //       psCur->resetRequested
  //-----------------------------------------------------------------------
    if (psCur->resetRequested)  {
      if (!psCur->activeIntlk)  {
        psNext->interlocked = false;  psNext->resetRequested = 0; }
      else  {
        if (msSince(psCur->resetRequested) > resetGraceTime)  {
          printf("PS%d (%.16ls): Active Interlock still exists\n",
                 ui+1, psNames[ui]);
          psNext->resetRequested = 0;
        }
      }
    }

    if (psCur->activeIntlk)  psNext->interlocked = true;


  //----------------------------------------------------------------------
  // psNext->cantBeUnipolar
  //    Description:
  //       Keeps track of whether the input/output conditions that indicate
  //       the presence of the wiring for a unipolar supply have failed since
  //       the start of the last test.
  //
  //       Unipolar supply wiring connects the polarity command output to the
  //       PS On/Off status input (not used for unipolar supplies), and the
  //       PS Interlock status is tied high (no interlock).
  //
  //       This means that, if the On/Off status input always follows the
  //       polarity output, and the PS Interlock status input is always high
  //       (no interlock) then the wiring is for a unipolar supply.
  //
  //       Since the we don't want to be changing the polarity on a uniploar
  //       supply while it is turned on, we limit testing for these conditions
  //       to when the supply is off.
  //    Inputs:
  //       psCur->testStarted
  //       psCur->psInterlock
  //       psCur->testIn
  //       psCur->testOut
  //-----------------------------------------------------------------------
    if (psCur->testStarted AND (msSince(psCur->testStarted) < 50))
      psNext->cantBeUnipolar = false;
    else  {
      if ((psCur->testIn != psCur->testOut) OR psCur->psInterlock)
        psNext->cantBeUnipolar = true;
    }


  //-----------------------------------------------------------------
  // psNext->supplyType
  //    Description:
  //       Keeps track of the type of supply connected (unipolar or bipolar).
  //       Whenever a test of the supply type is NOT running, the type is
  //       set based on whether or not the last test determined that a unipolar
  //       supply was connected.
  //    Inputs:
  //       psCur->testStarted
  //       psCur->cantBeUnipolar
  //-----------------------------------------------------------------
    if (psPrev->testStarted AND !psCur->testStarted)
      psNext->supplyType = (psCur->cantBeUnipolar ? ST_Bipolar : ST_Unipolar);

  }

}


//=============================================================================
//  Next we update output signal values to be written to the outputs at
//  the beginning of the next pass (just after the inputs are read).
//=============================================================================
void updateOutputState(void)
{
  uint  ui;
  bool  assertReset;
  float  setpoint;
  ulong  testRunningFor;
  PSState *psNext;


  if (current != applTask)  return;  // must only be called by "applic" task

  for (ui=0; ui<NumPS; ++ui)  {
    psNext = Next.ps + ui;

    if (!psNext->testStarted)
      assertReset = false;
    else  {
      testRunningFor = msSince(psNext->testStarted);
      assertReset = ((testRunningFor >= 250) AND (testRunningFor <= 750));
    }


  //--- On/Off digital output ---
    if (psNext->assertOn)  TurnSupplyOn(ui);  else  TurnSupplyOff(ui);

  //--- Polarity (unipolar)/Reset (biploar) /Test (both) digital output ---
    if (psNext->testStarted)  {
      if (assertReset)  ApplyReset(ui);  else  ClearReset(ui);  }
    else  if (psNext->supplyType == ST_Unipolar)  {
      if (psNext->assertPosPolarity)
        SetPosPolarity(ui);  else  SetNegPolarity(ui);
    }
    else
      ClearReset(ui);

  //--- Analog output that controls power supply output ---
    setpoint = psNext->activeSetpoint;
    if (psNext->supplyType == ST_Unipolar)  setpoint = FABS(setpoint);
    setVal[ui] = setpoint;
  }

}


//=============================================================================
// The last thing we do is update values read by other tasks
//=============================================================================
// IMPORTANT NOTES and WARNINGS:
//
//   To avoid concurrency issues, other tasks need to use locks when reading
//   values set here.  But be careful to keep such locks short to avoid
//   delaying the timing of the main application task.  If necessary, split
//   such code in to sections, getting and releasing the lock(s) in each
//   section to avoid increasing latency for the main application task.
//
//   Values in the "next" instance of the application state fall in to 2
//   categories:  Inputs and Outputs.
//
//   The "next" Outputs are those that may be changed in updateOutputState().
//   The reset are Inputs, and may be changed by other tasks within a lock.
//
//   Since each pass through the application logic starts by setting
//   cur = next, any "next" values set by other tasks become part of the
//   "cur" instance used by the rest of the applic logic.  Because the
//   "applic" task at a higher priority than any other real-time task, the
//   changes to "next" values by other tasks can ONLY happen between passes
//   through the applic logic loop.
//
//   NOTES:
//     - Some of these "next" input values values must be initialized in
//       updateCurState() in order to tell when a value is set by another
//       task.  For example, if we have a boolean flag that is set when we
//       get an ON command, then (after its "next" value is copied to "cur"),
//       we have to clear the flag in order to tell if another ON command is
//       received between passes.
//
//     - As with reads to these values, other tasks must use locks when
//       writing to the "next" input values to insure they are not preempted
//       by the "applic" task in the middle of a write.
//
//=============================================================================
void updateExternals(void)
{
  uint  ui;
  ulong  interlockedMask, statusMask, ramp0Mask;
  PSState *psCur;


  if (current != applTask)  return;  // must only be called by "applic" task

  digInBits = Cur.digitalIn;
  digOutBits = Cur.digitalOut;

//APPLCODE

  statusBitsAppl = statusBitsAppl2 = 0;
  interlockedMask = 0x01000000L;
  statusMask = 0x00000001L;
  ramp0Mask = 0x00000001L;

  for (ui=0; ui<NumPS; ++ui)  {
    psCur  = Cur.ps + ui;

//--- update settings based on values that were USED by the application ---
//--- logic (i.e. "cur" values that serve as "inputs" to the logic) ---

    if (psCur->externInterlock)  statusBitsAppl |= statusMask;
    statusMask <<= 1;

    if (psCur->psInterlock)  statusBitsAppl |= statusMask;
    statusMask <<= 1;

    if (psCur->psIsOn)  statusBitsAppl |= statusMask;
    statusMask <<= 1;

    if (psCur->interlocked)  statusBitsAppl |= interlockedMask;
    interlockedMask <<= 1;

    if (psCur->prepForOff)  statusBitsAppl2 |= ramp0Mask;
    ramp0Mask <<= 1;

    ISET(ui) = psCur->requestedAmps;
    IREAD(ui) = psCur->actualAmps;
    VREAD(ui) = psCur->actualVolts;
    IREADAVGD(ui) = psCur->avgdAmps;

//--- update settings based on values that were UPDATED by the application ---
//--- logic (i.e. "next" values that serve as "outputs" from the logic) ---

  }

}

#pragma warn 775


//=============================================================================
_nearcall float readGetRaw(uint ui)  {
  return (ui < NumHdwADC ? anaIn[ui] : 0.0);  }

_nearcall float readMinRaw(uint ui)  {
  return (ui < NumHdwADC) ? (adcMinRaw(ui) - adcOffset[ui]) * adcGain[ui] : -9.8; }

_nearcall float readMaxRaw(uint ui)  {
  return (ui < NumHdwADC) ? (adcMaxRaw(ui) - adcOffset[ui]) * adcGain[ui] : 9.8; }

_nearcall void readDefaults(uint ui, float *gain, float *offset)  {
  *gain = (ui & 1) ? IReadGainDefault : VReadGainDefault;  *offset = 0.0; }

_nearcall void readCalibPoints(uint ui, float *low, float *high)  {
  *low = 1.0;  *high = (ui & 1) ? 19.0 : 9.0;  }

_nearcall char *readUnits(uint ui, char *buf)  { return "V"; }


//=============================================================================
_nearcall float setGetRaw(uint ui)  {
  return (ui < NumHdwADC ? anaOut[ui] : 0.0);  }

_nearcall float setMinRaw(uint ui)  {
  return (ui < NumHdwDAC) ? (dacMinRaw(ui) - dacOffset[ui]) * dacGain[ui] : -9.8; }

_nearcall float setMaxRaw(uint ui)  {
  return (ui < NumHdwDAC) ? (dacMaxRaw(ui) - dacOffset[ui]) * dacGain[ui] : 9.8; }

_nearcall void setDefaults(uint ui, float *gain, float *offset)  {
  *gain = ISetGainDefault;  *offset = 0.0;  }

_nearcall void setCalibPoints(uint ui, float *low, float *high)  {
    *low = 1.0;  *high = 19.0; }

_nearcall char *setUnits(uint ui, char *buf)  { return "V"; }


//=============================================================================
//  Set default labels for all the I/O signals
//=============================================================================
void setDefaultLabels(void)
{
  uint  ui, uidx, psNum;

  printf("Setting default labels for I/O signals\r\n");

//--- set application-appropriate default labels ---
//APPLCODE

//--- application-default offset and gain values and labels ---
  for (ui=0; ui<NumPS; ++ui)  {
    psNum = ui+1;

    sprintf(ainLabel[ui<<1], "V_RD_%u", psNum);
    sprintf(ainLabel[(ui<<1)+1], "I_RD_%u", psNum);

    sprintf(aoutLabel[ui], "I_SET_%u", psNum);

    uidx = ui * 3;
    sprintf(dinLabel[uidx],    "EXT_ILK_OK_%u", psNum);
    sprintf(dinLabel[uidx+1],  "PS_ILK_OK_%u", psNum);
    sprintf(dinLabel[uidx+2],  "PS_IS_ON_%u", psNum);

    sprintf(doutLabel[ui<<1],     "SET_ON_%u", psNum);
    sprintf(doutLabel[(ui<<1)+1], "RESET_%u", psNum);
  }
}


//=============================================================================
//  This function is called to force software configurable values to default
//  values appropriate for this application.
//=============================================================================
void setDefaultConfig(void)
{
  uint  ui;


  printf("Setting in-RAM device config to default values\r\n");

//===== set default values for persistent data values =====
//APPLCODE

  for (ui=0; ui<NumPS; ++ui)  {
    ISET_MIN(ui) = -1000.0;
    ISET_MAX(ui) = 1000.0;
    MaxReadErr(ui) = 1.000;
    MaxRampRate(ui) = 100.0;  // max rate of chg for setpoint chgs (per second)
    farsprintf(psNames[ui], "Supply%u", ui+1);
    IlkIfReadErr(ui) = 0;
    MinRampRate(ui) = 0.250;  // min rate of chg considererd ramping (in amps/sec)
    MaxErrTime(ui) = 5000;  // max # ms set/read error ignored
  }

  polChgDelay = 500;  // delay in ms between polarity chg and ON re-asserted

  resetGraceTime = 500;  // max # ms between reset and end of extern intlks

  repeatMsgsSecs = 600;  // # secs between msgs for a persistent err cond
  logStateSecs = 600;  // # secs between logging the state of the PS's
}


//=============================================================================
//  This function is called by the application loop just before a shutdown.
//  The appliation must do whatever is necessary for a clean shutdown and
//  then exit
//=============================================================================
void applicShutdown(void)
{
  printf("Application prepping for shutdown\r\n");  fflush(stdout);

  Exit(0);
}

//=============================================================================
//  This is the highest priority task, usually resumed at a regular interval
//  so that it can do what it needs and sleep again until the start of the
//  next interval.
//=============================================================================
void _task applic(int argc, char **argv)
{
  #define SharedApplLocals
  #include <sharedAppl.c>

//--- add local variable definitions here ---

  PSState *psCur;

  #define SharedApplLocals2
  #include <sharedAppl.c>

//--- add local array definitions here ---


//--- place code here to set initial default values ---
  allowSoftReset = true;

  for (ui=0; ui<NumPS; ++ui)  {
    psCur = Cur.ps + ui;
    psCur->activeSetpoint = setVal[ui] = 0.0;
  }

  #define SharedApplInit
  #include <sharedAppl.c>

  _EnterCritical_();  SchedMS = 20;  _ExitCritical_();

//--- add application initialization code here ---

  lastLogStateSecs = time(NULL);

  #define SharedApplInit2
  #include <sharedAppl.c>

//--- initializations we want to have effect even after the non-application ---
//--- specific defaults and any state values restored by a soft-reset ---
  for (ui=0; ui<NumPS; ++ui)  {
    psCur = Cur.ps + ui;
    setAvgPeriodADC(I_RD_IDX(ui), 2000);
    psCur->lastOnOKTime = psCur->lastReadOKTime = GetMS();
  }


//--- The following loop is run at a periodic interval (normally 100 times ---
//--- per second, or every 10ms).  The application process normally ---
//--- consists of reading the current state of the digital and analog ---
//--- inputs, changing the state of the analog and digital outputs to the ---
//--- values computed during the previous pass, then computing new values ---
//--- for the outputs and updating values that outside clients can view ---
//--- (currently accessable as a set of ModBus registers) ---

  for (;;)  {
    #define SharedApplLoopStart
    #include <sharedAppl.c>

  //--- read the current state of the analog and digital Input signals, and ---
  //--- write the values computed at the end of the last pass ---
    digitalOut = Next.digitalOut;  rwIO();

  //--- process next event message (if any) ---
    processMsg(eventsfd, (EventMsg *)buf, sizeof(buf));

  //--- update the current state and determine the next state ---
    updateCurState();  updateNextState();
    updateOutputState();  updateExternals();

    #define SharedApplLoopEnd
    #include <sharedAppl.c>
  }

}


//=============================================================================
//  list of application-specific commands that are normally entered via a
//  user shell
//=============================================================================
//void _task nsclSeriald(int argc, char **argv);

//--- add any application-specific commands here ---
const CmdInfo far applCmds[] = {
//  { "misc",        miscTest,       "..." },
//  { "pscalib",   pscalib,     "Calibbrate Setting/Readings for a Supply", },
  { "cmd",       cmd,         "Turn a PS ON or OFF, or RESET one", },
  { "psname",    psname,      "Show/Set names for each power supply" },
  { "state",     showState,   "Show state of controller" },
  { "tracesup",  traceSup,    "Enable/disable detailed logging for a supply" }
};

int const NumApplCmds = ( sizeof(applCmds) / sizeof(CmdInfo) );

//-----------------------------------------------------------------------------
#endif
