//----- modbus.c ----- 12/05/12 -----(05/08/06)-----

// based on earlier work for ZWorld/Rabbit-based controllers

//  Application-specific register definitions are in the application source
//  file  (in the application-specific directory)

//-----------------------------------------------------------------------------
//  Update log (most recent first):
//
//
//  2012-12-05 - Mark Davis
//               - Moved MB_cmdBitsStd() and MB_timeZone() from modbus.c
//                 to sharedApplic.c
//
//  2012-10-02 - Mark Davis
//               - perror() rather than printf() is now used to log the error
//                 when reading the start of the next modbus cmd times out.
//                 This may provide more information about why these TCP
//                 connections often fail during a restart of an EPICS SIOC.
//
//  2012-07-18 - Mark Davis
//               - Changes to PV_Desc and related macros to support locking
//                 and PVs without a separate read buffer.
//
//  2012-05-31 - Mark Davis
//               - pvCheckAllEvents() now limits the # of non-register PV
//                 events based on the total # of bytes those events represent.
//
//  2012-03-27 - Mark Davis
//               - Began changes to support PV types defined by structures
//                 other then RegInfo.
//
//  2012-02-23 - Mark Davis
//               - Added the anyChg option to pvCheckAllEvents(), which allows
//                 the use of the much faster mode of checking for ANY change
//                 in the value of the PVs (compared to computing what the
//                 change was and comparing it to the minChg setting for the
//                 PVs).
//
//  2012-02-23 - Mark Davis
//               - A list of pointers to PV's is now created and sorted by name.
//                 The index in to this array is now used as the ID for each PV.
//                 This allows us to do a binary search when looking for a PV
//                 by name, which makes pvGetID() faster (which is important
//                 for the CA protocol which has very short timeouts).
//
//  2012-02-14 - Mark Davis
//               - Changed references to pid in the PV functions to just id
//                 to avoid confusion with process IDs.  For the same reasons
//                 I made the following func name changes:
//                    pvGetPID() is now pvGetID()
//                    pidToRegInfo() is now pvToRegInfo()
//               - Added temporary code to provide "fake" PVs RTYP, STAT, and
//                 DESC to make Qt CA interface happy.
//
//  2011-05-23 - Mark Davis
//               - Fixed bug in fmtVal() so it handles type S32 properly.
//
//  2011-05-09 - Mark Davis (*)
//               - Made sure all the places that use the RegInfo.setRegFunc
//                 value check to make sure it is not NULL first (although
//                 this shouldn't actually be possible, since the macros that
//                 set them complain if you try to use NULL rather than roReg
//                 for read-only 4xxxx registers).
//
//  2011-05-04 - Mark Davis
//               - Logging of messages for writes to registers are now
//                 supressed until startupComplete is true.  Prior to this,
//                 the last change (use of the new restoreIfRegValue() func)
//                 would cause the syslog log buffer to overflow during startup
//                 before it could be copied to the server.
//
//  2011-04-22 - Mark Davis
//               - Added restoreIfRegValue() for other modules to call if they
//                 have a value to write to memory that may be a register.
//                 This is now used by the functions in the backup.c module
//                 that restore values from a file.  If no value-specific
//                 restore function is specified, restoreIfRegValue() is called
//                 to cause the register-specific processing to be invoked if
//                 the value is a register/PV value.  This insures that any
//                 value-specific processing is invoked for regsiter values
//                 that are saved/restored to/from files.
//
//  2011-04-19 - Mark Davis
//               - Added MB_setACoil() and MB_resetACoil()
//
//  2010-12-14 - Mark Davis
//               - Changed _TimerSeconds + secTimerOffset to TIME()
//
//  2010-08-04 - Mark Davis
//               - Added setF32_ABS()
//
//  2010-06-10 - Mark Davis
//               - Changed PID values for pvXXX functions from long to int and
//                 added a pointer to the PV's name to the PV_Info structure.
//
//  2010-05-27 - Mark Davis
//               - Added following functions to support EPICS CA/PV interface:
//                   pvCheckAllEvents(), pvCheckEvents()
//
//  2010-05-25 - Mark Davis
//               - Added following functions to support EPICS CA/PV interface:
//                   pvGetPID(), pvGetInfo(), pvGetVal(), pvSetVal()
//
//  2010-01-14 - Mark Davis
//               - Added setMBReg() and setRegXXX() functions to allow code
//                 to change a register value and invoke the change function
//                 (if any) specified in the definition of the register.
//
//  2009-10-26 - Mark Davis
//               - changes to work with register names ref via far ptr
//
//  2009-10-06 - Mark Davis
//               - Updated call to saveParams() to include new parameter that
//                 tells it to save to flash (which was the only option before)
//
//  2009-04-30 - Mark Davis
//               - modbusScanTables() now returns a value indicating if there
//                 is a problem.
//
//  2008-02-23 - Mark Davis
//               - added global values to support "software oscilliscope"
//                 for real-time capture of register values
//
//  2007-03-20 - Mark Davis
//               - removed all KLOCK()/KUNLOCK() calls in ModBusCmd() for 2
//                 reasons:  It improves concurrency of scheduling, and write
//                 functions can block on a shared resource which would lock
//                 up the system if the scheduler was disabled.
//
//  2007-02-12 - Mark Davis
//               - removed an extra KUNLOCK() call in ModBusCmd()
//
//  2006-05-08 - Mark Davis
//               - started work on new version for DEBROS base
//
//-----------------------------------------------------------------------------
//
//  Module to support standard ModBus register interface.
//
//  The primary goals of this particular implementation:
//
//  1. Using values accessable via the ModBus interface should not affect
//     the complexity, efficiency, or portability of the application code.
//
//  2. Defining the relationship between register #'s and values used in the
//     application must be as simple and straighforward as possible.  Requiring
//     additions or changes in multiple places (even within a single file)
//     should be avoided if at all possible.
//
//  3. Ability for multiple instances of a logical device to share a single
//     set of register definitions (i.e. avoid having to define one set of
//     registers for each instance).
//
//  4. Provide a generic diagnostic interface (separate from the ModBus
//     interface itself) for examining and modifying the logical values
//     accessed via the ModBus interface (e.g. if 2 registers map to a
//     4-byte floating-point value, then you should be able to display and/or
//     modify the value by giving the name of the floating point value).
//
//-----------------------------------------------------------------------------
//
//  When designing this interface I considered 2 basic approaches:
//
//  - Store all register values as a contiguous array of integer values
//    and define a bunch of macros to simplify the application code that
//    needs to access one or more words as a single logical value (for
//    example:   #define FloatValX  (*(float *)intArray+10))
//
//  - Store all values as standard C variables and map the register #'s to
//    the variables.
//
//  The advantage of the 1st method is that implementation of the ModBus
//  interface would (in theory) be simple and efficient, making it more
//  practical to support more or faster ModBus clients.  This was very
//  tempting, but I soon realized there where some problems with this
//  simple approach:
//
//  - Portability:  Referencing an arbitrary location in an array (even an
//    array of 16-bit values) via a pointer to a type that is larger than
//    16 bits can lead to allignment problems (if not outright failure, then
//    at least reducing the CPU's efficiency compared to the same operation
//    on a properly alligned value).  Not a problem on 8-bit, or even 16-bit
//    controllers, but on 32-bit ones, a definite possibility.
//
//  - Byte order:  Different CPU's store values in different byte order,
//    and different ModBus clients expect the bytes or words for multi-word
//    values in different orders.  So simply copying a group of bytes from
//    the array does not result in a usable response for all ModBus clients.
//
//  - Increased size and Decreased efficiency of the application code:
//    Having to use pointers to indirectly reference all the values
//    accessable via the ModBus interface requires more code and an added
//    level of indirection.  Best case would be a compiler that supports
//    declaring const pointers and treats dereferenced const pointer values
//    the same as static addresses (i.e. as a reference to a standard variable
//    type).  Don't know how many compilers (if any) could manage this.
//
//  - Definition of the logical values (from the application's view) and
//    the registers are separate tasks.  This requires keeping two
//    interdependant sets of definitions in sync (and we all know how likely
//    that is to cause a problem!).
//
//  - It does absolutely nothing to support the goal of avoiding having one
//    set of register definitions per logical device.
//
//
//  Given the above issues with the "straighforard" approach, and the fact
//  that it still does not help me at all with my goal of a diagnostic
//  interface, I opted for the 2nd approach:  Use standard C variables, at
//  the cost of a (possibly) less efficient ModBus interface.
//
//
//  The methodology I implemented works as follows:
//
//  - All logical values and how they relate to the ModBus register interface
//    are defined in the application source file.  With just a single line
//    entry for each logical value I was able to handle all the following:
//
//     - Definition and declaration of a standard C variable type
//
//     - A statically initialized structure of information about each value
//       that includes:
//        - the registers encompased by the variable
//        - the variable name
//        - the data type
//        - a pointer to variable
//        - a single set of register definitions that can be used by multiple
//          instances of a logical device
//
//     - A statically initialized array of pointers, where each entry in the
//       array corresponds to a single register, and points to the information
//       structure for the variable that contains the register value (note that
//       multiple entries can point to the same information structure).
//
//     - A single set of register definitions that can be used by multiple
//       instances of a logical device
//
//  The last item allows the ModBus interface to quickly locate (through
//  a few indirections, I admit) the needed register value given the
//  register # and the type of client asking for it.  The client type is
//  important since there is no standard for the register order for logical
//  values composed of mulitple registers (like a 32-bit float or integer).
//
//  To deal with the lack of standard register order, the ModBus interface
//  uses a table to determine how to reorder registers for each combination of
//  client type and data type.  This is why it is important to know what kind
//  of logical value (C data type) each register is a part of.
//
//-----------------------------------------------------------------------------
//  Open issues/To-Do:
//
//  Given that registers are global variables, reads and writes to them need
//  to be protected.  A Modbus read should be able to read all the values
//  as they are at a given moment, and writes should be an atomic transaction
//  (i.e. changes to a group of registers should NOT be interrupted).
//
//  To address this, the reads really have to disable the scheduler while
//  copying the requested values, and writes send a message to the main
//  application process, and wait for a response.  Simply locking the register
//  values does not accomplish everything:
//
//  - The application logic is not informed of the change.  Even if it were to
//    check for changes, it would not know the order the changes were made in,
//    which can often be critical.  It could even miss changes that toggled a
//    a value.
//
//  - The registers values are simply global variables.  Allowing more than
//    one process to write to them means ALL processes must imploy locking
//    when accessing or altering them.
//
//  Seems I can't escape the simple fact that I have to coordinate the
//  processing of all events that potentially change the state of the
//  application or device, REGARDLESS of what process actually requests the
//  change.  The easiest way to do that is to make one process responsible for
//  all such changes and use msgs to request changes (with responses where
//  feedback is needed).
//
//  Given that reads should be fairly quick, I might be able to get away with
//  simply disabling the scheduler during the read transaction to insure the
//  copying of the values is not interrupted.  Using just a lock would
//  protect the data as well, but would then allow the application process to
//  run and find it could not do a pending write request from a different
//  process (possibly OK if I simply defer ALL msg processing when the
//  current request cannot be processed).
//
//-----------------------------------------------------------------------------

#include <common.h>

#include <sys/types.h>
#include <unistd.h>
#include <fcntl.h>
#include <errno.h>

#include <string.h>

#include <sys/backup.h>
//#include <sys/hdwDefs.h>

#include <sys/modbus.h>

#include <math.h>


extern bool  startupComplete;  //temp


#define MaxCmdMsgLen    64
#define MaxRespMsgLen  256


#define  ModbusWriteBits    0x00000001
#define  ModbusReadBits     0x00000002

#define  ShowWrites   (DiagFlags & ModbusWriteBits)
#define  ShowReads    (DiagFlags & ModbusReadBits)


const char *typeDesc[NUM_MB_TYPES] =
       { "S16", "U16", "S32", "U32", "F32", "STR", "U8", "S8" };


//--- table to control how to swap register #'s by client and data types ---
//--- #'s represent amount to change RegInfo.wordIdx by to get the ---
//--- register value expected for different client types ---
const schar regOrder[4][5][2] = {  // [clientType][dataType][wordIdx]
//   --I16--   --U16--   --I32--   --U32--   --F32--
  { { 0, 0 }, { 0, 0 }, { 0, 0 }, { 0, 0 }, { 0, 0 } }, // CT_MBSerial
  { { 0, 0 }, { 0, 0 }, { 1,-1 }, { 1,-1 }, { 0, 0 } }, // CT_MBNSCL
  { { 0, 0 }, { 0, 0 }, { 1,-1 }, { 1,-1 }, { 0, 0 } }, // CT_MBTCP
  { { 0, 0 }, { 0, 0 }, { 0, 0 }, { 0, 0 }, { 0, 0 } }  // CT_Native
};
typedef schar ClientOrder[5][2];


//--- Save 16-bit value to a Byte array in Big-Endian (MSB-1st) format ---
#define  SetWordVal(offset,val,buf)  \
  (((buf)[offset] = (val) >> 8), ((buf)[(offset)+1] = (val) & 0xff))


#include <sys/farprintf.h>

PV_Ref far *pvList;   // array of PV_Ref structures


uint  numC0Coils, numR3Regs, numR4Regs, numRegPVs, numNonRegPVs, numPVs;


//--- defs for "SoftScope" values ---
uint  numScopeRegs, captureIdx, scopeBufSampleSize, bytesPerScopeSample;
ulong  captureCount;
RegInfo  *scopeList[MaxWatchSize];
char far  *scopeBuf, *capturePtr;


//-----------------------------------------------------------------------------
//  Parse a string value according to the data type of the specified register
//
//  Returns the value in native binary form in the given buffer
//-----------------------------------------------------------------------------
void *parseValue(RegInfo *reg, char *valStr, char *buf)
{
  switch (reg->dataType)  {

    case DT_S16:
    case DT_U16:
      *(short *)buf = ATOI(valStr);  break;

    case DT_S32:
    case DT_U32:
      *(ulong *)buf = ATOL(valStr);  break;

    case DT_F32:
      *(float *)buf = atof(valStr);
      break;
  }

  return buf;
}

//-----------------------------------------------------------------------------
//  format a numeric value as a string, blnak filling the string on the left
//  if necessary to insure the result is at least minWidth chars in length
//-----------------------------------------------------------------------------
char *fmtVal(char *buf, uchar type, int minWidth, void *val)
{
  int  sigDigits, len, fill;

  *buf = 0;

  switch (type)  {
    case DT_S16:  sprintf(buf, "%d",  *(S16 *)val);  break;
    case DT_U16:  sprintf(buf, "%u",  *(U16 *)val);  break;
    case DT_S32:  sprintf(buf, "%ld", *(S32 *)val);  break;
    case DT_U32:  sprintf(buf, "%lu", *(U32 *)val);  break;
    case DT_F32:
      sigDigits = 6;
      if (minWidth > 8)  sigDigits = minWidth - 6;
      Engineering(*(F32 *)val, sigDigits, buf);
  }

  if ((len = strlen(buf)) < minWidth)  {
    fill = minWidth - len;
    memmove(buf+fill, buf, len+1);
    memset(buf, ' ', fill);
  }

  return buf;
}

//-----------------------------------------------------------------------------
//  format a register value as a string
//-----------------------------------------------------------------------------
char *fmtRegVal(RegInfo *reg, S16 *useVal, uint chanNum, char *buf)
{
  S16  *valPtr;

  if (useVal)
    valPtr = useVal;
  else  {
    valPtr = reg->regVal;
    if (reg->flags & OnePerChan)  valPtr += chanNum * reg->numRegs;
  }

  fmtVal(buf, reg->dataType, 12, valPtr);

  if (reg->numRegs == 1)
    sprintf(buf+12, " (0x%04x)    ", *(U16 *)valPtr);
  else
    sprintf(buf+12, " (0x%08lx)", *(U32 *)valPtr);

  return buf;
}



//-----------------------------------------------------------------------------
//  generic routine to set the value of a coil.  Used when no special
//  processing is required when changing a specific value.
//-----------------------------------------------------------------------------
bool _nearcall setCoil(SetCoilFuncParams)
{
  if ((COIL->flags & LogChanges) AND startupComplete)
    fprintf(DiagLog, "setCoil: %ls[%d] = %d\n", COIL->name, chanNum, newVal);

  COIL->coilVal[chanNum] = newVal;

  return true;
}

//-----------------------------------------------------------------------------
//  Function called by MB_setACoil() or MB_resetACoil() for the purpose of
//  changing the value of the specified coil.  This differs from setCoil()
//  in that this function CALLs the function specified in the coil definition,
//  whereas setCoil() is just the default function specified in coil
//  definitions when no special processing is required at the time the coil
//  is set.
//-----------------------------------------------------------------------------
bool setCoilVal(uint coilNum, uint chanNum, uchar coilVal, fd_t eventsfd)
{
  bool  accepted;
  CoilInfo  *coil;


  if (coilNum > numC0Coils)  return false;

  coil = c0Coils[coilNum-1];
  chanNum = ((uint)coil->flags & OnePerChan ? chanNum : 0);

  if (!coil->setCoilFunc)  return false;

  accepted = (coil->setCoilFunc)(coil, chanNum, coilVal, eventsfd);

  if ((coil->flags & LogChanges) AND startupComplete)
    fprintf(DiagLog, "set %ls[%d] = %u: %s\r\n\n",
            coil->name, chanNum, coilVal, (accepted ? "OK" : "Rejected")  );

  return accepted;
}

//-----------------------------------------------------------------------------
//  Got a new value for the setACoil register, which attempts to set the coil
//  whose ID is the value writtent to the setACoil register.
//-----------------------------------------------------------------------------
bool _nearcall MB_setACoil(SetRegFuncParams)
{
  return setCoilVal(MB_VAL(U16), chanNum, 1, eventsfd);
}

//-----------------------------------------------------------------------------
//  Got a new value for the resetACoil register, which attempts to clear the
//  coil whose ID is the value writtent to the setACoil register.
//-----------------------------------------------------------------------------
bool _nearcall MB_resetACoil(SetRegFuncParams)
{
  return setCoilVal(MB_VAL(U16), chanNum, 0, eventsfd);
}

//-----------------------------------------------------------------------------
//  generic routine to serve as placeholder for registers that shouldn't be
//  set
//-----------------------------------------------------------------------------
bool _nearcall roReg(SetRegFuncParams)
{
  if ((REG->flags & LogChanges) AND startupComplete)
    fprintf(DiagLog, "attempt to change ReadOnly reg: %ls\n", REG->name);

  return false;
}

//-----------------------------------------------------------------------------
//  generic routine to set the value of a register.  Used when no special
//  processing is required when changing a specific value.
//-----------------------------------------------------------------------------
bool _nearcall setS16(SetRegFuncParams)
{
  if ((REG->flags & LogChanges) AND startupComplete)
    fprintf(DiagLog, "setS16: %ls[%d] = %d\n",
            REG->name, chanNum, MB_VAL(S16) );

  _EnterCritical_();
  ((int *)REG->regVal)[chanNum] = MB_VAL(S16);
  _ExitCritical_();

  return true;
}

//-----------------------------------------------------------------------------
//  generic routine to set the value of a register.  Used when no special
//  processing is required when changing a specific value.
//-----------------------------------------------------------------------------
bool _nearcall setU16(SetRegFuncParams)
{
  if ((REG->flags & LogChanges) AND startupComplete)
    fprintf(DiagLog, "setU16: %ls[%d] = %u\n",
            REG->name, chanNum, MB_VAL(U16) );

  _EnterCritical_();
  ((uint *)REG->regVal)[chanNum] = MB_VAL(U16);
  _ExitCritical_();

  return true;
}

//-----------------------------------------------------------------------------
//  generic routine to set the value of a register.  Used when no special
//  processing is required when changing a specific value.
//-----------------------------------------------------------------------------
bool _nearcall setS32(SetRegFuncParams)
{
  if ((REG->flags & LogChanges) AND startupComplete)
    fprintf(DiagLog, "setS32: %ls[%d] = %ld\n",
            REG->name, chanNum, MB_VAL(S32) );

  _EnterCritical_();
  ((long *)REG->regVal)[chanNum] = MB_VAL(S32);
  _ExitCritical_();

  return true;
}

//-----------------------------------------------------------------------------
//  generic routine to set the value of a register.  Used when no special
//  processing is required when changing a specific value.
//-----------------------------------------------------------------------------
bool _nearcall setU32(SetRegFuncParams)
{
  if ((REG->flags & LogChanges) AND startupComplete)
    fprintf(DiagLog, "setU32: %ls[%d] = %lu\n",
            REG->name, chanNum, MB_VAL(U32) );

  _EnterCritical_();
  ((ulong *)REG->regVal)[chanNum] = MB_VAL(U32);
  _ExitCritical_();

  return true;
}

//-----------------------------------------------------------------------------
//  generic routine to set the value of a register.  Used when no special
//  processing is required when changing a specific value.
//-----------------------------------------------------------------------------
bool _nearcall setF32(SetRegFuncParams)
{
  char  temp[16];


  if ((REG->flags & LogChanges) AND startupComplete)
    fprintf(DiagLog, "setF32: %ls[%d] = %s\n",
            REG->name, chanNum, Engineering(MB_VAL(F32), 6, temp) );

  _EnterCritical_();
  ((float *)REG->regVal)[chanNum] = MB_VAL(F32);
  _ExitCritical_();

  return true;
}


//-----------------------------------------------------------------------------
//  generic routine to set the value of a register.  Used when no special
//  processing is required when changing a specific value.
//-----------------------------------------------------------------------------
bool _nearcall setF32_ABS(SetRegFuncParams)
{
  char  temp[16];


  if ((REG->flags & LogChanges) AND startupComplete)
    fprintf(DiagLog, "setF32_ABS: %ls[%d] = %s\n",
            REG->name, chanNum, Engineering(MB_VAL(F32), 6, temp) );

  _EnterCritical_();
  ((float *)REG->regVal)[chanNum] = FABS(MB_VAL(F32));
  _ExitCritical_();

  return true;
}


//-----------------------------------------------------------------------------
//  Generic routine to set the value of a register.  Used to evoke the
//  func specified for changing the register value and insuring the data
//  type of the new value matches that of the register.
//-----------------------------------------------------------------------------
bool setMBReg(RegInfo *reg, uchar chanNum, uchar newValType, void *newVal)
{
  if (newValType >= NUM_MB_TYPES)  {
    printf("setReg() for %ls: invalid type: %u\r\n", reg->name, newValType);
    return false;
  }

  if (reg->dataType != newValType)  {
    printf("Attemped to set %ls[%d] (type %s) from value of type %s\r\n",
           reg->name, chanNum, typeDesc[reg->dataType], typeDesc[newValType]);
    return false;
  }

  if (reg->setRegFunc)  return (reg->setRegFunc)(reg, chanNum, newVal, -1);

  return false;
}


//-----------------------------------------------------------------------------
//  Generic routines to set the value of a register.  Used to evoke the
//  func specified for changing the register value and insuring the data
//  type of the new value matches that of the register.
//
//  *** WARNING ***
//  Note that these MUST be done as real functions rather than macros to
//  ensure that the format of the data passed to setMBReg() is correct.
//
//  These functions have a specified type for the newVal argument and require
//  the value be passed as a COPY of the value on the stack (rather than as
//  a pointer to the value).  This insures that value that is passed to
//  setMBReg() (by reference) is in the format we claim it is.
//
//  While replacing these functions with macros would be slightly more
//  efficient, the macros would still have to use a local variable (i.e. one
//  located on the stack) to store a copy of newVal in the correct format and
//  pass the address of the copy to setMBReg().  Simply using typecasts alone
//  is not sufficient, since telling setMBReg() that we are passing a pointer
//  to floating point when in fact the data is in integer format is not going
//  to result in the correct register value.
//-----------------------------------------------------------------------------
bool setRegS16(RegInfo *reg, uchar chanNum, S16 newVal)
{  return setMBReg(reg, chanNum, DT_S16, &newVal);  }

bool setRegU16(RegInfo *reg, uchar chanNum, U16 newVal)
{  return setMBReg(reg, chanNum, DT_U16, &newVal);  }

bool setRegS32(RegInfo *reg, uchar chanNum, S32 newVal)
{  return setMBReg(reg, chanNum, DT_S32, &newVal);  }

bool setRegU32(RegInfo *reg, uchar chanNum, U32 newVal)
{  return setMBReg(reg, chanNum, DT_U32, &newVal);  }

bool setRegF32(RegInfo *reg, uchar chanNum, F32 newVal)
{  return setMBReg(reg, chanNum, DT_F32, &newVal);  }


//-----------------------------------------------------------------------------
//  Search the specified array of RegInfo pointers for a value with the
//  specified address.
//
//  Returns the addr of the RegInfo structure for the matching register, or
//  NULL if no match is found.
//-----------------------------------------------------------------------------
RegInfo *isReg(RegInfo **regs, void far *addr)
{
  uint  u;
  RegInfo  *reg, *lastReg;


  lastReg = NULL;
  for (u=0,reg=regs[0]; reg; ++u,reg=regs[u])  {
    if (reg == lastReg)  continue;  // skip if not 1st in multi-reg value
    if ((ulong)reg->regVal == (ulong)addr)  return reg;
    lastReg = reg;
  }

  return NULL;
}

//-----------------------------------------------------------------------------
//  Called when some other module is restoring a value that MAY be a public
//  (register/PV) value.
//
//  Return values:
//     -1:  The value is a register, but the new value was rejected
//      0:  The value is not a register
//      1:  The value is a register and the new value was accepted
//-----------------------------------------------------------------------------
int restoreIfRegValue(void far *curVal, char *newVal, uint count)
{
  bool  accepted;
  uint  chan, valSize;
  RegInfo  *reg;


  if (!(reg = isReg(r3Regs, curVal)))
    if (!(reg = isReg(r4Regs, curVal)))  return 0;

  if (! reg->setRegFunc)  return false;

  chan = 0;  valSize = reg->numRegs * 2;  accepted = true;
  while (count >= valSize)  {
//    printf("    %ls[%u]\r\n", reg->name, chan);  //tdebug
    if (! (reg->setRegFunc)(reg, chan, newVal, -1))  accepted = false;
    ++chan;  count -= valSize;  newVal += valSize;
  }

  return (accepted ? 1 : -1);
}


//-----------------------------------------------------------------------------
//  Return a pointer to the name of the specified PV
//-----------------------------------------------------------------------------
char const *pvGetName(PV_Ref far *pvPtr, char *nameCopy)
{
  char const far *name = "<invalid>";

  do  {
    if (!pvPtr)  break;

    switch (pvPtr->pvType)  {
      case PVT_REG:  name = ((RegInfo *)(pvPtr->pvDef))->name;  break;

      case PVT_EXT:  name = ((PV_Desc *)(pvPtr->pvDef))->pvInfo.name;  break;
    }
  }  while(0);

  strcpyfn(nameCopy, name);  return nameCopy;
}

//-----------------------------------------------------------------------------
//  Search for a value with the specified name.  If found, return a 16-bit
//  ID that can be used to access the value, or -1 if no such value is found.
//-----------------------------------------------------------------------------
int pvGetID(char *pvName)
{
  int  cmp;
  uint  first, last, test, max;
  PV_Ref far  *pvRef;
  char  nameBuf[MAX_PV_NAME_LEN];



//--- do binary search using the list of PVs that is sorted by name ---
  first = 0;  last = max = numPVs - 1;
  for (;;)  {
    pvRef = pvList + (test = (first + last) >> 1);

    cmp = stricmp(pvName, pvGetName(pvRef, nameBuf));

    if (ShowDebugMsgs)
      printf("[%s] %d [%s]  %u,%u,%u\r\n",
             pvName, cmp, pvGetName(pvRef, nameBuf), first, test, last);

    if (!cmp)  return (int) test;  // a match!

    if (first >= last)  break;      // tested them all?

    if (cmp < 0)  {  // reduce the range to the appropriate half
      if (test)  last = test - 1;  else   break; }
    else
      if (test < max)  first = test + 1;  else  break;
  }

  if (StrEq(pvName, "RTYP"))  return 900;   //temp
  if (StrEq(pvName, "STAS"))  return 901;   //temp
  if (StrEq(pvName, "DESC"))  return 902;   //temp

  return -1;
}


//-----------------------------------------------------------------------------
//  Get information for a PV
//
//  Returns 0 on success, < 0 if an error
//-----------------------------------------------------------------------------
int pvGetInfo(int id, PV_Info *pvInfo)
{
  RegInfo  *reg;
  PV_Desc  *pv;
  PV_Ref far  *pvRef;



//temp
  if (id >= 900)  {
    pvInfo->dataType = DT_STR;
    pvInfo->flags = PV_RD;

    switch (id)  {
      case 900:  pvInfo->name = "RTYP";  pvInfo->size = 7;  break;
      case 901:  pvInfo->name = "STAS";  pvInfo->size = 1;  break;
      case 902:  pvInfo->name = "DESC";  pvInfo->size = 12;  break;
      default:  return -1;
    }
    pvInfo->elemSize = pvInfo->size;
    pvInfo->numElems = 1;
    return 0;
  }


  if ((uint)id >= numPVs)  return -1;

  pvRef = pvList + (uint)id;

  switch (pvRef->pvType)  {

    case PVT_REG:
      reg = (RegInfo *)pvRef->pvDef;
      pvInfo->name = reg->name;
      pvInfo->dataType = reg->dataType;
      pvInfo->elemSize = pvInfo->size = reg->numRegs << 1;
      pvInfo->numElems = 1;
      pvInfo->flags = PV_RD;
      if (reg->wrMask == REG_RW)  pvInfo->flags |= PV_WR;
      return 0;

    case PVT_EXT:
      pv = (PV_Desc *)pvRef->pvDef;
      *pvInfo = pv->pvInfo;
      return 0;
  }

  return -1;
}


//-----------------------------------------------------------------------------
//  Attempt to obtain or release a read or write lock on a PV.
//
//  Returns the depth of the lock or < 0 if not successsful.
//
//  For lockMode LM_TryLock, a return value < 0 and errno == EAGAIN means
//  it could not get the lock w/o blocking.
//-----------------------------------------------------------------------------
int pvLockVal(int id, int lockMode, int lockType)
{
  PV_Desc  *pv;
  PV_Ref far  *pvRef;


  if ((uint)id >= numPVs)  {
    if ((id >= 900) AND (id <= 902))  return 0;  //temp
    errno = EINVAL;  return -1;
  }

  pvRef = pvList + (uint)id;

//--- consider ModBus register-defined values as always locked ---
  if (pvRef->pvType != PVT_EXT)  return 0;

  pv = (PV_Desc *)pvRef->pvDef;

//--- All non-reg PV's MUST provide a lock func! ---
  if (!pv->lockVal)  return -1;

  return (*pv->lockVal)(pv, lockMode, lockType);
}


//-----------------------------------------------------------------------------
//  Called to get a copy of a PV's current value.
//
//  The job of this function is to cause the specified copyFunc() to be called
//  as needed to copy all the bytes for the PV's current value.
//
//  If the PV's description does NOT specify a value-specific read function,
//  then the current PV value is assumed to be stored in its entirety at the
//  address indicated by the PV's description and all we have to do is call
//  the copyFunc() once with the location of the value and its size in bytes.
//
//  If the PV's description DOES specify a value-specific read function, then
//  that function will be called and IT will be responsible for generating
//  the current PV value in a suitable format and calling the copyFunc() as
//  needed to transfer the resulting data.  Note that the copyFunc() may be
//  called many times before the entire value is transferred.
//
//-----------------------------------------------------------------------------
int pvCopyVal(int id, uint numElems, void *conInfo, PV_CopyFunc copyFunc)
{
  uint  fromSize;
  RegInfo  *reg;
  PV_Desc  *pv;
  PV_Info  *pvInfo;
  PV_Ref far  *pvRef;
  void far  *fromVal;


  if ((uint)id >= numPVs)  {
    if ((id >= 900) AND (id <= 902))  return 0;  //temp
    errno = EINVAL;  return -1;
  }


  pvRef = pvList + (uint)id;

  switch (pvRef->pvType)  {

    case PVT_REG:
      reg = (RegInfo *)pvRef->pvDef;
      fromSize = reg->numRegs << 1;
//      fromVal = (event == DBE_VALUE) ? reg->lastChgVal : reg->regVal;
      fromVal = reg->lastChgVal;
      break;

    case PVT_EXT:
      pv = (PV_Desc *)pvRef->pvDef;
      pvInfo = &pv->pvInfo;
      fromSize = numElems * pvInfo->elemSize;
      //finish:  Support for values that don't set pv->val...
      fromVal = pv->val;  if (fromVal)  break;
      if (!pv->sendVal)  return -1;
      return (*pv->sendVal)(id, numElems, conInfo, copyFunc);

    default:
      return -1;
  }

  return (*copyFunc)(conInfo, fromVal, fromSize);
}


//-----------------------------------------------------------------------------
//  Return the value for the specified event on a PV in the requested format.
//
//  Returns 0 on success, < 0 if an error
//-----------------------------------------------------------------------------
int pvGetVal(int id, uint toType, uint toSize, void *toVal, uint event)
{
  uchar  fromType;
  uint  fromSize;
  RegInfo  *reg;
  PV_Desc  *pv;
//  PV_Info  *pvInfo;
  PV_Ref far  *pvRef;
  void  *fromVal;
  char  *testVal;

#define  FROM_VAL(type)  ( *((type *)fromVal) )
#define  TO_VAL(type)    ( *((type *)toVal) )


//temp
  if (id >= 900)  {
    if (toType != DT_STR)  return -1;

    if (id == 900)  testVal = "DEBROS";
    else  if (id == 902)  testVal = "Rabbit Ctlr";
    else  testVal = "";

    strncpy(toVal, testVal, toSize);  return 0;
  }


  if ((uint)id >= numPVs)  return -1;

  pvRef = pvList + (uint)id;

  switch (pvRef->pvType)  {

    case PVT_REG:
      reg = (RegInfo *)pvRef->pvDef;
      fromType = reg->dataType;
      fromSize = reg->numRegs << 1;
      fromVal = (event == DBE_VALUE) ? reg->lastChgVal : reg->regVal;
      break;

    case PVT_EXT:
      pv = (PV_Desc *)pvRef->pvDef;
      fromType = pv->pvInfo.dataType;
      fromSize = pv->pvInfo.size;
      if (((uint)pv->val) >= 0xE000)  return -1;
      fromVal = (void *)((uint)pv->val);
      break;

    default:
      return -1;
  }


//--- if types are the same, then just copy the value and return ---
  if ((toType == fromType) AND (toSize >= fromSize))  {
    memcpy(toVal, fromVal, fromSize);  return 0; }

//--- otherwise convert as best we can ---
  switch (toType)  {

    case DT_S16:
      if (toSize < 2)  return -1;
      switch (fromType)  {
        case DT_S16:  TO_VAL(S16) = (S16) FROM_VAL(S16);  break;
        case DT_U16:  TO_VAL(S16) = (S16) FROM_VAL(U16);  break;
        case DT_S32:  TO_VAL(S16) = (S16) FROM_VAL(S32);  break;
        case DT_U32:  TO_VAL(S16) = (S16) FROM_VAL(U32);  break;
        case DT_F32:  TO_VAL(S16) = (S16) FROM_VAL(F32);  break;
        default:  return -1;
      }
      return 0;

    case DT_U16:
      if (toSize < 2)  return -1;
      switch (fromType)  {
        case DT_S16:  TO_VAL(U16) = (U16) FROM_VAL(S16);  break;
        case DT_U16:  TO_VAL(U16) = (U16) FROM_VAL(U16);  break;
        case DT_S32:  TO_VAL(U16) = (U16) FROM_VAL(S32);  break;
        case DT_U32:  TO_VAL(U16) = (U16) FROM_VAL(U32);  break;
        case DT_F32:  TO_VAL(U16) = (U16) FROM_VAL(F32);  break;
        default:  return -1;
      }
      return 0;

    case DT_S32:
      if (toSize < 4)  return -1;
      switch (fromType)  {
        case DT_S16:  TO_VAL(S32) = (S32) FROM_VAL(S16);  break;
        case DT_U16:  TO_VAL(S32) = (S32) FROM_VAL(U16);  break;
        case DT_S32:  TO_VAL(S32) = (S32) FROM_VAL(S32);  break;
        case DT_U32:  TO_VAL(S32) = (S32) FROM_VAL(U32);  break;
        case DT_F32:  TO_VAL(S32) = (S32) FROM_VAL(F32);  break;
        default:  return -1;
      }
      return 0;

    case DT_U32:
      if (toSize < 4)  return -1;
      switch (fromType)  {
        case DT_S16:  TO_VAL(U32) = (U32) FROM_VAL(S16);  break;
        case DT_U16:  TO_VAL(U32) = (U32) FROM_VAL(U16);  break;
        case DT_S32:  TO_VAL(U32) = (U32) FROM_VAL(S32);  break;
        case DT_U32:  TO_VAL(U32) = (U32) FROM_VAL(U32);  break;
        case DT_F32:  TO_VAL(U32) = (U32) FROM_VAL(F32);  break;
        default:  return -1;
      }
      return 0;

    case DT_F32:
      if (toSize < 4)  return -1;
      switch (fromType)  {
        case DT_S16:  TO_VAL(F32) = (F32) FROM_VAL(S16);  break;
        case DT_U16:  TO_VAL(F32) = (F32) FROM_VAL(U16);  break;
        case DT_S32:  TO_VAL(F32) = (F32) FROM_VAL(S32);  break;
        case DT_U32:  TO_VAL(F32) = (F32) FROM_VAL(U32);  break;
        case DT_F32:  TO_VAL(F32) = (F32) FROM_VAL(F32);  break;
        default:  return -1;
      }
      return 0;

    case DT_STR:
      if (toSize < 12)  return -1;
      fmtVal(toVal, fromType, 0, fromVal);
      return 0;
  }

  return -1;
}


//-----------------------------------------------------------------------------
//  Update the value of a PV using the provided value.
//
//  Returns 0 on success, < 0 if an error
//-----------------------------------------------------------------------------
int pvSetVal(int id, uint fromType, uint fromSize, void *fromVal)
{
  uchar  toType;
  uint  u16;
  int  s16;
  ulong  u32;
  long  s32;
  float  f32;
  PV_Desc  *pv;
  RegInfo  *reg;
  PV_Ref far  *pvRef;
//  uint  regSize;
//  void  *regVal;
//  char  temp[32];  //tdebug

#define  FROM_VAL(type)   ( *((type *)fromVal) )


  if ((uint)id >= numPVs)  return -1;

  pvRef = pvList + (uint)id;

  switch (pvRef->pvType)  {

    case PVT_REG:
      reg = (RegInfo *)pvRef->pvDef;
      toType = reg->dataType;
//      toSize = reg->numRegs << 1;
//      toVal = reg->regVal;
      break;

    case PVT_EXT:
      pv = (PV_Desc *)pvRef->pvDef;
      toType = pv->pvInfo.dataType;
//      toSize = pv->pvInfo.size;
//      toVal = pv->val;
      return -1;  //finish
      break;

    default:
      return -1;
  }


//printf("pvSetVal(%d, %u, %u, [%02x %02x %02x %02x])\r\n",  //tdebug
//       id, fromType, fromSize,
//       *(char *)fromVal[0], *(char *)fromVal[1],
//       *(char *)fromVal[2], *(char *)fromVal[3] );


//--- if types are the same, then just copy the value and return ---
//  if ((fromType == toType) AND (fromSize >= toSize))  {
//    memcpy(toVal, fromVal, toSize);  return 0; }
  // This skips the set function specified for the register!!!!!

//--- otherwise convert as best we can ---
  switch (fromType)  {

    case DT_S16:
      if (fromSize < 2)  return -1;
      s16 = FROM_VAL(S16);
//      printf("DT_S16: [%d]\r\n", s16);  //tdebug
      switch (toType)  {
        case DT_S16:  setRegS16(reg, 0, (S16) s16);  break;
        case DT_U16:  setRegU16(reg, 0, (U16) s16);  break;
        case DT_S32:  setRegS32(reg, 0, (S32) s16);  break;
        case DT_U32:  setRegU32(reg, 0, (U32) s16);  break;
        case DT_F32:  setRegF32(reg, 0, (F32) s16);  break;
        default:  return -1;
      }
      return 0;

    case DT_U16:
      if (fromSize < 2)  return -1;
      u16 = FROM_VAL(U16);
//      printf("DT_U16: [%u]\r\n", u16);  //tdebug
      switch (toType)  {
        case DT_S16:  setRegS16(reg, 0, (S16) u16);  break;
        case DT_U16:  setRegU16(reg, 0, (U16) u16);  break;
        case DT_S32:  setRegS32(reg, 0, (S32) u16);  break;
        case DT_U32:  setRegU32(reg, 0, (U32) u16);  break;
        case DT_F32:  setRegF32(reg, 0, (F32) u16);  break;
        default:  return -1;
      }
      return 0;

    case DT_S32:
      if (fromSize < 4)  return -1;
      s32 = FROM_VAL(S32);
//      printf("DT_S32: [%ld]\r\n", s32);  //tdebug
      switch (toType)  {
        case DT_S16:  setRegS16(reg, 0, (S16) s32);  break;
        case DT_U16:  setRegU16(reg, 0, (U16) s32);  break;
        case DT_S32:  setRegS32(reg, 0, (S32) s32);  break;
        case DT_U32:  setRegU32(reg, 0, (U32) s32);  break;
        case DT_F32:  setRegF32(reg, 0, (F32) s32);  break;
        default:  return -1;
      }
      return 0;

    case DT_U32:
      if (fromSize < 4)  return -1;
      u32 = FROM_VAL(U32);
//      printf("DT_U32: [%lu]\r\n", u32);  //tdebug
      switch (toType)  {
        case DT_S16:  setRegS16(reg, 0, (S16) u32);  break;
        case DT_U16:  setRegU16(reg, 0, (U16) u32);  break;
        case DT_S32:  setRegS32(reg, 0, (S32) u32);  break;
        case DT_U32:  setRegU32(reg, 0, (U32) u32);  break;
        case DT_F32:  setRegF32(reg, 0, (F32) u32);  break;
        default:  return -1;
      }
      return 0;

    case DT_F32:
      if (fromSize < 4)  return -1;
      f32 = FROM_VAL(F32);
//      printf("DT_F32: [%.32s]\r\n", Engineering(f32, 6, temp));  //tdebug
      switch (toType)  {
        case DT_S16:  setRegS16(reg, 0, (S16) f32);  break;
        case DT_U16:  setRegU16(reg, 0, (U16) f32);  break;
        case DT_S32:  setRegS32(reg, 0, (S32) f32);  break;
        case DT_U32:  setRegU32(reg, 0, (U32) f32);  break;
        case DT_F32:  setRegF32(reg, 0, (F32) f32);  break;
        default:  return -1;
      }
      return 0;

    case DT_STR:
//      printf("DT_STR: [%.32s]\r\n", fromVal);  //tdebug
      switch (toType)  {
        case DT_S16:  setRegS16(reg, 0, (S16) ATOI(fromVal));   break;
        case DT_U16:  setRegU16(reg, 0, (U16) ATOU(fromVal));   break;
        case DT_S32:  setRegS32(reg, 0, (S32) ATOL(fromVal));   break;
        case DT_U32:  setRegU32(reg, 0, (U32) ATOUL(fromVal));  break;
        case DT_F32:  setRegF32(reg, 0, (F32) atof(fromVal));   break;
        default:  return -1;
      }
      return 0;
  }

  return -1;
}

//-----------------------------------------------------------------------------
//  Test all the PV values in the specified list to see which (if any) of the
//  events we are looking for have occured since we last checked.
//
//  For each event that occured, the corresponding bit in the PV's
//  eventsOccured bitmask is set.
//
//  This function clears the checkEvents bitmap for each PV after it
//  checks for the specified events.  The application must call pvCheckEvents()
//  for each PV to see which events happened and to re-assert the bits of
//  interest in the checkEvents bitmap.
//
//  This function returns a count of the number of events that occured.
//-----------------------------------------------------------------------------
int regTestEvents(RegInfo **regs, bool anyChg)
{
  bool  occured;
  int  numEvents = 0;
  uint  u;
  S16  diffS16;
  S32  diffS32;
  F32  diffF32;
  RegInfo  *reg, *lastReg;

#define S16_(val)   (*(S16 *)reg->val)
#define U16_(val)   (*(U16 *)reg->val)
#define S32_(val)   (*(S32 *)reg->val)
#define U32_(val)   (*(U32 *)reg->val)
#define F32_(val)   (*(F32 *)reg->val)


  lastReg = NULL;
  for (u=0,reg=regs[0]; reg; ++u,reg=regs[u])  {
    if (reg == lastReg)  continue;  // skip if not 1st in multi-reg value

    KLOCK();

    reg->eventsOccured = 0;

    if (!reg->checkEvents)  {  KUNLOCK();  continue; }


    if (reg->checkEvents & DBE_VALUE)  {

      if (anyChg)
        occured = memcmp(reg->lastChgVal, reg->regVal, reg->numRegs << 1);
      else  {
        occured = false;

        switch (reg->dataType)  {
          case DT_S16:
          case DT_U16:
            diffS16 = S16_(regVal) - S16_(lastChgVal);
            if (abs(diffS16) >= S16_(minChg))  occured = true;
            break;

          case DT_S32:
          case DT_U32:
            diffS32 = S32_(regVal) - S32_(lastChgVal);
            if (labs(diffS32) >= S32_(minChg))  occured = true;
            break;

          case DT_F32:
            diffF32 = F32_(regVal) - F32_(lastChgVal);
            if (fabs(diffF32) >= F32_(minChg))  occured = true;
            break;
        }

      }
      if (occured)  {
        memcpy(reg->lastChgVal, reg->regVal, reg->numRegs << 1);
        reg->eventsOccured |= DBE_VALUE;  ++numEvents;
      }

    }

    lastReg = reg;  reg->checkEvents = 0;

    KUNLOCK();
  }

  return numEvents;
}

//-----------------------------------------------------------------------------
//  Check all the registers with subscriptions for a change in value since
//  the last time this function was called.
//
//  If anyChg is true, then any change at all in a PV value will cause the
//  value change bit to be changed in the eventsOccured mask for the PV.
//
//  Otherwise, it uses the minChg value for each PV as the basis for whether
//  or not the value change bit should set for the PV.
//-----------------------------------------------------------------------------
void pvCheckAllEvents(time_t *dateTime, uint *ms, bool anyChg)
{
  uint  u, ttlExtBytes;
  PV_Desc  *pv;
  static int  nextExtPV = 0;


//--- update eventsOccured mask for PVs defined as ModBus registers ---
  regTestEvents(r3Regs, anyChg);
  regTestEvents(r4Regs, anyChg);

//--- update eventsOccured mask for all other PVs ---
//NOTE:  nonReg values are responsible for updating their own newEvents mask
  for (pv=nonRegPVs; pv->val; ++pv)  pv->eventsOccured = 0;

  ttlExtBytes = 0;
  for (u=0; u<numNonRegPVs; ++u)  {
    pv = nonRegPVs + nextExtPV;
    if (pv->updateVal)  (*pv->updateVal)(pv);

    _EnterCritical_();
    pv->eventsOccured = pv->newEvents;  pv->newEvents = 0;
    _ExitCritical_();

    if (++nextExtPV == numNonRegPVs)  nextExtPV = 0;

    // limit # of events we process in one pass based on ttl # of bytes
    if (pv->eventsOccured)  {
      ttlExtBytes += pv->pvInfo.size;
      if (ttlExtBytes > 1000)  break;
    }
  }

  _EnterCritical_();  *dateTime = TIME();  *ms = milliTicks;  _ExitCritical_();
}


//-----------------------------------------------------------------------------
//  Check a PV for the specified events and update the mask that determines
//  which events to check for next time pvCheckAllEvents() is called.
//
//  Returns value:
//    If no events occured: 0
//    If events occured:  bitmask of the events that occured
//    If an error: < 0
//-----------------------------------------------------------------------------
int pvCheckEvents(int id, uint checkEvents)
{
  RegInfo *reg;
  PV_Ref far  *pvRef;
  PV_Desc  *pv;


  if ((uint)id >= numPVs)  return 0;  //temp - so no error for temp "fake" PVs
//  if ((uint)id >= numPVs)  return -1;  //finish

  pvRef = pvList + (uint)id;

  switch (pvRef->pvType)  {

    case PVT_REG:
      reg = (RegInfo *)pvRef->pvDef;
      reg->checkEvents |= checkEvents;
      return (int) (reg->eventsOccured & checkEvents);

    case PVT_EXT:
      pv = (PV_Desc *)pvRef->pvDef;
      return (int) (pv->eventsOccured & checkEvents);
  }

  return -1;
}



//-----------------------------------------------------------------------------
//  Process a ModBus command
//-----------------------------------------------------------------------------
int ModBusCmd(uchar clientType, char *cmdBuf,
              char *respBuf, int *respLen, fd_t eventsfd)
{
  int  i, cmdLen, firstReg, numRegs, funcCode, regNum, regVal,
       respBytes, coilNum, firstCoil, numCoils, chanNum, chNum, ttlLen;
//  int  transId, slaveAddr, byteCount, protocolId;
  int  idx, offset, wordIdx;
  uint  cliOffset;
  bool  cmdOK, accepted;
  uchar  coilVal;
  RegInfo  *reg;
  CoilInfo  *coil;
  uchar  *bufPtr;
  ClientOrder  *cliOrder;
  U16  logicalVal[8];
  char  temp[32];



  *respLen = respBytes = 0;

//  transId = GetShort_(cmdBuf, 0);
//  protocolId = GetShort_(cmdBuf, 2);
  cmdLen = GetShort_(cmdBuf, 4);
//  slaveAddr = cmdBuf[6];
  funcCode = cmdBuf[7];

  ttlLen = cmdLen + 6;  // cmdLen value bound's checked by caller
  memcpy(respBuf, cmdBuf, ttlLen);

// Note that "respBytes" should be set to the # bytes following the func code
// in the respBuf (the func code is at offset 7)

//  fprintf(DiagLog, "Got ModBus cmd func: %d\r\n", funcCode);  //tdebug

  cmdOK = false;
  switch (funcCode)  {

//=================================================
// === read coils (single bits) (0xxxx series) ===
//=================================================
    case 1:
      if (cmdLen < 6)  break;

      firstCoil = GetShort_(cmdBuf, 8);  numCoils = GetShort_(cmdBuf, 10);
      respBytes = (numCoils + 7) / 8;

      if (ShowReads)
        fprintf(DiagLog, "Read coils: %05d - %05d\r\n",
                          firstCoil, firstCoil + numCoils - 1);

      chanNum = firstCoil / C0ChanMult;  firstCoil = firstCoil % C0ChanMult;

      if (respBytes > MaxRespMsgLen-10)  {
        fprintf(DiagLog, "Response too long\r\n");
        respBytes = 1;
        respBuf[7] |= 0x80;  // top bit on indicates exception
        respBuf[8] = 2;
        cmdOK = true;  break;
      }

      respBuf[8] = respBytes;

      bufPtr = respBuf + 9;
      memset(bufPtr, 0, respBytes);

    //--- note that range may cross chanNum boundaries ---
      chanNum = firstCoil / C0ChanMult;  coilNum = firstCoil % C0ChanMult;
      KLOCK();
      for (i=0; i<numCoils; ++i)  {
        coilVal = 0;
        for (;;)  {
          if (chanNum >= NumC0Chans)  break;
          if (coilNum >= numC0Coils)  break;
          coil = c0Coils[coilNum];
          chNum = ((uint)coil->flags & OnePerChan ? chanNum : 0);
          coilVal = coil->coilVal[chNum];
          break;
        }
        if (coilVal)  SetBit(bufPtr, i);
        if (++coilNum == C0ChanMult)  { coilNum = 0;  ++chanNum; }
      }
      KUNLOCK();
      ++respBytes;

      cmdOK = true;  break;


//=====================================================
// === read multiple R/W registers (4xxxx series) ===
//=====================================================
    case 3:
      if (cmdLen < 6)  break;

      firstReg = GetShort_(cmdBuf, 8);  numRegs = GetShort_(cmdBuf, 10);

      if (ShowReads)
        fprintf(DiagLog, "Read Regs: 4%04d - 4%04d\r\n",
                          firstReg, firstReg + numRegs - 1);

      respBytes = numRegs * 2;

      if (respBytes > MaxRespMsgLen-10)  {
        fprintf(DiagLog, "Response too long\r\n");
        respBytes = 1;
        respBuf[7] |= 0x80;  // top bit on indicates exception
        respBuf[8] = 2;
        cmdOK = true;  break;
      }

      respBuf[8] = respBytes;  ++respBytes;
      memset(respBuf + 9, 0, respBytes);

    //--- note that range of registers may cross chanNum boundaries ---
      chanNum = firstReg / R4ChanMult;  regNum = firstReg % R4ChanMult;
      offset = 9;  cliOrder = (ClientOrder *)(&regOrder[clientType][0][0]);
      KLOCK();
      for (i=0; i<numRegs; ++i)  {
        regVal = 0;
        for (;;)  {
          if (chanNum >= NumR4Chans)  break;
          if (regNum >= numR4Regs)  break;
          reg = r4Regs[regNum];
          chNum = ((uint)reg->flags & OnePerChan ? chanNum : 0);
          wordIdx = regNum - reg->regNum;  // which word in multi-word value
          idx = wordIdx + (chNum * reg->numRegs);  // offset by channel
          idx += (*cliOrder)[reg->dataType][wordIdx]; // reorder by client type
          regVal = reg->regVal[idx];
          break;
        }
        SetWordVal(offset, regVal, respBuf);  offset += 2;
        if (++regNum == R4ChanMult)  { regNum = 0;  ++chanNum; }
      }
      KUNLOCK();

      cmdOK = true;  break;


//======================================================
// === read multiple input registers (3xxxx series) ===
//======================================================
    case 4:
      if (cmdLen < 6)  break;

      firstReg = GetShort_(cmdBuf, 8);  numRegs = GetShort_(cmdBuf, 10);

      if (ShowReads)
        fprintf(DiagLog, "Read Regs: 3%04d - 3%04d\r\n",
                          firstReg, firstReg + numRegs - 1);

      respBytes = numRegs * 2;

      if (respBytes > MaxRespMsgLen-10)  {
        fprintf(DiagLog, "Response too long\r\n");
        respBytes = 1;
        respBuf[7] |= 0x80;  // top bit on indicates exception
        respBuf[8] = 2;
        cmdOK = true;  break;
      }

      respBuf[8] = respBytes;  ++respBytes;
      memset(respBuf+9, 0, respBytes);

    //--- note that range of registers may cross chanNum boundaries ---
      chanNum = firstReg / R3ChanMult;  regNum = firstReg % R3ChanMult;
      offset = 9;  cliOrder = (ClientOrder *)(&regOrder[clientType][0][0]);
      KLOCK();
      for (i=0; i<numRegs; ++i)  {
        if ((chanNum >= NumR3Chans) OR (regNum >= numR3Regs))
          regVal = 0;
        else  {
          reg = r3Regs[regNum];  // info about registers for this value
          chNum = ((uint)reg->flags & OnePerChan ? chanNum : 0);
          wordIdx = regNum - reg->regNum;  // which word in multi-word value
          idx = wordIdx + (chNum * reg->numRegs);  // offset by channel
          idx += (*cliOrder)[reg->dataType][wordIdx]; // reorder by client type
          regVal = reg->regVal[idx];
        }
        SetWordVal(offset, regVal, respBuf);  offset += 2;
        if (++regNum == R3ChanMult)  { regNum = 0;  ++chanNum; }
      }
      KUNLOCK();

      cmdOK = true;  break;


//======================================================
// === write coil (single bit value) (0xxxx series) ===
//======================================================
    case 5:
      if (cmdLen < 6)  break;

      coilNum = GetShort_(cmdBuf, 8);  coilVal = cmdBuf[10];
      chanNum = coilNum / C0ChanMult;  coilNum = coilNum % C0ChanMult;
      if ((chanNum >= NumC0Chans) OR (coilNum >= numC0Coils))  {
        if (ShowErrors)
          fprintf(DiagLog, "set coil %05d[%d] to %d  *IGNORED*\n",
                  coilNum+1, chanNum, coilVal );
      }
      else  {
        coil = c0Coils[coilNum];
        chNum = ((uint)coil->flags & OnePerChan ? chanNum : 0);
        if ((coil->flags & LogChanges) AND startupComplete)
          fprintf(DiagLog, "set coil %05d[%d] (%ls): %d\n",
                  coil->coilNum+1, chanNum, coil->name, coilVal );
        (coil->setCoilFunc)(coil, chNum, coilVal, eventsfd);
      }
      respBytes = 4;
      cmdOK = true;  break;


//==============================================
// === write single register (4xxxx series) ===
//==============================================
    case 6:
      if (cmdLen < 6)  break;

      regNum = GetShort_(cmdBuf, 8);  regVal = GetShort_(cmdBuf, 10);
      respBytes = 4;

      chanNum = regNum / R4ChanMult;  regNum = regNum % R4ChanMult;

      if ((chanNum >= NumR4Chans) OR (regNum > numR4Regs))  {
        if (ShowErrors)
          fprintf(DiagLog, "set reg 4%04d[%d] to 0x%04X %6d  *IGNORED*\n",
                  regNum+1, chanNum, regVal, regVal );
      }
      else  {
        reg = r4Regs[regNum];
        chNum = ((uint)reg->flags & OnePerChan ? chanNum : 0);

      //--- upate appropriate portion in multi-register value buffer ---
        wordIdx = regNum - reg->regNum;  // which word in multi-word value
        cliOffset = regOrder[clientType][(int)reg->dataType][wordIdx];
        logicalVal[wordIdx + cliOffset] = regVal;

        if (ShowWrites)
          fprintf(DiagLog, "set reg 4%04d[%d] (%ls.%d): 0x%04X %6d\n",
                  regNum+1, chanNum, reg->name, wordIdx, regVal, regVal );

        if (reg->wrMask == REG_RO)  {
          fprintf(DiagLog, "write to RO reg %ls[%d] = ignored\r\n\n",
                  reg->name, chNum);
          cmdOK = true;  break;
        }

      //--- if last reg in multi-reg value, update the value ---
        if (wordIdx == (reg->numRegs - 1))  if (reg->setRegFunc)  {
          accepted = (reg->setRegFunc)(reg, chNum, logicalVal, eventsfd);
          if ((reg->flags & LogChanges) AND startupComplete)
            fprintf(DiagLog, "set %ls[%d] = %s: %s\r\n\n",
                    reg->name, chNum,
                    fmtRegVal(reg, (S16 *)logicalVal, chNum, temp),
                    (accepted ? "OK" : "Rejected")  );
        }
      }

      cmdOK = true;  break;


//=================================================
// === write multiple registers (4xxxx series) ===
//=================================================
    case 16:
      if (cmdLen < 6)  break;

      firstReg = GetShort_(cmdBuf, 8);  numRegs = GetShort_(cmdBuf, 10);
//      byteCount = cmdBuf[12];

     //--- note that range of registers may cross chanNum boundaries ---
      chanNum = firstReg / R4ChanMult;  regNum = firstReg % R4ChanMult;
      offset = 13;  cliOrder = (ClientOrder *)(&regOrder[clientType][0][0]);

      for (i=0; i<numRegs; ++i)  {
        if ((chanNum >= NumR4Chans) OR (regNum > numR4Regs))  {
          if (ShowErrors)
            fprintf(DiagLog, "set reg 4%04d[%d] to 0x%04X %6d  *IGNORED*\n",
                    regNum+1, chanNum, regVal, regVal );
        }
        else  {
          reg = r4Regs[regNum];
          chNum = ((uint)reg->flags & OnePerChan ? chanNum : 0);

        //--- upate appropriate portion in multi-register value buffer ---
          regVal = GetShort_(cmdBuf, offset);
          wordIdx = regNum - reg->regNum;  // which word in multi-word value
          cliOffset = (*cliOrder)[(int)reg->dataType][wordIdx];
          logicalVal[wordIdx + cliOffset] = regVal;

          if (ShowWrites)
            fprintf(DiagLog, "set reg 4%04d[%d] (%ls.%d): 0x%04X %6d\n",
                    regNum+1, chNum, reg->name, wordIdx, regVal, regVal );

          if (reg->wrMask == REG_RO)  {
            fprintf(DiagLog, "write to RO reg %ls[%d] = ignored\r\n\n",
                    reg->name, chNum);
            continue;
          }

        //--- if last reg in multi-reg value, update the value ---
          if (wordIdx == (reg->numRegs - 1))  if (reg->setRegFunc)  {
            accepted = (reg->setRegFunc)(reg, chNum, logicalVal, eventsfd);
            if ((reg->flags & LogChanges) AND startupComplete)
              fprintf(DiagLog, "set %ls[%d] = %s: %s\r\n\n",
                      reg->name, chNum,
                      fmtRegVal(reg, (S16 *)logicalVal, chNum, temp),
                      (accepted ? "OK" : "Rejected")  );
          }
        }

        offset += 2;
        if (++regNum == R4ChanMult)  { regNum = 0;  ++chanNum; }
      }

      respBytes = 4;

      cmdOK = true;  break;


    default:
      if (ShowErrors)
        fprintf(DiagLog,
          "Unrecognized or unsupported command: func code: %d\r\n", funcCode);
      respBytes = 1;
      respBuf[7] |= 0x80;  // top bit on indicates exception
      respBuf[8] = 1;  cmdOK = true;
  }


  if (! cmdOK)  {
    if (ShowErrors)  fprintf(DiagLog, "Badly formed command\r\n");
    respBytes = 1;
    respBuf[7] |= 0x80;  // top bit on indicates exception
    respBuf[8] = 3;
  }

  SetWordVal(4, respBytes+2, respBuf);  *respLen = respBytes + 8;
/*
  if (showDetails)  {
    sprintf(msg, " %3d", respBytes+8);  DispStr(75,y, msg, false);  }
*/

  return 0;
}


//-----------------------------------------------------------------------------
void showReg4Info(uint idx)
{
  uchar  *lcall;
  ulong  absAddr;
  RegInfo  *reg;
  char  name[64];


  reg = r4Regs[idx];  lcall = (uchar *)(reg->setRegFunc);

  absAddr = ((ulong)lcall[5] << 12) + ((ulong)lcall[4]<<8) + (ulong)lcall[3];

  strcpyfn(name, reg->name);
  printf("r4Regs[%02d]: 4%04d+%d %-16s [%04X]->[%04X]\r\n",
          idx, reg->regNum, idx-reg->regNum, name,
          (uint)lcall, (uint)absAddr );
}


//-----------------------------------------------------------------------------
//  Scan a list of registers to make sure they are contiguous (no overlaps or
//  breaks in the list) and update the global numPVs value.
//
//  Returns 0 if no problems found, or < 0 if an error.
//-----------------------------------------------------------------------------
int scanRegList(RegInfo **regs, uint *numRegs, char prefix)
{
  uint  u;
  RegInfo  *reg, *lastReg = NULL;


  for (u=0,reg=*regs; reg; ++u,reg=*(++regs))  {
//    printf("reg: [%u] %c%04d: %ls\r\n",  //tdebug
//           u, prefix, reg->regNum, reg->name);

    if (reg != lastReg)  ++numPVs;  // count only 1st reg in multi-reg values

    if ((u < reg->regNum) OR (u >= reg->regNum + reg->numRegs))  {
      fprintf(DiagLog,
          "***** modbus REG%c table error: %c%04u at pos for %c%04u *****\n",
              prefix, prefix, reg->regNum+1, prefix, u+1);
      return -1;
    }

    lastReg = reg;
  }

  *numRegs = u;

  return 0;
}

//-----------------------------------------------------------------------------
//  Set the "deadband" value for each PV in a group of registers and add them
//  to the global list of all PVs.
//-----------------------------------------------------------------------------
PV_Ref far * setDeadbandDefaults(RegInfo **regs,
                                    PV_Ref far *pvListPtr)
{
  RegInfo  *reg, *lastReg = NULL;


  for (reg=*regs; reg; ++regs,reg=*regs)  {
    if (reg == lastReg)  continue;  // skip if not 1st in multi-reg value
    switch (reg->dataType)  {
      case DT_S16:  case DT_U16:  *(int *)reg->minChg = 1;  break;
      case DT_S32:  case DT_U32:  *(long *)reg->minChg = 1;  break;
      case DT_F32:  *(float *)reg->minChg = 0.0001;  break;
    }
    lastReg = reg;
    pvListPtr->pvDef = reg;  pvListPtr->pvType = PVT_REG;  ++pvListPtr;
  }

  return pvListPtr;
}


//-----------------------------------------------------------------------------
//  Initialization for the lists of PVs (Process Variables).  Includes scanning
//  the tables of values accessable via the ModBus interface to determine the
//  # of logical values represented by the registers and to make sure they are
//  consistent (i.e. no holes or overlaps in the sequence).
//
//  Also sets the defaults for minChg value used to test for value change
//  events for the EPICS CA (Channel Access) interface.
//-----------------------------------------------------------------------------
int pvInitLists(void)
{
  uint  u, last;
  CoilInfo  *coil, **coils;
  PV_Ref far  *pvPtr;
  PV_Ref  hold;
  char  name0[MAX_PV_NAME_LEN], name1[MAX_PV_NAME_LEN];


  numC0Coils = numR3Regs = numR4Regs = numNonRegPVs = numPVs = 0;

  printf("\nInitialization for ModBus and PV Interfaces...\r\n");


  coils = c0Coils;
  for (u=0,coil=*coils; coil; ++u,++coils,coil=*coils)  {
//    printf("coil: [%u][0x%04X] 0%04u [%ls]\r\n", //tdebug
//           u, (uint)coil, coil->coilNum, coil->name);  //tdebug
    if (coil->coilNum != u)  {
      printf(
         "\n***** modbus COIL table error: 0%04u at pos for 0%04u *****\n\n",
              coil->coilNum+1, u+1);
      return -1;
    }
  }
  numC0Coils = u;

//--- scan the ModBus register lists for errors and count total # of PVs ---
  if (scanRegList(r3Regs, &numR3Regs, '3'))  return -2;
  if (scanRegList(r4Regs, &numR4Regs, '4'))  return -3;
  numRegPVs = numPVs;
  for (u=0; nonRegPVs[u].val; ++u)  ++numNonRegPVs;
  numPVs += numNonRegPVs;

  printf("PV counts:\n"
         "  C0Coils: %d\n"
         "  R3Regs: %d\n"
         "  R4Regs: %d\n"
         "  NonReg PVs: %d\n",
         numC0Coils, numR3Regs, numR4Regs, numNonRegPVs);

//--- allocate an array of pointers for the PVs ---
  printf("\nFor sorted list of %u PVs:\n", numPVs);
  if (!(pvList = (PV_Ref far *)xalloc(numPVs * sizeof(PV_Ref))))  return -4;

//--- initialize the list of pointers (and some values for the ModBus regs) ---
  pvPtr = pvList;
  pvPtr = setDeadbandDefaults(r3Regs, pvPtr);
  pvPtr = setDeadbandDefaults(r4Regs, pvPtr);
  for (u=0; u<numNonRegPVs; ++u)  {
    pvPtr->pvDef = nonRegPVs+u;  pvPtr->pvType = PVT_EXT;  ++pvPtr; }

//--- let the kernel task run ---
  resetWDT();  current->state = TASK_UNINTERRUPTIBLE;  schedule();

//--- sort the list so we can use a binary search to locate a PV by name ---
  printf("Sorting PV names...\n");
  for (last=numPVs-1; last; --last)
    for (u=0,pvPtr=pvList; u<last; ++u,++pvPtr)  {
      if (stricmp(pvGetName(pvPtr, name0), pvGetName(pvPtr+1, name1)) > 0)  {
        hold = pvPtr[0];  pvPtr[0] = pvPtr[1];  pvPtr[1] = hold; }
    }

//--- let the kernel task run again ---
  resetWDT();  current->state = TASK_UNINTERRUPTIBLE;  schedule();

/*tdebug
  for (u=0; u<numPVs; ++u)  {
    pvPtr = pvList + u;
    printf("  %3u: Type:%u [%s]\n", u, pvPtr->pvType, pvGetName(pvPtr, name0));
    WDT_2S();
  }
  resetWDT();  current->state = TASK_UNINTERRUPTIBLE;  schedule();
*/

  printf("\n");

  return 0;
}


//-----------------------------------------------------------------------------
void _task modbusTCPd(int argc, char **argv)
{
  int  cmdLen, respLen;
  fd_t  eventsfd;
  uchar  cmdBuf[MaxCmdMsgLen], respBuf[MaxRespMsgLen];


  redirectStdio();  // so printf, perror, etc don't end up in data stream

  eventsfd = open(EventsFIFO, O_WRONLY, 0);

  DiagFlags |= ErrorBits;  // show errors by default

  for (;;)  {
    if (read_(0, cmdBuf, 6, 15000) < 6)  {  // get 1st 6 bytes of command
      perror("timeout reading modbusTCP cmd");  Exit(-1); }

    cmdLen = GetShort_(cmdBuf, 4);

    if (cmdLen > MaxCmdMsgLen-6)  {
      if (ShowErrors)  printf("Max cmd len exceeded: %d\n", cmdLen);
      skipBytes(0, cmdLen, 1000);  continue;
    }

    if (read_(0, cmdBuf+6, cmdLen, 15000) < cmdLen)  {
      perror("error reading cmd");  Exit(-2); }

    if (ModBusCmd(CT_MBNSCL, cmdBuf, respBuf, &respLen, eventsfd))  continue;

    if (write(1, respBuf, respLen) != respLen)  {
      perror("error writing response");  Exit(-3); }
  }

}

//-----------------------------------------------------------------------------
