//----- simLLRF2.c ----- 11/16/12 -----(09/15/11)-----

//-----------------------------------------------------------------------------
//  This module provides a bare simulation of functality that is specific to
//  the LLRF2 hardware platform.  This allows some basic testing of the
//  application logic on other platforms.
//
//  Note that the simulation is not meant to be of the hardware itself, but of
//  its behavior when used by the LLRF2 application.
//-----------------------------------------------------------------------------
//  Update log (most recent first)
//
//
//  2012-11-16 - Mark Davis
//               - Added write_FPGA_Regs()
//
//  2012-11-15 - Mark Davis
//               - Reduced timeouts for locks.
//               - Added read_FPGA_Regs()
//
//  2012-11-14 - Mark Davis
//               - Added fakeRegVal() to provide more reasonable values when
//                 running on simulated hardware.
//
//  2011-09-15 - Mark Davis
//               - Original version started.
//
//-----------------------------------------------------------------------------

#include <common.h>

#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <errno.h>

#include <sys/backup.h>

#include <sys/spi.h>
#include <sys/hdwDefs.h>

#include <busybox.h>

#include <Rabbit.h>
#include <dcdefs.h>

#include <stcpip.h>

#include <sys/socket.h>
#include <sys/srfs.h>
#include <sys/flashFile.h>
#include <sys/asflash.h>


#include <sys/modbus.h>

#include <LLRF2_Appl.h>
#include <LLRF2.h>


#include <sys/farprintf.h>


#define  RespOffset  2  // # packets DSP responses are delayed by


extern const RD_DSP_REG far rdDSPregs[];

extern const uint  Num_RD_DSP_Regs;


//-----------------------------------------------------------------------------
//  For testing purposes:  Replaces actual read of a DSP register value for
//  hdw platforms other than the actual Phase Controller.
//-----------------------------------------------------------------------------
ulong fakeRegVal(ulong regNum)
{
  uint  r;
  ulong  rcvWord;
  F32  fval, percNoise;
  const RD_DSP_REG far *dspReg;


  for (r=0,dspReg=rdDSPregs; r<Num_RD_DSP_Regs; ++r,++dspReg)  {
    if (dspReg->id != regNum)  continue;

    rcvWord = dspReg->typical;
    if (!dspReg->percNoise)  return htonl(rcvWord);

    percNoise = (F32)dspReg->percNoise * (F32)rand() / (F32)RAND_MAX / 100.0;

    switch (dspReg->type)  {

      case RT_S32:
        rcvWord = (U32)((S32)rcvWord + (S32)(percNoise * rcvWord));
        break;

      case RT_U32:
        rcvWord += (U32)(percNoise * rcvWord);
        break;

      case RT_F32:
        fval = *(F32 *)&rcvWord;
        fval += percNoise * fval;
        rcvWord = *(ulong *)&fval;
        break;

      case RT_PHASE:
        rcvWord += (U32)(percNoise * rcvWord);
        break;
    }

    return htonl(rcvWord);
  }

//  *((float *)&rcvWord) = 0.1234 + (float)(uint)(GetMS_()>>5 & 0x3F);
//  return htonl(rcvWord);

  return 0xFFFFFFFFL;
}


//=============================================================================
//  Hardware platform-specifc functions.  The simLLRF2.c module contains
//  hardware-independent versions of these that may provide some minimal
//  simulation for the purpose of doing basic testing of the application
//  logic on other platforms.
//=============================================================================

//-----------------------------------------------------------------------------
//  Get the APPL_LOCK and initialize the SPI port used to talk to the DSP
//
//  Returns lock depth on success, < 0 if an error
//-----------------------------------------------------------------------------
int getLockDSP(ulong timeOut)
{
  return getLock(APPL_LOCK, timeOut);
}

//-----------------------------------------------------------------------------
//  release the APPL_LOCK and allow other tasks to use the DSP interface
//
//  Returns 0 on success, < 0 if an error
//-----------------------------------------------------------------------------
int releaseLockDSP(void)
{
  return releaseLock(APPL_LOCK);
}


//-----------------------------------------------------------------------------
//  Reset the DSP so it (re)loads its config from the EEPROM.
//
//  Returns 0 on success, <0 if an error occurs.
//-----------------------------------------------------------------------------
int resetDSP(void)
{
  uint  startTime;


  if (getLock(APPL_LOCK, 5000) <= 0)  return -1;

  prepForReset();

  startTime = GetMS_();  while (msSince_(startTime) < 500);  // wait 500 ms

  releaseLock(APPL_LOCK);  return 0;
}



//-----------------------------------------------------------------------------
//  Place the DSP and FPGA in a reset state for the purpose of allowing the
//  Rabbit to access their respective EEPROMs via the SPI2 bus.  It is the
//  responsibility of the caller to obtain the proper lock BEFORE calling this
//  function.  Failure to do so can result in multiple master devices trying to
//  use the SPI2 bus at the same time with unpredictable results!
//-----------------------------------------------------------------------------
int stop_DSP_FPGA(void)
{
  if (NotLockOwner(APPL_LOCK))  return -1;

  printf("Placing DSP and FPGA in RESET mode ...\r\n");

  fpgaInitialized = false;

  return 0;
}

//-----------------------------------------------------------------------------
//  Reset the DSP and FPGA, causing them to load and use the configurations
//  currently stored in the EEPROMs.  To reset the DSP and FPGA, call
//  stop_DSP_FPGA() and then restart_DSP_FPGA()
//
//  Returns 0 on success, or <0 if an error
//-----------------------------------------------------------------------------
int restart_DSP_FPGA(ulong waitMS)
{
  if (NotLockOwner(APPL_LOCK))  return -1;
  if (LockDepth(APPL_LOCK) != 1)  return -1;

  printf("Restarting DSP and FPGA ...\r\n");

  printf("FPGA restarted\r\n");

  fpgaInitialized = true;

  return 0;
}


//-----------------------------------------------------------------------------
//  write a single 32-bit register value to the DSP (without any additional
//  cycles to read a response, which can interfere with any larger sequence
//  of packets this may be imbedded in).
//
//  Returns 0 on success, < 0 if an error
//-----------------------------------------------------------------------------
int writeDSPRegQuick(ulong regNum, ulong value)
{
  if (getLockDSP(1000) <= 0)  return -1;

  releaseLockDSP();

  return 0;
}


//-----------------------------------------------------------------------------
//  write a single 32-bit register value to the DSP and read back the echoed
//  addr and value (which at least lets us know we are talking to the DSP).
//
//  Returns 0 on success, < 0 if an error
//-----------------------------------------------------------------------------
int writeDSPReg(ulong regNum, ulong value, bool showErrs)
{
  ulong  xmitWord, rcvWord;


  if (getLockDSP(1000) < 0)  return -1;

  regNum |= 1L;

//--- send the register # in the 1st packet ---
  xmitWord = htonl(regNum);
  rcvWord = 0;  //simulation
  if (showFiller)  printf("  [%08lX", ntohl(rcvWord));

//--- followed by the value to be written in the 2nd packet ---
  xmitWord = htonl(value);
  rcvWord = 0;  //simulation
  if (showFiller)  printf("  %08lX]\r\n", ntohl(rcvWord));

//--- read back the echoed register # ---
  xmitWord = 0;
  rcvWord = htonl(regNum);  //simulation
  rcvWord = ntohl(rcvWord);

  if (rcvWord != regNum)  {
    if (showErrs)
      printf("  echoed write address error: 0x%08lX  echoed for 0x%08lX\r\n",
             rcvWord, regNum);
    releaseLockDSP();  return -2;
  }

//--- read back the echoed value ---
  xmitWord = 0;
  rcvWord = htonl(value);  //simulatioon
  rcvWord = ntohl(rcvWord);

  if (rcvWord != value)  {
    if (showErrs)
      printf("  echoed value error: 0x%08lX  echoed for 0x%08lX\r\n",
             rcvWord, value);
    releaseLockDSP();  return -3;
  }

  releaseLockDSP();

  if (xmitWord)  { }   //dummy


  return 0;
}


//-----------------------------------------------------------------------------
//  Read a single 32-bit register value from the DSP
//
//  Returns 0 and sets *regVal on success, < 0 if an error
//-----------------------------------------------------------------------------
int readDSPReg(ulong regNum, ulong *regVal)
{
  uint  i, wordCount;
  ulong  xmitWord, rcvWord;  //, respRegNum;
//  char  temp[32];


  if (getLockDSP(1000) < 0)  return -1;

  regNum &= 0xFFFFFFFEL;

  wordCount = 1 + RespOffset;

  xmitWord = htonl(regNum);

  for (i=0; i<wordCount; ++i,regNum+=2)  {
    rcvWord = (i < RespOffset) ? 0 : fakeRegVal(regNum-2*RespOffset);
    xmitWord = 0;
  }

  releaseLockDSP();

  rcvWord = ntohl(rcvWord);

//  respRegNum = regNum - (ulong)(RespOffset * 2);
//  printf("  0x%08LX:  0x%08LX  %12ld  %.32s\r\n",
//         respRegNum, rcvWord, rcvWord, fmtDSPval(rcvWord, respRegNum, temp));

  *regVal = rcvWord;

  if (xmitWord)  { }   //dummy

  return 0;
}


//-----------------------------------------------------------------------------
//  Write and Read a group of long integers to/from the DSP
//
//  The counts are the # of long (32-bit) words to send/rcv
//
//  Returns 0 on success, < 0 if an error
// ---------------------------------------------------------------------------
//  Exchange of data with the DSP is done using a bi-directional (simultaneous
//  send/recv) clocked SPI interface.  All data is exchanged in 32-bit (4-byte)
//  packets.  While we are sending a 32-bit packet to the DSP, it is using the
//  same clock pulses to simultaneously send a 32-bit packet in response.
//
//  Each exchange of a 32-bit packet consists of setting the DSP /Enable signal
//  low, writing (while simultaneously reading) 32-bits of data, and then
//  setting the DSP /Enable signal high.
//
//  Due to the way the DSP buffers are set up, all responses from the DSP are
//  always 2 packets "behind" the ones that contained the commands that they
//  are a response to.  We can write as many commands as we want in a
//  sequence of packets, but the response to each one will always be received
//  2 packets later.  Given that we generally want to read the response to
//  a set of commands as soon as possible, we normally have to add some
//  dummy or filler packets to the sequence to get the desired response
//  packets.  For extended responses (like reading of waveform data), we must
//  send lots of filler packets to get all the data.
//-----------------------------------------------------------------------------
int readPlot(ulong regNum, ulong far *rcvBuf, uint rcvCount)
{
  int  r;
  uint  i, wordCount;
  ulong  xmitWord, rcvWord;


  if (getLockDSP(1000) <= 0)  return -1;

  wordCount = rcvCount + RespOffset;

  xmitWord = htonl(regNum);

  for (i=0,r=-RespOffset; i<wordCount; ++i,++r)  {
    rcvWord = r;  //simulation
    if (r >= 0) rcvBuf[(uint)r] = rcvWord;
  }

  releaseLockDSP();

  if (xmitWord)  { }   //dummy

  return 0;
}

//-----------------------------------------------------------------------------
//  Read the values for the RO DSP registers and update the value of the
//  corresponding ModBus registers.
//
//  Returns 0 on success, < 0 if an error
// ---------------------------------------------------------------------------
//  See WrRdDSP() above for details on how the DSP interface works
//-----------------------------------------------------------------------------
int read_DSP_Regs(const RD_DSP_REG far *rdDSPregs)
{
  uint  i, wordCount;
  ulong  xmitWord, dspVal;
  RegInfo *mbReg;
  const RD_DSP_REG far *reqReg;
  const RD_DSP_REG far *dspReg;


  if (getLockDSP(1000) < 0)  return -1;

  wordCount = Num_RD_DSP_Regs + RespOffset;

  for (i=0,reqReg=rdDSPregs; i<wordCount; ++i,++reqReg)  {
    xmitWord = (i < Num_RD_DSP_Regs) ? htonl(reqReg->id) : 0;

    dspVal = (i < RespOffset) ? 0 : fakeRegVal((reqReg-RespOffset)->id);

    if (i < RespOffset)  continue;  // responses are offset by RespOffset pkts

    dspReg = reqReg - RespOffset;

  //--- convert value as needed and update the ModBus register value ---
    mbReg = dspReg->mbReg;  dspVal = ntohl(dspVal);
    dsp2mb(mbReg, mbReg->regVal, dspReg, &dspVal);
  }

  releaseLockDSP();

  if (xmitWord)  { }   //dummy

  return 0;
}

//-----------------------------------------------------------------------------
//  Send the current values to the DSP for each DSP R/W register that
//  represents a run-time setting (i.e. ones with writeMode == WM_EVERY).
//
//  Returns 0 on success, < 0 if an error
// ---------------------------------------------------------------------------
//  See WrRdDSP() above for details on how the DSP interface works
//-----------------------------------------------------------------------------
int write_DSP_Regs(const WR_DSP_REG far *wrDSPregs)
{
  uint  i;
  ulong  xmitWord, rcvWord, dspVal, regNum;
  const WR_DSP_REG far *dspReg;
  RegInfo  *mbReg;


  if (getLockDSP(1000) < 0)  return -2;

  for (i=0,dspReg=wrDSPregs; i<Num_WR_DSP_Regs; ++i,++dspReg)  {
    if (dspReg->writeMode != WM_EVERY)  continue;

    regNum = dspReg->id | 1L;

  //--- convert ModBus value to DSP format ---
    mbReg = dspReg->mbReg;
    mb2dsp(mbReg, mbReg->regVal, dspReg, &dspVal);

  //--- send the register # in the 1st packet ---
    xmitWord = htonl(regNum);
    rcvWord = 0;  //simulation

  //--- send the value in the 2nd packet ---
    xmitWord = htonl(dspVal);
    rcvWord = 0;  //simulation
  }

  releaseLockDSP();

  if (xmitWord)  { }   //dummy

  if (rcvWord)  { }   //dummy

  return 0;
}




//-----------------------------------------------------------------------------
//  Read an 8-bit value from the FPGA
//
//  Returns the value (>= 0) on success, < 0 if an error
//-----------------------------------------------------------------------------
int readFPGA(uint regNum)
{
  uchar  regVal;


  if (getLock(APPL_LOCK, 5000) < 0)  return -1;

  outi(PFDR, srPFDR = ((uchar)regNum << 2));  // assert 6-bit address

  regVal = ~regNum;  //simulation

  releaseLock(APPL_LOCK);

  return (int)regVal;
}


//-----------------------------------------------------------------------------
//  Read a sequence of FPGA registers.
//
//  Returns 0 on success, <0 if an error.
//-----------------------------------------------------------------------------
int read_FPGA_Regs(U16 *regVal[], uint firstReg, uint numRegs)
{
  uint  regNum, endReg;


  if (getLock(APPL_LOCK, 100) <= 0)  return -1;

  endReg = firstReg + numRegs;

  for (regNum=firstReg; regNum<endReg; ++regNum,++regVal)
    **regVal = regNum;  //simulation

  releaseLock(APPL_LOCK);

  return 0;
}


//-----------------------------------------------------------------------------
//  Write an 8-bit value to the FPGA
//
//  Returns 0 on success, -1 if an error
//-----------------------------------------------------------------------------
int writeFPGA(uint regNum, uint regVal)
{
  if (getLock(APPL_LOCK, 5000) < 0)  return -1;


  releaseLock(APPL_LOCK);

  return 0;
}

//-----------------------------------------------------------------------------
//  Check the list of FPGA register values.  For each one that is not 0,
//  write the value to the FPGA, then set it to 0 and wirte it again
//  (effectively "pulsing" any non-0 values).
//
//  Returns 0 on success, <0 if an error.
//-----------------------------------------------------------------------------
int write_FPGA_Regs(U16 *regVal[], uint firstReg, uint numRegs)
{
  uint  regNum, endReg;


  if (getLock(APPL_LOCK, 100) <= 0)  return -1;

  endReg = firstReg + numRegs;

  for (regNum=firstReg; regNum<endReg; ++regNum,++regVal)  {
    if (! **regVal)  continue;

    **regVal = 0;  // set the value back to 0
  }

  releaseLock(APPL_LOCK);

  return 0;
}


//-----------------------------------------------------------------------------
//  Update the firmware image in one of the EEPROMs.
//
//  The stopAt parameter specifies the address after which nothing in the
//  memory should be touched.  A value of 0 means the entire chip will be
//  erased, while values > 0 means nothing after the specified address will
//  be erased or written over.  The value will be rounded down to the nearest
//  multiple of 32K.
//
//  Returns 0 on success, < 0 if an error occured (with errno set)
//-----------------------------------------------------------------------------
int updateEEPROM(uint romNum, fd_t fromfd, ulong stopAt,
                 bool showProgress, bool verify)
{
#define  BufSize  256  // can't be very large without overflowing the stack!
                       // AND it MUST be an even # of bytes!

  bool  downloadOK;
  int  i, status, count;
  uint  errCount;
  ulong  ttlBytes, fileSize;
  bool  busy, eraseOK, protectOK;
  struct stat  fsInfo;
  char  cmdBuf[6];
  char  buf[BufSize], buf2[BufSize];


  if (romNum > 1)  return -1;

  if (getLock(APPL_LOCK, 10000) < 0)  return -1;

  downloadOK = false;

  stopAt &= 0xFF8000L;  // round down to nearest 32K boundary

//--- stop the FPGA and DSP so we can access the EEPROM via the SPI2 bus ---
  stop_DSP_FPGA();

//--- make sure EEPROM is ready to accept commands ---

  do {
  //--- read device ID ---
    memset(buf2, 0x5A, 4);  //simulation
    printf("  Device ID: %02X %02X %02X %02X\r\n",
           buf2[0], buf2[1], buf2[2], buf2[3] );

  //===== unprotect all sectors =====
    printf("  Unprotecting all sectors...\r\n");  eraseOK = false;

  //===== erase part or all of the selected EEPROM =====
    if (stopAt)  {
      printf("  Erasing up to 0x%06lX (%lu)", stopAt, stopAt);
      memset(cmdBuf, 0, sizeof(cmdBuf));
      for (ttlBytes=0; ttlBytes<stopAt; ttlBytes+=32768L)  {
        //simulation
      }
      eraseOK = true;
    }
    else  {
      printf("  Erasing entire EEPROM");
      eraseOK = true;
    }

    printf("\r\n");
    if (!eraseOK)  { printf("Erase timed out\r\n");  break; }

    if (fstat(fromfd, &fsInfo) < 0)  { perror(NULL);  Exit(-2); }
    fileSize = (ulong)fsInfo.st_size;
    printf("  File size: %lu bytes\r\n", fileSize);


  //===== copy the firmware file from the source fd to the EEPROM =====
    printf("  Copying file to EEPROM...\r\n");
    ttlBytes = 0;  busy = false;

    memset(cmdBuf, 0, sizeof(cmdBuf));

    while (((count = read(fromfd, buf, BufSize))) > 0)  {

      for (i=0; i<count; ++i)  {
        ++ttlBytes;
        if (stopAt AND (ttlBytes >= stopAt))  break;
      }

      if (showProgress)  { printf("%10lu\r", ttlBytes);  fflush(stdout); }
    }

    if (busy OR (ttlBytes < fileSize) OR (ttlBytes > fileSize+1))
      printf("Error writing to EEPROM\r\n");
    else
      if (showProgress)  printf("  Wrote %lu bytes to EEPROM\r\n", ttlBytes);


  //===== protect all sectors =====
    printf("  Protecting all sectors...\r\n");  protectOK = false;

    protectOK = true;  //simulation

    if (!protectOK)  { printf("  Unable to protect EEPROM\r\n");  break; }



  //===== verify contents of EEPROM =====
    if (verify)  {
      printf("  Verifying EEPROM copy...\r\n");

      if (lseek(fromfd, 0, SEEK_SET) != 0)  {
        printf("Unable to reset source file pos to verify EEPROM\r\n");  break; }

      ttlBytes = 0;  errCount = 0;

      while (((count = read(fromfd, buf, BufSize))) > 0)  {
        for (i=0; i<count; ++i)  { sleepMS(100); }
        if (errCount >= 100)  {
          printf("*** Over 100 errors! ***\r\n");  break; }
        ttlBytes += count;
        if (showProgress)  { printf("%10lu\r", ttlBytes);  fflush(stdout); }
        if (ttlBytes >= fileSize)  break;  //tdebug
      }
      printf("\r\n");

      if ((ttlBytes == fileSize) AND !errCount)  {
        downloadOK = true;  printf("Verified %lu bytes\r\n", ttlBytes);  }
      else
        printf("Copies do not match!\r\n");
    }

  } while(0);


  if (downloadOK)  printf("FPGA firmware download complete!\r\n");


//--- Restart the FPGA and DSP so they (re)load their firmware from their ---
//--- respective EEPROMs. ---
  status = restart_DSP_FPGA(30000);

  releaseLock(APPL_LOCK);

  return status;
}




//-----------------------------------------------------------------------------
//  "dspread": command to read one or more 32-bit values from the DSP
//-----------------------------------------------------------------------------
void _task dspRead(int argc, char **argv)
{
  ulong  regNum;
  uint  i, count, wordCount;
  ulong  xmitWord, rcvWord, respRegNum;
  char  temp[64];


  if (showHelp(3, argc, argv))  {
    printf("Usage:  dspread <reg> <count>\r\n");  Exit(-1); }

  regNum = ATOUL(argv[1]) & 0xFFFFFFFEL;
  count = (uint) ATOI(argv[2]);

  if (getLockDSP(1000) < 0)  { perror("getLockDSP");  Exit(-1); }

  wordCount = count + RespOffset;

  for (i=0; i<wordCount; ++i,regNum+=2)  {
    xmitWord = (i >= count) ? 0 : htonl(regNum);

    *((float *)&rcvWord) = 0.1234 + (float)(uint)(GetMS() & 0x3F);  //simulation
    rcvWord = htonl(rcvWord);  //simulation

    if (i < RespOffset)  {
      if (showFiller)  printf("  filler: 0x%08lX\r\n", ntohl(rcvWord));
      continue;
    }

    rcvWord = ntohl(rcvWord);

    respRegNum = regNum - (ulong)(RespOffset * 2);

    printf("  0x%08LX:  0x%08LX  %.32s\r\n",
           respRegNum, rcvWord, fmtDSPval(rcvWord, respRegNum, temp));
  }
  printf("\r\n");

  releaseLockDSP();

  if (xmitWord)  { }   //dummy

  Exit(0);
}

//-----------------------------------------------------------------------------
//  "dspwrite": command to write a value to the DSP
//-----------------------------------------------------------------------------
void _task dspWrite(int argc, char **argv)
{
  ulong  regNum, setVal;


  if (showHelp(3, argc, argv))  {
    printf("Usage:  dspwrite <reg> <value> [-f]\r\n");  Exit(-1); }

  regNum = ATOUL(argv[1]) | 1L;

  if (strstr(argv[2], "."))
    *(float *)&setVal = atof(argv[2]);
  else
    setVal = ATOUL(argv[2]);

  printf("  write 0x%08lX to 0x%08lX\r\n", setVal, regNum);

  writeDSPReg(regNum, setVal, true);

  //NOTE:  Can't just read back the result to check if it worked because
  //       reading from some registers triggers events and processing
  //       within the DSP.

  Exit(0);
}


//-----------------------------------------------------------------------------

