//----- ModularPS.c ----- 10/02/12 -----(03/26/06)-----


//-----------------------------------------------------------------------------
//  Board Support Package for the NSCL Modular Power Supply circuit board
//  which uses a ZWorld Rabbit Core Module 3200 (RCM3200).
//
//  This board has 22 12-bit ADC inputs, 4 16-bit ADC inputs, and 3 16-bit DAC
//  outputs
//-----------------------------------------------------------------------------
//  Change Log (most recent first):
//
//
//  2012-10-02 - Mark Davis (*)
//               - Reduced the threshold for excessive nested scheduler locks
//                 (used to detect excessive knob interrupts) from 50 to 30.
//                 Prior to this there was at least one instance of a stack
//                 overflow that occured around the 50 level.  Also, the
//                 source of the interrupt is not reenabled until the applic
//                 task has run at least 20 times since the interrupt was
//                 disabled (to reduce the affect on other tasks).
//
//  2011-12-22 - Mark Davis (*)
//               - Changed IP for knob interrupt from 3 to 1 so the knob
//                 interrupt is not immune to critical section locks.
//               - Changed the patch that checks for excessive knob interrupts
//                 so it is based only on the level of nested scheduler locks.
//
//  2011-10-27 - Mark Davis (*)
//               - Added patch from nscl16.c module to guard against stack
//                 overflows that can occur from a continuous rapid stream
//                 of interrupts from the front panel knob.
//               - Changed all the DC-style WrPortI() func calls to the more
//                 efficient Softools outi() macro.
//
//  2011-02-25 - Mark Davis
//               - Moved handling of forced bits for extraIn to ReadDigInputs()
//
//  2011-02-07 - Mark Davis (*)
//               - Removed critical section in qdPulse() ISR.  The only purpose
//                 it served was to protect the qdCount value from access by
//                 OTHER ISRs (which I don't do), so this isn't really
//                 necessary and runs the risk of overflow the IP "stack"
//                 register (which MIGHT be the cause of the rare cases where
//                 there has been an unexplained interrupt "runaway" that
//                 leads to a stack overflow.
//              2011-04-15 NOTE:  While the critical section is still overkill,
//              the ISR saves the IP register on the stack and restores it
//              just before it returns, so its value should be correct when it
//              finishes executing the ISR.  The root problem is that if the ISR
//              VECTOR repeatedly get interrupted before IT returns, which can
//              happen because the ISR restores the IP state on return to the
//              ISR vector.  So the real question is still HOW/WHY the knob
//              can get in to a state of generating a continuous stream of
//              pulses?
//
//  2010-02-06 - Mark Davis -
//               - Fixes to convertADS7825(), ReadDigInputs(),
//                 WriteDigOutputs(), & hdwInit() to avoid race conditions
//
//  2008-06-23 - Mark Davis
//               - Changed WDT_2S() call to resetWDT()
//
//  2008-01-08 - Mark Davis
//               - removed unnecessary extra A/D channel conversion
//
//  2007-09-24 - Mark Davis
//               - updated default ADC/DAC calibrations
//
//  2007-05-21 - Mark Davis
//               - used pieces from nscl16, bl2105, and ps8rack to create
//                 initial version.
//
//-----------------------------------------------------------------------------

#include <common.h>

#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>

#include <sys/backup.h>

#include <sys/spi.h>
#include <sys/hdwDefs.h>

#include <busybox.h>

#include <Rabbit.h>
#include <dcdefs.h>

#include <stcpip.h>

#include <sys/socket.h>
#include <sys/srfs.h>
#include <sys/flashFile.h>

#include <sys/farprintf.h>


//--- see hdwInit() for Digital I/O pin usage ---

const char ArchName[] = { "ModularPS" };

#define  NUM_ADC   26
#define  NUM_DAC    3

#define  MAX_DIN   32
#define  MAX_DOUT  16


//--- these values can be set LOWER by the application ---
const uint  NumHdwADC = NUM_ADC;
const uint  NumHdwDAC = NUM_DAC;

const uint  MaxDIN =  MAX_DIN;
const uint  MaxDOUT = MAX_DOUT;


//--- NSCL 16-bit analog RCM3200-based ctlrs use port B for SPI to ADCs, DACs
#define  ADC_DAC_SPIPort  1  // for most RCM3200-based NSCL ctlr's w/Analog


unsigned char  __brdInitFlag;

#include <hdwDefs.c>


SerialInfo  *spiPort;

uchar  restoreI1CR;



#define  IFDBK_INPUT  0
#define  VOUT_INPUT   1
#define  VCTRL_INPUT  2
#define  VLIM_INPUT   3

//=============================================================================
#define  BiasInputs    0x02551550L
#define  CurInputs     0x01aa8aa0L
#define  SpareInputs   0x00006000L
//=============================================================================
_nearcall float adcGetRaw(uint ui)  { return (float) adcRaw[ui]; }

_nearcall float adcMinRaw(uint ui)  { return 0.0; }
_nearcall float adcMaxRaw(uint ui)  { return (ui > 3) ? 0xfff : 0xffff; }

//-------------------------------------------------------------
_nearcall void adcDefaults(uint ui, float *gain, float *offset)
{
  if (ui == VOUT_INPUT)  {
    *gain = 0.0004578;  *offset = 32767.5;  return; }   // +/-15V

  if (ui < 4)  {
    *gain = 0.0003052;  *offset = 32767.5;  return; }   // +/-10V

  if ((1L<<ui) & BiasInputs)  {
    *gain = 0.003664;  *offset = 0.0;  return; }        // 0-15V

  if ((1L<<ui) & CurInputs)  {
      *gain = 0.00004885;  *offset = 2047.5;  return; } // +/-100mA

  *gain = 0.0012212;  *offset = 0.0;                    // 0-5V
}
//---------------------------------------------------------
_nearcall void adcCalibPoints(uint ui, float *low, float *high)
{
  if (ui == VOUT_INPUT)
    { *low = -10.0;  *high = 10.0;  return; }   // +/-15V

  if (ui < 4)
    { *low = -9.5;  *high = 9.5;  return; }     // +/-10V

  if ((1L<<ui) & BiasInputs)
    { *low = 0.5;  *high = 9.5;  return; }      // 0-15V

  if ((1L<<ui) & CurInputs)  {
     *low = -0.095;  *high = 0.095;  return; }  // +/-100mA

  *low = 0.5;  *high = 4.5;                     // 0-5V
}
//---------------------------------------------------------
_nearcall char *adcUnits(uint ui, char *buf)  {
   return ((1L<<ui) & CurInputs) ? "A" : "V"; }


//=============================================================================
_nearcall float dacGetRaw(uint ui)  { return (float) dacRaw[ui]; }

_nearcall float dacMinRaw(uint ui)  { return 0.0; }
_nearcall float dacMaxRaw(uint ui)  { return (float)0xffffL; }

_nearcall void dacDefaults(uint ui, float *gain, float *offset)
  {  *gain = 0.0003125;  *offset = 32767.5;  }  // +/- 10.24V

_nearcall void dacCalibPoints(uint ui, float *low, float *high)
  {  *low = -9.5;  *high = 9.5; }

_nearcall char *dacUnits(uint ui, char *buf)  { return "V"; }

//=============================================================================


//-----------------------------------------------------------------------------

/* START FUNCTION DESCRIPTION ********************************************
Function: int _readExtioPort ( int port );

Note:
This function is setup to provide additional hold time when
accessing external I/O addresses 0000 and 0002. This routine
must stay         and not be changed, for the additional
address hold-time to be valid.

Improved by Softools to not require a copy to RAM because
real tools have align and other stuff for real programmers.
************************************************************/
/*
unsigned readExtIO (unsigned port)      // Port is 0 or 2
{
#asm
    push    ip
    ipset   3
    ld      a,l
    and     a,0x03
    jr      z,..read0
    jr      ..skip
    .align  4           ; Align to clear A0 & A1
..read0:
    nop                 ; A0 = 0, A1 = 0
    nop                 ; A0 = 1, A1 = 0
    ioe     ld l,(hl)   ; A0 = 0, A1 = 1
    pop     ip          ; A0 = 0, A1 = 0
    jr      ..exit
    .align  4           ; Align to clear A0 & A1
..skip:
    ioe     ld l,(hl)   ; A0 = 0, A1 = 0 (2 bytes)
    pop     ip          ; A0 = 0, A1 = 1
..exit:
#endasm
}
*/

//-----------------------------------------------------------------------------
//  determine the device ID based on external hardware settings
//-----------------------------------------------------------------------------
ulong ReadDeviceID()
{
  uchar  inA, devID;


  inA = ini(PADR);

//--- device ID comes from 8-bit switch input ----
//--- bits need to be reversed and then inverted ---
  devID = 0;
  if (inA & 0x01)  devID |= 0x80;
  if (inA & 0x02)  devID |= 0x40;
  if (inA & 0x04)  devID |= 0x20;
  if (inA & 0x08)  devID |= 0x10;
  if (inA & 0x10)  devID |= 0x08;
  if (inA & 0x20)  devID |= 0x04;
  if (inA & 0x40)  devID |= 0x02;
  if (inA & 0x80)  devID |= 0x01;
  devID = ~devID;

  return (ulong) devID;
}

//-----------------------------------------------------------------------------
//  Return the current state of the digital inputs
//
//  The digital inputs are connected to a series of octal line drivers whose
//  outputs are connected to the parallel G signal lines.  To read all the
//  inputs, we enable the drivers one at a time and read the state of the G
//  lines, adding the result to previously read bits.
//
//  The digital inputs and outputs are address in groups of 8 bits.  Which
//  group of 8 bits we are setting or reading is controlled by a 3-to-8 line
//  decoder, where toggling one of its 8 outputs low latches the desired
//  data.
//
//  To avoid activating the wrong one of the 8 outputs even for a moment, we
//  have to make sure the 3 inputs that specify which output to toggle low
//  are all stable BEFORE we enable the outputs and while we are disabling
//  the outputs.
//-----------------------------------------------------------------------------
ulong ReadDigInputs()
{
  ulong  inBits0, inBits1;
  uchar  otherBits = srPBDR & 0x87;


//--- To insure that only the intended Y output changes state, we need to ---
//--- apply the bits that select which output, enable the Y outputs, ---
//--- then disable the Y outputs without changing the selection bits. ---
//--- If we change the selection bits at the same time that we enable ---
//--- the Y outputs, then we run the risk of causing other outputs to ---
//--- briefly change state.  This could be especially bad if one of them ---
//--- causes one of the sets of digital output latches to change state ---

//--- read Front Panel inputs --- (Y6)
  outi(PBDR, srPBDR = otherBits | 0x38 );  // select /Y6
  outi(PBDR, srPBDR = otherBits | 0x18 );  // /Y6 low to enable driver
  inBits1 = ini(PGDR);                     // read the data
  outi(PBDR, srPBDR = otherBits | 0x38 );  // /Y6 high again

//--- read input bits 24-31 --- (Y5)
  outi(PBDR, srPBDR = otherBits | 0x70 );  // select /Y5
  outi(PBDR, srPBDR = otherBits | 0x50 );  // /Y5 low to enable driver
  inBits0 = (ulong)ini(PGDR);              // read the data
  outi(PBDR, srPBDR = otherBits | 0x70 );  // /Y5 high again

//--- read input bits 16-23 --- (Y4)
  outi(PBDR, srPBDR = otherBits | 0x30 );  // select /Y4
  outi(PBDR, srPBDR = otherBits | 0x10 );  // /Y4 low to enable driver
  inBits0 = (inBits0 << 8) | (ulong) ini(PGDR);  // read the data
  outi(PBDR, srPBDR = otherBits | 0x70 );  // /Y4 high again

//--- read input bits 08-15 --- (Y3)
  outi(PBDR, srPBDR = otherBits | 0x68 );  // select /Y3
  outi(PBDR, srPBDR = otherBits | 0x48 );  // /Y3 low to enable driver
  inBits0 = (inBits0 << 8) | (ulong) ini(PGDR);  // read the data
  outi(PBDR, srPBDR = otherBits | 0x68 );  // /Y3 high again

//--- read input bits 00-07 --- (Y2)
  outi(PBDR, srPBDR = otherBits | 0x28 );  // selct /Y2
  outi(PBDR, srPBDR = otherBits | 0x08 );  // /Y2 low to enable driver
  inBits0 = (inBits0 << 8) | (ulong) ini(PGDR);  // read the data
  outi(PBDR, srPBDR = otherBits | 0x28 );  // /Y2 high again


  outi(PBDR, srPBDR = otherBits | 0x20);  // disable all drivers

  inBits0 = (inBits0 & ~forcedDigInMask) | forcedDigInBits;
  inBits1 = (inBits1 & ~forcedExtraInMask) | forcedExtraInBits;

  extraIn = inBits1;

  return inBits0;
}


//-----------------------------------------------------------------------------
//  set the state of the digital outputs
//
//  The digital inputs and outputs are address in groups of 8 bits.  Which
//  group of 8 bits we are setting or reading is controlled by a 3-to-8 line
//  decoder, where toggling one of its 8 outputs low latches the desired
//  data.
//
//  To avoid activating the wrong one of the 8 outputs even for a moment, we
//  have to make sure the 3 inputs that specify which output to toggle low
//  are all stable BEFORE we enable the outputs and while we are disabling
//  the outputs.
//-----------------------------------------------------------------------------
void WriteDigOutputs(ulong digOut)
{
  ulong  outBits;
  uchar  otherBits = srPBDR & 0x87;


  digOut = (digOut & ~forcedDigOutMask) | forcedDigOutBits;

  outBits = ~digOut;

//--- set digital I/O bus (port G) to output mode ---
  outi(PGDDR, srPGDDR = 0xFF);

//--- update output bits 00-07 --- (Y0)
  outi(PGDR, srPGDR = (uchar)outBits);          // put value on bus
  outi(PBDR, srPBDR = otherBits | 0x20);        // select /Y0
  outi(PBDR, srPBDR = otherBits );              // enable outputs (/Y0 low)
  outi(PBDR, srPBDR = otherBits | 0x20);        // disable outputs (all high)

//--- update output bits 08-15 --- (Y1)
  outi(PGDR, srPGDR = (uchar)(outBits >> 8));   // put value on bus
  outi(PBDR, srPBDR = otherBits | 0x60);        // select /Y1
  outi(PBDR, srPBDR = otherBits | 0x40 );       // enable outputs (/Y1 low)
  outi(PBDR, srPBDR = otherBits | 0x60);        // disable outputs (all high)

//--- set digital I/O bus (port G) back to input mode ---
  outi(PGDDR, srPGDDR = 0x00);
}


//-----------------------------------------------------------------------------
//  Function to communicate with a TLC2543 (an 11-channel, 12-bit ADC)
//
//  This functions initiates a conversion for the specified input and saves
//  the result of the previous conversion (initiated by the previous call to
//  this function).
//
//  Inputs:
//    chipSel:  0-1
//    nextChan: 0-14 (12-14 return test voltages)
//-----------------------------------------------------------------------------
void convertTLC2543(uchar chipSel, uint nextChan)
{
  uchar  yNum, eocMask;
  uint  i, raw;  //chan
  uchar  xmitBuf[2], rcvBuf[2];
  uint  *prevCh;
  bool  convDone;
  static uint  prevChan[2] = { 4, 15 };


  chipSel &= 1;  eocMask = 0x40 << chipSel;
  prevCh = prevChan + (uint)chipSel;

  xmitBuf[0] = (nextChan << 4) | 0x0C;  xmitBuf[1] = 0;

//--- wait for EOC to go high ---
  for (i=0,convDone=false; i<100; ++i)  {
    if (ini(PFDR) & eocMask)  { convDone = true;  break; }
    resetWDT();
  }

//--- select one of the 2 ADC chips (Y3 and Y4 outputs of 3-to-8 decoder) ---
//--- (Y# output set by PD5-PD4-PE3) ---
  yNum = chipSel + 3;
  outi(PEDR, srPEDR = (srPEDR & 0xF7) | (yNum&1)<<3 );
  outi(PDDR, srPDDR = (srPDDR & 0xCF) | (yNum&6)<<3 );

//--- enable selected decoder output line to set specified /CS low ---
  outi(PBDR, srPBDR &= 0x7F);  // PB7 ctls decoder output

//--- tell ADC what input to begin conversion on and read result of the ---
//--- conversion started by the previous call to this func ---
  _spiWrRd_(spiPort, xmitBuf, rcvBuf, 2);

//--- disable decoder outputs to bring /CS high ---
  outi(PBDR, srPBDR |= 0x80);  // PB7 ctls decoder output

//--- extract the raw value from the bytes just read ---
  if (convDone)
    raw = ((uint)rcvBuf[0] << 4) | (rcvBuf[1] >> 4);
  else
    raw = 0xBAD;
  adcRaw[(*prevCh)] = raw;

//--- keep track of what channel we started the conversion on, cause that ---
//--- is the channel whose value will be read during the next call for ---
//--- the same chip ---
  *prevCh = chipSel * 11 + nextChan + 4;
}


//-----------------------------------------------------------------------------
//  Function to communicate with an ADS7825 chip (a 4-channel, 16-bit ADC)
//
//  Call:
//    convertADS7825(uint nextChan)
//
//    Input:
//       nextChan: Channel # to sample between calls and to do conversion
//                 on during the next call.
//
//    Result:
//       The raw value for the channel specified during the previous call
//       is updated.
//
//
//  To maximize the conversion rate, the ADS7825 begins acquiring/sampling
//  the signal to be converted as soon as the previous conversion is
//  completed.  This requires that it already know what channel to begin
//  sampling, so you have to tell it what channel to sample for the NEXT
//  conversion, NOT the current conversion.
//
//  NOTE that this does NOT mean that the conversion initiated by this call
//  is for outdated data.  The ADS7825 is sampling the channel specified
//  during the last call right up to the time it is told to start the
//  conversion.  The only catch is you have to know what channel that was
//  so when you read the conversion data you know what channel it is for.
//
//  The ADC7825 is fairly unique in that it has a TAG input that allows
//  multiple ADC's to be connected in a daisy-chain configuration, allowing
//  you to initiate a conversion on multiple chips at the same time and then
//  read the results from all of them in a single operation.  This greatly
//  decreases the additonal time needed to support more than 4 channels when
//  the chips are connected in this manner.
//
//  NOTE:  The downside, of course, is the cost per input, both in $ and in
//         board space.  Only 4 inputs means never more than 4 passes to
//         read any # of channels.  It also means one more (expensive) chip
//         for each additional group of 4 channels.
//
//-----------------------------------------------------------------------------
void convertADS7825(uint nextChan)
{
  uchar  buf[10];
  static uint  prevChan = 0;


  nextChan &= 3;

//--- Set the bits that will tell the ADC which channel to sample between ---
//--- calls to this func, and to hold and convert during the next call ---
//--- (PE6 and PE7 are connected to ADR0 and ADR1 on all ADC's) ---
  outi(PEDR, srPEDR = (srPEDR & 0x3F) | (nextChan << 6));

//--- set R/C- low first so we can use CS- to start conversion ---
  outi(PFDR, srPFDR &= 0xF7);  // R/C- (PF3) low (convert)

//--- Toggle /CS low then high very quickly.  This starts the conversion ---
//--- on the previously specified input and latches the input # to be used ---
//--- for the next conversion.  /CS must be low for at least 40ns, but must ---
//--- go high again before BUSY does (aprox 25us).  If /CS is still low when ---
//--- BUSY goes high, the next conversion will start and the results of the ---
//--- previous one will be lost ---
  _EnterCritical_();
  outi(PFDR, srPFDR &= 0xFB);  // /CS (PF2) low
  outi(PFDR, srPFDR |= 0x04);  // /CS (PF2) high
  _ExitCritical_();

//--- set R/C- high to enable read mode ---
  outi(PFDR, srPFDR |= 0x08);  // R/C- (PF3) high

//--- The design did not include the ability to read the BUSY signal, so ---
//--- add a short delay to avoid trying to read the data before it is ready ---
  for (uint xx=0; xx<23; ++xx);

//--- set /CS low to enable transmission of data ---
  outi(PFDR, srPFDR &= 0xFB);  // /CS (PF2) low

//--- The ModularPS controller does not have ADC's chained together, so we ---
//--- get just one channel's value each time.  NOTE that the ADC sends us ---
//--- 17 bits (2+ bytes), the first of which is NOT part of the data. ---
  _spiRead_(spiPort, buf, 3, false);

//--- set /CS high so ADC does not react to clock pulses not meant for it ---
  outi(PFDR, srPFDR |= 0x04);  // /CS (PF2) high

//--- extract the raw value from the bytes just read ---
  adcRaw[prevChan] = (uint) (0x8000 ^ (
      ((uint)buf[0] << 9) |
      ((uint)buf[1] << 1) |
      ((uint)buf[2] >> 7)
    ) );

//--- keep track of what channel we started the conversion on, cause that ---
//--- is the channel whose value will be read during the next call ---
  prevChan = nextChan;
}

//-----------------------------------------------------------------------------
//  Read the Analog inputs
//
//  The Modular Power Supply board has 3 ADC chips:
//    (1)  4-channel, 16-bit ADC7825
//    (2) 11-channel, 12-bit TLC2543
//
//  In contrast, the TLC2543's are inexpensive and provide nearly 3 times as
//  many inputs.  Aside from the lower resolution, the main disadvantage is
//  that reading a given number of channels takes much longer, since 11 read/
//  convert cycles must be done per chip, and the cycles cannot be applied to
//  more than one chip at a time like they can with the ADC7825.
//
//  In this case, the requirements called for reading lots of channels with as
//  few chips as possible.  Given that the ADC7825 is quite a bit more
//  expensive than the TCL2543 (especially on a per-channel basis), the
//  less expensive TLC2543 was chosen for those inputs where speed and
//  resolution was not an issue.
//-----------------------------------------------------------------------------
void readAnalogInputs(ulong mask)
{
  uint  nextChan;
  static uint  ticks;
  static uint  waitTicks = 20, lastLogged = 0;


//--- We use the fact that this is called periodically but NOT by ISRs. --
//--- This allows us to check for a temporarily disabled INT1 ISR ---
//--- and re-enable it (after a bunch of nested INTs have "unwound") ---
  if (restoreI1CR)  if (--waitTicks == 0)  {
    if (msSince_(lastLogged) > 30000)  {
      printf("*** Excessive INT1 occurances detected ***\r\n");
      lastLogged = GetMS_();
    }
    _EnterCritical_();
    outi(I1CR, srI1CR = restoreI1CR);  // reenable INT1
    restoreI1CR = 0;  waitTicks = 20;
    _ExitCritical_();
  }


  if (! spiPort)  return;


//===== process the 4 inputs on the ADC7825 first =====

//--- Remember we have to tell the ADC which channel's data we will read ---
//--- NEXT time we call it, so consistently using the sequence 1, 2, 3, 0 ---
//--- each time means we get the data for 0, 1, 2, 3.  Also remember that ---
//--- this means channel 0 will be continously sampled (READY for conversion) ---
//--- between calls to this function ---
  _spiConfig_(spiPort, CLK_1MHz, NORM_LOW_LATCH_FALL);
  convertADS7825(1);  // chan to convert NEXT call
  convertADS7825(2);  // chan to convert NEXT call
  convertADS7825(3);  // chan to convert NEXT call
  convertADS7825(0);  // chan to convert NEXT call


//--- There are 22 channels total for the 2 TLC2543 ADC's, and they don't ---
//--- support the TAG feature, so getting RECENT data for all 22 channels ---
//--- would require making 24 consecutive calls, which would take too long. ---

//--- Since the timing and sample rate for these inputs is not critical, ---
//--- we read just one from each chip for each call to this function ---
  _spiConfig_(spiPort, CLK_1MHz, NORM_LOW_LATCH_RISE);
  nextChan = ++ticks % 11;
  convertTLC2543(0, nextChan);
  convertTLC2543(1, nextChan);
}


//-----------------------------------------------------------------------------
//  Update the state of the AD5542 DAC chips
//-----------------------------------------------------------------------------
void writeAnalogOutputs(ulong mask)
{
  uchar  buf[2];
  uint   rawVal, bitTest;
  uchar  i;


  if (! spiPort)  return;

//--- set SPI clock mode ---
  _spiConfig_(spiPort, CLK_1MHz, NORM_LOW_LATCH_RISE);

  for (i=0,bitTest=1; i<NUM_DAC; ++i,bitTest<<=1)  if (bitTest & (uint)mask)  {

  //--- select one of the 3 DAC chips ---
  //--- (chip # set by PD5-PD4-PE3) ---
    outi(PEDR, srPEDR = (srPEDR & 0xF7) | (i&1)<<3 );
    outi(PDDR, srPDDR = (srPDDR & 0xCF) | (i&6)<<3 );

  //--- enable selected decoder output line to set /CS low for one DAC ---
    outi(PBDR, srPBDR &= 0x7F);  // PB7 ctls decoder output

  //--- send cmd to DAC chip ---
    rawVal = dacRaw[i];   buf[0] = rawVal >> 8;  buf[1] = rawVal & 0xff;
    _spiWrite_(spiPort, buf, 2, false);

  //--- disable decoder outputs to bring /DS high for all DAC's ---
    outi(PBDR, srPBDR |= 0x80);  // PB7 ctls decoder output
  }

}


//-----------------------------------------------------------------------------
/*
float ADCtoVolts(uchar chan, uint rawADC)
{
  return (ADCOffset(chan) - (float)(rawADC)) * ADCGain(chan);
}
*/

//-----------------------------------------------------------------------------
//  Interrupt Service Routine to catch state of IQ optical encoder.
//  Assumes that A is tied to an input bit and B is tied to the INT-PE5 input
//  (and triggered the interrupt that called this func)
//-----------------------------------------------------------------------------
void _int qdPulse()
{
/*
  If the same interrupt occurs again before the ISR shim that called this ISR
  returns to the interrupted code, the next instance will cause this ISR to
  be called right after the ipres/ret at the end of this ISR.  This means the
  ISR shim which called this ISR will not return to the originally interrupted
  code before this ISR is invoked again.  If this happens too many times in a
  row, the stack of the originally interrupted task will overflow.

  To guard against this possibility, we check to see if the level of nested
  scheduler locks has gotten unreasonably large.  If it has, this ISR disables
  the interrupt that calls it, which (if the above IS the cause of the
  scheduler not running), will allow all the previous occurances to return.

  Once all the previous occurances have finished, the scheduler will be active
  again and the application task will run.  One of the functions that the
  application task calls is used to check to see if the interrupt was disabled.
  If it was, it reenables it, and things will be back to normal (unless, of
  couse, we are still getting a continuous stream of interrupts, in which case
  the cycle keeps repeating.

  While this can hurt performance, that is still far better than allowing the
  OS to crash and reboot.  And experience has shown that this situation is
  often temporary, and can easily be stopped just by touching the knob on the
  front of the controller.
 */
//  uint  stackLeft;

//  stackLeft = CurSP() - (uint) current->stack_start;

//  if ((stackLeft < 100) AND (_disabled > 30))  {
  if (_disabled > 30)  {  // _disabled is incr by the ISR shim...
    restoreI1CR = srI1CR;  // save the value that will reenable this ISR
    outi(I1CR, srI1CR = 0x00);  // disable INT1
  }

//--- got here because B just went high - if A already high (i.e. A leads B)---
//--- then knob was turned clockwise and count should be increased ---
  if (ini(PEDR) & 0x10)  ++qdCount;  else  --qdCount;   // A tied to PE4
}


//-----------------------------------------------------------------------------
//  Set an External 2/4-wire RS-485 adapter to transmit mode
//-----------------------------------------------------------------------------
void setExt485Tx(void)
{
}
//-----------------------------------------------------------------------------
//  Set an External 2/4-wire RS-485 adapter to receive mode
//-----------------------------------------------------------------------------
void setExt485Rx(void)
{
}

//-----------------------------------------------------------------------------
//  functions to enable/disable the transmitter for an RS-485 connection
//-----------------------------------------------------------------------------
void _nearcall SerSetDirA(int enableXmit)  {  }
void _nearcall SerSetDirB(int enableXmit)  {  }
void _nearcall SerSetDirC(int enableXmit)  {  }
void _nearcall SerSetDirD(int enableXmit)
{  // PF5 controls RS-485 line driver for port D(3) RS-485 connections
  char  val;
  val = enableXmit ? (srPFDR | 0x20) : (srPFDR & ~0x20);
  outi(PFDR, srPFDR = val);
}


//-----------------------------------------------------------------------------
void hdwInit(ulong digOutState)
{
  hdwType = HT_MODULAR_PS;

  __brdInitFlag = TRUE;

// (NI) = Normal Input
// (NO) = Normal Output
// (AI) = Alternate Input
// (AO) = Alternate Output
// (xx) = Used by ZWorld core - do not change

  //--- Port A --- (*must be inputs due to bug in earlier Rabbit 3000's)
  //  PA0 - (NI) *
  //  PA1 - (NI) *
  //  PA2 - (NI) *
  //  PA3 - (NI) *
  //  PA4 - (NI) *
  //  PA5 - (NI) *
  //  PA6 - (NI) *
  //  PA7 - (NI) *
    outi(SPCR, srSPCR &= ~0x0c);  // PortA: byte input

  //--- Port B ---
  //  PB0 - (AO)  Serial-B (SPI) clock
  //  PB1 - (AO)  Serial-A (Diag/Prog) clock
  //  PB2 - (NO)  /OE input on all Digital output line drivers
  //  PB3 - (NO)  A1 input on 3-to-8 line decoder for Dig I/O Bus
  //  PB4 - (NO)  A2 input on 3-to-8 line decoder for Dig I/O Bus
  //  PB5 - (NO)  /OE2A input on 3-to-8 line decoder for Dig I/O Bus
  //  PB6 - (NO)  A0 input on 3-to-8 line decoder for Dig I/O Bus
  //  PB7 - (NO)  G2A on 3-to-8 decoder for Analog output selector
    outi(PBDR, srPBDR |= 0xA0);   // PB7 and PB5 high
    outi(PBDDR, srPBDDR |= 0xFC); // dir: OOOO-OOxx

  //--- Port C ---
  //  PC0 - (AO)  Serial-D (RS-232/485) transmit (to rear connectors)
  //  PC1 - (xI)  Serial-D (RS-232/485) receive (to rear connectors)
  //  PC2 - (AO)  Serial-C (RS-232) transmit (to front panel LCD)
  //  PC3 - (xI)  Serial-C (RS-232) receive (from front panel LCD)
  //  PC4 - (AO)  Serial-B (SPI) transmit (to all DAC's)
  //  PC5 - (xI)  Serial-B (SPI) receive (from 1st ADC in the chain)
  //  PC6 - (AO)  Serial-A (Diag/Prog) transmit
  //  PC7 - (xI)  Serial-A (Diag/Prog) receive
    outi(PCFR, srPCFR = 0x55);  // func: xAxA-xAxA

  //--- Port D ---
  //  PD0 - (xx)  Ethernet Controller Reset
  //  PD1 - (xx)  Not Avail (pulled high - not even connected???)
  //  PD2 - (xx)  Ethernet Controller (TPOUT-)
  //  PD3 - (xx)  Ethernet Controller (TPOUT+)
  //  PD4 - (NO)  B input on 3-to-8 decoder for Analog output selector
  //  PD5 - (NO)  C input on 3-to-8 decoder for Analog output selector
  //  PD6 - (xx)  Ethernet Controller (TPIN-)
  //  PD7 - (xx)  Ethernet Controller (TPIN+)
    outi(PDDDR, srPDDDR |= 0x30);    // dir:  xxOO-xxxx
    outi(PDFR, srPDFR &= ~0x30);     // func: xxNN-xxxx

  //--- Port E ---
  //  PE0 - (xx)  Interrupt Input 0A (Ethernet ctlr Interrupt)
  //  PE1 - (AI)  Interrupt Input 1A (buffered from external source)
  //  PE2 - (xx)  Ethernet Controller Chip Select (output)
  //  PE3 - (NO)  A on 3-to-8 decoder for Analog output selector
  //  PE4 - (NI)  Normal input ("A" input from quadrature-encoded knob)
  //  PE5 - (AI)  Interrupt Input 1B (buffered from external source)
  //  PE6 - (NO)  ADR0 on all ADC's (output)
  //  PE7 - (NO)  ADR1 on all ADC's (output)
    outi(PEDDR, srPEDDR = (srPEDDR | 0xC8) & ~0x32); // dir:  OOII-OxIx
    outi(PEFR, srPEFR = (srPEFR | 0x22) & ~0xD8);    // func: NNAN-NxAx

  //--- Port F ---
  //  PF0 - (AO)  Serial-D clock (to RS-485/RS-232 port)
  //  PF1 - (AO)  Serial-C clock (to RS-485/RS-232 port)
  //  PF2 - (NO)  to CS on all ADC's (output)
  //  PF3 - (NO)  to R/C on all ADC's (output)
  //  PF4 - (NO)  xmit enable for Serial C (output)
  //  PF5 - (NO)  xmit enable for Serial D (output)
  //  PF6 - (NI)  EOC for 1st 12-bit ADC
  //  PF7 - (NI)  EOC for 2nd 12-bit ADC
    outi(PFDR, srPFDR |= 0x0C);   // PF2/3 high (ADC R/C-, /CS)
    outi(PFDDR, srPFDDR = 0x3F);  // dir:  IIOO-OOOO
    outi(PFFR, srPFFR = 0x03);    // func: NNNN-NNAA

  //--- Port G ---
  //  PG0 - (NIO)  Digital I/O bus
  //  PG1 - (NIO)  Digital I/O bus
  //  PG2 - (NIO)  Digital I/O bus
  //  PG3 - (NIO)  Digital I/O bus
  //  PG4 - (NIO)  Digital I/O bus
  //  PG5 - (NIO)  Digital I/O bus
  //  PG6 - (NIO)  Digital I/O bus
  //  PG7 - (NIO)  Digital I/O bus
    outi(PGDDR, srPGDDR = 0x00);  // dir:  IIII-IIII (for now)
    outi(PGFR, srPGFR = 0x00);    // func: NNNN-NNNN


  //--- all parallel port outputs driven hi/lo (no open-drain outputs) ---
    outi(PDDCR, srPDDCR = 0x00);
    outi(PFDCR, srPFDCR = 0x00);
    outi(PGDCR, srPGDCR = 0x00);

  //--- prep state of digital outputs ---
    digitalOut = digOutState;  WriteDigOutputs(digitalOut);

  //--- enable output latches so digital output states take effect ---
    outi(PBDR, srPBDR &= ~0x04);  // bring PB2 low

//    ser485Rx();  // disable the RS485 transmitter
//    setExt485Rx();  // set external 2/4-wire RS-485 adapter to Rx mode

    spiPort = NULL;  // set later by a call from the kernel process


//--- set INT1B interrupt handler for rising edge on PE5 using priority 1 ---
  qdCount = 0;  restoreI1CR = 0;
  setInterruptHandler(INT_INT1, qdPulse);
  outi(I1CR, srI1CR = 0x21);  // enable extern INT1B on PE5, rising edge, IP1


  flashSize = 524288L;

  printf("\r\nInitialized %s hardware\r\n", ArchName);
}


//-----------------------------------------------------------------------------
//  Initialize and register hardware-dependent file systems
//-----------------------------------------------------------------------------
void registerFileSystems(void)
{
  socketInit_();
  srfsInit_();
  serialInit_();

  flashFileInit_();
}

//-----------------------------------------------------------------------------
//  Initialize the network hardware
//-----------------------------------------------------------------------------
void netHdwInit()
{
//  Select ethernet hardware:
//  usingAsix, usingDevKit, usingRealtek, usingSMSC
//  usingAll() can be used to support all cmd core modules, but req's more code

  usingAsix();
}


//-----------------------------------------------------------------------------
//  allocate any resources that should be owned by the kernel process
//
//  To be called ONLY by the kernel process!!!!
//-----------------------------------------------------------------------------
void hdwInit2(void)
{
  fd_t  spifd;


  spifd = _serial_(ADC_DAC_SPIPort, 0, SER_SPI, SER_NO_FLOW);

  spiPort = _FileFromFD_(spifd)->ioDev;

  printf("spiPort: %04lX (%d)\r\n", (ulong)spiPort, spiPort->port);  //tdebug
}

//-----------------------------------------------------------------------------

