//----- flash.c ----- 12/21/12 ----(??/??/??)-----

//-----------------------------------------------------------------------------
//  Flash device programming support program.
//  Copyright (c) 2000-2003, Softools, Inc.  All rights reserved
//
//  Modifications by Mark Davis, National Superconducting Laboratory.
//
//  Supports writing to flash from the application.
//
//  Currently, the supplied driver does not support sector write
//  devices.  This was done because byte-write are currently the
//  only type supplied by Z-World and because byte-write are the
//  predominate ones used.  To support sector write devices, ...
//
//  Flash memories supported:  SST39SF010A / SST39SF020A / SST39SF040
//  (or compatible).
//
//-----------------------------------------------------------------------------
//  Change log (most recent first)
//
//
//  2012-12-21 - Mark Davis (*)
//               - Changed mapFlashAddress() so it uses the highest possible
//                 bank # to minimize the chance of selecting a bank that
//                 might contain code that an un-blockable (fault handler)
//                 ISR might try to access.
//                 NOTE that the PROPER way to deal with this (so it works with
//                 hardware that DOES normally have RAM mapped in to the
//                 highest bank/quadrant) is to insure the ISRs for the fault
//                 interrupts (the ones that can occur while a portion of flash
//                 is mapped in) is to insure that the ISR vector points to a
//                 function in the root code (in the 1st quadrant which is
//                 never mapped out) and that the first thing the ISR does is
//                 to insure the MMU regs are configured the way they normally
//                 are, and (if the ISR doesn't reboot the OS) to put them
//                 back the way it found them before returning.
//
//  2010-07-19 - Mark Davis
//               - Refined the code for calling the write function to eliminate
//                 excess overhead in the computation of the arguments for the
//                 call.
//
//  2010-06-11 - Mark Davis
//               - Made all write functions the same in terms of saving and
//                 restoring the value of XPC.  The (currently unused) functions
//                 for writing to sectors where not the same as the (currently
//                 used) byte-at-atime version.
//
//  2010-06-03 - Mark Davis
//               - Memory write protection is temporarily disabled in the
//                 functions that do the work of writing to flash
//
//  2010-05-20 - Mark Davis (*)
//               - Created alternate version of Softools file for DEBROS use:
//                  - mapFlashAddress() checks to make sure the memory bank
//                    the stack is in does not get mapped out.  This was a
//                    problem for DEBROS since stacks can be located anywhere
//                    in RAM.
//                  - mapFlashAddress() takes a pointer to the function that
//                    will use the values it sets and insure the memory bank
//                    it is in will not be mapped out.  Before it assumed that
//                    the code would always be in a buffer with a fixed root
//                    location, which was not true even before my changes
//                    because the read function used a buffer on the stack.
//                  - Root memory is no longer allowed to be mapped out.
//                  - DEBROS always runs from RAM so I removed the code that
//                    supported running from flash by copying code to a buffer
//                    and executing it from there (which just added overhead).
//                  - Added calls to disable the scheduler
//
//-----------------------------------------------------------------------------
//  To-Do:
//
//    - WriteFlash() still has lots of room for improvement.  There is an
//      awful lot of redundant computation going on for each byte written,
//      particulary if we just specify a few more restictions on the call
//      (which are already true for DEBROS).  Basically the LACK of assumptions
//      about how it is called adds LOTS of overhead for flexability that we
//      don't use!
//
//-----------------------------------------------------------------------------

#include <Rabbit.h>
#include <string.h>

#include <sys/types.h>
#include <common.h>

#include <sys/serial.h>  // required to use flushConsole()

#include <linux/sched.h>
#include <unistd.h>

#include <flash.h>

#include <sys/hdwDefs.h>



#undef YES
#undef NO
#define YES 1
#define NO  0



//  Types of functions in RAM
typedef uint (*tDetectFlash) (uint);
typedef uchar (*tWriteFlashByte) (uint, uchar *, uint, uint);
typedef void (*tReadFlash)(ulong, char *, uint);

//  Accessors for address as pointer, ulong, and in parts
typedef union {
  char const far  *p;
  ulong   dw;
  uint    w;
  uchar   b[4];
  uint    hl[2];
} FlAddress;

//  Define a list of Flashes with their address range and type
typedef struct  {
  FlType     type;        //  Type in range
  uint       sectorSize;  //  Size of sector
  ulong      deviceSize;  //  Size of device
  FlAddress  start;       //  Start of a flash
  FlAddress  end;         //  End address of device (+1)
  uchar      mbxcr;       //  Lower nibble of MBxCR
} FlDeviceInt;

//  Define storage for 2 flash devices.  These are by range and need not
//  be contiguous addresses.  deviceCount is the number of flashes detected.
static FlDeviceInt    device[2];
static int        deviceCount;

//  The global unlock DWord used to add some safety to the programming process
ulong       _FlashUnlock;

//  This buffer size depends on the size of function  _DetectFlashDevice,
// _ByteWriteFlash, and _SectorWriteFlash.  The lager one "wins"

//  Set to 0 when DetectFlash is called and the flash is detected.
static uchar   notReady = 1;   //  Mark flash not detected

//  Set to address of last programming error when an error occurs
ulong       _FlashError;

//  Declare detectFlash prototype for initializer following.
FlDeviceInfo  *detectFlash(ulong _4kBlock);

extern const FlDeviceInfo far supportedDevices[];
extern const int numDevices;

//
//  Detect a single flash device.  This is not used directly
//  by an application.
//
const far FlDeviceInfo * far DetectFlashDevice(ulong baseAddr, uchar mbxcr);

//--- set by mapFlashAddress() ---
// WARNING:  These must values must be in root memory (the 1st 58K) and
// protected against change between when they are set and when they are used.
// See mapFlashAddress() for description of these values:
static uchar  MBcr;
static uchar  MBval;
static uchar  MBsVal;
static ulong  MBaddr;


//-----------------------------------------------------------------------------
//  Determine which memory bank we can safely map the specified flash address
//  to long enough to read or write a block of bytes that are all in the same
//  4K block of memory.  This is generally called just once for the 1st byte
//  in a 4K block, as the values are the same for all bytes in the same block.
//
//  The decision is based on the following assumptions:
//    - The size of each memory bank is 256K
//    - The code is running from RAM, NOT from flash
//    - The function that codeAddr refers to does not cross a memory bank
//    - The stack does not cross a memory bank
//
//  The function sets the following global variables:
//    - MBcr:    WHICH memory bank register to use to access the flash
//    - MBVal:   The VALUE to use for the selected memory bank's MBxCR register
//    - MBaddr:  The ADDR in the meory bank's range to use to access the flash
//    - MBsVal:  The value to restore the memory bank register to
//-----------------------------------------------------------------------------
void  mapFlashAddress(ulong flashAddr_, uchar mbxcr, char far *codeAddr_)
{
  FlAddress  flashAddr, codeAddr, stackAddr;
  uchar  test, bank, flashBank, codeBank, stackBank;
  uchar  ws = SerialGet19200TC() > 76 ? _WS2 : _WS1;


  codeAddr.p = codeAddr_;
  flashAddr.dw = flashAddr_;
  stackAddr.dw = 0;  stackAddr.b[1] = (uint)ini(SEGSIZE) & 0xF0;
  stackAddr.dw += ((ulong)ini(STACKSEG) << 12);

  flashBank = (flashAddr.b[2] & 0x0C) >> 2;
  codeBank = (codeAddr.b[2] & 0x0C) >> 2;
  stackBank = (stackAddr.b[2] & 0x0C) >> 2;

//--- Determine which bank is safe to use to access the flash ---
//--- Note we don't allow 0 to be used to access flash because that is ---
//--- where ROOT is (which holds some of the values we set here and use ---
//--- elsewhere) ---
  for (test=1; test<4; ++test)  {
    if ((test == codeBank) OR (test == stackBank))  continue;
//    if ((bank = test) == flashBank)  break;
    bank = test;  // use the largest possible # bank to minimize ISR clashes
  }

//--- If we have to use a different bank than the "natural" one (i.e. the ---
//--- one that would not require any address translation), then figure out ---
//--- which bits need to be inverted, invert them now in the address we will ---
//--- use in the code to access the flash, and set the MBxCR register value ---
//--- so the same bits will be re-inverted when using the selected bank ---
//--- (i.e. so writing to the modified address in the selected bank causes ---
//--- the MMU to generate the correct physical addresses) ---

  if (bank != flashBank)  {
    uchar invert = bank ^ flashBank;
    flashAddr.b[2] ^= (invert <<= 2);
    mbxcr |= (invert << 2);  flashBank = bank;
  }

  MBaddr = flashAddr.dw;
  MBcr = MB0CR + flashBank;
  MBval = mbxcr | ws;
  MBsVal = *(&srMB0CR + flashBank);
}



/*
#if SECTOR_DEVICE_SUPPORT
//-----------------------------------------------------------------------------
//  _SectorWrite: Write a sector (block) to a device.  This uses SDP mode
//  which is safer and prevents random/inadvertent writes to flash from
//  going into page write mode.
//
//  SDP Write : 5555H=AAH 2AAAH=55H 5555H=A0H
//-----------------------------------------------------------------------------
//  WARNING:  This function must be located in root memory (the 1st 64K) and
//            must NOT call functions that may be in the memory bank that is
//            used to write to flash.  Near functions (that don't call any far
//            or nearcall functions) are safe because mapFlashAddress()
//            currently does not allow bank 0 to be used for this.  Macros and
//            (if the compiler ever supports them) in-line functions are safe.
//            Anything else (including nearcall functions) are NOT safe to call!
//-----------------------------------------------------------------------------
writeReturn near *_SectorWrite(uint _4kBlock, ulong writeOffset, uint count)
{
  char    xpc0000 = _4kBlock - 0x0E;
  char    xpc5555 = _4kBlock - (0x0E - 0x05);
  char    xpc2AAA = _4kBlock - (0x0E - 0x02);
  char    xpcSave;
  ulong   writeAddr = _4kBlock + writeOffset;
  char    xpcWrite = (uint) (writeAddr >> 12) + (0x80 - 0xE);
  uchar    *log5555 = (uchar *) 0xE555;
  uchar    *log2AAA = (uchar *) 0xEAAA;
  uchar    *addr = (uchar *) (((uint) writeAddr & 0xFFF) | 0xE000);
  uchar    wr, rd;
  uint    offset;
  uchar    *buff = buffer;
  static writeReturn wrReturn;


  wrReturn.length = sizeof (writeReturn);
  wrReturn.error = 0;

  getXPC(xpcSave);  ipset3();  outi(MBcr, MBval);  DisableWriteProtection();

  setXPC(xpc5555);  *log5555 = 0xAA;    //  Write SDP
  setXPC(xpc2AAA);  *log2AAA = 0x55;
  setXPC(xpc5555);  *log5555 = 0xA0;

  setXPC(xpcWrite);  memcopy( addr, buffer, count );   //  Uses LDIR!

  while ((*addr ^ *addr) & 0x40); //  Use toggle mode to wait for prev write

  for (offset = 0; count; --count, ++offset)
    if ((rd = *addr++) != (wr = *buff++))  {
      wrReturn.error = 1;   //  Can use more info later if nec
      wrReturn.address = writeAddr + offset;
      wrReturn.writtenByte = wr;
      wrReturn.readByte = rd;
    }

  EnableWriteProtection();  setXPC(xpcSave);  outi(MBcr, MBsVal);  ipres();

  return &wrReturn;
}
#endif
*/

//-----------------------------------------------------------------------------
//  This is left for the user to implement.  Since it's a byte-write
//  function, reading in the sector, changing the byte, and writing
//  back is the algorithm.  But it is highly inefficient.  Some method
//  of buffering and writing a sector at a time should be provided.
//  Since this function is likley to be called sequentially, this
//  could buffer a sector and write it when the last byte is written.
//  Of course, finishing the final buffer's worth has to be handled.
//
//  _FlashByteSecWrite: Write a byte to a device.
//
//  _4kBlock is the address >> 12 (since we can't do a runtime call in this).
//
//  If erase is non-0, the sector containing the byte to program will be erased.
//  It's unlikely that this paramter is useful for sector-write devices.
//
//  This must not contain branching code more than 128 bytes.
//
//  Returns 0 on no errors, or 1 if the byte doesn't match.
//-----------------------------------------------------------------------------
//  WARNING:  This function must be located in root memory (the 1st 64K) and
//            must NOT call functions that may be in the memory bank that is
//            used to write to flash.  Near functions (that don't call any far
//            or nearcall functions) are safe because mapFlashAddress()
//            currently does not allow bank 0 to be used for this.  Macros and
//            (if the compiler ever supports them) in-line functions are safe.
//            Anything else (including nearcall functions) are NOT safe to call!
//-----------------------------------------------------------------------------
uchar near _FlashByteSecWrite(uint _4kBlock, uchar *logAddr,
                              uint byte, uint DontErase)
{
#if SECTOR_DEVICE_SUPPORT

  char    xpcBase = _4kBlock - 0x0E;
  char    xpcDev = _4kBlock >> 8;
  char    xpc5555 = xpcDev - (0x0E - 0x05);
  char    xpc2AAA = xpcDev - (0x0E - 0x02);
  char    xpcSave;
  uchar    *log5555 = (uchar *) 0xE555;
  uchar    *log2AAA = (uchar *) 0xEAAA;
  uchar    read;
  uint    loop;


  getXPC(xpcSave);  ipset3();  outi(MBcr, MBval);  DisableWriteProtection();

  //  The body of code goes here!
#error _FlashByteSecWrite is not implemented!

  EnableWriteProtection();  setXPC(xpcSave);  outi(MBcr, MBsVal);  ipres();

  return read != (uchar) byte;
#else
  return 1;
#pragma warn 760
#endif // SECTOR_DEVICE_SUPPORT
}

#pragma warn +760


//-----------------------------------------------------------------------------
//  RAM portion of FlashByteDevWrite.
//
//  _FlashByteDevWrite: Write a byte to a device.  This uses SDP mode
//  which is safer and prevents random/inadvertent writes to flash from
//  going into page write mode.
//
//  SDP Write Byte:   5555H=AAH 2AAAH=55H 5555H=A0H
//
//  SDP Sector erase: 5555H=AAH 2AAAH=55H 5555H=80H
//                    5555H=AAH 2AAAH=55H 5555H=30H
//
//  _4kBlock is the address >> 12 (since we can't do a runtime call in this).
//
//  If erase is non-0, the sector containing the byte to program will be erased.
//
//  This must not contain branching code more than 128 bytes.
//
//  Returns 0 on no errors, or 1 if the byte doesn't match.
//-----------------------------------------------------------------------------
//  WARNING:  This function must be located in root memory (the 1st 64K) and
//            must NOT call functions that may be in the memory bank that is
//            used to write to flash.  Near functions (that don't call any far
//            or nearcall functions) are safe because mapFlashAddress()
//            currently does not allow bank 0 to be used for this.  Macros and
//            (if the compiler ever supports them) in-line functions are safe.
//            Anything else (including nearcall functions) are NOT safe to call!
//-----------------------------------------------------------------------------
uchar near _nostep _FlashByteDevWrite(uint _4kBlock, uchar *logAddr,
                                      uint byte, uint DontErase)
{
  // These values are the same every call for a given 4K block of bytes...
  char  xpcBase = _4kBlock - 0x0E;  // XPC for logAddr
  char  xpcDev = _4kBlock & 0xF0;
  char  xpc5555 = xpcDev - (0x0E - 0x05);
  char  xpc2AAA = xpcDev - (0x0E - 0x02);
  uchar  *log5555 = (uchar *) 0xE555;
  uchar  *log2AAA = (uchar *) 0xEAAA;

  char  xpcSave;
  uchar  read;
  uint  loop;


  getXPC(xpcSave);  ipset3();  outi(MBcr, MBval);  DisableWriteProtection();

  //  Erase sector - 2 steps of SDP
  if (! DontErase)  {       //  Or, if( erase )
    for (loop=2; loop; )  {
      setXPC(xpc5555);  *log5555 = 0xAA;    //  Write SDP
      setXPC(xpc2AAA);  *log2AAA = 0x55;
      if (--loop)  {    //  First pass, write 0x80
        setXPC(xpc5555);  *log5555 = 0x80;  }
      else  {           //  Second pass, write 0x30 to sector
        setXPC(xpcBase);  *logAddr = 0x30;  }
    }
    while ((*logAddr ^ *logAddr) & 0x40); //  Use toggle mode to wait for prev write
  }

  setXPC(xpc5555);  *log5555 = 0xAA;  //  Prep to write the byte
  setXPC(xpc2AAA);  *log2AAA = 0x55;
  setXPC(xpc5555);  *log5555 = 0xA0;

  setXPC(xpcBase);  *logAddr = byte;  // Write the byte

  while ((*logAddr ^ *logAddr) & 0x40);  // Wait for write to finish

  read = *logAddr;

  EnableWriteProtection();  setXPC(xpcSave);  outi(MBcr, MBsVal);  ipres();

  return read != (uchar) byte;
}


//-----------------------------------------------------------------------------
//  WriteFlash: Write a block of data to Flash.  If a sector boundary is
//  crossed, this needs to split on the boundary and tell the ByteWrite
//  RAM routine to erase the sector.
//
//  *** WARNING ***  *** WARNING ***  *** WARNING ***  *** WARNING ***
//
//    All calls should be for a complete sector worth of bytes.  Each sector
//    must be erased before new data can be written, so the caller is
//    responsible for all caching (i.e. read the whole sector, modify it, then
//    rewrite the whole sector).  Anything else will leave a mess!
//
//    The only thing this function does to facilitate this is to erase
//    the sector it is about to write to IF (and ONLY IF) the address
//    of the first byte to be written is the first byte in a sector.
//
//    While the rewriting of a sector COULD be spread across multiple
//    calls, care must be taken to insure the first call is for the
//    address of the first byte of a sector, and that no other task
//    tries to read from or write to the sector until it has been
//    completely rewritten (i.e. it basically needs to be locked
//    by whatever task is modifying it until it is done).
//
//  This returns:
//     0 if no error
//    -1 if a fatal error (didn't detect flash),
//    -2 if the _FlashUnlock ulong wasn't set before the call
//     1 if a programming error.  Calling FlashError() will return the address
//       in flash that didn't verify.
//-----------------------------------------------------------------------------
#pragma offset_labels on
#pragma warn -757

int far WriteFlash(char const far *addr, uchar const far *buffer, ulong count)
{
  FlAddress  writeAddr, endAddr;
  int     usingDevice;      //  Device in device[] we're using.
  FlDeviceInt *pDev;
  uint    sectorMask;
  FlAddress bytes;
  char    wrFailed;
  tWriteFlashByte  writeByteFunc;


  //  Ensure the driver has been copied!
  if (notReady || deviceCount == 0)  return -1;

  //  Store start and end address (+1)
  writeAddr.p = addr;

  //  Check that this write has been unlocked or that the address isn't 0xFFFFFFFF
  if( ~_FlashUnlock != writeAddr.dw || ! (writeAddr.dw+1) )  return -2;

  _FlashUnlock = 0;           //  Reset it
  endAddr.dw = writeAddr.dw + count;

  while (count)  {
    //  Find device in device[] list.  This loops when a write spans more than one
    //  device.  So the tests all use writeAddr since it is updated while programming.
    usingDevice = GetFlashDevice( writeAddr.dw, NULL );

      //  If address not in a detected flash, return
    if (usingDevice == -1)  return -1;

    //  Get the sector boundary test mask and device xpc
    pDev = &device[usingDevice];
    sectorMask = pDev->sectorSize - 1;

    //  Find amount of bytes in addressed device.  This allows writing while
    //  spanning different flash configurations (i.e. different sector sizes).
    if( pDev->end.dw < endAddr.dw )
      bytes.dw = pDev->end.dw - pDev->start.dw;
    else
      bytes.dw = count;

    if (pDev->type == FlBytePrg)
      writeByteFunc = _FlashByteDevWrite;
    else
      writeByteFunc = _FlashByteSecWrite;

  //--- The count is decremented to 0 by the loop, so we update the count of ---
  //--- not-yet written bytes before the loop that actually writes them. ---
    count -= bytes.dw;

  //--- We can only write to one sector of flash at a time, and part of ---
  //--- RAM is often mapped out of the physical address space while we are ---
  //--- doing it, so we insure we can't be preempted while writing a block ---
  //--- of bytes (which also protects the values set by mapFlashAddress, ---
  //--- which are currently global values in root memory to insure they ---
  //--- are accessible while a part of flash is mapped in) ---
    KLOCK();

    mapFlashAddress(writeAddr.dw, pDev->mbxcr, (char far *)writeByteFunc);

    while (bytes.dw)  {
      if (bytes.hl[0]-- == 0)  --bytes.hl[1];

      //  A bit ugly, but significantly more efficient than the compiled
      //  version, in part because it eliminates a call to the shift function.

      //finish:  Changing the loop so it handles only bytes all within a given
      //         4K block will allow me to eliminate an awful lot of redundant
      //         stuff here and in _FlashByteDevWrite()
      //         NOTE:  Spec sheet says MAX 20usec Byte-Program time, so
      //         optimization here could yield very significant results (incl
      //         minimizing interferance with ISRs and reducing latency).

      wrFailed = (writeByteFunc) (
        // MBaddr >> 12  (4k block #)
          _asm( [_HL],
                "ld   hl,(_MBaddr+1)",
                "rr   hl",
                "rr   hl",
                "rr   hl",
                "rr   hl",
                "ld   h,0",
              _HL ),
        // MBaddr & 0xFFF | 0xE000  (logical 64K space Addr)
        (uchar *) _asm( [_HL, _DE, _A],
                "ld   hl,(_MBaddr)",
                "ld   de,0FFFh",
                "and  hl,de",
                "ld   a,0E0h",
                "or   a,h",
                "ld   h,a",
              _HL ),
        // *(uint far *)buffer  (byte to write)
          _asm( [_HL, _A],
                "ld   hl,(ix+.buffer+2)",
                "ld   a,l",
                "ld   hl,(ix+.buffer)",
                "ldp  hl,(hl)",
              _HL ),
        // MBaddr & sectorMask  (DontErase true unless 1st byte of a sector)
          _asm( [_HL, _DE],
                "ld   hl,(_MBaddr)",
                "ex   de,hl",
                "ld   hl,(ix+.sectorMask)",
                "and  hl,de",
              _HL )
      );

      if (wrFailed)  {
        _FlashError = MBaddr;
        count = 0;  //  Terminate outer loop
        break;
      }
      //  This is the fastest way to increment a 32-bit value.
      //  Since the compiler doesn't always have the address
      //  and it might need the value again, it can't generate this.
      _asm( _HL = (unsigned) &MBaddr, // does ++MBaddr
                  "inc (hl)", "jr nz,..x",
        "inc hl", "inc (hl)", "jr nz,..x",
        "inc hl", "inc (hl)", "jr nz,..x",
        "inc hl", "inc (hl)", "..x:" );
      ++buffer;
    }

    KUNLOCK();
  }

  return wrFailed;
}

#pragma warn 757
#pragma offset_labels off


//-----------------------------------------------------------------------------
//  Return _FlashError.  Called after WriteFlash fails to get location of error.
//-----------------------------------------------------------------------------
ulong FlashError( void )
{
  return _FlashError;
}


//-----------------------------------------------------------------------------
//  Fill the next unused entry in the table of flash devices
//-----------------------------------------------------------------------------
static void far fillFlashInfo( ulong address, const far FlDeviceInfo *dev,
                               uchar mbxcr )
{
  FlDeviceInt *flash = &device[deviceCount];

  //  Fill in a device member of the array
  flash->start.dw = address;
  flash->type = dev->type;
  flash->sectorSize = dev->sectorSize;
  flash->deviceSize = dev->deviceSize;
  flash->end.dw = address + dev->deviceSize;
  flash->mbxcr = mbxcr;

  ++deviceCount;
}

//-----------------------------------------------------------------------------
//  RAM portion of DetectFlash.
//
//  This returns the Flash ID and Mfc ID at _4kBlock.
//  (since we can't do a runtime call in this).
//
//  This must not contain branching code more than 128 bytes.
//-----------------------------------------------------------------------------
//  WARNING:  This function must be located in root memory (the 1st 64K) and
//            must NOT call functions that may be in the memory bank that is
//            used to write to flash.  Near functions (that don't call any far
//            or nearcall functions) are safe because mapFlashAddress()
//            currently does not allow bank 0 to be used for this.  Macros and
//            (if the compiler ever supports them) in-line functions are safe.
//            Anything else (including nearcall functions) are NOT safe to call!
//-----------------------------------------------------------------------------
unsigned int _nostep near  _DetectFlashDevice(uint _4kBlock)
{
  char    xpc0000 = _4kBlock - 0x0E;
  char    xpc5555 = _4kBlock - (0x0E - 0x05);
  char    xpc2AAA = _4kBlock - (0x0E - 0x02);
  char    xpcSave;
  uchar    *log5555 = (uchar *) 0xE555;
  uchar    *log2AAA = (uchar *) 0xEAAA;
  uchar    mfcID;
  uchar    flashID;


  getXPC(xpcSave);  ipset3();  outi(MBcr, MBval);

  setXPC(xpc5555);  *log5555 = 0xAA;
  setXPC(xpc2AAA);  *log2AAA = 0x55;
  setXPC(xpc5555);  *log5555 = 0x90;

  setXPC(xpc0000);        //  Atmel says to read 0 and 1
  mfcID = * (uchar *) 0xE000;    //  Read mfcID
  flashID = * (uchar *) 0xE001;  //  Read flashID

  setXPC(xpc5555);  *log5555 = 0xAA;
  setXPC(xpc2AAA);  *log2AAA = 0x55;
  setXPC(xpc5555);  *log5555 = 0xF0;

  setXPC(xpcSave);  outi(MBcr, MBsVal);  ipres();

  return (flashID << 8) | mfcID;
}


//-----------------------------------------------------------------------------
//  Detect flash devices in the system.
//
//  This checks baseAddr looking for a flash ID.  If found,
//  a pointer to flash device info is filled in and 1 is returned.
//  Otherwise 0 is returned.
//  This does no translation for running in Flash versus RAM.
//-----------------------------------------------------------------------------
const far FlDeviceInfo * far DetectFlashDevice(ulong baseAddr, uchar mbxcr)
{
  uchar      mfcID;
  uchar      flashID;
  unsigned int  val;


  KLOCK();

  mapFlashAddress(baseAddr, mbxcr, (char far *)_DetectFlashDevice);
  notReady = 0;           //  Mark Detect Flash called

  val = _DetectFlashDevice((uint)(MBaddr >> 12));

  KUNLOCK();

  flashID = val >> 8;  mfcID = (char) val;

  for (unsigned int idx = 0; idx != numDevices; ++idx)  {
    const FlDeviceInfo far *item = &supportedDevices[ idx ];
    if (item->mfcID == mfcID)  if (item->flashID == flashID)  return item; // Matched
  }

  return 0;
}


//-----------------------------------------------------------------------------
//  Detect all flashes in the system and fill out the device list for the
//  range, type and sector size.
//
//  This ensures we get each MBxCR that is allocated to the 256k halves.
//  Unused device2 and device3 should be set to 0xFFFF.
//
//  Returns the number of devices detected.
//-----------------------------------------------------------------------------
int far DetectFlash( void )
{
  const far FlDeviceInfo *dev;

  deviceCount = 0;

  dev = DetectFlashDevice( 0, _CS0 | _OE0_WE0 );

  if (dev != NULL)  {
    fillFlashInfo( 0, dev, _CS0 | _OE0_WE0 );

    if (device[0].deviceSize == _256K)  {
      dev = DetectFlashDevice( _256K, _CS2 | _OE0_WE0 );
      if( dev != NULL )  fillFlashInfo( _256K, dev, _CS2 | _OE0_WE0 );
    }
    else if (device[0].deviceSize == _512K)  {
      dev = DetectFlashDevice( 0, _CS0 | _OE1_WE1 );
      if( dev != NULL )  fillFlashInfo( 0, dev, _CS0 | _OE1_WE1 );
    }
  }

  notReady = 0;

  return deviceCount;
}


//-----------------------------------------------------------------------------
//  GetFlashDevice.
//
//  If there is are no detected devices, this will detect the
//  device at the passed address.
//
//  Returns -1 if there is no device at the passed address.
//
//  Returns the device index if not -1, and stores the sector
//  size in sectorSize unless this is NULL.
//-----------------------------------------------------------------------------
int far GetFlashDevice(FlDWord address, FlWord *sectorSize)
{
  FlDeviceInt *pDev;
  int     usingDevice;

  //  If we have no devices, detect the one in question
  if (deviceCount == 0)  if (DetectFlash() == 0)  return -1;

  //  Find device in device[] list.  This loops when a write spans more than
  //  one device.  So the tests all use writeAddr since it is updated while
  //  programming.
  for (usingDevice=0; usingDevice!=deviceCount; ++usingDevice)  {
    pDev = &device[ usingDevice ];
    if ((pDev->start.dw <= address) AND (address < pDev->end.dw))  {
      if (sectorSize)  *sectorSize = pDev->sectorSize;
      return usingDevice;
    }
  }

  return -1;
}


//-----------------------------------------------------------------------------
//  GetFlashDeviceEnd.
//
//  If there is are no detected devices, this will detect the
//  device at the passed address.
//
//  Returns 0 if there is no device at the passed address.
//
//  Returns the end address of the passed device + 1, and stores the sector
//  size in the passed pointer unless this is NULL.
//-----------------------------------------------------------------------------
FlDWord far GetFlashDeviceEnd(FlDWord address, FlWord *sectorSize)
{
  FlDeviceInt  *pDev;
  int  usingDevice;

  //  If we have no devices, detect the one in question
  if (deviceCount == 0)  if (DetectFlash() == 0)  return 0;

  //  Find device in device[] list.  This loops when a write spans more than one
  //  device.  So the tests all use writeAddr since it is updated while programming.
  for (usingDevice=0; usingDevice!=deviceCount; ++usingDevice)  {
    pDev = &device[ usingDevice ];
    if( (pDev->start.dw <= address) && (address < pDev->end.dw) )  {
      if (sectorSize)  *sectorSize = pDev->sectorSize;
      return pDev->end.dw;
    }
  }

  return 0;
}


//-----------------------------------------------------------------------------
//  This uses the MMU settings computed by mapFlashAddress() to insure the
//  desired portion of flash is accessable.  The value MBaddr is the absolute
//  address of the first byte to copy.  The toAddr has to point to a buffer
//  that does not cross a quadrant boundary and is in one of the quadrants
//  that remains at the same location and is accessible while the needed
//  portion of flash is also mapped in.
//
//  This function is designed to be copied to and run from anywhere in
//  memory.  The only requirement is that all the following be in quadrants
//  that can all be mapped in at the same time:
//
//     - fromAddr:  The absolute addr where the needed portion of flash will
//       be mapped to
//     - toAddr:  The addr of a buffer to copy the data to
//     - The stack with all the passed argument values
//
//-----------------------------------------------------------------------------
//  WARNING:  This function must be located in root memory (the 1st 64K) and
//            must NOT call functions that may be in the memory bank that is
//            used to write to flash.  Near functions (that don't call any far
//            or nearcall functions) are safe because mapFlashAddress()
//            currently does not allow bank 0 to be used for this.  Macros and
//            (if the compiler ever supports them) in-line functions are safe.
//            Anything else (including nearcall functions) are NOT safe to call!
//-----------------------------------------------------------------------------
#pragma offset_labels on
#pragma warn 760

static void _nostep near _ReadFlash(ulong fromAddr, char *toAddr, uint count)
{
  ipset3();  outi(MBcr, MBval);

  _BCDE = fromAddr;  _HL = (unsigned) toAddr;

#asm
  push  ix               ;; save stack frame
  push  hl               ;; save toAddr
  ld    hl,(ix+.count)   ;; HL = count;
  pop   ix               ;; IX = toAddr;
  ex    de,hl            ;; HL = fromAddr & 0xFFFF; DE = count;
  ld    iy,hl            ;; IY = HL (logical fromAddr);
  ld    de',1            ;; DE' = 1
..lp:                    ;; for (;;)  {
  ld    a,d              ;;   if (count == 0)  break;
  or    a,e              ;;
  jr    z,..done         ;;
  ld    a,c              ;;   A = (fromAaddr & 0xFF0000) >> 16;
  ldp   hl,(iy)          ;;   HL = *fromAaddr;
  ld    (ix),l           ;;   *toAddr = L;
  dec   de               ;;   --count;
  inc   ix               ;;   ++toAddr;
  exx                    ;;   exchange BC, DE, HL
  add   iy,de            ;;   ++fromAddr;
  exx                    ;;   exchange BC, DE, HL
  jr    nc,..lp          ;;   if (!(fromAddr & 0x10000)) continue;
  inc   c                ;;
  jr    ..lp             ;; }
..done:
  pop   ix               ;; restore stack frame
#endasm

  outi(MBcr, MBsVal);  ipres();
}
#pragma warn +760

//-----------------------------------------------------------------------------
//  ReadFlash.
//
//  This reads flash to memory.
//  This is done since the flash could be mapped out if running in RAM.
//-----------------------------------------------------------------------------
void far ReadFlash(char const far *fromAddr, uchar far *buffer, ulong count)
{
  FlDeviceInt *pDev;
  int   usingDevice;      //  Device in device[] we're using.
  ulong  read;
  char readBuff[0x100];


  while( count )  {
    //  Find device in device[] list.  This loops when a write spans more than one
    //  device.  So the tests all use writeAddr since it is updated while programming.
    usingDevice = GetFlashDevice( (ulong) fromAddr, NULL );

      //  If address not in a detected flash, return
    if (usingDevice == -1)  return;

    //  Get the sector boundary test mask and device xpc
    pDev = &device[ usingDevice ];

  // Make sure the range of addresses for the bytes we are about to copy FROM
  // are all in one 256K Memory Bank.  NOTE that this DOES assume that all
  // the bytes on the stack are always within a single 4K aligned block, so
  // readBuff[] will not cross a boundary.  Things would get a LOT messier if
  // we had to deal with the possiblity that BOTH the source and either the
  // destination or the code that does the copy could cross a boundary!

    read = _256K - ((ulong) fromAddr & 0x3FFFFl);  // # bytes to end of bank
    if (read > count)  read = count;  // no more than the total remaining
    if (read > sizeof(readBuff))  read = sizeof(readBuff);  // or size of buf

    KLOCK();

    mapFlashAddress((ulong)fromAddr, pDev->mbxcr, (char far *)_ReadFlash);
    _ReadFlash(MBaddr, readBuff, (uint)read);

    KUNLOCK();

    fmemcpy(buffer, readBuff, read);

    count -= read;  fromAddr += read;  buffer += read;
  }
}


//-----------------------------------------------------------------------------

