//----- hdwDefs.c ----- 04/20/11 -----(07/17/06)-----

//-----------------------------------------------------------------------------
// Definitions common to most hardware platforms, some of which require
// platform-specific values.
//
// This file should only be included in the hardware-specific file that
// provides common functions for a specific hardware platform.  All other
// files should include the <sys/hdwDefs.h> file instead.
//-----------------------------------------------------------------------------
//  Update log (most recent first)
//
//
//  2011-04-20 - Mark Davis
//               - Added getAvgSamplesADC()
//
//  2011-04-15 - Mark Davis
//               - Added resetAvgDataADC()
//
//  2011-04-08 - Mark Davis
//               - Moved functions for tracking and managing averaged and noise
//                 values for analog inputs to the hdwDefs module.
//
//  2011-04-07 - Mark Davis
//               - Added adcNoisePeriod[], adcRawNoise[], adcNoiseTimeMS[] for
//                 tracking noise for each ADC input for configurable periods
//                 of time.
//
//  2010-06-02 - Mark Davis
//               - Updated descriptions for Rabbit 3000A interrupts
//
//  2010-03-31 - Mark Davis
//               - Moved IntVectDesc[] to this file (from usr/inittab.c)
//
//  2009-12-07 - Mark Davis
//               - Changed names for forced ADC/DAC variables and func to
//                 avoid confusion (i.e. is it the ADC/DAC or read/set values
//                 that are affected?)
//               - The flag indicating the analog output needs to be updated
//                 is now set when the forced state of one of the DACs is chgd
//
//  2009-10-27 - Mark Davis
//               - Added readValMin[] and readValMax[]
//
//  2009-05-20 - Mark Davis
//               - added updateAIn[], updateAOut[]
//
//  2008-04-07 - Mark Davis
//               - update hdwcfg param table entries to agree with recent
//                 change that keeps the name of the entry in the table
//                 of info about each entry (but NOT in the file - the file
//                 structure has NOT changed!)
//
//-----------------------------------------------------------------------------


#ifndef _hdwDefsC_
#define _hdwDefsC_


#include <applic.h>


//--- flags to let the applic code know that something about one of the ---
//--- analog inputs or outputs changed and that values need to be recomputed ---
bool  updateAIn[NUM_ADC], updateAOut[NUM_DAC];

//--- benchtop calibration data for each analog input and output ---
//--- restored from a file on startup ---
float  adcGain[NUM_ADC], adcOffset[NUM_ADC];
float  dacGain[NUM_DAC], dacOffset[NUM_DAC];


//--- Computed by converting min/max raw ADC/DAC values to engineering ---
//--- units using calibration data for each channel ---
float   ainMin[NUM_ADC], ainMax[NUM_ADC],
        aoutMin[NUM_DAC], aoutMax[NUM_DAC];
float far  readValMin[NUM_ADC], readValMax[NUM_ADC],
           setValMin[NUM_DAC], setValMax[NUM_DAC];



//--- real (or forced) state for each channel ---
float   anaIn[NUM_ADC], anaOut[NUM_DAC], lastAnaOut[NUM_DAC];
uint   adcRaw[NUM_ADC], dacRaw[NUM_DAC], lastadcRaw[NUM_ADC];
bool  newRdg[NUM_ADC];

//--- keep track of which channels are using forced values ---
bool  adcForced[NUM_ADC], dacForced[NUM_DAC];
float   adcForcedVal[NUM_ADC], dacForcedVal[NUM_DAC];


//--- data for keeping a running average for ADC inputs ---
uint far  adcRawHist[NUM_ADC][ADCHistSize];
ADCAvgData  adcAvgData[NUM_ADC];

ADCNoiseData  adcNoiseData[NUM_ADC];


ulong  ramSize, flashSize;

ulong  digitalIn, digitalOut;
ulong  forcedDigInMask, forcedDigInBits,
       forcedDigOutMask, forcedDigOutBits;
ulong  extraIn;
ulong  forcedExtraInMask, forcedExtraInBits;


int  qdCount;


// The IIR and EIR registers are shifted 8 bits and added to the values below
// to determine the location in root memeory where the code is stored to
// process an interrupt.  Since each "vector" is only 16 bytes long, the
// vectors typically contain little more than a jump or call to an ISR located
// elsewhere.

// The Rabbit 2000 has 16 Internal and 8 External vectors
// The Rabbit 3000 has 32 Internal and 8 External vectors

// Because only the first 2 of the External vectors are used by both
// processors, the External vectors are allowed to overlap the Internal
// ones on the Rabbit 3000 to avoid wasting 256 bytes of root memory.

const char far *IntVectDesc[NumIntVectors] = {
//--- IIR ---
  "Periodic Timer",  // IIR 0x0000
  "Secondary WDT",   // IIR 0x0010  Rabbit 3000A
  "RST 10",          // IIR 0x0020
  "RST 18",          // IIR 0x0030
  "RST 20",          // IIR 0x0040
  "RST 28",          // IIR 0x0050
  "Syscall",         // IIR 0x0060  Rabbit 3000A
  "RST 38",          // IIR 0x0070
  "Slave Port",      // IIR 0x0080
  "Write Protect",   // IIR 0x0090  Rabbit 3000A
  "Timer A",         // IIR 0x00A0
  "Timer B",         // IIR 0x00B0
  "Serial A",        // IIR 0x00C0
  "Serial B",        // IIR 0x00D0
  "Serial C",        // IIR 0x00E0
  "Serial D",        // IIR 0x00F0

//--- IIR/EIR overlapped ---
  "INT0",            // IIR 0x0100 / EIR 0x0000
  "INT1",            // IIR 0x0110 / EIR 0x0010
  "",                // IIR 0x0120 / EIR 0x0020
  "",                // IIR 0x0130 / EIR 0x0030
  "",                // IIR 0x0140 / EIR 0x0040
  "",                // IIR 0x0150 / EIR 0x0050
  "",                // IIR 0x0160 / EIR 0x0060
  "PWM",             // IIR 0x0170 / EIR 0x0070  Rabbit 3000A
  "Mode Violation",  // IIR 0x0180               Rabbit 3000A
  "Quad Decoder",    // IIR 0x0190               Rabbit 3000
  "Input capture",   // IIR 0x01A0               Rabbit 3000
  "Stack Limit",     // IIR 0x01B0               Rabbit 3000A
  "Serial E",        // IIR 0x01C0               Rabbit 3000
  "Serial F"         // IIR 0x01D0               Rabbit 3000
//  "Network Port A",  // IIR 0x01E0               Rabbit 4000
//  "Timer C"          // IIR 0x01F0               Rabbit 4000
};



//--- parameter ID's for data stored in hardware info file ---
//--------------------------------------------------------------------------
// WARNING:  ID's can be added, but existing ID's (and the format of the
//           data the represent) should NEVER change!  NOTE:  Existing ones
//           can be "eliminated" by changing their name and no longer
//           using them, but never REUSE an old one for something new!!!
//--------------------------------------------------------------------------
#define  ADC_GAINS    1
#define  ADC_OFFSETS  2
#define  DAC_GAINS    3
#define  DAC_OFFSETS  4

//NOTE:  Granularity indicates the smallest amount that should be read at
//       once to avoid possible corruption of individual elements

const ParamData far  hdwCfgParams[] = {
  PERSISTENT_TBLN( ADC_GAINS,    float,adcGain,    0,  NULL ),
  PERSISTENT_TBLN( ADC_OFFSETS,  float,adcOffset,  0,  NULL ),
  PERSISTENT_TBLN( DAC_GAINS,    float,dacGain,    0,  NULL ),
  PERSISTENT_TBLN( DAC_OFFSETS,  float,dacOffset,  0,  NULL ),

  END_PARAM_DATA
};

//-----------------------------------------------------------------------------
//  update data structures related to the specified DAC - call after the
//  (bench or field level) Gain and/or Offset have been altered.
//-----------------------------------------------------------------------------
void _nearcall newDACCalib(uint dacIdx)
{
  float  aMin, aMax, sMin, sMax, hold;


  if (dacIdx >= NumHdwDAC)  return;

  aMin = DACtoAOut(dacIdx, (uint) dacMinRaw(dacIdx));
  aMax = DACtoAOut(dacIdx, (uint) dacMaxRaw(dacIdx));

  if (aMin > aMax)  { hold = aMin;  aMin = aMax;  aMax = hold; }

  sMin = AOutToSet(dacIdx, aMin);
  sMax = AOutToSet(dacIdx, aMax);

  if (sMin > sMax)  { hold = sMin;  sMin = sMax;  sMax = hold; }

  KLOCK();

  aoutMin[dacIdx] = aMin;    aoutMax[dacIdx] = aMax;
  setValMin[dacIdx] = sMin;  setValMax[dacIdx] = sMax;

  updateAOut[dacIdx] = true;

  KUNLOCK();
}

//-----------------------------------------------------------------------------
//  update data structures related to the specified ADC - call after the
//  (bench or field level) Gain and/or Offset have been altered.
//-----------------------------------------------------------------------------
void _nearcall newADCCalib(uint adcIdx)
{
  float  aMin, aMax, rMin, rMax, hold;


  if (adcIdx >= NumHdwADC)  return;

  aMin = ADCtoAIn(adcIdx, (uint) adcMinRaw(adcIdx));
  aMax = ADCtoAIn(adcIdx, (uint) adcMaxRaw(adcIdx));

  if (aMin > aMax)  { hold = aMin;  aMin = aMax;  aMax = hold; }

  rMin = AInToRead(adcIdx, aMin);
  rMax = AInToRead(adcIdx, aMax);

  if (rMin > rMax)  { hold = rMin;  rMin = rMax;  rMax = hold; }

  KLOCK();

  ainMin[adcIdx] = aMin;  ainMax[adcIdx] = aMax;
  readValMin[adcIdx] = rMin;  readValMax[adcIdx] = rMax;

  updateAIn[adcIdx] = true;

  KUNLOCK();
}

//-----------------------------------------------------------------------------
//  Set or Clear a forced value for a DAC
//-----------------------------------------------------------------------------
void _nearcall setStateDAC(uint uidx, float val, bool clearForced)
{
  if (clearForced)
    dacForced[uidx] = false;
  else  {
    dacForced[uidx] = true;  dacForcedVal[uidx] = val; }

  updateAOut[uidx] = true;
}


//-----------------------------------------------------------------------------
//  Get initial state of input signals and initialize values used by rwIO()
//  to insure the DAC's will be written to during the next pass through the
//  application loop.
//
//  This function is to be used ONLY at startup!  It is NOT designed to be
//  called at any other time!
//-----------------------------------------------------------------------------
void initIOvalues(void)
{
  uint  ui, rawVal;
  ADCAvgData  *ad;
  ADCNoiseData  *nd;
//  char  temp[32];  //tdebug


//--- read digital and raw analog input values ---
  readAnalogInputs(ADC_Mask);
  digitalIn = ReadDigInputs();

//--- initialize the analog input values ---
  for (ui=0; ui<NumHdwADC; ++ui)  {
    ad = adcAvgData + ui;  nd = adcNoiseData + ui;
    rawVal = adcRaw[ui];

  //--- scale the raw values to volts and application engineering units ---
    newRdg[ui] = true;
    lastadcRaw[ui] = nd->rawMin = nd->rawMax = rawVal;
    anaIn[ui] = ADCtoAIn(ui, rawVal);
    if (ui < NumApplAIN)  readVal[ui] = AInToRead(ui, anaIn[ui]);

  //--- Initialize the data for handling averaging of the ADC readings ---
  //--- NOTE that the period set for averaged values applies ONLY to the ---
  //--- values returned by getAvgdAnaIn().  The anaIn[] values are NOT ---
  //--- affected by the averaging settings! ---
    ad->sum = adcRawHist[ui][0] = rawVal;
    ad->oldest = 0;  ad->next = 1;
    ad->curNumSamples = 1;  setAvgPeriodADC(ui, 200);

  //--- Initialize the data for tracking the noise level of ADC readings ---
    setNoisePeriodADC(ui, 2000);


  //tdebug
//    printf(" %2u: adcRaw: 0x%04X,  anaIn: %.16s\r\n",
//           ui, adcRaw[ui], Engineering(anaIn[ui], 6, temp) );

  }

//--- set values that will force DAC's to be written to in 1st pass ---
  for (ui=0; ui<NumHdwDAC; ++ui)  lastAnaOut[ui] = -9999.9;
}

//-----------------------------------------------------------------------------
//  Reset the averaging data for the specified analog input
//-----------------------------------------------------------------------------
void resetAvgDataADC(uint idx)
{
  ADCAvgData  *ad = adcAvgData + idx;

  if (idx >= NumHdwADC)  return;

  KLOCK();

  ad->sum = adcRawHist[idx][0] = adcRaw[idx];
  ad->oldest = 0;  ad->next = ad->curNumSamples = 1;

  KUNLOCK();
}

//-----------------------------------------------------------------------------
//  Set the period of time over which to compute AVERAGED versions of the ADC
//  values (i.e. values returned by getAvgdRawIn() and getAvgAnaIn())
//-----------------------------------------------------------------------------
uint setAvgPeriodADC(uint idx, uint numMS)
{
  uint  numSamples;

  if (idx >= NumHdwADC)  return 1;

  numSamples = (numMS + ADCSampleInterval/2) / ADCSampleInterval;

  if (numSamples < 1)  numSamples = 1;
  if (numSamples > ADCHistSize)  numSamples = ADCHistSize;

  _EnterCritical_();
  adcAvgData[idx].useNumSamples = numSamples;
  _ExitCritical_();

  return numSamples * ADCSampleInterval;
}

//-----------------------------------------------------------------------------
//  Return the period being used to compute AVERAGED versions of the ADC
//  values (i.e. values returned by getAvgdRawIn() and getAvgAnaIn())
//-----------------------------------------------------------------------------
uint getAvgPeriodADC(uint idx)
{
  uint  numSamples;

  if (idx >= NumHdwADC)  return  1;

  _EnterCritical_();
  numSamples = adcAvgData[idx].useNumSamples;
  _ExitCritical_();

  return numSamples * ADCSampleInterval;
}


//-----------------------------------------------------------------------------
//  Return the # of samples used to compute AVERAGED versions of the ADC
//  values (i.e. values returned by getAvgdRawIn() and getAvgAnaIn())
//-----------------------------------------------------------------------------
uint getAvgSamplesADC(uint idx)
{
  uint  numSamples;

  if (idx >= NumHdwADC)  return  1;

  _EnterCritical_();
  numSamples = adcAvgData[idx].useNumSamples;
  _ExitCritical_();

  return numSamples;
}


//-----------------------------------------------------------------------------
//  Return the averaged reading for the specified analog input in ADC units
//  (computed in floating point to minimize loss of accuracy)
//-----------------------------------------------------------------------------
float getAvgdRawIn(uint idx)
{
  ulong  sampleSum;
  uint  numSamples;
  ADCAvgData  *ad;


  if (idx >= NumHdwADC)  return  0.0;

  ad = adcAvgData + idx;

  _EnterCritical_();
  sampleSum = ad->sum;  numSamples = ad->curNumSamples;
  _ExitCritical_();

  return  (float)sampleSum / numSamples;
}


//-----------------------------------------------------------------------------
//  Return the averaged reading for the specified analog input
//-----------------------------------------------------------------------------
float getAvgdAnaIn(uint idx)
{
  ulong  sampleSum;
  uint  numSamples;
  ADCAvgData  *ad;


  if (idx >= NumHdwADC)  return  0.0;

  if (adcForced[idx])  return adcForcedVal[idx];

  ad = adcAvgData + idx;

  _EnterCritical_();
  sampleSum = ad->sum;  numSamples = ad->curNumSamples;
  _ExitCritical_();

  return  ADCtoAIn(idx, (float)sampleSum / numSamples);
}


//-----------------------------------------------------------------------------
//  Update the running average for each raw ADC input value
//  Must ONLY be called by the applic task!
//
//  NOTE: Separate indexes/etc are used for each input so the sample interval
//        can be set differently for each input
//-----------------------------------------------------------------------------
void updateAvgsADC(void)
{
  uint  u, raw, curTime;
  uint far  *histPtr;
  ADCAvgData  *ad;
  ADCNoiseData  *nd;

  curTime = (uint) GetMS();

  for (u=0; u<NumHdwADC; ++u)  {
    histPtr = adcRawHist[u];  ad = adcAvgData + u;  nd = adcNoiseData + u;

  //--- reduce # of values making up the running average until it ---
  //--- contains 1 less then the specified # of values ---
    while (ad->curNumSamples >= ad->useNumSamples)  {
      ad->sum -= histPtr[ad->oldest];
      if (++ad->oldest == ADCHistSize)  ad->oldest = 0;
      --ad->curNumSamples;
    }

  //--- add current sample to the sum and insert it at the end of the ---
  //--- circular queue of recent values ---
    raw = adcRaw[u];
    ad->sum += (ulong) (histPtr[ad->next] = raw);
    if (++ad->next == ADCHistSize)  ad->next = 0;
    ++ad->curNumSamples;

  //--- update noise data ---
    if (raw < nd->rawMin)  nd->rawMin = raw;
    if (raw > nd->rawMax)  nd->rawMax = raw;

    if ((curTime - nd->noiseTimeMS) >= nd->noisePeriod)  {
      nd->rawNoise = nd->rawMax - nd->rawMin;
      nd->rawMin = nd->rawMax = raw;
      nd->noiseTimeMS = curTime;
    }
  }

}

//-----------------------------------------------------------------------------
//  Set the interval of time used for tracking noise levels on an analog input
//-----------------------------------------------------------------------------
uint setNoisePeriodADC(uint ui, uint numMS)
{
  if (ui >= NumHdwADC)  return 0;

  _EnterCritical_();  adcNoiseData[ui].noisePeriod = numMS;  _ExitCritical_();

  return numMS;
}

//-----------------------------------------------------------------------------
//  Get the interval of time used for tracking noise levels on an analog input
//-----------------------------------------------------------------------------
uint getNoisePeriodADC(uint ui)
{
  uint  numMS;

  if (ui >= NumHdwADC)  return 0;

  _EnterCritical_();  numMS = adcNoiseData[ui].noisePeriod;  _ExitCritical_();

  return numMS;
}

//-----------------------------------------------------------------------------
//  Get the "noise" (the difference between the min and max raw values) for
//  the specified ADC as recorded for the previous noise period.
//
//  If not NULL, the ms16 argument must point to a 16-bit unsigned integer
//  where this function will save a copy of the end time of the period the
//  noise value was computed for.  NOTE that this is only the least significant
//  16 bits of the MS time value returned by GetMS().
//-----------------------------------------------------------------------------
uint getNoiseADC(uint ui, uint *ms16)
{
  uint  noise;
  ADCNoiseData  *nd;


  if (ui >= NumHdwADC)  return 0;

  nd = adcNoiseData + ui;

  _EnterCritical_();
  noise = nd->rawNoise;  if (ms16)  *ms16 = nd->noiseTimeMS;
  _ExitCritical_();

  return noise;
}

//-----------------------------------------------------------------------------
//  restart the interval for tracking the noise on an analog input signal
//-----------------------------------------------------------------------------
void resetNoiseDataADC(uint ui)
{
  ADCNoiseData  *nd;


  if (ui >= NumHdwADC)  return;

  nd = adcNoiseData + ui;

  KLOCK();

  nd->rawMin = nd->rawMax = adcRaw[ui];
  nd->noiseTimeMS = (uint) GetMS();

  KUNLOCK();
}


//-----------------------------------------------------------------------------

#endif
