//----- flashFile.c ----- 02/08/12 -----(07/13/06)-----

//-----------------------------------------------------------------------------
//  Standard DEBROS virtual file/device I/O API wrapper for the SHDesigns-based
//  Flash File system API.
//-----------------------------------------------------------------------------
//  Update log (most recent first)
//
//
//  2012-02-08 - Mark Davis
//               - errno is now set to EBADF rather than ENXIO when a function
//                 is passed an invalid device pointer.
//
//  2010-07-13 - Mark Davis
//               - Added support for FIONSPACE request for ioctl() calls.
//                 This allows poll() to work properly if called for flash
//                 files.
//
//  2010-04-01 - Mark Davis
//               - _flashFileFstat_() no longer changes the value of errno if
//                 ffile_fstat() returns a non-0 value.  This avoids a
//                 confusing errno value (and message from perror() if used)
//                 when ffile_fstat() detects a problem.
//
//  2010-03-23 - Mark Davis
//               - Added locks around call to ffile_info() in _flashFileDescr_()
//                 to eliminate error msgs and minimize risk of concurrency
//                 problems when formatting the description for a file.  NOTE
//                 that the concurrency issue is not critical as this is for
//                 display purposes only.
//
//  2009-06-18 - Mark Davis
//               - The open func now recognizes the O_APPEND flag.
//
//  2009-03-26 - Mark Davis
//               - Added unlink and rename functions
//
//  2009-03-24 - Mark Davis
//               - added _flashFileEntry_() function for getting info about
//                 flash file driver fds
//
//  2008-04-07 - Mark Davis
//               - changes to support dynamically registered file systems
//
//-----------------------------------------------------------------------------

#include <common.h>

#include <unistd.h>
#include <fcntl.h>
#include <sys/stat.h>

#include <errno.h>

#include <sys/flashFile.h>

#include <Rabbit.h>
#include <dcdefs.h>


#include <DLMIII.h>
#include <SHD_DLM.h>


#include <sys/farprintf.h>


IODEV_FUNCS(flashFile)

FILESYS_FUNCS(flashFile)

//typedef  int (*GetCFunc)(void);
//typedef  int (*PutCFunc)(int ch);


// The file_io.c device layer has its own table of open files and we don't
// (at least currently) need to add any functionality to it so no table is
// needed at this layer.
//
// static int ffid[MAX_FILES];


#define FFILE     ((FILE_DATA *)ioDev)
#define FFHandle(ff)  ((int)((ff) - openFiles))

//-----------------------------------------------------------------------------
//  func called once by init process to initialize data structures
//-----------------------------------------------------------------------------
void flashFileInit_()
{
//  uchar  i;

//  for (i=0; i<MAX_FILES; ++i)  ffid[i] = -1;

//  flashFileFSysFuncs.tick = NULL;

  registerFileSys(&flashFileFSysFuncs);
}


//-----------------------------------------------------------------------------
//  Returns true if provided pointer is a valid, in-use FlashFile device.
//  Returns false if is not.
//-----------------------------------------------------------------------------
bool  validFFILE(void *ioDev)
{
  if (ioDev)
    if (FFILE == openFiles + (int)(FFILE - openFiles))
      if (FFILE->flags & FFLAG_OPEN)  return true;

  return false;
}

//-----------------------------------------------------------------------------
//  Open file in flash and return a file descriptor for it
//
//  Should only be called from the fcntl.c layer!
//-----------------------------------------------------------------------------
fd_t _flashFileOpen_(const char *path, int flags, int mode)
{
  fd_t  fd;
  int  handle;  // flash file handle
  FILE_DATA  *ioDev;


  KLOCK();

//--- try to open file - create it if asked to and it doesn't exist yet ---
  handle = ffile_open((char *)path);
  if (handle == FERR_NOTFOUND)  {
    if (! (flags & O_CREAT))  { KUNLOCK();  ERRNO = ENOENT;  return -1; }
    handle = ffile_create((char *)path);
  }
  if (handle < 0)  {
    KUNLOCK();
    switch (handle)  {
      case FERR_TOOMANYFILES:  ERRNO = ENFILE;  break;
      case FERR_NOXMEM:  ERRNO = ENOMEM;  break;
    }
    return -1;
  }

  ioDev = openFiles + handle;

//--- allocate a file descriptor to reference the Flash File data ---
  if ((fd = _allocFile_(3, &flashFileIOFuncs, ioDev)) < 0)  {
    ffile_close(handle);  KUNLOCK();  return -1;  }

  KUNLOCK();

  if (flags & O_APPEND)  _flashFileLseek_(ioDev, 0, FSEEK_END);

//--- everything OK - return the descriptor assigned for this process ---
  return fd;
}


//-----------------------------------------------------------------------------
//  func called each kernel tick to provide file system with CPU cycles
//  Should ONLY be called by the kernel task!
//-----------------------------------------------------------------------------
void _flashFileTick_(void)
{

}


//-----------------------------------------------------------------------------
//  Return the state of the specified ioDev entry or NULL if not a valid
//  entry #.
//
//  For valid entrys, the entryState struct is filled in
//-----------------------------------------------------------------------------
void *_flashFileEntry_(uint idx, EntryState *entryState)
{
  FILE_DATA  *fileData;

  if (idx >= DOPEN_FILES)  return NULL;

  fileData = openFiles + idx;

  entryState->openCount = ((fileData->flags & FFLAG_OPEN) ? 1 : 0);
  entryState->unused = (entryState->openCount == 0);

  return fileData;
}


//-----------------------------------------------------------------------------
//  Delete a file
//
//  To be called ONLY by fcntl functions
//-----------------------------------------------------------------------------
int _flashFileUnlink_(char *path)
{
  int  stat;

  KLOCK();  stat = ffile_delete(path);  KUNLOCK();

  return stat;
}


//-----------------------------------------------------------------------------
//  Change the name of a file
//
//  To be called ONLY by fcntl functions
//-----------------------------------------------------------------------------
int _flashFileRename_(const char *oldPath, const char *newPath)
{
  int  stat;

  KLOCK();  stat = ffile_rename_(oldPath, newPath);  KUNLOCK();

  return stat;
}




//=============================================================================
//  IODev interface functions (for use by fcntl module)
//=============================================================================

//-----------------------------------------------------------------------------
//  Return wait queues to by used by fcntl layer when blocking on I/O calls.
//
//  Should only be called from the fcntl.c layer!
//-----------------------------------------------------------------------------
int _flashFileGetQueues_(void *ioDev, WaitQueues far *queues)
{
  queues->readQueue = queues->writeQueue = &sleepQueue;
  return 0;
}

//-----------------------------------------------------------------------------
//  Attempt to read bytes from a file in flash
//
//  Should only be called from the fcntl.c/read() function!
//
//  RETURN VALUEs:
//  - if bytes read: # of bytes read
//  - if no bytes read:
//    - if more possible in the future: -1 and ERRNO == EAGAIN
//    - if no more possible (EOF):  0
//-----------------------------------------------------------------------------
int _flashFileRead_(void *ioDev, char *buf, int numBytes, bool dontBlock)
{
  int  stat, handle;


  KLOCK();

  if (! validFFILE(ioDev))  { KUNLOCK();  ERRNO = EBADF;  return -1; }

  handle = FFHandle(FFILE);

  stat = ffile_read(handle, (char far *)buf, numBytes);

  KUNLOCK();  return stat;
}

//-----------------------------------------------------------------------------
//  Attempt to write bytes to file in flash
//
//  Should only be called from the fcntl.c/write() function!
//
//  RETURN VALUEs:
//  - if bytes written: # of bytes written
//  - if no bytes written:
//    - if more possible in the future: -1 and ERRNO == EAGAIN
//    - if no more possible (EOF):  0
//-----------------------------------------------------------------------------
int _flashFileWrite_(void *ioDev, char *buf, int numBytes, bool dontBlock)
{
  int  stat, handle;


  KLOCK();

  if (! validFFILE(ioDev))  { KUNLOCK();  ERRNO = EBADF;  return -1; }

  handle = FFHandle(FFILE);

  stat = ffile_write(handle, (char far *)buf, numBytes);

  KUNLOCK();  return stat;
}


//-----------------------------------------------------------------------------
//  close a flashFile object
//
//  Should only be called from the fcntl.c layer!
//-----------------------------------------------------------------------------
int _flashFileClose_(void *ioDev)
{
  KLOCK();

  if (! validFFILE(ioDev))  { KUNLOCK();  ERRNO = EBADF;  return -1; }

  ffile_close( FFHandle(FFILE) );

  KUNLOCK();

  return 0;
}

//-----------------------------------------------------------------------------
//  Kernel-space function to reposition a file offset
//
//  Should only be called from the fcntl.c layer!
//-----------------------------------------------------------------------------
off_t  _flashFileLseek_(void *ioDev, off_t offset, int whence)
{
  int  stat, handle;
  off_t  newOffset;


  KLOCK();

  if (! validFFILE(ioDev))  { KUNLOCK();  ERRNO = EBADF;  return -1; }

  handle = FFHandle(FFILE);

  stat = ffile_seek(handle, (long)offset, whence);

  if (stat != 0)  { KUNLOCK();  ERRNO = EINVAL;  return (off_t)-1; }

  newOffset = (off_t) ffile_tell(handle);

  KUNLOCK();

  return newOffset;
}

//-----------------------------------------------------------------------------
//  Kernel-space function to get the status of a flash file
//
//  Should only be called from the fcntl.c layer!
//-----------------------------------------------------------------------------
int  _flashFileFstat_(void *ioDev, struct stat *fsbuf)
{
  int  stat;


  KLOCK();

  if (! validFFILE(ioDev))  { KUNLOCK();  ERRNO = EBADF;  return -1; }

  stat = ffile_fstat(FFHandle(FFILE), fsbuf);

  KUNLOCK();

  return stat;
}


//-----------------------------------------------------------------------------
//  Kernel-space function to handle ioctl calls for flashFile ports
//
//  Should only be called from the fcntl.c layer!
//-----------------------------------------------------------------------------
int _flashFileioctl_(void *ioDev, int request, void *arg)
{
  int  stat;
  long  unread, freeSpace;


  KLOCK();

  if (! validFFILE(ioDev))  { KUNLOCK();  errno = EBADF;  return -1; }

  stat = -1;

  switch (request)  {

    case FIONREAD:
      unread = ffile_unread(FFHandle(FFILE));
      if (unread > 0x7fff)  unread = 0x7fff;
      if (unread < 0)  break;
      *((int *)arg) = (int) unread;  stat = 0;  break;

    case FIONSPACE:
      freeSpace = ffile_free();
      if (freeSpace > 0x7fff)  freeSpace = 0x7fff;
      if (freeSpace < 0)  break;
      *((int *)arg) = (int) freeSpace;  stat = 0;  break;

    default:
      errno = EINVAL;
  }

  KUNLOCK();  return stat;
}


//-----------------------------------------------------------------------------
//  Kernel-space function to return a description of the specified file
//
//  Should only be called from the fcntl.c layer!
//-----------------------------------------------------------------------------
char *_flashFileDescr_(void *ioDev, char *buf, int bufSize)
{
  FILE_ENTRY far *fe;
  char  *ptr;

  sprintf(buf, "flash 0x%04lX ", (ulong)ioDev);

  ptr = buf + strlen(buf);  strcpy(ptr, "<bad> <bad>");

//finish: need to limit output to bufSize...

  if (validFFILE(FFILE))  {
    sprintf(ptr, "flash[%d] ", FFHandle(FFILE));  while (*ptr) ++ptr;
    KLOCK();
    fe = ffile_info(openFiles[FFHandle(FFILE)].fatIdx);
    KUNLOCK();
    if (fe != NULL)
      sprintf(ptr, "%.32ls", fe->name);
    else
      strcpy(ptr, "<bad>");
  }

  return buf;
}


//-----------------------------------------------------------------------------

