//----- srfs.c ----- 02/08/12 -----(07/31/06)-----

/*-----------------------------------------------------------------------------
   Client interface for Simple Remote File System (SRFS), an inefficient, but
   dirt simple protocol for accessing files remotely over a TCP connection.
  -----------------------------------------------------------------------------

===============================================================================
  For details on how this driver works, see detailed description following
  the Update log.
===============================================================================

 Update log (most recent first):


  2012-02-08 - Mark Davis
               - errno is now set to EBADF rather than ENXIO when a function
                 is passed an invalid device pointer.

  2011-10-25 - Mark Davis
               - The BYE command is now sent just before closing the shared
                 socket connection if no error has occured on the connection.

  2011-03-23 - Mark Davis
               - After MANY interruptions for other projects, finally a
                 clean, stable version of the single-socket changes.

  2010-10-25 - Mark Davis
               - Began work on changes to use a single socket connection to the
                 SRFS server, shared by all tasks and protected by a semaphore.

  2010-02-05 - Mark Davis (*)
               - Added fatalError flag and set it if the conversation with the
                 SRFS server fails at any point.  Since failing to read or
                 write the expected # of bytes means the conversation is
                 basically out of sync, this means the connection is no
                 longer usable, so an EIO error is returned for the call that
                 had the problem and all subsequent calls.  Prior to this fix,
                 the results of a failed read or write on the socket connection
                 could mean all kinds of bizarre behavior following the error.
               - remfd now initialized to -1 and checked in the close call
                 to avoid sending a close command for a remote file that was
                 never opened.  This gets rid of all the "Bad file number"
                 errors in the srfsd.log on the server.


  2010-01-22 - Mark Davis (*)
               - Changed all read()/write() calls to readBytes()/writeBytes()
                 with a timeout.

  2009-04-15 - Mark Davis
               - Fixed unlink() so it works again (the pathnames it is called
                 with now include the "/srfs/" prefix, which I believe where
                 removed by an earlier version of fcntl.c

  2009-04-13 - Mark Davis
               - Fixed rename() so it works again (the pathnames it is called
                 with now includes the "/srfs/" prefix, which I believe was
                 removed by an earlier version of fcntl.c

  2009-03-24 - Mark Davis
               - added _srfsEntry_() function for getting info about SRFS fds

  2007-10-05 - Mark Davis
               - added support for unlink() and lseek() calls

===============================================================================
 Calls made by the client:

  int  srfsOpen(const char *path, int oflag, int mode);
  int  srfsClose(int fd);
  int  srfsRead(int fd, void *buf, int nbyte);
  int  srfsWrite(int fd, void *buf, int nbyte);
  long srfsLseek(int fd, long offset, int whence);
  int  srfsUnlink(const char *pathname);
  int  srfsRename(const char *oldName, const char *newName);

  The protocol is kept very simple:

  - All parameters are sent in network (Most Significant 1st) byte order

  - Each request from a client consists of a 2-byte Func #, a 2-byte Size
    value, then Size # of bytes (if any):
    - 2 byte func #
    - 2 byte size
    - size # bytes

  - Each response from the server starts with 6 result bytes:
    - 2 byte return value
    - 2 byte errno value
    - 2 byte data len
    - data len # bytes

    The 6 result bytes are followed by "len" bytes of data (if any)

    For the lseek function, the 4-byte return value is in the data bytes

===============================================================================
 Notes on how this driver operates (newer, single shared socket approach):

    The newer version of this driver now uses a single socket connection
    shared by all tasks.

    This eliminates several of the issues with the older one socket per open
    file approach:

      - Only one TCP socket resource is required, regardless of the number
        of files open on the server.

      - Opening a remote file does not require 2 free process fds (the older
        version used an additional one for the socket connection that
        applied only to a single open file)

      - Connections are not constantly being opened and closed, so the server
        is no longer stuck with hundreds of old closed connections in the
        one of the XXX_WAIT states.


 One connection to rule them all:

    Overview of how the revamped SRFS driver works:

       - A single connection is used by all tasks to communicate with the
         server.

       - The shared connection to the SRFS server is opened whenever a
         task needs to open a file on the server and there isn't currently
         an open connection to the server.

       - Only ONE task can be using the connection at a time.  Once
         started, a read/write/etc call must complete before another task
         can use the connection.

       - A semaphore is used as a global SRFS driver lock to insure that
         only one task is trying to use the shared connection at a time.

       - Any error on the shared connection means we are probably out of
         sync with the server, so the connection is closed and all files
         currently open by the SRFS driver are marked as bad.  (While it
         would be possible to modify the protocol so it could - at least
         in theory - recover and resync, the liklihood is that the
         occurance of an error that the TCP/IP layers could not recover
         from means the connection is almost certainly no longer usable).


   A more detailed description of how all this is done:

      The first time a file on the server is opened, the SRFS driver
      will open a connection to the server and leave it open.  A pointer
      to the FDEV structure (effectively a global fd) for the connection
      is saved in sockFDEV and used by all tasks.

      Calls to read(), write(), etc for remote files requires sending
      one or more messages to the server and receiving one or more
      messages in return.  To insure that messages (or even portions
      of messages) for different files are not mixed together, a global
      lock is used to insure only one operation for one task is using
      the connection at any time.

      So, each read/write/etc does this:

        - Get the global SRFS lock (or return an error if it can't
          get the lock)

        - Send the request and any data for the request to the
          server.

        - Read the response from the server (often 2 reads: one to
          get back a header with the status, and one to read any
          data that is part of the response).

        - Release the global SRFS lock so other tasks can use the
          connection.

      If an error occurs when writing to or reading from the shared
      connection, then there is a good chance that the client and server
      are no longer in sync.

        Remember that "successful" writes to an fd do NOT guarantee that
        the data made it all the way to its final destination - only that
        the driver accepted the data.

        Even getting a response back from the server for a write operation
        does not mean that the data was actually written to the file.  For
        all we know the server could  have crashed before it could update
        the disk with data that was accepted, acknowledged, and cached.

        This means that if any error occurs on the shared connection, all
        the fds for files that were using it are now unusable.

      So, whenever an error occurs, all the currently open SRFS files
      are marked as bad and the connection is closed.  Any tasks blocked
      on a pending operation will eventually get the global lock, find
      that the connection they were using was closed, and return an
      error.

      As with any error of this kind, recovery is up to the tasks that
      were using the files.  Only they know the context in which the
      file was being used and how to recover from the error.  Typical
      responses include trying again, or simply reporting the error
      and exiting.


   Sharing a connection across tasks:

      Opening the shared connection is currently done using the standard
      BSD-style socket library calls, which results in a process-specific
      fd being allocated for the task that opened the connection.

      Since this fd is usable only by the task that opened the connection,
      it is simply freed up, leaving the FDEV structure that it pointed to
      (and the socket driver-level structures IT points to).  It is the
      FDEV structure that is used by the SRFS driver functions to read and
      write to the shared connection.

      One consequence of freeing up the process-specific fd is that the
      FDEV structure it pointed to is now effectively an "orphan" (i.e.
      there are no direct references to it by any of the process level
      fds).  To avoid any cleanup or scanning routines from interpreting
      this as an error, a new flag was added to the FDEV structure that
      can be set to indicate that the FDEV is SUPPOSED to be an orphan.

      One of the advantages of this approach is that the only part of
      the code that is dependent on what type of connection is used is
      the srfsConnect function.  Everything else uses the generic API
      that the fcntl layer uses for all the drivers.  This means it
      would be fairly simple to add support for different connection
      types.

      Another advantage is that, with no pfd references to it, the
      shared FDEV will not be closed even when all the SRFS files have
      been closed.  This means that, unless an error occurs, the
      connection to the server should always be up and ready after it
      is first opened.


   Blocking - The fcntl layer vs the driver layer:

      From the point of view of the fcntl layer, drivers come in 2
      varieties:  Those that handle blocking on their own and those
      that don't.

      For the later group, the driver's read and write functions
      never block:  They simply read or write however bytes they can
      each time they are called.  It is up to the fcntl layer to handle
      the difference between blocking and non-blocking calls.

      For non-blocking reads and writes, the fcntl layer simply
      returns whatever the driver supplies.

      For blocking calls, it will call the driver's read or write
      function repeatedly until the specified # of bytes are processed
      or an error occurs.  Between calls to the driver, the task is
      suspended using the specified wait queue.  The more intelligent
      drivers provide their own wait queues which allows them to wake
      up tasks as soon as (and only when) there is a good chance the
      driver can process more bytes.  This improves response time and
      minimizes the number of CPU cycles wasted calling the driver
      when it cannot process more bytes.


      The SRFS driver is one of the ones that handles blocking on its
      own.  The only difference between the 2 modes is whether or not
      the task blocks if the global SRFS lock is in use.  After that,
      the two modes are the same, because once a conversation with the
      server is started, it has to be finished.


  Locking and wait queues:

     As noted above, a lock is used to insure that only one task at a time
     is using the shared connection.  To avoid the possibility that packets
     from different conversations can get mixed together, once a task owns
     the lock, it must not release it until it finishes a conversation.
     This means it sent a command and received a complete response, or an
     error occured (in which case no more conversations are allowed on the
     connection).

     The lock for the shared connection is also used to protect most of the
     global values used in this module.  Because the locking mechanism used
     is a semaphore, when a task releases the lock, the next task waiting
     for it (if any) will be woken up and given the semaphore.


  Open Questions:

      The blocking/non-blocking mode currently is NOT passed on to the
      server, so it opens all files in blocking mode.  But even so,
      is it possible that the server would NOT read or write the
      specified number of bytes (except when EOF is reached for a
      read, or a signal/etc interrupted the process)?

      If this CAN happen, then I the SRFS driver would need to
      return wait queues (even if just the sleepQueue) so the fcntl
      layer would do its thing (e.g. read, suspend, read, suspend, ...)
      to make sure all the bytes are processed for blocking calls.


=============================================================================*/


#include <common.h>

#include <unistd.h>
#include <sys/stat.h>
#include <sys/socket.h>
#include <fcntl.h>
#include <arpa/inet.h>
#include <semaphore.h>
#include <sys/time.h>

#include <errno.h>

#include <sys/farprintf.h>

#include <services.h>

#include <linux/limits.h>



//#define  SRFS_DEBUG


typedef struct {
  bool      inUse;     // non-0 if this entry is in use
  int       errNum;    // non-0 means connection failed
  int       remfd;     // server's descriptor for an open file
  char far *name;      // name of remote file
} SRFS_File;


char far srfsFileName[MaxSRFSFiles][NAME_MAX];


int sendCmd(const char *cmd, int cmdLen, const char *data, int dataLen);


IODEV_FUNCS(srfs)

FILESYS_FUNCS(srfs)


static SRFS_File srfsFile[MaxSRFSFiles];


#define SRFS  ((SRFS_File *)ioDev)

//--- request IDs for SRFS function "calls" ---
#define SRFS_KEEPALIVE  0
#define SRFS_OPEN       1
#define SRFS_CLOSE      2
#define SRFS_READ       3
#define SRFS_WRITE      4
#define SRFS_LSEEK      5
#define SRFS_FSTAT      6
#define SRFS_UNLINK     7
#define SRFS_RENAME     8
#define SRFS_BYE       99  // close connection to server


#define  ShowProcess  (DiagFlags & 0x00000001)

ulong  srfsIPAddr, srfsAltAddr;

#define  DEF_SRFS_ADDR   "192.168.4.34"  // icarus
#define  DEF_ALT_ADDR    "192.168.4.33"  // epics3rd


#define  MAX_RW_MS     10000   // max ms for r/w opers on shared connection


sem_t srfsSem;          // semaphore used to protect SRFS driver's shared data
#define  SRFS_LOCK(secs)  ( srfsGetLock(secs) )
#ifdef SRFS_DEBUG
  #define  SRFS_UNLOCK()  ( srfsUnlock() )
#else
  #define  SRFS_UNLOCK()  ( sem_post(&srfsSem) )
#endif


FDEV far *sockFDEV;     // global ftable entry for connection to srfs server
ulong  lastSendTime;    // time of last xmit on sockFDEV
bool  srfsConOK;

struct sockaddr_in  srvAddr;  // address of server

#define  LogErr(str)  \
   fprintf(DiagLog, "%s(): %.32s: %ls\r\n", __func__, (str), strerror(errNum))

//-----------------------------------------------------------------------------
//  func called once by init process to initialize data structures
//-----------------------------------------------------------------------------
void srfsInit_()
{
  BZERO(srfsFile);

  inet_ptoh(AF_INET, DEF_SRFS_ADDR, &srfsIPAddr);
  inet_ptoh(AF_INET, DEF_ALT_ADDR, &srfsAltAddr);

  sockFDEV = NULL;  srfsConOK = false;

  if (sem_init(&srfsSem, 1, 1))  { perror(__func__);  return; }

  registerFileSys(&srfsFSysFuncs);
}


//-----------------------------------------------------------------------------
//  Returns true if provided pointer is a valid, in-use SRFS device.
//  Returns false if is not.
//-----------------------------------------------------------------------------
bool  validSRFS(void *ioDev)
{
  uint  idx;


  do  {
    if (!ioDev)  break;

    if ((idx = SRFS - srfsFile) >= MaxSRFSFiles)  break;

    if (SRFS != srfsFile + idx)  break;

    if (SRFS->inUse)  return true;
  } while(0);

  return false;
}


//-----------------------------------------------------------------------------
//  Wait up to the specified number of seconds for the global SRFS lock.
//
//  Returns 0 on success, -1 if an error or timeout
//-----------------------------------------------------------------------------
int  srfsGetLock(int secs)
{
  struct timespec  srfsTimeout;


  srfsTimeout.tv_sec = time(NULL) + secs;  srfsTimeout.tv_nsec = 0;

  if (sem_timedwait(&srfsSem, &srfsTimeout))  {
    fprintf(DiagLog, "\n*** Unable to get SRFS lock: %ls ***\n",  //tdebug
            strerror(errno));  //tdebug
    return -1;
  }

  #ifdef SRFS_DEBUG
    if ((current != kernelTask) AND (userPix != 5))
      fprintf(DiagLog, "  *** got SRFS lock ***\n");
  #endif

  return 0;
}

#ifdef SRFS_DEBUG
//-----------------------------------------------------------------------------
int srfsUnlock(void)
{
  #ifdef SRFS_DEBUG
    if ((current != kernelTask) AND (userPix != 5))
      fprintf(DiagLog, "  *** released SRFS lock ***\n");
  #endif

  return sem_post(&srfsSem);
}
#endif

//-----------------------------------------------------------------------------
//  Establish a connection to the SRFS server if one not already established
//
//  The calling process MUST have the lock for the SRFS driver's shared data
//
//  On success:
//    - Returns a pointer to the FDEV for the shared socket connection
//
//  On failure (no existing connection and one cannot be established):
//    - Returns NULL
//-----------------------------------------------------------------------------
FDEV far *srfsConnect(void)
{
  fd_t  sockfd;
  char  buf[90];


//--- If we already have an open connection, then nothing to do ---
  if (sockFDEV)  return sockFDEV;

//--- Otherwise we try to open a new connection ---
  #ifdef SRFS_DEBUG
    fprintf(DiagLog, "  srfsConnect(): opening a new connection...\n");
  #endif

  if ((sockfd = socket(AF_INET, SOCK_STREAM, 0)) < 0)  {
    fprintf(DiagLog, "  srfsConnect(): socket() error: %ls\n",
                    strerror(errno) );
    return NULL;
  }

//--- set addr info for remote end of connection ---
  srvAddr.sin_family = AF_INET;
  srvAddr.sin_port   = htons(SRFS_PORT);
  srvAddr.sin_addr   = htonl(srfsIPAddr);

//--- and establish a connection to the SRFS server ---
  if (connect(sockfd, (struct sockaddr *) &srvAddr,
              sizeof(struct sockaddr_in) ) < 0)  {
    fprintf(DiagLog, "  srfsConnect(): connect() error: [%d] %ls\n",
                    sockfd, strerror(errno) );
    close(sockfd);  return NULL;
  }

  KLOCK();  sockFDEV = _fdToOpenFile_(sockfd);  KUNLOCK();

  (sockFDEV->ioFuncs->descr)(sockFDEV->ioDev, buf, sizeof(buf));
  fprintf(DiagLog, "\r\n"
                   "*** Connected to SRFS server ***\r\n"
                   "  %.128s\r\n\n", buf);

/*
   So we have a new connection, but the process-specific fd for it is pretty
   useless because it IS process-specific.  So we just free it up and use
   the FDEV ptr which any task can use (as long as it has the global SRFS
   lock to avoid mixing messages for different tasks).
*/
  sockFDEV->shared = true;  // indicates this is SUPPOSED to be an orphan FDEV
  *((FDEV far **)current->FDSet + sockfd) = NULL;

  srfsConOK = true;

  return sockFDEV;
}


//-----------------------------------------------------------------------------
//  close the shared connection to the server
//-----------------------------------------------------------------------------
void srfsDisconnect(void)
{
  char  *ptr;
  char  buf[90];


  KLOCK();

  if (_validFDEV_(sockFDEV))  {
    if (srfsConOK)  {
      ptr = buf;  addShort(&ptr, SRFS_BYE);  addShort(&ptr, 0);
      sendCmd(buf, 4, NULL, 0);
    }
    sockFDEV->openCount = 0;  // frees FDEV
    (sockFDEV->ioFuncs->descr)(sockFDEV->ioDev, buf, sizeof(buf));
    fprintf(DiagLog, "\r\n"
                     "*** Closing con to SRFS server ***\r\n"
                     "  %.128s\r\n\n", buf);
    (sockFDEV->ioFuncs->close)(sockFDEV->ioDev);
  }

  srfsConOK = false;  sockFDEV = NULL;

  KUNLOCK();
}

//-----------------------------------------------------------------------------
//  The connection to the SRFS server failed, which means the connection for
//  all the currently open files is no longer usable so we close the fd for
//  the shared connection for each one and mark them as failed.
//
//  The caller must have the global lock for the SRFS driver's shared data
//-----------------------------------------------------------------------------
void _srfsConFailed_(void)
{
  uint  ui;
  int  errNum;
  SRFS_File  *srfs;


  srfsConOK = false;

  fprintf(DiagLog, "*** Con to SRFS server failed ***\n");

  errNum = (errno ? errno : EIO);

/*
  Since any failure on the shared connection means none of the files
  currently open on it can reliably use the connection to continue
  interacting with the server, we set the error state for all of them.
*/
  for (ui=0,srfs=srfsFile; ui<MaxSRFSFiles; ++ui,++srfs)
    if (srfs->inUse)  srfs->errNum = errNum;

//--- free up the FDEV and close the (failed) shared connection ---
  srfsDisconnect();
}


//-----------------------------------------------------------------------------
//  Send a command and (optional) data to the SRFS server
//
//  The calling process MUST have the lock for the SRFS driver's shared data!
//
//  Parameters:
//    - sockfd:    Process-specific fd for the shared connection to the server
//    - timeoutMS: Max time (in ms) for bytes to be accepted for transmission
//                 to the server (no guarantee they have been or will be
//                 transmitted to or received by the server).
//    - cmd:       Bytes that contain the command to the server
//    - cmdLen:    # of cmd bytes
//    - data:      Parameter values for the command
//    - dataLen:   # of data bytes
//
//  On success (bytes accepted for transmission):
//    - returns 0
//
//  On failure (unable to queue all the bytes for transmission):
//    - returns -1
//    - all open SRFS files marked as failed
//
//  NOTE: All files use a single connection, and each command and its response
//        constitutes a single (usually brief) conversation with the server.
//        These conversations are atomic (i.e. they cannot be concurrent),
//        so any failure in any conversation means the connection itself has
//        failed, since any attempt to use the connection from that time on
//        could result in the command and/or data bytes from the failed
//        conversation getting mixed into subsequent ones.
//-----------------------------------------------------------------------------
int sendCmd(const char *cmd, int cmdLen, const char *data, int dataLen )
{
  bool  writeOK = false;

  #ifdef SRFS_DEBUG
    if (userPix != 5)
      fprintf(DiagLog,
            "  SRFS: sendCmd(): sockFDEV:[%05lX], cmd:%d, len:%d, dataLen:%d\n",
            (ulong)sockFDEV, ntohs(*(const short *)cmd), cmdLen, dataLen);
  #endif

  do  {
    if (writeFDEV(sockFDEV, (void *)cmd, cmdLen, MAX_RW_MS) != cmdLen)
      break;
    if (dataLen)
      if (writeFDEV(sockFDEV, (void *)data, dataLen, MAX_RW_MS) != dataLen)
        break;
    writeOK = true;
  } while(0);

//--- communications failure ---
  if (!writeOK)  {
    if (errno == EAGAIN)  errno = ETIMEDOUT;
    fprintf(DiagLog, "*** SRFS: sendCmd(%u): %ls ***\n",
            ntohs(*(const short *)cmd), strerror(errno) );
    _srfsConFailed_();  return -1;
  }

  lastSendTime = GetMS();

  return 0;
}


//-----------------------------------------------------------------------------
//  Read the response to the last cmd sent to the SRFS server
//
//  The calling process MUST have the lock for the SRFS driver's shared data!
//
//  On success:
//    - returns 0
//    - *stat = return value from requested func on the server
//    - *errnum = errno value from requested func on server
//    - *dataLen = length of any data returned by request
//    - dataBuf contains any data returned by the request
//
//  On failure:
//    - returns -1
//    - *stat = -1
//    - *errnum = local error
//    - *dataLen = 0
//    - dataBuf untouched
//    - all open SRFS files marked as failed
//
//  NOTE: All files use a single connection, and each command and its response
//        constitutes a single (usually brief) conversation with the server.
//        These conversations are atomic (i.e. they cannot be concurrent),
//        so any failure in any conversation means the connection itself has
//        failed, since any attempt to use the connection from that time on
//        could result in the command and/or data bytes from the failed
//        conversation getting mixed into subsequent ones.
//-----------------------------------------------------------------------------
int readResp(int *stat, int *errnum, int *dataLen, char *dataBuf, int maxLen)
{
  char  buf[6];
  char  *ptr;


  #ifdef SRFS_DEBUG
    if (userPix != 5)
      fprintf(DiagLog, "  SRFS: readResp(): sockFDEV:[%05lX]\n",
              (ulong)sockFDEV);
  #endif

  if (readFDEV(sockFDEV, buf, 6, MAX_RW_MS) < 6)  {
    if (errno == EAGAIN)  errno = ETIMEDOUT;
    fprintf(DiagLog, "*** SRFS: readResp(): %ls\n", strerror(errno));
    _srfsConFailed_();
    *stat = -1;  *errnum = errno;  *dataLen = 0;  return -1;
  }

  ptr = buf;
  *stat = getShort(&ptr);  *errnum = getShort(&ptr);  *dataLen = getShort(&ptr);

  #ifdef SRFS_DEBUG
    if (userPix != 5)
      fprintf(DiagLog, "    stat: %d, errnum: %d, dataLen: %d\r\n",
              *stat, *errnum, *dataLen);
  #endif

  if (*dataLen > maxLen)  {
    if (ShowErrors)
      fprintf(DiagLog, "*** SRFS: readResp(): resp too large\r\n");
    _srfsConFailed_();
    *stat = -1;  *errnum = ENOMEM;  *dataLen = 0;  return -1;
  }

  if (*dataLen > 0)
    if (readFDEV(sockFDEV, dataBuf, *dataLen, MAX_RW_MS) != *dataLen)  {
      if (errno == EAGAIN)  errno = ETIMEDOUT;
      fprintf(DiagLog, "*** SRFS: readResp(): %ls\n", strerror(errno));
      _srfsConFailed_();
      *stat = -1;  *errnum = errno;  *dataLen = 0;  return -1;
    }

  return 0;
}


//=============================================================================
//  FileSysFuncs that do not pass an ioDev object
//=============================================================================

//-----------------------------------------------------------------------------
//  Open a a file on the SRFS server.  Note that, while the socket connection
//  to the server is shared by all open SRFS driver files, we keep a copy of
//  the socket fd last used by each file so the next call to an SRFS driver
//  function knows if the server fd applies to the current connection or a
//  previous one.  In the later case, any further operations on the DEBROS fd
//  for the file will fail (since the only reason the socket connection for
//  it would be closed is that an error occured).
//-----------------------------------------------------------------------------
fd_t _srfsOpen_(const char *path, int flags, int mode)
{
  int  i, fd, nameLen, stat, errNum, respLen;
  SRFS_File  *srfs;
  char  *ptr;
  char  buf[16];


  #ifdef SRFS_DEBUG
    fprintf(DiagLog, "\n_srfsOpen_(\"%.32s\", 0x%X, 0x%X)\n",
            path, flags, mode);
  #endif

//--- ignore names that do not start with "/srfs/" ---
  if (strncmp(path, "/srfs/", 6))  { errno = ENXIO;  return -1; }
  path += 6;  // skip over "/srfs/" portion

//--- get lock for the SRFS driver's shared data ---
  if (SRFS_LOCK(30))  return -1;

//--- locate an unused entry in the list ---
  for (i=0; i<MaxSRFSFiles; ++i)  if (! srfsFile[i].inUse)  break;
  if (i == MaxSRFSFiles)  { SRFS_UNLOCK();  ERRNO = ENFILE;  return -1; }
  srfs = srfsFile + i;  srfs->name = srfsFileName[(uint)i];

//--- make sure the shared connection to the server is open ---
  if (! srfsConnect())  { SRFS_UNLOCK();  return -1; }

  KLOCK();

//--- alloc a process fd and global FDEV for the SRFS driver's structure ---
  if ((fd = _allocFile_(3, &srfsIOFuncs, srfs)) < 0)  return -1;

  srfs->inUse = true;  srfs->remfd = -1;  srfs->errNum = 0;
  fstrcpy(srfs->name, path);

  KUNLOCK();

//cur method for telling srfs server that we will send keepalive packets
#define SRFS_O_TIMEOUT   0x8000   // abort if idle for too long

//--- send OPEN request ---
  ptr = buf;  nameLen = strlen(path);
  addShort(&ptr, SRFS_OPEN);  addShort(&ptr, 4+nameLen);
  addShort(&ptr, flags | SRFS_O_TIMEOUT);  addShort(&ptr, mode);

  if (sendCmd(buf, 8, (char *)path, nameLen))  {
    SRFS_UNLOCK();  close(fd);  return -1;  }

//--- wait for response to OPEN request ---
  if (readResp(&stat, &errNum, &respLen, buf, sizeof(buf)))  {
    SRFS_UNLOCK();  close(fd);  return -1; }

//--- server was unable to open the specified file ---
  if (stat < 0)  { SRFS_UNLOCK();  close(fd);  errno = errNum;  return -1; }

//--- file was opened: save the remote fd provided for the file ---
  srfs->remfd = stat;

  #ifdef SRFS_DEBUG
    fprintf(DiagLog, "  opened: ioDev:0x%4X, fd:%d, remfd:%d\n",
            (uint)srfs, fd, srfs->remfd);
  #endif

//--- everything OK - return the local fd assigned for the remote file ---
  SRFS_UNLOCK();  return fd;
}


//-----------------------------------------------------------------------------
//  func called each kernel tick to provide the driver with CPU cycles
//  Should ONLY be called by the kernel task!
//-----------------------------------------------------------------------------
void _srfsTick_(void)
{
  ushort  u;
  char  *ptr;
  SRFS_File  *ioDev;
  char  buf[4];
  static uint  tickCount;


  if (current != kernelTask)  return;

//--- should we close the shared connection in preperation for a shutdown? ---
  if (doShutdown AND sockFDEV)  {
    for (u=0,ioDev=srfsFile; u<MaxSRFSFiles; ++u,++ioDev)
      if (ioDev->inUse)  break;
    if (u == MaxSRFSFiles)  srfsDisconnect();  // not in use, safe to close
  }


  if ((++tickCount & 0x3F) == 0)  return;  // don't need to check very often

//===== Send a "keepalive" packet if the connection to the SRFS server =====
//===== is open, not busy, and has been idle for at least 10 seconds =====

//--- If another process already has the lock for the shared data, then ---
//--- a keepalive packet isn't needed (since the link is in use). ---
//--- Besides, we CANNOT allow the kernel task to block indefinitely! ---
  if (sem_trywait(&srfsSem))  return;

//--- If we have an open connection and it has been idle, send a KEEPALIVE ---
//--- packet to the server so it knows we are still alive and well ---
  if (sockFDEV AND  (msSince(lastSendTime) >= 10000))  {
  //--- send KEEPALIVE "cmd" (no response sent by server) ---
    ptr = buf;  addShort(&ptr, SRFS_KEEPALIVE);  addShort(&ptr, 0);
    sendCmd(buf, 4, NULL, 0);
  }

  SRFS_UNLOCK();  return;
}


//-----------------------------------------------------------------------------
//  Return the state of the specified ioDev entry or NULL if not a valid
//  entry #.
//
//  For valid entrys, the entryState struct is filled in
//-----------------------------------------------------------------------------
void *_srfsEntry_(uint idx, EntryState *entryState)
{
  SRFS_File  *srfs;

  if (idx >= MaxSRFSFiles)  return NULL;

  srfs = srfsFile + idx;

  entryState->openCount = (srfs->inUse ? 1 : 0);
  entryState->unused = (entryState->openCount == 0);

  return srfs;
}


//-----------------------------------------------------------------------------
//  Attempt to delete a remote file via the SRFS protocol
//
//  Should only be called from the fcntl.c/unlink() function!
//
//  RETURN VALUEs:
//  - if successful: 0
//  - if an error: -1 and ERRNO set accordingly
//-----------------------------------------------------------------------------
int _srfsUnlink_(const char *path)
{
  char  localBuf[4];
  int   stat, respLen, errNum, pathLen;
  char  *ptr;


  #ifdef SRFS_DEBUG
    fprintf(DiagLog, "\n_srfsUnlink_(%.*s)\r\n", PATH_MAX, path);
  #endif

//--- ignore names that do not start with "/srfs/" ---
  if (strncmp(path, "/srfs/", 6))  { errno = ENXIO;  return -1; }
  path += 6;  // skip over "/srfs/" portion

//--- get lock for the SRFS driver's shared data ---
  if (SRFS_LOCK(30))  return -1;

//--- make sure the shared connection to the server is open ---
  if (! srfsConnect())  { SRFS_UNLOCK();  return -1; }

  pathLen = strlen(path) + 1;  stat = errNum = 0;

//--- send UNLINK request and get the response ---
  ptr = localBuf;  addShort(&ptr, SRFS_UNLINK);  addShort(&ptr, pathLen);
  if (sendCmd(localBuf, 4, path, pathLen))
    stat = -1;  else
  if (readResp(&stat, &errNum, &respLen, localBuf, sizeof(localBuf)))
    stat = -1;  else
  if (errNum)  ERRNO = errNum;

  SRFS_UNLOCK();  return stat;
}


//-----------------------------------------------------------------------------
//  Attempt to rename a remote file via the SRFS protocol
//
//  Should only be called from the fcntl.c/rename() function!
//
//  RETURN VALUEs:
//  - if successful: 0
//  - if an error: -1 and ERRNO set accordingly
//-----------------------------------------------------------------------------
int _srfsRename_(const char *oldPath, const char *newPath)
{
  bool  writeOK;
  char  localBuf[4];
  int   stat, respLen, errNum, oldPathLen, newPathLen;
  char  *ptr;


  #ifdef SRFS_DEBUG
    fprintf(DiagLog, "\n_srfsRename_(%.*s, %.*s)\r\n",
           PATH_MAX, oldPath, PATH_MAX, newPath);
  #endif

//--- ignore names that do not start with "/srfs/" ---
  if (strncmp(oldPath, "/srfs/", 6))  { errno = ENXIO;  return -1; }
  if (strncmp(newPath, "/srfs/", 6))  { errno = ENXIO;  return -1; }
  oldPath += 6;  newPath += 6;  // skip over "/srfs/" portion


  oldPathLen = strlen(oldPath) + 1;  newPathLen = strlen(newPath) + 1;

//--- get lock for the SRFS driver's shared data ---
  if (SRFS_LOCK(30))  return -1;

//--- make sure the shared connection to the server is open ---
  if (! srfsConnect())  { SRFS_UNLOCK();  return -1; }

//--- send WRITE request and the data ---
  ptr = localBuf;  addShort(&ptr, SRFS_RENAME);
  addShort(&ptr, oldPathLen + newPathLen);
  do  {
    writeOK = true;
    if (writeFDEV(sockFDEV, localBuf, 4, MAX_RW_MS) != 4)
      writeOK = false;  else
    if (writeFDEV(sockFDEV, (char *)oldPath, oldPathLen, MAX_RW_MS) != oldPathLen)
      writeOK = false;  else
    if (writeFDEV(sockFDEV, (char *)newPath, newPathLen, MAX_RW_MS) != newPathLen)
      writeOK = false;
  } while(0);

  if (!writeOK)  {
    fprintf(DiagLog, "_srfsRename_(): %ls\n", strerror(errno) );
    _srfsConFailed_();  SRFS_UNLOCK();  return -1;
  }

//--- wait for response to RENAME request ---
  if (readResp(&stat, &errNum, &respLen, localBuf, sizeof(localBuf)))
    stat = -1;
  else
    if (errNum)  ERRNO = errNum;

  SRFS_UNLOCK();  return stat;
}



//=============================================================================
//  IODev interface functions (for use by fcntl module)
//=============================================================================

//-----------------------------------------------------------------------------
//  The only difference between blocking and non-blocking mode for the SRFS
//  driver files is whether or not the call waits for the global SRFS lock
//  if it is not available immediately.  So there is no point in having the
//  fcntl layer attempt a read or write multiple times.
//-----------------------------------------------------------------------------
int _srfsGetQueues_(void *ioDev, WaitQueues far *queues)
{
  queues->readQueue = queues->writeQueue = NULL;  //&sleepQueue;

  return 0;
}

//-----------------------------------------------------------------------------
//  close an SRFS connection to a remote file
//-----------------------------------------------------------------------------
int _srfsClose_(void *ioDev)
{
  int  stat, respLen, errNum;
  char  *ptr;
  char  localBuf[16];


  if (!validSRFS(ioDev))  { ERRNO = EBADF;  return -1; }

  #ifdef SRFS_DEBUG
    fprintf(DiagLog, "\n_srfsClose_(0x%04X): remfd:%d, sockFDEV:[%05lX]\r\n",
            (uint)ioDev, SRFS->remfd, (ulong)sockFDEV);
  #endif


//--- get lock for the SRFS driver's shared data ---
  if (SRFS_LOCK(30))  return -1;

//--- previous error on the server connection we were using? ---
  if ((SRFS->remfd < 0) OR SRFS->errNum)  {
    ERRNO = SRFS->errNum;  SRFS->inUse = false;
    SRFS_UNLOCK();  return -1;
  }

//--- Send CLOSE cmd and wait for the response ---
  ptr = localBuf;  addShort(&ptr, SRFS_CLOSE);
  addShort(&ptr, 2);  addShort(&ptr, SRFS->remfd);

  stat = 0;
  if (sendCmd(localBuf, 6, NULL, 0))
    stat = -1;  else
  if (readResp(&stat, &errNum, &respLen, localBuf, sizeof(localBuf)))
    stat = -1;  else
  if (errNum)  ERRNO = errNum;


//--- clear the driver's ioDev for the remote file so it can be reused ---
  SRFS->remfd = -1;  SRFS->inUse = false;


  SRFS_UNLOCK();  return stat;
}

//-----------------------------------------------------------------------------
//  Attempt to read bytes from a remote file via the SRFS protocol
//
//  Should only be called from the fcntl.c/read() function!
//
//  RETURN VALUEs:
//  - if bytes read: # of bytes read
//  - if no bytes read:
//    - if more possible in the future: -1 and ERRNO == EAGAIN
//    - if no more possible (EOF):  0
//-----------------------------------------------------------------------------
int _srfsRead_(void *ioDev, char *buf, int numBytes, bool dontBlock)
{
  char  localBuf[16];
  int  stat, respLen, errNum;
  char  *ptr;


  #ifdef SRFS_DEBUG
    fprintf(DiagLog, "\n_srfsRead_(.., .., %d)\r\n", numBytes);
  #endif

  if (!validSRFS(ioDev))  { ERRNO = EBADF;  return -1; }

//--- Attempt to get the global lock (with no timeout if non-blocking mode) ---
  if (dontBlock ? sem_trywait(&srfsSem) : sem_wait(&srfsSem))  return -1;

//--- previous error on the server connection we were using? ---
  if ((SRFS->remfd < 0) OR SRFS->errNum)  {
    ERRNO = SRFS->errNum;  SRFS_UNLOCK();  return -1; }


//--- Send READ request and read parameters ---
  ptr = localBuf;
  addShort(&ptr, SRFS_READ);  addShort(&ptr, 4);
  addShort(&ptr, SRFS->remfd);  addShort(&ptr, numBytes);

  stat = 0;
  if (sendCmd(localBuf, 8, NULL, 0))
    stat = -1;  else
  if (readResp(&stat, &errNum, &respLen, buf, numBytes))
    stat = -1;
  else  {
    if (errNum)  errno = errNum;
    if (errno == EAGAIN)  errno = EIO;
  }

//fprintf(DiagLog, "srfsd-read returned: %d/%d\r\n", stat, errno);  //tdebug

  SRFS_UNLOCK();  return stat;
}

//-----------------------------------------------------------------------------
//  Attempt to write bytes to remote file via the SRFS protocol
//
//  Should only be called from the fcntl.c/write() function!
//
//  RETURN VALUEs:
//  - if bytes written: # of bytes written
//  - if no bytes written:
//    - if more possible in the future: -1 and ERRNO == EAGAIN
//    - if no more possible (EOF):  0
//-----------------------------------------------------------------------------
int _srfsWrite_(void *ioDev, char *buf, int numBytes, bool dontBlock)
{
  char  localBuf[16];
  int   stat, respLen, errNum;
  char  *ptr;


  #ifdef SRFS_DEBUG
    if (userPix != 5)
      fprintf(DiagLog, "\n_srfsWrite_(0x%04X, .., %d)\r\n",
              (uint)ioDev, numBytes);
  #endif

  if (!validSRFS(ioDev))  { ERRNO = EBADF;  return -1; }


//--- Attempt to get the global lock (with no timeout if non-blocking mode) ---
  if (dontBlock ? sem_trywait(&srfsSem) : sem_wait(&srfsSem))  return -1;

//--- previous error on the server connection we were using? ---
  if ((SRFS->remfd < 0) OR SRFS->errNum)  {
    ERRNO = SRFS->errNum;  SRFS_UNLOCK();  return -1; }


//--- Send WRITE request and wait for the response ---
  ptr = localBuf;
  addShort(&ptr, SRFS_WRITE);  addShort(&ptr, 4+numBytes);
  addShort(&ptr, SRFS->remfd);  addShort(&ptr, numBytes);

  stat = 0;
  if (sendCmd(localBuf, 8, buf, numBytes))
    stat = -1;  else
  if (readResp(&stat, &errNum, &respLen, localBuf, sizeof(localBuf)))
    stat = -1;
  else  {
    if (errNum)  errno = errNum;
    if (errno == EAGAIN)  errno = EIO;
  }

//fprintf(DiagLog, "srfsd-write returned: %d/%d\r\n", stat, errno);  //tdebug

  SRFS_UNLOCK();  return stat;
}


//-----------------------------------------------------------------------------
//  Set the current offset for a remote file using the SRFS protocol
//-----------------------------------------------------------------------------
off_t  _srfsLseek_(void *ioDev, off_t offset, int whence)
{
  char  localBuf[16];
  int    stat, respLen, errNum;
  char  *ptr;


  #ifdef SRFS_DEBUG
    fprintf(DiagLog, "_srfsLseek_(.., %lu, %d)\r\n", (ulong)offset, whence);
  #endif

  if (!validSRFS(ioDev))  { ERRNO = EBADF;  return -1; }

//--- Attempt to get the global lock ---
  if (SRFS_LOCK(15))  return -1;

//--- previous error on the server connection we were using? ---
  if ((SRFS->remfd < 0) OR SRFS->errNum)  {
    ERRNO = SRFS->errNum;  SRFS_UNLOCK();  return -1; }


//--- send LSEEK request and wait for the response ---
  ptr = localBuf;
  addShort(&ptr, SRFS_LSEEK);  addShort(&ptr, 8);
  addShort(&ptr, SRFS->remfd);
  addLong(&ptr, offset);  addShort(&ptr, whence);

  stat = 0;
  if (sendCmd(localBuf, 12, NULL, 0))
    offset = (off_t)(-1);  else
  if (readResp(&stat, &errNum, &respLen, localBuf, sizeof(localBuf)))
    offset = (off_t)(-1);
  else  {
    if (errNum)  ERRNO = errNum;
    if (errno == EAGAIN)  errno = EIO;
    ptr = localBuf;  offset = getLong(&ptr);
  }

//fprintf(DiagLog, "srfsd-write returned: %d/%d\r\n", stat, errno);  //tdebug

  SRFS_UNLOCK();  return offset;
}


//-----------------------------------------------------------------------------
//  Get information about a remote file using the SRFS protocol
//-----------------------------------------------------------------------------
int  _srfsFstat_(void *ioDev, struct stat *fsbuf)
{
  char  buf[sizeof(struct stat)];
  int   respStat, respLen, errNum;
  char  *ptr;


  #ifdef SRFS_DEBUG
    fprintf(DiagLog, "\n_srfsFstat_(.., ..)\n");
  #endif

  if (!validSRFS(ioDev))  { ERRNO = EBADF;  return -1; }


//--- fill buffer with FSTAT request ---
  ptr = buf;
  addShort(&ptr, SRFS_FSTAT);  addShort(&ptr, 4);
  addShort(&ptr, SRFS->remfd);  addShort(&ptr, sizeof(buf));

//--- Try to get the lock for the shared connection ---
  if (SRFS_LOCK(15))  return -1;

//--- previous error on the server connection we were using? ---
  if ((SRFS->remfd < 0) OR SRFS->errNum)  {
    ERRNO = SRFS->errNum;  SRFS_UNLOCK();  return -1; }


//--- send cmd and wait for response ---
  do  {
    if (sendCmd(buf, 8, NULL, 0))  break;
    if (readResp(&respStat, &errNum, &respLen, buf, sizeof(buf)))  break;

    SRFS_UNLOCK();

    if (errNum)  errno = errNum;
    if (errno == EAGAIN)  errno = EIO;

    if (respStat == 0)  {
      ptr = (char *)buf;
      fsbuf->st_mode  = getLong(&ptr);  // protection and file type bits
      fsbuf->st_rdev  = getLong(&ptr);  // device type
      fsbuf->st_size  = getLong(&ptr);  // total size in bytes
      fsbuf->st_atime = getLong(&ptr);  // time of last access
      fsbuf->st_mtime = getLong(&ptr);  // time of last modification
      fsbuf->st_ctime = getLong(&ptr);  // time of last status change
    }

    return respStat;
  } while(0);

  SRFS_UNLOCK();  return -1;
}

//-----------------------------------------------------------------------------
//  Kernel-space function to handle ioctl calls for srfs ports
//-----------------------------------------------------------------------------
int _srfsioctl_(void *ioDev, int request, void *arg)
{
  if (!validSRFS(ioDev))  { ERRNO = EBADF;  return -1; }

 //finish:  need to implement this...

  ERRNO = EINVAL;  return -1;
}


//-----------------------------------------------------------------------------
//  Kernel-space function to return a description of the srfs port
//
//  WARNING!  To be called ONLY by KERNELFUNC functions (or functions called BY
//            KERNELFUNC functions) !!!!
//-----------------------------------------------------------------------------
char *_srfsDescr_(void *ioDev, char *buf, int bufSize)
{
  char  *ptr;


  sprintf(buf, "srfs 0x%04lX ", (ulong)ioDev);

  ptr = buf + strlen(buf);

//finish: limit output to bufSize

  KLOCK();

  if (validSRFS(ioDev))
    sprintf(ptr, "srfs[%d] %ls", (int)(SRFS - srfsFile), SRFS->name);
  else
    strcpy(ptr, "bad bad");

  KUNLOCK();

  return buf;
}


//-----------------------------------------------------------------------------

