//----- file_io.c ----- 03/12/09 -----(xx/xx/xx)-----

//----- original code from SHDesigns -----

//--- modifications, additions, and reformatting by Mark Davis, NSCL ---

//finish:  need proper locking/unlocking of kernel structures around
//         changes to the file table

//-----------------------------------------------------------------------------
//  Update Log (most recent first)
//
//  2009-03-12 - Mark Davis
//               - A file can no longer be deleted if it is already open
//
//  2008-04-30 - Mark Davis
//               - changes to make seek work properly and so all funcs can
//                 properly use the data in the sector buffer after EOF is
//                 reached.
//
//  2008-01-15 - Mark Davis
//               - some cleanup and commenting of code done while examining it
//                 for potential concurrency issues
//
//  2007-08-03 - Mark Davis
//               - added ffile_setName() and ffile_getName()
//
//  2006-xx-xx - Mark Davis
//               - original modifications
//
//-----------------------------------------------------------------------------

#include <common.h>

#include <sys/types.h>
#include <unistd.h>
#include <sys/stat.h>
#include <utime.h>
#include <fcntl.h>

#include <stcpip.h>

#include "string.h"

#include <SHD_DLM.h>
#include <DLMiii.h>

#include <rabbit.h>
#include <dcdefs.h>

#include <sys/farprintf.h>


//--- list of open references to files ---
//--- The ffile "handles" are indexes in to this list ---
FILE_DATA  _ffile_list[DOPEN_FILES];

ulong  param_start = 0l;
ushort  sector_size;

//=============================================================================
//      ---------  !!!!! WARNING !!!!!    !!!!! WARNING !!!!!  ------------
//
//   The following table is used by both the Bootloader/Download Manager
//   and the application/OS!
//
//                !!!!! Do NOT change the structure at all !!!!!
//=============================================================================
CONFIG_TABLE far *config_table = (CONFIG_TABLE far *)0l;


//-----------------------------------------------------------------------------
static unsigned getNextBlock(unsigned curBlockNum)
{
  if (curBlockNum >= MAX_BLK_NUM)  return EOF_BLK;

  return config_table->nextBlock[curBlockNum];
}

//-----------------------------------------------------------------------------
static void setNextBlock(ushort blocknum, ushort data)
{
  config_table->nextBlock[blocknum] = data;
}

//-----------------------------------------------------------------------------
//  Return the index/# of the next unused block/sector in flash
//  Returns 0 if no unused blocks/sectors
//-----------------------------------------------------------------------------
static ushort get_freeblock(void)
{
  ushort far * blocks = config_table->nextBlock;
  ushort blocknum = 0;

  while (blocknum < MAX_BLK_NUM)  {
    if (*blocks == 0)  return blocknum;
    blocks++;  blocknum++;
  }

  return 0;
}

//-----------------------------------------------------------------------------
static void  write_block(unsigned blocknum, char far * buff)
{
  ulong addr = blocknum;

  addr *= sector_size;  _FlashUnlock = ~addr;

  WriteFlash((FlByte const far *)addr, (FlByte far *)buff, sector_size);
}

//-----------------------------------------------------------------------------
//  copy the entire config_table to flash
//-----------------------------------------------------------------------------
static void  update_config(void)
{
  _FlashUnlock = ~param_start;

  WriteFlash((FlByte const far *)param_start, (FlByte far *)config_table,
             sizeof(CONFIG_TABLE));
}

//-----------------------------------------------------------------------------
//  Initialize the flash file system
//-----------------------------------------------------------------------------
int init_file_data(void)
{
  static char file_init = 0;

  if (file_init)  return 1;

  GetFlashDeviceEnd((FlDWord )0x0l, &sector_size);
  printf("Flash Sector Size: %u\r\n", sector_size);

  memset(_ffile_list, 0, sizeof(_ffile_list));

  if (!config_table)
    config_table = (CONFIG_TABLE far *)xalloc(sizeof(CONFIG_TABLE));

  param_start = 0xc000L;  // where to start search from

  while (config_table AND !file_init)  {
    ReadFlash((FlByte const far *)param_start, (FlByte far *)config_table,
              sizeof(CONFIG_TABLE));
    if (config_table->key == CONFIG_KEY)  { file_init = 1;  break; }
    param_start += 0x1000;  // 4k alignment
    if (param_start >= 0x20000l)  break;  // not found in 128k then exit
  }

  return file_init;
}

//-----------------------------------------------------------------------------
//  Return the number of bytes in unused sectors of flash
//-----------------------------------------------------------------------------
long int ffile_free(void)
{
  unsigned  i;
  long int  free = 0l;

  if (!init_file_data())  return FERR_NOXMEM;

  for (i=0; i<config_table->num_blocks; i++)
    if (config_table->nextBlock[i] == 0)  free += sector_size;

  return free;
}


//-----------------------------------------------------------------------------
//  Return the next unused file handle (an index in to the table of opened
//  files).  Returns < 0 if all handles are in use.
//-----------------------------------------------------------------------------
static int next_handle(void)
{
  for (int i=0; i<DOPEN_FILES; i++)
    if ((_ffile_list[i].flags & FFLAG_OPEN) == 0) return i;

  return FERR_TOOMANYFILES;
}

//-----------------------------------------------------------------------------
//  Return the index for the next unused file entry in the FAT
//-----------------------------------------------------------------------------
static int next_file(void)
{
  for (unsigned j=0; j<MAX_FILES; j++)
    if ((config_table->files[j].flags & FILE_ACTIVE) == 0)  return j;

  return FERR_TOOMANYFILES;
}

//-----------------------------------------------------------------------------
//  Return the index in the FAT for the specified file name, or < 0 if the
//  file is not found.
//-----------------------------------------------------------------------------
int ffile_index(char * name)
{
  for (unsigned j=0; j<MAX_FILES; j++)
    if ( ((config_table->files[j].flags & (FILE_ACTIVE | FILE_SFLASH))
                == (FILE_ACTIVE))
         AND (fstrcmp(config_table->files[j].name, name) == 0) )
      return j;

  return FERR_NOTFOUND;
}

//-----------------------------------------------------------------------------
//  Return a handle for the specified file.  Returns < 0 if the file cannot
//  be opened.
//-----------------------------------------------------------------------------
int ffile_open(char * name)
{
  int i, j;
  FILE_DATA  *ffile;


  if (!init_file_data())  return FERR_NOXMEM;

//--- get a handle for opening a file ---
  i = next_handle();  if (i < 0)  return i;

//--- get the index in to the FAT for the specified file ---
  j = ffile_index(name);  if (j < 0)  return j;

  ffile = _ffile_list + i;

//--- allocate a buffer for partial sector reads and writes ---
  if (!ffile->sector_buff)  ffile->sector_buff = xalloc(sector_size);
//  if (!ffile->sector_buff)  ffile->sector_buff = farmalloc(sector_size);
  if (!ffile->sector_buff)  return FERR_NOXMEM;

//--- initialize the rest of the file handle ---
  ffile->buff_ptr = ffile->sector_buff;
  ffile->entry_num = j;
  ffile->block_num = 0;  // indicates no data in r/w buffer
  ffile->pos = 0;
  ffile->flags = FFLAG_OPEN;

  return i;
}

//-----------------------------------------------------------------------------
//  Read a sector from flash in to the provided buffer
//-----------------------------------------------------------------------------
static void read_block(unsigned block, char far * buff)
{
  ulong  addr;

  addr = (ulong)block * sector_size;
//printf("ReadFlash(%lu, [%lu], %u)\r\n", (ulong)addr, (ulong)buff, sector_size);  //tdebug
//fflush(stdout);
  ReadFlash((FlByte const far *)(addr), (FlByte far *)buff, sector_size);
}

//-----------------------------------------------------------------------------
//  copy a file's write buffer to flash and update its last-modified time
//-----------------------------------------------------------------------------
static void fflush_file(FILE_DATA * ffile)
{
  unsigned entry = ffile->entry_num;

  write_block(ffile->block_num, ffile->sector_buff);

  config_table->files[entry].date = readRTC(NULL);

  ffile->flags &= ~FFLAG_WRITE;   // reset file-needs-flushing flag

  ffile->flags |= FFLAG_UPDATED;  // set FAT-entry-needs-flushing flag
}

//-----------------------------------------------------------------------------
//  Read the specified # of bytes from an open file in to the provided buffer
//  Returns the # of bytes read.
//-----------------------------------------------------------------------------
unsigned ffile_read(int handle, char far *buff, unsigned bytes)
{
  unsigned  numleft, read, nextBlock;
  FILE_DATA  *ffile;  // equiv to file descriptor
  FILE_ENTRY  far *entry;  // global info on the file


  read = 0;  ffile = _ffile_list + handle;
  entry = config_table->files + ffile->entry_num;

//printf("ffile_read: pos: %lu  size: %lu  offset:%ld\r\n",
//       ffile->pos, entry->size, (long)(ffile->buff_ptr - ffile->sector_buff));

  if (ffile->pos + bytes > entry->size)
    bytes = (uint)(entry->size - ffile->pos);
  if (!bytes)  return read;

//--- if the r/w buffer has no data in it, then read the 1st block of ---
//--- the file from the flash ---
  if (ffile->block_num == 0) {  // first read
    ffile->block_num = (int)entry->startblock;
    read_block(ffile->block_num, ffile->sector_buff);
    ffile->buff_ptr = ffile->sector_buff;
  }

//--- copy bytes from the r/w buffer and read the next block from flash ---
//--- each time we have read all the bytes in the buffer ---
  while (bytes)  {
    numleft = sector_size - ((uint)ffile->pos & (sector_size-1));
    if (numleft > bytes)  numleft = bytes;
    fmemcpy(buff, ffile->buff_ptr, numleft);
    ffile->buff_ptr += numleft;
    buff += numleft;
    bytes -= numleft;
    ffile->pos += numleft;
    read += numleft;

//    if (bytes)  {
  //--- (NSCL: less than optimal fix to SHD bug:) ---
  //--- If we used up what was in the buffer, read the next sector (even if ---
  //--- we don't need it now, as another read() right after this one will ---
  //--- return the wrong stuff if we don't!)
    if (((uint)ffile->pos & (sector_size-1)) == 0)  {

    //--- if buffer has newer data than flash, update the flash copy first ---
      if (ffile->flags & FFLAG_WRITE)  fflush_file(ffile);

    //--- get next block num and check for EOF marker ---
      nextBlock = getNextBlock(ffile->block_num);
      if ((nextBlock == EOF_BLK) OR !nextBlock)  return read;

    //--- read the flash sector in to the buffer and reset the next-byte ---
    //--- to-read-or-write pointer to the start of the buffer ---
      read_block(nextBlock, ffile->sector_buff);
      ffile->buff_ptr = ffile->sector_buff;
      ffile->block_num = nextBlock;
    }
  }

  return read;
}

//-----------------------------------------------------------------------------
//  Close an open file, flushing any unwritten data from the write buffer
//  and updating the FAT table in flash if any info on the file has changed
//-----------------------------------------------------------------------------
void ffile_close(int handle)
{
  FILE_DATA  *ffile = _ffile_list + handle;

  if (ffile->flags & FFLAG_WRITE)  fflush_file(ffile);
  if (ffile->flags & FFLAG_UPDATED)  update_config();

//  if (ffile->sector_buff)  {
//    farfree(ffile->sector_buff);  ffile->sector_buff = NULL;  }

  ffile->flags = 0;
}

//-----------------------------------------------------------------------------
//  Delete a flash file by name.
//
//  Returns 0 on success, < 0 if an error
//-----------------------------------------------------------------------------
int ffile_delete(char * name)
{
  int index = ffile_index(name);

  if (index < 0)  { ERRNO = ENOENT;  return -1; }

  if (ffile_delete_index(index) < 0)  return -1;

  return 0;
}

//-----------------------------------------------------------------------------
//  Delete the file for the specified file table index
//
//  Returns the index # on success (>= 0), or < 0 if an error;
//-----------------------------------------------------------------------------
int ffile_delete_index(int index)
{
  unsigned block;
  FILE_DATA  *ffile;
  FILE_ENTRY far *file;


  if (!init_file_data())  { ERRNO = ENOMEM;  return FERR_NOXMEM; }
  if (index < 0)  { ERRNO = ENOENT;  return index; }

//--- make sure the file is not open ---
  for (uint ui=0; ui<DOPEN_FILES; ++ui)  {
    ffile = _ffile_list + ui;
    if (!(ffile->flags & FFLAG_OPEN))  continue;
    if (ffile->entry_num == index)  { ERRNO = EBUSY;  return -1; }
  }

  file = config_table->files + (uint)index;

  block = (uint)file->startblock;

  // FAT will have 0 size
  if (file->size)
    while (block AND (block != EOF_BLK))  {
      unsigned nxt = getNextBlock(block);
      setNextBlock(block, 0);  block = nxt;
    }

  fmemset(file, 0, sizeof(FILE_ENTRY));

  if (config_table->backup == index + 1)  config_table->backup = 0;
  if (config_table->primary == index + 1)  config_table->primary = 0;

  update_config();

  return index;
}

//-----------------------------------------------------------------------------
//  Create an entry in the file table for a new file, open it, and return the
//  handle for the open file
//-----------------------------------------------------------------------------
int ffile_create(char * name)
{
  int i, j;
  FILE_ENTRY far  *entry;
  FILE_DATA  *ffile;


  if (!init_file_data())  return FERR_NOXMEM;

//--- get a handle for opening a file ---
  i = next_handle();  if (i < 0)  return i;

//--- make sure a file with the specified name doesn't already exist ---
  j = ffile_index(name);  if (j >= 0)  return FERR_FILE_EXISTS;

//--- get an unused entry in the file table ---
  j = next_file();  if (j < 0)  return j;

  entry = config_table->files + (uint) j;
  ffile = _ffile_list + i;

//--- allocate a buffer for partial sector reads and writes ---
  if (!ffile->sector_buff)  ffile->sector_buff = xalloc(sector_size);
  if (!ffile->sector_buff)  return FERR_NOXMEM;

//--- initialize the new file table entry and the file handle ---
  fmemset(entry, 0, sizeof(FILE_ENTRY));
  fstrcpy(entry->name, name);
  entry->flags = FILE_ACTIVE;

  ffile->buff_ptr = ffile->sector_buff;
  ffile->entry_num = j;
  ffile->block_num = 0;
  ffile->pos = 0;
  ffile->flags = FFLAG_OPEN | FFLAG_UPDATED;

  return i;
}

//-----------------------------------------------------------------------------
//  Create an entry in the file table for a new file and return its index
//  in the file table
//-----------------------------------------------------------------------------
int ffile_FATLink(char * name)
{
  int j;
  FILE_ENTRY far  *entry;


  if (!init_file_data())  return FERR_NOXMEM;

//--- make sure a file with the specified name doesn't already exist ---
  j = ffile_index(name);  if (j >= 0)  return FERR_FILE_EXISTS;

//--- get an unused entry in the file table ---
  j = next_file();  if (j < 0)  return j;

  entry = config_table->files + (uint) j;

//--- initialize the entry ---
  fmemset(entry, 0, sizeof(FILE_ENTRY));
  fstrcpy(entry->name, name);
  entry->flags = FILE_ACTIVE | FILE_SFLASH;

//--- save the update FAT to flash ---
  update_config();

  return j;
}

//-----------------------------------------------------------------------------
//  Return current position for the specified open file
//-----------------------------------------------------------------------------
long ffile_tell(int handle)
{
  FILE_DATA *ffile = _ffile_list + handle;

  return ffile->pos;
}

//-----------------------------------------------------------------------------
//  Change current file position relative to start, cur pos, or end of file
//  Returns on success or < 0 if an error
//-----------------------------------------------------------------------------
int ffile_seek(int handle, long pos, int mode)
{
  long  distance, bufOffset;
  unsigned  nextBlock;
  FILE_DATA *ffile = _ffile_list + handle;
  FILE_ENTRY far *entry;


  ffile = _ffile_list + handle;
  entry = config_table->files + ffile->entry_num;

//printf("ffile_seek_(%d, %ld, %d), cur pos: %lu)\r\n",  //tdebug
//       handle, pos, mode, ffile->pos);

  if (ffile->block_num == 0) {  // first read
    ffile->pos = 0;
    ffile->block_num = (int)entry->startblock;
    read_block(ffile->block_num, ffile->sector_buff);
    ffile->buff_ptr = ffile->sector_buff;
  }

  switch (mode)  {
    case FSEEK_END:
      // pos better be negative, will check result below
      pos += config_table->files[ffile->entry_num].size;  break;

    case FSEEK_CUR:
      if (!pos)  return 0; // no move, so just exit
      pos += ffile->pos;  break;
  }

  if ((pos > config_table->files[ffile->entry_num].size) OR (pos < 0))
    return FERR_SEEK;


  bufOffset = ffile->buff_ptr - ffile->sector_buff;

//===== new position is closer to start of file =====
  if (pos < ffile->pos)  {
    distance = ffile->pos - pos;

  //--- move back within the existing contents of the sector buffer? ---
    if (distance < bufOffset)  {
      ffile->buff_ptr -= distance;
      ffile->pos -= distance;
      return 0;
    }

    if (ffile->flags & FFLAG_WRITE)  fflush_file(ffile);

  //--- seek to start of file and reseek ---
    ffile->block_num = config_table->files[ffile->entry_num].startblock;
    ffile->pos = 0;  ffile->buff_ptr = ffile->sector_buff;

  //--- only read first block if we are going to use it ---
    if (pos < sector_size)  read_block(ffile->block_num, ffile->sector_buff);
  }


//===== new position is closer to end of file =====
  if (pos > ffile->pos)  {
    distance = pos - ffile->pos;

  //--- move forward within the existing contents of the sector buffer? ---
    if (distance < ((long)sector_size - bufOffset))  {
      ffile->buff_ptr += distance;
      ffile->pos += distance;
      return 0;
    }

    if (ffile->flags & FFLAG_WRITE)  fflush_file(ffile);

  //--- back up to the beginning of the sector buffer ---
    ffile->pos -= bufOffset;
    distance += bufOffset;
    ffile->buff_ptr = ffile->sector_buff;

  //--- follow sector links until we run out of sectors or we figure out ---
  //--- which one we need to read ---
    while (distance >= sector_size)  {
      nextBlock = getNextBlock(ffile->block_num);
      if ((nextBlock == EOF_BLK) OR !nextBlock)  return FERR_SEEK;
      distance -= sector_size;
      ffile->pos += sector_size;
      ffile->block_num = nextBlock;
    }

  //--- got the right sector number - just need to read it in to buffer ---
    ffile->pos += distance;
    ffile->buff_ptr = ffile->sector_buff + distance;
    read_block(ffile->block_num, ffile->sector_buff);

    return 0;
  }


//===== if we get here, the seek was to the same point =====
  return 0;
}


//-----------------------------------------------------------------------------
//  Write bytes to a file and return the # of bytes written
//-----------------------------------------------------------------------------
unsigned ffile_write(int handle, char far * buff, unsigned bytes)
{
  unsigned  numleft, wrote = 0, nxt, entry;
  FILE_DATA  *ffile = _ffile_list + handle;

  entry = ffile->entry_num;
  if (!bytes)  return wrote;

  if (ffile->block_num == 0)  {  // first write
    nxt = config_table->files[entry].startblock;
    if (!nxt)  {
      nxt = get_freeblock();
      if (nxt == 0)  return wrote;
      config_table->files[entry].startblock = nxt;
      setNextBlock(nxt, EOF_BLK);
    }
    else
      read_block(nxt, ffile->sector_buff);
    ffile->block_num = nxt;
  }

  while (bytes)  {
    numleft = sector_size - (uint)(ffile->buff_ptr - ffile->sector_buff);

    if (numleft)  {
      if (numleft > bytes)  numleft = bytes;
      fmemcpy(ffile->buff_ptr, buff, numleft);
      ffile->buff_ptr += numleft;
      bytes -= numleft;  buff += numleft;
      ffile->pos += numleft;
      if (ffile->pos > config_table->files[entry].size)
        config_table->files[entry].size = ffile->pos;
      wrote += numleft;
      ffile->flags |= FFLAG_WRITE;
    }

    if (bytes)  {
      write_block(ffile->block_num, ffile->sector_buff);
      ffile->buff_ptr = ffile->sector_buff;
      nxt = getNextBlock(ffile->block_num);
      if ((nxt == EOF_BLK) OR !nxt)  {
        nxt = get_freeblock();
        if (!nxt)  return wrote;
        setNextBlock(ffile->block_num, nxt);
        setNextBlock(nxt, EOF_BLK);
      }
      else
        read_block(nxt, ffile->sector_buff);
      ffile->block_num = nxt;
    }
  }

  return wrote;
}



//-----------------------------------------------------------------------------
//  Set the primary boot image to the file specified by its FAT index
//-----------------------------------------------------------------------------
int ffile_setPrimary(int index)
{
  if (!init_file_data())  return FERR_NOXMEM;

  config_table->primary = index + 1;
  if (config_table->backup == index + 1)  config_table->backup = 0;

  update_config();

  return 0;
}

//-----------------------------------------------------------------------------
//  Set the backup boot image to the file specified by its FAT index
//-----------------------------------------------------------------------------
int ffile_setBackup(int index)
{
  if (!init_file_data())  return FERR_NOXMEM;

  config_table->backup = index +1;
  if (config_table->primary == index +1)  config_table->primary = 0;

  update_config();

  return 0;
}

//-----------------------------------------------------------------------------
//  Return the FAT index of the primary boot image file (or < 0 if no primary
//  boot image file).
//-----------------------------------------------------------------------------
int ffile_getPrimary(void)
{
  int index;

  if (!init_file_data())  return FERR_NOXMEM;

  index = config_table->primary;

  return index - 1;
}

//-----------------------------------------------------------------------------
//  Return the FAT index of the backup boot image file (or < 0 if no backup
//  boot image file).
//-----------------------------------------------------------------------------
int ffile_getBackup(void)
{
  int index;

  if (!init_file_data())  return FERR_NOXMEM;

  index = config_table->backup;

  return index - 1;
}


//=============================================================================
//============================ NSCL additions =================================
//=============================================================================

#include <common.h>

//-----------------------------------------------------------------------------
//  Get info for file given its index in the flash file table.
//  Returns far ptr to file data if index is for a valid file, NULL if an error
//-----------------------------------------------------------------------------
FILE_ENTRY far *ffile_info(unsigned idx)
{
  FILE_ENTRY  far *fe;


  if (!init_file_data())  { ERRNO = ENOMEM;  return NULL; }

  if (idx >= MAX_FILES)  return NULL;

  fe = config_table->files + idx;

  if ((fe->flags & (FILE_ACTIVE | FILE_SFLASH)) != (FILE_ACTIVE))  return NULL;

  return fe;
}

//-----------------------------------------------------------------------------
//  Return info in standard fstat stat struct for an open file
//-----------------------------------------------------------------------------
int  ffile_fstat(int handle, struct stat *fsbuf)
{
  FILE_ENTRY  far *fe;


  if (!init_file_data())  { ERRNO = ENOMEM;  return -1; }

  fe = config_table->files + _ffile_list[handle].entry_num;

  fsbuf->st_size  = fe->size;  // total size in bytes
  fsbuf->st_atime = fe->date;  // time of last access
  fsbuf->st_mtime = fe->date;  // time of last modification
  fsbuf->st_ctime = fe->date;  // time of last status change

  fsbuf->st_mode = S_IFREG | 00777;

  return 0;
}

//-----------------------------------------------------------------------------
//  Return the number of bytes the cur pos is from the end of the file
//-----------------------------------------------------------------------------
long ffile_unread(int handle)
{
  FILE_DATA  *ffile;  // equiv to file descriptor
  FILE_ENTRY  far *entry;  // global info on the file


  ffile = _ffile_list + handle;
  entry = config_table->files + ffile->entry_num;

  return (entry->size - ffile->pos);
}


//-----------------------------------------------------------------------------
//  Return the IPv4 the SBC is configured for in host byte order
//-----------------------------------------------------------------------------
ulong getIPAddr(void)
{
  if (!init_file_data())  { ERRNO = ENOMEM;  return -1; }

  return ntohl(config_table->IPAddress);
}


//-----------------------------------------------------------------------------
//  Change the IPv4 stored in the configuration data.  The address must be in
//  host byte order.
//-----------------------------------------------------------------------------
int  setIPAddr(ulong ipAddr)
{
  if (!init_file_data())  { ERRNO = ENOMEM;  return -1; }

  config_table->IPAddress = htonl(ipAddr);
  update_config();

  return 0;
}

//-----------------------------------------------------------------------------
//  Return the name of the primary boot image file (or NULL 0 if no primary
//  boot image file).
//-----------------------------------------------------------------------------
char *ffile_getPrimaryName(char *name, int nameLen)
{
  int index;

  if (!init_file_data())  { ERRNO = ENOMEM;  return NULL; }

  index = config_table->primary;
  if (index < 1)  return NULL;

  fstrcpy(name, config_table->files[(uint)index-1].name);

  return name;
}

//-----------------------------------------------------------------------------
//  Return the name of the backup boot image file (or NULL 0 if no backup
//  boot image file).
//-----------------------------------------------------------------------------
char *ffile_getBackupName(char *name, int nameLen)
{
  int index;

  if (!init_file_data())  { ERRNO = ENOMEM;  return NULL; }

  index = config_table->backup;
  if (index < 1)  return NULL;

  fstrcpy(name, config_table->files[(uint)index-1].name);

  return name;
}

//-----------------------------------------------------------------------------
//  change the name of a file
//-----------------------------------------------------------------------------
int ffile_rename_(const char *oldpath, const char *newpath)
{
  int index, stat;


  stat = -1;  KLOCK();

  for (;;)  {
    if (!init_file_data())  { ERRNO = ENOMEM;  break; }

//printf("rename(%.16s, %.16s)\r\n", oldpath, newpath);  //tdebug
    index = ffile_index((char *)oldpath);
    if (index < 0)  { ERRNO = ENOENT;  break; }

//printf("  delete(%.16s)\r\n", newpath);  //tdebug
    ffile_delete((char *)newpath);

//printf("  change name of files[%d](%.16s) to %.16s\r\n",  //tdebug
//       index, config_table->files[(uint)index].name, newpath);  //tdebug
    fstrcpy(config_table->files[(uint)index].name, newpath);

    update_config();

    stat = 0;  break;
  }

  KUNLOCK();

  return stat;
}


//-----------------------------------------------------------------------------
//  Set the primary boot image to the file specified by its name
//-----------------------------------------------------------------------------
int ffile_makePrimary(const char *fileName)
{
  int  index;


  if (!init_file_data())  { ERRNO = ENOMEM;  return -1; }

  index = ffile_index((char *)fileName);
  if (index < 0)  { ERRNO = ENOENT;  return -1; }

  ffile_setPrimary(index);

  return 0;
}

//-----------------------------------------------------------------------------
//  Set the backup boot image to the file specified by its name
//-----------------------------------------------------------------------------
int ffile_makeBackup(const char *fileName)
{
  int  index;

  if (!init_file_data())  { ERRNO = ENOMEM;  return -1; }

  index = ffile_index((char *)fileName);
  if (index < 0)  { ERRNO = ENOENT;  return -1; }

  ffile_setBackup(index);

  return 0;
}

//-----------------------------------------------------------------------------
//  Set date/time stamps on a file
//-----------------------------------------------------------------------------
int  ffile_utime(const char *fileName, const struct utimbuf *buf)
{
  int  index;
  FILE_ENTRY  far *fe;


  if (!init_file_data())  { ERRNO = ENOMEM;  return -1; }

  index = ffile_index((char *)fileName);
  if (index < 0)  { ERRNO = ENOENT;  return -1; }

  fe = config_table->files + (uint)index;

  fe->date = (ulong) buf->modtime;

  update_config();

  return 0;
}


//-----------------------------------------------------------------------------
//  set the board name
//-----------------------------------------------------------------------------
int  ffile_setName(char *newName)
{
  int  len;

  if ((len = strlen(newName)) >= BRD_NAME_LEN)  return -1;

  fmemcpy(config_table->board_name, newName, len);
  config_table->board_name[(uint)len] = 0;

  update_config();

  return 0;
}


//-----------------------------------------------------------------------------
//  get the board name
//-----------------------------------------------------------------------------
char *ffile_getName(char *name, int maxLen)
{
  int  len;

  if ((len = (uint)fstrlen(config_table->board_name)) >= maxLen)  len = maxLen-1;

  fmemcpy(name, config_table->board_name, len);  name[len] = 0;

  return name;
}


//-----------------------------------------------------------------------------
//  check the integrity of the file system
//
//  Note that if the file system is altered while this is running the results
//  may be incorrect.
//
//  Returns the # of errors found.
//
//  The repair parameter is currently ignored...
//-----------------------------------------------------------------------------
int ffile_fsck(bool repair)
{
  int  numErrs;
  uint  ui, numBlks, ttlBlks, usedBlks, freeBlks, rsvdBlks;
  ulong  spaceUsed, usedBytes, freeBytes, rsvdBytes;
  ushort  blkNum, nextBlk, sectorSize;
  uchar  mask;
  uchar *chkdFlag;
  FILE_ENTRY far *file;
  uchar chkdBlocks[MAX_BLK_NUM/8];


  BZERO(chkdBlocks);  numErrs = usedBlks = freeBlks = rsvdBlks = 0;

  ttlBlks = config_table->num_blocks;  sectorSize = config_table->sector_size;

  printf("\r\n%u bytes/sector,  %u sectors,  Total %lu bytes\r\n",
         config_table->sector_size, config_table->num_blocks,
         (ulong) ttlBlks * sectorSize );


//--- verify each file entry by following the linked list of blocks ---
  for (ui=0; ui<MAX_FILES; ++ui)  {
    file = config_table->files + ui;
    printf("\r\nentry %2u: ", ui);

    if ((file->flags & FILE_ACTIVE) == 0)  {
      printf("<unused>\r\n");  continue; }

    printf("%ls\r\n", file->name);
    spaceUsed = 0;  blkNum = file->startblock;  numBlks = 0;

    for (;;)  {
      if (blkNum == EOF_BLK)  break;

      if (!blkNum OR (blkNum >= MAX_BLK_NUM))  {
        printf("  Invalid sector # %u\r\n", blkNum);  ++numErrs;  break; }

      ++numBlks;

      chkdFlag = chkdBlocks + (blkNum >> 3);  mask = 1 << (blkNum & 7);
      if (*chkdFlag & mask)  {
        printf("  Cross-linked sector # %u\r\n", blkNum);  ++numErrs; }

      *chkdFlag |= mask;  spaceUsed += sectorSize;
      blkNum = config_table->nextBlock[blkNum];
    }
    printf("  %3d sectors:  Size: %6lu   on disk: %6lu\r\n",
           numBlks, file->size, spaceUsed);

    usedBlks += numBlks;

    if ((spaceUsed >= file->size) AND ((spaceUsed - file->size) <= sectorSize))
      continue;

    printf("  Size error\r\n");  ++numErrs;
  }
  printf("\r\n");


//--- check all the unused blocks ---
  for (ui=0; ui<ttlBlks; ++ui)  {
    chkdFlag = chkdBlocks + (ui >> 3);  mask = 1 << (ui & 7);
    if (*chkdFlag & mask)  continue;
    nextBlk = config_table->nextBlock[ui];
    if (!nextBlk) { ++freeBlks;  continue; }
    if (nextBlk == EOF_BLK)  { ++rsvdBlks;  continue; }
    printf("Unused sector # %4u linked to sector # %4u\r\n", ui, nextBlk);
    ++numErrs;
  }


  usedBytes = (ulong) usedBlks * sectorSize;
  freeBytes = (ulong) freeBlks * sectorSize;
  rsvdBytes = (ulong) rsvdBlks * sectorSize;

  printf(" used: %6lu bytes (%4u blocks)\r\n"
         " rsvd: %6lu bytes (%4u blocks)\r\n"
         " free: %6lu bytes (%4u blocks)\r\n",
          usedBytes, usedBlks, rsvdBytes, rsvdBlks, freeBytes, freeBlks );

  printf("\r\n %d errors\r\n\n", numErrs);

  return numErrs;
}


//-----------------------------------------------------------------------------

