/*----- fcntl.h ----- 12/05/12 ----(03/11/06)-----*/

#ifndef _fcntlH_
#define _fcntlH_


#include <sys/types.h>

#include <linux/wait.h>

#include <linux/limits.h>

#include <sys/ioctl.h>

//-----------------------------------------------------------------------------
//  Update log (most recent first):
//
//
//  2012-12-05 - Mark Davis
//               - Moved the following from fcntl to common:
//                    vt100 escape sequence definitions
//                    DispStr() and DispStrAttr()
//                    read_() and write_()
//                    redirectStdio() and logStdout()
//
//  2012-09-20 - Mark Davis
//               - Changed DARK_xxx macros to not use the DIM attribute,
//                 which causes OS-X terminals to ignore the color attribute.
//
//  2012-04-20 - Mark Davis
//               - Added gfdCounts() to get total and free counts for the
//                 global file descriptor table.
//
//  2012-03-07 - Mark Davis
//               - Added rtosVers and invInfoKey to DiagLogInfo struct
//
//  2011-03-17 - Mark Davis
//               - Removed the nestedFDs[][] table as it is no longer needed.
//
//  2011-03-11 - Mark Davis
//               - Added readFDEV() and writeFDEV()
//
//  2011-03-11 - Mark Davis
//               - Changed the bufSize argument for the driver's descr function
//                 from type char to type int
//
//  2011-02-04 - Mark Davis
//               - Added nestedFDs[][] for better supporting drivers like the
//                 revised SRFS one where one fd can be referenced by other fds
//
//  2010-12-14 - Mark Davis
//               - Changes to DiagLogInfo:  Replaced bootTime with hdwResetTime
//                 and added rebootTime (which is reset for any reboot,
//                 regardless of whether or not it involved a CPU reset).
//
//  2010-12-08 - Mark Davis
//               - Added didSoftReset flag to DiagLogInfo
//
//  2010-10-26 - Mark Davis
//               - Added _validFDEV_() and _addFileRef_()
//
//  2010-08-13 - Mark Davis
//               - Changed 'logStdio' to 'logStdout' in the task structure
//               - Added redirStdio flag to the task structure, which the
//                 redirectStdio() macro now sets in addition to the logStdout
//                 flag.  Setting both now causes redirectStdio() to work as
//                 originally intended:  Leave fd's 0,1,2 alone - just don't
//                 use them for stdin, stdout, stderr.
//               - Added the logStdout() macro that does what redirectStdio
//                 WAS doing: DUPLICATES stdout in the system log, rather than
//                 redirecting it.
//
//  2010-06-02 - Mark Davis
//               - Added openCrashLog() and closeCrashLog()
//
//  2010-05-13 - Mark Davis
//               - Added read_() and changed write_() to they can be used as
//                 more efficient replacements for readBytes() and writeBytes()
//
//  2010-05-12 - Mark Davis
//               - Added FIONSPACE for getting the # of unused bytes in a
//                 devices output buffer.  This is apparently standard on BSD
//                 systems but the closest I can find for debian systems is
//                 TIOCOUTQ (which is supposed to tell you the # of unsent
//                 bytes, rather than the # of free bytes).
//
//  2010-01-11 - Mark Davis
//               - Added intStackIdx and intStack[] to DiagLogInfo struct for
//                 keeping track of nested interrupts
//
//  2009-12-07 - Mark Davis
//               - Added copy of timeZone_ value to the values stored in
//                 the battery-backed memory (and restored on bootup)
//
//  2009-03-25 - Mark Davis
//               - Added unlink() and rename() to the FileSysFuncs API
//
//  2009-03-24 - Mark Davis
//               - added entry() func to FileSysFuncs API to be used to get
//                 pointers to and info about entries in a drivers list of
//                 ioDev's.
//      ...
//
//  2006-03-11 - Mark Davis
//               - original version started
//
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
// Unix compatible implementation of a file system:
// ------------------------------------------------
// - Every device of interest outside the kernel supports a standard API
//   (which does not mean it can't support other API's in addition, like
//   sockets do)
//
// - There is a global table of files keeping track of every open device
//   being accessed via the file system
//
// - Each process has a table of file descriptors, where each open file
//   descriptor points to one of the files in the global file table.  All
//   access by a process to devices via the file system uses one of the file
//   descriptors (specified either as in index or a pointer).  Pointers to
//   entries in the global file table are NEVER used outside the kernel.
//
//   stdin is fd[0] (file descriptor 0), stdout is fd[1], and stderr is fd[2]
//
//   NOTE that multiple descriptors (within a process and/or across processes)
//   can point to the same global file table entry.  The most common example is
//   stdio group (stdin, stdout, and stderr).
//-----------------------------------------------------------------------------


typedef struct  {
  WaitQueue  *readQueue;
  WaitQueue  *writeQueue;
}  WaitQueues;


extern const WaitQueues sleepQueues, NullQueues;


//---------------------------------------------------------------------
// ===== Driver functions for operations on an already open file =====
//---------------------------------------------------------------------

typedef struct  {
  int   (*read) (void *ioDev, char *buf, int nbyte, bool dontBlock);
  int   (*write)(void *ioDev, char *buf, int nbyte, bool dontBlock);
  int   (*close)(void *ioDev);
  off_t (*lseek)(void *ioDev, off_t offset, int whence);
  int   (*fstat)(void *ioDev, struct stat *buf);
  int   (*ioctl)(void *ioDev, int request, void *arg);
  char *(*descr)(void *ioDev, char *buf, int bufSize);
  int   (*getQueues)(void *ioDev, WaitQueues far *queues);
}  IODevFuncs;


#define IODEV_FUNCS(name)  \
  int   _##name##Read_ (void *ioDev, char *buf, int numBytes, bool dontBlock); \
  int   _##name##Write_(void *ioDev, char *buf, int numBytes, bool dontBlock); \
  int   _##name##Close_(void *ioDev);  \
  off_t _##name##Lseek_(void *ioDev, off_t offset, int whence);  \
  int   _##name##Fstat_(void *ioDev, struct stat *buf);  \
  int   _##name##ioctl_(void *ioDev, int request, void *arg);  \
  char *_##name##Descr_(void *ioDev, char *buf, int bufSize);  \
  int   _##name##GetQueues_(void *ioDev, WaitQueues far *queues);  \
  \
  const IODevFuncs  name##IOFuncs = {  \
    _##name##Read_,      \
    _##name##Write_,     \
    _##name##Close_,     \
    _##name##Lseek_,     \
    _##name##Fstat_,     \
    _##name##ioctl_,     \
    _##name##Descr_,     \
    _##name##GetQueues_  \
  };


//------------------------------------------------------------------------
// ===== Driver level functions NOT related to an already open file =====
//------------------------------------------------------------------------

typedef struct  {
  int   openCount;  // # global FDs referencing the device entry
  bool  unused;     // # device entry free for use
} EntryState;

typedef struct  {
  const char *name;            // name for the file system driver
  const IODevFuncs  *ioFuncs;  // for operating on an open device/file
  void  (*tick)(void);         // called periodically to give driver CPU cycles
  void *(*entry)(uint idx, EntryState *entryState);    // get state of fs node
  fd_t  (*open)(const char *path, int flags, int mode);
  int   (*unlink)(char *path);
  int   (*rename)(const char *oldPath, const char *newPath);
//  int   (*utime)(const char *fileName, const struct utimbuf *buf);
}  FileSysFuncs;


#define FILESYS_FUNCS(name)  \
  void  _##name##Tick_(void); \
  void *_##name##Entry_(uint idx, EntryState *entryState);  \
  fd_t  _##name##Open_(const char *path, int flags, int mode);  \
  int   _##name##Unlink_(char *path);  \
  int   _##name##Rename_(const char *oldPath, const char *newPath);  \
  \
  const FileSysFuncs  name##FSysFuncs = {  \
    #name,             \
    &##name##IOFuncs,  \
    _##name##Tick_,    \
    _##name##Entry_,   \
    _##name##Open_,    \
    _##name##Unlink_,  \
    _##name##Rename_   \
  };




//--- information about an open "file" ---
typedef struct {
  int          openCount;      // 0 means unused entry
  const IODevFuncs  *ioFuncs;  // driver funcs for specific device
  void        *ioDev;          // instance of a specific device
  WaitQueues   waitQueues;     // queues used for blocking caller
  int          flags;          // flags that affect oper (e.g. O_NONBLOCK)
  int          mode;           // permissions
  ulong        readCount;      // ttl # of bytes read since opened
  PID          rdInProgress;   // pid of process in middle of read() call
  int          rdStat;         // status last returned by driver's read func
  int          rdErr;          // errno last returned by driver's read func
  bool         dontCallRead;   // driver returned EOF or error last time
  ulong        writeCount;     // ttl # of bytes written since opened
  PID          wrInProgress;   // pid of process in middle of write() call
  int          wrStat;         // status last returned by driver's write func
  int          wrErr;          // errno last returned by driver's write func
  bool         dontCallWrite;  // driver returned EOF or error last time
  bool         eof;            // EOF reached
  bool         istty;          // is a terminal interface
  bool         shared;         // normally "orphaned" (no task fd's ref it)
} FDEV;




//-----------------------------------------------------------------------------

//--- all open files in the sytem ---
extern FDEV  far fileTbl[MaxOpenFiles];

//--- all references to open files by process ---
extern FDEV far *FD[NUM_TASKS][FD_SETSIZE];   // NULL ==> not in use


//-----------------------------------------------------------------------------
//NOTE: !!! DiagLogInfo struct must stay at 249 bytes for softReset to work !!!
//-----------------------------------------------------------------------------
typedef struct  {
  ulong  infoKey;         // used to detect buffer from before reboot/reset

  ulong  bbCount;         // # of bytes allocated by bb_xalloc()

  uint   toIdx;           // idx to write next char to
  uint   fromIdx;         // idx of next char to copy to backup copy on server
  bool   wrapped;         // entire buffer used (i.e. it wrapped around)

  bool   ctldShutdown;    // reboot followed a controlled shutdown

  uint   wdtTicks;        // # of ms since watchdog timer last reset

  ulong  seconds;         // time of last timer int (whole seconds portion)
  uint   milliSecs;       // time of last timer int (frac seconds portion)

  ulong  msTime;          // copy of MS_TIMER
  char   filler1[80];     // chgd 2012-03-15:  was: ulong taskTime[NUM_TASKS]

  int    tz;              // copy of timeZone_ value

  ulong  hdwResetTime;    // time of last CPU reset (in seconds)
  ulong  rebootTime;      // time we last rebooted (w/ or w/o CPU reset)

  bool   didSoftReset;    // Restarted OS w/o doing a CPU reset

  ulong  digOutState;     // state of digital outputs during a soft reset

  // The following were added on 2012-03-15
  ulong  rtosVers;        // date/time of RTOS vers that filled this struct
  ulong  timerSeconds;    // seconds since hdwResetTime (above)

  ulong  invInfoKey;      // inverted copy of infoKey values

  // New items taken from the start of filler2[]
  char   filler2[116];    // For future expansion
} DiagLogInfo;
//-----------------------------------------------------------------------------
//NOTE: !!! DiagLogInfo struct must stay at 249 bytes for softReset to work !!!
//-----------------------------------------------------------------------------

#define  LogInfoKey  (0xFACE1236L)

//--- values for rebootFlag ---
//--- A value other than shown here means we don't know why we rebooted ---
//--- or even IF we rebooted (i.e. maybe a jump to 0 occured) ---
#define  RDLM_Reboot  0xA5B1  // set by an RDLM Init program (for future use)
#define  WD_Reboot    0xA5B2  // called REBOOT() to force a WD timer reboot


extern DiagLogInfo far  *diagLogInfo;
extern uint  DiagLogSize;
extern char far  *diagLogBuf;   // pointer to buffer of DiagLogSize bytes
extern bool  logDateTime;

extern bool  noGettyOnA;

//-----------------------------------------------------------------------------

//--- Macros/funcs for accessing an FDEV's attributes ---
#define  File_OpenCount(file)  ( ((FDEV far *)(file))->openCount )
#define  File_Funcs(file)      ( ((FDEV far *)(file))->ioFuncs )
#define  File_Dev(file)        ( ((FDEV far *)(file))->ioDev )
#define  File_ID(file)         \
       ( ((uint)((ulong)(file) - (ulong)fileTbl)) / sizeof(FDEV) )
#define  GFD(pfd)  ( File_ID(FD[userPix][pfd]) )

void fcntlInit(void);

int registerFileSys(const FileSysFuncs *fsFuncs);  // register a file system

void fileSysTick(void);  // called to give file system drivers CPU cycles


void setIstty(fd_t fd, bool istty);


bool  _validFDEV_(FDEV far *file);

fd_t _addFileRef_(FDEV far *file);

fd_t _allocFile_(fd_t minFD, const IODevFuncs *ioFuncs, void *ioDev);

FDEV far *_FileFromFD_(fd_t fd);

FDEV far *_fdToOpenFile_(fd_t fd);


int _File_Swap_(FDEV far *f1, FDEV far *f2);

int _closeFD_(task_t *task, fd_t fd);


// void flush_stdout(void);
// void _flushStdout_(void);

int _nearcall rcvChar(void *userData);
int _nearcall xmitChar(uchar ch, void *userData);
//long _nearcall xmitBytes(const uchar far *buf, long count, void *userData);
//long _nearcall rcvBytes(uchar far *buf, long count, void *userData);

void openCrashLog(void);
void closeCrashLog(void);

int _nearcall diagIn(void *dummy);
int _nearcall diagOut(uchar ch, void *dummy);

int writeFDEV(FDEV far *file, void *buf, int numBytes, long maxIdleMS);
int readFDEV(FDEV far *file, void *buf, int numBytes, long maxIdleMS);


uint _nearcall gfdCounts(uint *ttl);

void fileSysCheck(bool closeOrphans, char *buf, int bufSize);


//-----------------------------------------------------------------------------
//  Unix standard function prototypes
//-----------------------------------------------------------------------------

int isatty(fd_t fd);

fd_t dup2(fd_t fromFD, fd_t toFD);


fd_t open(const char *path, int oflag, int mode);

//--- values for oflag ---
//--- must not conflict with definitions in Softools .h files
#define O_RDONLY         1   // Same value as O_READ in zserver.h
#define O_WRONLY         2   // Same value as O_WRITE in zserver.h
#define O_RDWR           2   // Alias for O_WRITE, since O_WRITE implies O_READ.
#define O_CREAT     0x0004   // open with file create (uses third arg)
#define O_TRUNC     0x0008   // open with truncation
#define O_APPEND    0x0010   // append (writes guaranteed at the end)
#define O_NONBLOCK  0x0020   // non-blocking I/O (POSIX)
#define O_NDELAY    0x0020   // non-blocking I/O
#define O_EXCL      0x0040   // exclusive open


int close(fd_t fd);

int read(fd_t fd, void *buf, int nbyte);

int write(fd_t fd, void *buf, int nbyte);


off_t lseek(fd_t fd, off_t offset, int whence);

//--- values for whence ---
#define  SEEK_SET  0   // from start of file
#define  SEEK_CUR  1   // from current position
#define  SEEK_END  2   // from end of file


int fcntl(fd_t fd, int cmd, int arg);

//--- values for cmd ---
#define F_GETFL         3       // Get file status flags.
#define F_SETFL         4       // Set file status flags.


//--- values for arg when cmd == F_GETFL or F_SETFL ---
#define FNDELAY     0x04


int unlink(const char *pathname);


//-------------------- stdio declarations ------------------------

int fflush(FILE *stream);

int rename(char const *oldpath, const char *newpath);


//--------------------- ioctl declarations ------------------------

int ioctl(int fd, int request, void *arg);


//finish:  These should be changed to use the macros in the sys/ioctl.h file
//         that provide a standard way to define the request values

//--- values for request ---
#define  IOC_GET_BAUD        1001  // get baud rate for async port
#define  IOC_SET_BAUD        1002  // set baud rate for async port
#define  IOC_SET_PKTMODE     1003  // set pkt mode for async port
#define  IOC_END_PKT_RX      1004  // tell ser lib we got an entire pkt
#define  IOC_START_PKT_TX    1005  // tell serial lib to chg to xmit mode
#define  IOC_END_PKT_TX      1006  // tell serial lib we wrote the entire pkt
#define  IOC_SET_ASYNC_FMT   1007  // set # data bits and parity options
#define  IOC_SET_FLOW_CHARS  1008  // set flow control chars

#define  IOC_SET_CLK_RATE    1100  // SPI port clk rate (see spi.h for values)
#define  IOC_SET_CLK_MODE    1101  // SPI port clk mode (see spi.h for values)

#define  FIONREAD          0x541B  // get # of unread bytes in input buffer
#define  TIOCINQ         FIONREAD
#define  FIONSPACE         0x6000  // get # of unused bytes in output buffer


//--- bit usage for IOC_SET_ASYNC_FMT --- (arg must point to unsigned int) ---
#define  ASYNC_FMT_7O1    0x0711  // 7 data bits, Odd parity,  1 stop bit
#define  ASYNC_FMT_7E1    0x0721  // 7 data bits, Even parity, 1 stop bit
#define  ASYNC_FMT_8O1    0x0811  // 8 data bits, Odd parity,  1 stop bit
#define  ASYNC_FMT_8E1    0x0821  // 8 data bits, Even parity, 1 stop bit

//--- values for FIFOs ---
#define  PIPE_MODE        0xFF11  // FIFO used as a one-way pipe


//-----------------------------------------------------------------------------
#endif
