//----- sched.h ----- 06/07/12 -----(02/20/07)-----

#ifndef  _schedH_
#define  _schedH_


//-----------------------------------------------------------------------------
// DEBROS version of Linux sched.h file
//-----------------------------------------------------------------------------
//  Update log (most recent first):
//
//
//  2012-06-07 - Mark Davis
//               - Added taskCounts() to return max and free counts for the
//                 task table.
//
//  2012-04-06 - Mark Davis
//               - Changed typedef PIX to int to generate more efficient
//                 assembly/machine code.
//
//  2012-03-01 - Mark Davis
//               - Added Schedule_PushCount for use in other modules.
//
//  2011-03-25 - Mark Davis
//               - Added myAddr ptr for use with VALID() and INVALID() macros
//
//  2011-02-21 - Mark Davis
//               - Added numReadyTasks
//
//  2010-05-13 - Mark Davis
//               - Eliminated useTimeout.  Each function that needs to use
//                 a timeout must now call schedule_timeout() (which sets
//                 the timeoutTicks values in the task struct).
//
//  2010-04-22 - Mark Davis
//               - Removed waitq from the task structure because it has not
//                 been used for some time, and fixed the comments on the
//                 wq field so they are correct.
//
//-----------------------------------------------------------------------------

#define  __KERNEL__
#define  __inline__

#include <signal.h>
#include <linux/list.h>
#include <linux/wait.h>

#include <linux/limits.h>


//--- The number of bytes schedule() pushes on to the stack of a task before ---
//--- it suspends it ---
#define  Schedule_PushCount  5



#define F_NULL ((void far *)0l)

typedef int    PID;
typedef int    PIX;


typedef void _nearcall (* ProcMain)(int argc, char **argv);


//--- values for task_t.state
#define  TASK_RUNNING              0  // Executing or waiting to be executed
#define  TASK_INTERRUPTIBLE   0x0001  // waiting for cond but processing signals
#define  TASK_UNINTERRUPTIBLE 0x0002  // waiting for cond and ignoring signals
#define  TASK_STOPPED         0x0004  // stopped in response to a signal
#define  TASK_TRACED          0x0008
#define  EXIT_ZOMBIE          0x0010  // Terminated before parent called waitXXX()
#define  EXIT_DEAD            0x0020  // Terminated & being removed
#define  TASK_NONINTERACTIVE  0x0040
#define  TASK_DEAD            0x0080
#define  TASK_UNUSED          0x8000  // never used entry in the task table

#define  ActiveTask(taskPtr)   ((taskPtr)->state < TASK_STOPPED)
#define  BlockedTask(taskPtr)  ((taskPtr)->state & 0x03)


//--- offsets in to task structure ---
#define STACK_OFF        0
#define SSEG_OFF         6

#define STKLLR_OFF       7

#define WPSAR_OFF        9
#define WPSALR_OFF      10

#define WPSBR_OFF       12
#define WPSBLR_OFF      13

struct task_struct  {
  uint  *stack;         // SP value when last preempted or called schedule()
  uint  *stack_start;   // "logical" portion of addr (w/o segment portion)
  uint   stack_size;    // size of stack (in bytes)
  uchar  sseg;          // STACKSEG portion of stack address

// for hdw enforced stack limits
  uchar  stkllr;        // value for STKLLR (STacK Limit Low Register)
  uchar  stkhlr;        // value for STKHLR (STacK Limit High Register)

// for hdw enforced write protection (4K res for 2 different 64K segments)
  uchar  wpsar;         // value for WPSAR (Write-Protect Segment A Register)
  uint   wpsalhr;       // value for WPSALR and WPSAHR
  uchar  wpsbr;         // value for WPSBR (Write-Protect Segment B Register)
  uint   wpsblhr;       // value for WPSBLR and WPSBHR

  void far **FDSet;     // ptr to array of ptrs to global FD's
  uchar   useCount;     // # times this entry has been used before
  bool    claimed;      // slot claimed for use - may not be initialized yet
  uint    state;        // used for scheduling - see above

  struct task_struct  *myAddr;  // for use with VALID() and INVALID() macros

  //-------------------------------------------------
  bool    zeroAfter;    // everything after this cleared to 0's for new task...

  int     argc;         // count of arguments
  char  **argv;         // argument list
  char    args[MaxArgBytes];  //finish: relocate to stack?
  // (but anything on stack requires a far ptr to access from another task)

  WaitQueue  waitQs;    // list of wait queues task is blocked on

  sigset_t  sig_pending;  // mask of signals that are pending
  sigset_t  sig_blocked;  // mask of signals that are blocked

  int     preempt_count; // preempted b4 call to schedule(), still runable
  bool    inUse;        // slot is in use by a process
  bool    exit;         // exit value

  bool    logStdout;     // duplicate stdout in the system log
  bool    redirStdio;    // stdio NOT Connected to fd's 0,1,2

  ulong   diagFlags;    // type of diagnostic output desired

  ProcMain  procMain;   // address of main() func for process

  PID     pid;          // process id: MSB=useCount, LSB=idx into this table
  PID     ppid;         // parent process id

  uchar   priority;     // task priority
  int     lastErr;      // errno value for this process

  time_t  started;      // time this process was started
  ulong   lastResumed;  // time process was last resumed (in ms)
  ulong   lastStopped;  // time process was last suspended (in ms)
  ulong   lastRunning;  // time process was last running (in ms)

  ulong   resumedTicks; // time process was last resumed (in clock ticks)
  ulong   ticksUsed;    // # ticks used (only bits below 1ms)
  ulong   msUsed;       // total time used in ms

  uchar   outCount;     // # of chars in outBuf
  char far *outPtr;     // ptr to next avail position in outBuf

  ulong   sleepTicks;   // # of scheduler ticks to sleep for
  ulong   sleepStarted; // to determine time sleeping

  ulong   timeoutTicks; // ticks remaining until blocking operation times out
};
typedef  struct task_struct  task_t;



#define NO_DEBUG_PARSE // for udpdebug

//----- global variables -----
extern char  _disabled;  // != 0 means scheduler is disabled
extern char  _pending;   // sched() was called while scheduler was disabled

extern uint  schedCalls;    // # of calls to schedule() since last zeroed
extern uint  taskSwitches;  // # of task switches since last zeroed

extern task_t   tasks[];

extern task_t  *current;
extern PIX  userPix;

extern task_t  *endTasks, *idleTask, *applTask, *kernelTask;

extern ulong  newTaskDiagFlags;   // diagFlags for new tasks

extern uint  SchedMS;  // scheduler tick interval in ms
extern uint  numReadyTasks;  // # tasks ready to run during last call to schedule()

extern bool  stacksCross64K;
extern byte  stackSegL, stackSegH, stackPageL, stackPageH;


//----- initialization -----
int init_tasks(void);


//----- task switching -----
void _nearcall schedule(void);

//----- task restart -----
int reset_task(int tasknum);

//----- stack ckecking -----
int check_stack(int tasknum);

//--- max and cur # of tasks ---
uint _nearcall taskCounts(uint *ttl);


//========================= some standard Linux stuff =========================

#define set_current_state(state_value)      \
  do { current->state = (state_value); } while (0)

#define __set_current_state(state_value)      \
  set_current_state(state_value)


//#define __sched   __attribute__((__section__(".sched.text")))
#define __sched

#define MAX_SCHEDULE_TIMEOUT  LONG_MAX
extern signed long FASTCALL(schedule_timeout(signed long timeout));
extern signed long schedule_timeout_interruptible(signed long timeout);
extern signed long schedule_timeout_uninterruptible(signed long timeout);

//=============================================================================

#endif
