//----- poll.c ----- 02/08/12 -----(04/30/10)-----

//-----------------------------------------------------------------------------
// DEBROS version of standard Linux poll.c source file
//-----------------------------------------------------------------------------
//  Change log (most recent first):
//
//
//  2012-02-08 - Mark Davis
//               - poll() now sets the POLLHUP bit if the events flag for an fd
//                 has POLLIN or POLLOUT set and the device driver ioctl() func
//                 sets errno to ENOLINK
//
//  2012-01-20 - Mark Davis
//               - Changed path for poll.h
//
//  2010-05-21 - Mark Davis
//               - first version completed and passed basic tests
//
//-----------------------------------------------------------------------------

#include <common.h>

#include <linux/dummy.h>

#include <sys/types.h>
#include <string.h>


#include <linux/sched.h>
#include <linux/wait.h>
#include <linux/sched.h>

#include <poll.h>

#include <errno.h>

#include <sys/stat.h>
#include <fcntl.h>


#include <unistd.h>

#include <limits.h>


//-----------------------------------------------------------------------------
//  wait for some event to occur on one or more file descriptors
//-----------------------------------------------------------------------------
int poll(struct pollfd *ufds, uint nfds, int timeout)
{
  bool  block;
  int  err, stat;
  uint  u, n, numNodes, numBytes;
  ulong  timeoutTicks = 0;
  struct pollfd  *ufd;
  WaitQueueNode  *node[WQ_POOL_SIZE];
  WaitQueue  *queue[WQ_POOL_SIZE];
  FDEV far  *file;


  if (nfds > RLIMIT_NOFILE)  { errno = EINVAL;  return -1; }

//--- Create a set of wait queues and wait queue nodes based on the event ---
//--- bits that are set for each valid open fd ---

  err = 0;  if (timeout > 0)  timeoutTicks = msToTicks((ulong)timeout);

  for (u=n=0,ufd=ufds; u<nfds; ++u,++ufd)  {
    ufd->revents = 0;

    if (ufd->fd < 0)  continue;  // -1 fd means ignore the entry

    if (! (file = _fdToOpenFile_(ufd->fd)))  {
      err = EBADF;  ufd->revents |= POLLNVAL;  continue; }

    if (ufd->events & POLLIN)  {
      if (! (file->flags & (O_RDONLY | O_RDWR)) )  {
        err = EBADRQC;  ufd->revents |= POLLERR;  continue; }
      if (file->dontCallRead)  {
        err = file->rdErr;  ufd->revents |= POLLERR;  continue; }
      if (timeout)  {
        if (! (queue[n] = file->waitQueues.readQueue))  continue;
        if (! (node[n] = alloc_WaitQueueNode()))  {
          err = ENOMEM;  ufd->revents |= POLLERR;  break; };
        node[n]->flags &= ~WQ_FLAG_EXCLUSIVE;  ++n;
      }
    }

    if (ufd->events & POLLOUT)  {
      if (! (file->flags & (O_WRONLY | O_RDWR)) )  {
        err = EBADRQC;  ufd->revents |= POLLERR;  continue; }
      if (file->dontCallWrite)  {
        err = file->wrErr;  ufd->revents |= POLLERR;  continue; }
      if (timeout)  {
        if (! (queue[n] = file->waitQueues.writeQueue))  continue;
        if (! (node[n] = alloc_WaitQueueNode()))  {
          err = ENOMEM;  ufd->revents |= POLLERR;  break; };
        node[n]->flags &= ~WQ_FLAG_EXCLUSIVE;  ++n;
      }
    }

  }
  numNodes = n;

  if (err)  {
    for (n=0; n<numNodes; ++n)  free_WaitQueueNode(node[n]);
    errno = err;  return -1;
  }


//===== Block until woken by an event, timeout, error, etc =====
  block = (timeout AND numNodes);

  for (;;)  {

  //--- The equivalent to prepare_to_wait() ---
    if (numNodes)  {
      KLOCK();
      for (n=0; n<numNodes; ++n)  __add_wait_queue(queue[n], node[n]);
      __set_current_state(TASK_UNINTERRUPTIBLE);
      KUNLOCK();
    }

  //--- Test to see if any of the requested events are satisfied or if any ---
  //--- error conditions exist ---
    for (u=0,ufd=ufds; u<nfds; ++u,++ufd)  {
      if (ufd->fd < 0)  continue;  // -1 fd means ignore the entry

      if (ufd->revents)  { block = false;  continue; }

      if (ufd->events & POLLIN)  {
        numBytes = 0;
        if (ioctl(ufd->fd, FIONREAD, &numBytes))  {
          ufd->revents |= POLLERR;
          if (errno == ENOLINK)  ufd->revents |= POLLHUP;
          block = false;
        }
        else
          if (numBytes > 0)  { ufd->revents |= POLLIN;  block = false; }
      }

      if (ufd->events & POLLOUT)  {
        numBytes = 0;
        if (ioctl(ufd->fd, FIONSPACE, &numBytes))  {
          ufd->revents |= POLLERR;
          if (errno == ENOLINK)  ufd->revents |= POLLHUP;
          block = false;
        }
        else
          if (numBytes > 0)  { ufd->revents |= POLLOUT;  block = false; }
      }
    }

    if (!block)  break;  // takes care of timeout == 0 as well

  //--- block, waiting for one of the wait queues or a timeout to wake us up ---
    if (timeout < 0)
      schedule();
    else  // if no ticks left, then timeout woke us rather than an event
      if (! schedule_timeout(timeoutTicks))  break;

    if (numNodes)  removeFromQueues(current, false);
  }

//--- The equivalent of finish_wait() ---
  __set_current_state(TASK_RUNNING);
  if (numNodes)  removeFromQueues(current, true);

  if (err)  return -1;

  for (stat=0,u=0,ufd=ufds; u<nfds; ++u,++ufd)  if (ufd->revents)  ++stat;

  return stat;
}

//-----------------------------------------------------------------------------

