//----- sched.c ----- 06/12/12 -----(02/20/07)-----

//-----------------------------------------------------------------------------
//  DEBROS implementation of the schedule() function, called directly by a
//  user task or an interrupt handler (usually for the scheduler tick) to
//  initiate a task switch.
//-----------------------------------------------------------------------------
//  Update log (most recent first)
//
//
//  2012-06-12 - Mark Davis
//               - Minor tweaks to chooseTask() to eliminate some time-consuming
//                 long int operations.
//
//  2012-06-11 - Mark Davis
//               - current->lastRunning is now set at the start of schedule().
//                 As calling schedule() is the only way to change which task
//                 has the CPU, this is the only place it is actually needed.
//
//  2012-06-07 - Mark Davis
//               - Added taskCounts() to return max and free counts for the
//                 task table.
//               - added schedCalls counter
//
//  2011-03-29 - Mark Davis (*)
//               - Fixed bug in chooseTask() that could (under the right
//                 conditions) cause a suspended task to stay suspended even
//                 when it is ready to run.
//
//  2011-03-25 - Mark Davis
//               - Added setting of myAddr in task struct to support use of
//                 the VALID() and INVALID() macros
//
//  2011-02-21 - Mark Davis
//               - Added numReadyTasks
//
//  2010-08-02 - Mark Davis
//               - Removed disable/enable of write protection during a task
//                 switch (since it is not actually needed).
//
//  2010-07-15 - Mark Davis
//               - Changes to memory usage to allow ALL of the FARCODE and
//                 FARCONST segments to be write protected as well as all of
//                 the CODE segment and most of the CONST segment.
//
//  2010-07-08 - Mark Davis
//               - The ability to use 4K resolution on the memory protection
//                 for 2 of the 64K segments is now used to protect as much
//                 as possible of the FARCONST and FARCODE segments.  Prior
//                 to this they were used to write protect the stacks of most
//                 of the tasks and (when possible) the CODE and CONST segments.
//                 But this resulted in a larger amount of memory not being
//                 protected so this change was made (although I may add
//                 settings in the future to allow this to be configured
//                 either at compile time or run time).
//
//  2010-06-29 - Mark Davis
//               - Add log messages for memory allocations
//
//  2010-06-09 - Mark Davis
//               - Basic stack and memory hdw protections implemented
//
//  2010-06-08 - Mark Davis
//               - Software check for stack pointer over/underflow is now
//                 skipped if the hdw supported method is available.
//
//  2010-06-07 - Mark Davis (*)
//               - Hdw enforced Stack Limits are now used when possible
//
//  2010-05-14 - Mark Davis (*)
//               - timeoutTicks is no longer set to 0 when resuming a task.
//                 This caused schedule_timeout() to always return 0, which
//                 would cause problems (or at least wasted effort) when
//                 doing a blocking call with a timeout.
//
//  2010-05-07 - Mark Davis (*)
//               - Corrected "fix" from 2009-06-28 regarding when sleepTicks
//                 and timeoutTicks are set to 0.  The fix was checking the
//                 the state of the task that was running rather than the
//                 one it is about to run.  Prior to this fix the problems
//                 stated for the 2009-06-28 fix could still happen (i.e. a
//                 resumed task whose state was not already RUNNING could
//                 hang.  This probably explains why the output from the
//                 scope command could still hang (and possibly other problems
//                 I am not aware of).
//               - Changed code so DATASEG is always the same as the ROOT
//                 segment (0), rather than trying to follow the STACK segment.
//                 This is simpler, slightly faster. and avoids any possibility
//                 of there being a problem if the schedule() function is
//                 interrupted at just the wrong moment.
//               - The timer_interrupt() is now disabled from the time just
//                 before we choose what task to run next and actually making
//                 it the current task.  This avoids the possibility of the
//                 timer_interrupt() happening between when have choosen a
//                 new task and when we have switch to running that task.
//
//  2010-03-11 - Mark Davis (*)
//               - chooseTask() now checks the state of the task that was
//                 preempted to see if it has been corrupted in some fashion
//                 we can easily detect.
//
//  2010-03-09 - Mark Davis (*)
//               - Tasks with a run state of STOPPED are now ignored when
//                 choosing a candidate for resuming.  This eliminates
//                 additional CPU cycles spent considering such tasks and the
//                 possibility of log messages repeatedly complaining about
//                 the corrupted state of such tasks prior to properly
//                 killing them off.
//
//  2010-03-06 - Mark Davis (*)
//               - chooseTask() now checks to see if the task it has choosen
//                 to resume has had the last/top 2 bytes of its stack
//                 overwritten, indicating either that it overflowed while
//                 running, or was corrupted by something else writing to its
//                 stack.
//               - chooseTask() now checks to make sure the stack pointer
//                 is within the valid range of values before resuming a
//                 task. The previous check only checked to see if it was
//                 LESS than the top of the stack, whereas it will now
//                 catch one that has wrapped around, popped everything on
//                 the stack, or just plain been corrupted.
//
//  2010-03-02 - Mark Davis (*)
//               - chooseTask() now checks for a near stack overflow before
//                 resuming a task.  If a task is about to overflow its stack,
//                 its run state is changed to STOPPED and it is sent a
//                 TERMINATE signal so the kernel will clean it up (and of
//                 course it chooses a differrent task to resume).
//
//  2010-02-14 - Mark Davis
//               - Default SchedMS value now set to 20 for BL2105 hdw
//
//  2009-06-28 - Mark Davis (*)
//               - sleepTicks and timeoutTicks are no longer set to 0 when
//                 resuming a task unless its state is RUNNABLE.  Setting them
//                 to 0 for a task whose state is not already RUNNABLE (e.g.
//                 one that was preempted) could result in a task that gets
//                 hung (in the case of timeoutTicks) or one that wakes up
//                 too early (in the case of sleepTicks).
//
//  2009-06-08 - Mark Davis
//               - Added SchedMS so the scheduler tick interval could be
//                 changed as needed for the hardware or the application
//
//  2007-06-17 - Mark Davis
//               - began migration from simple periodic polling loops for
//                 blocking to Linux-style wait queues (to reduce wasted CPU
//                 cycles, decrease average latency, improve response times).
//
//  2007-02-20 - Mark Davis
//               - work started on Linux-style method for managing task
//                 switches.  Done to eliminate need for Softools' TurboTask
//                 kernel and allow stacks to be located anywere in memory.
//                 Based on "Simple Task Switcher" code from SH Design's.
//
//-----------------------------------------------------------------------------


#include <common.h>

#include <unistd.h>

#include <linux/sched.h>

#include <linux/wait.h>

#include <string.h>

#include <rabbit.h>
#include <dcdefs.h>
#include <limits.h>

#include <sys/farprintf.h>


/*---------------------------------------------------------------------------*
 *              *** WARNING: NEVER modify these directly ***                 *
 *   _disabled is managed by schedule() itself and KLOCK() and KUNLOCK()     *
 *   _pending is managed by schedule(), sched_tick(), and KUNLOCK()          *
 *---------------------------------------------------------------------------*/
char _disabled = 0;  // a non-0 value disables the scheduler
char _pending = 0;   // a non-0 value indicates schedule() needs to be called
                     // as soon as it the scheduler is re-enabled


uint  schedCalls;    // # of calls to schedule() since last zeroed
uint  taskSwitches;  // # of task switches since last zeroed


task_t tasks[NUM_TASKS];

static task_t  *next;
task_t  *current = tasks;
PIX    userPix;

task_t  *endTasks = tasks + NUM_TASKS;
task_t  *idleTask, *applTask, *kernelTask, *inetTask;

ulong  newTaskDiagFlags;   // diagFlags for new tasks

uint  SchedMS;  // scheduler tick interval in ms
uint  numReadyTasks;  // # tasks ready to run during last call to schedule()

//--- set by init_tasks() ---
static uchar far  *stackPool = (uchar far *)0l;
static long  stackPoolSize = 0l;

bool  stacksCross64K;
byte  stackSegL, stackSegH, stackPageL, stackPageH;

static uint  stackSize;

static ulong  wpAllStacks;

//void dummy()  { getMilliSeconds(); }  // to make compiler happy....


//-----------------------------------------------------------------------------
//  Get the max and free counts for the table of all tasks
//  The *ttl is set to the max, and the free count is the return value.
//-----------------------------------------------------------------------------
uint taskCounts(uint *ttl)
{
  uint  freeCount = 0;
  task_t  *task;


  for (task=tasks; task!=endTasks; ++task)
    if (!task->inUse)  ++freeCount;

  *ttl = NUM_TASKS;

  return freeCount;
}


//-----------------------------------------------------------------------------
//  Test various aspects of the state of the specified task to see if it has
//  been corrupted.
//-----------------------------------------------------------------------------
void checkTask(task_t *task)
{
  uint far *stackStart;

//--- Check to see if the task's stack pointer is within the valid range ---
  if (!hdwProtection AND
      ((task->stack < task->stack_start + 32)
      OR (task->stack > task->stack_start + task->stack_size)) ) {
    fprintf(DiagLog, "***** %.32s[%d]: STACK POINTER OUT OF RANGE!!! *****\n",
            task->argv[0], task->pid);
    if (KeyTask(task))  postMortem("Stack ptr out of range");
    task->state = TASK_STOPPED;  task->sig_pending |= sigmask(SIGKILL);
    return;
  }

//--- If the task we chose to resume has had its stack corrupted or it ---
//--- overflowed while running (i.e. the top 2 bytes were written over) ---
//--- then log a message, kill the task, and choose another ---
  stackStart = (uint far *)
     ((ulong)task->stack_start + (((ulong)task->sseg) << 12));
  if (*stackStart != 0xffff)  {
    fprintf(DiagLog, "***** %.32s[%d]: STACK CORRUPTED OR OVERFLOWED!!! *****\n",
            task->argv[0], task->pid);
    if (KeyTask(task))  postMortem("Stack Corrupted or Overflowed");
    task->state = TASK_STOPPED;  task->sig_pending |= sigmask(SIGKILL);
    return;
  }

}


//-----------------------------------------------------------------------------
//  called by schedule() to decide which process to resume
//-----------------------------------------------------------------------------
void chooseTask(void)
{
  ulong  tickTime, msTime, ticksUsed;
  uint  readyCount, msSinceNextResumed, msSinceResumed;
  task_t  *task;


//--- check to see if the current task has been corrupted ---
  checkTask(current);


//--- starting with the process following the current one, scan the ----
//--- process list determining which ones can be woken up and which is the ---
//--- next-in-line, ready-to-run of the highest priority ---
//--- (crude, but adequate for now) ---

  next = idleTask;  task = current;  readyCount = 0;
  msSinceNextResumed = msSince_(next->lastResumed);

  for (;;)  {

 /*------------------------------------------------------------------*
   Set the idle task as the default candidate for the next task to run
   run and then determine if there is a better choice based on the
   following:
    - task is ready to run (includes ones that were preempted)
    - priority >= current candidate
    - run less recently than current candidate
  *------------------------------------------------------------------*/

    for (;;)  {
      if (++task == endTasks)  task = tasks;

  /*finish
      if (task->sig_pending)
        if ((task->state == TASK_RUNNING)
            OR (task->state == TASK_INTERRUPTIBLE))  {
          // ... check for signal handler for the task ...
        }
  */
      if (task->state != TASK_STOPPED)
        if ((task->state == TASK_RUNNING) OR task->preempt_count)  {
          ++readyCount;
          if ( (task->priority < next->priority) OR
              ((task->priority == next->priority)
                AND (msSinceResumed = msSince_(task->lastResumed))
                     > msSinceNextResumed) )
             if (ActiveTask(task))  {
               next = task;  msSinceNextResumed = msSinceResumed; }
        }

      if (task == current)  break;  // wrapped around to where we started
    }

  //--- Make sure the task we choose is not obviously corrupted ---
    checkTask(next);  if (next->state >= TASK_STOPPED)  continue;


  //--- If we are switching tasks, then update the CPU usage data ---
  //--- and set the current process info ---

    if (next != current)  {
      tickTime = clockTicks();  msTime = MS_TIMER;

      ++taskSwitches;

      current->lastStopped = msTime;

      if (current->lastResumed)  {
        ticksUsed = UL_Interval(current->resumedTicks, tickTime);
        if (ticksUsed & 0xffff0000L)  // value > 2 secs unlikely
          current->msUsed += msSince(current->lastResumed);
        else  {
          current->ticksUsed += ticksUsed;
          current->msUsed += current->ticksUsed >> 5;  // aprox the # of ms portion
          current->ticksUsed &= 0x1f;  // remove the aprox ms portion
        }
      }

      next->lastResumed = msTime;
      next->resumedTicks = tickTime;

      if (next->state == TASK_RUNNING)  next->sleepTicks = 0;

      current = next;  userPix = current - tasks;
    }

    break;
  }

  numReadyTasks = readyCount;  // for use in computing load averages
}


//-----------------------------------------------------------------------------

#pragma offset_labels(task_t)

/*-----------------------------------------------------------------------------
  This is what schedule() pushes on to the stack of the calling task before it
  suspends the task.  If this is changed, you must update the macro
  Schedule_PushCount in linux/sched.h:

  --- pushed by schedule() before suspending a task ---
  stackPtr[0]:    IX (LSB)
  stackPtr[1]:    IX (MSB)

  stackPtr[2]:    IY (LSB)
  stackPtr[3]:    IY (MSB)

  stackPtr[4]:    IP
*/
//-----------------------------------------------------------------------------
//  Called either by timer_interrupt() or by the task itself.
//
//  In the 1st case, the task was Preempted by the periodic timer interrupt.
//
//  In the 2nd case, schedule() was most likely called by some system fucntion
//  that the task called and the task is blocked, waiting for some event to
//  occur.
//
//                              ***** WARNING *****
//   If you change what schedule() pushes on to the stack before suspending a
//   task then you must update the macro Schedule_PushCount (in linux/sched.h)
//   and the function postMortem() unistd.c
//                              ***** WARNING *****
//-----------------------------------------------------------------------------
void _nearcall schedule(void)
{
//--- Save the state of the IP (Interrupt Priority) register on the stack ---
//--- of the process that called or was interrupted ---
  _asm( "push ip" );

  StartInterval(TrackScheduler);

  current->lastRunning = GetMS();

// IMPORTANT:  Caller must save all registers they might need on return except
//             for IY and IX (which are saved here on the stack of the current
//             process)

#asm

;//--- roughly equivalent C code for what this func does ---
;>>  do  {
;>>    EnterCritical();
;>>    ++schedCalls;
;>>    _pending = 1;
;>>    ++_disabled;  locked = _disabled - 1;
;>>    ExitCritical();
;>>
;>>    if (! locked)  {
;>>      _pending = 0;
;>>      current->stack = SP;
;>>
;>>      DisableTimerInts();
;>>
;>>      chooseTask();  // changes 'current'
;>>
;>>      _EnterCritical_();
;>>      SP = current->stack;  STACKSEG = current->sseg;
;>>
;>>      if (hdwProtection)  {
;>>        outi(MMIDR, 0x80);   // enable access to registers > 0x00FF
;>>
;>>  //      outi(WPCR, 0);  // disable Write Protection
;>>
;>>        outi(STKLLR, task->stkllr);  // set stack limits
;>>        outi(STKHLR, task->stkhlr);
;>>
;>>        outi(WPSAR, task->wpsar);  // set task-specific 4K page protections
;>>        outi(WPSALR, LSBI(task->wpsalhr));
;>>        outi(WPSAHR, MSBI(task->wpsalhr));
;>>
;>>        outi(WPSBR, task->wpsbr);  // set task-specific 4K page protections
;>>        outi(WPSBLR, LSBI(task->wpsblhr));
;>>        outi(WPSBHR, MSBI(task->wpsblhr));
;>>
;>>  //      outi(WPCR, 1);  // enable write protection
;>>
;>>        outi(MMIDR, 0x00);   // disable access to registers > 0x00FF
;>>      }
;>>      _ExitCritical_();
;>>
;>>      EnableTimerInts();
;>>    }
;>>    if Monotonic(--_disabled != 0)  break;  // ignore _pending if locked/disabled
;>>  } while (_pending);  // can be set by timer interrupt after we cleared it


;//  In English, here is what happens:
;//
;//  - Set a flag indicating that a task switch is pending
;//
;//  - Try to get the lock for the portion of code that actually performs the
;//    task switch.
;//
;//  - If the scheduler is not locked then:
;//    - clear the pending flag
;//    - choose the task to be resumed
;//    - switch to its stack
;//
;//  - Release the lock (or our attempt to get it earlier).  If the scheduler
;//    is locked, then exit
;//
;//  - Repeat until there are no unserviced (pending) task switch requests
;//    (this can happen if the timer interrupt occurs after we have cleared
;//     pending, in which case the run state of tasks may have changed and
;//     we need to check again to see which one should be resumed).
;//
;//
;//  Notes:
;//    - Two macros are provided for locking and unlocking the scheduler:
;//      KLOCK() and KUNLOCK()
;//
;//    - KLOCK() locks the scheduler by incrementing the _disabled flag as
;;//     a monoatomic operation.  This disables task switches which effectively
;//      protects ALL global variables from access by user-space code (NOTE
;//      that it does NOT protect anything from interrupt handlers).
;//
;//    - KUNLOCK() potentially unlocks the scheduler by decrementing the
;//      _disabled flag in a critical section.  It then checks the _pending
;//      and _disabled flags.  If _pending is set, but _disabled is 0, then a
;//      call to schedule() was due while the scheduler was disabled, so
;//      KUNLOCK() calls schedule() so the pending task switch is serviced
;//      without any further delay.
;//
;//    - Calls to KLOCK() and KUNLOCK() can be nested.  This allows functions
;//      to call KLOCK() and KUNLOCK() (in pairs) to insure the scheduler is
;//      disabled when needed without having to worry about whether or not the
;//      scheduler was already disabled (or not).  Of course, all such nesting
;//      is in the context of a single process, since no process switching
;//      occurs between the first call to KLOCK() and its corresponding call
;//      to KUNLOCK().  Also, it is critical that the calls always happen in
;//      pairs to avoid permanently disabling the scheduler.
;//
;//    - The purpose of a call to schedule() is NOT actually to force a task
;//      switch, but to indicate that the state of a task has changed, and that
;//      the CPU should be given to the highest ready-to-run process (which
;//      may or may not be the process that was running when schedule() was
;//      called).  3 cases must be considered when attempting to achieve this
;//      goal:
;//
;//        1. A process needs to block until some event occurs without using
;//           up CPU time.  This is generally done by some kernel function on
;//           behalf of the process that called it.
;//
;//        2. The currently running process was interrupted by the timer
;//           interrupt.
;//
;//        3. An event happened that one or more processes might be waiting
;//           for.
;//
;// ...
;//


;>>  do {
.sched_start:

;>>    EnterCritical();
     ipset 3

;>>    ++schedCalls;
    ld      hl,(_schedCalls)
    inc     hl
    ld      (_schedCalls),hl

;>>    _pending = 1;
    ld      a,1
    ld      (__pending),a

;>>    ++_disabled;  locked = _disabled - 1;
    ld      hl,__disabled
    inc     (hl)
    ld      a,(hl)
    dec     a

;>>    ExitCritical();
    ipres

;>>    if (! locked)  {
    jp      nz,.chk_pending

;>>      _pending = 0;
    ld      (__pending),a    ;; (A == 0 from above)

    push    iy
    push    ix

;>>      current->stack = SP;
    ld      h,a                 ;; A == 0 from above
    ld      l,a
    add     hl,sp               ;; hl = current's SP
    ld      ix,(_current)       ;; ix = current
    ld      (ix+STACK_OFF),hl   ;; current->stack = SP

;>>      DisableTimerInts();
    ipset 1

;>>      chooseTask();  (changes current)
    lcall    $chooseTask

;>>      SP = current->stack;  STACKSEG = current->sseg;
    ld      ix,(_current)           ;; ix = current
    ld      hl,(ix+STACK_OFF)       ;; hl = current->stack
    ld      a,(ix+SSEG_OFF)         ;; a = current->sseg

;>>      EnterCritical();
    ipset 3   ;; insures some ISR doesn't cause a stack limit fault

    ld      sp,HL         ;; disables ints for the next instruction
    ioi ld  (STACKSEG),a

;>>     if (hdwProtection)  {
    ld      a, (_hdwProtection)
    or      a
    jp      z,.skipHdwProtect

;>>        outi(MMIDR, 0x80);   // enable access to registers > 0x00FF
    ld      a,0x80
    ioi ld  (MMIDR),a

;;  //         outi(WPCR, 0);  // disable Write Protection
;;  //  xor     a
;;  //  ioi ld  (WPCR),a

;>>        outi(STKLLR, task->stkllr);  // set stack limits
;>>        outi(STKHLR, task->stkhlr);
    ld      hl,(ix+STKLLR_OFF)  ;; gets both STKLLR and STKHLR values
    ioi ld  (STKLLR),hl         ;; write both STKLLR and STKHLR

;>>        outi(WPSAR, task->wpsar);  // set task-specific 4K page protections
;>>        outi(WPSALR, LSBI(task->wpsalhr));
;>>        outi(WPSAHR, MSBI(task->wpsalhr));
    ld      a,(ix+WPSAR_OFF)    ;; get value for WPSAR
    ioi ld  (WPSAR),a
    ld      hl,(ix+WPSALR_OFF)  ;; gets both WPSALR and WPSAHR values
    ioi ld  (WPSALR),hl         ;; write both WPSALR and WPSAHR

;>>        outi(WPSBR, task->wpsbr);  // set task-specific 4K page protections
;>>        outi(WPSBLR, LSBI(task->wpsblhr));
;>>        outi(WPSBHR, MSBI(task->wpsblhr));
    ld      a,(ix+WPSBR_OFF)    ;; get value for WPSBR
    ioi ld  (WPSBR),a
    ld      hl,(ix+WPSBLR_OFF)  ;; gets both WPSBLR and WPSBHR values
    ioi ld  (WPSBLR),hl         ;; write both WPSBLR and WPSBHR

;>>  //      outi(WPCR, 1);  // enable Write Protection in System and User modes
;;   //  ld      a,1
;;   //  ioi ld  (WPCR),a

;>>        outi(MMIDR, 0x00);   // disable access to registers > 0x00FF
    xor     a
    ioi ld  (MMIDR),a

;>>     }

.skipHdwProtect:
;>>      ExitCritical();
    ipres

;>>      EnableTimerInts();
    ipres


;--- We are now using the stack of the resumed task, NOT the one that was ---
;--- running when schedule() was called ---
    pop     ix
    pop     iy

    ld      hl,__disabled
;>>    }  // end-if

.chk_pending:
;>>    if Monotonic(--_disabled != 0)  break;
    dec     (hl)
    jp      nz,.exit_ts_busy

;>>  } while (_pending);
    ld      a,(__pending)
    sla     a
    jp      nz,.sched_start


.exit_ts_busy:
;;--- restore the state of the IP register from the stack of the process ---
;;--- we are about to resume ---
    pop     ip
#endasm

  EndInterval(TrackScheduler);
}


#pragma offset_labels on

//-----------------------------------------------------------------------------
//  This launches a new process by calling its main function with any
//  arguments provided in the call to createProc()
//-----------------------------------------------------------------------------
void near taskStart(int pix)
{
  task_t  *task = tasks + pix;

// printf("Launcing task %d...\r\n", pix);  //tdebug

//finish - make task_ funcs return int rather than void?
  (task->procMain)(task->argc, task->argv);

  current->state = EXIT_DEAD;
  current->exit = -9999;  // indicates task returned rather than did an Exit()

  schedule();  // should never return since this task is now DEAD
}

//-----------------------------------------------------------------------------
//  Modify the write protect masks for the 2 4K resolution pages as follows:
//
//  - The first is used to protect the start of the FARCODE/FARCONST segments.
//    If the start is in the root page, then it is also used to protect as
//    much of the CODE and CONST segments as possible (and the kernel's stack
//    for all task's but the kernel's)
//
//  - The second is used to protect the end of the FARCODE/FARCONST segments.
//    Eventually, this may be expanded (when possible) to compute and use
//    a seperate mask for a portion of the FARBSS values that are writable
//    only by kernel functions.
//
//  While these values are curently the same for most tasks, this is called
//  for each task to allow the possibility of an earlier approach where at
//  least one of the 4K resolution pages was used to protect as many of the
//  stacks as possible (i.e. those that don't belong to the currently running
//  task).
//-----------------------------------------------------------------------------
void set4KwpMasks(task_t *task)
{
  uint  u, pageL, numPages, mask;


  if (!hdwProtection)  return;

//--- Use the 1st 4K resolution page to protect the beginning of the far ---
//--- segments.  Note that this asusmes that the first of the far segments ---
//--- starts on a 4K boundary (or at least does not share a 4K block with ---
//--- any other writeable data) ---
  task->wpsar = (uchar) MSWL(roStartFar);
  pageL = (uint)((roStartFar + 0x0FFF) >> 12) & 0x0F;  numPages = 16 - pageL;
  mask = 1 << pageL;
  for (u=1; u<numPages; ++u)  mask |= mask << 1;
  task->wpsalhr = mask;

//--- If the start of the far segments is in root, then we can also use the ---
//--- first 4K resolution page to protect some of CODE and CONST (and the ---
//--- kernel's stack from all other tasks) ---
  if (task->wpsar == 0)  {
    task->wpsalhr |= wpRootMask;
    if (task != tasks)  task->wpsalhr |= 0x2000;  // write protect kernel's stack
  }

//--- This assumes that the 4K block that contains the end of the ReadOnly ---
//--- portion of extended memory does not contain any writable data (i.e. ---
//--- that the following segment starts at the next 4K boundary) ---
  task->wpsbr = (uchar) MSWL(roEndFar);
  numPages = ((uint)(roEndFar >> 12) & 0x0F) + 1;
  if (numPages == 16)  {  // entire 64K segment?
    wpSegMask |= 1 << task->wpsbr;  mask = 0;  }
  else  {
    mask = 1;
    for (u=1; u<numPages; ++u)  mask |= mask << 1;
  }
  task->wpsblhr = mask;

//  fprintf(DiagLog, "  Task %2d: %02X:%02X-%02X  %02X:%04X  %02X:%04X\r\n",
//          task - tasks, task->sseg, task->stkllr, task->stkhlr,
//          task->wpsar, task->wpsalhr,
//          task->wpsbr, task->wpsblhr );
}


//-----------------------------------------------------------------------------
//  Prepare an entry in the task table to run a new task.  For entries that
//  have never been used, assign a portion of the memory allocated for stacks
//  for use only by the specified task.
//
//  In either case, the stack is initialized with the values that will cause
//  taskStart() to be executed when schedule() "resumes" the new task.
//
//  Returns the specified task number if successful, < 0 if an error
//-----------------------------------------------------------------------------
int initTaskEntry(int tnum, unsigned stack_size)
{
  ulong  absAddrSP, startPage, endPage;  //, wpMask;
  uint  logAddrSP, wasted;
  uchar  stackSeg;  //, absPage;
  uint far  *stackPtr;
  task_t  *task;


  if (tnum >= NUM_TASKS)   return -1;

  task = tasks + tnum;  task->myAddr = task;

//===== Reusing a task table entry that already has a stack allocated =====
  if (stack_size == 0) {  // this is a restart
    absAddrSP = (ulong) task->stack_start;
    absAddrSP += ((ulong) task->sseg) << 12;
    stackPtr = (uint far *) absAddrSP;

    stack_size = task->stack_size;
  }

//===== Assign stack space for an entry that has never been used =====
  else  for (;;)  {
    if (stackPoolSize < stack_size)  {
      printf("insufficient stack space left: %ld\r\n", stackPoolSize);
      return -2;
    }

  //--- make sure stack starts on a 256 byte boundary if using hdw checks ---
    if (hdwProtection)
      stackPtr = (uint far *) (((ulong)stackPool + 0xFFL) & 0xFFF00L);
    else
      stackPtr = (uint far *) stackPool;

  //--- compute start addr of the 4K pages the 1st & last bytes reside in ---
    startPage = (ulong)stackPtr & 0xFF000L;
    endPage =  ((ulong)stackPtr + stack_size - 1L) & 0xFF000L;

  //--- if it crosses 4K boundary, then use space starting at next boundary ---
    if (startPage != endPage)  {  // does stack cross a 4K boundary?
      wasted = 0x1000 - ((uint)stackPtr & 0xFFF);
      stackPoolSize -= wasted;
      stackPool += wasted; // advance ptr to next 4K boundary
      continue;  // try again
    }

//    printf("  Task %2d: Stack %05lX - %05lX\r\n",  //tdebug
//           tnum, (ulong)stackPtr, (ulong)stackPtr + stack_size - 1L);

  //--- update pointer and remaining byte count for next call ---
    stackPool += stack_size;  stackPoolSize -= stack_size;

    break;
  }

  if (stackPtr == NULL)   return -4;


//--- compute STACKSEG and SP values from the abs addr of start of stack ---
  stackSeg = (uchar)((ulong)stackPtr >> 12) - 0x0D;
  logAddrSP = 0xD000 | (((uint) stackPtr) & 0x0FFF);

  task->sseg = (uchar) stackSeg;

//--- init values for stack limits if hdw supports it ---
  if (hdwProtection)  {
    task->stkhlr = (uchar) ((logAddrSP + stack_size) >> 8);
    task->stkllr = (uchar) (logAddrSP >> 8);
/*
    absPage = (uchar) ((ulong)stackPtr >> 12);  // absolute 4K page #

    wpMask = 1L <<
        ((uint)absPage - (uint)stackPageL + ((uint)stackPageL & 0x0F));

    task->wpsar = stackSegL;
    task->wpsalhr = LSWL(wpAllStacks) & ~LSWL(wpMask);

    if (stacksCross64K)  {
      task->wpsbr = stackSegH;
      task->wpsblhr = MSWL(wpAllStacks) & ~MSWL(wpMask);
    }
    else  {
      task->wpsbr = 0;
      task->wpsblhr = wpRootMask;
    }
*/
    set4KwpMasks(task);
  }


//--- make size an even # of bytes (because its faster to use 16-bit instrs) ---
  stack_size &= 0xFFFE; // make it an even #

//--- Filling a new task's stack with the value 0xFF gives us an easy way ---
//--- to scan it any time we want to see how many bytes were unused.  Of ---
//--- course its not guaranteed, but unless a task actually uses a ---
//--- significant portion at the top of its stack for the value 0xFF, then ---
//--- its a pretty reliable way to check the peek usage.  The scheduler ---
//--- also checks to make sure a task's stack pointer is within the valid ---
//--- range for a task before it will resume it ---

  KLOCK();

//  DisableWriteProtection();  // disable so we can write to stack

  fmemset(stackPtr, 0xff, stack_size);

//--- Now we fill in the stack with what is needed to "resume" the task ---
  stack_size >>= 1; // change size to # of words
  stackPtr += stack_size - 1;     // makes ttl unused 16 byes to avoid a fault

  *stackPtr-- = tnum;              // save task number
  *stackPtr-- = 0;                 // dummy return addr

  // so when schedule() does an LRET it will be executing taskStart()
  *stackPtr = 0;                  // XPC = 0
  stackPtr = (uint far *) (((char far *) stackPtr) - 1);  // --1 byte instead of 2
  *stackPtr-- = (uint) taskStart; // our startup routine

  // values popped by schedule() when "resuming" a process
  *stackPtr = 0;                 // IP = 0
  stackPtr = (uint far *)(((char far *)stackPtr) - 1);    // --1 byte instead of 2
// initial IX and IY values
  *stackPtr-- = 0; // initial IY value
  *stackPtr-- = 0; // initial IX value

//  EnableWriteProtection();  // reenable protection

  KUNLOCK();

//--- set values in the task's structure ---
  task->stack_start = (uint *) logAddrSP;
  task->stack_size = stack_size << 1;  // change it back to # of bytes

  absAddrSP = (ulong) ++stackPtr;  // incr so not pointing at LSB of IX value
  absAddrSP = (absAddrSP & 0x0FFFL) | 0xD000L;
  task->stack = (uint *) absAddrSP;


/*<<<<< The task's stack should now look like this >>>>>

  ..        (<=== task->stack)

===== must match what schedule() pushes on the stack =====
  IX (LSB)
  IX (MSB)

  IY (MSB)
  IY (MSB)

  IP

===== causes taskStart() to execute when schedule() does an LRET =====
  taskStart (LSB)
  taskStart (MSB)

  XPC   0

===== values taskStart() expects to be on the stack when it starts ====
  0     (LSB)      // dummy return address (taskStart() should never return!)
  0     (MSB)

  tnum  (LSB)      // argument for taskStart()
  tnum  (MSB)
*/



#ifdef DEBUG
  printf("Task #%2d at [%04X], %d byte stack at %02X:%04X\r\n",
          tnum, main_, stack_size<<1, sseg, task->stack_start );
#endif

  return tnum;
}



//-----------------------------------------------------------------------------
//  reinitialize the specified entry in the task table for use by a new task
//-----------------------------------------------------------------------------
int reset_task(int tasknum)
{
  return  initTaskEntry(tasknum, 0);
}

//-----------------------------------------------------------------------------
//  return the # of bytes on the stack of the specified task that have not
//  been used yet
//-----------------------------------------------------------------------------
int check_stack(int tasknum)
{
   int words = 0;
   uint far *stack;

   ulong absAddrSP = (ulong) tasks[tasknum].stack_start;
   absAddrSP += ((ulong) tasks[tasknum].sseg) << 12;

   stack = (uint far *) absAddrSP;

   while (*stack++ == 0xffff)  words++;

   return words << 1;
}

//-----------------------------------------------------------------------------
//  Allocate memory for all the tasks stacks and initialize each entry in
//  the task table
//-----------------------------------------------------------------------------
int init_tasks(void)
{
  uint  u, startBit, endBit, numFullPages, xmemAdjust, bytesToNextPage,
        numPages, pageOffset, numFilledPages;
  int  stat;
  ulong  stackPoolEnd, mask;
  task_t  *task;


  task = tasks + 1;
  for (u=1; u<NUM_TASKS; ++u,++task)  {
    task->state = TASK_UNUSED;
    init_waitqueue_head(&task->waitQs);
  }


//--- The hardware stack limit checking is limited to a resolution of 256 ---
//--- bytes.  In addition, DEBROS write protects all the 4K pages that ---
//--- contain stacks other than the one with the currently running task's, ---
//--- which means the protected pages can't be allowed to have anything ---
//--- BUT stacks in them (if they did, then we would be disabling writes ---
//--- to any other variables that were in the same page as some stacks!) ---

//--- Also keep in mind that the kernel task already has its own 4K stack ---
//--- and is not included in the stack pool used for all the other tasks ---
//--- (hence the use of (NUM_TASKS-1) below)

  xmemAdjust = 0;  pageOffset = (uint)_xmemHeap & 0x0FFF;

  bytesToNextPage = (pageOffset ? 0x1000 - pageOffset : 0);

  if (hdwProtection)  {
    stackSize = STACK_SIZE & 0x1F00;  // 256 byte resolution for hdw limits
    numFullPages = (((NUM_TASKS-1) + StacksPer4K - 1) / StacksPer4K);
    stackPoolSize =  numFullPages * 0x1000;
    // to insure xmem heap starts on a page boundary
    if (pageOffset)  xmemAdjust = bytesToNextPage;
  }
  else  {
    stackSize = STACK_SIZE;
    numFilledPages = (NUM_TASKS-1) / StacksPer4K;
    stackPoolSize = (numFilledPages * 0x1000)
                  + ((NUM_TASKS-1) % StacksPer4K) * stackSize;
    // adjust start of heap to skip any 1st page bytes that can't be used ---
    xmemAdjust = bytesToNextPage % stackSize;
  }

  _xmemHeap += xmemAdjust;  _xmemSize -= xmemAdjust;

  printf("Stack size: %d (0x%04X) bytes\r\n"
         "For stacks:\r\n", stackSize);
  stackPool = (char far *) xalloc(stackPoolSize);
  if (stackPool == (char far *)0l)  {
    printf("***** Unable to allocate stacks!!! *****\r\n");  return -1;  }

  stackPoolEnd = (ulong)stackPool + stackPoolSize - 1;


//--- first/last 4K pages used by stacks, relative to start of memory ---
  stackPageL = (uchar) ((ulong)stackPool >> 12);
  stackPageH = (uchar) (stackPoolEnd >> 12);

//--- first/last 64K segments used by stacks, relative to start of memory ---
  stackSegL = (uchar) (stackPageL >> 4);
  stackSegH = (uchar) (stackPageH >> 4);

//--- first/last 4K pages relative to first page in first 64K segment ---
  numPages = (uint)stackPageH - (uint)stackPageL + 1;
  startBit = (uint)stackPageL & 0x0F;
  endBit = startBit + numPages - 1;

  stacksCross64K = (stackSegL != stackSegH);

  if (hdwProtection)  {
//    if (stacksCross64K)
//      printf("***** Can't protect anything in root segment *****\r\n");

  //--- compute write-protect mask that covers ALL the 4K stack pages ---
  //--- other than the kernel's ---
    wpAllStacks = 0;  mask = 1 << startBit;
    for (u=startBit; u<=endBit; ++u,mask<<=1)  wpAllStacks |= mask;

//    printf("wpAllStacks: 0x%08lX (%d pages: %02X-%02X)\r\n", //tdebug
//           wpAllStacks, numPages, stackPageL, stackPageH);

  //--- set values for kernel task ---
  /*
    tasks[0].wpsar = stackSegL;
    tasks[0].wpsalhr = LSWL(wpAllStacks);

    if (stacksCross64K)  {
      tasks[0].wpsbr = stackSegH;
      tasks[0].wpsblhr = MSWL(wpAllStacks);
    }
    else  {
      tasks[0].wpsbr = 0;
      // clear bit for page at 0xD000 so the kernel can write to its own stack!
      tasks[0].wpsblhr = wpRootMask & ~0x2000;
    }
  */
    set4KwpMasks(tasks);  // for kernel task
  }

  newTaskDiagFlags = 0;

//--- set default scheduler interval in ms (can be changed later) ---
  SchedMS = (hdwType == HT_BL2105 ? 20 : 10);

//--- set fixed values and init stack for each task_t structure  ---
  for (u=1; u<NUM_TASKS; ++u)
    if ((stat = initTaskEntry(u, stackSize)) != u)  {
      printf("*****  Unable to initialize task #%u: %d *****\r\n", u, stat);
      postMortem("initTaskEntry failure");
    }


  return 0;
}



//=============================================================================
//  portions borrowed from or based on standard Linux kernel/sched.c code
//=============================================================================

//-----------------------------------------------------------------------------
/***
 * try_to_wake_up - wake up a thread
 * @p: the to-be-woken-up thread
 * @state: the mask of task states that can be woken
 * @sync: do a synchronous wakeup?
 *
 * Put it on the run-queue if it's not already there. The "current"
 * thread is always on the run-queue (except when the actual
 * re-schedule is in progress), and as such you're allowed to do
 * the simpler "current->state = TASK_RUNNING" to mark yourself
 * runnable without the overhead of this.
 *
 * returns failure only if the task is already active.
 */
//-----------------------------------------------------------------------------
/*
static int try_to_wake_up(struct task_struct *p, unsigned int state, int sync)
{
    int cpu, this_cpu, success = 0;
    unsigned long flags;
    long old_state;
    struct rq *rq;

    rq = task_rq_lock(p, &flags);
    old_state = p->state;
    if (!(old_state & state))  goto out;

    if (p->array)  goto out_running;

    cpu = task_cpu(p);
    this_cpu = smp_processor_id();

  if (old_state == TASK_UNINTERRUPTIBLE)  rq->nr_uninterruptible--;

    activate_task(p, rq, cpu == this_cpu);
    //
    // Sync wakeups (i.e. those types of wakeups where the waker
    // has indicated that it will leave the CPU in short order)
    // don't trigger a preemption, if the woken up task will run on
    // this cpu. (in this case the 'I will reschedule' promise of
    // the waker guarantees that the freshly woken up task is going
    // to be considered on this CPU.)
    //
    if (!sync || cpu != this_cpu) {
        if (TASK_PREEMPTS_CURR(p, rq))
            resched_task(rq->curr);
    }
    success = 1;

out_running:
    p->state = TASK_RUNNING;
out:
    task_rq_unlock(rq, &flags);

    return success;
}
*/

//-----------------------------------------------------------------------------
int default_wake_function(WaitQueueNode *curr, unsigned mode, int sync,
              void *key)
{
//  return try_to_wake_up(curr->private, mode, sync);
  return 1;
}
EXPORT_SYMBOL(default_wake_function);

//-----------------------------------------------------------------------------

