//----- timer.c ----- 05/14/10 -----(10/30/07)-----

//-----------------------------------------------------------------------------
//  DEBROS implementation of some of the linux timer.c code
//-----------------------------------------------------------------------------
//  Update log (most recent first)
//
//
//  2010-05-14 - Mark Davis (*)
//               - The task's timeoutTicks value is now set to 0 after saving
//                 a copy of it when the task resumes.  This value is now
//                 decremented by the kernel task ONLY if the task is in
//                 a blocked state (i.e. INTERRUPTIBLE or UNINTERRUPTIBLE).
//                 This not only allows schedule_timeout() to return a non-0
//                 value, but it means the value no longer has to be protected
//                 by a critical section in this code (because the task's state
//                 WILL be RUNNING when it returns from the call to schedule(),
//                 so the kernel will not access the timeoutTicks value).
//
//  2009-06-26 - Mark Davis
//               - Added a critical section around the read of task's
//                 timeoutTicks value to guard against preemption.  Without
//                 this, the read could get a non-zero value just before its
//                 gets preempted and set to 0, causing the task
//
//
//  2007-10-30 - Mark Davis
//               - first version containing just schedule_timeout()
//
//-----------------------------------------------------------------------------


#include <common.h>

#include <unistd.h>

#include <linux/sched.h>

#include <linux/wait.h>

#include <linux/dummy.h>

#include <string.h>

#include <rabbit.h>
#include <dcdefs.h>
#include <limits.h>



/*-----------------------------------------------------------------------------
 * schedule_timeout - sleep until timeout
 * @timeout: timeout value in jiffies
 *
 * Make the current task sleep until @timeout jiffies have
 * elapsed. The routine will return immediately unless
 * the current task state has been set (see set_current_state()).
 *
 * You can set the task state as follows -
 *
 * %TASK_UNINTERRUPTIBLE - at least @timeout jiffies are guaranteed to
 * pass before the routine returns. The routine will return 0
 *
 * NSCL NOTE:  This is misleading.  @timeout jiffies are NOT guaranteed
 *             to pass if current has registered itself in a wait queue!
 *
 * %TASK_INTERRUPTIBLE - the routine may return early if a signal is
 * delivered to the current task. In this case the remaining time
 * in jiffies will be returned, or 0 if the timer expired in time
 *
 * The current task state is guaranteed to be TASK_RUNNING when this
 * routine returns.
 *
 * Specifying a @timeout value of %MAX_SCHEDULE_TIMEOUT will schedule
 * the CPU away without a bound on the timeout. In this case the return
 * value will be %MAX_SCHEDULE_TIMEOUT.
 *
 * In all cases the return value is guaranteed to be non-negative.
 ----------------------------------------------------------------------------*/
fastcall signed long __sched schedule_timeout(signed long timeout)
{
//  struct timer_list timer;
//  unsigned long expire;


  switch (timeout)
  {
  case MAX_SCHEDULE_TIMEOUT:
    /*
     * These two special cases are useful to be comfortable
     * in the caller. Nothing more. We could take
     * MAX_SCHEDULE_TIMEOUT from one of the negative value
     * but I' d like to return a valid offset (>=0) to allow
     * the caller to do everything it want with the retval.
     */
    schedule();
    goto out;
  default:
    /*
     * Another bit of PARANOID. Note that the retval will be
     * 0 since no piece of kernel is supposed to do a check
     * for a negative retval of schedule_timeout() (since it
     * should never happens anyway). You just have the printk()
     * that will tell you if something is gone wrong and where.
     */
    if (timeout < 0) {
//      printk(KERN_ERR "schedule_timeout: wrong timeout "
//        "value %lx\n", timeout);
      fprintf(DiagLog, "schedule_timeout: wrong timeout value %lx\n", timeout);
//      dump_stack();
      current->state = TASK_RUNNING;
      goto out;
    }
  }

/*
  expire = timeout + jiffies;

  setup_timer(&timer, process_timeout, (unsigned long)current);
  __mod_timer(&timer, expire);
  schedule();
  del_singleshot_timer_sync(&timer);

  timeout = expire - jiffies;
*/

//  printf("-----block for upto %ld ticks\r\n", timeout);  //tdebug

  _EnterCritical_();  current->timeoutTicks = timeout;  _ExitCritical_();

  schedule();

  //NOTE:  timeoutTicks is decremented only by the kernel and only if the
  //       task is blocked (i.e. INTERRUPTIBLE or UNINTERRUPTIBLE)
  timeout = (long) current->timeoutTicks;  // # of ticks left
  current->timeoutTicks = 0;

  if (timeout < 0)  timeout = 0;

//  printf("----- ticks left: %ld\r\n", timeout);  //tdebug

 out:
  return timeout;  // returns # ticks left
}
EXPORT_SYMBOL(schedule_timeout);


/*-----------------------------------------------------------------------------
 * We can use __set_current_state() here because schedule_timeout() calls
 * schedule() unconditionally.
 ----------------------------------------------------------------------------*/
signed long __sched schedule_timeout_interruptible(signed long timeout)
{
  __set_current_state(TASK_INTERRUPTIBLE);
  return schedule_timeout(timeout);
}
EXPORT_SYMBOL(schedule_timeout_interruptible);

//-----------------------------------------------------------------------------
signed long __sched schedule_timeout_uninterruptible(signed long timeout)
{
  __set_current_state(TASK_UNINTERRUPTIBLE);
  return schedule_timeout(timeout);
}
EXPORT_SYMBOL(schedule_timeout_uninterruptible);


//-----------------------------------------------------------------------------

