//----- _farcopy.c ----- 03/28/12 -----(03/28/12)-----

//-----------------------------------------------------------------------------
//  A version of _farcopy() that is compatible with write protection.
//-----------------------------------------------------------------------------
//  Version log (most recent first):
//
//
//  2012-03-28 - Mark Davis
//               - First version of replacement for Softools _farcopy().  This
//                 version is compatible with hardware write protection, which
//                 the Softools version at the time of this writing was not.
//
//-----------------------------------------------------------------------------


/*---------------------------------------------------------------------------*
  The problem that needs to be addressed is this:  If the destination addr
  ends in 0xFFFF, then it will wrap around (i.e. it write to the last byte in
  the 64K block and then the 1st byte in the SAME 64K block).  Even if the
  byte it would written to the first byte in the block is the same as what is
  already there, it will cause a write protection fault if the 1st byte in
  the block is in a 4K segment that is write protected.

  To avoid the possible write protection fault, what we must do is check the
  destination addr each time through the loop.  If the offset in to the 64K
  block is 0xFFFF, then for that pass through the loop, we execute an
  alternative version of the read/write loop that does this:

      - Decrement the destination addr
      - Read 2 bytes from the destination addr
      - Save the 1st byte from the dest
      - Read 2 bytes from the source addr
      - Set 2nd byte from dest = 1st byte from source
      - Write the 2 bytes to the dest
      - incr dest addr twice

 *---------------------------------------------------------------------------*/

#asm

.extern $fmemcpy

    .cseg

__farcopy::

;; C:DE = void far *toPtr,  (pushed dword) = void far *fromPtr,  HL = count

          push  ix
          push  bc
          push  de

          ex    de,hl         ;; de = HL (count)

          ld    ix,hl         ;; c:ix = dest addr

          ld    hl,(sp+8)
          ld    iy,hl
          ld    hl,(sp+10)
          ld    b,l           ;; b:iy = source addr

          ld    de',0x0001    ;; de' = 1

;;----------------------------------------------------------------------------
;;  Each pass through the loop does the following:
;;
;;    - read 2 bytes from the source addr
;;    - save the 1st byte from the source
;;    - read 2 bytes from the dest addr
;;    - replace 1 of the dest bytes read with the 1st one from the source
;;    - write the 2 bytes back to the dest
;;
;;  The only instructions for directly reading from and writing to extended
;;  memory always do 2 bytes.  And these instructions have a little problem:
;;  The don't handle crossing a 64K boundary.  What this means is that when
;;  the address used by instruction ends in 0xFFFF, it will operate on the
;;  last byte in the 64K block and 1st byte in the SAME 64K block (rather
;;  than the 1st byte in the NEXT 64K block).
;;
;;  Since we always read 2 bytes from the destination, change the value of
;;  just one of them, and write the 2 bytes back, we normally don't care '
;;  what addr the byte that isn't being changed came from (so long as it
;;  it the same addr for the read and write operation).
;;
;;  However, DEBROS uses the write protection features of the Rabbit 3000
;;  processor, which means if the write operation wraps, and the 1st 4K segment
;;  in a 64K block is write protected, this function will cause a write
;;  protection fault (the protection feature doesn't care that the value we
;;  are trying to write is the same as the one that is already there).
;;
;;  So to make this function safe when write protection is in use, we have to
;;  check for this circumstance and make sure the write instruction NEVER
;;  wraps.
;;---------------------------------------------------------------------------

;;===== about 103 cycles per byte =====

;; For a 2K block, that would be 210,944 cycles (about 4x as long as fmemcpy)


;;--- If the dest addr doesn't end in 0xFFFF, then the write won't wrap
@loop:
          exx                  ;; [2]
          ld    hl,ix          ;; [4]  if (ix + 1) == 0, use alternate path
          add   hl,de          ;; [2]
          exx                  ;; [2]
          jr    c,@avoidWrap   ;; [5]


        ;; read 2 bytes from source addr
          ld    a,b            ;; [2]
          ldp   hl,(iy)        ;; [10]

        ;; save 1st byte from source
          ld    a,l            ;; [2]
          ex    af,af'         ;; [2]

        ;; read 2 bytes from dest addr
          ld    a,c            ;; [2]
          ldp   hl,(ix)        ;; [10]

        ;; replace 1st byte with 1st byte from source
          ex    af,af'         ;; [2]
          ld    l,a            ;; [2]

        ;; write modified word (with 1st byte from source, 2nd from dest)
        ;; back to the dest addr
          ld    a,c            ;; [2]
          ldp   (ix),hl        ;; [12]

          jr   @decrCount      ;; [5]

;;--- Otherwise, we have to decr the dest addr by one (so it won't wrap) ---
;;--- operate on the 2nd destination byte rather than the 1st one, and ---
;;--- then restore the dest addr for the next pass through the loop ---

@avoidWrap:
          dec   ix             ;; [4]  change offset 0xFFFF to 0xFFFE

        ;; read 2 bytes from source addr
          ld    a,b            ;; [2]
          ldp   hl,(iy)        ;; [10]

        ;; save 1st byte from source
          ld    a,l            ;; [2]
          ex    af,af'         ;; [2]

        ;; read 2 bytes from dest-1 addr
          ld    a,c            ;; [2]
          ldp   hl,(ix)        ;; [10]

        ;; replace 2nd byte with 1st byte from source
          ex    af,af'         ;; [2]
          ld    h,a            ;; [2]

        ;; write modified word (with 1st byte from dest, 2nd from source)
        ;; back to the dest addr
          ld    a,c            ;; [2]
          ldp   (ix),hl        ;; [12]

          inc   ix             ;; [4]  change offset back to 0xFFFF

@decrCount:
          dec   de             ;; [2]  --count
          ld    a,d            ;; [2]
          or    e              ;; [2]
          jr    z,@return      ;; [5]   done if count == 0

          exx                  ;; [2]
          add   ix,de          ;; [4]   ++to
          exx                  ;; [2]
          jr    nc,@incrFrom   ;; [5]
          inc   c              ;; [2]

@incrFrom:
          exx                  ;; [2]
          add   iy,de          ;; [4]   ++from
          exx                  ;; [2]
          jr    nc,@loop       ;; [5]
          inc   b              ;; [2]
          jr    @loop          ;; [5]

@return:
          pop   de
          pop   bc
          pop   ix

          ret

#endasm

//-----------------------------------------------------------------------------

