//----- _fmlt.c ----- 07/13/10 -----(01/09/08)-----

//-----------------------------------------------------------------------------
//  A faster (but less precise) version of the "internal" __fmlt() function
//  for multiplying 2 IEEE 32-bit float point numbers.
//-----------------------------------------------------------------------------
//  Version log (most recent first):
//
//
//  2010-07-13 - Mark Davis
//               - Modified version of my fmult() function that serves as a
//                 direct replacement for Softool's internal __fmlt() (i.e. so
//                 ALL floating point multiplications will be done using this
//                 version rather than the Softools one, without having to call
//                 it explicitly).  This version is about 2.5 times as fast as
//                 the Softools version.
//
//  2008-01-09 - Mark Davis
//               - Completed original version of fmult(), a faster alternative
//                 to Softools' more accurate but slower floating point
//                 multiply.  Tests show that this version is about 2.5 times
//                 as fast as the Softools version (on an RCM3200 about 17-18
//                 ms vs around 44 ms for the Softools version).  The
//                 compromise is that this version can have errors in the last
//                 2 or 3 bits.
//
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
#asm
    .cseg

;;float near __fmlt(f1)

__fmlt::   ;;  BC:DE = f1 * BC:DE


.COMMENT ~
---------- stack usage ----------
  ix-13: manRes (ls byte)  working 32-bit result mantissa
  ix-12: manRes
  ix-11: manRes
  ix-10: manRes (ms byte)

  ix-9:  man2A (ls byte)
  ix-8:  man2A (ms byte)

  ix-7:  man1A (ls byte)
  ix-6:  man1A (ms byte)

  ix-5:  expRes (result exponent)

  ix-4:  f2 (ls byte)
  ix-3:  f2
  ix-2:  f2
  ix-1:  f2 (ms byte)

  ix+0:  saved IX (LSB)
  ix+1:  saved IX (MSB)

  ix+2:  return addr (LSB)
  ix+3:  return addr (MSB)

  ix+4:  f1 (ls byte)
  ix+5:  f1
  ix+6:  f1
  ix+7:  f1 (ms byte)
--------------------------------
  ~

; stack frame offsets
.manRes     equ   -13     ; temp 32-bit mantissa for result
.man2A      equ    -9     ; mantissa 2  (16-bits)
.man1A      equ    -7     ; mantissa 1  (16-bits)
.expRes     equ    -5     ; 1 byte
.f2         equ    -4     ; IEEE float (4 bytes)
.f1         equ     4     ; IEEE float (4 bytes)


    push    ix                 ;[10]

    ld      ix,0               ;[8]
    add     ix,sp              ;[4]

    push    bc                 ;[10] save f2 on the stack
    push    de                 ;[10]

    add     sp,-9              ;[4]  so interrupts don't trash local vars


;  if ((f1 == 0.0) OR (f2 == 0.0))  return 0.0;
    ld      hl,(ix+.f1)        ;[9]
    ex      de,hl              ;[2]
    ld      hl,(ix+.f1+2)      ;[9]
    ld      bc,hl              ;[4]  BC:DE = f1
    ld      a,b                ;[2]
    and     a,7Fh              ;[4]  ignore sign bit
    or      a,c                ;[2]
    or      a,d                ;[2]
    or      a,e                ;[2]
    jr      z,ret0             ;[5]  everything but sign bit 0?

    ld      hl,(ix+.f2)        ;[9]
    ex      de,hl              ;[2]
    ld      hl,(ix+.f2+2)      ;[9]
    ld      bc,hl              ;[4]  BC:DE = f2
    ld      a,b                ;[2]
    and     a,7Fh              ;[4]  ignore sign bit
    or      a,c                ;[2]
    or      a,d                ;[2]
    or      a,e                ;[2]
    jr      nz,not0            ;[5]  everything but sign bit 0?

ret0:
    ld      de,0               ;[6]  return 0.0
    ld      c,e                ;[2]
    ld      b,e                ;[2]
    jp      done               ;[7]

not0:

;#define  f1_0  (*(((uchar *)&f1)+3))
;#define  f1_1  (*(((uchar *)&f1)+2))
;#define  f1_2  (*(((uchar *)&f1)+1))
;#define  f1_3  (*(((uchar *)&f1)+0))

;#define  f2_0  (*(((uchar *)&f2)+3))
;#define  f2_1  (*(((uchar *)&f2)+2))
;#define  f2_2  (*(((uchar *)&f2)+1))
;#define  f2_3  (*(((uchar *)&f2)+0))

; C = MSB(MSW(f1) << 1));    // get f1's exponent
    ld      hl,(ix+.f1+2)      ;[9]  HL = MSW(f1)
    ld      a,h                ;[2]  save byte with f1's sign bit
    ex      de,hl              ;[2]  DE = HL
    rl      de                 ;[2]  D = MSB(DE << 1) (f1's exponent)
    ld      c,d                ;[2]  C = f1's exponent

; D = MSW(f2) << 1    ;  // f2 exponent
    ld      hl,(ix+.f2+2)      ;[9]  HL = MSW(f2)
    xor     h                  ;[2]  MS bit of A has sign bit for result
    sla     a                  ;[4]  now carry bit has sign bit
    ex      af,af'             ;[2]  save carry bit with sign bit for result
    ex      de,hl              ;[2]  DE = HL
    rl      de                 ;[2]  D = MSB(DE << 1) (f2's exponent)

; expRes = (int)((uint)exp1 + (uint)exp2) - 254;
    bool    hl                 ;[2]  H = 0
    ld      b,h                ;[2]  now BC = f1's exponent
    ld      l,d                ;[2]  now HL = f2's exponent
    add     hl,bc              ;[2]  HL = f1 exp + f2 exp
    ld      de,0xff82          ;[6]  DE = 128-254
    add     hl,de              ;[2]  HL = result exponent
    ld      (ix+.expRes),l     ;[10] save 8 bit result exponent

;---------------------------------------------------------------
; close approximation of result Mantissa computed as:
;
;    manRes = man1A*man2A + (man1A*man2B + man2A*man1B) >> 8
;
; where man1A, man2A are the 15 msb of each mantissa (incl implied 1 bit)
;       man2B, man2B are the  9 lsb of each mantissa
;
; For ref:  IEEE 32-bit float format (and breakdown of use by this func):
;
;   s-eeeeeee e-
;             1-AAAAAAA AAAAAAA-B BBBBBBB
;
; where
;   s = sign bit
;   e = exponent bits
;   1 = implied 1 in most significant bit of mantissa
;   A = 14 most significant bits of mantissa (not including implied 1)
;   B =  9 least significant bits of mantissa
;---------------------------------------------------------------

;=== man1A = (*(int *)((char *)&f1+1) | 0x8000) >> 1; // 15 MSB of f1 mantissa
    bool    hl                 ;[2]  to clear carry flag
    ld      hl,(ix+.f1+1)      ;[9]  HL = 2nd and 3rd bytes of f1
    set     7,h                ;[4]  replace exponent bit with implied 1 bit
    rr      hl                 ;[2]  carry -> [HL] -> carry
    ld      bc,hl              ;[4]  BC = 01xx-xxxx xxxx-xxxx
    ld      (ix+.man1A),hl     ;[11] man1A = HL

;=== man2A = (*(int *)((char *)&f2+1) | 0x8000) >> 1; // 15 MSB of f2 mantissa
    bool    hl                 ;[2]  to clear carry flag
    ld      hl,(ix+.f2+1)      ;[9]  2nd and 3rd bytes of f2
    set     7,h                ;[4]  replace exponent bit with implied 1 bit
    rr      hl                 ;[2]  carry -> [HL] -> carry
    ld      de,hl              ;[4]  DE = 01xxxxxx xxxxxxxx
    ld      (ix+.man2A),hl     ;[11] man2A = HL

;=== multiply 2 15-bit mantissa values (NOTE that MS bit is 0 for both values)
    mul                        ;[16] HL:BC = BC * DE

    ld      (ix+.manRes+2),hl  ;[11] MSW(manRes) = HL
    ld      hl,bc              ;[4]
    ld      (ix+.manRes),hl    ;[11] MSW(manRes) = BC

;=== add in (man1A * man2B) >> 8
    ld      hl,(ix+.f2)        ;[9]  HL = LSW(f2)
    ld      a,h                ;[2]
    and     a,0x01             ;[4]  keep only 9th bit from MS byte
    ld      b,a                ;[2]
    ld      c,l                ;[2]  BC = 9 LS bits of f1 mantissa
    ld      hl,(ix+.man1A)     ;[9]
    ex      de,hl              ;[2]  DE = 15 MS bits f1 mantissa (w/ implied 1)
    mul                        ;[16] HL:BC = man1A * man2B
    ld      e,b                ;[2]
    ld      d,l                ;[2]  DE has 12 MS bits of 20-bit result
    rr      de                 ;[2]
    res     7,d                ;[4]
  ; add to 32-bit result mantissa
    ld      hl,(ix+.manRes)    ;[9]  HL = LSW(manRes)
    add     hl,de              ;[2]
    ld      (ix+.manRes),hl    ;[11]
    jr      nc,nocarry1        ;[5]
    ld      hl,(ix+.manRes+2)  ;[9]  HL = MSW(manRes)
    inc     hl                 ;[2]
    ld      (ix+.manRes+2),hl  ;[9]
nocarry1:

;=== add in (man2A * man1B) >> 8
    ld      hl,(ix+.f1)        ;[9]  HL = LSW(f1)
    ld      a,h                ;[2]
    and     a,0x01             ;[4]  keep only 9th bit from MS byte
    ld      b,a                ;[2]
    ld      c,l                ;[2]  BC = 9 LS bits of f1 mantissa
    ld      hl,(ix+.man2A)     ;[9]
    ex      de,hl              ;[2]  DE = 15 MS bits f2 mantissa (w/ implied 1)
    mul                        ;[16] HL:BC = man2A * man1B
    ld      e,b                ;[2]
    ld      d,l                ;[2]  DE has 12 MS bits of 20-bit result
    rr      de                 ;[2]
    res     7,d                ;[4]
  ; add to 32-bit result mantissa
    ld      hl,(ix+.manRes)    ;[9]  HL = LSW(manRes)
    add     hl,de              ;[2]
    ld      (ix+.manRes),hl    ;[11]
    jr      nc,nocarry2        ;[5]
    ld      hl,(ix+.manRes+2)  ;[9]  HL = MSW(manRes)
    inc     hl                 ;[2]
    ld      (ix+.manRes+2),hl  ;[9]
nocarry2:

;=== set HL:DE = manRes
    ld      hl,(ix+.manRes)    ;[9]
    ex      de,hl              ;[2]
    ld      hl,(ix+.manRes+2)  ;[9]

;=== shift HL:DE left 2 or 3 bits (until msb is a 1)

    rl      de                 ;[2]  carry <- [DE] <- carry
    ex      de,hl              ;[2]
    rl      de                 ;[2]  carry <- [DE] <- carry
    ex      de,hl              ;[2]

    rl      de                 ;[2]  carry <- [DE] <- carry
    ex      de,hl              ;[2]
    rl      de                 ;[2]  carry <- [DE] <- carry
    ex      de,hl              ;[2]

    ld      b,(ix+.expRes)     ;[9]  B = 8-bit result exponent

    ;--- if MS bit is not 1, then shift one more time ---
    bit     7,h                ;[4]  F.z
    jr      nz,clear1stbit     ;[5]

    rl      de                 ;[2]  carry <- [DE] <- carry
    ex      de,hl              ;[2]
    rl      de                 ;[2]  carry <- [DE] <- carry
    ex      de,hl              ;[2]

    dec     b                  ;[2]  and decr exponent

clear1stbit:
    res     7,h                ;[4]  reset msb (the implied 1 bit for mantissa)

;=== return result in to BC:DE
    ex      af,af'             ;[2]  carry bit contains sign bit for result
    rr      b                  ;[4]  carry -> [B] -> carry
    ld      c,0                ;[4]
    rr      c                  ;[4]  carry -> [C] -> carry
    ld      a,c                ;[2]
    or      h                  ;[2]
    ld      c,a                ;[2]
    ld      e,d                ;[2]
    ld      d,l                ;[2]

  done:
    ld      sp,ix

    pop     ix

;; The caller expects this func to remove f1 from the stack.  Since the return
;; address is pushed on to the stack AFTER f1 is, we must pop (and save) the
;; address first, then skip over the f1 value and THEN "return" to the caller.

    pop     hl                 ; save return addr
    add     sp,4               ; remove f1 pushed by caller

    jp      (hl)               ; return to caller

#endasm

//-----------------------------------------------------------------------------

