//----- fmemset.c ----- 07/16/10 -----(07/16/10)-----

//-----------------------------------------------------------------------------
//  A version of fmemset() that avoids single write operations that span a
//  4K boundary
//-----------------------------------------------------------------------------
//  Version log (most recent first):
//
//
//  2010-07-16 - Mark Davis (*)
//               - DEBROS version of fmemset() that is compatible with memory
//                 write protection (i.e. it doesn't do writes that cross a
//                 4K boundary).
//
//-----------------------------------------------------------------------------

#include <sys/types.h>
#include <unistd.h>


#pragma offset_labels on

//-----------------------------------------------------------------------------
//  A version of fmemset() that uses the safe version of __farst1 to avoid
//  writing to 4K pages outside the specified range.
//-----------------------------------------------------------------------------
/*
void far * far fmemset(void far *to, int ch, ulong size)
{
  uchar far  *to_ = (uchar far *)to;

  while (size--)  *to_++ = (uchar)ch;

  return to;
}
*/

#pragma warn -758

//-----------------------------------------------------------------------------
//  DEBROS version of fmemset() function that is both faster than the Softools
//  version and compatible with Rabbit 3000A memory write protection
//-----------------------------------------------------------------------------
void far * far fmemset(void far *to, int ch, ulong size)
{
  uchar far *saveTo;
  uint   byteCount;
  ulong  toEnd, safeToEnd;


  saveTo = to;
;
#asm
;   toEnd = (long)to + size - 1;
        ld      hl,(ix+.to)
        ex      de,hl
        ld      hl,(ix+.to+2)
        ld      bc,hl
        ld      b,0
        push    bc
        push    de
        ld      hl,(ix+.size)
        ex      de,hl
        ld      hl,(ix+.size+2)
        ld      bc,hl
        call    __ladd##
        call    __ldec##
        ld      hl,bc
        ld      (ix+.toEnd+2),hl
        ex      de,hl
        ld      (ix+.toEnd),hl
        ex      de,hl

;   while (size)  {
  while:
        ld      hl,(ix+.size)
        ex      de,hl
        ld      hl,(ix+.size+2)
        or      hl,de
        jp      z,done

;   //--- # of target bytes that won't cross a 4K boundary ---
;     safeToEnd = (ulong)to | 0x0FFF;
        ld      hl,(ix+.to)
        ld      de,0x0FFF
        or      hl,de
        ld      (ix+.safeToEnd),hl
        ld      hl,(ix+.to+2)
        ld      (ix+.safeToEnd+2),hl

;     if (safeToEnd > toEnd)  safeToEnd = toEnd;
        ex      de,hl                ;; DE = MSW(safeToEnd) from above
        ld      hl,(ix+.toEnd+2)     ;; HL = MSW(toEnd)
        ex      de,hl                ;; HL = MSW(safeToEnd), DE = MSW(toEnd)
        or      a,a                  ;; clear Carry (for DE == HL case)
        sbc     hl,de                ;; Carry set if HL < DE
        jr      c,toOK               ;; MSW(safeToEnd) < MSW(toEnd) ?
        jr      nz,copyToEnd         ;; MSW(safeToEnd) > MSW(toEnd) ?

        ld      hl,(ix+.safeToEnd)   ;; MSW's equal, so compare LSW's
        ex      de,hl                ;; DE = LSW(safeToEnd)
        ld      hl,(ix+.toEnd)       ;; HL = LSW(toEnd)
        scf                          ;; set Carry (for DE == HL case)
        sbc     hl,de                ;; Carry set if DE > HL
        jr      nc,toOK              ;; LSW(safeToEnd) <= LSW(toEnd)

    copyToEnd:                       ;; set safeToEnd = toEnd
        ld      hl,(ix+.toEnd+2)
        ld      (ix+.safeToEnd+2),hl
        ld      hl,(ix+.toEnd)
        ld      (ix+.safeToEnd),hl
    toOK:

;     byteCount = (uint)safeToEnd - (uint)to + 1;
        ld      hl,(ix+.safeToEnd)
        ex      de,hl
        ld      hl,(ix+.to)
        ex      de,hl
        or      a,a
        sbc     hl,de
        inc     hl
        ld      (ix+.byteCount),hl

;   //--- if at least 2 bytes, use loop to copy 2 bytes at a time ---
;     if (byteCount >> 1)  {
        ld      hl,(ix+.byteCount) ;[9]
        or      a,a                ;[2]
        rr      hl                 ;[2]
        jp      z,copy1            ;[7]  HL == 0 is assumed by copy1!!!!

;===== copy up to 4K bytes, 2 bytes at a time =====
;  //--- set up register for loop ---
;  // reg use:  A[3-0]:IX = from,  A'[3-0]:IY = to,  DE = # of words to copy

       ex    de,hl                 ;[2]   DE = # words

     ;; to address
       ld    a,(ix+.to+2)          ;[9]   A = MSB(to)
       ld    hl,(ix+.to)           ;[9]
       ld    iy,hl                 ;[4]   IY = LSW(to)

       ld    bc,2                  ;[6]   BC = 2 (# of bytes copied each pass)

       ld    hl,(ix+.ch)           ;[10]  L = char to write
       ld    h,l                   ;[2]   now both H and L have byte to write

;;------ tight loop that fills up to 4K bytes, 2 bytes each pass ------
  loop:
       ldp   (iy),hl           ;[12]   *(uint *)to = HL;
       add   iy,bc             ;[4]    to += 2;

       dec   de                ;[2]    --count
       rl    de                ;[2]    set Z flag if DE is 0
       rr    de                ;[2]
       jr    nz,loop           ;[5]    if (count) continue
;;----------------------------------------------------------------------

       ld    hl,iy                 ;[4]   update LSW(to)
       ld    (ix+.to),hl           ;[11]
       or    hl,de                 ;[2]   DE still 0 from above
       jr    nz,updSize            ;[5]   did LSW(to) wrap to 0?
       inc   (ix+.to+2)            ;[12]  inc MSW(to)

   updSize:   ;; size -= (byteCount & 0xFFFE);
       ld    hl,(ix+.byteCount)    ;[9]   HL = byteCount
       ex    de,hl                 ;[2]   DE = byteCount
       res   0,e                   ;[4]   DE = byteCount & 0xFFFE

       ld    hl,(ix+.size)         ;[9]   upate LSW(size)
       or    a,a                   ;[2]   clear Carry
       sbc   hl,de                 ;[4]   Carry set if LSW(size) < DE
       ld    (ix+.size),hl         ;[11]
       jp    nc,while              ;[7]
       dec   (ix+.size+2)          ;[12]  update MSW(size)
       jp    while                 ;[7]


;;===== writing 2 bytes would cross a 4K boundary =====
;;   *to++ = ch;
  copy1:
       ex    de,hl                 ;[2]   DE = 0 (because HL is 0 from the jp)

   ; IY = to;
       ld    a,(ix+.to+2)          ;[9]   A = MSB(to)
       ld    hl,(ix+.to)           ;[9]   HL = LSW(to)
       ld    iy,hl                 ;[4]   IY = LSW(to)

   ; ++to;
       inc   hl                    ;[2]   ++LSW(to)
       ld    (ix+.to),hl           ;[11]
       or    hl,de                 ;[2]   DE still 0 from above
       jr    nz,testTo             ;[5]   did LSW(to) wrap to 0?
       inc   (ix+.to+2)            ;[12]  incr MSB(to)

   ; is target addr even or odd? (remember it is already incremented!)
    testTo:
       bit   0,l                   ;[4]   test bit that is already reversed
       jr    nz,evenAddr           ;[5]   nz ==> bit is 1, ==> it WAS 0 b4 ++!

;; The target addr is an odd #, so we decrement it to insure both bytes
;; we are about to write will be in the same 4K block.  Using the decremented
;; address, L = the preeceding byte, H = the target byte
       dec    iy                   ;[4]   target addr now even
       ldp    hl,(iy)              ;[10]  read word from target addr - 1
       ld     h,d                  ;[2]   chg byte to be updated (D still 0)
       jr     writeWord            ;[5]

;; The passed addr is an even #, so the read/write will not span a 4K boundary.
;; L represents the target addr, H represents the one after it
    evenAddr:
       ldp    hl,(iy)              ;[10]  read word from target addr
       ld     l,d                  ;[2]   chg byte to be updated (D still 0)

    writeWord:
       ldp    (iy),hl              ;[12]  write copied and updated bytes


;;   --size;
       ld     hl,(ix+.size)        ;[9]  -- LSW(size)
       dec    hl                   ;[2]
       ld     (ix+.size),hl        ;[11]

       ld     a,h                  ;[4]  is HL now 0xFFFF?
       and    l                    ;[2]
       inc    a                    ;[2]  Z set if HL is 0xFFFF
       jp     nz,while             ;[7]

       ld     hl,(ix+.size+2)      ;[9]  -- MSW(size)
       dec    hl                   ;[2]
       ld     (ix+.size+2),hl      ;[11]
       jp     while

  done:

#endasm

  return saveTo;
}

#pragma warn +758

//-----------------------------------------------------------------------------

#pragma offset_labels off


