//----- btof.c ----- 05/31/11 -----(03/14/06)-----

//-----------------------------------------------------------------------------
//  Change log (most recent first):
//
//
//  2011-05-31 - Mark Davis
//               - Now checks for original ZWorld/Rabbit ADC/DAC calibration
//                 data even if NSCL userblock entries not found.  Prior to
//                 this, there was no easy way to extract the factory-set
//                 calibration data on a brand new board.
//
//  2011-03-07 - Mark Davis
//               - Made blockList (a local variable that is declared static
//                 simply to avoid overflowing the stack) is now declared
//                 far (to reduce use of root memory).
//
//  2010-08-04 - Mark Davis
//               - Corrected formatting for dump of ADC/DAC calib data
//
//  2010-07-29 - Mark Davis
//               - The ADC and DAC calibrations are now written to a file
//                 named benchcalibs in a format that can be run directly
//                 using the sh command.
//
//  2010-07-28 - Mark Davis
//               - Update to use standard open/write/close calls to write to
//                 files.
//               - Added code to show ADC/DAC calibration values.
//
//  2006-03-14 - Mark Davis
//               - original version
//
//-----------------------------------------------------------------------------

#include <time.h>
#include <common.h>

#include <sys/types.h>
#include <unistd.h>
#include <sys/stat.h>
#include <sys/socket.h>
#include <fcntl.h>
#include <arpa/inet.h>
#include <errno.h>
#include <signal.h>
#include <utime.h>


#include <busybox.h>

#include <sys/hdwDefs.h>

#include <sys/srfs.h>

#include <dcdefs.h>

#include <ctype.h>
#include <limits.h>

#include <sys/SHD_headers.h>

#include <sys/modbus.h>

#include <tsh.h>

#include <applic.h>

#include <sys/farprintf.h>


//=============================================================================
//  command-line task functions moved from busybox1 to split size of busybox
//  modules to address link-time problems (busybox was over 200K before being
//  split in two)
//=============================================================================

//-----------------------------------------------------------------------------
#define  MaxCfgBlocks   64  // from older ZWorld compiled code

typedef struct  {
  uint  blockSig;     // marker to indicate existance and vers of this struct
  int   blkListSize;  // size of this structure
  int   numBlks;      // # blocks currently in use
  int   blkSize[MaxCfgBlocks];  // size of each block
  uint  goofyCRC[MaxCfgBlocks];   // CRCs (goofed vers for DLM compatability)
}  BlockList;

#define  CfgBlockSig    0xfe01  // version 01 (1st byte = 1's complment of 2nd)


//-----------------------------------------------------------------------------
//  Test the specified address to see if it contains a table of reasonable ADC
//  and DAC calibration values.
//  Returns the # of valid looking values found
//-----------------------------------------------------------------------------
int readCalibs(ulong calibAddr, uchar *buf, uint bufSize, bool saveValues)
{
  bool  validData;
  uint  u, adcVals, numVals;
  int  fd, validCount = 0;
  float far *fptr;
  float  gain, offset;
  char temp[128];


  if (saveValues)  {
    unlink("benchmarks");
    if ((fd = open("benchcalibs", O_CREAT | O_RDWR, 0x666)) < 0)  {
      perror("benchcalibs");  Exit(-1);  }
  }

  ReadFlash((FlByte far *)calibAddr, buf, bufSize);

  fptr = (float *)buf;

  for (adcVals=0; adcVals<2; ++adcVals)  {  // 2 sets: DACs, then ADCs
    numVals = (adcVals ? 11 : 8);

    for (u=0; u<numVals; ++u,fptr+=2)  {
      gain = FABS(fptr[0]);  offset = FABS(fptr[1]);  validData = false;

      if (cpuID)  {  // NOT a BL2105
        if ((gain > 280e-6) AND (gain < 330e-6)
          AND (offset > 30e+3) AND (offset < 34e+3))  validData = true;
      }
      else  {
        if (adcVals)  {  // ADC values
          if ((gain > 4.8e-3) AND (gain < 5.2e-3)
            AND (offset > 1.95e+3) AND (offset < 2.15e+3))  validData = true; }
        else  // DAC values
          if ((gain > 2.3e-3) AND (gain < 2.8e-3)
            AND (offset > 3.90e+3) AND (offset < 4.3e+3))  validData = true;
      }

      if (!validData)  continue;

      if (!cpuID)  gain = -gain;

      ++validCount;

      if (saveValues)  {
        sprintf(temp, "  %scalib %2u  %s  %s -quiet\r\n",
               (adcVals ? "adc" : "dac"), u,
               Engineering(gain, 7, temp+64),
               Engineering(offset, 7, temp+80) );
        printf(temp);  write(fd, temp, strlen(temp));
      }
    }

    if (saveValues)  { printf("\r\n");  write(fd, "\r\n", 2); }
  }

  if (saveValues)  close(fd);

  return validCount;
}



//-----------------------------------------------------------------------------
//  convert old UserBlock blocks to files
//-----------------------------------------------------------------------------
void _task _btof(int argc, char **argv)
{
  uint  u, remain, count;
  int  fd;
  ulong  dataOffset, addr, calibAddr = 0;
  char  *fileName;
  static BlockList  far blockList;  // static only to avoid overflowing stack
  uchar  buf[256];


  printf("Searching for old UserBlock 'files'...\r\n");

  for (addr=0; addr<0x80000L; addr+=0x1000)  {
    *buf = 0;
    ReadFlash((FlByte far *)addr, (char far *)&blockList, sizeof(blockList));
    if ( (blockList.blockSig != CfgBlockSig)
      OR (blockList.blkListSize != sizeof(blockList)) )  continue;

    printf("\r\nOld UserBlock 'files' at Flash addr: 0x%05lX\r\n\n", addr);

    dataOffset = addr + sizeof(blockList);
    calibAddr = (addr & 0xF0000L) | 0xFC00;

    for (u=0; u<MaxCfgBlocks; ++u)  if (blockList.blkSize[u])  {
      remain = blockList.blkSize[u];

      printf("  Block %u: %4u bytes: ", u, remain);

      if (u > 3)  { printf("<ignored>\r\n");  continue; }

      switch (u)  {
        case 0:  fileName = "IPConfig";  break;
        case 1:  fileName = "ADC_DAC_Calib";  break;
        case 2:  fileName = "DLPInfo";  break;
        case 3:  fileName = "ApplCfg";  break;
      }

      printf("%s\r\n", fileName);  unlink(fileName);

      fd = open(fileName, O_RDWR | O_CREAT, 0x666);
      if (fd < 0)  { perror(fileName);  continue; }


      while (remain)  {
        if ((count = remain) > sizeof(buf))  count = sizeof(buf);
        ReadFlash((FlByte far *)dataOffset, buf, count);
        if (write(fd, buf, count) != count)  { perror("write");  break; }
        dataOffset += count;  remain -= count;
      }

      close(fd);
    }

    printf("\r\n");
  }

//--- Scan flash looking for ADC and DAC calibrations ---
  printf("Searching for ADC/DAC calibrations...\r\n");
  for (calibAddr=0; calibAddr<0x7FFFFL; calibAddr+=256)  {
    printf("  0x%05lX\r", calibAddr);  fflush(stdout);  CheckETX();
    if (readCalibs(calibAddr, buf, sizeof(buf), false) > 3)  break;
  }
//--- if we found valid-looking data, then save it to a shell-script file ---
  if (calibAddr < 0x7FFFFL)  {
    printf("\r\nReading ADC/DAC calibs from 0x%05lX:  benchcalibs\r\n\n",
            calibAddr);
    readCalibs(calibAddr, buf, sizeof(buf), true);
  }


  Exit(0);
}

//-----------------------------------------------------------------------------
