//----- ps.c ----- 03/01/12 -----(03/14/06)-----

//-----------------------------------------------------------------------------
//  Update log (most recent first)
//
//
//  2012-03-01 - Mark Davis
//               - Corrections to assumptions about the contents of the stack
//                 for each task.
//
//  2012-01-03 - Mark Davis
//               - PC (Program Counter) value now shows the correct value for
//                 preempted tasks and for the current task.
//               - The IP (Interrupt Priority) register value that was in
//                 effect for each task is now shown.
//
//  2010-05-13 - Mark Davis
//               - Removed column for useTimeout from the -diag output option
//
//  2010-03-31 - Mark Davis
//               - Added more info about the stack to the -diag option output
//
//  2009-06-27 - Mark Davis
//               - Added -diag option to provide more information about the
//                 state of tasks for debugging purposes.
//
//  2009-xx-xx - Mark Davis
//               - original version of ...
//
//-----------------------------------------------------------------------------

#include <time.h>
#include <common.h>

#include <sys/types.h>
#include <unistd.h>
#include <sys/stat.h>
#include <sys/socket.h>
#include <fcntl.h>
#include <arpa/inet.h>
#include <errno.h>
#include <signal.h>
#include <utime.h>


#include <busybox.h>

#include <sys/hdwDefs.h>

#include <TIMERB.h>

#include <dcdefs.h>

#include <ctype.h>
#include <limits.h>

#include <sys/SHD_headers.h>

#include <tsh.h>

#include <applic.h>

#include <sys/farprintf.h>

//-----------------------------------------------------------------------------
// copy the cmd for the specified process and all its arguments to a string
//-----------------------------------------------------------------------------
char *fmtCmd(char *str, int strLen, task_t *task)
{
  uchar  i;
  char  *ptr = str, *arg;

  for (i=0; i<task->argc; ++i)  {
    arg = task->argv[i];
    if (ptr - str + strlen(arg) > strLen-2)  break;
    sprintf(ptr, "%s ", arg);
    while (*ptr) ++ptr;
  }

  return str;
}

//-----------------------------------------------------------------------------
//  show process status
//
//  See notes in unistd.c:postMortem() regarding the contents of the stack
//-----------------------------------------------------------------------------
void _task _ps(int argc, char **argv)
{
  PID  i;
  task_t  *task;
  bool  diagMode;
  uint  sp, xpc, pc, ip;
  ulong  absPC, idle, secsUsed, stackStart, stackEnd;
  char  *ptr;
  uchar far *stackPtr;
  task_t  tsk;
  char  msg[200];

/*
     UID        PID  PPID  C STIME TTY          TIME CMD
     root         1     0  0 Mar01 ?        00:00:00 init [2]
     root         2     1  0 Mar01 ?        00:00:00 [ksoftirqd/0]
*/

  if (showHelp(1, argc, argv))  {
    printf("\nusage:  ps [-diag]\r\n");  Exit(-1); }

  diagMode = (argc > 1)  ? StriEq(argv[1], "-diag") : false;

  printf("\r\n");

  if (!diagMode)  printf(
    "UID        PID  PPID  PRI  STKAV  STATUS      IDLE        TIME CMD\r\n");
   //xxxxxxxx ##### #####  ###   ####  ########  ######  ####:##:## xxxxxxxxxxxxxxxxx
  else printf(
    "                              ---------Stack--------- \r\n"
    "  PID  PRI  STATUS      IDLE  Start  End   SP  Unused  --PC--  --IP---  state sTcks pr TOTicks  CMD\r\n");
   //#####  ###  ########  ######  ##### ##### #####  ####  ######  x-x-x-x  ##### #####  # #######  ############

  for (i=0; i<NUM_TASKS; ++i)  {
    task = tasks + (uint)i;  if (!task->inUse)  continue;

    ptr = msg;

    KLOCK();  memcpy(&tsk, task, sizeof(task_t));  KUNLOCK();

    idle = msSince(tsk.lastRunning);

    secsUsed = tsk.msUsed / 1000;

    if (!diagMode)
      sprintf(ptr, "root     %5d %5d  %3d   %4d  ",
              tsk.pid, tsk.ppid, tsk.priority, _stackFree_(tsk.pid) );
    else
      sprintf(ptr, "%5d  %3d  ", tsk.pid, tsk.priority );

    strcat(ptr, statusDescr(i));  strcat(ptr, "  ");  while (*ptr) ++ptr;

    fmtIdleTime(ptr, idle);  while (*ptr) ++ptr;

    if (!diagMode)
      fmtHMSUsed(ptr, secsUsed);
    else  {
      KLOCK();

      if (task == current)  {
        sp = _asm( [_HL], "ld hl,0", "add hl, sp", _HL );
        stackPtr = (uchar far *)( ((ulong)(task->sseg) << 12) + (ulong)sp );
        pc = getPC();
        xpc = _asm( [_A], "ld a,xpc", _A );
        ip = _asm( [_HL], "push ip", "push ip", "pop hl", _HL ) & 0xFF;
      }
      // See notes in unistd.c:postMortem() regarding stack contents
      else  {  // suspended tasks (i.e. all BUT the current one)
        sp = (uint)task->stack;
        stackPtr = (uchar far *)( ((ulong)(task->sseg) << 12) + (ulong)sp );
        if (!task->preempt_count)  {  // schedule() called by kernel func
          pc = *(uint far *)(stackPtr + Schedule_PushCount);
          xpc = (uint) stackPtr[Schedule_PushCount + 2];
        }
        else if (task->preempt_count >= 8)  {  // schedule() called by KUNLOCK()
          pc = *(uint far *)(stackPtr + Schedule_PushCount + 3);
          xpc = (uint) stackPtr[7];
        }
        else  {  // schedule() called by timer_interrupt()
          pc = *(uint far *)(stackPtr + Schedule_PushCount + 3 +
                             TimerBISR_PushCount + IntShim_PushCount);
          xpc = (uint) stackPtr[10];
        }
        ip = (uint)stackPtr[4];
      }

      absPC = (ulong)pc;
      if (pc >= 0xE000)  absPC += ((ulong)xpc << 12);

      stackStart = ((ulong)task->stack_start + (((ulong)task->sseg) << 12));
      stackEnd = stackStart + task->stack_size - 1;

      KUNLOCK();

      sprintf(ptr,
              "%05lX %05lX %05lX  %4d  %06lX  %X-%X-%X-%X  %5u %5lu  %d %7lu  ",
              stackStart, stackEnd, (ulong)stackPtr, _stackFree_(tsk.pid),
              absPC, ip >> 6, (ip >> 4) & 3, (ip >> 2) & 3, ip & 3,
              tsk.state, tsk.sleepTicks, tsk.preempt_count,
              tsk.timeoutTicks );
    }
    while (*ptr) ++ptr;

    fmtCmd(ptr, sizeof(msg)-(ptr-msg), task);

    strcat(ptr, "\033[K\r\n");  printf(msg);
  }

  Exit(0);
}


//-----------------------------------------------------------------------------
