//----- PhaseCtlrAppl.c ----- 01/08/13 -----(03/05/06)----

// scope 5 appllooptime latency tcpticktime kernellooptime fwdampl

//-----------------------------------------------------------------------------
//  Application-specific code run at a regular interval
//  Application code to measure and control phase of RF frequency signals.
//  Incorporates the following hardware:
//
//  - Single Board Computer (RCM3200: 44 MHz 8-bit CPU, 768K RAM, 512K flash)
//  - Serial EEPROM (ATMEL AT26F004, 512K bytes)
//  - LMK02002 (Precision Clock Conditioner with Integrated PLL)
//  - Xilinx Spartan-II FPGA
//  - DSP (make/model ?)
//  - high-speed ADC (still using ADC644x?)
//-----------------------------------------------------------------------------
//  Update log (most recent first):
//
//
//  2013-01-08 - Mark Davis
//               - updateCurState() now reads a waveform OR does the periodic
//                 I/O in a pass, but not both.
//
//  2012-11-13 - Mark Davis
//               - Added #defines for DSP registers CavHighIlkLevel and
//                 FwdHighIlkLevel, entries in the table need to connect them
//                 to the corresponding new registers, and added the registers
//                 to the list of persistent values.
//
//  2012-09-26 - Mark Davis (test)
//               - Added OUTI() macro to make it easy to switch between the
//                 slower WrPortI() and faster outi() methods of writing to
//                 an internal register value.
//
//  2012-09-21 - Mark Davis (test)
//               - Went back to using a 1MHz clock for the progdsp command to
//                 see if that allows it to work reliably again.
//
//  2012-09-10 - Mark Davis
//               - Changed so setAtten no longer has any affect (per Nathan)
//
//  2012-08-03 - Mark Davis
//               - Modified readPlot() to use new variants on SPI interface to
//                 reduce overhead.
//               - Incr clock freq for talking to DSP to 2 MHz which further
//                 reduced the time needed to read the data for a waveform.
//
//  2012-08-01 - Mark Davis
//               - Tweaking locking and timeouts to address performance issues.
//
//  2012-07-27 - Mark Davis
//               - Improved efficiency of PV_UpdateValWF() by having it test
//                 the time stamps before even attempting to get the locks.
//
//  2012-07-25 - Mark Davis
//               - PV_UpdateValWF() now updates the size and numElems values
//                 for the current PV value
//               - For non HT_PHASE_CTLR hardware, readWaveform() now uses the
//                 typical # of elements for each waveform that the DSP
//                 specifies on HT_PHASE_CTLR hardware.
//
//  2012-07-23 - Mark Davis
//               - The basic structural changes to support locking of "extended"
//                 PV values (ones non defined via the ModBus register macros)
//                 are now working.
//               - Corrected the "simulated" 16-bit ampl/phase waveform values
//                 computed for testing the code on non-PhCtlr module hardware.
//
//  2012-07-09 - Mark Davis
//               - Added code to allow basic testing of triggering and reading
//                 of fake CTF waveforms on non-PhaseCtlr hdw.
//               - Added "finished" log msg for CTF calculation
//
//  2012-04-27 - Mark Davis
//               - replaced the 16-bit signed stepperReadPos and stepperSetPos
//                 registers with 32-bit signed registers
//
//  2012-04-19 - Mark Davis
//               - Rewrote getLock() and modified releaseLock() to use unnamed
//                 shared/global semaphores.  Prior to this there were still
//                 instances where the locks would not be correctly released.
//
//  2012-04-18 - Mark Davis
//               - Changed all WrPortI() calls to outi() calls which generates
//                 smaller/faster code.
//
//  2012-04-02 - Mark Davis
//               - Began adding structures to test use of extensions to new CA
//                 interface that handle values other than those defined via
//                 the ModBus register macros.
//
//  2012-03-05 - Mark Davis
//               - Updated default values for several settings per Nathan.
//
//  2012-01-16 - Mark Davis
//               - Changed argument for releaseLock() to uint.  This is so it
//                 is consistent with getLock() and to avoid any possible
//                 compiler errors (and excess code) converting between uchar
//                 and uint.
//               - Changed all getLock calls so a return value  of 0 is not
//                 treated as meaning that it got the lock.  This shouldn't be
//                 possible, but better safe than sorry.
//
//  2010-12-01 - Mark Davis
//               - Added Pulse_Ramp DSP RW and ModBus registers and made it
//                 a persistent value.
//
//  2010-10-29 - Mark Davis
//               - Added setDSPState to the saved settings file
//
//  2010-09-16 - Mark Davis
//               - Added #defines for DSP registers FwdAmplGain and CavAmplGain
//                 and entries in the table need to connect them to the register
//                 values fwdAmplGain and cavAmplGain.
//               - Added rfq_t1_offset and rfq_t2_offset to the list of
//                 persistent values.
//
//  2010-08-26 - Mark Davis
//               - The Mode for the DSP register values RFQ_T1_Offset and
//                 RFQ_T2_Offset was changed from ONCE to EVERY
//
//  2010-06-15 - Mark Davis
//               - Added a floating point version of the dspTemp register and
//                 displayed the floating point vesion on the state screen.
//
//  2010-06-10 - Mark Davis
//               - Added autocal command
//
//  2010-05-07 - Mark Davis
//               - Moved initialization of FPGA to after the PLL init (per
//                 Nathan, who said it isn't working the other way since I
//                 fixed the sleepMS() func to work properly when called from
//                 the applic task).
//
//  2010-04-26 - Mark Davis
//               - changes to accomodate updated wait queue interface
//
//  2010-04-20 - Mark Davis
//               - Support for new registers that provide tuner-related values
//                 that get forwarded to other devices via EPICS.
//
//  2010-03-31 - Mark Davis
//               - Added resetfpga command
//
//  2010-02-18 - Mark Davis
//               - Added maxOutAmplGain/Offset
//
//  2010-02-12 - Mark Davis
//               - fwdCavOffset is now a PHASE type
//               - all new phase values are now "wrapped" to 0 <= val < 360
//               - added maxAmplSet register, made it persistent, and used it
//                 to limit all ampl-related settings
//               - added ability to use a pllInitSeq stored in a flash file
//
//  2010-02-08 - Mark Davis
//               - Chgd piezoPos and stepperPos to piezoSetPos, piezoReadPos,
//                 stepperSetPos, stepperReadPos
//
//  2010-02-04 - Mark Davis
//               - More tweaks, fixes to scaling logic and settings
//               - Added dspcalib command to make calibrating the gain and
//                 values easy (fixed new command on 2010-02-05)
//
//  2010-02-03 - Mark Davis
//               - Tweaks, fixes to scaling logic and settings
//
//  2010-01-26 - Mark Davis
//               - First reasonably complete vers with revamped DSP interface
//
//  2009-12-14 - Mark Davis
//               - Changed order of param files to deal with problems caused
//                 by the order in which they were restored at startup.
//
//  2009-12-11 - Mark Davis
//               - started changes to accomodate overhaul of DSP registers
//
//  2009-10-29 - Mark Davis
//               - Updated pllInitSeq's from Dan Morris
//
//  2009-08-04 - Mark Davis
//               - Changed ADC init sequence from 0400 0443  to 0400 0403
//                 (in initDefaults() func)
//
//  2009-06-30 - Mark Davis
//               - Changed pllInitSeq0[] again per Nathan
//
//  2009-06-26 - Mark Davis
//               - Changed pllInitSeq0[] per Nathan
//
//  2009-06-23 - Mark Davis -
//               - Changed all watchdog timer resets to call resetWDT() to
//                 work with new reboot diagnostics.
//
//  2009-06-16 - Mark Davis -
//               - Changed default value for pllInitIdx to 1
//
//  2009-06-08 - Mark Davis -
//               - Added piezoPos and stepperPos values
//
//  2009-03-30 - Mark Davis -
//               - Added new Modbus register values:
//                 - Two floating-point and one int value for readings from an
//                   external device (updated by the control system)
//                 - Two floating-point and one int value for settings for an
//                   external device (forwarded to the device by the ctl system)
//
//  2009-03-21 - Mark Davis -
//               - Added digInDescr[] and digOutDescr[] to support the newer
//                 generic method for logging changes of interest in the state
//                 of the digital I/O signals.
//
//  2009-03-02 - Mark Davis -
//               - Added ability to use 1 of 2 different PLL init sequences
//               - The DSP is now reset after the PLL, ADC initialization
//
//  2008-10-01 - Mark Davis -
//               - added WriteDig() calls to resetDSP() so it works again...
//
//  2008-09-25 - Mark Davis -
//               - moved init of PLL settings to after init of FPGA
//
//  2008-09-24 - Mark Davis -
//               - corrrections to changes for new clock ctlr/PLL
//
//  2008-09-23 - Mark Davis -
//               - removed outdate progfpga and related code
//
//  2008-09-22 - Mark Davis -
//               - changes to fix FPGA download problems
//
//  2008-09-09 - Mark Davis -
//               - changes to support new clock ctlr/PLL (National LMK02002)
//
//  2008-05-21 - Mark Davis -
//               - changes to support new ADC chip on 3rd prototype
//
//  2008-05-19 - Mark Davis -
//               - changes to properly recognize when FPGA has initialized on
//                 latest prototype
//
//  2008-04-30 - Mark Davis -
//               - first set of changes for 3rd prototype with serial EEPROM,
//                 modified FPGA and DSP interfaces
//
//  2007-04-11 - Mark Davis -
//               - adding logic to implement pulsed settings for tuning the
//                 PID parameters
//
//  2007-03-16 - Mark Davis -
//               - began adding application-specific data to Modbus register
//                 set
//
//  2007-03-14 - Mark Davis -
//               - changed code to include (although commented out in this
//                 application for now) a call to rwIO() (which reads and
//                 writes the analog and digital I/O signals).
//
//  2007-03-13 - Mark Davis -
//               - added DISABLE_AUTO_IO to disable the normal reading/writing
//                 of digital and analog I/O signals by the shared application
//                 code.
//
//  2007-03-04 - Mark Davis -
//               - reworked X-Macro usage to allow application-specific
//                 register definitions to reside in the same file as the
//                 rest of the application-specific code.
//
//  2007-02-09 - Mark Davis -
//               - made WrRdDSP more general and more capable to simplify the
//                 funcs that call it
//
//  2007-02-02 - Mark Davis -
//               - streamlined SPI communications with DSP
//
//  2006-10-05 - Mark Davis -
//               - ported Adam Molzhan's old Dynamic-C version of code
//                 and did major cleanup & reorganization to streamline it
//                 (now manages 50 cycles/sec rather than 3)
//
//-----------------------------------------------------------------------------
/*
NOTE:  CTF = Cavity Transfer Function (which is no longer supported...)

Overview of the startup sequence:

  - The EEPROM contains the code for the FPGA.  On startup, one of the digital
    inputs to the FPGA (FPGA_SPI) is toggled, which causes it to take over the
    SPI connection to the EEPROM and load the binary image stored there.  A
    digital output from the FPGA is monitored to indicate when this has
    completed (it takes a little less than 2 seconds).

  - The code for the DSP is stored in the DSP itself.  Similar to how the
    DEBROS update command can be used to update the DEBROS boot image, the DSP
    code provides a mechanism for updating the copies of the DSP code stored
    within the DSP.  When the DSP is running a working copy of the DSP code,
    the progdsp command can be used to download a new copy of DSP code in to
    the DSP itself.

    So updating or storing the initial working copy of DSP code requires that
    a working copy of DSP code is already running in the DSP.  If a working
    copied is stored at the correct place within the DSP memory, then the DSP
    will load it and run it after being reset.  But if there isn't a working
    copy in the DSP memory, then getting it to run a working copy can only be
    accomplished by running the appropriate development software on a desktop
    PC and use of a special cable between the PC and the controller (again,
    very similar to how the DLM is downloaded to a Rabbit Core Module).


Communicating with the DSP:

  - Exchange of data with the DSP is done using a bi-directional (simultaneous
    send/recv) clocked SPI interface.  All data is exchanged in 32-bit (4-byte)
    packets.  While we are sending a 32-bit packet to the DSP, it is using the
    same clock pulses to simultaneously send a 32-bit packet in response.

    Each exchange of a 32-bit packet consists of setting the DSP /Enable signal
    low, writing (while simultaneously reading) 32-bits of data, and then
    setting the DSP /Enable signal high.

    Due to the way the DSP buffers are set up, all responses from the DSP are
    always 2 packets "behind" the ones that contained the commands that they
    are a response to.  We can write as many commands as we want in a
    sequence of packets, but the response to each one will always be received
    2 packets later.  Given that we generally want to read the response to
    a set of commands as soon as possible, we normally have to add some
    dummy or filler packets to the sequence to get the desired response
    packets.  For extended responses (like reading of waveform data), we must
    send lots of filler packets to get all the data.


Revamped register conventions (Jan 2010):

  - Communication with the DSP is done by reading and writing "register"
    values, each of which has an "address".  Even numbered addresses are for
    reading and odd numbered ones are for writing (i.e. the least significant
    bit specifies a read or write operation).

  - Reading a register value is done by sending the 4-byte address of the
    register and then reading (2 packets later) its 4 byte value.

  - Writing to a register is done by sending the 4-byte address and then the
    4-byte value.  The DSP responds (with the usual 2-packet delay) with the
    4-byte address and 4-byte value as a means of verifying that the we are
    communicating with the DSP and that it received the correct values.

    The meaning of the 4-byte value is dependent on the register.  The most
    common representations used are 32-bit floats, 32 and 16-bit integers,
    and bit-mapped values.  The values are always in network byte order (most
    significant first).


  - Registers come in several varieties:

      Read-Only-and-Continuously:
         The value should be read each pass to keep track of the current state.

      Write-Only-and-Continuously:
         These represent settings that we want asserted all the time.  If the
         DSP resets, these values will be quickly re-asserted because they are
         written to the DSP each pass.

      Write-Once, Read-never:
         Writing to the value triggers some action so writing to them each
         pass is not an option.  These registers generally provide the means
         to write to one instance in a table of values maintained within the
         DSP itself, where the row written to is based on the current state
         of the DSP.

      Read/Write-continuously:
         Such registers have both a separate read and write register and often
         corresponding separate read and write ModBus values.  The usual case
         is that the write value represents a setpoint and read value
         represents the current state or readback value.

      Read-Trigger:
         Reading from the register triggers some event such as starting or
         stopping a calculation, so it is done only when we want to trigger
         the corresponding process or event.


  - Most the DSP registers have a corresponding ModBus register.  Tables are
    used to provide the information that links them together.  About 20 times
    per second the main application task uses these tables to read and write
    values from/to the DSP.

    The table for the read-only registers causes the ModBus register values
    that correspond to each DSP register to be updated.  The table for the
    writable registers is used to write the current value of each of the
    corresponding ModBus registers to the DSP.  This means that, if the DSP
    resets for any reason, the current state will be restored very quickly.

  - The dspread and dspwrite commands provide a mechanism to manually read
    values from and write values to the DSP.  Just be careful about using
    registers that trigger some event within the DSP as this can interfere
    with normal operation.

  - To facilitate testing, the normal periodic reading from and writing to the
    DSP registers can be disabled or enabled using the dspio command.


  - The register values provide information about the state of the DSP.
    One of the registers is a bit-mapped status value, with bits that
    indicate if new waveform data is available for reading.

    For each waveform we check to see if we have an open connection to a
    client for the waveform and if there is new data or a fixed interval
    of time has elapsed.  If these conditions are met, then we read the
    corresponding waveform data as follows:

      - Using the write/read 4-bytes methodology described above,
        read the # of 32-bit words of data available for the waveform.

      - Using the value 0x0010##00 for the 4 bytes we write each
        time (where ## indicates which waveform we want data for),
        use the write/read 4-bytes methodology to read the
        indicated # of 32-bit words of data.

    A couple of the waveforms take around 40 seconds to compute and the
    DSP must be told when to compute them.  For these an additional
    register provides the %complete the computation is so we can monitor
    the progress of the computation and determine when it is available
    for reading.


 - Some of the bit-mapped registers we can write to represent multiple values
   each represented by more than one bit.  When writing to these registers,
   all 0's for a given group means do not change the value of any of the
   bits in that group.

   To insure that the DSP state is properly restored if it is reset, the last
   value of each bit-group within such registers must be kept track of.  This
   minimizes problems caused by DSP resets.


 - The SPI interface used to communicate with the DSP is also used to
   communicate with the FPGA, the PLL, and the ADC's.  To allow multiple tasks
   to communicate with with these devices, a locking mechanism is provided to
   insure that only one task is using the shared SPI connection and related
   data values at any given time.


 - The DSP xxxCalib value registers (0x0010022A - 0x001000239) are used
   to write a given value in to a table of values based on the current state
   of another value (attenuations?), so they cannot be included in the set
   that is constantly being written to.  Writing to one of these registers
   is done just once when you need to write a value to the table stored in
   the DSP.  Saving the table is done by reading from the register at the
   end of that range (SaveCalibToDSP: 0x0010023A).

------------------------------------------------------------------------------*/


//-----------------------------------------------------------------------------
//  TO-DO:
//
//  - Reading/writing the DSP registers takes around 15 ms (?) at present.
//    Given the various imposed limits on how often we read/write the DSP
//    registers and how often we read waveform data, I am not sure if the
//    reading of waveform data can (and does) happen during the same pass.
//    If it does, I should change it so it does NOT, which would reduce the
//    time the applic loop runs during a pass that we are reading a waveform.
//
//  - Consider creating a faster version of readPlot() in assembly, as this
//    is one of the functions that consumes much of the CPU time during many
//    of the executions of the main applic task's loop.
//
//  - Same for read_DSP_Regs() and write_DSP_regs(), although these are a bit
//    trickier as they have to read-from-DSP/convert/save-to-PV or
//    read-from-PV/convert/write-to-DSP for each value.
//
//  - If it turns out we don't need slower code for the setting/clearing the
//    SPI-enable signal for the DSP, then consider making it even faster by
//    using the set/reset internal reg bit macros ioSetBitI() and ioClearBitI().
//
//-----------------------------------------------------------------------------
//  Work in progress:
//
//  - Need to deal with multiple sources of R/W requests for the DSP.  The
//    situation is not as simple as representing the DSP as a std streaming
//    device that reads() and writes() can be interleaved on.  The following
//    constraints exist:
//
//    - SPI port 1 is connected to the FPGA, PLL, DSP, and ADC's
//
//    - Different control bits and clock modes are required for different
//      devices
//
//    - Reads and Writes to the DSP are not independant.  In at least some
//      cases, you have to know how many bytes to read from the DSP at the
//      time you begin writing to it.
//
//    - Reads and Writes to the DSP always consist of an even number of bytes
//
//    - Reads and Writes to the DSP require a control bit be toggled between
//      each pair of bytes transferred
//
//
//  Events/processes that require use of the SPI connection:
//
//  - Writing commands to the ADC's
//
//  - Writing commands to the FPGA
//
//  - Writing commands to and reading data from the DSP (where some of the
//    responses can be up to 1024 bytes in length)
//
//  - Downloading a new config file to the FPGA (or the DSP?)
//
//  - A write to a Modbus register that mirrors a value in the DSP
//
//
//  Two options and issues regarding sharing this single resource that many
//  tasks need to use or affect:
//
//  - Extension of current approach:  Use a semaphore to limit access to a
//    single task at a time.
//
//  - Limit all access to a single task and send msgs to the task requesting
//    certain actions
//
//  The semaphore approach has one major drawback:  Tasks that cannot block
//  (like the main applic process) must either give up on an action or keep the
//  data needed to process the request and retry until successful or an error
//  occurs.  It has the advantage that it means less code for tasks that CAN
//  block without a problem.
//
//  The message approach has the advantage that a task can send a message about
//  the request and pretty much forget about it.  The disadvantages are that
//  their is no easy way to provide feedback to the requesting task, and that
//  all the work has to be done in one task.
//
//
//  Basic problem scenario:
//
//    A command is issued to download a new image to the FPGA or the DSP.
//    This can tie up the SPI for several seconds.  While this is going on,
//    several Modbus writes to registers mirroring DSP values happen.  These
//    writes cause event messages to be written to the main application's
//    event msg queue.  The application loop, which is still running while
//    the download to the FPGA is going on, cannot use the SPI line, so it
//    cannot read values from the DSP to update its state data, much less
//    process the messages to write new values to the DSP.  In order to
//    not loose the write events, it has to stop processing the event
//    messages, which means the queue could overflow, causing processes
//    that write to the queue to block.
//
//  I don't see any way around this.  The SPI line serves several purposes
//  and it is either in use or it is not.  If it is in use, a task that
//  wishes to use it can either block or not.  For processes that cannot
//  (or should not) block, they can either ignore/forget what they were
//  going to do with the SPI connection or keep track of what they were
//  about to do and keep polling periodically, waiting for their chance
//  to use the SPI connection.
//
//  This is really nothing but the basic limitation of any resource that
//  must be protected against concurrent usage.  A locking mechanism is
//  required that insures only one user at a time, and tasks contending for
//  the resource have 3 choices:
//
//  - block on the resource (and get woken up when it is available)
//
//  - poll in a loop, waiting until they get access, possibly doing other
//    useful stuff that is not dependent on the the event that cannot be
//    completed until the locked resource is available
//
//  - forget what they needed it for and continue on
//
//
//  So, when the main application process needs the SPI line, it will either
//  have to block on a wait queue, or simply suspend processing of event
//  messages.  The advantage of the blocking approach is that the order of
//  requests for the SPI line will be enforced by the wait queue.  On the
//  other hand, I might want to grant access based on priority as well, where
//  higher priority tasks get unblocked before lower priority ones.  But as
//  a general rule, ignoring the sequence of requests can lead to other
//  problems.
//
//-----------------------------------------------------------------------------
//
//  More notes on SPI port usage:
//
//  The model I am using for the SPI port is not right.  It should be treated
//  more like a disk drive or a network connection.
//
//  When a serial port is used as a point-to-point connection between 2 devices
//  it often makes sense to allow the application code to directly manage the
//  connection.  But when there are multiple devices connected via the same
//  communications medium, then the connection is a bus, and the connection
//  needs to be handled by a driver that provides a layer of abstraction for
//  communicating with the devices attached to the bus.
//
//  In this case, something more like a device special file per device attached
//  to the SPI port makes more sense, where the major device # is for the SPI
//  port and the minor specifies which device on the "bus".  Then individual
//  architecture modules can deal with them however I want (for devices that
//  a particular board does not have, it can either return errors for attempts
//  to access the device, or I could have a test mode where they mimic the
//  device to some degree, or simply return no errors).
//
//-----------------------------------------------------------------------------
// Keeping Modbus register values (global variables) in sync with the DSP
// settings vs reading waveform data after changing a setting (is this still,
// or was it EVER true?  We read waveform data if/when it ):
//
//  - Ampl and phase values in DSP and ModBus register (global variable) values
//    must be kept in sync
//
//  - When we want to read the waveform data from the DSP to see the effect
//    of changing the ampl and/or phase value, we have to make the settings
//    and read the waveform data as a single transaction
//
//  - Oone possibility is to read the waveform data EVERY time the ampl or
//    phase setpoints are altered.  Then there is no distinction between "just
//    setting" one of them vs setting-and-reading
//    BUT:  The problem with that is it doesn't allow for changing BOTH the
//    the phase and amplitude and reading the result.
//
//-----------------------------------------------------------------------------
//
//  "Pulsed" tuning test mode:
//
//    The DSP normally controls the Amplitude and Phase values using a standard
//    PID feedback-loop algorithm.  To facilitate the tuning of the PID weights
//    used by the feedback loop, the DSP and controller provide a "pulsed"
//    tuning/test mode where the DSP periodically cycles the Amplitude and/or
//    Phase setpoints between two settings and captures a portion of the
//    resulting outputs in a buffer that the controller's CPU can read and
//    send to a remote client for the purpose of providing a near real-time
//    plot of the resulting waveforms.
//
//
//  Basic operation:
//
//    The DSP contains all the logic and code for actually implementing the
//    pulsed test functionality.  It requires the following values to be set
//    before this mode is enabled:
//
//    - Nominal Amplitude and Phase setpoints:  These are the "normal"
//      setpoints, used when the pulsed mode is not active.
//
//    - Pulsed Amplitude and Phase setpoints:  When the pulsed test mode is
//      activated, the DSP toggles the setpoints between the Nominal values
//      and Pulsed setpoints.
//
//    - Pulse Frequency:  Controls the # of times per second that the
//      Amplitude and/or Phase setpoints are toggled between the 2 setpoints
//
//    - Trigger position:  Specifies the # of samples that should be taken
//      before changing the setpoint(s).  This provides a simple means of
//      specifying a trigger delay, or a sliding window around the trigger
//      event.
//
//    The pulsed test mode is activated by setting one or both of the run bits
//    in the DSP's cmd word.  When activated, the DSP samples the outputs and
//    writes the values to one of two buffers.  One buffer is for the sample
//    window around the rising edge, and one is for the window around the
//    falling edge.  When the DSP has filled one of the buffers, it sets a bit
//    in the status word indicating the buffer contains new data to be read.
//    The DSP must then leave the contents of the buffer untouched until it
//    has been read by the controller CPU (see next paragraph).
//
//    The controller CPU periodically reads the status word from the DSP.
//    Whenever it sees that the new-data bit for one of the buffers is set,
//    it reads the data for that buffer from the DSP and then sets a bit in
//    the command word indicating it has read the contents of the buffer.
//    To avoid having the controller CPU attempt to read the same data from
//    the same buffer the next cycle, the DSP must respond by immeadiately
//    resetting the new-data bit in the status word.
//
//    Once the new-data bit has been reset, the DSP is free to write new
//    data to the buffer, repeating the above cycle.
//
//-----------------------------------------------------------------------------


//test
#define  OUTI(reg, value)  \
           outi(reg, reg##Shadow = (value))         //faster
//           WrPortI(reg, &(reg##Shadow), (value))    //slower
//           outi(reg, value)      // fastest, but doesn't update shadow value


#include <common.h>

#include <unistd.h>
#include <sys/types.h>

#include <linux/wait.h>

#include <semaphore.h>



#ifndef  ApplicSource
  #define  ApplicSource  "PhaseCtlrAppl.c"
#endif

#ifndef DidRegDefs
  #define  DidRegDefs
  #include <modbusRegs.c>
#endif
#ifndef RegDefs



//APPLCODE - change to the application name
const char far applName[] = { "Phase Controller" };
const char ApplAbbrev[] = { "PhCtlr" };

//APPLCODE - add ID to unistd.h and place it here
#define DeviceType  DT_PhaseCtlr

//APPLCODE - set the # of input and outputs used
#define  NUM_APPL_AIN     1   // analog inputs
#define  NUM_APPL_AOUT    1   // analog outputs
#define  NUM_APPL_DIN    14   // digital inputs
#define  NUM_APPL_DOUT   24   // digital outputs

#define  ADC_MASK_APPL  0x00000000L
#define  DAC_MASK_APPL  0x00000000L

//APPLCODE - set to state we want outputs in during initialization
#define InitialDigOutState  ( _ADC_CS | FPGA_SPI | _RESET_DSP )

#define  RespOffset  2  // # packets DSP responses are delayed by

#define SharedINC
#include <sharedAppl.c>

#include <math.h>


//--- definitions for enabling diagnostic messages ---
#define  DSPErrorBits      0x80

#define  ShowDSPErrors     (DiagFlags & DSPErrorBits)


/*-------------------------------------------------
 NOTES:
   port B ("spi1")  connected to FPGA, PLL, ADC's, and DSP
   port C ("spi2")  connected to EEPROM
---------------------------------------------------*/

//--- multipliers/divisors for converting between unsigned 16-bit int values ---
//--- used by the DSP and engineering units (volts and degress in this case) ---
#define  AmplConvFactor   ( 0.0000426233445431 )
//#define  PhaseConvFactor  ( 0.0054931640625 )  // 360 / 2^16
#define  PhaseConvFactor  ( 0.0000000838190317 )  // ~ 360 / 2^32
#define  FRACU16Divisor   ( 0.0000152590219 )  // div by this = mult by 65535.0
//#define  FreqConvFactor   (    0.0152590219 )  // scales 0-1000 -> 0-65535


#define  DSP_Prog_File   "dspprog.bin"
#define  FPGA_Prog_File  "/srfs/llrf.bin"



//--- instruction codes for ATMEL SPI Serial EEPROMs ---
#define  EEPROM_WRSR       0x01   // WRite Status Register
#define  EEPROM_WRITE      0x02   // Write a page to memory
#define  EEPROM_READ       0x03   // Read bytes from memory
#define  EEPROM_WRDI       0x04   // WRite DIsable latch
#define  EEPROM_RDSR       0x05   // ReaD Status Register
#define  EEPROM_WREN       0x06   // WRite ENable
#define  EEPROM_PROT       0x36   // PROTect sector
#define  EEPROM_UNPR       0x39   // UNPRotect sector
#define  EEPROM_ERASE_ALL  0x60   // Erase entire chip
#define  EEPROM_RDID       0x9F   // Read device ID
#define  EEPROM_SEQ_PROG   0xAF   // Write a sequence of bytes

#define  EEPROM_Enable()   \
    OUTI(PFDR, PFDRShadow & 0xFE)  // Chip Select (PF0) low
#define  EEPROM_Disable()   \
    OUTI(PFDR, PFDRShadow | 0x01)  // Chip Select (PF0) high



#define  ReadDig()  { digitalIn = ReadDigInputs(); }
#define  WriteDig()  {  WriteDigOutputs(digitalOut); }


#define  WriteSPI1(buf, len)    _spiWrite_(spi1Port, buf, len, false)
//#define  WriteSPI1(buf, len)   write(spi1fd, buf, len)
#define  ReadSPI1(buf, len)     _spiRead_(spi1Port, buf, len, false)
//#define  ReadSPI12(buf, len)    read(spi1fd, buf, len)

#define  WriteSPI2(buf, len)    _spiWrite_(spi2Port, buf, len, false)
//#define  WriteSPI2(buf, len)   write(spi2fd, buf, len)
#define  ReadSPI2(buf, len)     _spiRead_(spi2Port, buf, len, false)
//#define  ReadSPI2(buf, len)    read(spi2fd, buf, len)


#define  PI_  3.1415926

//-----------------------------------------------------------------------------
// data types and function prototypes:
//-----------------------------------------------------------------------------

long fltofi(float value);
float fitofl(long value);


int initDSP(fd_t fromfd, uint progNum, ulong ttlBytes, bool showProgress);

int updateFPGAfirmware(fd_t fromfd, bool showProgress, bool verify);

//int writeFPGAParallel(uchar id, uchar val);

int write_ADCReg(uint regNum, uint regVal);

int write_Config(uint newConfig);


int readPlot(ulong regNum, ulong far *rcvBuf, uint rcvCount);

int writeDSPRegQuick(ulong regNum, ulong regVal);
int writeDSPReg(ulong regNum, ulong regVal, bool showErrs);
int readDSPReg(ulong regNum, ulong *regVal);

int read_DSP_Regs(void);
int write_DSP_Regs(void);


void prepForReset(void);
void resetDSP(void);


bool MB_DSP_REG(SetRegFuncParams);


int  getLock(uint id, ulong msTimeout);
int  releaseLock(uint id);

typedef struct  {
  int  count;
  PID  owner;
//  WaitQueue  tasks;
  sem_t  sem;
}  LockData;

enum LockIDS { APPL_LOCK,
               WAVE_DATA_0, WAVE_DATA_1, WAVE_DATA_2, WAVE_DATA_3,
               WAVE_DATA_4, WAVE_DATA_5, WAVE_DATA_6, WAVE_DATA_7,
               WAVE_DATA_8, WAVE_DATA_9,
               NUM_LOCKS };


#define  PLLInitSeqLen  40

#define  PLLInitSeqFile  "pllInitSeqs"

#define  NumFixedPLLSeqs  3

const char pllInitSeqs[NumFixedPLLSeqs][PLLInitSeqLen] = {
  {  // 80.5 MHz
    0x80,0x00,0x00,0x00,
    0x00,0x03,0x08,0x04,
    0x00,0x03,0x0a,0x05,
    0x00,0x03,0x03,0x06,
    0x00,0x03,0x03,0x07,
    0x10,0x00,0x09,0x08,
    0x00,0x82,0x00,0x0b,
    0x02,0x94,0x00,0x0d,
    0x08,0x30,0xa1,0x0e,
    0xc8,0x06,0x40,0x0f
  },
  {  // 161 MHz
    0x80,0x00,0x00,0x00,
    0x00,0x03,0x08,0x04,
    0x00,0x03,0x0a,0x05,
    0x00,0x03,0x03,0x06,
    0x00,0x03,0x03,0x07,
    0x10,0x00,0x09,0x08,
    0x00,0x82,0x00,0x0b,
    0x02,0xa8,0x40,0x0d,
    0x08,0x30,0xa1,0x0e,
    0xc8,0x03,0x20,0x0f
  },
  {  // 241.5 MHz
    0x80,0x00,0x00,0x00,
    0x00,0x03,0x08,0x04,
    0x00,0x03,0x0a,0x05,
    0x00,0x03,0x03,0x06,
    0x00,0x03,0x03,0x07,
    0x10,0x00,0x09,0x08,
    0x00,0x82,0x00,0x0b,
    0x02,0xb2,0x00,0x0d,
    0x08,0x31,0xe3,0x0e,
    0xc8,0x06,0x40,0x0f
  }
};


//=============================================================================
//  definitions related to use of the Digital and Analog I/O signals
//=============================================================================

//======================= digitalIn usage ===========================
#define  ID0          0x000001  // DI-00: PA0:  ID bit 0
#define  ID1          0x000002  // DI-01: PA1:  ID bit 1
#define  ID2          0x000004  // DI-02: PA2:  ID bit 2
#define  ID3          0x000008  // DI-03: PA3:  ID bit 3
//-----------------------------
#define  ID4          0x000010  // DI-04: PA4:  ID bit 4
#define  ID5          0x000020  // DI-05: PA5:  ID bit 5
#define  ID6          0x000040  // DI-06: PA6:  ID bit 6
#define  ID7          0x000080  // DI-07: PA7:  ID bit 7
//-----------------------------
#define  DSP_GP1      0x000100  // DI-08: PD4:
#define  DSP_GP2      0x000200  // DI-09: PD5:
#define  unused_i10   0x000400  // DI-10: PE3:
#define  FPGA_DONE    0x000800  // DI-11: PE4:
//-----------------------------
#define  FPGA_CL2     0x001000  // DI-12: PF3:  ACK from FPGA (i.e. last param value latched) (unused now?)
#define  unused_i13   0x002000  // DI-13: PF4:  DSP is in Reset state
//-----------------------------

const uchar  LogicOrderDIn[NUM_APPL_DIN] = {
  0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13 };

  //--- msgs to log automatically when a digital input changes state ---
//--- use NULL for bits that we don't want to log changes for ---
const char far *digInDescr[NUM_APPL_DIN*2] = {
//--0 state descr--    --1 state descr--
  NULL,                NULL,
  NULL,                NULL,
  NULL,                NULL,
  NULL,                NULL,
  NULL,                NULL,
  NULL,                NULL,
  NULL,                NULL,
  NULL,                NULL,
  NULL,                NULL,
  NULL,                NULL,
  NULL,                NULL,
  NULL,                NULL,
  NULL,                NULL,
  NULL,                NULL
};



//====================== digitalOut usage ===========================

//------------------------------
#define  DSP_GP5      0x000001L  // DO-00: PG0:
#define  DSP_GP6      0x000002L  // DO-01: PG1:
#define  DSP_GP7      0x000004L  // DO-02: PG2:
#define  DSP_GP8      0x000008L  // DO-03: PG3:
//------------------------------
#define  _DSP_WR_RD   0x000010L  // DO-04: PG4:  DSP SPI enable
#define  DSP_GP9      0x000020L  // DO-05: PG5:
#define  DSP_GP10     0x000040L  // DO-06: PG6:
#define  DSP_GP11     0x000080L  // DO-07: PG7:
//------------------------------
#define  _RESET_DSP   0x000100L  // DO-08: PB2:  0=Reset DSP, 1=Run
#define  ActivityLED  0x000200L  // DO-09: PB3:
#define  READY_LED    0x000400L  // DO-10: PB4:  1 turns on CPU Ready LED
#define  ON_OFF_LED   0x000800L  // DO-11: PB5:  1 turns on RF ON LED
//------------------------------
#define  DSP_NMI      0x001000L  // DO-12: PB6:
#define  unused_o13   0x002000L  // DO-13: PB7:
#define  FPGA_SPI     0x004000L  // DO-14: PE1:  low-to-high causes FPGA reload
#define  _ADC_CS      0x008000L  // DO-15: PE6:  ADC /CS (Chip Select)
//------------------------------
#define  PLL_LE       0x010000L  // DO-16: PE7:  PLL Load Enable
#define  _EEPROM_CS   0x020000L  // DO-17: PF0   /CS on EEPROM
#define  none_o18     0x040000L  // DO-18: ---   (ignored)
#define  FPGA_CL1     0x080000L  // DO-19: PF2:  (now unused?)
//------------------------------
#define  none_o20     0x100000L  // DO-20: ---   (ignored)
#define  unused_o21   0x200000L  // DO-21: PF5:
#define  DSP_GP3      0x400000L  // DO-22: PF6:  LSB used to select DSP prog # to run (?)
#define  DSP_GP4      0x800000L  // DO-23: PF7:  MSB used to select DSP prog # to run (?)
//------------------------------
const uchar  LogicOrderDOut[NUM_APPL_DOUT] = {
  0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12,
  13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23
};

#define  TurnOnActivityLED()    ClearBits(Next.digitalOut, ActivityLED)
#define  TurnOffActivityLED()   SetBits(Next.digitalOut, ActivityLED)

//--- msgs to log automatically when a digital output changes state ---
//--- use NULL for bits that we don't want to log changes for ---
const char far *digOutDescr[NUM_APPL_DOUT*2] = {
//--0 state descr--    --1 state descr--
  NULL,                NULL,
  NULL,                NULL,
  NULL,                NULL,
  NULL,                NULL,
  NULL,                NULL,
  NULL,                NULL,
  NULL,                NULL,
  NULL,                NULL,
  NULL,                NULL,
  NULL,                NULL,
  NULL,                NULL,
  NULL,                NULL,
  NULL,                NULL,
  NULL,                NULL,
  NULL,                NULL,
  NULL,                NULL,
  NULL,                NULL,
  NULL,                NULL,
  NULL,                NULL,
  NULL,                NULL,
  NULL,                NULL,
  NULL,                NULL,
  NULL,                NULL,
  NULL,                NULL
};


//=============== Analog input usage ===============
const uchar  LogicOrderAIn[NUM_APPL_AIN] = { 0 };


//=============== Analog output usage ===============
const uchar  LogicOrderAOut[NUM_APPL_AOUT] = { 0 };



//=============================================================================
//  ID's for values sent to or read from the DSP
//  OR'ing an ID with 0x8000 allows reading without writing
//=============================================================================

//--- DSP register numbers (without the 0x01000000 mask shared by all regs) ---


// Even # values used to read a register, Odd # values used to write to a reg

// P32 means the "PHASE" format, a 32-bit integer scaled to/from 0 to 360

//===== Read-Only registers =====
#define  FwdPhase         0x00100000L  // P32: Forward Phase
#define  ReflPhase        0x00100002L  // P32: Reflected Phase
#define  CavPhase         0x00100004L  // P32: Cavity Phase
#define  RefPhase         0x00100006L  // P32: Reference Phase

#define  FwdAmpl          0x00100008L  // F32: Forward Amplitude
#define  ReflAmpl         0x0010000AL  // F32: Reflected Amplitude
#define  CavAmpl          0x0010000CL  // F32: Cavity Amplitude
#define  RefAmpl          0x0010000EL  // F32: Reference Amplitude

#define  MaxFwdAmpl       0x00100010L  // F32: Max Forward Amplitude
#define  MaxReflAmpl      0x00100012L  // F32: Max Reflected Amplitude
#define  MaxCavAmpl       0x00100014L  // F32: Max Cavity Amplitude
#define  MaxOutAmpl       0x00100016L  // F32: Max Output Amplitude

#define  DSP_Status       0x00100018L  // Bits: misc DSP status
#define  AFC_State        0x0010001AL  // Bits: AFC state bits

#define  DSP_Load         0x0010001CL  // F32: DSP load percentage
#define  PCB_Temp         0x0010001EL  // F32: PCB temperature


//===== R/W registers (the addresses shown are the read addresses) =====
#define  TriggerPos       0x00100100L  // U32: trigger position (tuning plot)

#define  WarmUpAmpl       0x00100102L  // F32: Warm-up Amplitude
#define  WarmUpTime       0x00100104L  // F32: time at warm-up Amplitude
#define  RampTime         0x00100106L  // F32: time to ramp to set-point

#define  FwdAtten         0x00100108L  // U32: Forward Attenuator setting
#define  ReflAtten        0x0010010AL  // U32: Reflected Attenuator setting
#define  CavAtten         0x0010010CL  // U32: Cavity Attenuator setting

#define  PhaseSet         0x0010010EL  // P32: Phase setpoint
#define  AmplSet          0x00100110L  // F32: Ammplitude setpoint

#define  TunePulseFreq    0x00100112L  // F32: Tuning Pulse Freq
#define  FreqSearchAmpl   0x00100114L  // F32: Freq Search Amplitude setpoint

#define  RF_P             0x00100116L  // F32: RF P value
#define  RF_I             0x00100118L  // F32: RF I value
#define  RF_D             0x0010011AL  // F32: RF D value

#define  Tuner_P          0x0010011CL  // F32: Tuner P value
#define  Tuner_I          0x0010011EL  // F32: Tuner I value
#define  Tuner_D          0x00100120L  // F32: Tuner D value

#define  PulsedAmplSet    0x00100122L  // F32: Pulsed Amplitude setpoint
#define  PulsedPhaseSet   0x00100124L  // P32: Pulsed Phase setpoint

#define  AFC_Freq0        0x00100126L  // F32: AFC cancellation frequency 0
#define  AFC_Freq1        0x00100128L  // F32: AFC cancellation frequency 1
#define  AFC_Freq2        0x0010012AL  // F32: AFC cancellation frequency 2
#define  AFC_Freq3        0x0010012CL  // F32: AFC cancellation frequency 3
#define  AFC_Freq4        0x0010012EL  // F32: AFC cancellation frequency 4
#define  AFC_Freq5        0x00100130L  // F32: AFC cancellation frequency 5
#define  AFC_Freq6        0x00100132L  // F32: AFC cancellation frequency 6
#define  AFC_Freq7        0x00100134L  // F32: AFC cancellation frequency 7
#define  AFC_Freq8        0x00100136L  // F32: AFC cancellation frequency 8
#define  AFC_Freq9        0x00100138L  // F32: AFC cancellation frequency 9

#define  CTFTunerAmpl     0x0010013AL  // F32: Trans Func Tuner Amplitude

#define  CavLowIlkLevel   0x0010013CL  // F32: Cavity Low Interlock Level
#define  CavLowIlkTime    0x0010013EL  // F32: Cavity Log Interlock Time
#define  DvDt_IlkRate     0x00100140L  // F32: -dV/dT Interlock Rate
#define  VV_IlkRatio      0x00100142L  // F32: V+/V- Interlock Ratio

#define  FwdCavOffset     0x00100144L  // U32: Forward Cavity Offset
#define  FeedFwdPhase     0x00100146L  // P32: Feed Forward Phase
#define  FeedFwdGain      0x00100148L  // F32: Feed Fwd Gain

#define  AFC_Phase0       0x00100150L  // P32: AFC cancellation phase 0
#define  AFC_Phase1       0x00100152L  // P32: AFC cancellation phase 1
#define  AFC_Phase2       0x00100154L  // P32: AFC cancellation phase 2
#define  AFC_Phase3       0x00100156L  // P32: AFC cancellation phase 3
#define  AFC_Phase4       0x00100158L  // P32: AFC cancellation phase 4
#define  AFC_Phase5       0x0010015AL  // P32: AFC cancellation phase 5
#define  AFC_Phase6       0x0010015CL  // P32: AFC cancellation phase 6
#define  AFC_Phase7       0x0010015EL  // P32: AFC cancellation phase 7
#define  AFC_Phase8       0x00100160L  // P32: AFC cancellation phase 8
#define  AFC_Phase9       0x00100162L  // P32: AFC cancellation phase 9


//===== Registers for ADRC =====
#define  OBW_Ampl         0x00100164L  // (RW) F32: Observer Bandwidth
#define  CBW_Ampl         0x00100166L  // (RW) F32: Controller Bandwidth
#define  P_Gain_Ampl      0x00100168L  // (RO) F32: Proportional Gain
#define  D_Gain_Ampl      0x0010016AL  // (RO) F32: Derivative Gain
#define  X1_Ampl          0x0010016CL  // (RO) F32: x1 state value (estimated state)
#define  X2_Ampl          0x0010016EL  // (RO) F32: x2 state value (estimated state)
#define  X3_Ampl          0x00100170L  // (RO) F32: x3 state value (estimated state)

#define  OBW_Phase        0x00100172L  // (RW) F32: Observer Bandwidth
#define  CBW_Phase        0x00100174L  // (RW) F32: Controller Bandwidth
#define  P_Gain_Phase     0x00100176L  // (RO) F32: Proportional Gain
#define  D_Gain_Phase     0x00100178L  // (RO) F32: Derivative Gain
#define  X1_Phase         0x0010017AL  // (RO) F32: x1 state value (estimated state)
#define  X2_Phase         0x0010017CL  // (RO) F32: x2 state value (estimated state)
#define  X3_Phase         0x0010017EL  // (RO) F32: x3 state value (estimated state)

#define  DSP_Smpl_Rate    0x00100180L  // (RO) F32: DSP sample reate (Hz)
#define  DSP_Smpl_Period  0x00100182L  // (RO) F32: DSP sample period (secs)

#define  Beta_Ampl        0x00100184L  // (RO) F32: beta
#define  LC1_Ampl         0x00100186L  // (RO) F32: LC1
#define  LC2_Ampl         0x00100188L  // (RO) F32: LC2
#define  LC3_Ampl         0x0010018AL  // (RO) F32: LC3

#define  Beta_Phase       0x0010018CL  // (RO) F32: beta
#define  LC1_Phase        0x0010018EL  // (RO) F32: LC1
#define  LC2_Phase        0x00100190L  // (RO) F32: LC2
#define  LC3_Phase        0x00100192L  // (RO) F32: LC3

#define  B_param_Ampl     0x00100194L  // (RW) F32: B parameter (ampl ctl)
#define  B_param_Phase    0x00100196L  // (RW) F32: B parameter (phase ctl)


//===== RFQ Tuner values, some of which are forwarded by the control =====
//===== system to other devices/records =====
#define  RFQ_ManTuneSet   0x00100198L  // (RW) F32: manual tuner setpoint

#define  RFQ_T1_Offset    0x0010019AL  // (RW) F32: tuner 1 calib offset
#define  RFQ_T2_Offset    0x0010019CL  // (RW) F32: tuner 2 calib offset

#define  RFQ_T1_Output    0x0010019EL  // (RO) F32: tuner 1 output
#define  RFQ_T2_Output    0x001001A0L  // (RO) F32: tuner 2 output

#define  RFQ_T1_Feedback  0x001001A2L  // (RW) F32: tuner 1 feedback
#define  RFQ_T2_Feedback  0x001001A4L  // (RW) F32: tuner 2 feedback

#define  Pulse_Ramp       0x001001ABL  // (RW) U32: Pulse Ramp (rate?)

//===== More R/W registers =====
#define  FwdAmplGain      0x001001ADL  // (RW) F32: Forward Amplitude Gain
#define  CavAmplGain      0x001001AFL  // (RW) F32: Cavity Amplitude Gain

#define  CavHighIlkLevel  0x001001B5L  // (RW) F32: Cavity High Interlock Level
#define  FwdHighIlkLevel  0x001001B7L  // (RW) F32: Forward High Interlock Level


//===== Special Registers =====
#define  SetDSPState      0x00100200L  // (WO) Bits: set/reset DSP state bits

#define  PiezoTunerPos    0x00100202L  // (RW) U32: piezo tuner position
#define  StepperPos       0x00100204L  // (RW) U32: stepper position

#define  SetAFCsOnOff     0x00100206L  // (WO) Bits: turn AFC frequencies on/off

#define  Unused_208       0x00100208L  //

#define  Rd_L_Buf         0x0010020AL  // (RO) I16x2: read next tuning plot value (rising edge)
#define  Rd_T_Buf         0x0010020CL  // (RO) I16x2: read next tuning plot value (falling edge)

#define  Reserved_20E     0x0010020EL  //
#define  Reserved_210     0x00100210L  //

#define  WrMemBlock       0x00100212L  // (WO) U32: start prog mode for a memory block
#define  WriteStatus      0x00100214L  // (RO) I32: (% done or -1 ==> busy)

#define  CalcCTF          0x00100216L  // (**) N/A: begin calc of Cavity Transfer Func
#define  PercentCTF       0x00100218L  // (RO) U32: percent done for CTF calculation
#define  AbortCTF         0x0010021AL  // (**) N/A: abort CTF calc; == -1 if too late
#define  SizeCTF          0x0010021CL  // (RO) U32: # of 32-bit words int plot data
#define  PhaseCTF         0x0010021EL  // (RO) P32: read next Phase value from CTF plot
#define  AmplCTF          0x00100220L  // (RO) F32: read next Ampl value from CTF plot

#define  PercentFFT       0x00100222L  // (RO) U32: percent done for FFT calculation
#define  SizeFFT          0x00100224L  // (RO) U32: # of 32-bit words of in plot data
#define  AmplFFT          0x00100226L  // (RO) F32: read next Ampl value from FFT plot

#define  CTFCalcRate      0x00100228L  // (RW) F32: CTF calc rate

#define  FwdMagCalib      0x0010022AL  // (RW) F32: Forward Mag calib value
#define  FwdPhaseCalib    0x0010022CL  // (RW) P32: Forward Phase calib value
#define  ReflMagCalib     0x0010022EL  // (RW) F32: Reflected Mag calib value
#define  ReflPhaseCalib   0x00100230L  // (RW) P32: Reflected Phase calib value
#define  CavMagCalib      0x00100232L  // (RW) F32: Cavity Mag calib value
#define  CavPhaseCalib    0x00100234L  // (RW) P32: Cavity Phase calib value
#define  OutMagCalib      0x00100236L  // (RW) F32: Output Mag calib value
#define  OutPhaseCalib    0x00100238L  // (RW) P32: Output Phase calib value

#define  SaveCalibToDSP   0x0010023AL  // (**) N/A: Save calib to DSP's ROM

#define  ROMProgErrs      0x0010023CL  // (RO) U32: ROM programming errors


//  (**) means a READ from the register triggers the stated functionality


//=============================================================================
//  List of daemon processes.  The inetd process listens for connection
//  requests on each port in the list and creates a new instance of the
//  specified process to service each new connection.
//=============================================================================
#include <services.h>

//--- service funcs declared the same as standard command funcs:
//     e.g.:   void _task fastdatad(int argc, char **argv);
void _task tsh(int argc, char **argv);

void _task readWave(int argc, char **argv);


ServiceInfo inetdConf[] = {
//--- add any application-specific inetd-type network services here ---
//--- (add port # names to services.h) ---
//  NAME          PORT           PROTOCOL  FUNC        PRI -1  Config Params
//  ------------  -------------  --------  ----------- --- --  --------------
  { "readWave",   READ_WAVE,     ProtoTCP, readWave,   12, -1, NULL },

//--- standard ones defined in sharedAppl.c ---
//{ "console",    TTY0,          ASYNC,    tsh,        10, -1, &consCfg },
//{ "fastdatad",  FASTDATA_PORT, ProtoTCP, fastdatad,   5, -1, NULL },
//{ "modbusTCPd", MODBUS_PORT,   ProtoTCP, modbusTCPd,  6, -1, NULL },
//{ "tsh",        TELNET_PORT,   ProtoTCP, tsh,        10, -1, NULL }

  #define SharedServices
  #include <sharedAppl.c>
};

const int  NumServices = sizeof(inetdConf) / sizeof(ServiceInfo);


//=============================================================================
//  Application-specific data structures and definitions
//=============================================================================

// values that can be different between prev/cur or cur/next during processing


// ***** WARNING ***** WARNING ***** WARNING ***** WARNING *****
//  Changing the applState structure means that a soft reset will fail!  A
//  change to the size will cause a soft restart to abort and do a CPU
//  restart instead.  And while a change that does NOT result in a new
//  size will not cause the soft restart to abort, it does mean that any
//  shift in the position of values can cause all kinds of mayhem.
// ***** WARNING ***** WARNING ***** WARNING ***** WARNING *****


typedef struct {
//--- values updated in "cur" instance at start of each pass ---
  ulong  digitalIn;       // state of digital inputs

//--- values updated in "next" instance at start of each pass ---
  ulong  digitalOut;      // state for digital outputs
} ApplState;


ApplState  prevState, curState, nextState;

#define  Prev  prevState
#define  Cur   curState
#define  Next  nextState

const uint ApplStateDataSize =  ( sizeof(ApplState)*3 + sizeof(setVal) );


//--- values and flags set by the 1st stage of the control loop when ---
//--- interpreting the state of the inputs and processing the next ---
//--- event msg ---

ulong  calcStartTimeCTF;


//=============================================================================
//  Defintions for variables that are accessable as ModBus registers
//=============================================================================
#else  // end of before-RegDefs portion

/*
  The purpose of this section is to define a set of C variables that are
  accessable as ModBus registers.  This approach avoids the need to use
  special functions or macros throughout the code when accessing register
  values, which keeps the application code fast and simple.

  The model used provides support for multiple logical devices.  The 2nd
  parameter in each register definition macro indicates if one instanace
  applies to all logical devices (specified by an "x" for "don't care"), or
  if each logical device requires its own instance (specified by an under-
  score "_" to signify "fill in the blank").  In the 1st case, the variable
  is defined as a scalar.  In the 2nd case, it is defined as an array.

  The definition macros support data types (floats and longs in particular)
  that require more than one register to store them.  To avoid the possibility
  of corrupted values, all reads of and writes to such values must be done
  using the multi-register commands and including all the registers that
  make up the value in the command.  Differences in register order between
  different client types is handled by a table that automatically swaps
  register values as needed based on the client type.

  The definition for each writable register also includes the function to
  call when the value is set via the ModBus interface.  Many registers can
  use the default setXXX() functions, but you can specify your own to provide
  any special logic needed (including ignoring the new setting if invalid).

  For registers, the format for the function definitions is:

    bool mySetFunc(SetRegFuncParams);

       where SetRegFuncParams is currently defined as:
              void *info, uchar chanNum, void *newVal, fd_t eventsfd

  For coils, the format for the function definitions is:

     bool (*SetCoilFunc)(SetCoilFuncParams);

         where SetCoilFuncParams is currently define as:
               void *info, uchar chanNum, uchar newVal, fd_t eventsfd

  The parameters to the call are as follows:
    *info:     A pointer to the Coil's/Register's CoilInfo/RegInfo structure
    chanNum:   The logical device # (0 is the 1st)
    *newVal:   A pointer to the new value requested in the Modbus packet
               (changed to native-byte order where needed)
    eventsfd:  The fd for the FIFO used to send msgs to the main application
               process

  The function returns true if the new value was accepted, false if not.

  The register numbers for multiple "channels" or logical devices are
  handled automatically based on the xxChanMult values.  These values specify
  the offset between a register value for one logical device and the "same"
  register for the next.

  Example:

    Assume the following definitions:
      #define  NumLogicalDevices  2
      #define  R4ChanMult  100
      REG4( motorPos,       _, 40050, F32, RW, setMotorPos )

    Then we would get:
      global variable definition:  float motorPos[2];
      register pair:  40050,40051  refers to:  motorPos[0]
      register pair:  40150.40151  refers to:  motorPos[1]

    If instead we had specified:
      REG4( motorPos,       x, 40050, F32, RW, setMotorPos )

    Then we would get:
      global variable definition:  float motorPos;
      register pairs:  40050,40051 & 40150,40151  BOTH refer to:  motorPos


  To keep the definitions as simple as possible, and avoid requiring multiple
  macros for each register definition, each macro is used more than once,
  with a different definition (#define) each time (a practice generally
  referred to as X-Macros).  While trying to follow the order of inclusion
  of files (including a recursive reference back to this file) can be rather
  confusing, the X-Macro method keeps the source code simple to read and
  maintain by doing everything with a single macro for each register.

  ==========================================================================
                             !!! WARNING !!!
  ==========================================================================
    Each sequence of register definitions MUST be contiguous!  There must
    NOT be any holes (i.e. no unused register #'s between used ones) and no
    overlap!
-----------------------------------------------------------------------------*/


#ifndef NumLogicalDevices
  #define  NumLogicalDevices  1
#endif

//-----------------------------------------------------------------------------
#ifndef NumC0Chans
  #define NumC0Chans_  NumLogicalDevices
  #define C0ChanMult_  100
#endif

#ifdef COIL0

COIL0( DoNothing,       x, 1, setCoil )
//COIL0( Coil00002,      _, 2, setCoil )
//COIL0( Coil00003,      _, 3, setCoil )

#undef COIL0
#endif


//-----------------------------------------------------------------------------
#ifndef NumR3Chans
  #define NumR3Chans_  NumLogicalDevices
  #define R3ChanMult_  100
#endif

#ifdef REG3

REG3( dlmVersTime,     x, 30001, U32 )
REG3( rtosVersTime,    x, 30003, U32 )
REG3( applVersTime,    x, 30005, U32 )
REG3( devType_,        x, 30007, U16 )
REG3( deviceID_,       x, 30008, U32 )
REG3( upSecs,          x, 30010, U32 )
REG3( percentIdle_,    x, 30012, U16 )

REG3( numChans,        x, 30013, U16 )
REG3( coilChanMult,    x, 30014, U16 )
REG3( reg3ChanMult,    x, 30015, U16 )
REG3( reg4ChanMult,    x, 30016, U16 )

REG3( digInBits,       x, 30017, U32 )

REG3( statusBitsStdA,  x, 30019, U16 )
REG3( statusBitsStdB,  x, 30020, U16 )
REG3( statusBitsAppl,  x, 30021, U32 )

REG3( fwdAmpl,         x, 30023, F32 )  // forward amplitude
REG3( reflAmpl,        x, 30025, F32 )  // reflected amplitude
REG3( cavAmpl,         x, 30027, F32 )  // cavity amplitude
REG3( refAmpl,         x, 30029, F32 )  // reference amplitude

REG3( fwdPhase,        x, 30031, F32 )  // forward phase
REG3( reflPhase,       x, 30033, F32 )  // reflected phase
REG3( cavPhase,        x, 30035, F32 )  // cavity phase
REG3( refPhase,        x, 30037, F32 )  // reference phase

REG3( dspLoad,         x, 30039, U16 )
REG3( dspTempInt,      x, 30040, U16 )

REG3( dspTemp,         x, 30041, F32 )
REG3( unused_30043,    x, 30043, F32 )  // changed to a R/W register
REG3( unused_30045,    x, 30045, F32 )  // changed to a R/W register
REG3( unused_30047,    x, 30047, F32 )  // changed to a R/W register
REG3( unused_30049,    x, 30049, F32 )  // changed to a R/W register
REG3( unused_30051,    x, 30051, F32 )  // changed to a R/W register
REG3( unused_30053,    x, 30053, F32 )  // changed to a R/W register
REG3( unused_30055,    x, 30055, F32 )  // changed to a R/W register
REG3( unused_30057,    x, 30057, F32 )  // changed to a R/W register
REG3( unused_30059,    x, 30059, F32 )  // changed to a R/W register

REG3( percentCTF,      x, 30061, U16 )
REG3( percentFFT,      x, 30062, U16 )

REG3( latency,         x, 30063, U16 )  // actual vs scheduled appl start in ms
REG3( tcpTickTime,     x, 30064, U16 )  // # ms inside tcp_tick() call
REG3( applLoopTime,    x, 30065, U16 )  // # ms used by applic task loop

REG3( maxFwdAmpl,      x, 30066, F32 )  // max Forward Amplitude
REG3( maxReflAmpl,     x, 30068, F32 )  // max Reflected Amplitude
REG3( maxCavAmpl,      x, 30070, F32 )  // max Cavity Amplitude
REG3( maxOutAmpl,      x, 30072, F32 )  // max Output Amplitude

REG3( statusBitsDSP,   x, 30074, U32 )  // DSP status
REG3( statusBitsAFC,   x, 30076, U16 )  // state of each AFC channel

REG3( writeStatus,     x, 30077, U16 )  // % of prog buffer saved to DSP mem
REG3( sizeCTF,         x, 30078, U16 )  // # 32-bit words in CTF plot
REG3( sizeFFT,         x, 30079, U16 )  // # 32-bit words in FFT plot
REG3( romProgErrs,     x, 30080, U16 )  // # programming errors

REG3( piezoReadPos,    x, 30081, S16 )  // current piezo pos read from DSP
REG3( unused_30082,    x, 30082, S16 )  // older (undersize) stepperReadPos

REG3( kernelLoopTime,  x, 30083, U16 )  // # ms used by last kernel pass

REG3( stepperReadPos,  x, 30084, S32 )  // current stepper pos read from DSP

#undef REG3
#endif


#ifdef RegMacros
//===== application-specific defintions related to register values =====

//  #define  READING(idx)  REG_ARRAY(float, rdg_1, idx)

#endif

//--- values for older statusBitsAppl ---
#define  OLD_Phase_PID_ON    0x00000001  // 1 = Closed loop, 0 = Open loop
#define  OLD_Ampl_PID_ON     0x00000002  // 1 = Closed loop, 0 = Open loop
#define  OLD_AFC_ON          0x00000004  // 1 = enable 0-9 On/Offs, 0 = disable
#define  OLD_RF_ON           0x00000008  // 1 = RF On, 0 = RF Off

#define  OLD_PHASE_PULSES    0x00000010  // 1 = pulsing, 0 = normal
#define  OLD_AMPL_PULSES     0x00000020  // 1 = pulsing, 0 = normal
#define  OLD_NEW_L_DATA      0x00000040  // new data in Leading Edge sample buf
#define  OLD_NEW_T_DATA      0x00000080  // new data in Trailing Edge sample buf

#define  OLD_PROG_ID0        0000000100  // LSB of prog # currently running
#define  OLD_PROG_ID1        0000000200  // MSB of prog # currently running
#define  OLD_CALC_XFER_FUNC  0x00000400  // Calculating xfer func
#define  OLD_SAVG_XFER_FUNC  0x00000800  // Saving xfer func to ROM

#define  OLD_SAVING_PROG     0x00001000  // Saving program to ROM
#define  OLD_FILL_FFT_BUF    0x00002000  // Filling buffer for FFT
#define  OLD_COMPUTING_FFT   0x00004000  // FFT being computed
#define  OLD_FFT_DATA_VALID  0x00008000  // FFT buffer has valid data

#define  OLD_AFC_0_ON        0x00010000L  // AFC freq 0 on
#define  OLD_AFC_1_ON        0x00020000L  // AFC freq 1 on
#define  OLD_AFC_2_ON        0x00040000L  // AFC freq 2 on
#define  OLD_AFC_3_ON        0x00080000L  // AFC freq 3 on

#define  OLD_AFC_4_ON        0x00100000L  // AFC freq 4 on
#define  OLD_AFC_5_ON        0x00200000L  // AFC freq 5 on
#define  OLD_AFC_6_ON        0x00400000L  // AFC freq 6 on
#define  OLD_AFC_7_ON        0x00800000L  // AFC freq 7 on

#define  OLD_AFC_8_ON        0x01000000L  // AFC freq 8 on
#define  OLD_AFC_9_ON        0x02000000L  // AFC freq 9 on



//----- values for statusBitsDSP --- (DSP reg 0x00100018)
#define  DSP_ROM0_BUSY       0x00000001  // 1 = program memory ROM busy
#define  DSP_ROM1_BUSY       0x00000002  // 1 = lookup table/cal ROM busy
#define  DSP_UNUSED_2        0x00000004  //
#define  DSP_RF_ON           0x00000008  // 1 = RF On, 0 = RF Off

#define  DSP_RF_PULSES       0x00000010  // RF phase 1 = pulsing, 0 = normal
#define  DSP_TUNER_PID_ON    0x00000020  // 1 = closed loop, 0 = open loop
#define  DSP_NEW_L_DATA      0x00000040  // new data in Leading Edge sample buf
#define  DSP_NEW_T_DATA      0x00000080  // new data in Trailing Edge sample buf

#define  DSP_PROG_ID0        0000000100  // LSB of prog # currently running
#define  DSP_PROG_ID1        0000000200  // MSB of prog # currently running
#define  DSP_CALC_XFER_FUNC  0x00000400  // Calculating xfer func
#define  DSP_SAVG_XFER_FUNC  0x00000800  // Saving xfer func to ROM

#define  DSP_SAVING_PROG     0x00001000  // Saving program to ROM
#define  DSP_FILL_FFT_BUF    0x00002000  // Filling buffer for FFT
#define  DSP_COMPUTING_FFT   0x00004000  // FFT being computed
#define  DSP_FFT_DATA_VALID  0x00008000  // FFT buffer has valid data

#define  DSP_RF_OUT_WARM     0x00010000L // 1 = RF output at warm level
#define  DSP_OUT_RAMPING     0x00020000L // 1 = RF ouput ramping
#define  DSP_AFC_ON          0x00040000L // 1 = 0-9 On/Offs enabled
#define  DSP_CAV_LOCKED      0x00080000L // 1 = cavity locked

#define  DSP_AMPL_PID_ON     0x00100000L // 1 = Closed loop, 0 = Open loop
#define  DSP_PHASE_MODE_0    0x00200000L // 000 = open loop   100 = freq search
#define  DSP_PHASE_MODE_1    0x00400000L // 001 = PID         101 = PID(auto)
#define  DSP_PHASE_MODE_2    0x00800000L // 010 = SEL         110 = SEL(auto)

#define  DSP_UNUSED_24       0x01000000L  //
#define  DSP_UNUSED_25       0x02000000L  //
#define  DSP_UNUSED_26       0x04000000L  //
#define  DSP_UNUSED_27       0x08000000L  //

#define  DSP_UNUSED_28       0x10000000L  //
#define  DSP_UNUSED_29       0x20000000L  //
#define  DSP_UNUSED_30       0x40000000L  //
#define  DSP_UNUSED_31       0x80000000L  //


#define  RF_ON           (statusBitsDSP & DSP_RF_ON)
#define  RF_PULSES       (statusBitsDSP & DSP_RF_PULSES)
#define  TUNER_PID_ON    (statusBitsDSP & DSP_TUNER_PID_ON)
#define  NEW_L_DATA      (statusBitsDSP & DSP_NEW_L_DATA)
#define  NEW_T_DATA      (statusBitsDSP & DSP_NEW_T_DATA)
#define  CUR_PROG_NUM    ((statusBitsDSP >> 8) & 0x03)
#define  CALC_XFER_FUNC  (statusBitsDSP & DSP_CALC_XFER_FUNC)
#define  SAVG_XFER_FUNC  (statusBitsDSP & DSP_SAVG_XFER_FUNC)
#define  SAVING_PROG     (statusBitsDSP & DSP_SAVING_PROG)
#define  FILL_FFT_BUF    (statusBitsDSP & DSP_FILL_FFT_BUF)
#define  COMPUTING_FFT   (statusBitsDSP & DSP_COMPUTING_FFT)
#define  FFT_DATA_VALID  (statusBitsDSP & DSP_FFT_DATA_VALID)
#define  RF_OUT_WARM     (statusBitsDSP & DSP_RF_OUT_WARM)
#define  OUT_RAMPING     (statusBitsDSP & DSP_OUT_RAMPING)
#define  AFC_ON          (statusBitsDSP & DSP_AFC_ON)
#define  CAV_LOCKED      (statusBitsDSP & DSP_CAV_LOCKED)
#define  AMPL_PID_ON     (statusBitsDSP & DSP_AMPL_PID_ON)
#define  PHASE_MODE      ((statusBitsDSP >> 21) & 0x07)



//----- values for statusBitsAFC --- (DSP reg 0x0010001A) ---
#define  DSP_AFC_0_ON        0x00000001  // AFC freq 0 on
#define  DSP_AFC_1_ON        0x00000002  // AFC freq 1 on
#define  DSP_AFC_2_ON        0x00000004  // AFC freq 2 on
#define  DSP_AFC_3_ON        0x00000008  // AFC freq 3 on

#define  DSP_AFC_4_ON        0x00000010  // AFC freq 4 on
#define  DSP_AFC_5_ON        0x00000020  // AFC freq 5 on
#define  DSP_AFC_6_ON        0x00000040  // AFC freq 6 on
#define  DSP_AFC_7_ON        0x00000080  // AFC freq 7 on

#define  DSP_AFC_8_ON        0x00000100  // AFC freq 8 on
#define  DSP_AFC_9_ON        0x00000200  // AFC freq 9 on

#define  AFC_0_ON        (statusBitsAFC & DSP_AFC_0_ON)
#define  AFC_1_ON        (statusBitsAFC & DSP_AFC_1_ON)
#define  AFC_2_ON        (statusBitsAFC & DSP_AFC_2_ON)
#define  AFC_3_ON        (statusBitsAFC & DSP_AFC_3_ON)
#define  AFC_4_ON        (statusBitsAFC & DSP_AFC_4_ON)
#define  AFC_5_ON        (statusBitsAFC & DSP_AFC_5_ON)
#define  AFC_6_ON        (statusBitsAFC & DSP_AFC_6_ON)
#define  AFC_7_ON        (statusBitsAFC & DSP_AFC_7_ON)
#define  AFC_8_ON        (statusBitsAFC & DSP_AFC_8_ON)
#define  AFC_9_ON        (statusBitsAFC & DSP_AFC_9_ON)

//-----------------------------------------------------------------------------
#ifndef NumR4Chans
  #define NumR4Chans_  NumLogicalDevices
  #define R4ChanMult_  400
#endif

#ifdef REG4

REG4( timeZone_,      x, 40001, S16, RW, MB_timeZone )
REG4( curTime,        x, 40002, U32, RW, setU32 )

REG4( ipAddr,         x, 40004, U32, RW, setU32 )
REG4( netmask,        x, 40006, U32, RW, setU32 )
REG4( gateway,        x, 40008, U32, RW, setU32 )

REG4( setACoil,       _, 40010, U16, RW, setU16 )
REG4( resetACoil,     _, 40011, U16, RW, setU16 )

REG4( digOutBits,     x, 40012, U32, RW, setU32 )

REG4( cmdBitsStd,     x, 40014, U32, RW, MB_cmdBitsStd )
REG4( cmdBitsAppl,    x, 40016, U32, RW, MB_cmdBitsAppl )

REG4( setAtten,       x, 40018, U16, RW, MB_setAtten )

REG4( trigPos,        x, 40019, U16, RW, MB_DSP_REG   )
REG4( cmdBitsAppl3,   x, 40020, U16, RW, MB_cmdBitsAppl3 )

REG4( phaseSet,       x, 40021, F32, RW, MB_DSP_REG )
REG4( amplSet,        x, 40023, F32, RW, MB_maxAmpl )

REG4( oldPulseFreq,   x, 40025, U16, RW, MB_oldPulseFreq )

REG4( rf_P,           x, 40026, F32, RW, MB_DSP_REG )
REG4( rf_I,           x, 40028, F32, RW, MB_DSP_REG )
REG4( rf_D,           x, 40030, F32, RW, MB_DSP_REG )

REG4( unusedreg2,     x, 40032, F32, RW, setF32 )
REG4( pulsedAmplSet,  x, 40034, F32, RW, MB_maxAmpl )

REG4( tuner_P,        x, 40036, F32, RW, MB_DSP_REG )
REG4( tuner_I,        x, 40038, F32, RW, MB_DSP_REG )
REG4( tuner_D,        x, 40040, F32, RW, MB_DSP_REG )

REG4( unusedreg3,     x, 40042, F32, RW, setF32 )
REG4( pulsedPhaseSet, x, 40044, F32, RW, MB_DSP_REG )

REG4( setAFCsOnOff,   x, 40046, U32, RW, MB_setAFCsOnOff )

REG4( afcFreq0,       x, 40048, F32, RW, MB_DSP_REG )
REG4( afcFreq1,       x, 40050, F32, RW, MB_DSP_REG )
REG4( afcFreq2,       x, 40052, F32, RW, MB_DSP_REG )
REG4( afcFreq3,       x, 40054, F32, RW, MB_DSP_REG )
REG4( afcFreq4,       x, 40056, F32, RW, MB_DSP_REG )
REG4( afcFreq5,       x, 40058, F32, RW, MB_DSP_REG )
REG4( afcFreq6,       x, 40060, F32, RW, MB_DSP_REG )
REG4( afcFreq7,       x, 40062, F32, RW, MB_DSP_REG )
REG4( afcFreq8,       x, 40064, F32, RW, MB_DSP_REG )
REG4( afcFreq9,       x, 40066, F32, RW, MB_DSP_REG )

REG4( warmLevel,      x, 40068, F32, RW, MB_maxAmpl )
REG4( warmMS,         x, 40070, U16, RW, MB_DSP_REG )
REG4( rampMS,         x, 40071, U16, RW, MB_DSP_REG )

REG4( dspComErrors,   x, 40072, U16, RW, setU16 )
REG4( dspProgNum,     x, 40073, U16, RW, MB_dspProgNum )

REG4( pllInitIdx,     x, 40074, U16, RW, setU16 )

// readings from another device sent to us by the control system
REG4( rfq_t1_feedback, x, 40075, F32, RW, MB_DSP_REG )  // t1_output feedback
REG4( rfq_t2_feedback, x, 40077, F32, RW, MB_DSP_REG )  // t2_output feedback
REG4( devStat1,        x, 40079, U16, RW, setU16 )

// settings forwarded to another device by the control system
REG4( rfq_t1_output,   x, 40080, F32, RO, setF32 )  // tuner 1 output
REG4( rfq_t2_output,   x, 40082, F32, RO, setF32 )  // tuner 2 output
REG4( devCtl1,         x, 40084, U16, RW, setU16 )

// last known values of settings fwded to another device by the ctl system
REG4( devRdSet1,       x, 40085, F32, RW, setF32 )
REG4( devRdSet2,       x, 40087, F32, RW, setF32 )
REG4( devRdCtl1,       x, 40089, U16, RW, setU16 )

REG4( piezoSetPos,     x, 40090, S16, RW, MB_DSP_REG )
REG4( unused_40091,    x, 40091, S16, RW, MB_DSP_REG ) // older (undersized) stepperSetPos

REG4( fwdAtten,        x, 40092, U16, RW, MB_DSP_REG )
REG4( reflAtten,       x, 40093, U16, RW, MB_DSP_REG )
REG4( cavAtten,        x, 40094, U16, RW, MB_DSP_REG )

REG4( freqSearchAmpl,  x, 40095, F32, RW, MB_DSP_REG )
REG4( cavLowIlkLevel,  x, 40097, F32, RW, MB_DSP_REG )
REG4( cavLowIlkTime,   x, 40099, F32, RW, MB_DSP_REG )
REG4( dvdt_IlkRate,    x, 40101, F32, RW, MB_DSP_REG )
REG4( vv_IlkRatio,     x, 40103, F32, RW, MB_DSP_REG )
REG4( fwdCavOffset,    x, 40105, F32, RW, MB_DSP_REG )
REG4( feedFwdPhase,    x, 40107, F32, RW, MB_DSP_REG )
REG4( feedFwdGain,     x, 40109, F32, RW, MB_DSP_REG )

REG4( ctfCalcRate,     x, 40111, F32, RW, MB_DSP_REG )
REG4( fwdMagCalib,     x, 40113, F32, RW, MB_DSP_REG )
REG4( fwdPhaseCalib,   x, 40115, F32, RW, MB_DSP_REG )
REG4( reflMagCalib,    x, 40117, F32, RW, MB_DSP_REG )
REG4( reflPhaseCalib,  x, 40119, F32, RW, MB_DSP_REG )
REG4( cavMagCalib,     x, 40121, F32, RW, MB_DSP_REG )
REG4( cavPhaseCalib,   x, 40123, F32, RW, MB_DSP_REG )
REG4( outMagCalib,     x, 40125, F32, RW, MB_DSP_REG )
REG4( outPhaseCalib,   x, 40127, F32, RW, MB_DSP_REG )

REG4( ctfTunerAmpl,    x, 40129, F32, RW, MB_DSP_REG )

REG4( setDSPState,     x, 40131, U32, RW, MB_setDSPState )

REG4( fwdAmplGain,     x, 40133, F32, RW, MB_setGain )
REG4( fwdAmplOffset,   x, 40135, F32, RW, setF32 )
REG4( fwdPhaseGain,    x, 40137, F32, RW, MB_setGain )
REG4( fwdPhaseOffset,  x, 40139, F32, RW, setF32 )

REG4( reflAmplGain,    x, 40141, F32, RW, MB_setGain )
REG4( reflAmplOffset,  x, 40143, F32, RW, setF32 )
REG4( reflPhaseGain,   x, 40145, F32, RW, MB_setGain )
REG4( reflPhaseOffset, x, 40147, F32, RW, setF32 )

REG4( cavAmplGain,     x, 40149, F32, RW, MB_setGain )
REG4( cavAmplOffset,   x, 40151, F32, RW, setF32 )
REG4( cavPhaseGain,    x, 40153, F32, RW, MB_setGain )
REG4( cavPhaseOffset,  x, 40155, F32, RW, setF32 )

REG4( refAmplGain,     x, 40157, F32, RW, MB_setGain )
REG4( refAmplOffset,   x, 40159, F32, RW, setF32 )
REG4( refPhaseGain,    x, 40161, F32, RW, MB_setGain )
REG4( refPhaseOffset,  x, 40163, F32, RW, setF32 )

REG4( pulseFreq,       x, 40165, F32, RW, MB_DSP_REG )

REG4( afcPhase0,       x, 40167, F32, RW, MB_DSP_REG )
REG4( afcPhase1,       x, 40169, F32, RW, MB_DSP_REG )
REG4( afcPhase2,       x, 40171, F32, RW, MB_DSP_REG )
REG4( afcPhase3,       x, 40173, F32, RW, MB_DSP_REG )
REG4( afcPhase4,       x, 40175, F32, RW, MB_DSP_REG )
REG4( afcPhase5,       x, 40177, F32, RW, MB_DSP_REG )
REG4( afcPhase6,       x, 40179, F32, RW, MB_DSP_REG )
REG4( afcPhase7,       x, 40181, F32, RW, MB_DSP_REG )
REG4( afcPhase8,       x, 40183, F32, RW, MB_DSP_REG )
REG4( afcPhase9,       x, 40185, F32, RW, MB_DSP_REG )

REG4( maxAmplSet,      x, 40187, F32, RW, setF32 )

REG4( rfq_manTuneSet,  x, 40189, F32, RW, MB_DSP_REG )

REG4( rfq_t1_offset,   x, 40191, F32, RW, MB_DSP_REG )
REG4( rfq_t2_offset,   x, 40193, F32, RW, MB_DSP_REG )

REG4( pulse_ramp,      x, 40195, U32, RW, MB_DSP_REG )

REG4( stepperSetPos,   x, 40197, S32, RW, MB_DSP_REG )

REG4( cavHighIlkLevel, x, 40199, F32, RW, MB_DSP_REG )
REG4( fwdHighIlkLevel, x, 40201, F32, RW, MB_DSP_REG )

#undef REG4
#endif


#ifdef RegMacros
//===== application-specific defintions related to register values =====

//  #define  ISET(idx)   REG_ARRAY(float, iSet_1, idx)

  #define  AFCFreq(idx)    REG_ARRAY(float, afcFreq0, idx)
  #define  AFC_Phase(idx)  REG_ARRAY(float, afcPhase0, idx)

#endif

#undef RegMacros


//-----------------------------------------------------------------------------
#ifdef PV_DECLS

char pvStrTest[] = { "....Initial Value...." };  //test

//-------------------------------
// NOTES:
//   - To allow slow clients to read a copy of waveform data without blocking
//     the application task for a significant amount of time, "double-buffering"
//     is used as follows:
//     - any access to waveformData[#] requires acquiring a global lock
//       (basically a simple semaphore with a wait queue)
//     - the application task always reads in to the even #'d rows
//     - the clients always read from the odd #'d rows
//     - when an even #'d row has newer data, both the even and odd rows
//       are locked, the even #'d row being locked just long enough to copy
//       from one to the other
//     - the client currently holding the lock for an odd #'d row can then
//       read from the odd #'d row without blocking the application task
//       (of course it can block other clients wanting to read the same row,
//       although support for multiple reader locks on the odd #'d rows vs
//       single writer locks could improve concurrency quite a bit).
//-----------------------------
#define  MaxDataPoints  1024  // WARNING: max value determined by DSP code
#define  NUM_WAVEFORMS     5

uint  numDataPoints[NUM_WAVEFORMS*2];
ulong far waveformData[NUM_WAVEFORMS*2][MaxDataPoints];
ulong  dataTimeMS[NUM_WAVEFORMS*2], dataTimeSecs[NUM_WAVEFORMS*2];
uint  lastWaveReadTime[NUM_WAVEFORMS*2];
bool   readerRdy[NUM_WAVEFORMS*2];
uint  lastReadPlotTime;

#define  WF_IDX(ptr)  \
  ((uint)(((ulong far *)(ptr) - waveformData[0]) / MaxDataPoints))


//--- indexes in to waveformData[] ---
#define  WAVE_AP_L_SRC    0   // Ampl-Phase Leading-edge Source
#define  WAVE_AP_L_CPY    1   // Ampl-Phase Leading-edge Copy
#define  WAVE_AP_T_SRC    2   // Ampl-Phase Trailing-edge Source
#define  WAVE_AP_T_CPY    3   // Ampl-Phase Trailing-edge Copy
#define  WAVE_FFT_SRC     4   // FFT Source
#define  WAVE_FFT_CPY     5   // FFT Copy
#define  WAVE_CTF_P_SRC   6   // Cavity Transfer Func Phase Source
#define  WAVE_CTF_P_CPY   7   // Cavity Transfer Func Phase Copy
#define  WAVE_CTF_A_SRC   8   // Cavity Transfer Func Ampl Source
#define  WAVE_CTF_A_CPY   9   // Cavity Transfer Func Ampl Copy


ulong  startCTFWaitTime;

#undef PV_DECLS
#endif



#ifdef PV_DEFS

PV_STRING(pvStrTest, PV_RD, PV_UpdateVal, PV_LockVal, PV_SendVal, PV_RcvVal)

PV_ARRAY(AmplPhaseLead, waveformData[1], U32, PV_RD,
                          PV_UpdateValWF, PV_LockWave, PV_SendVal, PV_RcvVal)

PV_ARRAY(AmplPhaseTrail, waveformData[3], U32, PV_RD,
                          PV_UpdateValWF, PV_LockWave, PV_SendVal, PV_RcvVal)

PV_ARRAY(FFT, waveformData[5], F32, PV_RD,
                          PV_UpdateValWF, PV_LockWave, PV_SendVal, PV_RcvVal)

PV_ARRAY(CTF_Phase, waveformData[7], U32, PV_RD,
                          PV_UpdateValWF, PV_LockWave, PV_SendVal, PV_RcvVal)

PV_ARRAY(CTF_Ampl, waveformData[9], F32, PV_RD,
                          PV_UpdateValWF, PV_LockWave, PV_SendVal, PV_RcvVal)

#undef PV_DEFS
#endif


#ifdef PV_Macros


const uint  PV_WF_Idx[NUM_WAVEFORMS] = {
  AmplPhaseLead_IDX,
  AmplPhaseTrail_IDX,
  FFT_IDX,
  CTF_Phase_IDX,
  CTF_Ampl_IDX
};

#define  WAVE_PV(w)  (nonRegPVs + PV_WF_Idx[(w)/2])


#undef PV_Macros
#endif

//-----------------------------------------------------------------------------

#endif  // end of else portion of:  #ifndef RegDefs ... #else ... #endif

#ifndef RegDefs // start of after-RegDefs portion



//---- bit usage for older cmdBitsAppl register ---
#define OLDC_AMPL_OPEN_LOOP     0x00000001  // disable Amplitude PID loop
#define OLDC_AMPL_CLOSED_LOOP   0x00000002  // enable  Amplitude PID loop
#define OLDC_PHASE_OPEN_LOOP    0x00000004  // disable Phase PID loop
#define OLDC_PHASE_CLOSED_LOOP  0x00000008  // enable  Phase PID loop
#define OLDC_AMPL_TUNE_RUN      0x00000010  // run  Ampl  PID tuning pulse test
#define OLDC_AMPL_TUNE_STOP     0x00000020  // stop Ampl  PID tuning pulse test
#define OLDC_PHASE_TUNE_RUN     0x00000040  // run  Phase PID tuning pulse test
#define OLDC_PHASE_TUNE_STOP    0x00000080  // stop Phase PID tuning pulse test
#define OLDC_RF_ON              0x00000100  // Turn RF On
#define OLDC_RF_OFF             0x00000200  // Turn RF Off
#define OLDC_L_DATA_READ        0x00000400  // Leading edge data buffer read
#define OLDC_T_DATA_READ        0x00000800  // Trailing edge data buffer read
#define OLDC_CAV_TUNER_ON       0x00001000  // Turn Cavity Tuner On
#define OLDC_CAV_TUNER_OFF      0x00002000  // Turn Cavity Tuner Off
#define OLDC_FFT_DATA_READ      0x00004000  // FFT data buffer read complete

//---- bit usage for setDSPState (DSP 0x00100201) ---
#define  CMD_AMPL_PID_OFF       0x00000001  // disable Ampl PID loop
#define  CMD_AMPL_PID_ON        0x00000002  // enable  Ampl PID loop
#define  CMD_AMPL_PID_BITS      0x00000003L

#define  CMD_PHASE_CTL_OFF      0x00000004  // disable Phase PID loop
#define  CMD_PHASE_CTL_PID      0x00000008  // enable  Phase PID/disable SEL
#define  CMD_PHASE_CTL_SEL      0x00000010  // enable SEL/disable phase PID
#define  CMD_PHASE_CTL_AUTO     0x00000020  // auto SEL/phase PID control
#define  CMD_PHASE_CTL_BITS     0x0000003CL

#define  CMD_RF_PULSING_ON      0x00000040  // run  Phase PID tuning pulse test
#define  CMD_RF_PULSING_OFF     0x00000080  // stop Phase PID tuning pulse test
#define  CMD_RF_PULSING_BITS    0x000000C0L

#define  CMD_UNUSED_8           0x00000100  //
#define  CMD_UNUSED_9           0x00000200  //

#define  CMD_L_DATA_READ        0x00000400  // Done reading Leading edge plot
#define  CMD_T_DATA_READ        0x00000800  // Done reading Trailing edge plot

#define  CMD_AFC_ON             0x00001000  // turn AFC On
#define  CMD_AFC_OFF            0x00002000  // turn AFC Off
#define  CMD_AFC_BITS           0x00003000L

#define  CMD_FFT_DATA_READ      0x00004000  // Done reading FFT plot
#define  CMD_UNUSED_15          0x00008000  //

#define  CMD_TUNER_PID_OFF      0x00010000L // disable tuner PID
#define  CMD_TUNER_PID_ON       0x00020000L // enable tuner PID
#define  CMD_TUNER_PID_BITS     0x00030000L

#define  CMD_UNUSED_18          0x00040000L //
#define  CMD_UNUSED_19          0x00080000L //

#define  CMD_UNUSED_20          0x00100000L //
#define  CMD_UNUSED_21          0x00200000L //
#define  CMD_UNUSED_22          0x00400000L //
#define  CMD_UNUSED_23          0x00800000L //

#define  CMD_UNUSED_24          0x01000000L //
#define  CMD_UNUSED_25          0x02000000L //
#define  CMD_UNUSED_26          0x04000000L //
#define  CMD_UNUSED_27          0x08000000L //

#define  CMD_UNUSED_28          0x10000000L //
#define  CMD_UNUSED_29          0x20000000L //
#define  CMD_UNUSED_30          0x40000000L //
#define  CMD_UNUSED_31          0x80000000L //

//--- bit-masks to determine which groups have a bit set ---
const ulong SetDSPStateMasks[] = {
  CMD_AMPL_PID_BITS,   // enable/disable Amplitude PID loop bits
  CMD_PHASE_CTL_BITS,  // Phase control mode bits
  CMD_RF_PULSING_BITS, // RF pulsing start/stop bits
  CMD_AFC_BITS,        // Turn AFC On/Off biuts
  CMD_TUNER_PID_BITS,  // enable/disable Tuner PID loop bits
  0  //indicates end of list
};



const char far *SetDSPStateDesc[32] = {
  "Ampl-PID Off",         "Ampl-PID On",

  "Phase-PID Off",        "Phase-PID On",
  "SEL mode enabled",     "Auto Phase-PID/SEL mode",

  "RF Pulsing On",        "RF Pulsing Off",

  "",                     "",
  "",                     "",

  "AFC/Cavity Tuner On",  "AFC/Cavity Tuner Off",

  "",                     "",

  "Tuner-PID Off",        "Tuner-PID On",

  "",                     "",
  "",                     "",
  "",                     "",
  "",                     "",
  "",                     "",
  "",                     "",
  "",                     ""
};

//---- bit usage for setAFCsOnOff --- (DSP 0x0207)
#define  CMD_AFC_0_ON           0x00000001L  // Turn On  AFC freq 0
#define  CMD_AFC_0_OFF          0x00000002L  // Turn Off AFC freq 0
#define  CMD_AFC_1_ON           0x00000004L  // Turn On  AFC freq 1
#define  CMD_AFC_1_OFF          0x00000008L  // Turn Off AFC freq 1
#define  CMD_AFC_2_ON           0x00000010L  // Turn On  AFC freq 2
#define  CMD_AFC_2_OFF          0x00000020L  // Turn Off AFC freq 2
#define  CMD_AFC_3_ON           0x00000040L  // Turn On  AFC freq 3
#define  CMD_AFC_3_OFF          0x00000080L  // Turn Off AFC freq 3
#define  CMD_AFC_4_ON           0x00000100L  // Turn On  AFC freq 4
#define  CMD_AFC_4_OFF          0x00000200L  // Turn Off AFC freq 4
#define  CMD_AFC_5_ON           0x00000400L  // Turn On  AFC freq 5
#define  CMD_AFC_5_OFF          0x00000800L  // Turn Off AFC freq 5
#define  CMD_AFC_6_ON           0x00001000L  // Turn On  AFC freq 6
#define  CMD_AFC_6_OFF          0x00002000L  // Turn Off AFC freq 6
#define  CMD_AFC_7_ON           0x00004000L  // Turn On  AFC freq 7
#define  CMD_AFC_7_OFF          0x00008000L  // Turn Off AFC freq 7
#define  CMD_AFC_8_ON           0x00010000L  // Turn On  AFC freq 8
#define  CMD_AFC_8_OFF          0x00020000L  // Turn Off AFC freq 8
#define  CMD_AFC_9_ON           0x00040000L  // Turn On  AFC freq 9
#define  CMD_AFC_9_OFF          0x00080000L  // Turn Off AFC freq 9

const char far *SetAFCsOnOffDesc[] = {
  "AFC_0 On",  "AFC_0 Off",
  "AFC_1 On",  "AFC_1 Off",
  "AFC_2 On",  "AFC_2 Off",
  "AFC_3 On",  "AFC_3 Off",
  "AFC_4 On",  "AFC_4 Off",
  "AFC_5 On",  "AFC_5 Off",
  "AFC_6 On",  "AFC_6 Off",
  "AFC_7 On",  "AFC_7 Off",
  "AFC_8 On",  "AFC_8 Off",
  "AFC_9 On",  "AFC_9 Off",
  "",          "",
  "",          "",
  "",          "",
  "",          "",
  "",          "",
  "",          ""
};


//---- bit usage for cmdBitsAppl3 ---
#define  CMD_RESET_DSP          0x0001  // Reset the DSP
#define  CMD_CALC_CTF           0x0002  // begin CTF calculation



//--- parameter ID's for data in PhaseCtlr Device Data file ---
//--------------------------------------------------------------------------
// WARNING:  ID's can be added, but existing ID's (and the format of the
//           data the represent) should NEVER change!  NOTE:  Existing ones
//           can be "eliminated" by changing their name and no longer
//           using them, but never REUSE an old one for something new!!!
//--------------------------------------------------------------------------
//  NOTE:  ID's >= 1000 are rsvd for values shared by most applications
//--------------------------------------------------------------------------
const ParamData far  devDataParams[] = {
//                ID   ---value---  --RESTORE_FUNC--
//  PERSISTENT_VAR(  1,  pidParams,      NULL ),
//  PERSISTENT_VAR(  2,  afcParams,      NULL ),
//  PERSISTENT_VAR(  3,  devSettings,    NULL ), // ID for older vers of struct
//  PERSISTENT_VAR(  4,  warmUpParams,   NULL ),

  PERSISTENT_VAR(  5,  pllInitIdx,     NULL ),

//  PERSISTENT_VAR(  6,  devSettings,    NULL ),

  PERSISTENT_VAR( 7,  trigPos,         NULL ),

  PERSISTENT_VAR( 8,  warmLevel,       NULL ),
  PERSISTENT_VAR( 9,  warmMS,          NULL ),
  PERSISTENT_VAR( 10, rampMS,          NULL ),

  PERSISTENT_VAR( 11, fwdAtten,        NULL ),
  PERSISTENT_VAR( 12, reflAtten,       NULL ),
  PERSISTENT_VAR( 13, cavAtten,        NULL ),

//  PERSISTENT_VAR( 14, phaseSet,        NULL ),  // moved to settings
//  PERSISTENT_VAR( 15, amplSet,         NULL ),  // moved to settings
//  PERSISTENT_VAR( 16, oldPulseFreq,    NULL ),  // replaced by float value

  PERSISTENT_VAR( 17, freqSearchAmpl,  NULL ),

  PERSISTENT_VAR( 18, rf_P,            NULL ),
  PERSISTENT_VAR( 19, rf_I,            NULL ),
  PERSISTENT_VAR( 20, rf_D,            NULL ),

  PERSISTENT_VAR( 21, tuner_P,         NULL ),
  PERSISTENT_VAR( 22, tuner_I,         NULL ),
  PERSISTENT_VAR( 23, tuner_D,         NULL ),

  PERSISTENT_VAR( 24, pulsedAmplSet,   NULL ),
  PERSISTENT_VAR( 25, pulsedPhaseSet,  NULL ),

//  PERSISTENT_VAR( 26, afcFreq0,       <write-once-reg-func> ),
//  PERSISTENT_VAR( 27, afcFreq1,       <write-once-reg-func> ),
//  PERSISTENT_VAR( 28, afcFreq2,       <write-once-reg-func> ),
//  PERSISTENT_VAR( 29, afcFreq3,       <write-once-reg-func> ),
//  PERSISTENT_VAR( 30, afcFreq4,       <write-once-reg-func> ),
//  PERSISTENT_VAR( 31, afcFreq5,       <write-once-reg-func> ),
//  PERSISTENT_VAR( 32, afcFreq6,       <write-once-reg-func> ),
//  PERSISTENT_VAR( 33, afcFreq7,       <write-once-reg-func> ),
//  PERSISTENT_VAR( 34, afcFreq8,       <write-once-reg-func> ),
//  PERSISTENT_VAR( 35, afcFreq9,       <write-once-reg-func> ),

  PERSISTENT_VAR( 36, cavLowIlkLevel,  NULL ),
  PERSISTENT_VAR( 37, cavLowIlkTime,   NULL ),
  PERSISTENT_VAR( 38, dvdt_IlkRate,    NULL ),
  PERSISTENT_VAR( 39, vv_IlkRatio,     NULL ),

  PERSISTENT_VAR( 40, fwdCavOffset,    NULL ),
  PERSISTENT_VAR( 41, feedFwdPhase,    NULL ),
  PERSISTENT_VAR( 42, feedFwdGain,     NULL ),

  PERSISTENT_VAR( 43, ctfCalcRate,     NULL ),

//  PERSISTENT_VAR( 44, fwdMagCalib,     NULL ),  // not "single instance" values
//  PERSISTENT_VAR( 45, fwdPhaseCalib,   NULL ),
//  PERSISTENT_VAR( 46, reflMagCalib,    NULL ),
//  PERSISTENT_VAR( 47, reflPhaseCalib,  NULL ),
//  PERSISTENT_VAR( 48, cavMagCalib,     NULL ),
//  PERSISTENT_VAR( 49, cavPhaseCalib,   NULL ),
//  PERSISTENT_VAR( 50, outMagCalib,     NULL ),
//  PERSISTENT_VAR( 51, outPhaseCalib,   NULL ),

  PERSISTENT_VAR( 52, ctfTunerAmpl,    NULL ),

  PERSISTENT_VAR( 53, fwdAmplGain,     NULL ),
  PERSISTENT_VAR( 54, fwdAmplOffset,   NULL ),
  PERSISTENT_VAR( 55, fwdPhaseGain,    NULL ),
  PERSISTENT_VAR( 56, fwdPhaseOffset,  NULL ),

  PERSISTENT_VAR( 57, reflAmplGain,    NULL ),
  PERSISTENT_VAR( 58, reflAmplOffset,  NULL ),
  PERSISTENT_VAR( 59, reflPhaseGain,   NULL ),
  PERSISTENT_VAR( 60, reflPhaseOffset, NULL ),

  PERSISTENT_VAR( 61, cavAmplGain,     NULL ),
  PERSISTENT_VAR( 62, cavAmplOffset,   NULL ),
  PERSISTENT_VAR( 63, cavPhaseGain,    NULL ),
  PERSISTENT_VAR( 64, cavPhaseOffset,  NULL ),

  PERSISTENT_VAR( 65, refAmplGain,     NULL ),
  PERSISTENT_VAR( 66, refAmplOffset,   NULL ),
  PERSISTENT_VAR( 67, refPhaseGain,    NULL ),
  PERSISTENT_VAR( 68, refPhaseOffset,  NULL ),

  PERSISTENT_VAR( 69, pulseFreq,       NULL ),

  PERSISTENT_VAR( 70, maxAmplSet,      NULL ),

  PERSISTENT_VAR( 71, rfq_t1_offset,   NULL ),
  PERSISTENT_VAR( 72, rfq_t2_offset,   NULL ),

  PERSISTENT_VAR( 73, pulse_ramp,      NULL ),

  PERSISTENT_VAR( 74, cavHighIlkLevel, NULL ),
  PERSISTENT_VAR( 75, fwdHighIlkLevel, NULL ),


  #define SharedDevParams
  #include <sharedAppl.c>

  END_PARAM_DATA
};

//----------------------------------------------------

//RESTORE_FUNC(xxxSet);

const ParamData far  settingsParams[] = {
//APPLDATA
//                ID   ---value---  --RESTORE_FUNC--
//  PERSISTENT_VAR(  1,  xxxSet,     PVRF(xxxSet) ),
  PERSISTENT_VAR(  1,  piezoSetPos,    NULL ),
//  PERSISTENT_VAR(  2,  stepperSetPos,  NULL ),  // older S16 value
  PERSISTENT_VAR(  3,  phaseSet,       NULL ),
  PERSISTENT_VAR(  4,  amplSet,        NULL ),
  PERSISTENT_VAR(  5,  setDSPState,    NULL ),
  PERSISTENT_VAR(  6,  stepperSetPos,  NULL ),

  END_PARAM_DATA
};


//-----------------------------------------------------------------
//  Application-specific values needed to restore the state of the
//  controller following a "soft" reset (i.e. a restart without a
//  CPU reset)
//-----------------------------------------------------------------
const ParamData far  ctlrStateParams[] = {
// sample values from sharedAppl.c, common to all applications
//                   --ID--  ---value,granularity---  --RESTORE_FUNC--
//    PERSISTENT_VARx(  1000,  prevState,               NULL ),
//    PERSISTENT_VARx(  1001,  curState,                NULL ),
//    PERSISTENT_VARx(  1002,  nextState,               NULL ),
//    PERSISTENT_TBL (  1003,  setVal,4,                NULL ),
// // PERSISTENT_TBL (  1004,  shadowRegValues,4,       NULL ),

//APPLDATA
//    PERSISTENT_VAR(  1,  myApplStateVar,        NULL ),


  #define SharedCtlrStateParams
  #include <sharedAppl.c>

  END_PARAM_DATA
};


//-----------------------------------------------------------------------------
//  A list of parameterized "persistent" data.  Backup copies of such values
//  are stored in files and restored on startup.
//-----------------------------------------------------------------------------

//    Base File Name    Category  ParamData
//    ----------------  --------  ---------
//--- add any application-specific parameterized file defs here ---
//const ParamFile far  myAppCfgFile =
//  { "myFileName",     PDC_CTLR, myCfgParams,   0 };

ParamFile  settingsFile =
    { "settings",       PDC_SET,  settingsParams  };


#define SharedParams
#include <sharedAppl.c>


ParamFile *paramFiles[] = {
  #define SharedParamsList
  #include <sharedAppl.c>
//--- add any application-specific ParamData references here ---
//  &myAppCfgFile
  &settingsFile
};

const int NumParamFiles = sizeof(paramFiles) / sizeof(paramFiles[0]);

//=============================================================================
//  Data structures for managing communication with the DSP and for keeping
//  ModBus register values in sync with the corresponding DSP registers.
//=============================================================================

typedef struct {
  ulong    id;         // DSP "address"
  uchar    type;       // int32, float, bits, ...
  RegInfo *mbReg;      // corresponding Modbus register
  float   *gain;       // ptr to gain value or NULL
  float   *offset;     // ptr to offset value or NULL
  ulong    typical;    // for testing on other hdw
  uint     percNoise;  // max noise as percentage of typical value
} RD_DSP_REG;

typedef struct {
  ulong    id;         // reg # in DSP to write value to (if applicable)
  uchar    type;       // int32, float, bits, ...
  RegInfo *mbReg;      // Modbus register info
  float    minVal;     // minimum value allowed
  float    maxVal;     // maximum value allowed
  uchar    writeMode;  // how to handle writes
  float   *gain;       // ptr to gain value or NULL
  float   *offset;     // ptr to offset value or NULL
} WR_DSP_REG;

float  ilkRatioGain, ilkRatioOffset, maxOutAmplGain, maxOutAmplOffset;


//--- values for WR_DSP_REG.writeMode ---
#define  WM_ONCE    0  // write new value only once at time it was received
#define  WM_EVERY   1  // write cur value every pass

//--- values for RD_DSP_REG.type and WR_DSP_REG.type ---
#define  RT_BITS   DT_U32
#define  RT_S32    DT_S32
#define  RT_U32    DT_U32
#define  RT_F32    DT_F32
#define  RT_PHASE  99



#define  DSP_RD( dspName, dspType, mbReg, gain, offset, typ, noise )  \
    { dspName, RT_##dspType, &mbReg##_RegInfo, gain, offset, typ, noise }


const RD_DSP_REG far rdDSPregs[] = {
  //      DSP Reg ID      DSP fmt ModBus Reg      Gain            Offset
  //      --------------  ------- --------------  --------------  ---------------
  DSP_RD( FwdPhase,       PHASE,  fwdPhase,       &fwdPhaseGain,  &fwdPhaseOffset,    0x8230D9E6L,  5 ),
  DSP_RD( ReflPhase,      PHASE,  reflPhase,      &reflPhaseGain, &reflPhaseOffset,   0xE64288E8L,  5 ),
  DSP_RD( CavPhase,       PHASE,  cavPhase,       &cavPhaseGain,  &cavPhaseOffset,    0xDA93A22CL,  5 ),
  DSP_RD( RefPhase,       PHASE,  refPhase,       &refPhaseGain,  &refPhaseOffset,    0xEBF0809CL,  5 ),

  DSP_RD( FwdAmpl,        F32,    fwdAmpl,        &fwdAmplGain,   &fwdAmplOffset,     0x3F8D0BA5L,  5 ),
  DSP_RD( ReflAmpl,       F32,    reflAmpl,       &reflAmplGain,  &reflAmplOffset,    0x3E606060L,  5 ),
  DSP_RD( CavAmpl,        F32,    cavAmpl,        &cavAmplGain,   &cavAmplOffset,     0x4013A76BL,  5 ),
  DSP_RD( RefAmpl,        F32,    refAmpl,        &refAmplGain,   &refAmplOffset,     0x3EE4FE6EL,  5 ),

  DSP_RD( MaxFwdAmpl,     F32,    maxFwdAmpl,     &fwdAmplGain,   &fwdAmplOffset,     0x407918DEL,  0 ),
  DSP_RD( MaxReflAmpl,    F32,    maxReflAmpl,    &reflAmplGain,  &reflAmplOffset,    0x3FE3C7EEL,  0 ),
  DSP_RD( MaxCavAmpl,     F32,    maxCavAmpl,     &cavAmplGain,   &cavAmplOffset,     0x40DF6F19L,  0 ),
  DSP_RD( MaxOutAmpl,     F32,    maxOutAmpl,     &maxOutAmplGain, &maxOutAmplOffset, 0x3FCD6FCEL,  0 ),

  DSP_RD( CavHighIlkLevel,F32,    cavHighIlkLevel,&cavAmplGain,   &cavAmplOffset,     0x4013A76BL,  5 ),
  DSP_RD( FwdHighIlkLevel,F32,    fwdHighIlkLevel,&fwdAmplGain,   &fwdAmplOffset,     0x4013A76BL,  5 ),

  DSP_RD( DSP_Status,     BITS,   statusBitsDSP,  NULL,           NULL,               0x00302088L,  0 ),
  DSP_RD( AFC_State,      BITS,   statusBitsAFC,  NULL,           NULL,               0,            0 ),

  DSP_RD( DSP_Load,       F32,    dspLoad,        NULL,           NULL,               0x42736725L, 10 ),
  DSP_RD( PCB_Temp,       F32,    dspTemp,        NULL,           NULL,               0x4284B886L,  5 ),

  DSP_RD( RFQ_T1_Output,  F32,    rfq_t1_output,  NULL,           NULL,               0x00000000L,  0 ),
  DSP_RD( RFQ_T2_Output,  F32,    rfq_t2_output,  NULL,           NULL,               0x00000000L,  0 ),

//--- R/W registers that we want to read from each pass ---
  DSP_RD( PiezoTunerPos,  U32,    piezoReadPos,   NULL,           NULL,               0x00000000L,  0 ),
  DSP_RD( StepperPos,     S32,    stepperReadPos, NULL,           NULL,               0x00000000L,  0 ),

//--- RO registers in "special" registers group ---
  DSP_RD( WriteStatus,    U32,    writeStatus,    NULL,           NULL,               0x00000064L,  0 ),
  DSP_RD( PercentCTF,     U32,    percentCTF,     NULL,           NULL,               0x00000064L,  0 ),
  DSP_RD( SizeCTF,        U32,    sizeCTF,        NULL,           NULL,               0x000003E8L,  0 ),
  DSP_RD( PercentFFT,     U32,    percentFFT,     NULL,           NULL,               0x00000064L,  0 ),
  DSP_RD( SizeFFT,        U32,    sizeFFT,        NULL,           NULL,               0x000003E9L,  0 ),
  DSP_RD( ROMProgErrs,    U32,    romProgErrs,    NULL,           NULL,               0x00000000L,  0 ),

//--- R/W registers in "special" group that provide the current value from ---
//--- a table of values ---
  DSP_RD( FwdMagCalib,    F32,    fwdMagCalib,    NULL,           NULL,               0x311743D5L,  0 ),
  DSP_RD( FwdPhaseCalib,  PHASE,  fwdPhaseCalib,  NULL,           NULL,               0x8C7EAFCEL,  0 ),
  DSP_RD( ReflMagCalib,   F32,    reflMagCalib,   NULL,           NULL,               0x308A5217L,  0 ),
  DSP_RD( ReflPhaseCalib, PHASE,  reflPhaseCalib, NULL,           NULL,               0x96F5A5A2L,  0 ),
  DSP_RD( CavMagCalib,    F32,    cavMagCalib,    NULL,           NULL,               0x3187AE52L,  0 ),
  DSP_RD( CavPhaseCalib,  PHASE,  cavPhaseCalib,  NULL,           NULL,               0x8F7F91BCL,  0 ),
  DSP_RD( OutMagCalib,    F32,    outMagCalib,    NULL,           NULL,               0x30A93DD5L,  0 ),
  DSP_RD( OutPhaseCalib,  PHASE,  outPhaseCalib,  NULL,           NULL,               0x00000000L,  0 )
};

#define  NUM_RD_DSP_REGS  (sizeof(rdDSPregs) / sizeof(RD_DSP_REG))


#define  DSP_WR( dspName, dspType, mbReg, min, max, mode, gain, offset )  \
    { dspName, RT_##dspType, &mbReg##_RegInfo, min, max, WM_##mode, gain, offset }

const WR_DSP_REG far wrDSPregs[] = {
  //      DSP Reg ID      DSP fmt ModBus Reg       Min    Max     Mode  Gain            Offset
  //      --------------  ------- --------------   ----   ------  ----  --------------  ---------------
  DSP_WR( TriggerPos,     U32,    trigPos,         0,     1023,   EVERY, NULL,          NULL ),
  DSP_WR( WarmUpAmpl,     F32,    warmLevel,       0.0,   0.0,    EVERY, &cavAmplGain,  &cavAmplOffset ),
  DSP_WR( WarmUpTime,     F32,    warmMS,          0.0,   32767,  EVERY, NULL,          NULL ),
  DSP_WR( RampTime,       F32,    rampMS,          0.0,   32767,  EVERY, NULL,          NULL ),

  DSP_WR( FwdAtten,       U32,    fwdAtten,        0,     63,     EVERY, NULL,          NULL ),
  DSP_WR( ReflAtten,      U32,    reflAtten,       0,     63,     EVERY, NULL,          NULL ),
  DSP_WR( CavAtten,       U32,    cavAtten,        0,     63,     EVERY, NULL,          NULL ),

  DSP_WR( PhaseSet,       PHASE,  phaseSet,        0.0,   0.0,    EVERY, &cavPhaseGain, &cavPhaseOffset ),
  DSP_WR( AmplSet,        F32,    amplSet,         0.0,   0.0,    EVERY, &cavAmplGain,  &cavAmplOffset ),

  DSP_WR( TunePulseFreq,  F32,    pulseFreq,       0.0,   0.0,    EVERY, NULL,          NULL ),
  DSP_WR( FreqSearchAmpl, F32,    freqSearchAmpl,  0.0,   0.0,    EVERY, &cavAmplGain, &cavAmplOffset ),

  DSP_WR( RF_P,           F32,    rf_P,            0.0,   0.0,    EVERY, NULL,          NULL ),
  DSP_WR( RF_I,           F32,    rf_I,            0.0,   0.0,    EVERY, NULL,          NULL ),
  DSP_WR( RF_D,           F32,    rf_D,            0.0,   0.0,    EVERY, NULL,          NULL ),

  DSP_WR( Tuner_P,        F32,    tuner_P,         0.0,   0.0,    EVERY, NULL,          NULL ),
  DSP_WR( Tuner_I,        F32,    tuner_I,         0.0,   0.0,    EVERY, NULL,          NULL ),
  DSP_WR( Tuner_D,        F32,    tuner_D,         0.0,   0.0,    EVERY, NULL,          NULL ),

  DSP_WR( PulsedAmplSet,  F32,    pulsedAmplSet,   0.0,   0.0,    EVERY, &cavAmplGain,  &cavAmplOffset ),
  DSP_WR( PulsedPhaseSet, PHASE,  pulsedPhaseSet,  0.0,   0.0,    EVERY, &cavPhaseGain, &cavPhaseOffset ),

  DSP_WR( AFC_Freq0,      F32,    afcFreq0,        0.0,   1000.0,  ONCE, NULL,          NULL ),
  DSP_WR( AFC_Freq1,      F32,    afcFreq1,        0.0,   1000.0,  ONCE, NULL,          NULL ),
  DSP_WR( AFC_Freq2,      F32,    afcFreq2,        0.0,   1000.0,  ONCE, NULL,          NULL ),
  DSP_WR( AFC_Freq3,      F32,    afcFreq3,        0.0,   1000.0,  ONCE, NULL,          NULL ),
  DSP_WR( AFC_Freq4,      F32,    afcFreq4,        0.0,   1000.0,  ONCE, NULL,          NULL ),
  DSP_WR( AFC_Freq5,      F32,    afcFreq5,        0.0,   1000.0,  ONCE, NULL,          NULL ),
  DSP_WR( AFC_Freq6,      F32,    afcFreq6,        0.0,   1000.0,  ONCE, NULL,          NULL ),
  DSP_WR( AFC_Freq7,      F32,    afcFreq7,        0.0,   1000.0,  ONCE, NULL,          NULL ),
  DSP_WR( AFC_Freq8,      F32,    afcFreq8,        0.0,   1000.0,  ONCE, NULL,          NULL ),
  DSP_WR( AFC_Freq9,      F32,    afcFreq9,        0.0,   1000.0,  ONCE, NULL,          NULL ),

  DSP_WR( CTFTunerAmpl,   F32,    ctfTunerAmpl,    0.0,   32767,  EVERY, NULL,          NULL ),
  DSP_WR( CavLowIlkLevel, F32,    cavLowIlkLevel,  0.0,   0.0,    EVERY, &cavAmplGain,  &cavAmplOffset ),
  DSP_WR( CavLowIlkTime,  F32,    cavLowIlkTime,   0.0,   0.0,    EVERY, NULL,          NULL ),
  DSP_WR( DvDt_IlkRate,   F32,    dvdt_IlkRate,    0.0,   0.0,    EVERY, &cavAmplGain,  &cavAmplOffset ),
  DSP_WR( VV_IlkRatio,    F32,    vv_IlkRatio,     0.0,   0.0,    EVERY, &ilkRatioGain, &ilkRatioOffset ),

  DSP_WR( FwdCavOffset,   PHASE,  fwdCavOffset,    0.0,   0.0,    EVERY, NULL,          NULL ),
  DSP_WR( FeedFwdPhase,   PHASE,  feedFwdPhase,    0.0,   0.0,    EVERY, NULL,          NULL ),
  DSP_WR( FeedFwdGain,    F32,    feedFwdGain,     0.0,   0.0,    EVERY, NULL,          NULL ),

  DSP_WR( AFC_Phase0,     PHASE,  afcPhase0,       0.0,   0.0,    ONCE,  NULL,          NULL ),
  DSP_WR( AFC_Phase1,     PHASE,  afcPhase1,       0.0,   0.0,    ONCE,  NULL,          NULL ),
  DSP_WR( AFC_Phase2,     PHASE,  afcPhase2,       0.0,   0.0,    ONCE,  NULL,          NULL ),
  DSP_WR( AFC_Phase3,     PHASE,  afcPhase3,       0.0,   0.0,    ONCE,  NULL,          NULL ),
  DSP_WR( AFC_Phase4,     PHASE,  afcPhase4,       0.0,   0.0,    ONCE,  NULL,          NULL ),
  DSP_WR( AFC_Phase5,     PHASE,  afcPhase5,       0.0,   0.0,    ONCE,  NULL,          NULL ),
  DSP_WR( AFC_Phase6,     PHASE,  afcPhase6,       0.0,   0.0,    ONCE,  NULL,          NULL ),
  DSP_WR( AFC_Phase7,     PHASE,  afcPhase7,       0.0,   0.0,    ONCE,  NULL,          NULL ),
  DSP_WR( AFC_Phase8,     PHASE,  afcPhase8,       0.0,   0.0,    ONCE,  NULL,          NULL ),
  DSP_WR( AFC_Phase9,     PHASE,  afcPhase9,       0.0,   0.0,    ONCE,  NULL,          NULL ),

  DSP_WR( RFQ_ManTuneSet, F32,    rfq_manTuneSet,  0.0,   0.0,    EVERY, NULL,          NULL ),
  DSP_WR( RFQ_T1_Offset,  F32,    rfq_t1_offset,   0.0,   0.0,    EVERY, NULL,          NULL ),
  DSP_WR( RFQ_T2_Offset,  F32,    rfq_t2_offset,   0.0,   0.0,    EVERY, NULL,          NULL ),
  DSP_WR( RFQ_T1_Feedback,F32,    rfq_t1_feedback, 0.0,   0.0,    EVERY, NULL,          NULL ),
  DSP_WR( RFQ_T2_Feedback,F32,    rfq_t2_feedback, 0.0,   0.0,    EVERY, NULL,          NULL ),

  DSP_WR( FwdAmplGain,    F32,    fwdAmplGain,     0.0,   0.0,    EVERY, NULL,          NULL ),
  DSP_WR( CavAmplGain,    F32,    cavAmplGain,     0.0,   0.0,    EVERY, NULL,          NULL ),

//--- WO registers in the "special" registers group ---
  DSP_WR( SetDSPState,    U32,    setDSPState,     0,     0,      EVERY, NULL,          NULL ),
  DSP_WR( SetAFCsOnOff,   U32,    setAFCsOnOff,    0,     0,      EVERY, NULL,          NULL ),

//--- R/W registers in the "special" registers group ---
  DSP_WR( PiezoTunerPos,  U32,    piezoSetPos,     0,     0,      EVERY, NULL,          NULL ),
  DSP_WR( StepperPos,     S32,    stepperSetPos,   0,     0,      EVERY, NULL,          NULL ),
  DSP_WR( CTFCalcRate,    F32,    ctfCalcRate,     0.0,   0.0,    EVERY, NULL,          NULL ),

  DSP_WR( Pulse_Ramp,     U32,    pulse_ramp,      0,     0,      EVERY, NULL,          NULL ),

//--- R/W registers in "special" group that write a value into a table ---
//--- and cannot be continuously written to ---
  DSP_WR( FwdMagCalib,    F32,    fwdMagCalib,     0.0,   0.0,     ONCE, NULL,          NULL ),
  DSP_WR( FwdPhaseCalib,  PHASE,  fwdPhaseCalib,   0.0,   0.0,     ONCE, NULL,          NULL ),
  DSP_WR( ReflMagCalib,   F32,    reflMagCalib,    0.0,   0.0,     ONCE, NULL,          NULL ),
  DSP_WR( ReflPhaseCalib, PHASE,  reflPhaseCalib,  0.0,   0.0,     ONCE, NULL,          NULL ),
  DSP_WR( CavMagCalib,    F32,    cavMagCalib,     0.0,   0.0,     ONCE, NULL,          NULL ),
  DSP_WR( CavPhaseCalib,  PHASE,  cavPhaseCalib,   0.0,   0.0,     ONCE, NULL,          NULL ),
  DSP_WR( OutMagCalib,    F32,    outMagCalib,     0.0,   0.0,     ONCE, NULL,          NULL ),
  DSP_WR( OutPhaseCalib,  PHASE,  outPhaseCalib,   0.0,   0.0,     ONCE, NULL,          NULL )
};

#define  NUM_WR_DSP_REGS  (sizeof(wrDSPregs) / sizeof(WR_DSP_REG))


//=============================================================================
//  Definitions for event msgs.  Event msgs are the means that other processes
//  use to inform the application process of an event that it needs to handle.
//
//  In general, allowing multiple processes to concurrently modify the state
//  of the controller is a sure recipie for chaos.  By restricting all such
//  changes to the context of a single process, and sending messages to that
//  process to request changes, we eliminate all such concurrency issues (not
//  the least of which is keeping track of the order of events).
//
//  The most common event is a request for change of state that corresponds to
//  a message received from an external client via a protocol like ModBus.
//=============================================================================

//EVENT_DEF( myEvent )


//=============================================================================
//  This is where misc global variables should be defined
//=============================================================================
#define SharedGlobals
#include <sharedAppl.c>

//--- add definitions for application-specific global variables here ---

LockData  locks[NUM_LOCKS];



fd_t  spi1fd, spi2fd;
SerialInfo  *spi1Port, *spi2Port;  // can't share fd's between processes yet...



bool  fpgaInitialized, pulseAmpl, pulsePhase;

uint  ADCVal;

bool  autoReadDSP, autoWriteDSP, showFiller, readWaves;

bool  rfIsOn, rfPulsing, tunerPIDisON, rfAtWarm, rfRamping, afcEnabled,
      cavLocked, amplPIDisON;

uint  curProgNum, phaseMode, lastDSPioTime;

const char *PhaseModeDesc[] = { "    OPEN", "     PID", "     SEL", "????????",
                                "FREQSRCH", "PID/AUTO", "SEL/AUTO", "????????" };


//=============================================================================
//  Temporary functions to fill in for not-yet implemented system functions
//=============================================================================


#define LockOwnedBy(id)  ( locks[id].owner )
#define NotLockOwner(id)  (LockOwnedBy(id) != getpid())

//-----------------------------------------------------------------------------
//  Get the lock for the specified resource.
//
//  Returns:
//    If lock acquired:  depth of locks aquired by caller
//    On error:  < 0
//-----------------------------------------------------------------------------
int  getLock(uint id, ulong msTimeout)
{
  PID  mypid, owner;
  int  stat, count;
  task_t  *task;
  LockData  *lock;


  if (id >= NUM_LOCKS)  { errno = ENODEV;  return -1; }

  mypid = getpid();   lock = locks + id;

  _EnterCritical_();  owner = lock->owner;  _ExitCritical_();

//--- If we already own the lock, then just increment the count.  Since ---
//--- only the owner of the lock changes the count, we can safely incr ---
//--- and return the current value w/o worrying about preemption. ---
  if (owner == mypid)  { ++lock->count;  return lock->count; }

//--- If we aren't the owner, then try to get the lock.  If the previous ---
//--- owner died, then the LockData will be invalid but the semaphore for ---
//--- the lock will have been released so another task will be able to get ---
//--- it (and change the LockData values to indicate the new owner) ---
  stat = sem_wait_(&lock->sem, msToTicks(msTimeout));

  if (stat < 0)  {
    if (msTimeout > 100)  {
      _EnterCritical_();
      owner = lock->owner;  count = lock->count;
      _ExitCritical_();
      task = tasks + _activeTaskIdx_(owner);
      printf("Lock %d still held by %.15s[%d], count %d\r\n",
              id, task->argv[0], owner, count );
    }
    return stat;
  }


//--- Again, if we got here, we own the lock and only the owner will change ---
//--- the LockData values ---
  lock->owner = mypid;
  lock->count = 1;

//  fprintf(DiagLog, "%u now has lock %d\n", mypid, id);  //tdebug


  return lock->count;
}



//-----------------------------------------------------------------------------
//  If the calling process owns the specified lock, then reduce its lock count
//  (or release the lock if no more counts)
//
//  Returns:
//    On success:  New lock depth held by caller
//    On error:  < 0
//-----------------------------------------------------------------------------
int releaseLock(uint id)
{
  uint  owner;
  LockData  *lock;


  if (id >= NUM_LOCKS)  {
    fprintf(DiagLog, "*** releaseLock(%u): Invalid lock ID ***\n", id);
    return -1;
  }

  lock = locks + id;

  _EnterCritical_();  owner = lock->owner;  _ExitCritical_();

  if (owner != getpid())  {
    fprintf(DiagLog, "*** releaseLock(%u): Lock owned by: %u ***\n",
            id, owner);
    return -1;
  }

  // At this point, we know we own the lock, so no other task will make changes
  // to the values in the LockData structure until we release the semaphore

  if (lock->count < 1)  {
    fprintf(DiagLog, "*** releaseLock(%u) with count of %u ***\n",
            id, lock->count);
    return -1;
  }

  if (--lock->count == 0)  {
//    fprintf(DiagLog, "%u releasing lock %d\n",  getpid(), id);  //tdebug
    lock->owner = -1;  sem_post(&lock->sem);
  }

  return lock->count;
}


//-----------------------------------------------------------------------------
//  Get the APPL_LOCK and initialize the SPI port used to talk to the DSP
//
//  Returns lock depth on success, < 0 if an error
//-----------------------------------------------------------------------------
int getLockDSP(ulong timeOut)
{
  uchar  clockMode = NORM_LOW_LATCH_RISE;
  int  lockDepth;


  if (hdwType == HT_PHASE_CTLR)  if (!spi1Port)  { errno = EBADF; return -1; }

  lockDepth = getLock(APPL_LOCK, timeOut);
  if (lockDepth <= 0)  return lockDepth;

  if (hdwType == HT_PHASE_CTLR)  if (lockDepth == 1)
  //  ioctl(spi1fd, IOC_SET_CLK_MODE, &clockMode);
    _spiConfig_(spi1Port, CLK_2MHz, clockMode);

  return lockDepth;
}

//-----------------------------------------------------------------------------
//  release the APPL_LOCK and allow other tasks to use the DSP interface
//
//  Returns 0 on success, < 0 if an error
//-----------------------------------------------------------------------------
int releaseLockDSP(void)
{
  return releaseLock(APPL_LOCK);
}


//=============================================================================


//-----------------------------------------------------------------------------
//  Convert a new ModBus register value to a floating point number
//
//  Note that the size of the buffer pointed to by newVal must match the
//  size of the specified ModBus register value.
//-----------------------------------------------------------------------------
float mb2float(void *newVal, RegInfo *mbRegInfo)
{
  uchar  type = mbRegInfo->dataType;

  switch (type)  {
    case DT_S16:  return (float) MB_VAL(S16);
    case DT_U16:  return (float) MB_VAL(U16);
    case DT_S32:  return (float) MB_VAL(S32);
    case DT_U32:  return (float) MB_VAL(U32);
    case DT_F32:  return         MB_VAL(F32);
  }

  return 0.0;
}

//-----------------------------------------------------------------------------
//  Convert a new ModBus register value to a signed long number
//
//  Note that the size of the buffer pointed to by newVal must match the
//  size of the specified ModBus register value.
//-----------------------------------------------------------------------------
S32 mb2S32(void *newVal, RegInfo *mbRegInfo)
{
  uchar  type = mbRegInfo->dataType;

  switch (type)  {
    case DT_S16:  return (S32) MB_VAL(S16);
    case DT_U16:  return (S32) MB_VAL(U16);
    case DT_S32:  return       MB_VAL(S32);
    case DT_U32:  return (S32) MB_VAL(U32);
    case DT_F32:  return (S32) MB_VAL(F32);
  }

  return 0;
}

//-----------------------------------------------------------------------------
//  Convert a new ModBus register value to an unsigned long number
//
//  Note that the size of the buffer pointed to by newVal must match the
//  size for the type of the specified ModBus register value.
//-----------------------------------------------------------------------------
U32 mb2U32(void *newVal, RegInfo *mbRegInfo)
{
  uchar  type = mbRegInfo->dataType;

  switch (type)  {
    case DT_S16:  return (U32) MB_VAL(S16);
    case DT_U16:  return (U32) MB_VAL(U16);
    case DT_S32:  return (U32) MB_VAL(S32);
    case DT_U32:  return       MB_VAL(U32);
    case DT_F32:  return (U32) MB_VAL(F32);
  }

  return 0;
}

//-----------------------------------------------------------------------------
//  Convert a floating value to the format of the specified ModBus register
//
//  Note that the size of the buffer pointed to by newVal must match the
//  size of the specified ModBus register value.
//-----------------------------------------------------------------------------
void float2mb(void *newVal, RegInfo *mbRegInfo, float fromVal)
{
  uchar  type = mbRegInfo->dataType;

  switch (type)  {
    case DT_S16:  MB_VAL(S16) = (int) fromVal;  break;
    case DT_U16:  MB_VAL(U16) = (uint) fromVal;  break;
    case DT_S32:  MB_VAL(S32) = (long) fromVal;  break;
    case DT_U32:  MB_VAL(U32) = (ulong) fromVal;  break;
    case DT_F32:  MB_VAL(float) = fromVal;  break;
  }
}

//-----------------------------------------------------------------------------
//  Convert a DSP register value to a format that is compatible with the
//  specified ModBus register (i.e. make sure it is a floating point or
//  an integer value as needed).
//-----------------------------------------------------------------------------
void dsp2mb(RegInfo *mbReg, S16 *mbVal,
            const RD_DSP_REG far *dspReg, ulong *dspVal)
{
  uchar  mbType = mbReg->dataType, dspType = dspReg->type;
  bool  offset0, gain1, scaleIt;
  float  gain, offset, fval;


//--- if DSP value is a scaled PHASE value, rescale it into a float ---
  if (dspType == RT_PHASE)  {
    *(float *)dspVal = (float)*dspVal * PhaseConvFactor;  dspType = RT_F32; }


//--- determine if the value needs to be scaled ---
  offset = (dspReg->offset ? *dspReg->offset : 0.0);  // NULL == no offset
  gain = (dspReg->gain ? *dspReg->gain : 1.0);  // NULL == no gain

  if (gain == 0.0)  gain = 1.0;

  offset0 = (offset == 0.0);  gain1 = (gain == 1.0);

  scaleIt = (!offset0 OR !gain1);

  if (scaleIt)  {
    switch (dspType)  {
      case RT_S32:  fval = (float)((S32)*dspVal);  break;
      case RT_U32:  fval = (float)((U32)*dspVal);  break;
      case RT_F32:  fval = *(float *)dspVal;  break;
    }
    if (!offset0)  fval -= offset;
    if (!gain1)  fval = fval * gain;
    *(float *)dspVal = fval;  dspType = RT_F32;
  }


//--- Make sure the DSP format is the same type as the ModBus format ---
//--- (i.e. both are 16 or 32-bit integers or both are 32-bit floats) ---
  if ((dspType == DT_F32) != (mbType == DT_F32))  {
    if (mbType == DT_F32)  {  // non-float to float
      if (dspType == DT_S32)  {  // signed or unsigned?
        *(float *)dspVal = (float)((S32)*dspVal);  dspType = DT_F32;  }
      else  {
        *(float *)dspVal = (float)((U32)*dspVal);  dspType = DT_F32;  }
    }
    else  {  // float to non-float
      fval = *(float *)dspVal;
      if ((mbType == DT_S16) OR (mbType == DT_S32)) { // signed or unsigned?
        *(S32 *)dspVal = (S32) fval;  dspType = DT_S32; }
      else  {
        *(U32 *)dspVal = (U32) fval;  dspType = DT_U32; }
    }
  }

  if (mbReg->numRegs < 2)  // ModBus is 16-bit int, DSP is 32-bit int
    *(uint *)mbVal = (uint)(*dspVal);
  else  // both are 32-bits (either float or long ints)
    memcpy(mbVal, dspVal, 4);
}


//-----------------------------------------------------------------------------
//  Convert a ModBus register value to the specified DSP register format
//-----------------------------------------------------------------------------
void mb2dsp(RegInfo *mbReg, S16 *mbVal,
            const WR_DSP_REG far *dspReg, ulong *dspVal)
{
  uchar  dspType = dspReg->type;
  bool  offset0, gain1, scaleIt;
  float  gain, offset, fval;


//--- determine if the value needs to be scaled ---
  offset = (dspReg->offset ? *dspReg->offset : 0.0);  // NULL == no offset
  gain = (dspReg->gain ? *dspReg->gain : 1.0);  // NULL == no gain

  if (gain == 0.0)  gain = 1.0;

  offset0 = (offset == 0.0);  gain1 = (gain == 1.0);

  scaleIt = (!offset0 OR !gain1);

  if (scaleIt)  {
    fval = mb2float(mbVal, mbReg);
    if (!gain1)  fval = fval / gain;
    if (!offset0)  fval += offset;
  }

//--- convert either the scaled or unscaled ModBus value to the DSP format ---
  switch (dspType)  {
    case RT_S32:
      if (scaleIt)
        *(S32 *)dspVal = (S32)fval;
      else
        *(S32 *)dspVal = mb2S32(mbVal, mbReg);
      break;

    case RT_U32:
      if (scaleIt)
        *(U32 *)dspVal = (U32)fval;
      else
        *(U32 *)dspVal = mb2U32(mbVal, mbReg);
      break;

    case RT_F32:
      if (scaleIt)
        *(float *)dspVal = fval;
      else
        *(float *)dspVal = mb2float(mbVal, mbReg);
      break;

    case RT_PHASE:
      if (scaleIt)
        *dspVal = (ulong)(fval / PhaseConvFactor);
      else
        *dspVal = (ulong)(mb2float(mbVal, mbReg) / PhaseConvFactor);
      break;
  }

}


//=============================================================================
//  Functions referenced in the Modbus register definitions to handle writes
//  to specific global values defined as ModBus registers
//=============================================================================

//-----------------------------------------------------------------------------
//  Got a new value for one of the registers that mirrors one of the DSP's R/W
//  registers.  In response, we check the new value against the range of values
//  given for the DSP register and modify the ModBus register value if needed.
//
//  The writeMode specified for the DSP register determines when the new value
//  is written to the DSP as follows:
//
//  WM_ONCE:
//     The value must be written just once ASAP.  Writing to one of these DSP
//     registers triggers some action on the part of the DSP, so the register
//     must not be written to except in response to a write to the corresponding
//     ModBus register.  The most common use of this mode by the DSP is to
//     copy some related values to a table (possibly including the value just
//     written, or using the new value as an index in to the table?) as part
//     of a calibration process.
//
//  WM_EVERY:
//     The value represents some run-time setting and can be written to the DSP
//     repeatedly without causing a problem.  While writing such values to the
//     DSP just once would be sufficient if the the DSP never rebooted, such
//     values are not retained by the DSP across a reboot.  For that reason,
//     these values are resent to the DSP each pass through the application's
//     main loop (i.e. by the applic task).  The result is that if the DSP
//     restarts, or if a write should fail one time, the correct settings will
//     still be quickly (re)asserted.
//
//-----------------------------------------------------------------------------
bool MB_DSP_REG(SetRegFuncParams)
{
  uint  ui;
  RegInfo *mbRegInfo = (RegInfo *)info;
  ulong  dspVal;
  float  fval;
  const WR_DSP_REG far *dspReg = wrDSPregs;
  char  temp[32];


  for (ui=0; ui<NUM_WR_DSP_REGS; ++ui,++dspReg)  {
    if (dspReg->mbReg != mbRegInfo)  continue;

  //--- if a phase value, then make sure it is within the range 0 to 360 ---
    if (dspReg->type == RT_PHASE)  {
      fval = mb2float(newVal, mbRegInfo);
      if ((fval < 0.0) OR (fval >= 360.0))  {
        while (fval < 0.0)  fval += 360.0;
        while (fval >= 360.0)  fval -= 360.0;
        float2mb(newVal, mbRegInfo, fval);
      }
    }

  //--- check the new value against the specified range of allowed values ---
    if (dspReg->minVal != dspReg->maxVal)  {
      fval = mb2float(newVal, mbRegInfo);
      if (fval < dspReg->minVal)  float2mb(newVal, mbRegInfo, dspReg->minVal);
      if (fval > dspReg->maxVal)  float2mb(newVal, mbRegInfo, dspReg->maxVal);
    }

    if (mbRegInfo->flags & LogChanges)
      printf("new value for DSP reg: 0x%08lX: %.32s\r\n",
             dspReg->id, fmtRegVal(mbRegInfo, newVal, 0, temp) );

  //--- write the value to the DSP now? ---
    if (dspReg->writeMode == WM_ONCE)  {
      mb2dsp(mbRegInfo, newVal, dspReg, &dspVal);
      if (writeDSPReg(dspReg->id, dspVal, true) < 0)  return false;
    }

  //--- update the modbus register value using the new value ---
  // NOTE:  Can't use setMBReg() since that would just call this func again!
    switch (mbRegInfo->dataType)  {
      case DT_S16:  setS16(_SetRegFuncParams_);  break;
      case DT_U16:  setU16(_SetRegFuncParams_);  break;
      case DT_S32:  setS32(_SetRegFuncParams_);  break;
      case DT_U32:  setU32(_SetRegFuncParams_);  break;
      case DT_F32:  setF32(_SetRegFuncParams_);  break;
    }
    break;
  }

  return true;
}

//-----------------------------------------------------------------------------
//  process a new value for "cmdBitsAppl"
//-----------------------------------------------------------------------------
bool MB_cmdBitsAppl(SetRegFuncParams)
{
  ulong newFmt = 0;

  cmdBitsAppl = MB_VAL(ulong);

//--- This is the older version of the bit-mapped cmds, so we translate ---
//--- the best we can in to the newer ones and pass them on ---

  if (cmdBitsAppl & OLDC_AMPL_OPEN_LOOP)  newFmt |= CMD_AMPL_PID_OFF;
  if (cmdBitsAppl & OLDC_AMPL_CLOSED_LOOP)  newFmt |= CMD_AMPL_PID_ON;

  if (cmdBitsAppl & OLDC_PHASE_OPEN_LOOP)  newFmt |= CMD_PHASE_CTL_OFF;
  if (cmdBitsAppl & OLDC_PHASE_CLOSED_LOOP)  newFmt |= CMD_PHASE_CTL_PID;

  if (cmdBitsAppl & OLDC_AMPL_TUNE_RUN)  newFmt |= CMD_RF_PULSING_ON;
  if (cmdBitsAppl & OLDC_AMPL_TUNE_STOP)  newFmt |= CMD_RF_PULSING_OFF;
  if (cmdBitsAppl & OLDC_PHASE_TUNE_RUN)  newFmt |= CMD_RF_PULSING_ON;
  if (cmdBitsAppl & OLDC_PHASE_TUNE_STOP)  newFmt |= CMD_RF_PULSING_OFF;

  if (cmdBitsAppl & OLDC_CAV_TUNER_ON)  newFmt |= CMD_AFC_ON;
  if (cmdBitsAppl & OLDC_CAV_TUNER_OFF)  newFmt |= CMD_AFC_OFF;

  SetMBReg(setDSPState, 0, U32, newFmt);

  return true;
}

//-----------------------------------------------------------------------------
//  process a new value for "setAtten"
//-----------------------------------------------------------------------------
bool MB_setAtten(SetRegFuncParams)
{
  setAtten = MB_VAL(U16);

//  SetMBReg(fwdAtten, 0, U16, setAtten);
//  SetMBReg(reflAtten, 0, U16, setAtten);
//  SetMBReg(cavAtten, 0, U16, setAtten);

  return true;
}

//-----------------------------------------------------------------------------
//  process a new value for "setDSPState"
//-----------------------------------------------------------------------------
bool MB_setDSPState(SetRegFuncParams)
{
  uint  i;
  ulong  mask, cmdBits;


  cmdBits = MB_VAL(ulong);

//--- setDSPState contains cmd bits in groups:  i.e. one bit is asserted to ---
//--- enable a state, the other to disable it.  None set within a group ---
//--- means do not change the current state for the group. ---

//--- Writing of values to the DSP is handled primarily by the application ---
//--- task.  To avoid slowing down the task that called this function ---
//--- waiting for access to the DSP interface, we just keep track of the ---
//--- most recent setting for each group of bits.  That way the actual ---
//--- writing of the bits (whenever it happens) will always use the last ---
//--- setting for each group, and we have a record of what state the DSP ---
//--- should be in any given time (and a way to re-assert that state if ---
//--- the DSP is reset) ---

  for (i=0; SetDSPStateMasks[i]; ++i)  {
    mask = SetDSPStateMasks[i];
    if (cmdBits & mask) setDSPState = (setDSPState & ~mask) | (cmdBits & mask);
  }

//--- Also, log a message showing what bits were set each time we get a new ---
//--- value ---
  for (i=0,mask=1; i<32; ++i,mask<<=1)  if (cmdBits & mask)
    fprintf(DiagLog, "setDSPstate: Set: %ls\n", SetDSPStateDesc[i]);


  return true;
}

//-----------------------------------------------------------------------------
//  process a new value for "setAFCsOnOff"
//-----------------------------------------------------------------------------
bool MB_setAFCsOnOff(SetRegFuncParams)
{
  uint  i;
  ulong  mask, cmdBits;


  cmdBits = MB_VAL(ulong);

//--- setAFCState contains cmd bits in groups:  i.e. one bit is asserted to ---
//--- enable a state, the other to disable it.  None set within a group ---
//--- means do not change the current state for the group. ---

//--- Writing of values to the DSP is handled primarily by the application ---
//--- task.  To avoid slowing down the task that called this function ---
//--- waiting for access to the DSP interface, we just keep track of the ---
//--- most recent setting for each group of bits.  That way the actual ---
//--- writing of the bits (whenever it happens) will always use the last ---
//--- setting for each group, and we have a record of what state the DSP ---
//--- should be in any given time (and a way to re-assert that state if ---
//--- the DSP is reset) ---

  for (i=0,mask=3; i<10; ++i,mask<<=2)   if (cmdBits & mask)
    setAFCsOnOff = (setAFCsOnOff & ~mask) | (cmdBits & mask);

  for (i=0,mask=1; i<20; ++i,mask<<=1)
    if (cmdBits & mask)  fprintf(DiagLog, "Set: %ls\n", SetAFCsOnOffDesc[i]);

  return true;
}

//-----------------------------------------------------------------------------
//  process a new value for "cmdBitsAppl3"
//-----------------------------------------------------------------------------
bool MB_cmdBitsAppl3(SetRegFuncParams)
{
  uint  cmdBits;
  ulong  dummy;

//NOTE:  Save ONLY those bits that indicate an ongoing state.  Bits that
//       are a one time command (like resetting the DSP) must NOT be
//       saved after acting on them.

  cmdBits = (uint) MB_VAL(ulong);


  if (cmdBits & CMD_RESET_DSP)  {
    if (getLockDSP(5000) <= 0)  return false;
    resetDSP();
    releaseLockDSP();
  }

  if (hdwType != HT_PHASE_CTLR)  {  // no longer supported on actual hdw
    if (cmdBits & CMD_CALC_CTF)  {
      if (getLockDSP(5000) <= 0)  return false;
      if (calcStartTimeCTF == 0)  {
        fprintf(DiagLog, "Starting CTF calculation...\n");
        readDSPReg(CalcCTF, &dummy);
        calcStartTimeCTF = GetMS();  percentCTF = 0;
      }
      releaseLockDSP();
    }
  }

  return true;
}


//-----------------------------------------------------------------------------
//  process a new value for an amplitude value that needs to be checked
//  against maxAmplSet
//-----------------------------------------------------------------------------
bool MB_maxAmpl(SetRegFuncParams)
{
  if (REG->dataType != DT_F32)  return false;

  if (MB_VAL(F32) > maxAmplSet)  MB_VAL(F32) = maxAmplSet;

  return setF32(_SetRegFuncParams_);
}


//-----------------------------------------------------------------------------
//  process a new value for "oldPulseFreq"
//-----------------------------------------------------------------------------
bool MB_oldPulseFreq(SetRegFuncParams)
{
  float  newPulseFreq;


  oldPulseFreq = MB_VAL(U16);

  newPulseFreq = (float) oldPulseFreq;

  return SetMBReg(pulseFreq, 0, F32, newPulseFreq);
}


//-----------------------------------------------------------------------------
//  process a new value for "dspProgNum"
//-----------------------------------------------------------------------------
bool MB_dspProgNum(SetRegFuncParams)
{
  if (MB_VAL(uint) > 3)  return false;

  dspProgNum = MB_VAL(uint);  return true;
}

//-----------------------------------------------------------------------------
//  process a new value for pllInitIdx register
//-----------------------------------------------------------------------------
/*
bool MB_pllInitIdx(SetRegFuncParams)
{
  uint  ui = MB_VAL(uint);

  if (ui > PLLInitIdxMax)  ui = PLLInitIdxMax;

  _EnterCritical_();  pllInitIdx = ui;  _ExitCritical_();

  return true;
}
*/

//-----------------------------------------------------------------------------
//  process a new value for one of the xxxGain registers
//-----------------------------------------------------------------------------
bool MB_setGain(SetRegFuncParams)
{
  if (REG->dataType != DT_F32)  return false;

  if (MB_VAL(float) == 0.0)  return false;

  return setF32(_SetRegFuncParams_);
}


//-----------------------------------------------------------------------------
//  process a new value for a register with a phase value
//-----------------------------------------------------------------------------
bool MB_setPhase(SetRegFuncParams)
{
  if (REG->dataType != DT_F32)  return false;

  if (MB_VAL(float) < -720.0)  MB_VAL(float) = -360.0;
  if (MB_VAL(float) >  720.0)  MB_VAL(float) =  360.0;

  while (MB_VAL(float) < -360.0)  MB_VAL(float) += 360.0;
  while (MB_VAL(float) >  360.0)  MB_VAL(float) -= 360.0;

  return setF32(_SetRegFuncParams_);
}


//-----------------------------------------------------------------------------
//
//-----------------------------------------------------------------------------
int _nearcall PV_UpdateVal(PV_OP_Params)
{
  //finish:  Need to update pv->newEvents and pv->val for any non-waveform
  //         PVs that don't do it elsewhere

//  printf("PV_UpdateVal(%ls)\r\n", pv->pvInfo.name);  //tdebug

  return -1;
}

//-----------------------------------------------------------------------------
//
//-----------------------------------------------------------------------------
int _nearcall PV_LockVal(PV_Lock_Params)
{
  //finish:  Implement locking for non-waveform values

//tdebug
//  printf("\r\nPV_LockVal(%ls, %d, %d)\r\n", pv->pvInfo.name, lockMode, lockType);

  return 0;
}

//-----------------------------------------------------------------------------
//
//-----------------------------------------------------------------------------
int _nearcall PV_SendVal(PV_RW_Params)
{
  PV_Info  pvInfo;


  //finish:  call (*copyFunc)(void *conInfo, void far *from, uint count)
  //         as needed to send PV value

  if (!pvGetInfo(id, &pvInfo))  pvInfo.name = "<invalid>";

  printf("PV_SendVal(%ls)\r\n", pvInfo.name);  //tdebug

  return -1;
}

//-----------------------------------------------------------------------------
//
//-----------------------------------------------------------------------------
int _nearcall PV_RcvVal(PV_RW_Params)
{
  PV_Info  pvInfo;


  //finish:  call (*copyFunc)(void *conInfo, void far *from, uint count)
  //         as needed to get new PV value
  //         (probably need more info so we know how big a payload is in
  //         the message that caused this call)

  if (!pvGetInfo(id, &pvInfo))  pvInfo.name = "<invalid>";

  printf("PV_RcvVal(%ls)\r\n", pvInfo.name);  //tdebug

  return -1;
}


//-----------------------------------------------------------------------------
//  If newer data available, update the copy of a PV's value that is sent to
//  external clients (and update the newEvents flag indicating what events the
//  new data applies to).
//-----------------------------------------------------------------------------
int _nearcall PV_UpdateValWF(PV_OP_Params)
{
  bool noChange;
  int  stat;
  uint  src, cpy, numBytes, srcLockID, cpyLockID;
  char far  *srcBuf, *cpyBuf;


  if ((src = WF_IDX(pv->val) - 1) >= NUM_WAVEFORMS*2)  return -1;
  cpy = src + 1;

//--- Don't even TRY to get the locks unless we KNOW there is new data ---
  _EnterCritical_();
  noChange = (dataTimeMS[cpy] == dataTimeMS[src]);
  _ExitCritical_();
  if (noChange)  return 0;


  srcBuf = (char far *) waveformData[src];    srcLockID = WAVE_DATA_0 + src;
  cpyBuf = (char far *) waveformData[cpy];    cpyLockID = WAVE_DATA_0 + cpy;


//--- Get a lock on the dest/cpy buffer so it won't change while we are ---
//--- accessing (and possibly updating) it ---

  if (getLock(cpyLockID, 0) <= 0)  return -1;  // try to get lock on dest buffer

//--- If the timestamps for the source and destination buffers are ---
//--- different, then update the contents and timestamp for the ---
//--- destination buffer.  NOTE that we don't really care about the ---
//--- possibility of the src date/time changing during the check as ---
//--- all we really care about is whether or not it IS different. ---

  stat = 0;
  if (dataTimeMS[cpy] != dataTimeMS[src])  do {
    if (getLock(srcLockID, 0) <= 0)  { stat = -1;  break; }  // try to get src buf lock

    numBytes = numDataPoints[src]<<2;

    fmemcpy(cpyBuf, srcBuf, numBytes);
    numDataPoints[cpy] = numDataPoints[src];
    dataTimeSecs[cpy] = dataTimeSecs[src];
    dataTimeMS[cpy] = dataTimeMS[src];

    releaseLock(srcLockID);  // release lock on src buffer

    pv->pvInfo.size = numBytes;
    pv->pvInfo.numElems = numDataPoints[src];

    pv->newEvents |= DBE_VALUE;

    if (ShowDebugMsgs)  //tdebug
      printf("Updated copy of PV val: [%ls]\r\n", pv->pvInfo.name);  //tdebug
  } while (0);

  releaseLock(cpyLockID);

  return stat;
}


//-----------------------------------------------------------------------------
//  Attempt to obtain or release a read or write lock on a waveform PV.
//
//  Returns the depth of the lock or < 0 if not successsful.
//
//  For lockMode LM_TryLock, a return value < 0 and errno == EAGAIN means
//  it could not get the lock w/o blocking.
//-----------------------------------------------------------------------------
int _nearcall PV_LockWave(PV_Lock_Params)
{
  uint  src, cpy, lockID;


//tdebug
//  printf("\r\nLock PV val: [%ls]  mode:[%d]  type:[%d]\r\n",
//         pv->pvInfo.name, lockMode, lockType);

  if ((src = WF_IDX(pv->val) - 1) >= NUM_WAVEFORMS*2)  return -1;
  cpy = src + 1;

  switch (lockType)  {
    case LT_READ:  lockID = WAVE_DATA_0 + cpy;  break;
    case LT_WRITE:  lockID = WAVE_DATA_0 + src;  break;
    default:  return -1;
  }

  switch (lockMode)  {
    case LM_Unlock:    return releaseLock(lockID);
    case LM_WaitLock:  return getLock(lockID, 5000);
    case LM_TryLock:   return getLock(lockID, 0);
  }

  return -1;
}


//-----------------------------------------------------------------------------


#define SharedFuncs
#include <sharedAppl.c>



//=============================================================================
//  Misc Application-specific functions
//=============================================================================


//-----------------------------------------------------------------------------
//  phaseDiff:  compute difference between 2 phases, returning a value between
//  -180.0 and +180.0
//-----------------------------------------------------------------------------
float  phaseDiff(float ph1, float ph2)
{
  float diff = ph1 - ph2;

  if (diff < -180.0)
    diff += 360.0;
  else
    if (diff > 180.0)  diff -= 360.0;

  return diff;
}


//-----------------------------------------------------------------------------
//  "fitofl": Converts a decimal number in fixed point format to floating point
//-----------------------------------------------------------------------------
float fitofl(long value)
{
  float out;
  bool neg = value < 0;
  ulong *p = (ulong *)&value;

  if (neg)  value *= -1;

//finish
  out = (*p) & 0x0000FFFFL;
  out /= 65536.0;
  out += (float) ((*p) >> 16);

  if (neg)  out *= -1.0;

  return out;
}

//-----------------------------------------------------------------------------
//  "fltofi": Converts a decimal number in floating point format to fixed point
//-----------------------------------------------------------------------------
long fltofi(float value)
{
  long fixed;
  bool neg = (value < 0.0);

  if (neg)  value *= -1.0;

//finish
  fixed = (long) value;
  value -= (float) fixed;
  fixed <<= 16;
  fixed += (long) (65536.0 * value);

  if (neg)  fixed *= -1;

  return fixed;
}


//=============================================================================
typedef struct {
  fd_t   fromfd;
  bool   eof;
}  ReadState;

#define RDSTATE ((ReadState *)rdState)

typedef struct {
  uint   progNum;     // 0 - 3
  uint   blkNum;      // 0 - 15
  ulong  ttlWritten;  // ttl bytes written to DSP mem for this program
  ulong  blkWritten;  // # bytes in current block written so far
  bool   showProgress;
}  WriteState;

#define WRSTATE ((WriteState *)wrState)

#define  DSP_BLK_SIZE  (0x10000L)


//-----------------------------------------------------------------------------
//  Read the sepcified # of bytes from a file descriptor.  If the end of the
//  file is reached, bytes with 0 value are returned instead.
//
//  Parameters:
//    *rdState:  Pointer to ReadState structure that contains the following:
//                 fromfd:  the file descriptor to read from
//                 eof:  flag to indicate that the End-Of-File was reached
//    *buf:      pointer to buffer to read the data in to
//    count:     # of bytes to read (or fill with zeros)
//
//  This function is designed to be called by the copyBytes() function which
//  provides a generic means of managing the transfer of a stream of bytes.
//
//  On return:
//    rdState values updated as needed
//    Returns the # of bytes successfully read or set to 0
//-----------------------------------------------------------------------------
int _nearcall readFromFile(void *rdState, char *buf, int count)
{
  int  stat;

  if (RDSTATE->eof)  { memset(buf, 0xff, count);  return count; }

  if ((stat = read(RDSTATE->fromfd, buf, count)) < 0)  return stat;

  if (stat < count)  { memset(buf+stat, 0, count-stat);  RDSTATE->eof = true; }

  return count;
}

//-----------------------------------------------------------------------------
//  Write bytes to the DSP program memory
//
//  Parameters:
//    *wrState:  pointer to WriteState structure that contains the following:
//                 progNum:  Which program # the block belongs to (0-3)
//                 blkNum:   64K block # within the program (0-15)
//                 ttlWritten:  ttl # bytes written so far
//                 blkWritten:  # bytes written to blkNum block so far
//    *buf:      pointer to buffer with data to be written
//    count:     # of bytes to be processed
//
//  This function is designed to be called by the copyBytes() function which
//  provides a generic means of managing the transfer of a stream of bytes.
//
//  NOTE:  count MUST be a multiple of 4!
//
//  On return:
//    wrState values are updated as needed
//    Returns the # of bytes successfully written
//-----------------------------------------------------------------------------
int _nearcall writeDSPMem(void *wrState, char *buf, int count)
{
  int  i, stat, remain, wrCount;
  uint  memBlock;
  uchar  enableLow, enableHigh;
  ulong  startTime, wordVal, percentDone;
  ulong  *ptr;


  if (NotLockOwner(APPL_LOCK))  return -1;

  ptr = (ulong *)buf;  remain = count;


  while (remain)  {
    resetWDT();

  //--- prep DSP for writing next block num as needed ---
    if (WRSTATE->blkWritten == 0)  {
      readDSPReg(WriteStatus, &percentDone);
      if (percentDone != 100)
        printf("DSP reports %ld%% before start of block!\r\n", percentDone);
      printf("Program: %d  Block: %d\r\n", WRSTATE->progNum, WRSTATE->blkNum);
      memBlock = (WRSTATE->progNum << 14) | (WRSTATE->blkNum & 0x0f);
      if (writeDSPRegQuick(WrMemBlock, memBlock))  break;
    }

  //--- write as many bytes as we can in to current block ---
    wrCount = (int) min((ulong)remain, DSP_BLK_SIZE - WRSTATE->blkWritten);

    enableLow = PGDRShadow & (uchar)(~_DSP_WR_RD);
    enableHigh = enableLow | (uchar)_DSP_WR_RD;

    for (i=0; i<wrCount; i+=4,++ptr)  {
      OUTI(PGDR, enableLow);  // signal start of 32-bit packet
      wordVal = htonl(*ptr);
      stat = WriteSPI1((char *)&wordVal, 4);
      OUTI(PGDR, enableHigh);  // signal end of 32-bit packet
      if (stat != 4)  return -1;
    }

    WRSTATE->blkWritten += (uint) wrCount;
    WRSTATE->ttlWritten += (uint) wrCount;
    remain -= (uint) wrCount;

  //--- if finished a block, wait for the DSP to indicate it has copied it ---
  //--- to flash and is ready to accept the next block ---
    if (WRSTATE->blkWritten == DSP_BLK_SIZE)  {
      startTime = GetMS();  percentDone = 0;
      while ((msSince(startTime) < 30000) AND (percentDone != 100))  {
        resetWDT();  sleepMS(100);  readDSPReg(WriteStatus, &percentDone);
        if (WRSTATE->showProgress)  {
          printf("%3ld%%      \r", percentDone);  fflush(stdout);  }
      }
      if (percentDone != 100)  {
        printf("DSP not responding: Only %3d%% saved to flash\r\n", percentDone);
        resetDSP();  errno = EIO;  return -1;
      }
      ++WRSTATE->blkNum;  WRSTATE->blkWritten = 0;
    }

  }

  return count - remain;
}

//-----------------------------------------------------------------------------
//  Initialize the DSP by copying the specified # of bytes from the source fd
//  to the DSP
//
//  Returns 0 on success, < 0 if an error occured (with errno set)
//-----------------------------------------------------------------------------
int initDSP(fd_t fromfd, uint progNum, ulong ttlBytes, bool showProgress)
{
  uchar  clockMode = NORM_LOW_LATCH_RISE;  //test
  ReadState  readState;
  WriteState  writeState;
  char  buf[256];


//  if (!spi1Port)  return -1;

  resetWDT();  // so WatchDog timer doesn't reboot us...

  ttlBytes = ttlBytes + DSP_BLK_SIZE - 1;
  ttlBytes = ttlBytes - (ttlBytes % DSP_BLK_SIZE);

  printf("Copying %lu bytes to DSP\r\n", ttlBytes);

//--- make sure the SPI port is not in use by another process ---
  if (getLockDSP(5000) <= 0)  return -1;

  readState.fromfd = fromfd;
  readState.eof = false;

  writeState.progNum = progNum;
  writeState.blkNum = 0;
  writeState.blkWritten = writeState.ttlWritten = 0;
  writeState.showProgress = showProgress;

  if (hdwType == HT_PHASE_CTLR)
  //  ioctl(spi1fd, IOC_SET_CLK_MODE, &clockMode);
    _spiConfig_(spi1Port, CLK_1MHz, clockMode);  //test

  if (copyBytes(readFromFile, &readState, writeDSPMem, &writeState,
                buf, sizeof(buf), ttlBytes, showProgress) != ttlBytes)
    perror("copyBytes");
  else
    printf("Copied %lu bytes to DSP\r\n", writeState.ttlWritten);

  close(fromfd);

  releaseLockDSP();

  return 0;
}

//-----------------------------------------------------------------------------
//  Wait for FPGA to release SPI bus
//-----------------------------------------------------------------------------
bool waitFPGA(ulong waitMS)
{
  ulong  startTime;

  startTime = GetMS();
  for (;;)  {
    resetWDT();  // so WatchDog timer doesn't reboot us...
//if (diagLogInfo->wdtTicks > 10)  //tdebug
//  printf("=== %lu %u\r\n", msSince(startTime), diagLogInfo->wdtTicks);  //tdebug
    digitalIn = ReadDigInputs();
    if (BOOL(digitalIn & FPGA_DONE))  return true;
    if (msSince(startTime) > waitMS)  break;
    sleepMS(250);
  }

  if (hdwType != HT_PHASE_CTLR)  return true;

  printf("FPGA never released SPI bus\r\n");

  return false;
}

//-----------------------------------------------------------------------------
//  Reset the FPGA, causing it to reload its firmware from the EEPROM
//-----------------------------------------------------------------------------
void _task resetFPGA(int argc, char **argv)
{
  if (hdwType != HT_PHASE_CTLR)  Exit(-1);

  if (getLock(APPL_LOCK, 10000) <= 0)  Exit(-1);

//--- tell FPGA to release SPI bus ---
  KLOCK();
  ClearBits(Next.digitalOut, FPGA_SPI);  digitalOut = Next.digitalOut;
  fpgaInitialized = false;
  KUNLOCK();
  WriteDig();

//--- allow FPGA to have the SPI bus again.  This also causes it to reload ---
//--- from the new image in EEPROM (good or bad) ---
  KLOCK();
  SetBits(Next.digitalOut, FPGA_SPI);  digitalOut = Next.digitalOut;
  KUNLOCK();
  WriteDig();

  releaseLock(APPL_LOCK);

//--- wait for FPGA to load new firmware and release SPI bus ---
  fpgaInitialized = waitFPGA(3000);

  if (fpgaInitialized)  { printf("FPGA firmware reloaded.\r\n");  Exit(0); }

  Exit(-1);
}


//-----------------------------------------------------------------------------
//  write and then read specified # of bytes to/from the EEPROM
//-----------------------------------------------------------------------------
char *writeReadEE(char *send, int sendCount, char *rcv, int rcvCount)
{
  EEPROM_Enable();

  WriteSPI2(send, sendCount);
  if (rcvCount)  ReadSPI2(rcv, rcvCount);

  EEPROM_Disable();

  return rcv;
}

//-----------------------------------------------------------------------------
//  write a single byte command to the EEPROM, and read and return the
//  value of the status register
//-----------------------------------------------------------------------------
char writeCmdEE(uchar cmd)
{
  char  readStatCmd, statReg;

  if (cmd)  writeReadEE(&cmd, 1, NULL, 0);

  readStatCmd = EEPROM_RDSR;  writeReadEE(&readStatCmd, 1, &statReg, 1);

  return statReg;
}

//-----------------------------------------------------------------------------
const char SectorAddr[11][3] = {
  { 0x00,0x00,0x00 },  //  0
  { 0x01,0x00,0x00 },  //  1
  { 0x02,0x00,0x00 },  //  2
  { 0x03,0x00,0x00 },  //  3
  { 0x04,0x00,0x00 },  //  4
  { 0x05,0x00,0x00 },  //  5
  { 0x06,0x00,0x00 },  //  6
  { 0x07,0x00,0x00 },  //  7
  { 0x07,0x80,0x00 },  //  8
  { 0x07,0xA0,0x00 },  //  9
  { 0x07,0xC0,0x00 }   // 10
};

//-----------------------------------------------------------------------------
//  Update the FPGA firmware image in the EEPROM
//  Returns 0 on success, < 0 if an error occured (with errno set)
//-----------------------------------------------------------------------------
int updateFPGAfirmware(fd_t fromfd, bool showProgress, bool verify)
{
#define  BufSize  256  // can't be very large without overflowing the stack!

  uchar  clockMode = NORM_LOW_LATCH_RISE;
  bool  downloadOK;
  int  i, count;
  ulong  startTime, ttlBytes, fileSize;
  bool  busy;
  char  statReg;
  struct stat  fsInfo;
  char  cmdBuf[5];
  char  buf[BufSize], buf2[BufSize];


  if (hdwType != HT_PHASE_CTLR)  return 0;

  if (!spi2Port)  return -1;

  if (getLock(APPL_LOCK, 10000) <= 0)  return -1;

//  ioctl(spi2fd, IOC_SET_CLK_MODE, &clockMode);
  _spiConfig_(spi2Port, CLK_1MHz, clockMode);

//--- tell FPGA to release SPI bus ---
  KLOCK();
  ClearBits(Next.digitalOut, FPGA_SPI);  digitalOut = Next.digitalOut;
  KUNLOCK();
  WriteDig();

  fpgaInitialized = downloadOK = false;

//--- make sure EEPROM is ready to accept commands ---
  EEPROM_Disable();  writeCmdEE(EEPROM_WRDI);

//===== erase EEPROM and copy bytes from the file in to it =====
  do {
  //--- read device ID ---
    *cmdBuf = EEPROM_RDID;  writeReadEE(cmdBuf, 1, buf2, 4);
    printf("  Device ID: %02X %02X %02X %02X\r\n",
           buf2[0], buf2[1], buf2[2], buf2[3] );

  //--- unprotect all sectors ---
    printf("  Unprotecting all sectors...\r\n");
    *cmdBuf = EEPROM_UNPR;
    for (i=0; i<11; ++i)  {
      if (! (writeCmdEE(EEPROM_WREN) & 0x02))  break;
      memcpy(cmdBuf+1, SectorAddr[i], 3);  writeReadEE(cmdBuf, 4, NULL, 0);
    }
    if (i != 11)  break;

  //--- erase the EEPROM ---
    printf("  Erasing EEPROM");  *buf = EEPROM_RDSR;

    if (! (writeCmdEE(EEPROM_WREN) & 0x02))  break;
    if (! (statReg = writeCmdEE(EEPROM_ERASE_ALL) & 0x01))  {
      printf("\r\nUnable to initiate Chip Erase: Status: 0x%02X\r\n", statReg);
      break;
    }

    startTime = GetMS();
    while (msSince(startTime) < 12000)  {
      if (! (busy = writeCmdEE(0) & 0x01))  break;
      printf(".");  fflush(stdout);  sleepMS(500);
    }
    printf("\r\n");
    if (busy)  { printf("Erase timed out\r\n");  break; }

    if (fstat(fromfd, &fsInfo) < 0)  { perror(NULL);  Exit(-2); }
    fileSize = (ulong)fsInfo.st_size;
    printf("  File size: %lu bytes\r\n", fileSize);


  //--- copy the firmware file from the source fd to the EEPROM ---
    printf("  Copying file to EEPROM...\r\n");
    ttlBytes = 0;
    if (! (writeCmdEE(EEPROM_WREN) & 0x02))  break;
    memset(cmdBuf, 0, sizeof(cmdBuf));  cmdBuf[0] = EEPROM_SEQ_PROG;
    while ((count = read(fromfd, buf, sizeof(buf))) > 0)  {
      for (i=0; i<count; ++i)  {
        if (! ttlBytes)  {  // 1st byte requires cmd code and 3 addr bytes
          cmdBuf[4] = *buf;  writeReadEE(cmdBuf, 5, NULL, 0);  }
        else  {  // subsequent bytes require only cmd code
          cmdBuf[1] = buf[i];  writeReadEE(cmdBuf, 2, NULL, 0);  }
//        busy = writeCmdEE(0) & 0x01;  if (busy)  break;
        ++ttlBytes;
      }
      if (busy)  break;
      if (showProgress)  { printf("%10lu\r", ttlBytes);  fflush(stdout); }
    }
    writeCmdEE(EEPROM_WRDI);
    if (busy OR (ttlBytes != fileSize))
      printf("Error writing to EEPROM\r\n");
    else
      if (showProgress)  printf("  Wrote %lu bytes to EEPROM\r\n", ttlBytes);

  //--- protect all sectors ---
    printf("  Protecting all sectors...\r\n");
    *cmdBuf = EEPROM_PROT;
    for (i=0; i<11; ++i)  {
      if (! (writeCmdEE(EEPROM_WREN) & 0x02))  break;
      memcpy(cmdBuf+1, SectorAddr[i], 3);  writeReadEE(cmdBuf, 4, NULL, 0);
    }



  //--- verify contents of EEPROM ---
    if (verify)  {
      printf("  Verifying EEPROM copy...\r\n");

      if (lseek(fromfd, 0, SEEK_SET) != 0)  {
        printf("Unable to reset source file pos to verify EEPROM\r\n");  break; }

      memset(cmdBuf, 0, sizeof(cmdBuf));  cmdBuf[0] = EEPROM_READ;
      EEPROM_Enable();  WriteSPI2(cmdBuf, 4);

      ttlBytes = 0;

      while ((count = read(fromfd, buf, sizeof(buf))) > 0)  {
        ReadSPI2(buf2, count);
        for (i=0; i<count; ++i)  {
//          if (!((uchar)ttlBytes & 0x0f))  printf("\r\n%06lX:", ttlBytes);
//          printf(" %02X", buf2[i]);
          if (buf[i] != buf2[i])  {
            printf("Mismatch at offset %lu:  EEPROM: 0x%02X,  File: 0x%02X\r\n",
                   ttlBytes, buf[i], buf2[i]);
            break;
          }
          if (++ttlBytes >= fileSize)  break;  //tdebug
        }
        if (showProgress)  { printf("%10lu\r", ttlBytes);  fflush(stdout); }
        if (ttlBytes >= fileSize)  break;  //tdebug
      }
      printf("\r\n");
      EEPROM_Disable();

      if (ttlBytes == fileSize)  {
        downloadOK = true;  printf("Verified %lu bytes\r\n", ttlBytes);  }
      else
        printf("Copies do not match!\r\n");
    }

  } while(0);


//--- allow FPGA to have the SPI bus again.  This also causes it to reload ---
//--- from the new image in EEPROM (good or bad) ---
  KLOCK();
  SetBits(Next.digitalOut, FPGA_SPI);  digitalOut = Next.digitalOut;
  KUNLOCK();
  WriteDig();

  releaseLock(APPL_LOCK);

  if (downloadOK)  printf("FPGA firmware download complete!\r\n");

//--- wait for FPGA to load new firmware and release SPI bus ---
  fpgaInitialized = waitFPGA(3000);

  return 0;
}


//-----------------------------------------------------------------------------
//  Write and Read a group of long integers to/from the DSP
//
//  The counts are the # of long (32-bit) words to send/rcv
//
//  Returns 0 on success, < 0 if an error
// ---------------------------------------------------------------------------
//  Exchange of data with the DSP is done using a bi-directional (simultaneous
//  send/recv) clocked SPI interface.  All data is exchanged in 32-bit (4-byte)
//  packets.  While we are sending a 32-bit packet to the DSP, it is using the
//  same clock pulses to simultaneously send a 32-bit packet in response.
//
//  Each exchange of a 32-bit packet consists of setting the DSP /Enable signal
//  low, writing (while simultaneously reading) 32-bits of data, and then
//  setting the DSP /Enable signal high.
//
//  Due to the way the DSP buffers are set up, all responses from the DSP are
//  always 2 packets "behind" the ones that contained the commands that they
//  are a response to.  We can write as many commands as we want in a
//  sequence of packets, but the response to each one will always be received
//  2 packets later.  Given that we generally want to read the response to
//  a set of commands as soon as possible, we normally have to add some
//  dummy or filler packets to the sequence to get the desired response
//  packets.  For extended responses (like reading of waveform data), we must
//  send lots of filler packets to get all the data.
//-----------------------------------------------------------------------------

extern const uchar RevBitTable[];

int readPlot(ulong regNum, ulong far *rcvBuf, uint rcvCount)
{
  uchar  enableLow, enableHigh;
  uint  u;
  ulong  xmitWord, rcvWord;
//  uchar *ptr;


  if (hdwType != HT_PHASE_CTLR)  return 0;

  if (!spi1Port)  return -1;

  if (getLockDSP(1000) <= 0)  return -1;

  enableLow = PGDRShadow & (uchar)(~_DSP_WR_RD);
  enableHigh = enableLow | (uchar)_DSP_WR_RD;

  xmitWord = htonl(regNum);

  if (_spiLock_(spi1Port))  { releaseLockDSP();  return -1; }

/*
  NOTE:  This takes about 80 ms for a 1024 point (4096 byte) waveform,
         but about 50 ms of that is just to do what has to be done for
         each byte:  reverse the bit of the byte being sent, tell the
         CPU to send then rcv a byte, wait for the rcv to finish, and
         reverse the bit order of the rcvd byte.

         Minimzing the effect of having to save the whole result to a
         far pointer by using a near buf in root of about 512 bytes
         (and then copying it each time it is filled to the far buffer
         using fmemcpy) shaved off only about 4 ms.

         One additional (possibly significant) change might be to
         reverse the bits of the bytes in xmitWord before starting
         the loop, and using a modified vers of the _spiWrRdFast_()
         function that doesn't do this for the transmitted byte.

         But keep in mind that, at the 2MHz clock rate we are using,
         even if we could eliminate all pauses (such as the one that
         HAS to be there to toggle the bit that indicates the start
         and end of each 4-byte packet), sending and receiving 4096
         bytes would take over 16 ms.

           NOTE:  At 2MHz SPI rate and 44MHz CPU clock rate, each
                  8-bit write/read takes 8 * 44 = 176 CPU clock
                  cycles!  That's a significant # of instructions.

                  Any way I could overlap prep for the next byte
                  while waiting to read the result of the prev one?
                  THAT could reduce overhead quite a bit (NOTE:  An
                  interrupt-driven approach could help with this).

         Given that we DO have to toggle the start/end signal each
         4 bytes, and we DO have to reverse the bits for each byte
         received, it seems unlikely that we can reduce that 50 ms
         by much.

            One thought:  See how much of that 50ms is the bit reversing
            by reversing the sent bytes before we start, and disabling
            the reversing in just the FAST version (so as not to mess up
            anything but the waveform data).  If the reduction in time
            is significant enough, I could put off the reversing of the
            bits until the next pass.  At the very least, this would even
            out the CPU time usage and reduce the max time the applic
            task blocks everything else.

*/

//--- pre-reverse the bits in the xmitWord ---
//  ptr = (char *)&xmitWord;
//  for (u=0; u<4; ++u,++ptr)  *ptr = RevBitTable[*ptr];


//--- "prime the pump":  Ignore the first group of readings ---
  for (u=0; u<RespOffset; ++u)  {
    OUTI(PGDR, enableLow);  // signal start of 32-bit packet
    _spiWrRdFast_((char *)&xmitWord, (char *)&rcvWord, 4);
    OUTI(PGDR, enableHigh);  // signal end of 32-bit packet
  }
//--- save the rest of the readings ---
  for (u=0; u<rcvCount; ++u)  {
    OUTI(PGDR, enableLow);  // signal start of 32-bit packet
    _spiWrRdFast_((char *)&xmitWord, (char *)&rcvWord, 4);
    OUTI(PGDR, enableHigh);  // signal end of 32-bit packet
    *rcvBuf = rcvWord;  ++rcvBuf;
  }


  _spiUnlock_(spi1Port);

  releaseLockDSP();

  return 0;
}

//-----------------------------------------------------------------------------
//  For testing purposes:  Replaces actual read of a DSP register value for
//  hdw platforms other than the actual Phase Controller.
//-----------------------------------------------------------------------------
ulong fakeRegVal(ulong regNum)
{
  uint  r;
  ulong  rcvWord;
  F32  fval, percNoise;
  const RD_DSP_REG far *dspReg;


  for (r=0,dspReg=rdDSPregs; r<NUM_RD_DSP_REGS; ++r,++dspReg)  {
    if (dspReg->id != regNum)  continue;

    rcvWord = dspReg->typical;
    if (!dspReg->percNoise)  return htonl(rcvWord);

    percNoise = (F32)dspReg->percNoise * (F32)rand() / (F32)RAND_MAX / 100.0;

//    rand()  RAND_MAX

    switch (dspReg->type)  {

      case RT_S32:
        rcvWord = (U32)((S32)rcvWord + (S32)(percNoise * rcvWord));
        break;

      case RT_U32:
        rcvWord += (U32)(percNoise * rcvWord);
        break;

      case RT_F32:
        fval = *(F32 *)&rcvWord;
        fval += percNoise * fval;
        rcvWord = *(ulong *)&fval;
        break;

      case RT_PHASE:
        rcvWord += (U32)(percNoise * rcvWord);
        break;
    }

    return htonl(rcvWord);
  }

//  *((float *)&rcvWord) = 0.1234 + (float)(uint)(GetMS_()>>5 & 0x3F);
//  return htonl(rcvWord);

  return 0xFFFFFFFFL;
}

//-----------------------------------------------------------------------------
//  Read the values for the RO DSP registers and update the value of the
//  corresponding ModBus registers.
//
//  Returns 0 on success, < 0 if an error
// ---------------------------------------------------------------------------
//  See WrRdDSP() above for details on how the DSP interface works
//-----------------------------------------------------------------------------
int read_DSP_Regs(void)
{
  uchar  enableLow, enableHigh;
  uint  i, wordCount;
  ulong  xmitWord, dspVal;
  RegInfo *mbReg;
  const RD_DSP_REG far *reqReg;
  const RD_DSP_REG far *dspReg;


  if (getLockDSP(1000) <= 0)  return -1;

  wordCount = NUM_RD_DSP_REGS + RespOffset;

  enableLow = PGDRShadow & (uchar)(~_DSP_WR_RD);
  enableHigh = enableLow | (uchar)_DSP_WR_RD;

  for (i=0,reqReg=rdDSPregs; i<wordCount; ++i,++reqReg)  {

    xmitWord = (i < NUM_RD_DSP_REGS) ? htonl(reqReg->id) : 0;

    if (hdwType != HT_PHASE_CTLR)
      dspVal = (i < RespOffset) ? 0 : fakeRegVal((reqReg-RespOffset)->id);
    else  {
      OUTI(PGDR, enableLow);  // signal start of 32-bit packet
      _spiWrRd_(spi1Port, (char *)&xmitWord, (char *)&dspVal, 4);
      OUTI(PGDR, enableHigh);  // signal end of 32-bit packet
    }

    if (i < RespOffset)  continue;  // responses are offset by RespOffset pkts

    dspReg = reqReg - RespOffset;

  //--- convert value as needed and update the ModBus register value ---
    mbReg = dspReg->mbReg;  dspVal = ntohl(dspVal);
    dsp2mb(mbReg, mbReg->regVal, dspReg, &dspVal);
  }

  releaseLockDSP();

  return 0;
}

//-----------------------------------------------------------------------------
//  Send the current values to the DSP for each DSP R/W register that
//  represents a run-time setting (i.e. onces with writeMode == WM_EVERY).
//
//  Returns 0 on success, < 0 if an error
// ---------------------------------------------------------------------------
//  See WrRdDSP() above for details on how the DSP interface works
//-----------------------------------------------------------------------------
int write_DSP_Regs(void)
{
  uchar  enableLow, enableHigh;
  uint  i;
  ulong  xmitWord, rcvWord, dspVal, regNum;
  const WR_DSP_REG far *dspReg;
  RegInfo  *mbReg;


  if (hdwType == HT_PHASE_CTLR)  if (!spi1Port)  return -1;

  if (getLockDSP(1000) <= 0)  return -2;

  enableLow = PGDRShadow & (uchar)(~_DSP_WR_RD);
  enableHigh = enableLow | (uchar)_DSP_WR_RD;

  for (i=0,dspReg=wrDSPregs; i<NUM_WR_DSP_REGS; ++i,++dspReg)  {
    if (dspReg->writeMode != WM_EVERY)  continue;

    regNum = dspReg->id | 1L;

  //--- convert ModBus value to DSP format ---
    mbReg = dspReg->mbReg;
    mb2dsp(mbReg, mbReg->regVal, dspReg, &dspVal);

  //--- send the register # in the 1st packet ---
    xmitWord = htonl(regNum);
    OUTI(PGDR, enableLow);  // signal start of 32-bit packet
    _spiWrRd_(spi1Port, (char *)&xmitWord, (char *)&rcvWord, 4);
    OUTI(PGDR, enableHigh);  // signal end of 32-bit packet

  //--- send the value in the 2nd packet ---
    xmitWord = htonl(dspVal);
    OUTI(PGDR, enableLow);  // signal start of 32-bit packet
    _spiWrRd_(spi1Port, (char *)&xmitWord, (char *)&rcvWord, 4);
    OUTI(PGDR, enableHigh);  // signal end of 32-bit packet
  }

  releaseLockDSP();

  return 0;
}


//-----------------------------------------------------------------------------
//  write an 8-bit value to the FPGA
//
//  Returns 0 on success, -1 if an error
//-----------------------------------------------------------------------------
int writeFPGAParallel(uchar id, uchar val)
{
  uchar  i;
  int  stat;
  ulong  lval, valBits;


  if (getLock(APPL_LOCK, 5000) <= 0)  return -1;

  stat = -1;

  lval = (ulong)val << 8;
  valBits = ((lval << 6) & 0x300000L) | (lval & 0x003f00L);

//--- place value on "bus" ---
  digitalOut = ((digitalOut & 0xcfc0f0L) | valBits) | ((uint)id & 0x0f);
  WriteDig();

//--- latch the bits into the FPGA ---
  digitalOut |= FPGA_CL1;  WriteDig();

//--- short wait for the FPGA to acknowledge the new value ---
  for (i=0; i<10; ++i)  {
    ReadDig();
    if (digitalIn & FPGA_CL2)  { stat = 0;  break; }
  }

//--- reset FPGA "clock" signal ---
  digitalOut &= ~FPGA_CL1;  WriteDig();

  releaseLock(APPL_LOCK);

  if (stat)
    printf("FPGA did not acknowledge write of param id %d\r\n", (uint)id);

  return stat;
}


//-----------------------------------------------------------------------------
//  write bytes to the PLL via a serial port in SPI mode
//  NOTE:  All data sent out the same port in SPI mode will get clocked in to
//         the PLL, but the PLL will not do anything with it unless the
//         PLL's Load Enable signal transitions from high to low
//-----------------------------------------------------------------------------
int writePLL(char *buf, uint len)
{
  uchar  clockMode = NORM_LOW_LATCH_RISE;


  if (hdwType != HT_PHASE_CTLR)  return 0;

  if (!spi1Port)  return -1;

  if (getLock(APPL_LOCK, 5000) <= 0)  return -1;

//  ioctl(spi1fd, IOC_SET_CLK_MODE, &clockMode);
  _spiConfig_(spi1Port, CLK_1MHz, clockMode);

  WriteSPI1(buf, len);  // clock bytes in to PLL

//--- set PLL Load Enable pin to load the internal initialization register ---
  digitalOut |= PLL_LE;  WriteDig();

//--- reset PLL Load Enable pin to prepare for the next register load ---
  digitalOut &= ~PLL_LE;  WriteDig();

  releaseLock(APPL_LOCK);

  return 0;
}


//-----------------------------------------------------------------------------
//  set one of the registers in the high-speed ADC's to the specified value
//  via a serial port in SPI mode
//-----------------------------------------------------------------------------
int write_ADCReg(uint regNum, uint regVal)
{
  uchar  clockMode = NORM_HIGH_LATCH_FALL;
  char  buf[2];


  if (!spi1Port)  return -1;

  if (getLock(APPL_LOCK, 5000) <= 0)  return -1;

  buf[0] = (char) ((regNum << 3) | ((regVal >> 8) & 0x07));
  buf[1] = (char) regVal;

//  ioctl(spi1fd, IOC_SET_CLK_MODE, &clockMode);
  _spiConfig_(spi1Port, CLK_1MHz, clockMode);

  digitalOut &= ~_ADC_CS;  WriteDig();  // set ADC /CS low

  WriteSPI1(buf, 2);  //write 2 bytes to ADC's

  digitalOut |= _ADC_CS;  WriteDig();  // set ADC /CS high

  releaseLock(APPL_LOCK);

  return 0;
}


//-----------------------------------------------------------------------------
int write_Config(uint newConfig)
{
  char  buf[3];

  buf[0] = 0x1f;
  buf[1] = 0x80;
  buf[2] = (0x70 & newConfig << 4) | 0x82;

  return  writePLL(buf, 3);
}

//-----------------------------------------------------------------------------
//  Read a single 32-bit register value from the DSP
//
//  Returns 0 and sets *regVal on success, < 0 if an error
//-----------------------------------------------------------------------------
int readDSPReg(ulong regNum, ulong *regVal)
{
  uchar  enableLow, enableHigh;
  uint  i, wordCount;
  ulong  xmitWord, rcvWord;  //, respRegNum;
//  char  temp[32];


  if (getLockDSP(1000) <= 0)  return -1;

  regNum &= 0xFFFFFFFEL;

  wordCount = 1 + RespOffset;

  enableLow = PGDRShadow & (uchar)(~_DSP_WR_RD);
  enableHigh = enableLow | (uchar)_DSP_WR_RD;

  xmitWord = htonl(regNum);

  for (i=0; i<wordCount; ++i,regNum+=2)  {
    if (hdwType != HT_PHASE_CTLR)
      rcvWord = (i < RespOffset) ? 0 : fakeRegVal(regNum-2*RespOffset);
    else  {
      OUTI(PGDR, enableLow);  // signal start of 32-bit packet
      _spiWrRd_(spi1Port, (char *)&xmitWord, (char *)&rcvWord, 4);
      OUTI(PGDR, enableHigh);  // signal end of 32-bit packet
    }
    xmitWord = 0;
  }

  releaseLockDSP();

  rcvWord = ntohl(rcvWord);

//  respRegNum = regNum - (ulong)(RespOffset * 2);
//  printf("  0x%08LX:  0x%08LX  %12ld  %.32s\r\n",
//         respRegNum, rcvWord, rcvWord, fmtDSPval(rcvWord, respRegNum, temp));

  *regVal = rcvWord;

  return 0;
}

//-----------------------------------------------------------------------------
//  write a single 32-bit register value to the DSP (without any additional
//  cycles to read a response, which can interfere with any larger sequence
//  of packets this may be imbedded in).
//
//  Returns 0 on success, < 0 if an error
//-----------------------------------------------------------------------------
int writeDSPRegQuick(ulong regNum, ulong value)
{
  uchar  enableLow, enableHigh;
  ulong  xmitWord, rcvWord;


  if (getLockDSP(1000) <= 0)  return -1;

  enableLow = PGDRShadow & (uchar)(~_DSP_WR_RD);
  enableHigh = enableLow | (uchar)_DSP_WR_RD;

  regNum |= 1L;

//--- send the register # in the 1st packet ---
  xmitWord = htonl(regNum);
  OUTI(PGDR, enableLow);  // signal start of 32-bit packet
  _spiWrRd_(spi1Port, (char *)&xmitWord, (char *)&rcvWord, 4);
  OUTI(PGDR, enableHigh);  // signal end of 32-bit packet

//--- send the value in the 2nd packet ---
  xmitWord = htonl(value);
  OUTI(PGDR, enableLow);  // signal start of 32-bit packet
  _spiWrRd_(spi1Port, (char *)&xmitWord, (char *)&rcvWord, 4);
  OUTI(PGDR, enableHigh);  // signal end of 32-bit packet

  releaseLockDSP();

  return 0;
}

//-----------------------------------------------------------------------------
//  write a single 32-bit register value to the DSP and read back the echoed
//  addr and value (which at least lets us know we are talking to the DSP).
//
//  Returns 0 on success, < 0 if an error
//-----------------------------------------------------------------------------
int writeDSPReg(ulong regNum, ulong value, bool showErrs)
{
  uchar  enableLow, enableHigh;
  ulong  xmitWord, rcvWord;


  if (getLockDSP(2000) <= 0)  return -1;

  if (hdwType != HT_PHASE_CTLR)  showErrs = false;

  enableLow = PGDRShadow & (uchar)(~_DSP_WR_RD);
  enableHigh = enableLow | (uchar)_DSP_WR_RD;

  regNum |= 1L;

//--- send the register # in the 1st packet ---
  xmitWord = htonl(regNum);
  OUTI(PGDR, enableLow);  // signal start of 32-bit packet
  _spiWrRd_(spi1Port, (char *)&xmitWord, (char *)&rcvWord, 4);
  OUTI(PGDR, enableHigh);  // signal end of 32-bit packet
  if (showFiller)  printf("  [%08lX", ntohl(rcvWord));

//--- send the value in the 2nd packet ---
  xmitWord = htonl(value);
  OUTI(PGDR, enableLow);  // signal start of 32-bit packet
  _spiWrRd_(spi1Port, (char *)&xmitWord, (char *)&rcvWord, 4);
  OUTI(PGDR, enableHigh);  // signal end of 32-bit packet
  if (showFiller)  printf("  %08lX]\r\n", ntohl(rcvWord));

//--- read back the echoed register # ---
  xmitWord = 0;
  OUTI(PGDR, enableLow);  // signal start of 32-bit packet
  _spiWrRd_(spi1Port, (char *)&xmitWord, (char *)&rcvWord, 4);
  OUTI(PGDR, enableHigh);  // signal end of 32-bit packet
  rcvWord = ntohl(rcvWord);

  if (rcvWord != regNum)  {
    if (showErrs)
      printf("  echoed write address error: 0x%08lX  echoed for 0x%08lX\r\n",
             rcvWord, regNum);
    releaseLockDSP();  return -2;
  }

//--- read back the echoed value ---
  xmitWord = 0;
  OUTI(PGDR, enableLow);  // signal start of 32-bit packet
  _spiWrRd_(spi1Port, (char *)&xmitWord, (char *)&rcvWord, 4);
  OUTI(PGDR, enableHigh);  // signal end of 32-bit packet
  rcvWord = ntohl(rcvWord);

  if (rcvWord != value)  {
    if (showErrs)
      printf("  echoed value error: 0x%08lX  echoed for 0x%08lX\r\n",
             rcvWord, value);
    releaseLockDSP();  return -3;
  }

  releaseLockDSP();

  return 0;
}

//-----------------------------------------------------------------------------
//  initialize various default settings
//-----------------------------------------------------------------------------
void initDefaults(void)
{
  char  *initSeq;
  fd_t  fd;
  off_t  offset;
  uint  ui, numWrites, numSeqs;
  struct stat  fsInfo;
  char  buf[PLLInitSeqLen];


//--- initialize the PLL, either from a fixed seq or one in a file ---
  fd = open(PLLInitSeqFile, O_RDONLY, 0);
  if (fd < 0)  {
    if (pllInitIdx >= NumFixedPLLSeqs)  {
      printf("No 'pllInitSeqs' file and pllInitIdx > fixed list size\r\n");
      return;
    }
    initSeq = (char *) pllInitSeqs[pllInitIdx];
  }
  else  {
    offset = pllInitIdx * PLLInitSeqLen;
    if (fstat(fd, &fsInfo) < 0)  { perror(PLLInitSeqFile);  Exit(-2); }
    numSeqs = (uint)fsInfo.st_size / PLLInitSeqLen;
    if (pllInitIdx >= numSeqs)  {
      printf("pllInitIdx > # of sequences in file 'pllInitSeqs'\r\n");
      close(fd);  return;
    }
    if (lseek(fd, offset, SEEK_SET) != offset)  {
      perror(PLLInitSeqFile);  close(fd);  return;  }
    if (read(fd, buf, PLLInitSeqLen) != PLLInitSeqLen)  {
      printf("Error reading PLL init sequence\r\n");
      perror(PLLInitSeqFile);  close(fd);  return;
    }
    close(fd);  initSeq = buf;
  }

  numWrites = PLLInitSeqLen / 4;

  for (ui=0; ui<numWrites; ++ui)  { writePLL(initSeq, 4);  initSeq += 4; }


//--- initialize the ADC ---
  write_ADCReg(0, 0x0400);  write_ADCReg(13, 0x0403);
}



//=============================================================================
//  Event-Handler functions called by the appl task to process event messages
//=============================================================================

//-----------------------------------------------------------------------------
//  Event-Handler for a request to ...
//-----------------------------------------------------------------------------
/*
void EVENT_MyEvent(EventMsg *event)
{
  uint  idx;
  float  setpoint;
  char  temp[16];


  idx = event->chanNum;

  if (error)  {
    fprintf(DiagLog, "error msg, chan:\n", idx+1);
    return;
  }

  setpoint = *((float *)event->data);

  fprintf(DiagLog, "New SetPoint for PS%d: %s\n",
     idx+1, fmtFloat(setpoint, 8,3, temp) );

  *((float *)((&iSet_1_RegInfo + idx)->regVal)) = setpoint;

  anaOut[idx] = ISETtoVolts(idx, setpoint);
}
*/



//=============================================================================
//  Task functions to handle application-specific commands
//=============================================================================


//-----------------------------------------------------------------------------
//  example func to process a command entered via a user shell
//-----------------------------------------------------------------------------
/*
void _task miscTest(int argc, char **argv)
{
  int  idx;
  bool  validCmd;
  // used by EVENT macro
  fd_t  eventsfd = -1;
  EventMsg  eventMsg;


  validCmd = false;

  for (;;)  {
    if (argc < 3)  break;
    if (! isdigit(argv[1][0]))  break;

    idx = atoi(argv[1]);
    if ((idx < 1) OR (idx > NumChan))  {
      printf("Invalid chan #\n");  Exit(-2);  }

    --idx;  validCmd = true;

    if (StriEq(argv[2], "on"))     { EVENT(TurnOn,  idx, NULL, 0);  break; }
    if (StriEq(argv[2], "off"))    { EVENT(TurnOff, idx, NULL, 0);  break; }
    if (StriEq(argv[2], "reset"))  { EVENT(Reset,   idx, NULL, 0);  break; }

    validCmd = false;  break;
  }

  if (! validCmd)  { printf("Usage:  misc # on|off|reset\r\n");  Exit(-1); }

  Exit(0);
}
*/



//-----------------------------------------------------------------------------
//  Check for a new header from the client.  If the maxIdleMS indicates how
//  long to wait to receive the start of the header.
//
//  Returns:
//    0 if nothing is received
//    The client version if a valid-looking header was received
//    < 0 if an error occured
//-----------------------------------------------------------------------------
int checkForHeader(bool sendWaveform[], long maxIdleMS)
{
  uint  u, uw, hdrSize, numWaveforms;
  int  w, stat, clientVers;
  char  *ptr;
  char  buf[64];


//--- wait a bit to see if we got anything from the client ---
  stat = read_(0, buf, 2, maxIdleMS);

  if (stat != 2)  {
    if (stat < 0) {
      if (errno == EAGAIN)  return 0;
      perror("read()");
    }
    return -1;
  }

  if ((hdrSize = ntohs(*(U16 *)buf)) > sizeof(buf))  return -1;

  if (hdrSize < 4)  return -1;

  if (read_(0, buf, hdrSize, 5000) != hdrSize)  { perror("read()");  return -1; }

  ptr = buf;

  clientVers = getShort(&ptr);
  numWaveforms = (uint) getShort(&ptr);

  if ((ptr + numWaveforms*2) > (buf + hdrSize))  return -1;

//--- > 0 means add waveform to list, < 0 means remove it from the list ---
  for (u=0; u<numWaveforms; ++u)  {
    w = getShort(&ptr);  uw = abs(w) - 1;
    if (uw >= NUM_WAVEFORMS)  return -1;
    sendWaveform[uw] = (w > 0);
  }

  return clientVers;
}

//-----------------------------------------------------------------------------
//  Task started by inetd to service remote connections that want to repeatedly
//  read amplitude and phase waveforms.
//-----------------------------------------------------------------------------
//  Notes on method to allow clients to "continuously" read amplitude and
//  phase waveforms:
//
//  - When a client wants to repeatedly read the amplitude and phase data as a
//    waveform, it opens a TCP socket connection to the appropriate port and
//    then repeatedly reads the data as it is sent.
//
//  - Each client with such a socket connection will repeatedly receive the
//    following until one end or the other closes the connection:
//    - A 12-byte header with 3 32-bit unsigned values in network byte order:
//      - the seconds portion of the time for the start of the data points
//      - the nanoseconds portion of the start time
//      - the # of bytes with data points to follow the header
//    - The specified # bytes of data.  Each data point is sent as a 16-bit
//      unsigned phase value followed by a 16-bit unsigned amplitude value.
//      These values can be converted to engineering units by multiplying
//      them by the apropriate conversion factor (see the PhaseConvFactor and
//      AmplConvFactor macros in this file).
//
//  - When the amplitude or phase values are being "pulsed" (the setpoint
//    periodically changed to one of 2 values), the waveform data sent back
//    will be for a period immeadiately following the change in the setpoint.
//
//  - When the values are NOT being pulsed, the waveform data sent back will
//    simply indicate the amplitude and phase values over a recent interval,
//    with no particular starting point as a reference.
//
//-----------------------------------------------------------------------------
void _task readWave(int argc, char **argv)
{
  ulong  dataHeader[3];
  ulong  lastDataTime, startTime;
  uint  lastSentTime;
  U16  src, cpy;
  int  stat, dataBytes;
  uint  srcLockID, cpyLockID;
  char far  *srcBuf, *cpyBuf;
  char  buf[256];


//--- A short delay if needed in case the client got a connection so quickly ---
//--- after we rebooted that the applic task hasn't finished initializing ---
  startTime = GetMS();
  while (!applicReady AND (msSince(startTime) < 5000))  sleepMS(200);
  if (!applicReady)  { printf("applic task not ready\r\n");  Exit(-1); }


  lastDataTime = GetMS();  lastSentTime = GetMS_();

//--- read data index ---
  if (read_(0, (char *)&src, 2, 5000) != 2)  Exit(-1);
  src = ntohs(src) & 0xFFFE;

  if (src >= NUM_WAVEFORMS*2)  src = 0;
  cpy = src + 1;

  fprintf(DiagLog, "New reader for Waveform #: %d\n", src);

  srcBuf = (char far *) waveformData[src];    srcLockID = WAVE_DATA_0 + src;
  cpyBuf = (char far *) waveformData[cpy];    cpyLockID = WAVE_DATA_0 + cpy;

  for (;;)  {
    _EnterCritical_();  readerRdy[src] = true;  _ExitCritical_();

  //--- The next thing we check is to see if we need to update the buffer ---
  //--- that holds a copy of the new data. ---

  //  The applic process is the one that copies data from the DSP buffers to
  //  global buffers in RAM.  To avoid blocking it for excessive amounts of
  //  time, we minimize the amount of time the buffer it uses is locked by
  //  making a copy of it and using the copy at our more leisurely pace.
  //  (Mearly doing a memory-to-memory copy goes fairly quickly - writing a
  //  copy of that data to a socket could block for some time).

  //  Changes that would improve this further:
  //
  //    - Support multiple read-only locks on the copy
  //
  //    - A single task that services multiple clients.  This would mean
  //      only one task for any number of clients.  But to avoid having a
  //      single task running all the time listening for new connections,
  //      I need a way to tell inetd to NOT keep listening for new
  //      connections (or for inetd to hand new connections off to an
  //      existing server task).  I don't know if there is a standard way
  //      to accomplish either of these approachs on a real Linux box.
  //
  //    - Support an arbitrary set of data elements on each connection
  //      (i.e. use the EPICS CA approach including subscriptions)
  //
  //  Of course, once we have EPICS waveform records for these, there
  //  will be at least one open connection all the time.  And if we don't
  //  change to the multipe waveforms per connection approach, there will
  //  be several connections open all the time!
  //
  //  Given the rather limited resources of the current hardware that
  //  DEBROS runs on, the EPICS CA-like approach is really the only
  //  practical one.


  //--- Get a lock on the dest/cpy buffer so it won't change while we are ---
  //--- accessing (and possibly updating) it ---

    while (getLock(cpyLockID, 5000) <= 0);  // get lock on dest buffer

  //--- If the timestamps for the source and destination buffers are ---
  //--- different, then update the contents and timestamp for the ---
  //--- destination buffer.  NOTE that we don't really care about the ---
  //--- possibility of the src date/time changing during the check as ---
  //--- all we really care about is if it HASN'T changed. ---

    if (dataTimeMS[cpy] != dataTimeMS[src])  {
      while (getLock(srcLockID, 5000) <= 0);  // get lock on src buffer
      fmemcpy(cpyBuf, srcBuf, numDataPoints[src]<<2);
      numDataPoints[cpy] = numDataPoints[src];
      dataTimeSecs[cpy] = dataTimeSecs[src];  dataTimeMS[cpy] = dataTimeMS[src];
      releaseLock(srcLockID);  // release lock on src buffer
      WAVE_PV(cpy)->newEvents |= DBE_VALUE;
    }

  //--- If the destination/copy buffer doesn't have newer data than what we ---
  //--- last sent to the client, and it has been at least 2 seconds since ---
  //--- we sent anything to the client, then send an "empty" waveform ---
  //--- response (this gives the client a way to tell if we died so it ---
  //--- can close its current connection and open a new one if needed) ---

    if (!dataTimeMS[cpy] OR (dataTimeMS[cpy] == lastDataTime))  {
      releaseLock(cpyLockID);
      if (msSince_(lastSentTime) >= 2000)  {
//        fprintf(DiagLog, "sending empty response\n");  //tdebug
        lastSentTime = GetMS_();
        dataHeader[0] = htonl(time(NULL));
        dataHeader[1] = htonl((lastSentTime%1000) * 1000000L);
        dataHeader[2] = htonl(0);
        if (write_(1, (char *)dataHeader, 12, 1000) != 12)  {
          fprintf(DiagLog, "Error writing empty response\n");  break; }
      }
      sleepMS(50);  continue;
    }

  //--- We have new data to send ---
    dataBytes = numDataPoints[cpy] * sizeof(ulong);

    dataHeader[0] = htonl(dataTimeSecs[cpy]);
    dataHeader[1] = htonl((dataTimeMS[cpy] % 1000) * 1000000L);
    dataHeader[2] = htonl(dataBytes);

    if (write_(1, (char *)dataHeader, 12, 2000) != 12)  {
      fprintf(DiagLog, "Error writing the header\n");  break; }

    stat = fwrite_(1, cpyBuf, dataBytes, buf, sizeof(buf), 2000);
    if (stat != dataBytes)  {
      fprintf(DiagLog, "Wrote %d of %d bytes\n", stat, dataBytes);  break; }

    lastDataTime = dataTimeMS[cpy];  lastSentTime = GetMS_();

    releaseLock(cpyLockID);  // release lock on buffer w/ copy of data

  //--- CTF connections are not allowed to be continuous ---
    if ((cpy >= WAVE_CTF_P_SRC) AND (cpy <= WAVE_CTF_A_CPY))  Exit(0);

    sleepMS(50);
  }


//--- We only get here if there was an error ---
  releaseLock(cpyLockID);

  fprintf(DiagLog, strerror(errno));  fprintf(DiagLog, "\n");

  Exit(-1);
}


//-----------------------------------------------------------------------------
//  auto-calibrate the fwd, refl, or cav atten value by cycling through
//  the range of atten values for a given pair of V-peak and phase values.
//-----------------------------------------------------------------------------
void _task autocal(int argc, char **argv)
{
  U16  u, curAtten;
  F32  vPeak, phase;
  byte  flags[3];
  RegInfo  *reg[3];


  if (showHelp(4, argc, argv))  {
    printf("Usage:  autocal fwd|refl|cav <V-peak> <Phase>\r\n");
    Exit(-1);
  }

  reg[0] = NULL;

  if (StriEq(argv[1], "fwd"))  {
    curAtten = fwdAtten;
    reg[0] = &fwdAtten_RegInfo;
    reg[1] = &fwdMagCalib_RegInfo;
    reg[2] = &fwdPhaseCalib_RegInfo;
  }  else
  if (StriEq(argv[1], "refl"))  {
    curAtten = reflAtten;
    reg[0] = &reflAtten_RegInfo;
    reg[1] = &reflMagCalib_RegInfo;
    reg[2] = &reflPhaseCalib_RegInfo;
  }  else
  if (StriEq(argv[1], "cav"))  {
    curAtten = cavAtten;
    reg[0] = &cavAtten_RegInfo;
    reg[1] = &cavMagCalib_RegInfo;
    reg[2] = &cavPhaseCalib_RegInfo;
  }

  if (! reg[0])  {
    printf("Invalid argument: %.32s\r\n", argv[1]);  Exit(-1); }

  vPeak = atof(argv[2]);  phase = atof(argv[3]);

//--- save the flags value for each register so we can restore it after ---
//--- temporarily changing it to avoid seeing the normal log messages on ---
//--- the terminal screen ---
  for (u=0; u<3; ++u)  flags[u] = reg[u]->flags;

//--- cycle the specified attenuation from 0 to 63, waiting 1 second after ---
//--- each change and then telling the DSP the v-peak and phase values ---
//--- that it should associate with the current conditions ---

  for (u=0; u<64; ++u)  {
    printf("%3d\r", u);  fflush(stdout);

    reg[0]->flags &= ~(byte)LogChanges;
    setRegU16(reg[0], 0, u);
    reg[0]->flags = flags[0];

    sleepMS(1000);

    reg[1]->flags &= ~(byte)LogChanges;
    setRegF32(reg[1], 0, vPeak);
    reg[1]->flags = flags[1];

    reg[2]->flags &= ~(byte)LogChanges;
    setRegF32(reg[2], 0, phase);
    reg[2]->flags = flags[2];

    CheckETX();
  }
  printf("\r\n");

  reg[0]->flags &= ~(byte)LogChanges;
  setRegU16(reg[0], 0, curAtten);
  reg[0]->flags = flags[0];


  Exit(0);
}


//-----------------------------------------------------------------------------
//  "cmd": set various DSP states
//-----------------------------------------------------------------------------
void _task cmd(int argc, char **argv)
{
  ulong  bits = 0;


  if (showHelp(3, argc, argv))  {
    printf("Usage:  cmd afc on|off\r\n"
           "            amplPID on|off\r\n"
           "            phaseCtl off|pid|sel|auto\r\n"
           "            rfPulse on|off\r\n"
           "            tunerPID on|off\r\n"
          );
    Exit(-1);
  }

  do {
    if (StriEq(argv[1], "afc"))  {
      if (StriEq(argv[2], "on"))  { bits = CMD_AFC_ON;  break; }
      if (StriEq(argv[2], "off"))  { bits = CMD_AFC_OFF;  break; }
    }

    if (StriEq(argv[1], "amplPID"))  {
      if (StriEq(argv[2], "on"))  { bits = CMD_AMPL_PID_ON;  break; }
      if (StriEq(argv[2], "off"))  { bits = CMD_AMPL_PID_OFF;  break; }
    }

    if (StriEq(argv[1], "phaseCtl"))  {
      if (StriEq(argv[2], "off"))  { bits = CMD_PHASE_CTL_OFF;  break; }
      if (StriEq(argv[2], "pid"))  { bits = CMD_PHASE_CTL_PID;  break; }
      if (StriEq(argv[2], "sel"))  { bits = CMD_PHASE_CTL_SEL;  break; }
      if (StriEq(argv[2], "auto"))  { bits = CMD_PHASE_CTL_AUTO;  break; }
    }

    if (StriEq(argv[1], "rfPulse"))  {
      if (StriEq(argv[2], "on"))  { bits = CMD_RF_PULSING_ON;  break; }
      if (StriEq(argv[2], "off"))  { bits = CMD_RF_PULSING_OFF;  break; }
    }

    if (StriEq(argv[1], "tunerPID"))  {
      if (StriEq(argv[2], "on"))  { bits = CMD_TUNER_PID_ON;  break; }
      if (StriEq(argv[2], "off"))  { bits = CMD_TUNER_PID_OFF;  break; }
    }

  } while (false);


  if (bits)
    SetMBReg(setDSPState, 0, U32, bits);
  else
    printf("Invalid parameters\r\n");

  Exit(0);
}



//-----------------------------------------------------------------------------
//  "config": set the configuration register (values 0-7)
//-----------------------------------------------------------------------------
void _task setConfig(int argc, char **argv)
{
  int  newConfig;


  if (showHelp(2, argc, argv))  {
    printf("Usage:  config #  (# = 0-7)\r\n");  Exit(-1); }

  newConfig = atoi(argv[1]);

  if ((newConfig < 0) OR (newConfig > 7))  {
    printf("Configuration command out of valid range:  %s\r\n", argv[1]);
    Exit(-2);
  }

  printf("Configuring register\r\n");

  write_Config(newConfig);

  Exit(0);
}


//-----------------------------------------------------------------------------
//  "read": picks ADC to send to DAC output (1-4)
//-----------------------------------------------------------------------------
void _task selectADC(int argc, char **argv)
{
  int  newADCVal;


  if (showHelp(2, argc, argv))  {
    printf("Usage:  read #  (# = 1-4)\r\n");  Exit(-1); }

  newADCVal = atoi(argv[1]) - 1;

  if ((newADCVal < 0) OR (newADCVal > 3))  {
    printf("%s is not a valid ADC #: Select 1-4\r\n", argv[1]);
    Exit(-2);
  }

  printf("Changing to ADC %d for readings\r\n", newADCVal + 1);

  writeFPGAParallel(1, (uchar)newADCVal);

  Exit(0);
}

//-----------------------------------------------------------------------------
//  "setadcreg": set the value of one of the registers in the ADC
//-----------------------------------------------------------------------------
void _task setADCReg(int argc, char **argv)
{
  uint  regNum, regVal;


  if (showHelp(3, argc, argv))  {
    printf("Usage:  setadcreg reg# value  (reg# = 0-31)\r\n");  Exit(-1); }

  regNum = (uint) ATOI(argv[1]);  regVal = (uint) ATOI(argv[2]);

  if (regNum > 31)  {
    printf("%s is not a valid ADC register #: Use 0-31\r\n", argv[1]);
    Exit(-2);
  }
  if (regVal > 0x07FF)  {
    printf("%s is not a valid ADC register value: Use 0-0x7FF\r\n", argv[2]);
    Exit(-2);
  }

  printf("Setting ADC reg %u to %03X\r\n", regNum, regVal);

  write_ADCReg(regNum, regVal);

  Exit(0);
}

//-----------------------------------------------------------------------------
//  "defaults": initialize FPGA settings to default values
//-----------------------------------------------------------------------------
void _task defaults(int argc, char **argv)
{
  initDefaults();  Exit(0);
}



//-----------------------------------------------------------------------------
//  "progdsp": load a program in to the DSP
//-----------------------------------------------------------------------------
void _task progDSP(int argc, char **argv)
{
  int  fromfd;
  uint progNum;
  char  *fileName;
  struct stat  fsInfo;


  if (showHelp(3, argc, argv))  {
    printf("Usage:  progdsp prog# file\r\n");  Exit(-1);  }

  progNum = ATOI(argv[1]);  fileName = argv[2];

  if (progNum > 3)  { printf("prog# must be 0-3\r\n");  Exit(-1); }

  if ((fromfd = open(fileName, O_RDONLY, 0)) < 0)  {
    perror(fileName);  Exit(-1); }

  if (fstat(fromfd, &fsInfo))  { perror(argv[1]);  Exit(-1); }

  initDSP(fromfd, progNum, (ulong)fsInfo.st_size, true);

  close(fromfd);  Exit(0);
}

//-----------------------------------------------------------------------------
//  "restartdsp": reset the DSP and run specified program # (0-3)
//-----------------------------------------------------------------------------
void _task restartDSP(int argc, char **argv)
{
  uint progNum;

  if (showHelp(2, argc, argv))  {
    printf("Usage:  restartdsp prog#\r\n");  Exit(-1);  }

  progNum = ATOI(argv[1]);

  if (progNum > 3)  { printf("prog# must be 0-3\r\n");  Exit(-1); }

  dspProgNum = progNum;

  resetDSP();  Exit(0);
}


//-----------------------------------------------------------------------------
//  "setfpgafirm": load a new FPGA firmware file in to the serial EEPROM
//-----------------------------------------------------------------------------
void _task setFPGAfirm(int argc, char **argv)
{
  int  i, fromfd;
  bool  verify = false;
  char  *fileName;


  if (showHelp(1, argc, argv))  {
    printf("\nusage:  setfpgafirm [-v] [fileName]\r\n");  Exit(-1); }

  fileName = FPGA_Prog_File;
  for (i=1; i<argc; ++i)  {
    if (argv[i][0] != '-')  { fileName = argv[i];  continue; }
    if (StrEq(argv[i], "-v"))  verify = true;
  }

  if ((fromfd = open(fileName, O_RDONLY, 0)) < 0)  {
    perror(fileName);  Exit(-1); }

  printf("\r\nDownloading %.40s to EEPROM\r\n", fileName);

  updateFPGAfirmware(fromfd, true, verify);

  close(fromfd);  Exit(0);
}


//-----------------------------------------------------------------------------
//  "setpllreg": set the value of one of the registers in the clock ctlr/PLL
//-----------------------------------------------------------------------------
void _task setPLLReg(int argc, char **argv)
{
  uint  regNum;
  ulong  regVal;
  char  buf[4];


  if (showHelp(3, argc, argv))  {
    printf("Usage:  setpllreg reg# value  (reg# = 0-15)\r\n");  Exit(-1); }

  regNum = (uint) ATOI(argv[1]);  regVal = (ulong) ATOL(argv[2]);

  if (regNum > 15)  {
    printf("%s is not a valid ADC register #: Use 0-31\r\n", argv[1]);
    Exit(-2);
  }
  if (regVal > 0x0FFFFFFFL)  {
    printf("%s is not a valid PLL register value: Use 0-0x0FFFFFFF\r\n",
            argv[2]);
    Exit(-2);
  }

  buf[0] = (uchar)(regVal >> 20);
  buf[1] = (uchar)(regVal >> 12);
  buf[2] = (uchar)(regVal >> 4);
  buf[3] = (uchar)(regVal << 4) | (uchar)regNum;

  printf("Setting PLL reg %u to %08lX\r\n", regNum, regVal);

  writePLL(buf, 4);

  Exit(0);
}

//=============================================================================
//  Generic structure to handle application-specific presentation of the
//  digital and analog input/output values.  Code to handle additional
//  formatting that cannot be handled in this generic manner should be
//  placed in the showState() task routine.
//=============================================================================

//-----------------------------------------------------------------------------
_nearcall char *fmtPhaseMode(DispElem far *elem, char *buf)
{
  return (char *) PhaseModeDesc[ELEM_VAL(uint)];
}

//-----------------------------------------------------------------------------
_nearcall char *fmtAFC(DispElem far *elem, char *buf)
{
  bool isOn = statusBitsAFC & (1 << elem->prec);
  return (isOn ? "ON " : "OFF");
}

//-----------------------------------------------------------------------------
_nearcall char *fmtWarming(DispElem far *elem, char *buf)
{
  return (ELEM_VAL(bool) ? " WARM  " : "WARMING");
}

//-----------------------------------------------------------------------------
_nearcall char *fmtRamping(DispElem far *elem, char *buf)
{
  return (ELEM_VAL(bool) ? "RAMPING" : "STEADY ");
}

//-----------------------------------------------------------------------------
_nearcall char *fmtLocked(DispElem far *elem, char *buf)
{
  return (ELEM_VAL(bool) ? "LOCKED  " : "UNLOCKED");
}
//-----------------------------------------------------------------------------
  const char far DevStateStr[] = {
"--INPUTS--  Max  Ampl   Phase Atten   -------- RF State --------   ---DSP---\r\n"
"Forward:  ##.## ##.## -###.##  ##     WarmLevel: #.##      ###     Load: ###%%\r\n"
"Reflected:##.## ##.## -###.##  ##     WarmMS:   #####    #######   Temp:##.#C\r\n"
"Cavity:   ##.## ##.## -###.##  ##     RampMS:   #####    #######   Prog ID: #\r\n"
"Reference:      ##.## -###.##                                      Errs:#####\r\n"
"Setpoints:##.## ##.## -###.##         CavLocked: ########\r\n"
"\r\n"
"<########>   -Freq-  -Phase-    -TUNE-  -Ampl-  -Phase-   Pulsing: ###\r\n"
"AFC_0: ###   ####.#  -###.##    State:  ###### ########   Pulse Rate:###.##\r\n"
"AFC_1: ###   ####.#  -###.##    Pulsed:   #.##  -###.##   SearchAmpl: #.##\r\n"
"AFC_2: ###   ####.#  -###.##\r\n"
"AFC_3: ###   ####.#  -###.##      --RF-- -Tuner-  ---- Tuner -----\r\n"
"AFC_4: ###   ####.#  -###.##    P: #.###  #.###   State:    ######    CTF:###%%\r\n"
"AFC_5: ###   ####.#  -###.##    I: #.###  #.###   Piezo:    -#####    FFT:###%%\r\n"
"AFC_6: ###   ####.#  -###.##    D: #.###  #.###   Stepper:-#######\r\n"
"AFC_7: ###   ####.#  -###.##\r\n"
"AFC_8: ###   ####.#  -###.##\r\n"
"AFC_9: ###   ####.#  -###.##\r\n"
"\r\n"
  };

const DispElem  far dispElems[] = {
  { 10, 3,  VFPV(maxFwdAmpl),    BLUE,  fmtF32, 5,2 },
  { 10, 4,  VFPV(maxReflAmpl),   BLUE,  fmtF32, 5,2 },
  { 10, 5,  VFPV(maxCavAmpl),    BLUE,  fmtF32, 5,2 },

  { 10, 7,  VFPV(maxOutAmpl), BLUE,  fmtF32, 5,2 },

  { 16, 3,  VFPV(fwdAmpl),  BLUE,  fmtF32, 5,2 },
  { 16, 4,  VFPV(reflAmpl), BLUE,  fmtF32, 5,2 },
  { 16, 5,  VFPV(cavAmpl),  BLUE,  fmtF32, 5,2 },
  { 16, 6,  VFPV(refAmpl),  BLUE,  fmtF32, 5,2 },
  { 16, 7,  VFPV(amplSet),  BLUE,  fmtF32, 5,2 },

  { 22, 3,  VFPV(fwdPhase),  BLUE,  fmtF32, 7,2 },
  { 22, 4,  VFPV(reflPhase), BLUE,  fmtF32, 7,2 },
  { 22, 5,  VFPV(cavPhase),  BLUE,  fmtF32, 7,2 },
  { 22, 6,  VFPV(refPhase),  BLUE,  fmtF32, 7,2 },
  { 22, 7,  VFPV(phaseSet),  BLUE,  fmtF32, 7,2 },

  { 31, 3,  VFPV(fwdAtten),   BLUE,  fmtU16, 2,2 },
  { 31, 4,  VFPV(reflAtten),  BLUE,  fmtU16, 2,2 },
  { 31, 5,  VFPV(cavAtten),   BLUE,  fmtU16, 2,2 },

  { 59, 3,  VFPV(rfIsOn), BLUE,  fmtBool, 3, FMT_ON_OFF },
  { 57, 4,  VFPV(rfAtWarm), BLUE,  fmtWarming, 7,7 },
  { 57, 5,  VFPV(rfRamping), BLUE,  fmtRamping, 7,7 },

  { 49, 7,  VFPV(cavLocked), BLUE,  fmtLocked, 8,8 },

  { 49, 3,  VFPV(warmLevel),  BLUE,  fmtF32, 4,2 },
  { 48, 4,  VFPV(warmMS),     BLUE,  fmtU16, 5,5 },
  { 48, 5,  VFPV(rampMS),     BLUE,  fmtU16, 5,5 },

  { 73, 3,  VFPV(dspLoad),     BLUE,  fmtU16, 3,3 },
  { 72, 4,  VFPV(dspTemp),     BLUE,  fmtF32, 4,1 },
  { 76, 5,  VFPV(curProgNum),  BLUE,  fmtU16, 1,1 },
  { 72, 6,  VFPV(romProgErrs), BLUE,  fmtU16, 5,5 },

  {  1, 9,  VFPV(afcEnabled), BLUE,  fmtBool, 8, FMT_ENA_DISABLED },

  {  7,10,  VFPV(statusBitsAFC), BLUE,  fmtAFC, 3, 0 },
  {  7,11,  VFPV(statusBitsAFC), BLUE,  fmtAFC, 3, 1 },
  {  7,12,  VFPV(statusBitsAFC), BLUE,  fmtAFC, 3, 2 },
  {  7,13,  VFPV(statusBitsAFC), BLUE,  fmtAFC, 3, 3 },
  {  7,14,  VFPV(statusBitsAFC), BLUE,  fmtAFC, 3, 4 },
  {  7,15,  VFPV(statusBitsAFC), BLUE,  fmtAFC, 3, 5 },
  {  7,16,  VFPV(statusBitsAFC), BLUE,  fmtAFC, 3, 6 },
  {  7,17,  VFPV(statusBitsAFC), BLUE,  fmtAFC, 3, 7 },
  {  7,18,  VFPV(statusBitsAFC), BLUE,  fmtAFC, 3, 8 },
  {  7,19,  VFPV(statusBitsAFC), BLUE,  fmtAFC, 3, 9 },

  { 13,10,  VFPV(afcFreq0),  BLUE,  fmtF32, 6,1 },
  { 13,11,  VFPV(afcFreq1),  BLUE,  fmtF32, 6,1 },
  { 13,12,  VFPV(afcFreq2),  BLUE,  fmtF32, 6,1 },
  { 13,13,  VFPV(afcFreq3),  BLUE,  fmtF32, 6,1 },
  { 13,14,  VFPV(afcFreq4),  BLUE,  fmtF32, 6,1 },
  { 13,15,  VFPV(afcFreq5),  BLUE,  fmtF32, 6,1 },
  { 13,16,  VFPV(afcFreq6),  BLUE,  fmtF32, 6,1 },
  { 13,17,  VFPV(afcFreq7),  BLUE,  fmtF32, 6,1 },
  { 13,18,  VFPV(afcFreq8),  BLUE,  fmtF32, 6,1 },
  { 13,19,  VFPV(afcFreq9),  BLUE,  fmtF32, 6,1 },

  { 21,10,  VFPV(afcPhase0),  BLUE,  fmtF32, 7,2 },
  { 21,11,  VFPV(afcPhase1),  BLUE,  fmtF32, 7,2 },
  { 21,12,  VFPV(afcPhase2),  BLUE,  fmtF32, 7,2 },
  { 21,13,  VFPV(afcPhase3),  BLUE,  fmtF32, 7,2 },
  { 21,14,  VFPV(afcPhase4),  BLUE,  fmtF32, 7,2 },
  { 21,15,  VFPV(afcPhase5),  BLUE,  fmtF32, 7,2 },
  { 21,16,  VFPV(afcPhase6),  BLUE,  fmtF32, 7,2 },
  { 21,17,  VFPV(afcPhase7),  BLUE,  fmtF32, 7,2 },
  { 21,18,  VFPV(afcPhase8),  BLUE,  fmtF32, 7,2 },
  { 21,19,  VFPV(afcPhase9),  BLUE,  fmtF32, 7,2 },

  { 40,10,  VFPV(amplPIDisON), BLUE,  fmtBool, 6, FMT_CLOSED_OPEN },
  { 42,11,  VFPV(pulsedAmplSet),  BLUE,  fmtF32, 4,2 },

  { 47,10,  VFPV(phaseMode), BLUE,  fmtPhaseMode, 8,8 },
  { 48,11,  VFPV(pulsedPhaseSet),    BLUE,  fmtF32, 7,2 },

  { 67, 9,  VFPV(rfPulsing),  BLUE,  fmtBool, 3, FMT_ON_OFF },
  { 69,10,  VFPV(pulseFreq),  BLUE,  fmtF32, 6,2 },
  { 70,11,  VFPV(freqSearchAmpl),  BLUE,  fmtF32, 4,2 },

  { 35,14,  VFPV(rf_P),  BLUE,  fmtF32, 5,3 },
  { 35,15,  VFPV(rf_I),  BLUE,  fmtF32, 5,3 },
  { 35,16,  VFPV(rf_D),  BLUE,  fmtF32, 5,3 },

  { 42,14,  VFPV(tuner_P),  BLUE,  fmtF32, 5,3 },
  { 42,15,  VFPV(tuner_I),  BLUE,  fmtF32, 5,3 },
  { 42,16,  VFPV(tuner_D),  BLUE,  fmtF32, 5,3 },

  { 60,14,  VFPV(tunerPIDisON),    BLUE,  fmtBool, 6, FMT_CLOSED_OPEN },
  { 60,15,  VFPV(piezoReadPos),    BLUE,  fmtS16,  6,6 },
  { 58,16,  VFPV(stepperReadPos),  BLUE,  fmtS32,  8,8 },

  { 74,14,  VFPV(percentCTF),  BLUE,  fmtU16, 3,3 },
  { 74,15,  VFPV(percentFFT),   BLUE,  fmtU16, 3,3 },

  { 0,0, NULL, NULL, NULL, 0,0 }
};

//bool  rfAtWarm, rfRamping, cavLocked


//--- ATTR_NAME macros currently defined in fcntl.h ---
const DigDispElem far digInElems[] = {
//{ xx,yy,  BIT_MASK,          "0STR"  ,ATTR_NAME,  "1STR"  ,ATTR_NAME },

  { 0,0, 0, NULL,NULL, NULL,NULL }
};

const DigDispElem far digOutElems[] = {
//{ xx,yy,  BIT_MASK,      "0STR"  ,ATTR_NAME,  "1STR" ,ATTR_NAME },

  { 0,0,  0,  NULL,NULL,  NULL,NULL }
};


//-----------------------------------------------------------------------------
//  "state":  Show continuously updated state of the system
//-----------------------------------------------------------------------------
void _task showState(int argc, char **argv)
{
  bool  blinkOff;
  char  temp[128];


  printf(clsCmd);  blinkOff = false;

  CursorTo(0, 2);  printf(DevStateStr);

  for (;;)  {
    CursorTo(0, 0);  blinkOff = !blinkOff;

    stdHeader();  displayElems((DispElem far *)dispElems, blinkOff, temp);

    displayDigElems(digitalIn, (DigDispElem far *)digInElems, blinkOff, temp);
    displayDigElems(digitalOut, (DigDispElem far *)digOutElems, blinkOff, temp);

    CursorTo(0,22);  fflush(stdout);  CheckETX();  sleepMS(500);
  }

}

//-----------------------------------------------------------------------------
//  return a string with a human-readable format for the specified DSP value
//-----------------------------------------------------------------------------
char *dspVal2Str(ulong dspVal, uchar dspType, char *buf)
{
  float  fval;

  switch (dspType)  {
    case DT_S32:
      sprintf(buf, "%12ld", (S32)dspVal);  break;

    case DT_U32:
      sprintf(buf, "%12lu", (U32)dspVal);  break;

    case DT_F32:
      Engineering(*(float *)&dspVal, 6, buf);  break;

    case RT_PHASE:
      fval = (float)dspVal * PhaseConvFactor;
      Engineering(fval, 6, buf);  break;
  }


  return buf;
}

//-----------------------------------------------------------------------------
//  Check the list of known DSP registers against the specified register ID.
//  If it matches one of the known registers, then format the provided DSP
//  value based on the expected data format.  Otherwise, format the value as
//  a float.
//-----------------------------------------------------------------------------
char *fmtDSPval(ulong dspVal, ulong regNum, char *buf)
{
  uchar  dspType = DT_F32;
  uint  i, j;
  const RD_DSP_REG far *rdReg;
  const WR_DSP_REG far *wrReg;
  const char far *name = NULL;


  for (i=0,rdReg=rdDSPregs; i<NUM_RD_DSP_REGS; ++i,++rdReg)  {
    if (rdReg->id != regNum)  continue;
    name = rdReg->mbReg->name;  dspType = rdReg->type;  break;
  }

  if (! name)  for (i=0,wrReg=wrDSPregs; i<NUM_WR_DSP_REGS; ++i,++wrReg)  {
    if (wrReg->id != regNum)  continue;
    name = wrReg->mbReg->name;  dspType = wrReg->type;  break;
  }

  if (name)  {
    strcpyfn(buf, name);
    for (j=strlen(buf); j<16; ++j)  buf[j] = ' ';
    buf[15] = ' ';
    dspVal2Str(dspVal, dspType, buf+16);
  }
  else  {
    sprintf(buf, "%14ld  ", dspVal);
    dspVal2Str(dspVal, DT_F32, buf+16);
  }

  return buf;
}

//-----------------------------------------------------------------------------
//  "dspcalib": calibrate gain/offset for conversion to/from certain DSP
//  values and Engineering units.
//
//  The value to calibrate is specified by giving the name of a ModBus
//  register value.  If the register is referenced in one of the DSP register
//  tables, and there is a gain and offset value listed for it, then the
//  gain and offset can be calibrated by running the device at 2 different
//  states and providing the actual (measured) value for each state.
//
//  NOTE that some calibrations apply to more than one register, so changing
//  them for one can affect other readings/settings.
//-----------------------------------------------------------------------------
void _task dspcalib(int argc, char **argv)
{
  bool  foundReg;
  uint  ui;
  float  *gain, *offset;
  float  mbVal, egu[2], raw[2];
  const RD_DSP_REG far *rdReg;
  const WR_DSP_REG far *wrReg;
  RegInfo *mbReg;
  char  temp[64];


  if (showHelp(2, argc, argv))  {
    printf("\r\n Usage:  dspcalib <regName>\r\n");  Exit(-1); }


  foundReg = false;

//--- check RO DSP register list ---
  for (rdReg=rdDSPregs,ui=0; ui<NUM_RD_DSP_REGS; ++ui,++rdReg)  {
    mbReg = rdReg->mbReg;  strcpyfn(temp, mbReg->name);
    if (stricmp(temp, argv[1]))  continue;
    gain = rdReg->gain;  offset = rdReg->offset;
    foundReg = true;  break;
  }

//--- check RW DSP register list ---
  if (!foundReg)  for (wrReg=wrDSPregs,ui=0; ui<NUM_WR_DSP_REGS; ++ui,++wrReg) {
    mbReg = wrReg->mbReg;  strcpyfn(temp, mbReg->name);
    if (stricmp(temp, argv[1]))  continue;
    gain = wrReg->gain;  offset = wrReg->offset;
    foundReg = true;  break;
  }

  if (!foundReg)  {
    printf("\r\n Register not related to a DSP value\r\n");  Exit(-2); }

  if (!gain AND !offset)  {
    printf("\r\n Value does not require conversion\r\n");  Exit(-3); }


  printf("\r\n Existing:");
  if (gain)  printf("  Gain: %s", Engineering(*gain, 6, temp));
  if (offset)  printf("  Offset: %s", Engineering(*offset, 6, temp));
  printf("\r\n\n");

//--- get real EGU and DSP values for two test points ---
  for (ui=0; ui<2; ++ui)  {
    printf(" Test point %u:  Enter measured/actual value for %ls: ",
           ui+1, mbReg->name);
    GETS(temp);  if (!*temp)  Exit(-4);
    egu[ui] = atof(temp);  mbVal = mb2float(mbReg->regVal, mbReg);
    if (gain)  if (*gain != 0.0)  mbVal = mbVal / *gain;
    if (offset)  mbVal += *offset;
    raw[ui] = mbVal;
  }

  if ((raw[0] == raw[1]) OR (egu[0] == egu[1]))  {
    printf("\r\n Unusable calibration data\r\n");  Exit(-4); }

  KLOCK();
  if (gain)  *gain = (egu[1] - egu[0]) / (raw[1] - raw[0]);
  if (offset)  *offset = raw[0] - (gain ? egu[0]/(*gain) : raw[1]);
  KUNLOCK();

  printf("\r\n New:");
  if (gain)  printf("  Gain: %s", Engineering(*gain, 6,temp));
  if (offset)  printf("  Offset: %s", Engineering(*offset, 6, temp));
  printf("\r\n\n");

  Exit(0);
}

//-----------------------------------------------------------------------------
//  "dspio": command to enable/disable normal communication with the DSP
//-----------------------------------------------------------------------------
void _task dspio(int argc, char **argv)
{
  bool  val;


  if (showHelp(3, argc, argv))  {
    printf("Usage:  dspio <read|write|filler|waves> <on|off>\r\n");  Exit(-1); }

  if (StriEq("on", argv[2]))  val = true;  else
  if (StriEq("off", argv[2])) val = false;  else  {
    printf("invalid option\r\n");  Exit(-1);  }

  if (StriEq("read", argv[1]))    { autoReadDSP = val;   Exit(0); }
  if (StriEq("write", argv[1]))   { autoWriteDSP = val;  Exit(0); }
  if (StriEq("filler", argv[1]))  { showFiller = val;    Exit(0); }
  if (StriEq("waves", argv[1]))   { readWaves = val;     Exit(0); }

  printf("invalid option\r\n");  Exit(-1);
}


//-----------------------------------------------------------------------------
//  "dspread": command to read one or more 32-bit values from the DSP
//-----------------------------------------------------------------------------
void _task dspRead(int argc, char **argv)
{
  ulong  regNum;
  uchar  enableLow, enableHigh;
  uint  i, count, wordCount;
  ulong  xmitWord, rcvWord, respRegNum;
  char  temp[64];


  if (showHelp(3, argc, argv))  {
    printf("Usage:  dspread <reg> <count>\r\n");  Exit(-1); }

  if (hdwType == HT_PHASE_CTLR)
    if (!spi1Port)  { fprintf(DiagLog, "SPI port not open!\r\n");  Exit(-1); }

  regNum = ATOUL(argv[1]) & 0xFFFFFFFEL;
  count = (uint) ATOI(argv[2]);

  if (getLockDSP(5000) <= 0)  Exit(-1);

  wordCount = count + RespOffset;

  enableLow = PGDRShadow & (uchar)(~_DSP_WR_RD);
  enableHigh = enableLow | (uchar)_DSP_WR_RD;

  for (i=0; i<wordCount; ++i,regNum+=2)  {
    xmitWord = (i >= count) ? 0 : htonl(regNum);

    if (hdwType != HT_PHASE_CTLR)
      rcvWord = (i < RespOffset) ? 0 : fakeRegVal(regNum-2*RespOffset);
    else  {
      OUTI(PGDR, enableLow);  // signal start of 32-bit packet
      _spiWrRd_(spi1Port, (char *)&xmitWord, (char *)&rcvWord, 4);
      OUTI(PGDR, enableHigh);  // signal end of 32-bit packet
    }
    if (i < RespOffset)  {
      if (showFiller)  printf("  filler: 0x%08lX\r\n", ntohl(rcvWord));
      continue;
    }

    rcvWord = ntohl(rcvWord);

    respRegNum = regNum - (ulong)(RespOffset * 2);

    printf("  0x%08LX:  0x%08LX  %.32s\r\n",
           respRegNum, rcvWord, fmtDSPval(rcvWord, respRegNum, temp));
  }
  printf("\r\n");

  releaseLockDSP();

  Exit(0);
}

//-----------------------------------------------------------------------------
//  "dspwrite": command to write a value to the DSP
//-----------------------------------------------------------------------------
void _task dspWrite(int argc, char **argv)
{
  ulong  regNum, setVal;


  if (showHelp(3, argc, argv))  {
    printf("Usage:  dspwrite <reg> <value> [-f]\r\n");  Exit(-1); }

  if (!spi1Port)  { fprintf(DiagLog, "SPI port not open!\r\n");  Exit(-1); }

  regNum = ATOUL(argv[1]) | 1L;

  if (strstr(argv[2], "."))
    *(float *)&setVal = atof(argv[2]);
  else
    setVal = ATOUL(argv[2]);

  printf("  write 0x%08lX to 0x%08lX\r\n", setVal, regNum);

  writeDSPReg(regNum, setVal, true);

  //NOTE:  Can't just read back the result to check if it worked because
  //       reading from some registers triggers events and processing
  //       within the DSP.

  Exit(0);
}


//-----------------------------------------------------------------------------
//  Function to read the data from one of the DSP's data buffers that contains
//  new waveform data for the Phase and Amplitude values surrounding a recent
//  change to the Phase and/or Amplitude setpoints.
//-----------------------------------------------------------------------------
int readWaveform(ulong bufID)
{
  static uint readCount;  //test
  ulong far  *buf;
  uint  w, lockID;
  U32  u32;
  float  fval;
  ulong  regNum, ackBits, val, count, countReg;


  ackBits = setDSPState;

  switch (bufID)  {
    case Rd_L_Buf:
      w = WAVE_AP_L_SRC;  ackBits |= CMD_L_DATA_READ;  countReg = 0;
      break;

    case Rd_T_Buf:
      w = WAVE_AP_T_SRC;  ackBits |= CMD_T_DATA_READ;  countReg = 0;
      break;

    case AmplFFT:
      w = WAVE_FFT_SRC;  ackBits |= CMD_FFT_DATA_READ;  countReg = SizeFFT;
      break;

    case PhaseCTF:
      if (hdwType == HT_PHASE_CTLR)  return -1;
      w = WAVE_CTF_P_SRC;  ackBits = 0;  countReg = SizeCTF;
      break;

    case AmplCTF:
      if (hdwType == HT_PHASE_CTLR)  return -1;
      w = WAVE_CTF_A_SRC;  ackBits = 0;  countReg = SizeCTF;
      break;

    default:
      return -1;
  }

  if (msSince_(lastWaveReadTime[w]) < 500)  return -1;

  if (msSince_(lastReadPlotTime) < 200)  return -1;

//  fprintf(DiagLog, "readWaveform(w=%d)\n", w);  //tdebug

  buf = waveformData[w];  lockID = WAVE_DATA_0 + w;

  regNum = bufID;


  if (getLock(lockID, 0) <= 0)  return -1;

  if (readWaves)  {

  //--- get the # of long words in the waveform ---
    if (countReg)  {
      if (hdwType == HT_PHASE_CTLR)
        readDSPReg(countReg, &count);
      else  {
        switch (bufID)  {
          case Rd_L_Buf:  case Rd_T_Buf:  count = 1024;  break;
          case AmplFFT:                   count = 1001;  break;
          case PhaseCTF:  case AmplCTF:   count = 1000;  break;
        }
      }
      if (count > MaxDataPoints)  count = MaxDataPoints;
      numDataPoints[w] = (uint) count;
    }

//--- read specified DSP buffer with waveform data ---
    readPlot(regNum, buf, numDataPoints[w]);  lastReadPlotTime = GetMS_();

    if (hdwType != HT_PHASE_CTLR)  {  // generate a simple waveform
      readCount += 5;  val = readCount;
      for (uint u=0; u<numDataPoints[w]; ++u)  {
        val = ++val & 0xff;

        switch (bufID)  {
          case Rd_L_Buf:  // 2 16-bit LSB 1st values for Ampl and Phase
          case Rd_T_Buf:  // ampl: raw ADC rdg, phase: int mult of 360/2^16
            u32 = (val << 16) | (~val & 0xff);  buf[u] = htonl(u32);  break;

          case AmplFFT:  // Amplitude FFT (32-bit IEEE float)
            fval = (float) val / 256.0;  buf[u] = htonl(*((ulong *)&fval));  break;

          case PhaseCTF:  // Phase CTF (expr as int mult of 360/2^32)
            u32 = ~(val << 24);  buf[u] = htonl(u32);  break;

          case AmplCTF:  // Ampl CTF (32-bit IEEE float) (?)
  //          u32 = val << 24;  buf[u] = htonl(u32);  break;
            fval = (float) val / 256.0;  buf[u] = htonl(*((ulong *)&fval));  break;
        }
      }
    }

  //--- let DSP know it is OK to write to the buffer again ---
    if (ackBits)  writeDSPRegQuick(SetDSPState, ackBits);

    _EnterCritical_();
    dataTimeSecs[w] = time(NULL);  dataTimeMS[w] = GetMS();
    _ExitCritical_();

    lastWaveReadTime[w] = GetMS_();
  }

  releaseLock(lockID);

  return 0;
}



//-----------------------------------------------------------------------------
void ForceDigOut(uchar bitNum, uchar val)
{
  ulong  mask;

  mask = 1L << bitNum;
  forcedDigOutMask |= mask;  forcedDigOutBits &= ~mask;
  if (val)  forcedDigOutBits |= mask;

  WriteDig();
}

//-----------------------------------------------------------------------------
void ReleaseDigOut(uchar bitNum)
{
  forcedDigOutMask &= ~(1L << bitNum);
  forcedDigOutBits &= forcedDigOutMask;

  WriteDig();
}

//=============================================================================
//  reset the DSP
//=============================================================================
void resetDSP(void)
{
  if (getLock(APPL_LOCK, 5000) <= 0)  return;

  dspProgNum &= 0x03;

  fprintf(DiagLog, "*** Reset DSP with %u errors ***\n", dspComErrors);
  fprintf(DiagLog, "Run program # %d\r\n", dspProgNum);

  prepForReset();

//--- assert program # to run after reset ---
  KLOCK();
  ClearBits(digitalOut, READY_LED);
  Next.digitalOut = digitalOut =
    (digitalOut & 0x3FFFFFL) | ((ulong)dspProgNum << 22);
  KUNLOCK();
  WriteDig();

//--- reset the DSP ---
  ForceDigOut(8, 0);  // force DO-08 (_RESET_DSP) low
  resetWDT();  sleepMS(500);  // hold it for 0.5 seconds
  ForceDigOut(8, 1);  // force DO-08 (_RESET_DSP) back high
  ReleaseDigOut(8);   // release the bit
  resetWDT();  sleepMS(500);

  KLOCK();  SetBits(Next.digitalOut, READY_LED);  KUNLOCK();

  _EnterCritical_();  dspComErrors = 0;  _ExitCritical_();

  releaseLock(APPL_LOCK);
}

//-----------------------------------------------------------------------------

//=============================================================================
//  Copy the "cur" state from last time to "prev", and update "cur"
//=============================================================================
//  Just before this function is called, readings were taken of the current
//  state of the hardware, and any changes to the output states specified by
//  the last pass were applied (just after taking the readings).
//
//  This func first copies "cur" to "prev" and then "next" to "cur" which does
//  2 things for us:
//
//    - We have a copy of the controller's logical state from the last pass
//      (so we can determine what has changed)
//
//    - We have a copy of values set by other tasks indicating events/requests
//      that happend/were made since the beginning of the last pass
//
//  We then reset the values (as needed) that are set by other tasks to
//  indicate events and requests, and update the current state based on the
//  readings just taken and the requests and events communicated by other
//  tasks.
//
//  All of this is in preperation for the updateNextState() function, which
//  implements the main "decision making" portion of the application logic,
//  determining what the next state should be based on the latest conditions
//  (as determined by the updateCurState() function).
//
//=============================================================================
void updateCurState(void)
{
  uint  x;
  ulong  maxWait1, maxWait2, elapsed;  //test
  PV_Desc  *pv;  //tdebug
  static time_t  lastChg;  //tdebug
  static uint  check1st = 0;



  if (current != applTask)  return;

//-----------------------------------------------------------------
// The following conditions exist at this point:
// - The inputs have just been read
// - The outputs were just set to values computed during the last pass
// - "cur" contains the state the inputs were in at the START of
//   of the previous pass, and the state the outputs were in DURING
//   the previous pass
// - "next" contains the values that the outputs were just set
//   to and any "input" values that were changed by other tasks
//------------------------------------------------------------------

// Now we update the values in "Next" that are based solely on the
// state of the hardware inputs we just read and values maintained by
// by other tasks (basically, we deal with READINGS here: Values that
// we need but are not written to "Next" by another task)

//  Next.processTime = GetMS();  curTime = time(NULL);

//  msSincePrev = msSince(Cur.processTime);

  Next.digitalIn = digitalIn;


// Note that a copy is done (rather than just switching pointer values) to
// insure all values are set in both the source and destination.  If we
// simply kept pointers and swapped the pointer values, then EVERY value
// would have to be updated independently by the code, which is a pain to
// program, adds overhead and code, and is a VERY easy way to introduce odd
// bugs.  Overall, copying everything here is actually faster overall,
// simpler, and much safer.

  Prev = Cur;  Cur = Next;

//--- At this point the Cur copy contains what the inputs were at the start ---
//--- of this pass, and the Prev copy contains the output values from the ---
//--- the previous pass (i.e. the output values that resulted in the input ---
//--- values in Cur) ---


//--------------------------------------------------------------------
// Next we update the rest of the "cur" values needed by the control
// logic (the ones that do NOT come from "next" values set by other
// tasks)
//
// NOTES:
//  - This portion of the logic has order-of-execution dependencies,
//    so be sure that all values referenced in an expression are ones
//    set before they are used.
//
//  - References to "prev" values are always safe since they are just
//    a copy of values from the last pass
//--------------------------------------------------------------------

  //--- log msgs for bit chgs only if not starving other processes for cycles ---
  if (applLoopTime < 8)  logDigIOChgs(false);


//APPLCODE
  if (!fpgaInitialized)  return;

  if (!autoReadDSP AND !autoWriteDSP)  return;

  if (getLockDSP(10) <= 0)  return;


//--- Amplitude and phase waveform data is read as a single stream of ---
//--- 32-bit values, where 16 of each 32 bits is an amplitude value and ---
//--- the other 16 is a phase value.  When operating in "pulsed" mode ---
//--- (where the Amplitude and/or Phase setpoints are being toggled ---
//--- between 2 values), the waveform data is updated based on when ---
//--- the change happens, and it provides data surrounding the "leading" ---
//--- edge and the "trailing" edge.  When not operating in pulsed mode, ---
//--- the exact timing of the updates to the buffers is not known. ---

//--- The FFT waveform is 32-bits per value and, like the Ampl/Phase data ---
//--- is updated by the DSP whenever we tell it we have finished reading ---
//--- the last copy in its buffer ---

//--- In each case, status bits from the DSP indicate when the buffers ---
//--- have new data, and we need to read the data then tell the DSP that ---
//--- we have done so and that it may write new data to the buffers ---

  if (hdwType == HT_PHASE_CTLR)  {  //test
    maxWait1 = 2000;  maxWait2 = 10000;  }
  else  {
    maxWait1 = 500;  maxWait2 = 2000;  }

  if (!calcStartTimeCTF)  {
  //--- Rotate the order we use to check for new waveform data and do not ---
  //--- read more than one each pass ---
    for (x=0; x<3; ++x)  {
      if (++check1st >= 3)  check1st = 0;

      if (check1st == 0)  if (NEW_L_DATA
            OR (msSince_(lastWaveReadTime[WAVE_AP_L_SRC]) > maxWait1))
        if (!readWaveform(Rd_L_Buf))  { lastDSPioTime = GetMS_();   break; }

      if (check1st == 1)  if (NEW_T_DATA
            OR (msSince_(lastWaveReadTime[WAVE_AP_T_SRC]) > maxWait1))
        if (!readWaveform(Rd_T_Buf))  { lastDSPioTime = GetMS_();   break; }

      if (check1st == 2)  if (FFT_DATA_VALID
            OR (msSince_(lastWaveReadTime[WAVE_FFT_SRC]) > maxWait2))
        if (!readWaveform(AmplFFT))   { lastDSPioTime = GetMS_();   break; }
    }
  }


//--- The CTF waveforms take around 40 secs to compute and interferes with ---
//--- the other waveforms (the DSP can't do them while computing the CTF) ---
//--- There is a separate Amplitude and Phase waveform, each one composed ---
//--- of 32-bits for each data point.  The waveforms are computed once ---
//--- each time we ask for them, and the DSP provides a status register ---
//--- to indicate how far along it is in calculating them ---
//    NOTE:  No longer implemented in the "real" hardware - but left here
//           for use in testing manually (re)armed waveforms

//--- check to see if we should start a new calc for the CTF waveform ---
  if ( (readerRdy[WAVE_CTF_P_SRC] OR readerRdy[WAVE_CTF_A_SRC])
      AND (calcStartTimeCTF == 0) )  {
    readerRdy[WAVE_CTF_P_SRC] = readerRdy[WAVE_CTF_A_SRC] = false;
  }

//--- for test purposes only ---
  if (hdwType != HT_PHASE_CTLR)
    if (!calcStartTimeCTF)
      percentCTF = 100;
    else  {
      elapsed = msSince(calcStartTimeCTF);
      percentCTF = (elapsed < 40000) ? (U16)(elapsed / 400) : 100;
    }

//--- Check to see if a pending CTF calc has completed ---
  if (calcStartTimeCTF)  do  {
    if (msSince(calcStartTimeCTF) > 60000)  {
      printf("CTF calc never finished\n");  calcStartTimeCTF = 0;  break; }
    if (msSince(calcStartTimeCTF) < 2000)  break;
    if (percentCTF != 100)  break;
    fprintf(DiagLog, "Finished CTF calculation\n");
    startCTFWaitTime = GetMS();  calcStartTimeCTF = 0;
  }  while(0);

//--- Wait 2 seconds after CTF calc finished before reading it ---
  if (startCTFWaitTime)  if (msSince(startCTFWaitTime) > 2000)  {
    if (msSince_(lastDSPioTime) > 20)  {
      readWaveform(PhaseCTF);  readWaveform(AmplCTF);
      startCTFWaitTime = 0;  lastDSPioTime = GetMS_();
    }
  }


/*
  Reading waveforms and doing the periodic I/O to/from the DSP are both
  lengthy processes, so we minimize latency on the rest of the system
  by not allowing them both to happen in the same pass.
*/
  if (msSince_(lastDSPioTime) > 20)  {

    ilkRatioGain = fwdAmplGain / reflAmplGain;
    ilkRatioOffset = 0;  //finish - what to set this to?

    maxOutAmplGain = cavAmplGain * feedFwdGain;
    maxOutAmplOffset = 0;  //finish - what to set this to?

    if (autoReadDSP)   if (!read_DSP_Regs())   lastDSPioTime = GetMS_();
    if (autoWriteDSP)  if (!write_DSP_Regs())  lastDSPioTime = GetMS_();

    if (dspLoad > 999)  dspLoad = 999;
    if (dspTemp > 99.9)  dspTemp = 99.9;

    if (percentFFT > 999)  percentFFT = 999;
    if (percentCTF > 999)  percentCTF = 999;

    dspTempInt = (uint) dspTemp;

    rfIsOn = BOOL(RF_ON);
    rfPulsing = BOOL(RF_PULSES);
    tunerPIDisON = BOOL(TUNER_PID_ON);
    curProgNum = (uint) CUR_PROG_NUM;
    rfAtWarm = BOOL(RF_OUT_WARM);
    rfRamping = BOOL(OUT_RAMPING);
    afcEnabled = BOOL(AFC_ON);
    cavLocked = BOOL(CAV_LOCKED);
    amplPIDisON = BOOL(AMPL_PID_ON);
    phaseMode = (uint) PHASE_MODE;

  }


  releaseLockDSP();


//tdebug - forced periodic change to pvStrTest for testing interface
  if (TIME() - lastChg >= 5)  {
    pv = PV_DESC(pvStrTest);
    farsprintf(pv->val, "..%lu..", TIME());
    pv->pvInfo.size = pv->pvInfo.elemSize = (uint) fstrlen(pv->val) + 1;
    pv->newEvents |= DBE_VALUE;
    lastChg = TIME();
  }


}


//=============================================================================
//  Now we update the values in "next" based on the "cur" and "prev"
//  values.
//=============================================================================
//  All values changed by updateNextState() are "outputs" from this func.
//  All other values referenced in this function are "inputs" to this func.
//
//  All the values in "next" provide information about the next logical state
//  for the controller:  Values to be applied during the next pass of the
//  control logic.  All "next" values fall in to one of 2 categories:
//
//    - They are set/updated ONLY by this function (which is executed only
//      by the "applic" task)
//
//    - They are set ONLY by other tasks (within a lock), and possibly
//      reset/cleared by the "applic" task in updateCurState() (after copying
//      "next" to "cur")
//
//
//  All values that act as input to this func must fall in to one of these
//  categories:
//
//    - "prev" values
//
//    - "cur" values updated directly by statements in updateCurState()
//
//    - "cur" values updated indirectly by updateCurState() due to the copy
//      of "next" to "cur", which includes the values in "next" that are
//      set by other tasks
//
//    - a shared/global value that is the equivalent of a "cur" value but
//      for which we don't need to track changes
//
//
//  All "outputs" from this function must be altered ONLY within this function.
//
//  All "inputs" to this function must NEVER be altered within this function.
//
//  Each output value (or VERY small "set" of ouput values) must be updated
//  in a single, self-contained block of code.  NO output may be referenced
//  OR altered in ANOTHER block of code:  This is a declarative style of
//  programming and has multiple advantages:
//
//     - There are no order-of-execution dependencies across nodes/blocks
//       within a pass.  They can be executed in any order and the result
//       will be the same (get used to it - this method will be the norm as
//       computer languages and hardware add support for atomic transactions
//       and computers evolve in to neural-net-like processor arrays - FPGAs
//       are just the beginning of this trend).
//
//     - Encapsulating all logic for an output within a single node/block
//       leaves just one place to start when tracking a problem backward from
//       the result (far simpler than guessing at the possible causes and
//       trying to work forward, looking for paths to the observed result).
//
//     - Drastically reduces the complexity of the code by greatly reducing
//       the number of possible paths through the code.  This simplifies
//       testing and verification processes.
//
//  Each node/block-of-code should be prefaced with the following information:
//   - The names of the values it modifies
//   - A descr of the meaning of the values and the logic used to set them
//   - A list of the values that serve as "inputs" to the node
//
//=============================================================================
void updateNextState(void)
{
  if (current != applTask)  return;

}


//=============================================================================
//  Next we update output signal values to be written to the outputs at
//  the beginning of the next pass (just after the inputs are read).
//=============================================================================
void updateOutputState(void)
{
//  digitalOut = 0;

  if (rfIsOn)
    SetBits(Next.digitalOut, ON_OFF_LED);
  else
    ClearBits(Next.digitalOut, ON_OFF_LED);

//APPLCODE

}


//=============================================================================
// The last thing we do is update values read by other tasks
//=============================================================================
// IMPORTANT NOTES and WARNINGS:
//
//   To avoid concurrency issues, other tasks need to use locks when reading
//   values set here.  But be careful to keep such locks short to avoid
//   delaying the timing of the main application task.  If necessary, split
//   such code in to sections, getting and releasing the lock(s) in each
//   section to avoid increasing latency for the main application task.
//
//   Values in the "next" instance of the application state fall in to 2
//   categories:  Inputs and Outputs.
//
//   The "next" Outputs are those that may be changed in updateOutputState().
//   The reset are Inputs, and may be changed by other tasks within a lock.
//
//   Since each pass through the application logic starts by setting
//   cur = next, any "next" values set by other tasks become part of the
//   "cur" instance used by the rest of the applic logic.  Because the
//   "applic" task at a higher priority than any other real-time task, the
//   changes to "next" values by other tasks can ONLY happen between passes
//   through the applic logic loop.
//
//   NOTES:
//     - Some of these "next" input values values must be initialized in
//       updateCurState() in order to tell when a value is set by another
//       task.  For example, if we have a boolean flag that is set when we
//       get an ON command, then (after its "next" value is copied to "cur"),
//       we have to clear the flag in order to tell if another ON command is
//       received between passes.
//
//     - As with reads to these values, other tasks must use locks when
//       writing to the "next" input values to insure they are not preempted
//       by the "applic" task in the middle of a write.
//
//=============================================================================
void updateExternals(void)
{
  if (current != applTask)  return;


  digInBits = Cur.digitalIn;
  digOutBits = Cur.digitalOut;

//--- set bits in older statusBitsAppl format ---
  statusBitsAppl = statusBitsDSP & 0x0000FFC8L;
  if (RF_PULSES)  statusBitsAppl |= (OLD_PHASE_PULSES | OLD_AMPL_PULSES);
  if (AFC_ON)  statusBitsAppl |= OLD_AFC_ON;
  statusBitsAppl |= (ulong)statusBitsAFC << 16;

//APPLCODE
}


//=============================================================================
_nearcall float readGetRaw(uint ui)  {
  return (ui < NumHdwADC ? anaIn[ui] : 0.0);  }

_nearcall float readMinRaw(uint ui)  {
  return (ui < NumHdwADC) ? (adcMinRaw(ui) - adcOffset[ui]) * adcGain[ui] : 0.0; }

_nearcall float readMaxRaw(uint ui)  {
  return (ui < NumHdwADC) ? (adcMaxRaw(ui) - adcOffset[ui]) * adcGain[ui] : 9.8; }

_nearcall void readDefaults(uint ui, float *gain, float *offset)  {
  *gain = 1.0;  *offset = 0.0; }

_nearcall void readCalibPoints(uint ui, float *low, float *high)  {
  *low = -9.5;  *high = 9.5; }

_nearcall char *readUnits(uint ui, char *buf)  { return "V"; }


//=============================================================================
_nearcall float setGetRaw(uint ui)  {
  return (ui < NumHdwADC ? anaOut[ui] : 0.0);  }

_nearcall float setMinRaw(uint ui)  {
  return (ui < NumHdwDAC) ? (dacMinRaw(ui) - dacOffset[ui]) * dacGain[ui] : 0.0; }

_nearcall float setMaxRaw(uint ui)  {
  return (ui < NumHdwDAC) ? (dacMaxRaw(ui) - dacOffset[ui]) * dacGain[ui] : 9.8; }

_nearcall void setDefaults(uint ui, float *gain, float *offset)  {
  *gain = 1.0;  *offset = 0.0;  }

_nearcall void setCalibPoints(uint ui, float *low, float *high)  {
  *low = -9.5;  *high = 9.5;  }

_nearcall char *setUnits(uint ui, char *buf)  { return "V"; }


//=============================================================================
//  Set default labels for all the I/O signals
//=============================================================================
void setDefaultLabels(void)
{
  printf("Setting default labels for I/O signals\r\n");

//--- set application-appropriate default labels ---
//APPLCODE

  fstrcpy(dinLabel[0],  "ID0");
  fstrcpy(dinLabel[1],  "ID1");
  fstrcpy(dinLabel[2],  "ID2");
  fstrcpy(dinLabel[3],  "ID3");

  fstrcpy(dinLabel[4],  "ID4");
  fstrcpy(dinLabel[5],  "ID5");
  fstrcpy(dinLabel[6],  "ID6");
  fstrcpy(dinLabel[7],  "ID7");

  fstrcpy(dinLabel[8],  "DSP_GP1");
  fstrcpy(dinLabel[9],  "DSP_GP2");
  fstrcpy(dinLabel[10],  "unused");
  fstrcpy(dinLabel[11],  "FPGA_DONE");

  fstrcpy(dinLabel[12],  "unused");
  fstrcpy(dinLabel[13],  "unused");


  fstrcpy(doutLabel[0], "DSP_GP5");
  fstrcpy(doutLabel[1], "DSP_GP6");
  fstrcpy(doutLabel[2], "DSP_GP7");
  fstrcpy(doutLabel[3], "DSP_GP8");

  fstrcpy(doutLabel[4], "_DSP_WR_RD");
  fstrcpy(doutLabel[5], "DSP_GP9");
  fstrcpy(doutLabel[6], "DSP_GP10");
  fstrcpy(doutLabel[7], "DSP_GP11");

  fstrcpy(doutLabel[8], "_RESET_DSP");
  fstrcpy(doutLabel[9], "ActivityLED");
  fstrcpy(doutLabel[10], "READY_LED");
  fstrcpy(doutLabel[11], "ON_OFF_LED");

  fstrcpy(doutLabel[12], "DSP_NMI");
  fstrcpy(doutLabel[13], "unused");
  fstrcpy(doutLabel[14], "FPGA_SPI");
  fstrcpy(doutLabel[15], "_ADC_CS");

  fstrcpy(doutLabel[16], "PLL_LE");
  fstrcpy(doutLabel[17], "_EEPROM_CS");
  fstrcpy(doutLabel[18], "--n/a--");
  fstrcpy(doutLabel[19], "FPGA_CL1");

  fstrcpy(doutLabel[20], "--n/a--");
  fstrcpy(doutLabel[21], "unused");
  fstrcpy(doutLabel[22], "DSP_GP3");
  fstrcpy(doutLabel[23], "DSP_GP4");

}


//=============================================================================
//  This function is called to force software configurable values to default
//  values.
//=============================================================================
void setDefaultConfig(void)
{
  printf("Setting in-RAM device config to default values\r\n");

//===== set default values for persistent data values =====
//APPLCODE

//--- default values for application-specific data elements ---
  rf_P = 0.9;   tuner_P = 0.9;
  rf_I = 0.1;   tuner_I = 0.1;
  rf_D = 0.01;  tuner_D = 0.01;

  pulseFreq = 10;  trigPos = 128;

  fwdAtten = reflAtten = cavAtten = 0;

  pulse_ramp = 0;  warmMS = rampMS = 0;

  pllInitIdx = 1;

  ilkRatioGain = maxOutAmplGain = feedFwdGain = maxAmplSet =
  fwdAmplGain = fwdPhaseGain = reflAmplGain = reflPhaseGain =
  cavAmplGain = cavPhaseGain = refAmplGain = refPhaseGain = 1.0;

  ilkRatioOffset = maxOutAmplOffset =
  fwdAmplOffset = fwdPhaseOffset = reflAmplOffset = reflPhaseOffset =
  cavAmplOffset = cavPhaseOffset = refAmplOffset = refPhaseOffset = 0.0;

  amplSet = pulsedAmplSet = pulsedAmplSet = pulsedPhaseSet = warmLevel =
  feedFwdPhase = 0.0;

  cavHighIlkLevel = fwdHighIlkLevel = 999.0;
}

//=============================================================================
//  This function should be called to prepare for a DSP or ZWorld reset
//=============================================================================
void prepForReset(void)
{
//finish  writeSetDSPState(CMD_RF_OFF);

// do something more MB_setDSPState(...)
}

//=============================================================================
//  This function is called by the application loop just before a shutdown.
//  The appliation must do whatever is necessary for a clean shutdown and
//  then exit
//=============================================================================
void applicShutdown(void)
{
  printf("Application prepping for shutdown\r\n");  fflush(stdout);

  prepForReset();

  Exit(0);
}

//=============================================================================
//  This is the highest priority task, usually resumed at a regular interval
//  so that it can do what it needs and sleep again until the start of the
//  next interval.
//=============================================================================
void _task applic(int argc, char **argv)
{
  #define SharedApplLocals
  #include <sharedAppl.c>

//--- add local variable definitions here ---
  int  clockRate;
//  fd_t  fpgaProgfd;
  LockData  *lock, *endLocks;

  #define SharedApplLocals2
  #include <sharedAppl.c>



  #define SharedApplInit
  #include <sharedAppl.c>

  _EnterCritical_();  SchedMS = 20;  _ExitCritical_();

//--- add application initialization code here ---

  autoReadDSP = autoWriteDSP = readWaves = true;  showFiller = false;

  rfq_t1_feedback_RegInfo.flags ^= LogChanges;
  rfq_t2_feedback_RegInfo.flags ^= LogChanges;
  devStat1_RegInfo.flags ^= LogChanges;

  rfq_t1_output_RegInfo.flags ^= LogChanges;
  rfq_t2_output_RegInfo.flags ^= LogChanges;
  devCtl1_RegInfo.flags ^= LogChanges;

  devRdSet1_RegInfo.flags ^= LogChanges;
  devRdSet2_RegInfo.flags ^= LogChanges;
  devRdCtl1_RegInfo.flags ^= LogChanges;


//--- initialize the locks ---
  endLocks = locks + NUM_LOCKS;
  for (lock=locks; lock<endLocks; ++lock)  {
    lock->count = 0;  lock->owner = 0;
    sem_init(&lock->sem, true, 1);
  }


  dspProgNum = 0;


  fpgaInitialized = pulseAmpl = pulsePhase = false;

//--- set defaults for # of data points in each waveform ---
  FBZERO(waveformData);  BZERO(readerRdy);
  BZERO(dataTimeSecs);  BZERO(dataTimeMS);
  for (ui=0; ui<NUM_WAVEFORMS*2; ++ui)  {
    numDataPoints[ui] = MaxDataPoints;
    lastWaveReadTime[ui] = GetMS_();
  }

//--- open serial ports for SPI connections ---
  clockRate = CLK_2MHz;

  spi1fd = open("/dev/spi1", O_WRONLY, 0660);  spi1Port = NULL;
  printf("spi1fd: %d\r\n", spi1fd);
  if (spi1fd < 0)
    printf("*** Unable to open '/dev/spi1' ***\r\n");
  else  {
    ioctl(spi1fd, IOC_SET_CLK_RATE, &clockRate);
    spi1Port = _FileFromFD_(spi1fd)->ioDev;
  }

  spi2fd = open("/dev/spi2", O_WRONLY, 0660);  spi2Port = NULL;
  printf("spi2fd: %d\r\n", spi2fd);
  if (spi2fd < 0)
    printf("*** Unable to open '/dev/spi2' ***\r\n");
  else  {
    ioctl(spi2fd, IOC_SET_CLK_RATE, &clockRate);
    spi2Port = _FileFromFD_(spi2fd)->ioDev;
  }




//--- initialize the ADC's and the PLL ---
  initDefaults();

//--- toggle FPGA SPI bus bit so it reloads from the EEPROM ---
  printf("reloading FPGA from EEPROM...\r\n");
  ClearBits(digitalOut, FPGA_SPI);  WriteDig();
  SetBits(digitalOut, FPGA_SPI);  WriteDig();
//--- wait for FPGA to load new firmware and release SPI bus ---
  fpgaInitialized = waitFPGA(3000);

  digitalOut |= _EEPROM_CS;


  #define SharedApplInit2
  #include <sharedAppl.c>


  lastDSPioTime = GetMS_();

  calcStartTimeCTF = startCTFWaitTime = 0;

//  resetDSP();

//===================================================
//===== start loop that runs once per timer msg =====
//===================================================

//--- The following loop is run at a periodic interval (normally 100 times ---
//--- per second, or every 10ms).  The application process normally ---
//--- consists of reading the current state of the digital and analog ---
//--- inputs, changing the state of the analog and digital outputs to the ---
//--- values computed during the previous pass, then computing new values ---
//--- for the outputs and updating values that outside clients can view ---
//--- (currently accessable as a set of ModBus registers) ---

  SetBits(Next.digitalOut, READY_LED);

  for (;;)  {
    #define SharedApplLoopStart
    #include <sharedAppl.c>

  //--- read the current state of the analog and digital Input signals, and ---
  //--- write the values computed at the end of the last pass ---
    digitalOut = Next.digitalOut;  rwIO();

  //--- limit DSP communication to aprox 20 times per second ---
    if (msSince_(lastDSPioTime) >= 50)  {

    //--- process next event message (if any) ---
      processMsg(eventsfd, (EventMsg *)buf, sizeof(buf));

    //--- update the current state and determine the next state ---
      updateCurState();  updateNextState();
      updateOutputState();  updateExternals();
    }

    #define SharedApplLoopEnd
    #include <sharedAppl.c>
  }

}


//=============================================================================

const CmdInfo far applCmds[] = {
  { "autocal",     autocal,     "calibrate fwd, refl, or cav Atten value" },
  { "cmd",         cmd,         "control DSP state" },
  { "config",      setConfig,   "set the output MUXOUT (values 0-7)" },
  { "defaults",    defaults,    "set FPGA params to default values" },
  { "dspcalib",    dspcalib,    "calibrate gain/offset for DSP/EGU conversion" },
  { "dspio",       dspio,       "enable/disable normal com with DSP" },
  { "dspread",     dspRead,     "read values from the DSP" },
  { "dspwrite",    dspWrite,    "write a value to the DSP" },
  { "progdsp",     progDSP,     "Load program in to DSP" },
  { "restartdsp",  restartDSP,  "Restart DSP and run specified program #" },
  { "read",        selectADC,   "picks ADC to send to DAC output (1-4)" },
  { "resetfpga",   resetFPGA,   "Reload the FPGA firmware from EEPROM" },
  { "setadcreg",   setADCReg,   "set value of an ADC register" },
  { "setfpgafirm", setFPGAfirm, "Update and reload FPGA firmware image" },
  { "setpllreg",   setPLLReg,   "Set value of a PLL register" },
  { "state",       showState,   "Show state of controller" }
};

int const NumApplCmds = ( sizeof(applCmds) / sizeof(CmdInfo) );

//-----------------------------------------------------------------------------
#endif

// progdsp 0 /srfs/RFController2.0.bin
// setfpgafirm /srfs/llrf2.0.bin
